package algo.alignment;

import java.util.*;

import lmu.utils.*;

import javax.xml.stream.*;

public class Alignment implements Comparable<Alignment>
{
    public Vector<Pair<Integer,Integer> > alignment = null;

    public double score;

    int l1;
    int l2;

    public Alignment(int l1, int l2)
    {
        setLengths(l1,l2);
        
    }
    public Alignment(int l1, int l2, Vector<Pair<Integer,Integer> > ali)
    {
        this(l1,l2);
        this.alignment = ali;
    }
    public void setLengths(int l1, int l2)
    {
        this.l1=l1; this.l2=l2;
    }

    public int getSeq1Length()
    {
        return l1;
    }

    public int getSeq2Length()
    {
        return l2;
    }
    public Vector<Pair<Integer,Integer>> initAlignment()
    {
        if(alignment!=null)
            return alignment;

        return alignment = new Vector<>();
    }

    public int compareTo(Alignment a)
    {
        return new Double(score).compareTo(new Double(a.score));


    }
    static java.util.regex.Pattern spaceP = java.util.regex.Pattern.compile(" ");

    public Pair<Alignment, String> cut_on_first(String s2)
    {
        int s21 = NumUtils.min(alignment, (p) -> p.getSecond());
        int s22 = NumUtils.max(alignment, (p) -> p.getSecond()) + 1;
        return Pair.create(cut(new Region1D(0, l1), new Region1D(s21, s22)), 
                           s2.substring(s21, s22));
    }

    public Pair<Alignment, UPair<String>> cut_local(String s1, String s2)
    {
        int s11 = NumUtils.min(alignment, (p) -> p.getFirst());
        int s12 = NumUtils.max(alignment, (p) -> p.getFirst()) + 1;

        int s21 = NumUtils.min(alignment, (p) -> p.getSecond());
        int s22 = NumUtils.max(alignment, (p) -> p.getSecond()) + 1;

        return Pair.create(cut(new Region1D(s11, s12), new Region1D(s21, s22)),
                           UPair.createU(
                               s1.substring(s11, s12),
                               s2.substring(s21, s22)
                               )
                           );

    }
    public Alignment cut(Region1D seq1_region, Region1D seq2_region)
    {
        final int s11 = seq1_region.getX1();
        final int s12 = seq1_region.getX2();
        final int s21 = seq2_region.getX1();
        final int s22 = seq2_region.getX2();

        Vector<Pair<Integer,Integer>> t = ObjectGetter.map(alignment, (p) -> (Pair<Integer,Integer>)Pair.create(p.getFirst() - s11, p.getSecond() - s21));
        return new Alignment(seq1_region.getLength(), seq2_region.getLength(), t);
    }

    public static Alignment parseAliFromXML(XMLStreamReader xsr) throws XMLStreamException
    {
        int l1 = XMLUtils.getIntAttrib(xsr, "l1", null);
        int l2 = XMLUtils.getIntAttrib(xsr, "l2", null);
        Alignment a = new Alignment(l1, l2);
        a.score = XMLUtils.getDoubleAttrib(xsr, "score", 0.0);
        String s1 = XMLUtils.getAttrib(xsr, "ali_1", null, false);
        if (s1==null)
            return a;
        
        a.alignment = new Vector<>();

        String s2 = XMLUtils.getAttrib(xsr, "ali_2", null, true);
        int[] i1 = StringUtils.convert_int(s1, spaceP);
        int[] i2 = StringUtils.convert_int(s2, spaceP);
        for (int i=0; i<i1.length; i++)
        {
            a.alignment.add(new Pair<Integer,Integer>(i1[i], i2[i]));
        }
        return a;
    }
    public static Alignment parseAliFromXML(org.jdom.Element e)
    {
        int l1 = XMLUtils.getIntAttrib(e, "l1", null);
        int l2 = XMLUtils.getIntAttrib(e, "l2", null);
        Alignment a = new Alignment(l1, l2);
        a.score = XMLUtils.getDoubleAttrib(e, "score", 0.0);
        String s1 = XMLUtils.getAttrib(e, "ali_1", null, false);
        if (s1==null)
            return a;
        
        a.alignment = new Vector<>();

        String s2 = XMLUtils.getAttrib(e, "ali_2", null, true);
        int[] i1 = StringUtils.convert_int(s1, spaceP);
        int[] i2 = StringUtils.convert_int(s2, spaceP);
        for (int i=0; i<i1.length; i++)
        {
            a.alignment.add(new Pair<Integer,Integer>(i1[i], i2[i]));
        }
        return a;
    }

    public org.jdom.Element getXML()
    {
        return getXML("ali");
    }
    public org.jdom.Element getXML(String rootelem)
    {
        org.jdom.Element ali = new org.jdom.Element(rootelem);
        ali.setAttribute("score", ""+score);
        ali.setAttribute("l1", ""+l1);
        ali.setAttribute("l2", ""+l2);

        if (alignment==null)
            return ali;
        
        StringBuffer sb1 = new StringBuffer();
        StringBuffer sb2 = new StringBuffer();
        for(Pair<Integer,Integer> p : alignment)
        {
            if (sb1.length()>0)
            {
                sb1.append(' ');
                sb2.append(' ');
            }
            sb1.append(""+p.getFirst());
            sb2.append(""+p.getSecond());
        }
        ali.setAttribute("ali_1", sb1.toString());
        ali.setAttribute("ali_2", sb2.toString());
        return ali;

    }
    public Alignment reverseAli()
    {
        Alignment a = new Alignment(l2, l1);
        a.score = score;
        if (alignment==null)
        {
            return a;
        }
        a.alignment = new Vector<>();

        for (Pair<Integer,Integer> p : alignment)
        {
            a.alignment.add(new Pair<Integer,Integer>(p.getSecond(), p.getFirst()));
        }
        return a;
    }

    public Vector<Region1D> toRegions(boolean onfirst, int maxgap) 
    {
        if (alignment==null || alignment.size()==0)
            return null;
        
        
        Vector<Integer> poslist = new Vector<Integer>();
        HashMap<Integer, Integer> m = ObjectGetter.buildMap(alignment, (p) -> (onfirst) ? p.getFirst() : p.getSecond(), (p) -> (onfirst) ? p.getSecond() : p.getFirst());
        for (Pair<Integer,Integer> p : alignment)
        {
            poslist.add((onfirst)?p.getFirst():p.getSecond());
        }
        if (poslist.get(0) > poslist.get(poslist.size()-1))
        {
            Collections.reverse(poslist);
        }

        Vector<Region1D>rv = new Vector<Region1D>();

        int laststart = poslist.get(0);
        int lastp = laststart;
        
        for (int i=1; i<poslist.size(); i++)
        {
            int p =  poslist.get(i);

            if(lastp<p-maxgap)
            {
                rv.add(new Region1D(laststart, lastp, new Region1D(m.get(laststart), m.get(lastp))));
                laststart = lastp = p;
                continue;
            }
            lastp = p;
        }
         
        int lp = poslist.get(poslist.size()-1);
         rv.add(new Region1D(laststart, lp, new Region1D(m.get(laststart), m.get(lp))));

         return rv;
    }


    public int getNumId(String s1, String s2)
    {
        if (alignment==null)
            return 0;

        int rv=0;
        final int s1l = s1.length();
        final int s2l = s2.length();
        for (Pair<Integer,Integer> p : alignment)
        {
            if (s1l<=p.getFirst() || s2l<=p.getSecond())
            {
                throw new RuntimeException("invalid sequences to ali...");
            }

            char ch1 = s1.charAt(p.getFirst());
            char ch2 = s2.charAt(p.getSecond());

            if (ch1!=ch2)
                continue;

            rv++;
        }
        return rv;
    }

    public static int getAlignedNumID(String s1, String s2)
    {
        final int l=Math.min(s1.length(), s2.length());


        int rv=0;

        char ch1=' ';
        char ch2=' ';
        for (int i=0; i<l; i++)
        {
            if ((ch1=s1.charAt(i))=='-' || (ch2=s2.charAt(i))=='-' || ch1!=ch2)
            {
                continue;
            }
            rv++;


        }
        return rv++;

    }
    public static String getIDString(String s1, String s2)
    {
        final int l=Math.min(s1.length(), s2.length());


        StringBuffer idstr = new StringBuffer();
        

        char ch1=' ';
        char ch2=' ';
        for (int i=0; i<l; i++)
        {
            if ((ch1=s1.charAt(i))=='-' || (ch2=s2.charAt(i))=='-' || ch1!=ch2)
            {
                idstr.append(' ');
                continue;
            }
            idstr.append('|');


        }
        return idstr.toString();
    }


    public String getFormatted(String id1, String seq1, String id2, String seq2)
    {
        Pair<String,String> ali = getAlignmentString(seq1, seq2);
        String idstr = getIDString(ali.getFirst(), ali.getSecond());

        int idlength = Math.max(id1.length(), id2.length());

        String fmt= String.format("%%%ds: %%s\n%%%ds  %%s\n%%%ds: %%s\n", idlength,idlength,idlength);
        return String.format(fmt, id1, ali.getFirst(), "", idstr, id2, ali.getSecond());
    }
    public String getFormattedBy(Integer format, String id1, String seq1, String id2, String seq2)
    {
        Vector<Pair<String,String>> v1 = new Vector<>(); v1.add(new Pair<String,String>(id1,seq1));
        Vector<Pair<String,String>> v2 = new Vector<>(); v2.add(new Pair<String,String>(id2,seq2));

        return getFormattedBy(format, v1,v2);
    }
    
    public String getFormattedBy(Integer format, Vector<Pair<String, String>> firstAnnots, Vector<Pair<String, String>> secondAnnots)
    {
        return getFormattedBy(format, firstAnnots, secondAnnots, null, null);
    }

    public String getFormattedBy(Integer format, Vector<Pair<String, String>> firstAnnots, Vector<Pair<String, String>> secondAnnots, String identityBase1, String identityBase2)
    {
        format--;
        
        //idstring is calculated between the last firstannots and first secondAnnots


        //String idstr = getIDString(ali.getFirst(), ali.getSecond());

        Vector<String> id1list = new Vector<String>();
        Vector<String> formatted1 = new Vector<String>();
        Vector<String> id2list = new Vector<String>();
        Vector<String> formatted2 = new Vector<String>();

        int idlength = 0;
        String alibase1 = null;
        for (Pair<String,String> p : firstAnnots)
        {
            String id = p.getFirst();
            idlength = Math.max(idlength, id.length());
            id1list.add(id);
            alibase1 = p.getSecond();
            formatted1.add(getAlignedString(p.getSecond(), true));
        }
        String alibase2 = null;
        for (Pair<String,String> p : secondAnnots)
        {
            String id = p.getFirst();
            idlength = Math.max(idlength, id.length());
            id2list.add(id);
            
            if (alibase2==null)
            {
                alibase2 = p.getSecond();
            }
            
            formatted2.add(getAlignedString(p.getSecond(), false));
        }

        if (identityBase1!=null)
        {
            alibase1 = identityBase1;
        }
        if (identityBase2!=null)
        {
            alibase2 = identityBase2;
        }

        if (alibase1 == null || alibase2 == null)
        {
            //no infos to draw the alignment
            return "";
        }

        String idstr = getIDString(getAlignedString(alibase1,true), getAlignedString(alibase2,false));

        final String rowFMT = String.format("%%%ds: %%s\n", idlength);
        
        Vector<String> rows = new Vector<String>();

        
        final int MAXL = idstr.length();
        for (int i=0;i<idstr.length(); i+=format)
        {
            StringBuilder sb_first = new StringBuilder();
            StringBuilder sb_id = new StringBuilder();
            StringBuilder sb_second = new StringBuilder();
            int lastpos = (i+format < MAXL) ? i+format : MAXL;

            for (int j=0; j<id1list.size(); j++)
            {
                rows.add(String.format(rowFMT, 
                                       id1list.get(j),
                                        formatted1.get(j).substring(i, lastpos)));
            }
            rows.add(String.format(rowFMT, " ", idstr.substring(i, lastpos)));

            for (int j=0; j<id2list.size(); j++)
            {
                rows.add(String.format(rowFMT, 
                                       id2list.get(j),
                                        formatted2.get(j).substring(i, lastpos)));
            }

            rows.add("\n");
        }  

        StringBuilder sb_final = new StringBuilder();

        for (String r : rows)
        {
            sb_final.append(r);
        }

        return sb_final.toString();
    }

    Vector<Integer> s1_poslist = null;
    Vector<Integer> s2_poslist = null;

    Vector<Integer> getPoslist(boolean first)
    {
        if (s1_poslist==null)
        {
            calcPoslist();
        }
        return (first)?s1_poslist:s2_poslist;
    }
    
    void calcPoslist()
    {
        int ali_length=l1+l2-alignment.size();

        Collections.sort(alignment);

        final int lasti=alignment.size()-1;
        
        StringBuffer ali = new StringBuffer();
                
        int s1_lastpos = 0;
        int s2_lastpos = 0;
        
        int ali_pos=0;

        s1_poslist = new Vector<>();
        s2_poslist = new Vector<>();

        
        for (Pair<Integer,Integer> p : alignment)
        {
            for (; s1_lastpos<p.getFirst(); s1_lastpos++)
            {
                s1_poslist.add(s1_lastpos);
                s2_poslist.add(null);
            }


            for (; s2_lastpos<p.getSecond(); s2_lastpos++)
            {
                s1_poslist.add(null);
                s2_poslist.add(s2_lastpos);
            }
            s1_poslist.add(p.getFirst());
            s2_poslist.add(p.getSecond());
            s1_lastpos=p.getFirst()+1;
            s2_lastpos=p.getSecond()+1;
        }

        for (; s1_lastpos<l1; ali_pos++, s1_lastpos++)
        {
            s1_poslist.add(s1_lastpos);
            s2_poslist.add(null);

        }


        for (; s2_lastpos<l2; ali_pos++, s2_lastpos++)
        {
            s1_poslist.add(null);
            s2_poslist.add(s2_lastpos);
        }
        
    }
    String getAlignedString(String seq, boolean first)
    {
        return getAlignedString(seq,first, '-');
    }

    String getAlignedString(String seq, boolean first, char gapChar)
    {
        int l = (first)?l1:l2;
        if (seq.length()!=l)
        {
            throw new RuntimeException("invalid seq length, expcected "+l+" got "+seq.length()+" first in alignment: "+first);
        }
        StringBuffer alibuff = new StringBuffer();
        for(Integer i : getPoslist(first))
        {
            if(i==null)
            {
                alibuff.append(gapChar);
                continue;
            }
            alibuff.append(seq.charAt(i));
        }

        return alibuff.toString();
    }
    public Pair<String,String> getAlignmentString(String seq1, String seq2)
    {
        return new Pair<String,String>(getAlignedString(seq1, true), getAlignedString(seq2,false));

    }
    public float getSeqId(String s1, String s2) {
        float numId = this.getNumId(s1, s2);
        return (float)numId/(0.5f*(s1.length()+s2.length()));
        
    }


}
