package isoformtransfer;

import genome.CodingExon;
import genome.GenomicUtils;
import genome.Protein;

import java.util.regex.*;

import lmu.utils.*;

import java.util.*;

import algo.alignment.Alignment;

import org.jdom.*;

import ortholog.PartialCodingExon;

public class InferredExon
{
    public static final String EXON_ELEM_NAME="exon";
    
    public String infoline;
    
    public Pair<Integer,Integer> source;
    public Pair<Integer,Integer> target;

    
    public String source_peptide;
    public String transferred_peptide;
    
    
    public double peptide_seqid;

    public int num_id_aa=0;
    public Alignment ali;
    public int peptide_frame = -1;

    public double dna_seqid;

        
    public HashSet<Integer> identity_codon_starts = new HashSet<Integer>();
    public Vector<Pair<Integer,Integer>> pos;


    public Vector<PartialCodingExon> input_exons = new Vector<PartialCodingExon>();

    public InferredExon(Pair<Integer,Integer> e)
    {
        target=e;
    }
        
    public InferredExon()
    {

    }
    
	public static RegionVector toRegionVector(List<InferredExon> exons){
		RegionVector ret = new RegionVector();
		for(InferredExon exon : exons){
			ret.addRegion(new Region1D<InferredExon>(exon.target,exon));
		}
		return ret;
	}
	
    
    public static ReflectedObject.XMLAttribHandler xml_attrib_handler = new ReflectedObject.XMLAttribHandler(InferredExon.class, true,
                                                                             "source_peptide", "transferred_peptide", "peptide_seqid", "num_id_aa", "peptide_frame", "dna_seqid");

    public InferredExon(Element e)
    {
        xml_attrib_handler.read(e,this);
        
        
        source=Pair.parseIntPair(XMLUtils.getAttrib(e, "source", null, true));
        target=Pair.parseIntPair(XMLUtils.getAttrib(e, "target", null, true));
        identity_codon_starts.addAll(StringUtils.convert_int_vector(XMLUtils.getAttrib(e, "identity_codon_starts", null, true), ","));

        for(Element che: XMLUtils.getChildren(e, PartialCodingExon.ELEM_NAME))
        {
            input_exons.add( new PartialCodingExon(che));
        }
    }
    public Element getXML()
    {
        Element exe = new Element("exon");

        xml_attrib_handler.write(exe,this);
        


        exe.setAttribute("source", ""+source);
        exe.setAttribute("target", ""+target);

        exe.setAttribute("identity_codon_starts", StringUtils.join(",", identity_codon_starts.iterator()));
        
        for (PartialCodingExon pce : input_exons)
        {
            exe.addContent(pce.getXML());
        }
        return exe;
    }                

    public int getSourceLength()
    {
        return source.getSecond()-source.getFirst();
    }

    public int getTargetLength()
    {
        return target.getSecond()-target.getFirst();

    }

    public String toString()
    {
        return target.toString();
    	//return source.toString();
    }
    
    /**
     * check which input exons contribute to the source
     * @param fromProtein
     * @param coding_exons
     */
    public static void addInputExons(Vector<InferredExon> ie,Protein fromProtein) {
        Vector<Pair<Integer,Integer> > fixed_regions = fromProtein.getCodingRegionVector(false).getRegionPairs();
        Vector<Region1D<CodingExon>> coding_exons = new Vector<Region1D<CodingExon>>();

        int tlength=0;

       
        for (int i=0; i<fixed_regions.size(); i++)
        {
            Pair<Integer,Integer> fr = fixed_regions.get(i);
            int frlength=fr.getSecond()-fr.getFirst();
            CodingExon src = null;
            for (int j=0; j<fromProtein.getNumExons(); j++)
            {
                CodingExon ce = fromProtein.getExon(j);

                int ov = Region1D.getOverlap(ce.start_in_gene, ce.end_in_gene+1, fr.getFirst(), fr.getSecond());
                if (ov==0)
                    continue;

                src=ce;
                break;
            }


            coding_exons.add(new Region1D<CodingExon>(tlength, tlength+frlength, src));
            tlength+=frlength;
        }
        
        for(InferredExon i : ie){
        	
        	int src_start=i.source.getFirst();
            int src_end=i.source.getSecond();
            
        	for (Region1D<CodingExon> r:  coding_exons)
	        {
	            if (r.getOverlap(src_start, src_end)<=0)
	                continue;
	
	
	            int ce_start=Math.max(r.getX1(), src_start);
	            int ce_end=Math.min(r.getX2(), src_end);
	            i.input_exons.add( new PartialCodingExon(r.getObject(), ce_start-r.getX1(), ce_end-r.getX1()) );
	        }
        }


    }
}

