package isoformtransfer;

import java.io.*;

import lmu.utils.*;
import genome.*;
import isoformtransfer.algo.EXALIN;
import isoformtransfer.algo.Exonerate;
import isoformtransfer.algo.Sim4cc;
import isoformtransfer.algo.Spaln2;
import isoformtransfer.algo.Splign;
import isoformtransfer.algo.TransferAlgo;
import protein.algo.ProteinAligner;

public class IsoformTransferSettings
{
    public GenomeSequenceExtractor fromdna;
    //public EnsemblDbHandler fromdb;
    public GenomeAnnotationHandler fromdb;
    public GenomeSequenceExtractor todna;
    //public GenomeAnnotationHandler todb;
    //public EnsemblDbHandler todb;
    public GenomeAnnotationHandler todb;
    public ProteinAligner paligner;
    public boolean save_transcript_seq=false;

    private Organism fromOrg;
    private Organism toOrg;
    
    public Organism getFromOrg(){
    	return fromOrg;
    }
    public Organism getToOrg(){
    	return toOrg;
    }
    
    
    public IsoformTransferSettings()
    {

    }

    public IsoformTransferSettings clone()
    {
        IsoformTransferSettings is = new IsoformTransferSettings();
        is.fromdb= fromdb;
        is.fromdna = fromdna;
        is.todb = todb;
        is.todna = todna;
        is.save_transcript_seq = save_transcript_seq;
        is.fromOrg = fromOrg;
        is.toOrg = toOrg;
        is.paligner = new ProteinAligner("freeshift", "dayhoff", -12, -1);
        return is;
    }
    public IsoformTransferSettings inverse()
    {

        IsoformTransferSettings rv = new IsoformTransferSettings();
        rv.fromdb=todb;
        rv.todb=fromdb;
        rv.todna=fromdna;
        rv.fromOrg = toOrg;
        rv.toOrg = fromOrg;
        rv.fromdna=todna;
        rv.paligner=paligner;
        rv.save_transcript_seq=save_transcript_seq;

        return rv;
    }
 
    /*
    public void setDbHandler(SimpleOptionParser cmdline, String dbkey, boolean isdst)
    {
        File f = new File(cmdline.getValue(dbkey));
        GenomeAnnotationHandler handler =null;
        //System.out.println("dbkey: "+dbkey+" value: "+f.getAbsolutePath());
        if (f.exists())
        {
            
            //gtf 
            handler= new genome.GTFAnnotation(f);

        }
        else
        {
            
            SQLConnectionOptionParser db = new SQLConnectionOptionParser();
            db.parseOption("db", cmdline.getValue(dbkey));
            
    
            if (cmdline.isOptionSet("host"))
            {
                db.parseOption("host", cmdline.getValue("host"));
            }
            if (cmdline.isOptionSet("dbuser"))
            {
                db.parseOption("user", cmdline.getValue("dbuser"));
                
            }
            if (cmdline.isOptionSet("dbpw"))
            {
                db.parseOption("pw", cmdline.getValue("dbpw"));
            }
    
               
            handler = new EnsemblDbHandler(db.getKeepAliveConnection());
        }

        if (isdst)
        {
            todb=handler;
        }
        else
        {
            fromdb=handler;
        }
    }
    */
    
    public void setDNAHandler(SimpleOptionParser cmdline, String key, boolean isdst)
    {
        setDNAHandler(cmdline.getFile(key), isdst);
    }
    public void setDNAHandler(File f, boolean isdst)
    {
       GenomeSequenceExtractor ex = new GenomeSequenceExtractor(f);
       if (isdst)
       {
           todna=ex;
       }
       else
       {
           fromdna=ex;
       }
    }

    public IsoformTransferSettings(Organism from, Organism to)
    {
        this(from,to, null);
    }
    public IsoformTransferSettings(Organism from, Organism to, ProteinAligner paligner)
    {
    	this.fromOrg = from;
    	
    	this.toOrg=to;
    	
        fromdna = from.getGenome();
        fromdb = from.getAnnot();
        
        
        todna = to.getGenome();
        todb = to.getAnnot();
        if (paligner==null)
        {
            paligner = new ProteinAligner("freeshift", "dayhoff", -12, -1);
        }
        this.paligner = paligner;
    }
    
    public IsoformTransferSettings(SimpleOptionParser cmdline)
    {
    	if ( cmdline.isOptionSet("fromdna")){
    		setDNAHandler(cmdline, "fromdna", false);
	        setDNAHandler(cmdline, "todna", true);
	        
	        throw new RuntimeException("DB access via from and to not supported anymore (use the GTF and fromdna and todna options)");
	        //setDbHandler(cmdline, "from", false);
	        //setDbHandler(cmdline, "to", true);
        
    	}else if ( cmdline.isOptionSet("organism")){ 
    		Organism.readOrganisms(cmdline.getFile("organism"));
    		
    		Organism from = Organism.getOrganism(cmdline.getInt("from_tax_id"));
    		Organism to = Organism.getOrganism(cmdline.getInt("to_tax_id"));

            todna = to.getGenome();
            todb = to.getAnnot();
    		
    		fromdna = from.getGenome();
            fromdb = from.getAnnot();
            
            fromOrg = from;
            toOrg = to;
           
    	}else{
    		throw new RuntimeException("Neither from/to dna settings are given, nor an organism XML file.");
    	}
        
        paligner = new ProteinAligner("freeshift", "dayhoff", -12, -1);
    }

    public static class TransferAlgoCommand {
        public TransferAlgo transferAlgo;
        public TransferAlgoCommand(SimpleOptionParser cmdline){
            String transferAlgoValue = cmdline.getValue("transfer_algo");
            if ( transferAlgoValue.equals("sim4cc")){
                transferAlgo = new Sim4cc(cmdline.getFile("transfer_algo_executable"));
            }else if ( transferAlgoValue.equals("exalin")){
                transferAlgo = new EXALIN(cmdline.getFile("transfer_algo_executable"));
            }else if ( transferAlgoValue.equals("spaln2")){
                transferAlgo = new Spaln2(cmdline.getFile("transfer_algo_executable"),false,!cmdline.isSet("optimal") );
            }else if ( transferAlgoValue.equals("spaln2Prot")){
                transferAlgo = new Spaln2(cmdline.getFile("transfer_algo_executable"), true, !cmdline.isSet("optimal"));
            }else if ( transferAlgoValue.equals("splign")){
                transferAlgo = new Splign(cmdline.getFile("transfer_algo_executable"));
            }else if ( transferAlgoValue.equals("exonerate")){
                transferAlgo = new Exonerate(cmdline.getFile("transfer_algo_executable"),false);
            }else if ( transferAlgoValue.equals("exonerateProt")){
                transferAlgo = new Exonerate(cmdline.getFile("transfer_algo_executable"),true);
            }/*else if ( transferAlgoValue.equals("ExonAlignment")){
             //   transferAlgo =new ExonGraphAligner();
            }*/else {
                throw new RuntimeException("Unknown transfer_algo parameter (" + transferAlgo + ")" );
            }
        }

        public static SimpleOptionParser prepareCmdLine()
        {
            return prepareCmdLine(new SimpleOptionParser());
        }
        public static SimpleOptionParser prepareCmdLine(SimpleOptionParser cmdline)
        {
            cmdline.addSimpleOptions("transfer_algo","transfer_algo_executable","optimal");
            cmdline.getOption("transfer_algo").setHelp("Transfer algorithm e.g. sim4cc,spaln2,exalin,splign,spaln2Prot");
            cmdline.getOption("transfer_algo_executable").setHelp("Some transfer_algo may use call an external executable");
            
            cmdline.getOption("optimal").setHelp("Use -Q0 for spaln2");
            
            cmdline.setSwitches("optimal");
            
            return cmdline;
        }
    }
    
    public static SimpleOptionParser prepareCmdLine(SimpleOptionParser cmdline)
    {
        if (cmdline == null)
        {
            cmdline = new SimpleOptionParser();
        }
        cmdline.addSimpleOptions("from", "to",  "fromdna", "todna", "host", "dbuser", "dbpw","organism","from_tax_id","to_tax_id");
        cmdline.setOptional( "from", "to",  "fromdna", "todna", "host", "dbuser", "dbpw","organism","from_tax_id","to_tax_id");
        cmdline.setDir("fromdna", "todna");
        cmdline.setFile("organism");
        cmdline.setInt("from_tax_id","to_tax_id");
        return cmdline;
    }
    public static SimpleOptionParser prepareCmdLine()
    {
        return prepareCmdLine(null);
    
    }

}
