package isoformtransfer.algo;

import isoformtransfer.InferredExon;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import lmu.utils.Pair;

public class EXALIN extends ExternalAlgo{

	static Pattern exalign_pat = Pattern.compile("EXON:\\s+(\\d+)\\s+-\\s+(\\d+)\\s+\\((\\d+)\\s+-\\s+(\\d+)\\)+\\s+[^\\s]++\\s+[^\\s]++\\s+[^\\s]++\\s+(\\d+.\\d+)%");
	private File exalin_path;
	
	public EXALIN(File exalin_path){
		this.exalin_path = exalin_path;
	}
	
	  @Override
	public List<InferredExon> parseExternalProgramOutputLine(String line) {
        Matcher m = exalign_pat.matcher(line);
      
        
        if (!m.matches()){
        	if ( line.startsWith("EXON")) throw new RuntimeException("Rex ex error:" + line);
        	return null;
        }
        
       
        InferredExon rv = new InferredExon();
        rv.infoline=line;
       
        rv.source=new Pair(Integer.valueOf(m.group(1)),Integer.valueOf(m.group(2))+1);//RegexUtils.getIntPair(m,1,2);
        rv.target=new Pair(Integer.valueOf(m.group(3)),Integer.valueOf(m.group(4))+1);

       //fixes strange bug in exalin 
        if ( rv.source.getFirst() == 1){ 
        	rv.source.setFirst(rv.source.first-1);
        	rv.target.setFirst(rv.target.first-1);
        	
        }
        
        rv.dna_seqid = Double.valueOf(m.group(5))/100;
       return Arrays.asList(rv);
       
     //   return rv;
	}

	@Override
	public List<String> getCommand(File cdna, File DNA) {
        ArrayList<String> command = new ArrayList<String>();
		
        command.add(exalin_path.getAbsolutePath());
        command.add(DNA.getAbsolutePath());
        command.add(cdna.getAbsolutePath());
        command.add("-p" );
        command.add(exalin_path.getParent() +"/HumanSPM.par" );
        command.add("-s");
        command.add("2");
 /*
        command.add("-a");
        command.add("1");
        command.add("-e");
   */     
        return command;
	}

}
