package isoformtransfer.algo;

import isoformtransfer.InferredExon;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import lmu.utils.Pair;

public class Exonerate extends ExternalAlgo{
	private File exonerate;
	public Exonerate(File exonerate, boolean useProtein){
		super.useProtein = useProtein;
		this.exonerate = exonerate;
		super.needFastaHeader = true;
	}
	private static Pattern p =Pattern.compile("\\w+\\s+[^\\s+]+\\s+similarity\\s+\\d+\\s\\d+\\s+\\d+\\s\\+\\s+\\.\\s+alignment_id\\s+0\\s+;\\s+Query\\s+\\w+\\s*((;\\s+Align\\s+\\d+\\s+\\d+\\s+\\d+\\s*)+)");
	@Override
	public List<InferredExon> parseExternalProgramOutputLine(String line) {
	
		Matcher m = p.matcher(line);
		if ( m.matches()){
			List<InferredExon> ret = new ArrayList<InferredExon>(); 
			for(String ali : m.group(1).substring(1).split(";")){
				String[] tokens = ali.trim().split("\\s+");
				 InferredExon rv = new InferredExon();
			     rv.infoline=ali;
			     Integer length = Integer.valueOf(tokens[3]);
			     Integer sourceStart = Integer.valueOf(tokens[2])-1;
			     Integer targetStart = Integer.valueOf(tokens[1])-1;
			     
			     if ( useProtein){
			    	 sourceStart*=3;
			     }
			     
			     rv.source=new Pair(sourceStart,sourceStart+length);//RegexUtils.getIntPair(m,1,2);
			     rv.target=new Pair(targetStart,targetStart+length);
			     
			     ret.add(rv);
			}
			return ret;
		
		}
		
		return null;
	}

	@Override
	public List<String> getCommand(File cdna, File DNA) {
		//./exonerate --model cdna2genome  ../../spaln2.1.2.linux64/bin/transript.faa ../../spaln2.1.2.linux64/bin/bovine.mfa --showalignment FALSE --showtargetgff TRUE -n 1
		
		// ./exonerate --model protein2genome ~/Downloads/spaln2.1.2.linux64/bin/protein.faa ~/Downloads/spaln2.1.2.linux64/bin/bovine.mfa --showalignment FALSE --showtargetgff TRUE --forwardcoordinates FALSE
		List<String> command = new ArrayList<String>();
		command.add(exonerate.getAbsolutePath());
		command.add("--model");
		if (! useProtein)
			command.add("est2genome");
		else
			command.add("protein2genome");
		command.add(cdna.getAbsolutePath());
		command.add(DNA.getAbsolutePath());
		command.add("--showalignment");
		command.add("FALSE");
		command.add("--showtargetgff");
		command.add("TRUE");
		command.add("-n");
		command.add("1");
		return command;
	}

}
