package isoformtransfer.algo;

import isoformtransfer.InferredExon;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import lmu.utils.RegexUtils;

public class Sim4cc extends ExternalAlgo {
	


    private static Pattern _s4cc_pat = Pattern.compile("(\\d+)-(\\d+)\\s+\\((\\d+)-(\\d+)\\)\\s+(\\d+)%(.*)");
    
	private File sim4cc_path;

	
	public Sim4cc(File sim4cc_path){
		this.sim4cc_path = sim4cc_path;
	}
   
	@Override
	public List<InferredExon> parseExternalProgramOutputLine(String line) {
	       Matcher m = _s4cc_pat.matcher(line);
	 
	        if (!m.matches())
	            return null;
	    //    System.out.println(line);
	        InferredExon rv = new InferredExon();
	        rv.infoline=line;
	        //sim4cc returns 1-based coordinates 
	        //and end non-inclusive
	        rv.source=RegexUtils.getIntPair(m,1,2);
	        rv.source.setFirst(rv.source.getFirst()-1);
	        rv.target=RegexUtils.getIntPair(m,3,4);
	        rv.target.setFirst(rv.target.getFirst()-1);
	        rv.dna_seqid = 0.01*RegexUtils.getDouble(m,5);
	        return Arrays.asList(rv);
	}
	@Override
	public List<String> getCommand(File cdna, File DNA) {
		ArrayList<String> command = new ArrayList<String>();

		command.add(sim4cc_path.getAbsolutePath());
	//	command.add("-R0");
		command.add(cdna.getAbsolutePath());
		command.add(DNA.getAbsolutePath());
		return command;
	}


}
