package isoformtransfer.algo;

import isoformtransfer.InferredExon;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;

import lmu.utils.Pair;

public class Spaln2 extends ExternalAlgo {

    private boolean use_cigar=false;
	private File splan2;
	private boolean fast = true;
	
	public Spaln2(File splan2){
		this.splan2 = splan2;
		super.useProtein = false;
		this.fast = true;
	}
	
	
	public Spaln2(File splan2, boolean useProtein, boolean fast){
		this.splan2 = splan2;
		super.useProtein = useProtein;
		this.fast = fast;
	}
	public void getCigarOutput(File cdna, File DNA){
		List<String> command = this.getCommand(cdna, DNA);
		command.add(0, "-O8");
	}
	
	@Override
	public List<InferredExon> parseExternalProgramOutputLine(String line) {
		
		InferredExon rv = new InferredExon();
		if ( use_cigar){
			//cigar: Query 0 117 + Target 237 346 + -280 M 14 I 5 M 43 I 2 M 26 I 1 M 26
			
			if ( line.startsWith("#") || line.startsWith("@")) return null;
			//System.out.println(line);
			String[] tokens = line.split("\\s+");
		 	rv.source=new Pair<Integer,Integer>(Integer.valueOf(tokens[2].trim()),Integer.valueOf(tokens[3].trim()));
		    rv.target=new Pair<Integer,Integer>(Integer.valueOf(tokens[6].trim()),Integer.valueOf(tokens[7].trim()));
		    Vector<Pair<Integer,Integer>> pos = new  Vector<Pair<Integer,Integer>>();
		    int currentSource = rv.source.first;
		    int currentTarget = rv.target.first;
		    for(int i = 10; i< tokens.length;i+=2){
		    	String type = tokens[i].trim();
		    	Integer size = Integer.valueOf(tokens[i+1]);
		    	if ( type.equals("M")){
		    		for(int j = 0 ; j< size;j++){
		    			pos.add(new Pair(currentSource++,currentTarget++));
		    		}
		    	}else if (type.equals("I") || type.equals("N")){
		    		for(int j = 0 ; j< size;j++){
		    			currentSource++;
		    		}
		    	}else if (type.equals("D")){
		    		for(int j = 0 ; j< size;j++){
		    			currentTarget++;
		    		}
		    	}else{
		    		throw new RuntimeException("Unknown type:" + type);
		    	}
		    }
		    rv.pos =pos;
		}else{
		//# rID	  gID	   %id	  ExonL	 MisMch	 Unpair	 ref_l	  ref_r	  tgt_l	  tgt_r	 eScore	 IntrnL	 iScore	 Sig3/I	 Sig5/T  # -  X P DiNuc
		//e.g.: ENSP00000346389	Gene	 100.00	     54	      0	      0	      1	     54	 155956	 156009	  116.9	      0	    0.0	   0.00	   8.90  0 0  0 0   .
			if ( line.startsWith("#") || line.startsWith("@")) return null;
			String[] tokens = line.split("\t");
			if ( tokens.length < 9) return null;
			
	        rv.infoline=line;
	        if ( useProtein){
	        	 rv.source=new Pair<Integer,Integer>(Integer.valueOf(tokens[6].trim()),Integer.valueOf(tokens[7].trim()));
	        	 rv.source.first = (rv.source.first-1)*3;
	        
	        	 rv.source.second = rv.source.second*3;
	        	 rv.target=new Pair<Integer,Integer>(Integer.valueOf(tokens[8].trim()),Integer.valueOf(tokens[9].trim()));
	        }else{
	        	 rv.source=new Pair<Integer,Integer>(Integer.valueOf(tokens[6].trim()),Integer.valueOf(tokens[7].trim()));
	        	 
	        }
	        
	        if ( Integer.valueOf(tokens[8].trim()) > 0){ 
	        	 rv.target=new Pair<Integer,Integer>(Integer.valueOf(tokens[8].trim())-1,Integer.valueOf(tokens[9].trim()));
	        }else{
	        	rv.target=new Pair<Integer,Integer>(Integer.valueOf(tokens[8].trim()),Integer.valueOf(tokens[9].trim()));
	        }
	    
	        rv.dna_seqid = Double.valueOf(tokens[2])/100f;
			
		}
		return Arrays.asList(rv);
		
	}
	@Override
	public void setEnviroment(ProcessBuilder builder) {
		File tabDir = new File(splan2.getParentFile()+ "/../table");
		
		builder.environment().put("ALN_TAB",tabDir.getAbsolutePath() );
	}
	public List<String> getCommand(File cdna, File DNA) {
		ArrayList<String> command = new ArrayList<String>();
		command.add(splan2.getAbsolutePath());
		if (use_cigar)command.add("-O8"); //cigar format
		//command.add("-pw"); //dont filter bad transfers
		command.add("-yX"); //cross-species parameter
		command.add("-S1"); //match only forward strand
		if ( !fast) command.add("-Q0"); //results in better predictions, but ~10x slower ..
		command.add(tmp_gene_dna.getAbsolutePath());
        command.add(cdna.getAbsolutePath());
		return command;
	}
	
	/*
	@Override
	public List<String> getCommand(File cdna, File DNA) {
		ArrayList<String> command = new ArrayList<String>();
		command.add(splan2.getAbsolutePath());
		command.add("-pw"); //dont filter bad transfers
		command.add("-yX"); //cross-species parameter
		command.add("-S1"); //match only forward strand
		if ( !fast) command.add("-Q0"); //results in better predictions, but ~10x slower ..
		command.add(tmp_gene_dna.getAbsolutePath());
        command.add(cdna.getAbsolutePath());
		return command;
	}
	*/
	
}
