package isoformtransfer.isar;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.awt.TexturePaint;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Vector;

import genome.Gene;
import genome.Organism;
import genome.Protein;
import isoformtransfer.isar.Formater.FASTATransferformation;
import isoformtransfer.isar.PartialOrderMultipleAlignment.Region;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.MapBuilder;
import lmu.utils.ObjectGetter;
import lmu.utils.Pair;
import lmu.utils.Region1D;
import lmu.utils.RegionVector;
import lmu.utils.UPair;
import lmu.utils.plotting.AbstractZoomedMultiPicPlane;
import lmu.utils.plotting.DGraphics;
import lmu.utils.plotting.MultiPicPlaneFrame;
import lmu.utils.plotting.RegionPicturePane;
import lmu.utils.plotting.RegionPlot;
import lmu.utils.plotting.RegionPlot.RegionDecorator;
import lmu.utils.plotting.ZoomedCoordinateConverter;
import lmu.utils.tuple.Tuple4;

public class Drawer {
    
    public static boolean grid = true;
    public static double SCALE = 1f;
    
    private static List<Region1D> getBetween(RegionVector geneExons,int from, int to)
    {
        List<Region1D> ret = new ArrayList<Region1D>();

        if ( from > to)
        {
            throw new RuntimeException(from + " " + to);
        }
        for (Region1D r : geneExons.getRegions())
        {

            if ( from >= r.getX2() ) continue;
            if ( r.getX1() > to) break;
            int x1 = r.getX1();
            if ( r.contains(from))
            {
                x1 = from;
            }
            if ( r.contains(to))
            {
                int x2 = Math.min(r.getX2(),to );
                ret.add(new Region1D(x1,to,null));
                return ret;
            }
            ret.add(new Region1D(x1,r.getX2(),null));

        }

        return ret;
    }
    private static Region1D getExon(RegionVector geneExons,Region1D subRegion)
    {
        for (Region1D r : geneExons.getRegions())
        {
            if ( r.getOverlap(subRegion) >0 )return r;
        }
        throw new RuntimeException("Should not happen");
    }
    static public class DecoratorExonsConnectes extends RegionDecorator
    {
        private Logger log = LogConfig.getLogger();

        private ZoomedCoordinateConverter rc;
        RegionVector regions;
        boolean debug;
        boolean showGaps;
        POMAGene pg;
        
        
        public DecoratorExonsConnectes(POMAGene pg, Color color, ZoomedCoordinateConverter rc, RegionVector regions, boolean debug, boolean showGaps)
        {
            super(color,true);
            this.debug = debug;
            this.rc = rc;
            this.regions = regions;
            this.pg = pg;
            this.showGaps = showGaps;
        }

        public Rectangle drawRegions(DGraphics g2, int y_shift, Vector<Rectangle> multi_regions)
        {
            if ( multi_regions.size() == 0) {
                
                log.warn("No exon information skip:" + label);
                return new Rectangle(0,0,0,0);
            }
            final int N=multi_regions.size()-1;
            int x=(int)multi_regions.get(0).getX();
            int y=y_shift+(int)multi_regions.get(0).getY();
            int height=(int)multi_regions.get(0).getHeight();

            int x2=(int)multi_regions.get(N).getX()+(int)multi_regions.get(N).width;

            int width=x2-x;



            for (Rectangle box : multi_regions)
            {
                draw(g2, y_shift, box);
            }
            //connect the boxes
            int middle_h=(int)(y+(0.5*height));
            int w= 0;
            
            
            for (int i=1; i<multi_regions.size(); i++)
            {
                Rectangle prev = multi_regions.get(i-1);
                Rectangle next = multi_regions.get(i);
                
                Region1D preExon = (Region1D)regions.get(i-1).getObject();
                Region1D exon = (Region1D)regions.get(i).getObject();
                
                if ( preExon == exon || preExon.getX2() == exon.getX1()) {
                    int start=(int)(prev.getX()+prev.getWidth());
                    int end = (int)next.getX();
                   
                    //connection line
                    g2.fillRect(start, middle_h-1, (end-start)+1, (int)Math.round(2*SCALE));
                    
                }

            }
            //color newly inferred exons
            for (int i=0; i<multi_regions.size(); i++)
            {
                Color before = g2.getColor();
                
                List<Region1D> overlap = pg.origGeneExons.getIntersectingRegions((Region1D)regions.get(i).getObject());
                
                if (overlap == null || overlap.size() == 0)
                {
                    g2.setColor(Color.RED);
                    
                    g2.fillRect((int)multi_regions.get(i).getX(),y_shift+(int)multi_regions.get(i).getY(),(int)multi_regions.get(i).getWidth(),(int)multi_regions.get(i).getHeight());
                    g2.setColor(before);
                    
                }
            }
            
            //color gapped regions
            if ( showGaps)
            {
                for (int i=0; i<multi_regions.size(); i++)
                {
                    
                    if ( regions.get(i).getObject() instanceof GappedRegion)
                    {
                        Color before = g2.getColor();
                        BufferedImage bi = new BufferedImage(5, 5,   BufferedImage.TYPE_INT_RGB); 
                         Graphics2D big = bi.createGraphics(); 
                         // Render into the BufferedImage graphics to create the texture 
                         big.setColor(Color.lightGray); 
                         
                         big.fillRect(0,0,5,5); 
                         /*
                         big.setColor(Color.lightGray); 
                         big.fillOval(0,0,5,5); 
                         */
                         
                         big.setColor(new Color(color.getRed(),color.getGreen(),color.getBlue(),120)); 
                         //big.setStroke(new BasicStroke(4));
                         
                         big.drawLine(0, 0, 5, 5);
                         Rectangle r = new Rectangle(0,0,5,5); 
                         TexturePaint tp = new TexturePaint(bi,r); 

                         g2.setPaint(tp);
                        
                        g2.fillRect((int)multi_regions.get(i).getX(),y_shift+(int)multi_regions.get(i).getY(),(int)multi_regions.get(i).getWidth(),(int)multi_regions.get(i).getHeight());
                        g2.setColor(before);
                        
                    }
                }
            }
            
            
            if ( debug)
            {
                for (int i=1; i<multi_regions.size(); i++)
                {
                    Rectangle prev = multi_regions.get(i-1);
                    Rectangle next = multi_regions.get(i);

                    int start=(int)(prev.getX()+prev.getWidth());
                    // int end = (int)next.getX();

                    if ( regions.get(i-1).getObject() == regions.get(i).getObject()) continue;

                    if ( GappedRegion.class.isInstance(regions.get(i-1).getObject()) || GappedRegion.class.isInstance(regions.get(i).getObject()))
                        continue;
                    
                    Region1D preExon = (Region1D)regions.get(i-1).getObject();
                    Region1D exon = (Region1D)regions.get(i).getObject();

                    
                    
                    int intronLength = exon.getX1()-preExon.getX2();

                    if ( true)
                    {

                        g2.setFont("monospaced", 8);

                        g2.setColor(Color.BLUE);

                        g2.drawLine(start, y, start, y-height);
                        g2.drawString(intronLength + "bp", start, (w%2==0)?y-2:(y-height));
                    }
                    w++;
                }   
            }
         //   int highlightHeight = Math.round(height+height*0.5f);


            return new Rectangle(x,y,width,height);
        }

    };
    
    static public class Decorator extends RegionDecorator
    {

        // public Color[] color = new Color[]{new Color(0.6f,0.6f,0.6f,0.6f),new Color(0.7f,0.7f,0.7f,0.5f)};//new Color[]{ new Color(1,0,0,0.5f),new Color(0,1,0,0.5f)};
//        public Color[] color = new Color[]{ new Color(1,0,0,0.5f),new Color(0,1,0,0.5f)};

        public Color[] color = new Color[]{ Color.BLACK, Color.GRAY};

        public String label="";
        RegionVector regions;
        boolean debug;
        public Decorator( RegionVector regions, boolean debug)
        {
            this.debug = debug;
            this.regions = regions;
        }

        public Rectangle drawRegions(DGraphics g2, int y_shift, Vector<Rectangle> multi_regions)
        {
            final int N=multi_regions.size()-1;
            int x=(int)multi_regions.get(0).getX();
            int y=y_shift+(int)multi_regions.get(0).getY();
            int height=(int)multi_regions.get(0).getHeight();

            int x2=(int)multi_regions.get(N).getX()+(int)multi_regions.get(N).width;

            int width=x2-x;



            //   g2.drawRect(x, y, width, height);

            for (Rectangle box : multi_regions)
            {
                draw(g2, y_shift, box);
            }
            //connect the boxes
            int middle_h=(int)(y+(0.5*height));
            g2.setColor(Color.BLACK);

            int w= 0;

            for (int i=1; i<multi_regions.size(); i++)
            {
                Rectangle prev = multi_regions.get(i-1);
                Rectangle next = multi_regions.get(i);

                int start=(int)(prev.getX()+prev.getWidth());
                int end = (int)next.getX();
                if ( (end-start)> 1)  g2.drawLine(start, middle_h, end, middle_h);
            }

            if ( debug)
            {
                for (int i=1; i<multi_regions.size(); i++)
                {
                    Rectangle prev = multi_regions.get(i-1);
                    Rectangle next = multi_regions.get(i);

                    int start=(int)(prev.getX()+prev.getWidth());
                    // int end = (int)next.getX();

                    if ( regions.get(i-1).getObject() == regions.get(i).getObject()) continue;

                    Region1D preExon = (Region1D)regions.get(i-1).getObject();
                    Region1D exon = (Region1D)regions.get(i).getObject();


                    int intronLength = exon.getX1()-preExon.getX2();

                    if ( true)
                    {

                        g2.setFont("monospaced", 8);

                        g2.setColor(Color.BLACK);

                        g2.drawLine(start, y, start, y-height);
                        g2.drawString(intronLength + " bp", start, (w%2==0)?y-2:(y-height));
                    }
                    w++;
                }   
            }
            //int highlightHeight = Math.round(height+height*0.5f);


            return new Rectangle(x,y,width,height);
        }
        int k = 0;
        int colI =0;

        Integer start=null;

        public Rectangle draw(DGraphics g, int y_shift, Rectangle box)
        {
            int x=(int)box.getX();
            int y=y_shift+(int)box.getY();
            int height=(int)box.getHeight();
            int width=(int)box.getWidth();

            if ( start == null && regions.get(k).getLength() > 2) start = x;

            // System.out.println(start + "\t" + k + "\t" + box);
            if (k >0 && regions.get(k).getObject() != regions.get(k-1).getObject())
            {
                colI++;
            }
            g.setColor(color[colI%2]);
            if (filled)
            {

                
                
                g.fillRect(x,y,width,height);
                g.setColor(Color.BLACK);
               // g.drawRect(x,y,width,height);

            }
            else
            {
                g.drawRect(x,y,width,height);
            }

            // System.out.println(k + " " + regions.size());
            /*
            if ( (k < regions.size()-1 && regions.get(k).getObject() != regions.get(k+1).getObject()) || k == regions.size()-1){
                g.setColor(Color.BLACK);
                if ( start==null)start=x;
                g.drawRect(start, y, (x-start)+width, height);
                start=null;
            }
            */
            k++;
            return box;
        }
    };


    public static BufferedImage drawGene(Gene gene, boolean small ) {
        
        ZoomedCoordinateConverter zcc = new ZoomedCoordinateConverter(gene.length,null,5); 
        Object obj = new Object();
        for(Protein p : gene.proteins.values())
        {
            for(Region1D exon : p.getCodingRegionVector(false).getRegions())
            {
                zcc.addRegion(exon, obj);
            } 
        }
        zcc.addCategory(obj, small?0.2:1.0, null);
        zcc.update();
        RegionPicturePane rpp = new RegionPicturePane(gene.getName() + " (" + gene.getOrganism().getShortName() + ")", zcc.getZoomedLength() , null);

        //rpp.changeRegionSettings(16, 40, 20);  

       // String label = small?gene.getOrganism().getShortName():gene.getOrganism().getShortName() + "(" + gene.id + ")";
        //RegionPlot rpGene1 = rpp.addRegion(label,null,zcc.getConvertedRegVec(gene.getMergedTranscript(false)));
        //rpGene1.setDefaultDecorator(new RegionDecorator(Color.BLACK,true));

        for(Protein p : gene.proteins.values())
        {
            RegionPlot rp = rpp.addRegion(small?"":p.id,null,zcc.getConvertedRegVec(p.getCodingRegionVector(false)));
            rp.setDefaultDecorator(new RegionDecorator(Color.GRAY,true,false));
        }
        rpp.update();
        MultiPicPlaneFrame frame = new MultiPicPlaneFrame();

        frame.addInfo(rpp);

        frame.update();
        frame.setAutoSize(500, 800);
        return frame.getCombinedImage();
    }
    public static BufferedImage drawAlignmentRegions(PartialOrderMultipleAlignment poma, boolean showIsoformLabels, boolean showGeneId)
    {
        return drawAlignmentRegions(poma,showIsoformLabels,showGeneId,poma.getGenes());
    }
    public static BufferedImage drawAlignmentRegions(PartialOrderMultipleAlignment poma, boolean showIsoformLabels, boolean showGeneId,List<POMAGene> genes)
    {
        String name =null;
        for (POMAGene ag : genes)
        {
            if ( ag.getName() != null)
            {
                name = ag.getName();
                break;
            }
        }

        HashMap<POMAGene,ZoomedCoordinateConverter> gene2zcc = new  HashMap<POMAGene,ZoomedCoordinateConverter> ();

        if ( name == null) name ="";

        int maxLength = 0;
        for (POMAGene gene : genes)
        {
            //ZoomedCoordinateConverter zcc = new ZoomedCoordinateConverter(gene.geneSequence.length()+1,SCALE*0.25,null);
            
            ZoomedCoordinateConverter zcc = new ZoomedCoordinateConverter(gene.geneSequence.length()+1,null,(int)Math.round(SCALE*10)); 
            Object obj = new Object();
            gene2zcc.put(gene,zcc );

            for(Region1D exon : gene.geneExons.getRegions())
            {
                zcc.addRegion(exon, obj);
            }
            
            for (String proteinId : gene.getProteins())
            {
                for (Region1D exon : gene.getRegionVector(proteinId).getRegions())
                {
                    zcc.addRegion(exon, obj);
                }
            }
            zcc.update();
            maxLength = Math.max(maxLength,zcc.getZoomedLength());
        }
        RegionPicturePane rpp = new RegionPicturePane(name, maxLength , null);
        
        rpp.changeRegionSettings((int)Math.round(SCALE*10), (int)Math.round(SCALE*20), (int)Math.round(SCALE*5));
        
        rpp.update();
        
        
        
        HashMap<POMAGene,RegionPlot> obj2Rp = new HashMap<POMAGene,RegionPlot> ();
        
        
        
        for (POMAGene gene : genes)
        {
            ZoomedCoordinateConverter zcc = gene2zcc.get(gene);
            
            RegionVector rvg = new RegionVector();

            for (Region1D r : gene.geneExons.getRegions())
            {
                r.setObject(r);
                rvg.addRegion(r);
            }
            
            //String spName = gene.getOrganism()!=null?gene.getOrganism().getShortName():gene.getTaxId()+"";
            //String label = small?spName + " - " + gene.getName().toUpperCase():spName + "(" + gene.getId() + ")";
            
            String label = null;
            String pgName = gene.getName();
            if ( pgName == null) pgName ="";
            pgName = pgName.toUpperCase();
            
            String orgName = ISAR.OrganismNameReader.getShortName(gene.getTaxId());
            if ( orgName != null)
            {
                label = (showGeneId?gene.getId():orgName) + " (" + pgName + ")" ;
            }else{
                label = (showGeneId?gene.getId():"") + " (" + pgName + ")" ;
            }
            
            RegionPlot rpGene1 = rpp.addRegion(label,null,zcc.getConvertedRegVec(rvg));
            
            rpGene1.setDefaultDecorator(new Drawer.DecoratorExonsConnectes(gene,Color.BLACK,zcc,rvg,false,false));
            
            obj2Rp.put(gene, rpGene1);
     
            if ( gene.getProteins().size() > 1){
                for (String proteinId : gene.getProteins())
                {
                    RegionVector rv = new RegionVector();
                    for (Region1D r : gene.getRegionVector(proteinId).getRegions())
                    {
                        r.setObject(r);
                        rv.addRegion(r);
                    }
                   
                    RegionPlot rp = rpp.addRegion(showIsoformLabels?proteinId:"", null, zcc.getConvertedRegVec(rv));
                    
                    rp.setDefaultDecorator(new Drawer.DecoratorExonsConnectes(gene,Color.GRAY,zcc,rv,false,false));

                }
            }
            

         
    
        }
        
        HashMap<UPair<POMAGene>,HashSet<Region>> regions = new HashMap<>();
        
        for(Region r : poma.getMatchedRegions())
        {
            UPair<POMAGene> up = UPair.createU(r.g1, r.g2);
            MapBuilder.update(regions, up,r);
        }
       
        
        for(int i = 0 ; i< genes.size();i++)
        {
        
            RegionVector covered = new RegionVector();
            Stroke stroke = new BasicStroke(Math.round(SCALE*3));
            int alpha= 120;
       
            for(int j = i+1 ; j< genes.size()-1;j++)
            {
                
                //int j = i+1;
                List<Region> relevant = new ArrayList<Region>();
                
                POMAGene from = genes.get(i);
                POMAGene to = genes.get(j);
                
                
                UPair<POMAGene> p1 = UPair.createU(genes.get(i),genes.get(j)) ;
                
                if ( regions.containsKey(p1))
                    relevant.addAll(regions.get(p1));
            
                
                
                relevant.sort((Region r1, Region r2) -> r1.getSrcRegion().getX1()-r1.getSrcRegion().getX1());
                
                for(Region r : relevant)
                {
                   
                    
                    if(! isRelevant(covered,r.getSrcRegion()))
                        continue;
                    
                   
                    
                    covered = POMAGene.extend(covered, r.getSrcRegion());
                    
                    rpp.connect(obj2Rp.get(from), gene2zcc.get(from).getConvertedCoordinate(r.getSrcRegion().getX1()), obj2Rp.get(to),gene2zcc.get(to).getConvertedCoordinate(r.getDstRegion().getX1()),new Color(0,0,255,alpha),stroke);
                    rpp.connect(obj2Rp.get(from), gene2zcc.get(from).getConvertedCoordinate(r.getSrcRegion().getX2()-1), obj2Rp.get(to),gene2zcc.get(to).getConvertedCoordinate(r.getDstRegion().getX2()-1),new Color(255,0,0,alpha),stroke);
                    
                }
                stroke = new BasicStroke(Math.round(SCALE*1), BasicStroke.CAP_SQUARE, BasicStroke.JOIN_BEVEL, 0, new float[]{3}, 0);
                alpha=60;
            }
            
            for(int j = i-1 ; j>= 0;j--)
            {
                
                //int j = i+1;
                List<Region> relevant = new ArrayList<Region>();
                
                POMAGene from = genes.get(i);
                POMAGene to = genes.get(j);
                
                
                UPair<POMAGene> p1 = UPair.createU(genes.get(j),genes.get(i)) ;
                
                if ( regions.containsKey(p1))
                    relevant.addAll(regions.get(p1));

                
                relevant.sort((Region r1, Region r2) -> r1.getDstRegion().getX1()-r1.getDstRegion().getX1());
                
                for(Region r : relevant)
                {
                   
                    
                    if(! isRelevant(covered,r.getDstRegion()))
                        continue;
                    
                    
                    covered = POMAGene.extend(covered, r.getDstRegion());
                    
                    rpp.connect(obj2Rp.get(from), gene2zcc.get(from).getConvertedCoordinate(r.getDstRegion().getX1()), obj2Rp.get(to),gene2zcc.get(to).getConvertedCoordinate(r.getSrcRegion().getX1()),new Color(0,0,255,alpha),stroke);
                    rpp.connect(obj2Rp.get(from), gene2zcc.get(from).getConvertedCoordinate(r.getDstRegion().getX2()-1), obj2Rp.get(to),gene2zcc.get(to).getConvertedCoordinate(r.getSrcRegion().getX2()-1),new Color(255,0,0,alpha),stroke);
                    
                }
                stroke = new BasicStroke(Math.round(SCALE*1), BasicStroke.CAP_SQUARE, BasicStroke.JOIN_BEVEL, 0, new float[]{3}, 0);
                
            }
            
            
        }
        
        rpp.update();
        MultiPicPlaneFrame frame = new MultiPicPlaneFrame();

        frame.addInfo(rpp);
        //frame.setSize(5000, 8000);
        
        frame.update();
        
        return frame.getCombinedImage();
    }
    private static boolean isRelevant(RegionVector rv, Region1D r)
    {
        int k = 0;
        for(Region1D inter : rv.getIntersectingRegions(r, true))
        {
            k+=inter.getLength();
        }
        return (r.getLength()-k)>0;
    }
    /*
    public static BufferedImage drawAlignmentHasse(PartialOrderMultipleAlignment poma, boolean small,boolean showIsoforms,List<POMAGene> genes)
    {
        
        String name =null;
        for (POMAGene ag : genes)
        {
            if ( ag.getName() != null)
            {
                name = ag.getName();
                break;
            }
        }

        HashMap<POMAGene,ZoomedCoordinateConverter> gene2zcc = new  HashMap<POMAGene,ZoomedCoordinateConverter> ();

        if ( name == null) name ="";

        int maxLength = 0;
        for (POMAGene gene : genes)
        {
            ZoomedCoordinateConverter zcc = new ZoomedCoordinateConverter(gene.geneSequence.length()+1,null,10); 
            Object obj = new Object();
            zcc.addCategory(obj, SCALE, null);
            gene2zcc.put(gene,zcc );

            for(Region1D exon : gene.geneExons.getRegions())
            {
                zcc.addRegion(exon, obj);
            }
            
            for (String proteinId : gene.getProteins())
            {
                for (Region1D exon : gene.getRegionVector(proteinId).getRegions())
                {
                    zcc.addRegion(exon, obj);
                }
            }
            zcc.update();
            maxLength = Math.max(maxLength,zcc.getZoomedLength());
        }
        RegionPicturePane rpp = new RegionPicturePane(name, maxLength , null);
        
        if ( small)
        {
            rpp.changeRegionSettings(9, 12, 5);  

        }
        else
        {
            rpp.changeRegionSettings(16, 40, 20);  
            
        }

        HashMap<POMAGene,RegionPlot> obj2Rp = new HashMap<POMAGene,RegionPlot> ();
        for (POMAGene gene : genes)
        {
            ZoomedCoordinateConverter zcc = gene2zcc.get(gene);
            
            RegionVector rvg = new RegionVector();

            for (Region1D r : gene.geneExons.getRegions())
            {
                r.setObject(r);
                rvg.addRegion(r);
            }
            String spName = gene.getOrganism()!=null?gene.getOrganism().getShortName():gene.getTaxId()+"";
            String label = small?spName:spName + "(" + gene.getId() + ")";
            RegionPlot rpGene1 = rpp.addRegion(label,null,zcc.getConvertedRegVec(rvg));
            rpGene1.setDefaultDecorator(new Drawer.DecoratorExonsConnectes(gene,Color.BLACK,zcc,rvg,false,false));
            
            obj2Rp.put(gene, rpGene1);
     
            if ( showIsoforms)
            {
                if ( gene.getProteins().size() > 1){
                    for (String proteinId : gene.getProteins())
                    {
                        RegionVector rv = new RegionVector();
                        for (Region1D r : gene.getRegionVector(proteinId).getRegions())
                        {
                            r.setObject(r);
                            rv.addRegion(r);
                        }
                        //         System.out.println(rv);
                        RegionPlot rp = rpp.addRegion(small?"":proteinId,null,zcc.getConvertedRegVec(rv));
                        rp.setDefaultDecorator(new Drawer.DecoratorExonsConnectes(gene,Color.GRAY,zcc,rv,false,false));
                        
                    }
                }
            }

         
    
        }
       
        for(HashSet<Pair<POMAGene, Integer>> it : poma.getAlignment())
        {
            ArrayList<Pair<POMAGene, Integer>> items = new ArrayList<Pair<POMAGene,Integer>>(it);

            ArrayList<Pair<POMAGene, Integer>> cNew   = new ArrayList<Pair<POMAGene, Integer>>();
            ObjectGetter.apply(items, (Pair<POMAGene, Integer> p) -> { if ( obj2Rp.containsKey(p.getFirst())) cNew.add(p);});
            items= cNew;

            Collections.sort(items, (Pair<POMAGene,Integer> p1 , Pair<POMAGene,Integer> p2) -> {
                                 Integer index1 = null;
                                 Integer index2 = null;

                                 for (int i = 0 ; i< genes.size(); i++)
                                 {
                                     if ( p1.first == genes.get(i)) index1= i;
                                     if ( p2.first == genes.get(i)) index2= i;

                                 }

                                 return index2-index1;
                             });


            for (int i = 0 ; i< items.size() -1 ; i+=2)
            {
                POMAGene from = items.get(i).first;
                POMAGene to = items.get(i+1).first;
                
              //  System.out.println("FROM:" + items.get(i).second + " TO: " + items.get(i+1).second);
                rpp.connect(obj2Rp.get(from), gene2zcc.get(from).getConvertedCoordinate(items.get(i).second), obj2Rp.get(to),gene2zcc.get(to).getConvertedCoordinate(items.get(i+1).second),new Color(0,0,255,120));
            }
        }
        rpp.update();
        MultiPicPlaneFrame frame = new MultiPicPlaneFrame();

        frame.addInfo(rpp);

        frame.update();
        frame.setAutoSize(500, 800);
        return frame.getCombinedImage();
    }
    public  static BufferedImage drawAlignment(PartialOrderMultipleAlignment poma, boolean small,boolean showIsoforms)
    {
        return drawAlignment(poma,small,showIsoforms,poma.getGenes());
    }
    
   
    public static BufferedImage drawAlignment(PartialOrderMultipleAlignment poma, boolean small, boolean hasse, POMAGene g1, POMAGene g2)
    {
    
        List<POMAGene> genes = new ArrayList<>();
        genes.add(g1);
        genes.add(g2);
        
        if ( hasse)
        {
            return drawAlignmentHasse(poma,small,true,genes); 
            
        }
        
        return drawAlignment(poma,small,true,genes);
    }

    public static BufferedImage drawAlignment(PartialOrderMultipleAlignment poma, boolean small,boolean showIsoforms,List<POMAGene> genes)
    {


        HashMap<POMAGene,Integer> currentPos = new HashMap<POMAGene,Integer>();
        HashMap<POMAGene,RegionVector> rvs = new HashMap<POMAGene,RegionVector>();
        HashMap<POMAGene,Region1D> lastExon = new HashMap<POMAGene,Region1D> ();
        HashMap<POMAGene,Integer> currentInP = new HashMap<POMAGene,Integer> ();

        HashMap<String,RegionVector> proteins = new HashMap<String,RegionVector>();

        HashMap<POMAGene,HashMap<Integer,Integer>> gene2PosMap = new HashMap<POMAGene,HashMap<Integer,Integer>> ();
        List<Tuple4<POMAGene,Integer,POMAGene,Integer>> connections = new ArrayList<Tuple4<POMAGene,Integer,POMAGene,Integer>>();


        for (POMAGene g : genes)
        {
            
            currentPos.put(g, 0);
            rvs.put(g,new RegionVector());
            currentInP.put(g,0);
            lastExon.put(g, g.geneExons.get(0));
            gene2PosMap.put(g, new HashMap<Integer,Integer>());
            for (String proteinId : g.getProteins())
            {
                if ( proteins.containsKey(proteinId))
                {
                    throw new RuntimeException("Not unique protein id (" + proteinId);
                }
                proteins.put(proteinId,new RegionVector());
            }
        }

        for ( HashSet<Pair<POMAGene, Integer>> cur : poma.getAlignment())
        {
            ArrayList<Pair<POMAGene, Integer>> current = new ArrayList<Pair<POMAGene,Integer>>(cur);

            ArrayList<Pair<POMAGene, Integer>> cNew   = new ArrayList<Pair<POMAGene, Integer>>();
            ObjectGetter.apply(current, (Pair<POMAGene, Integer> p) -> { if ( rvs.containsKey(p.getFirst())) cNew.add(p);});
            current= cNew;

            Collections.sort(current, (Pair<POMAGene,Integer> p1 , Pair<POMAGene,Integer> p2) -> {
                Integer index1 = null;
                Integer index2 = null;

                for (int i = 0 ; i< genes.size(); i++)
                {
                    if ( p1.first == genes.get(i)) index1= i;
                    if ( p2.first == genes.get(i)) index2= i;
                }

                return index1.compareTo(index2);
            });

            
            if ( current.size() == 0) continue;
            int maxNewStart=0;
            List<Integer> froms = new ArrayList<Integer>();
            List<Boolean> endedWithExon = new ArrayList<Boolean>();
            for (Pair<POMAGene, Integer> p  : current)
            {
                int start = currentPos.get(p.first);
                int end = p.second;

                List<Region1D> before = getBetween(p.first.geneExons, currentInP.get(p.first), end);
                boolean gap =true;
                for (Region1D r : before) 
                {
                    if ( r.getLength() == 0)continue;
                    if ( r.getX1() <currentInP.get(p.first) )
                    {
                        throw new RuntimeException("Cannot happen");
                    }
                    Region1D currentExon = getExon(p.first.geneExons, r);
                    Region1D beforeExon = lastExon.get(p.first);


                    if ( beforeExon != currentExon)
                    {
                        //start+=50;
                        gap=true;
                    }
                    else
                    {
                        gap =false;  
                    }
                    Region1D newR = new Region1D(start,start+r.getLength(),currentExon);
                    if ( showIsoforms)
                    {
                        for (String proteinId : p.first.getProteins())
                        {
                            RegionVector rv =  p.getFirst().getRegionVector(proteinId);//protein.getCodingRegionVector(false);
                            List<Region1D> intersecting =rv.getIntersectingRegions(r);
                            if ( intersecting != null && intersecting.size() > 0)
                            {
                                for (Region1D i : intersecting)
                                {
                                    Region1D rI = i.getIntersectedRegion(r);
                                    if (rI != null && rI.getLength() > 0)
                                    {
                                        //      System.out.println(proteins.get(protein.id) + " " + intersecting + " " + newR);
                                        int newStart= start+(rI.getX1()-r.getX1());

                                        Region1D newR2 = new Region1D(newStart,newStart+rI.getLength(),currentExon);

                                        proteins.get(proteinId).addRegion(newR2); 
                                    }
                                }


                            }
                        }
                    }


                    lastExon.put(p.first,currentExon);
                    rvs.get(p.first).addRegion(newR);

                    start=newR.getX2();
                    currentInP.put(p.first, r.getX2());


                }
                endedWithExon.add(gap);

                froms.add(start);

                maxNewStart=Math.max(maxNewStart, start);
            }

            for (Pair<POMAGene, Integer> p  : current)
            {

                currentPos.put(p.getFirst(), maxNewStart);
            }
            for (int i = 0 ; i < current.size()-1;i++)
            {

                int start1 =maxNewStart; //currentPos.get(current.get(i).first);
                //if ( endedWithExon.get(i))start1+=50;

                int start2 = maxNewStart;//currentPos.get(current.get(i+1).first);
                //if ( endedWithExon.get(i+1))start2+=50;

                //if ( start1 != start2) throw new RuntimeException("Cannot be (" + start1 + " " + start2);
                connections.add(Tuple4.create(current.get(i).first,start1, current.get(i+1).first, start2));
            }

        }
        int length =0;
        for ( RegionVector  rv  : rvs.values())
        {

            length=Math.max(length,rv.getX2());
        }

        ZoomedCoordinateConverter converter = new ZoomedCoordinateConverter(length,0.33,null);
        converter.addRegion(new Region1D(0,length,null), new Object());
        converter.update(); 

        String name =null;
        for (POMAGene ag : genes)
        {
            if ( ag.getName() != null)
            {
                name = ag.getName();
                break;
            }
        }
        if ( name == null) name ="";
        
        AbstractZoomedMultiPicPlane _plot = null;
        RegionPicturePane rpp = null;

        if ( grid){
          _plot = new AbstractZoomedMultiPicPlane(null, null, converter.getZoomedLength());
          rpp = _plot.createRegionPane(name);
        }else{
            rpp = new RegionPicturePane(name,  converter.getZoomedLength() , null);
        }
        
        
        if ( small)
        {
            rpp.changeRegionSettings(9, 12, 5);  
        }
        else
        {
            rpp.changeRegionSettings(16, 40, 20);  
        }
        int isoforms=0;
       HashMap<POMAGene,RegionPlot> agRpp = new HashMap<POMAGene,RegionPlot>();
        for ( POMAGene ag : genes)
        {
            if (! rvs.containsKey(ag)) continue;
            //  System.out.println("ADD:" + ag.getId() + " " + rvs.get(ag));
            String label = null;
            if ( ag.getOrganism() != null)
            {
                label = small?ag.getOrganism().getShortName():ag.getOrganism().getShortName() + "(" + ag.getName() + ")";
            }else{
                label = ag.getId();
            }
            
            RegionPlot rp1 = rpp.addRegion(label, null, converter.getConvertedRegVec(rvs.get(ag)));
            isoforms+=ag.getProteins().size();
            if ( showIsoforms)
            {
             //   if ( ag.getProteins().size() > 1){
                    for (String proteinId : ag.getProteins())
                    {
                        //     System.out.println(protein.id + " " + proteins.get(protein.id) + " " + protein.getCodingRegionVector(false));
                        RegionPlot rrpp = rpp.addRegion(small?"": proteinId, null, converter.getConvertedRegVec(proteins.get(proteinId)));
    
                        rrpp.setDefaultDecorator(new Decorator(proteins.get(proteinId),false));
                    }
               // }
            }
            //RegionPlot rp1 = rpp.addRegion( ag.getOrganism().name, null, converter.getConvertedRegVec(rvs.get(ag)));
            rp1.setDefaultDecorator(new Decorator(rvs.get(ag),false));
            agRpp.put(ag,rp1);
        }
//        System.out.println("ISOFORM:" + isoforms);
        for (Tuple4<POMAGene, Integer, POMAGene, Integer> c : connections)
        {
            rpp.connect(agRpp.get(c.get0()), converter.getConvertedCoordinate(c.get1()), agRpp.get(c.get2()),converter.getConvertedCoordinate(c.get3()), new Color(0, 0, 255,100));
        }

        rpp.update();
        
        if ( grid){
            _plot.update(true);
    
            return _plot.getCombinedImage();
        }
        MultiPicPlaneFrame frame = new MultiPicPlaneFrame();

        frame.addInfo(rpp);

        frame.update();
        frame.setAutoSize(500, 800);
        BufferedImage img = frame.getCombinedImage();
        frame.dispose();
        return img;
    }
    public static BufferedImage drawAlignmentHasse(PartialOrderMultipleAlignment poma,boolean small,boolean showIsoforms)
    {
        return drawAlignmentHasse(poma,small,showIsoforms,poma.getGenes());
    }
   */
    
    static class GappedRegion extends Region1D{
        public GappedRegion(Region1D<?> exon) {
            super(exon);
        }
    }
    static class IntronRegion extends Region1D{
        public IntronRegion(Region1D<?> exon) {
         //   this.exon = exon;
            super(exon);
        }
    }
    
    public static BufferedImage drawAlignmentComplete(PartialOrderMultipleAlignment poma,  List<HashSet<Pair<POMAGene, Integer>>> mapping,boolean showIsoformLabels, boolean showGeneId,boolean showGaps,int intronSize,List<POMAGene> orderedGenes)
    {
        FASTATransferformation tr = Formater.getFASTAAlignmentTransformation(poma, mapping,orderedGenes,intronSize) ;
        String name =null;
        for (POMAGene ag : orderedGenes)
        {
            if ( ag.getName() != null)
            {
                name = ag.getName();
                break;
            }
        }
        
        AbstractZoomedMultiPicPlane _plot = null;
        RegionPicturePane rpp = null;

        ZoomedCoordinateConverter converter = new ZoomedCoordinateConverter(tr.aliLength,SCALE*0.25,null);
        Object obj = new Object();
        converter.addRegion(new Region1D(0,tr.aliLength+1), obj);
        converter.update();
        
        if ( grid){
          _plot = new AbstractZoomedMultiPicPlane(null, null, converter.getZoomedLength());
          rpp = _plot.createRegionPane(name);
        }else{
            rpp = new RegionPicturePane(name,  converter.getZoomedLength() , null);
        }
        
        rpp.changeRegionSettings((int)Math.round(SCALE*10), (int)Math.round(SCALE*20), (int)Math.round(SCALE*5));
        
        for(POMAGene pg : orderedGenes)
        {
            StringBuffer ali = tr.fastaStr.get(pg);
            
            RegionVector rv = new RegionVector();
            Integer start = null;
            Integer end = null;
            Region1D realExon = null;
            
            HashMap<String,RegionVector> protRv = new HashMap<>();
            HashMap<String,Integer> protStart = new HashMap<>();
            HashMap<String,Integer> protEnd = new HashMap<>();
            HashMap<String,Region1D> protRealExon = new HashMap<>();
            
            for(String proteinId : pg.getProteins())
            {
                protStart.put(proteinId, null);
                protEnd.put(null, null);
                
                protRv.put(proteinId,new RegionVector());
                protRealExon.put(proteinId, null);
            }
            for(int i = 0 ; i < ali.length()+1;i++)
            {
                if (  i == ali.length() || ali.charAt(i) == '*' || (showGaps &&   ali.charAt(i) == '-') )
                {
                    if ( start != null)
                    {
                        Region1D region = new Region1D(start,end  ,realExon);
                        
                        if ( region.getLength()>= 3) {
                            /*
                            if ( rv.size() > 0   )
                            {
                                Region1D lastExon = (Region1D) rv.get(rv.size()-1).getObject();
                                if ( lastExon.equals(realExon) || lastExon.getX2() == realExon.getX1()) 
                                {
                                    Region1D r = new Region1D(rv.getX2(),region.getX1(),new GappedRegion(realExon));
                                    rv.addRegion(r);
                                }else{
                                    rv.addRegion(region);
                                }
                            }else{
                                rv.addRegion(region);
                            }
                            */
                            rv.addRegion(region);
                            
                        }
                        
                       
                    }
                    
                    for(String proteinId : pg.getProteins())
                    {
                        
                        if (protStart.get(proteinId) != null )
                        {
                            Region1D region = new Region1D(protStart.get(proteinId),protEnd.get(proteinId),protRealExon.get(proteinId));
                            
                            if ( region.getLength()>=3) {
                                protRv.get(proteinId).addRegion(region);
                                
                               // protRv.get(proteinId).addRegion(region);
                                /*
                                if ( protRv.get(proteinId).size() > 0)
                                {
                                    Region1D lastExon = (Region1D) protRv.get(proteinId).get(protRv.get(proteinId).size()-1).getObject();
                                    if ( lastExon.equals(protRealExon.get(proteinId)) || lastExon.getX2() == protRealExon.get(proteinId).getX1()) 
                                    {
                                        Region1D r = new Region1D(protRv.get(proteinId).getX2(),region.getX1(),new GappedRegion(realExon));
                                        protRv.get(proteinId).addRegion(r);
                                        //curProtRv.get(curProtRv.size()-1).setX2(protEnd.get(proteinId));
                                    }else{
                                        protRv.get(proteinId).addRegion(region);
                                    }  
                                }else
                                {
                                    protRv.get(proteinId).addRegion(region);
                                }
                                */
                                
                            }

                        }
                        
                        protStart.put(proteinId, null);
                        protEnd.put(proteinId, null);
                        
                    }
                    
                    start = null;
                    end = null;
                    continue;
                }
                
                if ( ali.charAt(i) == '-')
                {
                    for(String proteinId : pg.getProteins())
                    {
                        protEnd.put(proteinId, i);
                    }
                    
                    end = i;
                    continue;
                }
                
                Integer currentPos = tr.aliPosToGenePos.get(pg).get(i);
                Region1D curExon =getRegion(pg.geneExons,currentPos);
                 
                if ( start == null)
                {
                    start = i;
                }
            
                realExon = curExon;
                
                end = i;
                
                for(String proteinId : pg.getProteins())
                {
                    
                    if ( pg.getRegionVector(proteinId).isIn(currentPos))
                    {
                        realExon = getRegion(pg.getRegionVector(proteinId),currentPos);
                        
                        if (protStart.get(proteinId) == null )
                        {
                            protStart.put(proteinId, i); 
                        }
                        else
                        {
                            curExon = getRegion(pg.getRegionVector(proteinId),currentPos);
                            
                            if ( curExon != protRealExon.get(proteinId))
                            {
                                Region1D region = new Region1D(protStart.get(proteinId),protEnd.get(proteinId),protRealExon.get(proteinId));
                            
                                if ( region.getLength() >= 3)
                                   protRv.get(proteinId).addRegion(region);
                                
                                if ( protRealExon.get(proteinId) == null)
                                {
                                    throw new RuntimeException("Not possible");
                                }
                                protStart.put(proteinId, i); 
                                protRealExon.put(proteinId, curExon);
                            }  
                        }
                        protRealExon.put(proteinId,realExon);
                        
                        protEnd.put(proteinId, i);
                    }
                    
                }
            }
            
            String label = null;
            String pgName = pg.getName();
            if ( pgName == null) pgName ="";
            pgName = pgName.toUpperCase();
            
            String orgName = ISAR.OrganismNameReader.getShortName(pg.getTaxId());
            if ( orgName != null)
            {
                label = (showGeneId?pg.getId():orgName) + " (" + pgName + ")" ;
            }else{
                label = (showGeneId?pg.getId():"") + " (" + pgName + ")" ;
            }
            RegionPlot rp = rpp.addRegion(label, null, converter.getConvertedRegVec(rv));
            rp.setDefaultDecorator(new DecoratorExonsConnectes(pg,Color.BLACK,converter,rv,false,true));
            if (!showIsoformLabels && pg.getProteins().size() == 1)
                continue;
            
            for(String proteinId : pg.getProteins())
            {
                RegionPlot rpProt = rpp.addRegion(showIsoformLabels?proteinId:"", null, converter.getConvertedRegVec(protRv.get(proteinId)));
                rpProt.setDefaultDecorator(new DecoratorExonsConnectes(pg,Color.GRAY,converter,protRv.get(proteinId),false,false));

            }
            
        }
   
        rpp.update();
        
        if ( grid){
            _plot.update(true);
    
            return _plot.getCombinedImage();
        }
        MultiPicPlaneFrame frame = new MultiPicPlaneFrame();

        frame.addInfo(rpp);

        frame.update();
        frame.setAutoSize(500, 800);
        return frame.getCombinedImage();
    }
    
    private static Region1D getRegion(RegionVector rv, int p)
    {
        for(Region1D r : rv.getRegions())
        {
            if ( r.contains(p)) return r;
        }
        
        return null;
    }


}
