package isoformtransfer.isar;

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;
import java.util.Vector;

import javax.imageio.ImageIO;

import algo.alignment.Alignment;
import genome.Organism;
import isoformtransfer.isar.ISAR.ISAROptionParser;
import isoformtransfer.isar.PartialOrderMultipleAlignment.UnmatchedStrategy;
import lmu.utils.FileUtils;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.ObjectGetter;
import lmu.utils.OptionParser;
import lmu.utils.Pair;
import lmu.utils.Region1D;
import lmu.utils.RegionVector;
import lmu.utils.StringUtils;
import protein.algo.ProteinAligner;
import protein.aminoacid.AminoAcidType;

public class Formater {
    PartialOrderMultipleAlignment poma = null;
    
    List<HashSet<Pair<POMAGene, Integer>>> mapping;

    public Formater(File pomaXML){
         this( new PartialOrderMultipleAlignment(pomaXML));
    }
    public Formater(PartialOrderMultipleAlignment poma){
        this(poma,poma.getAlignment());
    }
    
    public Formater(PartialOrderMultipleAlignment poma,List<HashSet<Pair<POMAGene, Integer>>> mapping){
        this.poma = poma;
        this.mapping = mapping;
   }
    
    public List<Integer> getOverlaps(RegionVector rv, Region1D r)
    {
        List<Integer> positions = new ArrayList<Integer>();
        
        for(Region1D exon: rv.getIntersectingRegions(r, true))
        {
            for(int i = exon.getX1() ; i< exon.getX2(); i++)
            {
                positions.add(i);
            }
               
        }
        return positions;
    }
    
    public boolean isIn(RegionVector rv, int p)
    {
        for(Region1D r : rv.getRegions())
            if ( r.contains(p)) return true;
        return false;
    }
    public String getCDNAAlignmentStr(String g1, String p1, String g2, String p2)
    {
        POMAGene pg1 = this.poma.getGene(g1);
        POMAGene pg2 = this.poma.getGene(g2);
        
        String s1 = pg1.getSequence(p1, true);
        String s2 = pg2.getSequence(p2, true);
        
        Alignment ret = getCDNAAlignment(pg1,p1,pg2,p2);
            
        float numId = ret.getNumId(s1, s2);
        float seqId =numId / (0.5f*(s1.length()+s2.length())); 
        return ">" + seqId + "\n" + ret.getFormattedBy(80, p1, s1, p2, s2);
    }
    public String getAlignmentStr(String g1, String p1, String g2, String p2, boolean gotoh)
    {
        POMAGene pg1 = this.poma.getGene(g1);
        
        POMAGene pg2 = this.poma.getGene(g2);
        
        
        String s1 = pg1.getSequence(p1, false);
        String s2 = pg2.getSequence(p2, false);
        
        Alignment ret = null;
        if ( gotoh)
        {
            ProteinAligner aligner = new ProteinAligner();
            ret = aligner.align(s1, s2);
        }else{
            
            Alignment ali = getCDNAAlignment(pg1,p1,pg2,p2);
            
            ret = this.convertToAS(ali, s1.length(),s2.length());
        }
        float numId = ret.getNumId(s1, s2);
        float seqId =numId / (0.5f*(s1.length()+s2.length())); 
        return ">" + seqId + "\n" + ret.getFormattedBy(80, p1, s1, p2, s2);
    }
    
    public Alignment getAlignment(String p1, String p2)
    {
        POMAGene g1 = null;
        POMAGene g2 = null;
        for(POMAGene gene : poma.getGenes())
        {
            if ( gene.getProteins().contains(p1))
                g1 = gene;
            if ( gene.getProteins().contains(p2))
                g2 = gene;
            
        }
        return getCDNAAlignment(g1,p1,g2,p2);
    }
    public void printPairwise(PrintWriter out, boolean dna,  boolean verbose)
    {
        
         ArrayList<POMAGene> genes = new ArrayList<>(poma.getGenes());
      //  genes.add(poma.getGene("ENSG00000134201"));
      //  genes.add(poma.getGene("ENSMMUG00000019630"));
        
        
        for(int i =  0 ; i< genes.size(); i++)
        {
            for(int j =  i+1 ; j< genes.size(); j++)
            {
                for(String p1 : genes.get(i).getProteins())
                {
                    for(String p2 : genes.get(j).getProteins())
                    {
                        Alignment ali = getCDNAAlignment(genes.get(i),p1,genes.get(j),p2);
                        
                        String s1 = genes.get(i).getSequence(p1, true);
                        String s2 = genes.get(j).getSequence(p2, true);
                        String cDNAAli = ali.getFormattedBy(80,p1,s1 ,p2, s2);
                       // System.out.println(a);
        
                        s1 = genes.get(i).getSequence(p1, false);
                        s2 = genes.get(j).getSequence(p2, false);
                        
                        Alignment ali2 = convertToAS(ali,s1.length(),s2.length());
                        
                        String isarALI = ali2.getFormattedBy(70,p1,s1 ,p2, s2);
                        
                        
                        float seqId_isar = (float)ali2.getNumId(s1, s2)/(0.5f*(s1.length()+s2.length()));
                        
                        ProteinAligner aligner = new ProteinAligner();
                        Alignment ali3 = aligner.align(s1, s2);
                        float seqId = (float)ali3.getNumId(s1, s2)/(0.5f*(s1.length()+s2.length()));
                        
                      //  String gotoALI = ali3.getFormattedBy(70,p1,s1 ,p2, s2);
                        
                        synchronized(out)
                        {
                            out.printf(">%s %s %s %s %s %.4f %.4f\n",
                                    poma.getPOMAID(),
                                    poma.getGenes().get(i).getId(),
                                    poma.getGenes().get(j).getId(), 
                                    p1,
                                    p2,
                                    seqId_isar,
                                    seqId
                            );
                            if ( verbose)
                            {
                                if ( dna)
                                {
                                    out.print(cDNAAli);
                                }else
                                {
                                    out.print(isarALI);
                                    //out.println(gotoALI);
                                }
                                    
                            }
                            
                            out.flush();
                        }
                        
                    }
                }
            }
        }
    }
    public Alignment convertToAS(Alignment ali,int l1, int l2 )
    {
        
        Vector<Pair<Integer,Integer>> matches = new Vector<Pair<Integer,Integer>>();
        
        
        
        for(Pair<Integer, Integer> p : ali.alignment)
        {
            if ( p.getFirst()%3==0 && p.getSecond()%3 == 0)
            {
                int p1 = p.getFirst()/3;
                int p2 = p.getSecond()/3;
                if ( p1 >= l1 || p2 >= l2) continue;
                
                matches.add(Pair.create(p1, p2));
            }
        }
        
        return   new Alignment(l1,l2,matches);
    }
    public Alignment getCDNAAlignment(POMAGene g1, String p1, POMAGene g2, String p2)
    {
        List<HashSet<Pair<POMAGene, Integer>>> ali = mapping;
        
        Vector<Pair<Integer,Integer>> matches = new Vector<Pair<Integer,Integer>>();
        
        RegionVector exons1 = g1.getRegionVector(p1);
        RegionVector exons2 = g2.getRegionVector(p2);
        
        HashMap<Integer,Integer> m1 = new HashMap<>();
        HashMap<Integer,Integer> m2 = new HashMap<>();

        int l1 = g1.getSequence(p1, true).length();
        int l2 = g2.getSequence(p2, true).length();
        
        int k = 0 ; 
        for(Region1D r : exons1.getRegions())
        {
            for(int i  = r.getX1(); i < r.getX2(); i++)
            {
                if ( k < l1) m1.put(i, k++);
            }
        }
        k= 0;
        for(Region1D r : exons2.getRegions())
        {
            for(int i  = r.getX1(); i < r.getX2(); i++)
            {
                if ( k < l2 ) m2.put(i, k++);
            }
        }
        
        for(HashSet<Pair<POMAGene, Integer>> a : ali)
        {
            Integer f1 = null;
            Integer f2 = null;
            for(Pair<POMAGene, Integer> t : a)
            {
                if ( t.getFirst() == g1 && isIn(exons1,t.getSecond())) 
                        f1 = t.getSecond();
                
                
                if ( t.getFirst() == g2 && isIn(exons2,t.getSecond())) 
                    f2 = t.getSecond();
            
            }
            if( f1 != null && f2 != null && m1.get(f1) != null && m2.get(f2) != null){
                
                matches.add(Pair.create(m1.get(f1),m2.get(f2)));
                
            }
        }
        
        Alignment ret = new Alignment(l1,l2,matches);
    
        return ret;
    }
    
    static class FASTATransferformation
    {
        public FASTATransferformation(List<POMAGene> orderedGenes) {
            this.genes = orderedGenes;
        }
        HashMap<POMAGene,StringBuffer> fastaStr = new HashMap<>();
        
        
        HashMap<POMAGene,HashMap<Integer,Integer>> genePosToAliPos = new HashMap<>();
        HashMap<POMAGene,HashMap<Integer,Integer>> aliPosToGenePos = new HashMap<>();
        
        List<POMAGene> genes ;
        Integer aliLength;
        
    }
    
    private static Region1D getExon( RegionVector exons, int pos)
    {
        for(Region1D r : exons.getRegions())
        {
            if ( pos>=r.getX1() && pos < r.getX2()) 
                return r;
        }
        throw new RuntimeException(pos + " is outside of "+ exons);
    }
    
    private static List<Pair<Integer,Region1D>> getBetween(int lastPosition, Region1D lastExon,int currentPosition,Region1D currentExon, RegionVector exons)
    {
        List<Pair<Integer,Region1D>> between = new ArrayList<>();
        
        //last position with exon -- > 0
        between.add(Pair.create(lastPosition,lastExon));

        for(Region1D exon :exons.getRegions())
        {
            for(int i = exon.getX1(); i< exon.getX2(); i++)
            {
                if ( i >lastPosition && i<currentPosition)
                {
                    between.add(Pair.create(i, exon));
                }
                
            }
        }
        ///current position with exon --> last element in between
        between.add(Pair.create(currentPosition,currentExon));
        
        return between;
    }
    private static int addIntronGaps(FASTATransferformation ret,POMAGene targetGene, int gaps)
    {
        //if ( ret.fastaStr.get(targetGene).charAt(ret.fastaStr.get(targetGene).length()-2) == '*') return 0;
        if ( ret.fastaStr.get(targetGene).charAt(ret.fastaStr.get(targetGene).length()-1) == '*') return 0;
        //check if intron already added
        StringBuffer aliStr = ret.fastaStr.get(targetGene);
        /*
        int c = 0;
        for(int i = aliStr.length()-1;i>=0;i-- )
        {
            if ( aliStr.charAt(i) == '-') continue;
            if ( aliStr.charAt(i) == '*') return 0;
            c++;
            if ( c> 2) break;
        }
        */
        for(int k = 0 ; k< gaps; k++)
        {
            for(POMAGene gene :ret.genes)
            {
                ret.fastaStr.get(gene).append("*");
            }

        }   
        return gaps;
    }
    private static int fillBetween(FASTATransferformation ret,POMAGene targetGene, int lastPosition, Region1D lastExon, int position, Region1D currentExon,int cPos,int intronGaps)
    {
        List<Pair<Integer, Region1D>> between = getBetween(lastPosition,lastExon, position, currentExon,targetGene.geneExons);
        //System.out.println(between + "\t" + pos);
        //System.exit(1);
        
        Region1D f = between.get(0).getSecond();
        Region1D l = between.get(between.size()-1).getSecond();
        if ( between.size() == 2)
        {
            if ( f != l){
                cPos+=addIntronGaps(ret,targetGene,intronGaps);
            }
            return cPos;
        }
        for(int i = 1 ; i< between.size()-1; i++)
        {
            
            Region1D le = between.get(i-1).getSecond();
            Region1D ce = between.get(i).getSecond();
           
            //add intron-gap
            if ( le != ce) {
                cPos+=addIntronGaps(ret,targetGene,intronGaps);
            }
            
            //add alignment-gap
            
            for(POMAGene gene :ret.genes)
            {
                if ( gene == targetGene)
                    continue;
                ret.fastaStr.get(gene).append("-");
                    
            }
            
            
            //set positions for unique positions
            int p = between.get(i).getFirst();
            ret.fastaStr.get(targetGene).append(targetGene.geneSequence.charAt(p));
            ret.genePosToAliPos.get(targetGene).put(p, cPos);
            
            cPos++;
            
        }
        Region1D le = between.get(between.size()-1).getSecond();
        Region1D ce = between.get(between.size()-2).getSecond();
        if ( le != ce) {
            cPos+=addIntronGaps(ret,targetGene,intronGaps);
        }
        
        return cPos;
    }
    public static FASTATransferformation getFASTAAlignmentTransformation(PartialOrderMultipleAlignment poma, List<HashSet<Pair<POMAGene, Integer>>> mapping, List<POMAGene> orderedGenes,int intronGaps)
    {
        FASTATransferformation ret = new FASTATransferformation(orderedGenes);
        
        HashMap<POMAGene,Integer> lastMatch = new HashMap<POMAGene,Integer>();
        HashSet<POMAGene> genes = new HashSet<>(orderedGenes);
        
        HashMap<POMAGene,Region1D> lastExon = new HashMap<POMAGene,Region1D>();
        
        for(POMAGene gene : genes)
        {
            ret.fastaStr.put(gene, new StringBuffer());
            lastMatch.put(gene, -1);
            ret.genePosToAliPos.put(gene, new HashMap<>());
            
            lastExon.put(gene,gene.geneExons.get(0));
        }
        int cPos = 0;
        for(HashSet<Pair<POMAGene, Integer>> a : mapping)
        {
            if ( a.size() == 1 ) 
                continue;
            
            List<Pair<POMAGene,Integer>> pos = new ArrayList<>(a);
            
            Collections.sort(pos, (Pair<POMAGene,Integer> p1, Pair<POMAGene,Integer> p2)->{
                int i1 = 0;
                int i2 = 0;
                
                for(int i = 0 ; i < orderedGenes.size();i++)
                {
                    if ( orderedGenes.get(i)==p1.getFirst())i1=i;
                    if ( orderedGenes.get(i)==p2.getFirst())i2=i;
                    
                }
                return i1-i2;
            });
            
            for(Pair<POMAGene, Integer> m : pos)
            {
                if (! genes.contains(m.getFirst()))
                    continue;
                
                //last match
                Integer lastPosition = lastMatch.get(m.getFirst());
                Integer position = m.getSecond();
               // System.out.println(position + " " + m.getFirst());
                Region1D currentExon = getExon(m.getFirst().geneExons,position);
                
                cPos=fillBetween(ret,m.getFirst(),lastPosition,lastExon.get(m.getFirst()), position, currentExon,cPos,intronGaps);

            }
            Set<POMAGene> considered = new HashSet<>();
            for(Pair<POMAGene, Integer> m : pos)
            {
                if (! genes.contains(m.getFirst()))
                    continue;
                
                considered.add(m.getFirst());
                
                int p = m.getSecond();
                Region1D currentExon = getExon(m.getFirst().geneExons,p);
                
                ret.fastaStr.get(m.getFirst()).append(m.getFirst().geneSequence.charAt(p));
                ret.genePosToAliPos.get(m.getFirst()).put(p, cPos);
                
                lastMatch.put(m.getFirst(),m.getSecond());
                lastExon.put(m.getFirst(),currentExon);
                
            }
            for(POMAGene g : genes)
            {
                if ( considered.contains(g))
                    continue;
                
                ret.fastaStr.get(g).append("-");
            }
            
            cPos++;
        }   
        
        //add tail
        
        for(POMAGene gene : genes)
        {
            //last match
            Integer lastPosition = lastMatch.get(gene);
            Integer endPosition = gene.geneExons.getX2();
            
            //exons gaps?
            Region1D lE =  lastExon.get(gene);
            
            for(Region1D exon :gene.geneExons.getRegions())
            {

                for(int i = exon.getX1(); i< exon.getX2(); i++)
                {

                    if ( i >lastPosition )
                    {
                         cPos=fillBetween(ret,gene,lastPosition,lE, i, exon,cPos,intronGaps);
                         ret.fastaStr.get(gene).append(gene.geneSequence.charAt(i));
                         ret.genePosToAliPos.get(gene).put(i, cPos);
                            
                        for(POMAGene g : genes)
                        {
                            if (g == gene)
                                continue;
                            
                            ret.fastaStr.get(g).append("-");
                        }
                        

                        lastPosition = i;

                        lE = exon;
                        cPos++;
                    }
                }
            }
            
            
            ret.aliPosToGenePos.put(gene, ObjectGetter.reverseMap(ret.genePosToAliPos.get(gene)));
            
        }
       
        for(Entry<POMAGene, StringBuffer> k : ret.fastaStr.entrySet())
        {
            //System.out.println(k.getKey());
            //System.out.println(k.getValue().length() + " " + k.getKey().geneExons.getCoveredLength());
         //   List<Entry<Integer,Integer>> w = new ArrayList(ret.genePosToAliPos.get(k.getKey()).entrySet());
         //  Collections.sort(w,(Entry<Integer,Integer> e1, Entry<Integer,Integer> e2) ->e1.getKey() - e2.getKey() );
         // /  System.out.println(w);
           
        }
        
        ret.aliLength = cPos;
        return ret;
    }
  
    
    public void printISAR(PrintWriter pw, List<POMAGene> genes)  
    {
        printISAR(pw,false,genes,0);
    }
    public void printISAR(PrintWriter pw,boolean fastaAS,List<POMAGene> genes,int intronSize)  
    {
        if ( genes == null || genes.size() == 0) genes = this.poma.getGenes();
                
        FASTATransferformation fastaTransform = getFASTAAlignmentTransformation(poma,mapping,genes,intronSize);
        List<String> out = new ArrayList<>();
        
        int idLength = 0;
        for(POMAGene gene : genes)
        {
            idLength = Math.max(idLength,gene.getId().length()+2);
            Vector<String> exons = ObjectGetter.map(gene.geneExons.getRegions(), (Region1D exon)-> exon.getX1()+"-"+exon.getX2());
            if ( !fastaAS)
                out.add(String.format("#%s:%s\n",gene.getId(),StringUtils.join(",",exons) ));
            
            
            for(String p : gene.getProteins())
            {
                idLength = Math.max(idLength,p.length()+6);
                exons = ObjectGetter.map(gene.getRegionVector(p).getRegions(), (Region1D exon)-> exon.getX1()+"-"+exon.getX2());
                if ( !fastaAS)
                    out.add(String.format("#%s:%s\n",p,StringUtils.join(",",exons) ));
                
            }
        }
        
        final String rowFMT = String.format("%%%ds: %%s\n", idLength);
        
        //PrintWriter pw = FileUtils.getWriter(new File("ali.fasta"));//new PrintWriter(new File("ali.fasta"));
        
        List<String> pIds = new ArrayList<>();
        boolean onlyProteinName = true;
        
        for(POMAGene gene : genes)
        //for(Entry<POMAGene, StringBuffer> e : fastaTransform.fastaStr.entrySet())
        {
            StringBuffer ali = fastaTransform.fastaStr.get(gene);
            if ( !fastaAS)
                out.add(String.format(rowFMT,gene.getId() + "_g", ali.toString()));
          //  String gappedCDNA = e.getValue().toString();
            
            for(String proteinId : gene.getProteins())
            {
                if ( onlyProteinName)
                {
                    pIds.add(proteinId);
                }else{
                    if ( gene.getName() != null && gene.getName().length() > 0)
                    {
                        pIds.add(gene.getName() + "_" + proteinId);
                    }else{
                        pIds.add(gene.getId() + "_" + proteinId);
                    }
                }
                StringBuffer pAli = new StringBuffer();
                StringBuffer pAliAS = new StringBuffer();
                StringBuffer exonStr = new StringBuffer();
                
                Integer lastPos = -1;
                String codon ="";
                Integer[] pos = new Integer[2];
                
                int exonId = 0;
                for(Region1D exon : gene.getRegionVector(proteinId).getRegions())
                {
                    for(int i = exon.getX1(); i< exon.getX2(); i++)
                    {
                        Integer aPos = fastaTransform.genePosToAliPos.get(gene).get(i);
                        if ( aPos == null)
                        {
                            throw new RuntimeException("No alignment position for gene position " + i + " in gene " + gene);
                        }
                        for(int k = lastPos+1 ; k < aPos; k++)
                        {
                            pAli.append("-");
                            pAliAS.append("-");
                            exonStr.append("-");
                        }
                        pAli.append(gene.geneSequence.charAt(i));
                        
                        codon+=gene.geneSequence.charAt(i);
                        exonStr.append(exonId%2==0?"0":"1");
                        pAliAS.append("-");
                        
                        if ( codon.length()==1 )
                        {
                            pos[0] = pAliAS.length()-1;
                            
                            //pAliAS.append( ".");
                        }
                        if ( codon.length()==2)
                        {
                            pos[1] = pAliAS.length()-1;
                        }
                        if ( codon.length()==3)
                        {
                            //pAliAS.append(AminoAcidType.get(codon) );
                            
                            pAliAS.setCharAt(pos[0], AminoAcidType.get(codon));
                                
                            pAliAS.setCharAt(pos[1], ':');
                            pAliAS.setCharAt(pAliAS.length()-1,'.');
                            
                            
                            codon = "";
                        }
                        
                        lastPos = aPos;
                        
                    }
                    exonId++;
                }      
                for(int k = lastPos+1 ; k < fastaTransform.aliLength; k++)
                {
                    pAli.append("-");
                    pAliAS.append("-");
                    exonStr.append("-");
                    
                }
                if ( !fastaAS)
                {
                    out.add(String.format(rowFMT,proteinId  + "_pEXO",exonStr.toString()));
                    out.add(String.format(rowFMT,proteinId  + "_pDNA",pAli.toString()));
                    out.add(String.format(rowFMT,proteinId + "_pAS",pAliAS.toString() ));
                }else{
                    out.add(pAliAS.toString() );
                }
             
            }
        }
        if ( !fastaAS)
        {
            for(String o : out)
            {
                pw.print(o);
            }    
            return;
        }
        
        List<StringBuffer> as = new ArrayList<>();
        for(String o : out)
        {
            as.add(new StringBuffer());
        }
        
        for(int k = 0 ; k < fastaTransform.aliLength; k++)
        {
            boolean coding = false;
            for(String o : out)
            {
                if ( o.charAt(k) != '-' && o.charAt(k) != '.' && o.charAt(k) != ':')
                    coding = true;
            }
            if ( coding)
            {
                for(int i = 0 ; i < out.size(); i++)
                {
                    if ( out.get(i).charAt(k) != '-' && out.get(i).charAt(k) != '.' && out.get(i).charAt(k) != ':')
                    {
                        if ( out.get(i).charAt(k)==' ') throw new RuntimeException("White space");
                        as.get(i).append(out.get(i).charAt(k));
                    }else{
                        as.get(i).append("-");
                    }
                }
            }
        }
        for(int i = 0 ; i< as.size(); i++)
        {
            pw.println(">"+ pIds.get(i));
            
            for(int j = 0 ; j < as.get(i).length();j+=80)
            {
                pw.println(as.get(i).substring(j, Math.min(as.get(i).length(),j+80)));
            }
            
        }
        pw.flush();             
    }
    /*
    public void printISAR(PrintWriter pw)
    {
        List<HashSet<Pair<POMAGene, Integer>>> ali = poma.getAlignment();
        
        HashMap<POMAGene,Integer> lastMatch = new HashMap<POMAGene,Integer>();
        HashMap<POMAGene,List<Pair<Integer,Character>>> aliStr = new HashMap<>();
        System.out.println("PREP. PRINT");
        Integer idlength = Integer.MIN_VALUE;
        
        HashMap<POMAGene,Pair<Integer,Integer>> startEnd = new HashMap<>();
        for(POMAGene gene : poma.genes)
        {
            int min = Integer.MAX_VALUE;
            int max = Integer.MIN_VALUE;
            
            idlength = Math.max(idlength, gene.getId().length());
            
            for(String proteinId : gene.getProteins())
            {
                idlength = Math.max(idlength, proteinId.length());
                
                min = Math.min(min,gene.getRegionVector(proteinId).getX1());
                max = Math.max(max,gene.getRegionVector(proteinId).getX2());
                
                
               // aliStr.put(proteinId, new StringBuffer());
                
            }
            startEnd.put(gene, Pair.create(min, max));
            lastMatch.put(gene, min);
            aliStr.put(gene, new ArrayList<>());
            
        }
  
        
        for(HashSet<Pair<POMAGene, Integer>> a : ali)
        {
            if ( a.size() == 1 ) 
                continue;
            HashSet<POMAGene> tmp = new HashSet<>();
            
            for(Pair<POMAGene,Integer> pair : a)
            {
                POMAGene gene = pair.getFirst();
                Pair<Integer, Integer> se = startEnd.get(gene);
                if ( pair.getSecond() < se.getFirst() || pair.getSecond() > se.getSecond())
                    continue;
                
                Integer position = pair.getSecond();
                
                Integer lastPosition = lastMatch.get(gene);
                
                if ( position > (lastPosition+1) )
                {
                    Region1D r = new Region1D(lastPosition,position,null);
                    HashSet<Integer> positions = new HashSet<>();
                    for(String proteinId : gene.getProteins())
                    {
                        for(Region1D exon : gene.getRegionVector(proteinId).getRegions())
                        {
                            for(int i = exon.getX1(); i< exon.getX2(); i++)
                            {
                                if ( r.contains(i))
                                {
                                    positions.add(i); 
                                }
                            }
                        }
                    }
                    List<Integer> sortedPos = new ArrayList<>(positions);
                    Collections.sort(sortedPos);
                    for(Integer p : sortedPos)
                    {
                        aliStr.get(gene).add(Pair.create(p, gene.geneSequence.charAt(p)));
                    }
                    for(Entry<POMAGene, List<Pair<Integer, Character>>> e : aliStr.entrySet())
                    {
                        if (! e.getKey().equals(gene.getId()))
                        {
                            for(Integer p : sortedPos)
                            {
                                aliStr.get(e.getKey()).add(Pair.create((Integer)null, '-'));
                            }
                        }
                    }
                }
                        tmp.add(pair.getFirst());
                
                
                Integer pos = pair.getSecond();
                aliStr.get(gene).add(Pair.create(pos,gene.geneSequence.charAt(pos)));
                lastMatch.put(gene, position);
             
            }
           
            for(POMAGene gene : poma.genes)
            {
                if ( tmp.contains(gene))
                    continue;
                aliStr.get(gene).add(Pair.create((Integer)null, '-'));
            }
            
            
        }
        final String rowFMT = String.format("%%%ds: %%s\n", idlength);
        
        for(Entry<POMAGene, List<Pair<Integer, Character>>> a : aliStr.entrySet())
        {
            StringBuffer str = new StringBuffer();
            for(Pair<Integer, Character> p : a.getValue())
            {
                str.append(p.getSecond());
            }
            pw.printf(rowFMT,a.getKey().getId() , str.toString());
            
            for(String proteinId : a.getKey().getProteins())
            {
                RegionVector rv = a.getKey().getRegionVector(proteinId);
                str = new StringBuffer();
                    
                HashMap<Integer,Character> prot = new HashMap<>();
                int k = 0;
                for(Region1D r : rv.getRegions())
                {
                    String codon = null;
                    Integer lastStart = null;
                    for(int i = r.getX1() ; i< r.getX2(); i++)
                    {
                        if ( k++%3==0  )
                        {
                            if (  lastStart != null)
                            {
                                char as = AminoAcidType.get(codon);
                                prot.put(lastStart, as);
                            }
                            lastStart = i;
                            codon ="";
                        }
                        codon += a.getKey().geneSequence.charAt(i);
                        
                    }
                }
                StringBuffer as = new StringBuffer();
                
                for(Pair<Integer, Character> p : a.getValue())
                {
                    if ( p.getFirst() == null)
                    {
                        as.append("-");
                        str.append("-");
                        continue;
                    }
                    if ( prot.containsKey(p.getFirst()))
                    {
                        as.append(prot.get(p.getFirst()));
                    }else if( prot.containsKey(p.getFirst()-1))
                    {
                        as.append(":");
                    }else if( prot.containsKey(p.getFirst()-2))
                    {
                        as.append(".");
                    }else
                    {
                        as.append("-");
                    }
                    
                    if (p.getFirst() != null && rv.isIn(p.getFirst()) ) 
                    {
                        str.append(p.getSecond());
                    }else
                        str.append("-");
                    
                }   
                pw.printf(rowFMT,proteinId , as.toString());
                pw.printf(rowFMT,proteinId , str.toString());
                
            }
            
            
        }
        pw.flush();
        
    }
    */
    public static void main(String[] args) throws Exception  {
        ISAROptionParser cmdLine = new ISAROptionParser("isar","out_file","out_type","png_scale_factor","intron_size","show_isoform_labels","org_xml","genes","show_gaps");
        cmdLine.setFile("isar");
        cmdLine.setDouble("png_scale_factor");
        cmdLine.setOutFile("out_file");
        cmdLine.setOptional("org_xml");
        cmdLine.setFile("org_xml");
        cmdLine.getOption("isar").setHelp("input ISAR xml.gz file");
        cmdLine.getOption("out_file").setHelp("output file");
        cmdLine.getOption("out_type").setHelp("output type");
        cmdLine.getOption("org_xml").setHelp("organism configuration file");
        cmdLine.getOption("isar").setType("FILE");
        cmdLine.getOption("out_file").setType("FILE");
        cmdLine.getOption("out_type").setType("STR");
        cmdLine.getOption("genes").setType("STR");
        cmdLine.setOptional("genes");
        cmdLine.setInt("intron_size");
        cmdLine.setDefault("intron_size", "30");
        cmdLine.getOption("genes").setHelp("comma seperated list of genes to be included");
        cmdLine.getOption("intron_size").setHelp("size of introns in png");
        cmdLine.getOption("png_scale_factor").setHelp("Scale for output visualization");
        cmdLine.setDefault("png_scale_factor", "1");
        cmdLine.setDouble("png_scale_factor");
        cmdLine.getOption("show_gaps").setHelp("show alignment gaps in exon-intron projected visualization");
        cmdLine.getOption("out_type").setEnumParam(1, "fasta", "png","png_simple","png_event", "isar_mul_ali","report_unmatched");
        cmdLine.setSwitches("show_isoform_labels","show_gaps");

        cmdLine.addToCategory("INPUT/OUTPUT", "isar","out_file","out_type");
        cmdLine.addToCategory("PNG OPTIONS", "png_scale_factor","intron_size","show_isoform_labels","show_gaps");
        
        LogConfig lc = LogConfig.configureLog(null, true);

        ISAR.printISARInfo(System.err);
        
        if (!OptionParser.parseParams(args,true, false, true,  cmdLine))
            return;

        Logger log = LogConfig.getLogger();
        if ( cmdLine.isOptionSet("org_xml"))
        {
            Organism.readOrganisms(cmdLine.getFile("org_xml"));
        }
        String mode = cmdLine.getValue("out_type").toLowerCase();
        
        File out = cmdLine.getFile("out_file");
        File isar = cmdLine.getFile("isar");
        
        System.out.println("Read: " + isar);
     //   LinkedHashSet<String> genes = new LinkedHashSet<String>();
       
        
        PartialOrderMultipleAlignment poma = new PartialOrderMultipleAlignment(isar,mode.equals("report_unmatched")?UnmatchedStrategy.ONLY_EXTENDED:UnmatchedStrategy.ONLY_BASIC,null,false);
        log.info("Read POMA with %d genes",poma.getGenes().size());
        
        List<POMAGene> selectedGenes = new ArrayList<>(poma.getGenes());;
        
        if ( cmdLine.isOptionSet("genes"))
        {
            selectedGenes.clear();
            ObjectGetter.apply(cmdLine.getValue("genes").split(","), (String gene)-> {
                POMAGene pg = poma.getGene(gene);
                if ( pg == null) return;
                selectedGenes.add(pg);
                
            });
            System.out.println("Only consider the following genes:" + selectedGenes);
        }
        if( selectedGenes.size() == 0)
        {
            System.err.println("Abort. No genes selected. Possible genes:"  + poma.getGenes());
            System.exit(-1);
        }
      
        Drawer.SCALE = cmdLine.getDouble("png_scale_factor");
        Drawer.grid = false;
        if(mode.equals("png_simple")){
            BufferedImage bim = Drawer.drawAlignmentRegions(poma,cmdLine.isSet("show_isoform_labels"),false);
            
            ImageIO.write(bim, "png", out);
            System.out.println(out);
            System.exit(1);
        }else if ( mode.equals("png_event"))
        {
            throw new UnsupportedOperationException();
           // Conservation.printEventClassifications(poma, new File("."),cmdLine.getDouble("png_scale_size"));
        }
         
        List<HashSet<Pair<POMAGene, Integer>>> mapping = null;
        
        if ( mode.equals("fasta") || mode.equals("isar_mul_ali") || mode.equals("png"))
        {
            mapping = PartialOrderMultipleAlignment.getCompletMapping(isar, poma,new File(poma.getPOMAID() + ".complete"),true);
        }
        
        PrintWriter pw = FileUtils.getWriter(out); 
        
        if ( mode.equals("fasta"))
        {
            Formater formater = new Formater(poma,mapping);
            formater.printISAR(pw,true,selectedGenes,0);
            
            
        }else if ( mode.equals("isar_mul_ali"))
        {
            Formater formater = new Formater(poma,mapping);
            formater.printISAR(pw,false,selectedGenes,cmdLine.getInt("intron_size"));
        }else if ( mode.equals("png"))
        {
            BufferedImage bim = Drawer.drawAlignmentComplete(poma, mapping, cmdLine.isSet("show_isoform_labels"),false,cmdLine.isSet("show_gaps"),cmdLine.getInt("intron_size"),selectedGenes);
            ImageIO.write(bim, "png", out);
        }else  if ( mode.equals("report_unmatched"))
        {
            poma.reportUnmatched(pw,false);
        } else {
            System.err.println("Invalid mode:" + mode);
        }
        pw.close();
        System.out.println(out + " written");
        System.exit(1);

    }
}
