package isoformtransfer.isar.oracle;

import java.util.ArrayList;
import java.util.List;

import isoformtransfer.isar.POMAGene;
import isoformtransfer.isar.PartialOrderMultipleAlignment.UnmappedInformation;

public class MultiplePairwiseOracle implements Oracle {
    
    MultipleAlignmentOracle mulOracle;
    PairwiseAlignmentOracle pwOracle;
    boolean performedMultipleAlignment = true;
    
    public MultiplePairwiseOracle(MultipleAlignmentOracle mulOracle, PairwiseAlignmentOracle pwOracle)
    {
        this.mulOracle = mulOracle;
        this.pwOracle = pwOracle;
    }

    
    @Override
    public List<Suggestion> generateMappings(List<POMAGene> genes,    UnmappedInformation unmappedInformation) {
        List<Suggestion> ret = new ArrayList<>();
        
        if ( mulOracle != null && performedMultipleAlignment ){
            ret.addAll(mulOracle.generateMappings(genes, unmappedInformation));
            performedMultipleAlignment = false;
        }
        
        if ( pwOracle != null)
            ret.addAll(pwOracle.generateMappings(genes, unmappedInformation));
        

        
        return ret;
    }

    @Override
    public String getName() {
        if ( pwOracle != null && mulOracle != null)
            return "Multiple and pair-wise alignment oracle";
        if ( pwOracle != null)
            return pwOracle.getName();
        if ( mulOracle != null){
            return mulOracle.getName();
        }
        return "";
    }


    @Override
    public boolean isRecursiveOracle() {
        return pwOracle!=null?true:false;
    }

}
