package isoformtransfer.isar.tools;

import java.io.File;
import java.util.Vector;

import genome.Organism;
import isoformtransfer.algo.Spaln2;
import isoformtransfer.algo.TransferAlgo;
import isoformtransfer.isar.PartialOrderMultipleAlignment;
import isoformtransfer.isar.oracle.PairwiseAlignmentOracle;
import isoformtransfer.isar.oracle.SplicedAlignmentOracle;
import lmu.utils.FileUtils;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.OptionParser;
import lmu.utils.SimpleOptionParser;

public class CreateComplete {
    public static void main(String[] args) throws Exception
    {
        SimpleOptionParser cmdLine = new SimpleOptionParser("jobs","out_dir","spaln2","organism");
        cmdLine.setFile("jobs","organism","spaln2");
        cmdLine.setDir("out_dir");
        
        LogConfig lc = LogConfig.configureLog(null, true);

        if (!OptionParser.parseParams(args,true, false, true,  cmdLine))
            return;

        Logger log = LogConfig.getLogger();

        Organism.readOrganisms(cmdLine.getFile("organism"));

        Vector<String> lines = FileUtils.getLines(cmdLine.getFile("jobs"));
        
        File spaln2 = cmdLine.getFile("spaln2");
        
        for(String line : lines)
        {
            System.out.println(line);
            File pomaXML = new File(line);
            
            PartialOrderMultipleAlignment poma = new PartialOrderMultipleAlignment(pomaXML);
            File extPOMA = new File(cmdLine.getFile("out_dir") + "/" + poma.getPOMAID() + "ext.xml.gz");
            if ( extPOMA.exists())
            {
                System.out.println(extPOMA + " already exists.");
                continue;
            }
            ext(poma,spaln2);
            
            
            poma.writePOMA(extPOMA);
            
            PartialOrderMultipleAlignment.getCompletMapping(extPOMA, poma);
        }
        
    }
    private static void ext(PartialOrderMultipleAlignment poma, File spaln2) {
        poma.updateUnMappedRegions(false);
        PairwiseAlignmentOracle pao = new PairwiseAlignmentOracle();
        
        poma.setMIN_SEQ_ID( 0);
        poma.setMIN_BP_IN_REGION(0);
        poma.setMIN_COVERAGE(0);
        poma.setMIN_MATCHES(0);
        poma.applyOracle(pao);
        
        TransferAlgo algo = new Spaln2(spaln2);
        SplicedAlignmentOracle sp = new SplicedAlignmentOracle(1,algo);
        poma.updateUnMappedRegions(false);
         
        poma.applyOracle(sp);
        
    }
}
