package isoformtransfer.isar.tools;

import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import genome.Gene;
import genome.Organism;
import lmu.utils.FileUtils;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.OptionParser;
import lmu.utils.SimpleOptionParser;
import lmu.utils.StringUtils;
import lmu.utils.ThreadUtils;
import ortholog.OrthologGeneGroup;

public class CreateInput {
    public static void main(String[] args) throws Exception
    {
        SimpleOptionParser cmdLine = new SimpleOptionParser("organism","out_dir","in_dir","isars","threads");
        cmdLine.setFile("organism");
        cmdLine.setDir("out_dir","in_dir");
        cmdLine.setInt("threads");
        cmdLine.setDefault("threads", "4");
        cmdLine.setOutFile("isars");
        LogConfig lc = LogConfig.configureLog(null, true);

        if (!OptionParser.parseParams(args,true, false, true,  cmdLine))
            return;

        Logger log = LogConfig.getLogger();

        Organism.readOrganisms(cmdLine.getFile("organism"));

        PrintWriter stat = FileUtils.getWriter(cmdLine.getFile("isars"));

        List<File> files = new ArrayList<>(Arrays.asList(cmdLine.getFile("in_dir").listFiles()));

        for(Organism o : Organism.getAllOrganisms())
        {
            o.getAnnot();
            o.getGenome();
        }

        ThreadUtils.runLambdas(files, (File f)->
        {
            if ( !f.getName().endsWith(".xml"))
                return;

            OrthologGeneGroup ogg = new OrthologGeneGroup(f);

            List<String> names = new ArrayList<>();

            for(Gene gene : ogg.getRelevantGenes())
            {
                String name = gene.id + ".fasta";
                names.add(name);
                PrintWriter pw = FileUtils.getWriter(cmdLine.getFile("out_dir") + "/" + name);
                InputGenerator.writeISAROutput(gene, gene.getOrganism().getGenome(), gene.getOrganism().getTaxId(),true, pw);
                pw.close();
            }

            synchronized(stat)
            {
                stat.println(StringUtils.join(",", names));
                stat.flush();   
            }
        },cmdLine.getInt("threads"));
        stat.close();
    }
}
