package isoformtransfer.isar.tools;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

import org.jdom.Element;

import genome.Gene;
import genome.Organism;
import lmu.utils.FileUtils;
import lmu.utils.LogConfig;
import lmu.utils.OptionParser;
import lmu.utils.SimpleOptionParser;
import lmu.utils.StringUtils;
import lmu.utils.XMLUtils;

public class OGGConvert {
    public static void main(String[] args) throws IOException{
        SimpleOptionParser cmdLine = new SimpleOptionParser();
        
        cmdLine.addSimpleOptions("orga_xml","ogg_dir","out_dir","threads","isar_in_summary");
        cmdLine.setDir("ogg_dir");
        cmdLine.setDir("out_dir");
        cmdLine.setFile("orga_xml");
        cmdLine.setInt("threads");
        cmdLine.setOutFile("isar_in_summary");
        
        cmdLine.setDefault("threads", "4");
        
        LogConfig lc = LogConfig.configureLog(null, true);

        if (!OptionParser.parseParams(args,true, false, true,  cmdLine,lc))
            return;
        
        Organism.readOrganisms(cmdLine.getFile("orga_xml"));
        
        for(Organism o : Organism.getAllOrganisms())
        {
            o.getAnnot();
            o.getGenome();
            
            File f = new File(cmdLine.getFile("out_dir") + "/" + o.getTaxId() );
            f.mkdirs();
        }
        
        PrintWriter summary = FileUtils.getWriter("isar_in_summary");
        
        for(File f : cmdLine.getFile("ogg_dir").listFiles())
        {
            
            if ( !f.getName().endsWith(".xml"))
                continue;
            
      
            Element root = XMLUtils.getRootElement(f);
            List<Element> genes = root.getChildren("gene");
            
            //OrthologGeneGroup ogg = new OrthologGeneGroup(f);
            
            List<File> files = new ArrayList<>();
            for(Element g : genes)
            {
                String geneId = g.getAttributeValue("id");
                Integer taxId= Integer.valueOf(g.getAttributeValue("tax"));
                
                File out = new File(cmdLine.getFile("out_dir") + "/"+ taxId + "/" + geneId  + ".fasta.gz");
                files.add(out);
                
                if ( out.exists())
                    continue;
                Organism o = Organism.getOrganism(taxId);
                
                PrintWriter pw = FileUtils.getWriter(out);
                Gene gene = o.getAnnot().getGene(geneId);
                
                InputGenerator.writeISAROutput(gene, o.getGenome(),taxId, true, pw);
                
                pw.close();
            }
            summary.println(StringUtils.join(",", files, (File file) ->  file.toString().replaceAll(cmdLine.getFile("out_dir").toString() + "/","")));
            
        }
        summary.close();
    }
}
