package lmu.utils;

import java.io.*;
import java.util.*;

import javax.xml.stream.*;

import lmu.utils.index.BitSetIndex;


public class AutoResizeBitSet
{
    
    BitSet bs = null;
    int alloc_step=10000;

    static final String BS_SUFFIX = ".bsindex";

    public static File getBSIndex(File xml)
    {
        return new File(xml.getAbsolutePath() + BS_SUFFIX);
    }

    public interface BitSetInfoGetter
    {
        public int getCardinality();
        public BitSet getBitSet();
        public int updateWithFilter(BitSet bs);
    }

    public static class BitSetXML
    {
        
        int n = 0;
        HashMap<Tuple, IndexedBitSetContainer> tuple2bs =  new HashMap<>();

        public BitSetXML(int n)
        {
            this.n = n;
        }

        public void add(BitSet bs, String ... names)
        {
            Tuple k = new Tuple(names.length);
            for (int i=0; i< names.length; i++)
            {
                k.set(i, names[i]);
            }
            tuple2bs.put(k, new IndexedBitSetContainer(bs));
            System.out.printf("will add %s: %d\n", k, bs.cardinality());
        }

        public void write(File xml)
        {
            write(xml, getBSIndex(xml));
        }

        public String getInfo()
        {
            BufferPrintWriter bpw = new BufferPrintWriter();
            bpw.printf("indexed bitset base size: %d\n", n);
            double norm = 100.0 / n;
            for (Map.Entry<Tuple, IndexedBitSetContainer> e : tuple2bs.entrySet())
            {
                int c = e.getValue().getBitSet().cardinality();

                bpw.printf("\t%s: %d (%2.2f%%)\n", e.getKey(), c, c * norm);
            }
            return bpw.getBuffer();
        }
        public void write(File xml, File bs_indexfile)
        {
            try
            {
                BitSetIndex bsi = new BitSetIndex(bs_indexfile);
                XMLStreamWriter xsw = XMLUtils.getWriter(xml);
                xsw.writeStartDocument();
                xsw.writeStartElement("bsxml");
                xsw.writeAttribute("n", "" + n);

                int bsi_index = 0;
                for (Map.Entry<Tuple, IndexedBitSetContainer> e : tuple2bs.entrySet())
                {
                    IndexedBitSetContainer ibc = e.getValue();
                    
                    xsw.writeStartElement("bitset");
                    for (int i=0; i<e.getKey().cardinality(); i++)
                    {
                        xsw.writeEmptyElement("n");
                        xsw.writeAttribute("v", ""+e.getKey().get(i));
                    }
                    bsi_index = writeEncoded(xsw, ibc.getBitSet(), bsi, "data", bsi_index, 0, "");

                    xsw.writeEndElement();


                    //ibc.writeEncoded(xsw, BitSet bs, BitSetIndex bsi, String elem_name, int bsi_index, int value_index, String typename) throws XMLStreamException
                    //ibc.
                }
                xsw.writeEndElement();
                xsw.writeEndDocument();
                xsw.close();
            }
            catch(XMLStreamException xse)
            {
                throw new FRuntimeException("xml error:%s while writing bs-index xml: %s", xse.getMessage(), xml.getAbsolutePath());
            }
        }

        public BitSetXML(File xml)
        {
            this(xml, new File(xml.getAbsolutePath() + BS_SUFFIX));
        }

        public BitSetXML(File xml, File bs_indexfile)
        {
            BitSetIndex bsi = (bs_indexfile == null) ? null : new BitSetIndex(bs_indexfile);
        
            boolean rootelem = true;
            try
            {
                XMLStreamReader xsr = XMLUtils.getReader(FileUtils.getReader(xml));

                while (xsr.hasNext())
                {
                    int event = xsr.next();
        
                    if (event != XMLStreamConstants.START_ELEMENT)
                        continue;
        
                    if (rootelem)
                    {
                        this.n = XMLUtils.getIntAttrib(xsr, "n", null);
                        rootelem = false;
                    }
                    if (!xsr.getLocalName().equals("bitset"))
                    {
                        throw new FRuntimeException("unexcepted entry in bitsetxml: %s", xsr.getLocalName());
                    }



                    org.jdom.Element e = XMLUtils.readElemFromStream(xsr);
                    
                    LinkedList<org.jdom.Element> names = XMLUtils.getChildren(e, "n");
                    Tuple key = new Tuple(names.size());

                    for (int i=0; i<names.size(); i++)
                    {
                        key.set(i, XMLUtils.getAttrib(names.get(i), "value", null, true));
                    }

                    org.jdom.Element bse = e.getChild("data");
                    int num_n = XMLUtils.getIntAttrib(bse, "value", null);
                    Integer count = XMLUtils.getIntAttrib(bse, "count", null);
                    Integer bs_index = XMLUtils.getIntAttrib(bse, "bs_index", null);
                    
                    IndexedBitSetContainer ibs = new IndexedBitSetContainer(count, bsi, bs_index);
                    
                    tuple2bs.put(key, ibs);
                }
            }
            catch(XMLStreamException xse)
            {
                throw new FRuntimeException("xml error: %s while reading from %s", xse, xse.getMessage(), xml);
            }
        
        }
        

    }

    
    public static class IndexedBitSetContainer implements BitSetInfoGetter
    {
        BitSetIndex index;
        int bs_index;
        int cardinality = 0;
        int filtered_cardinality = 0;

        BitSet cached = null;

        public IndexedBitSetContainer(BitSet bs)
        {
            this.cached = bs;
            cardinality = bs.cardinality();
        }
        
        public IndexedBitSetContainer(int cardinality, BitSetIndex index, int bs_index)
        {
            this.cardinality = cardinality;
            this.index = index;
            this.bs_index = bs_index;
        }

        public int getCardinality()
        {
            return cardinality;
        }

        public int updateWithFilter(BitSet bs)
        {
            return filtered_cardinality = SetInfo.intersect(bs, getBitSet()).cardinality();
        }
        public BitSet getBitSet()
        {
            if (cached != null)
                return cached;
            
            return index.read(bs_index);
        }
    }

    public AutoResizeBitSet()
    {
        this(0, 10000);
    }
    public AutoResizeBitSet(int init_size, int alloc_step)
    {
        bs=new BitSet(init_size);
        this.alloc_step=alloc_step;
        
    }

    public AutoResizeBitSet(BitSet bs, int alloc_step)
    {
        this.bs = bs;
        this.alloc_step = alloc_step;
    }

    public int length()
    {
        return (bs == null) ? 0 : bs.length();
    }

    public void setBitSet(BitSet bs)
    {
        this.bs = bs;
    }
    
    public BitSet getBitSet()
    {
        return bs;
    }

    public int cardinality()
    {
        return (bs == null) ? 0 : bs.cardinality();
    }
    public AutoResizeBitSet(Collection<Integer> col)
    {
        int maxx=0;
        for (int i: col)
        {
            maxx=Math.max(i,maxx);
        }
        bs=new BitSet(maxx+1);
        set(col);
    }

    public Vector<Integer> filter(Collection<Integer> col)
    {
        Vector<Integer> rv = new Vector<Integer>();
        for (int i : col)
        {
            if(!isSet(i))
                continue;

            rv.add(i);
        }
        return rv;
    }
    public void readFromFile(File f)
    {
        //excepts one int per line
        String line=null;
        int ln=0;
        try
        {
            BufferedReader bfr = new BufferedReader(new FileReader(f));

            
            while (null!=(line=bfr.readLine()))
            {
                set(Integer.parseInt(line));
                ln++;
            }
            bfr.close();
        }
        catch(IOException e)
        {
            throw new RuntimeException("error: "+e.getMessage()+" at setting bits from file: "+f.getAbsolutePath()+" at line: "+ln+" last line: "+line,e);
        }
        
    }
    public int size()
    {
        return (bs==null)?0:bs.size();
    }
    public boolean isSet(int i)
    {
        if (bs==null || i<0 || i>=bs.size())
            return false;
        
        return bs.get(i);
    }
    public void set(Collection<Integer> toset)
    {
        if (toset==null)
            return;
        
        
        for (int i: toset)
        {
            set(i);
        }
    }
    public void set(int i)
    {
        if (bs==null || bs.size()<=i)
        {
            int nsize=(int)((1+(i/(double)alloc_step))*alloc_step);
            bs = extend(bs, nsize);
        }
        bs.set(i);
    }

    public static BitSet toBitSet(int maxsize, int[] c)
    {
        return toBitSet(maxsize, CollectionIterator.getIterator(c));
    }

    public static BitSet toBitSet(int maxsize, Collection<Integer> c)
    {
        return toBitSet(maxsize, c.iterator());
    }

    public static BitSet toBitSet(int maxsize, Iterator<Integer> it)
    {
        BitSet bs = new BitSet(maxsize);
        ObjectGetter.apply(it, (i) -> bs.set(i));
        return bs;
    }

    /*
     * all the same, but Gergely wants both names
     */
    

    public static Iterator<Integer> toIndexIterator(BitSet bs)
    {
        return new BitSetIndexOperator(bs);
    }

    public static Iterator<Integer> toIndexIterator(BitSet bs, int length)
    {
        return new BitSetIndexOperator(bs, false, length);
    }
    public static Iterator<Integer> toIndexIterator(BitSet bs, boolean invert, int length)
    {
        return new BitSetIndexOperator(bs, invert, length);
    }
    
    public static Iterator<Integer> iterator(BitSet bs)
    {
        return new BitSetIndexOperator(bs);
    }

    public static Iterator<Integer> iterator(BitSet bs, boolean invert, int length)
    {
        return new BitSetIndexOperator(bs, invert, length);
    }
    
    
    public static BitSet extend(BitSet old, int n)
    {
        BitSet nbs = new BitSet(n);
        if (old ==null)
            return nbs;
        
        ObjectGetter.apply(iterator(old), (Integer idx) -> nbs.set(idx));
        return nbs;
    }
    static class BitSetIndexOperator implements Iterator<Integer>
    {
        BitSet bs;
        int next_set = -1;
        boolean invert = false;
        int N = -1;
        int length;

        BitSetIndexOperator(BitSet bs)
        {
            this(bs, false, -1);
        }
        BitSetIndexOperator(BitSet bs, boolean invert, int length)
        {
            this.bs = bs;
            N = (length < 0) ? bs.size() : length;
            next_set = (invert) ? bs.nextClearBit(0) : bs.nextSetBit(0);
            next_set = (next_set >= N) ? -1 : next_set;
            
            this.invert = invert;
        }

        public boolean hasNext()
        {
            return next_set >= 0;
        }

        public Integer next()
        {
            int rv = next_set;
            if (next_set >= 0)
            {
                next_set = (invert) ? bs.nextClearBit(next_set+1) : bs.nextSetBit(next_set+1);
                next_set = (next_set >= N) ? -1 : next_set;
                
            }
            //System.out.printf("next bs: %d\n", rv);
            return rv;
        }
        
    }


    static public Vector<Integer> getAutoCumulativeCardinality(Vector<AutoResizeBitSet> bsets, boolean up)
    {
        return  getCumulativeCardinality(toBitSets(bsets), up);
    }


    static void invert(BitSet bs)
    {
        for (int i=0; i<bs.size(); i++)
        {
            bs.flip(i);
        }
    }
    public static int DIRECT_BS_FLAG = 1;
    public static int CUM_UP_BS_FLAG = 2;
    public static int CUM_DOWN_BS_FLAG = 4;
    public static int ALL_BS_FLAG = DIRECT_BS_FLAG + CUM_UP_BS_FLAG + CUM_DOWN_BS_FLAG;

    static int writeEncoded(XMLStreamWriter xsw, BitSet bs, BitSetIndex bsi, String elem_name, int bsi_index, int value_index, String typename) throws XMLStreamException
    {
        Logger log = LogConfig.getLogger();

        if (bs == null)
            return bsi_index;

        if (bsi == null)
        {
            xsw.writeStartElement(elem_name);
        }
        else
        {
            xsw.writeEmptyElement(elem_name);
        }


        log.info("will write %s:%d to idx: %s type: %s count: %d/%d (%2.2f%%)\n", elem_name, value_index, bsi_index, typename, bs.cardinality(), bs.size(), (100.0 * bs.cardinality()) / bs.size());
        xsw.writeAttribute("value", ""+value_index);
        xsw.writeAttribute("type", typename);
        xsw.writeAttribute("count", ""+bs.cardinality());
        

        if (bsi == null)
        {
            XMLUtils.writeEncodedCData(xsw, bs);
            xsw.writeEndElement();
            return bsi_index;
        }
            
        xsw.writeAttribute("bs_index", ""+bsi_index);
        bsi.write(bs, bsi_index);
        
        return bsi_index + 1;
        
    }

    static public int writeBitSetData(XMLStreamWriter xsw, String elem_name, Vector<BitSet> bitsets, BitSetIndex bsi, int bsi_index, int WRITE_BITSET_FLAGS)
    {
        
        int rv = bsi_index;
        boolean cum_up = (WRITE_BITSET_FLAGS & CUM_UP_BS_FLAG) > 0;
        boolean cum_down = (WRITE_BITSET_FLAGS & CUM_DOWN_BS_FLAG) > 0;
        boolean w_direct = (WRITE_BITSET_FLAGS & DIRECT_BS_FLAG) > 0;

        Pair<Vector<Integer>, Vector<BitSet>> p1 = (!cum_up) ? null: AutoResizeBitSet.getCumC(bitsets, true);
        

        Pair<Vector<Integer>, Vector<BitSet>> p2 = (!cum_down) ? null : AutoResizeBitSet.getCumC(bitsets, false);

        //Vector<Integer> cumulative_down = AutoResizeBitSet.getCumulativeCardinality(bitsets, false);


        
        try
        {
            for (int i=0; i<bitsets.size(); i++)
            {
                if (w_direct)
                {
                    rv = writeEncoded(xsw, bitsets.get(i), bsi, elem_name, rv, i, "direct");
                }
                
                if (cum_up)
                {
                    rv = writeEncoded(xsw, p1.getSecond().get(i), bsi, elem_name, rv, i, "cumup");
                }

                if (cum_down)
                {
                    rv = writeEncoded(xsw, p2.getSecond().get(i), bsi, elem_name, rv, i, "cumdown");
                }
                
            }
        }
        catch(XMLStreamException xse)
        {
            throw new FRuntimeException("xml error: %s while writing %d %s elems with bitset indeces", xse, xse.getMessage(), bitsets.size(), elem_name);
        }

        return rv;
    }
        
    
    public static Pair<Vector<Integer>, Vector<BitSet>> getCumC(Vector<BitSet> bsets, boolean up)
    {
        Vector<Integer> rv = new Vector<>();
        Vector<BitSet> rv_bsets = new Vector<>();

        int start = (up) ? 0 : bsets.size() -1;
        int end = (up) ? bsets.size() : -1;
        int inc = (up) ? 1 : -1;

        BitSet last = null;

        for (int i = start; i != end; i += inc)
        {
            BitSet current = bsets.get(i);
            int value = 0;
            if (last == null)
            {
                last = current;
                value = (current == null) ? 0 : current.cardinality();
                
            }
            else
            {
                if (current == null)
                {
                    value = (last == null) ? 0 : last.cardinality();
                }
                else
                {
                    last = SetInfo.union(last, current);
                    value = last.cardinality();
                }
            }
            if (last == null)
            {
                rv_bsets.add(new BitSet(1));
            }
            else
            {
                rv_bsets.add(last);
            }
            
            rv.add(value);
        }

        return Pair.create(rv, rv_bsets);
    }

    static public Vector<Integer> getCumulativeCardinality(Vector<BitSet> bsets, boolean up)
    {
        return getCumC(bsets,up).getFirst();
    }
    
    static public Vector<BitSet> getCumulativeBitSets(Vector<BitSet> bsets, boolean up)
    {
        return getCumC(bsets,up).getSecond();
    }

    public static Vector<BitSet> toBitSets(Vector<AutoResizeBitSet> v)
    {
        return ObjectGetter.map(v, (AutoResizeBitSet arb) -> (arb == null) ? null : arb.getBitSet());
    }

    /** returns the relevant bitsets */
    public static Vector<BitSet> getRelevantBitSet(int n, Vector<BitSet> direct_sets, double max_level, boolean up)
    {
      int maxlevel = (int)(max_level * n);
    
      BitSet full = AutoResizeBitSet.getFullBS(n);
      if (direct_sets.size() == 0)
          return ObjectGetter.toVector(full);
      
      
      Vector<BitSet> v = AutoResizeBitSet.getCumulativeBitSets(direct_sets, up);
    
      int sumbs_idx = (up) ? v.size() -1 : 0;
      full = SetInfo.minus(full, v.get(sumbs_idx));
    
      Vector<BitSet> rv = new Vector<BitSet>();
      if (up)
      {
          rv.add(full);
      }
      
    
      BitSet current = full;
    
      final int end = (up) ? direct_sets.size() : -1;
      final int step = (up) ? 1 : -1;
      for (int i= (up) ? 0 : direct_sets.size() -1;   i != end; i += step)
      {
          current = (direct_sets.get(i) == null) ? current : SetInfo.union(current, direct_sets.get(i));
          if (current.cardinality() >= maxlevel)
              break;
    
          rv.add(current);
      }
      if (!up)
      {
          Collections.reverse(rv);
          rv.insertElementAt(full, 0);
      }

      return rv;
    }

    public static BitSet sortedInttoBitSet(Vector<Integer> v)
    {
        if (v.size() == 0)
            return new BitSet(1);
        
        BitSet rv = new BitSet(v.get(v.size()-1)+1);
        ObjectGetter.apply(v, (Integer idx) -> rv.set(idx));
        return rv;
    }

    public static BitSet getFullBS(int size)
    {
        BitSet bs = new BitSet(size);
        for (int i = 0; i< size; i++)
        {
            bs.set(i);
        }
        return bs;
    }
    public static void main (String args[]) 
    {
        /*
        long t1 = System.currentTimeMillis();
        AutoResizeBitSet arbs = new AutoResizeBitSet(1000000, 1000000);
        arbs.readFromFile(new File(args[0]));
        long t2 = System.currentTimeMillis();
        System.out.printf("read in took: %2.3f s\n", (t2-t1)/1000.0);
        */
        /*
        BitSet bs = new BitSet(200);
        bs.set(9);
        bs.set(11);
        System.out.println(ObjectGetter.toVector(iterator(bs)));

        if (true)
            return;
        
        
        Vector<BitSet> v = new Vector<>();
        Random rnd = new Random(); 
        for (int i=0; i<10; i++)
        {
            bs = new BitSet(10);
            for (int j=0; j<5; j++)
            {
                bs.set(rnd.nextInt(10));
            }
            System.out.printf("%d: %s\n", i, bs);
            
            v.add(bs);
        }
        System.out.printf("cumulative up: %s\n", getCumulativeCardinality(v, true));
        System.out.printf("cumulative down: %s\n", getCumulativeCardinality(v, false));
        */
    }
}

