package lmu.utils;

import java.io.*;


public class BufferedRandomAccessFile extends RandomAccessFile 
{
    Logger log = LogConfig.getLogger();
    File file;
    private byte buffer[];
    private int bufferSize;
    private int bufEnd;
    private int bufPos;
    private long realPos;

    public File getFile()
    {
        return file;
    }
    public class RReader extends java.io.Reader
    {
        byte[] bdata = new byte[2048];
        BufferedRandomAccessFile barf;

        RReader(BufferedRandomAccessFile barf)
        {
            this.barf=barf;
        }
        public int read(char cbuf[], int off, int len) throws IOException
        {
            
            //log.info(String.format("request: cbuff %d off: %d len: %d\n", cbuf.length, off, len));
            //int rl=len-off;
            int rl=len;
            if (rl>bdata.length)
            {
                bdata=new byte[rl];
            }
            //System.out.printf("> RR1: off: %d len: %d rl: %d bdata: %d\n", off, len, rl, bdata.length);
            int n=barf.read(bdata, 0, rl);
            //System.out.printf("> RR: off: %d len: %d rl: %d bdata: %d n: %d\n", off, len, rl, bdata.length, n);
            for (int i=0,j=off; i<n ; j++, i++)
            {
                cbuf[j]=(char)bdata[i];
            }
            //log.info("return: "+n);
            return (n==0)?-1:n;
        }
        /**
         * Closes the stream and releases any system resources associated with
         * it.  Once the stream has been closed, further read(), ready(),
         * mark(), reset(), or skip() invocations will throw an IOException.
         * Closing a previously closed stream has no effect.
         *
         * @exception  IOException  If an I/O error occurs
         */
        public void close() throws IOException
        {
            barf.close();
        }
    }
    
    public java.io.Reader getReader()
    {
        return new RReader(this);
    }
    public BufferedRandomAccessFile(File file, String mode, int bufferSize) throws IOException {
        super(file, mode);
        this.file=file;
        invalidate();
        this.bufferSize = bufferSize;
        this.buffer = new byte[bufferSize];
    }

    public String readString(int n) throws IOException
    {
        byte[] hdata = new byte[n];
        readFully(hdata);
        BufferedReader brf = new BufferedReader(new InputStreamReader(new ByteArrayInputStream(hdata)));
        String line = null;
        StringBuffer sb = new StringBuffer();
        while (null!=(line=brf.readLine()))
        {
            if (sb.length()!=0)
            {
                sb.append('\n');
            }
            sb.append(line);
        }
        brf.close();
        return sb.toString();
    }
    public final int read() throws IOException {
        if (this.bufPos >= this.bufEnd)
        {
            if (fillBuffer() < 0)
                return -1;
        }

        if (this.bufEnd == 0)
            return -1;

        else
            return this.buffer[this.bufPos++];
    }

    private int fillBuffer() throws IOException {
        int n = super.read(this.buffer, 0, this.bufferSize);
        if (n >= 0)
        {
            this.realPos +=n;
            this.bufEnd = n;
            this.bufPos = 0;
        }
        return n;
    }

    private void invalidate() throws IOException {
        this.bufEnd = 0;
        this.bufPos = 0;
        this.realPos = super.getFilePointer();
    }

    public int read(byte b[], int off, int length) throws IOException 
    {

        //log.info(String.format("request: cbuff %d off: %d len: %d\n", b.length, off, length));

        int leftover = this.bufEnd - this.bufPos;
        if (length <= leftover)
        {
            //System.out.printf("got buffer l: %d pos: %d b: %d off: %d lenght: %d\n", buffer.length, bufPos, b.length, off, length);
            System.arraycopy(this.buffer, this.bufPos, b, off, length);
            this.bufPos += length;
            return length;
        }
        for (int i = 0; i < length; i++)
        {
            int c = this.read();
            if (c != -1)
                b[off+i] = (byte)c;
            else
            {
                return i;
            }
        }
        return length;
    }

    public long getFilePointer() throws IOException{
        long l = this.realPos;
        return(l - this.bufEnd + this.bufPos) ;
    }

    public void seek(long pos) throws IOException {
        int n = (int)(this.realPos - pos);
        if (n >= 0 && n <= this.bufEnd)
        {
            this.bufPos = this.bufEnd - n;
        }
        else
        {
            super.seek(pos);
            invalidate();
        }
    }

    @SuppressWarnings("deprecation")
    public final String getNextLine() throws IOException {
        String str = null;
        if (this.bufEnd - this.bufPos <= 0)
        {
            if (fillBuffer() < 0)
            {
                return null;
            }
        }
        int lineEnd = -1;
        for (int i = this.bufPos; i < this.bufEnd; i++)
        {
            if (buffer[i] == '\n')
            {
                lineEnd = i;
                break;
            }
        }
        if (lineEnd < 0)
        {
            StringBuilder input = new StringBuilder(256);
            int c;
            while (((c = read()) != -1) && (c != '\n'))
            {
                input.append((char)c);
            }
            if ((c == -1) && (input.length() == 0))
            {
                return null;
            }
            return input.toString();
        }
        if (lineEnd > 0 && this.buffer[lineEnd-1] == '\r')
            str = new String(this.buffer, 0, this.bufPos, lineEnd - this.bufPos -1);
        else str = new String(this.buffer, 0, this.bufPos, lineEnd - this.bufPos);
        this.bufPos = lineEnd + 1;
        return str;
    }


}
