package lmu.utils;

import java.io.*;
import java.nio.*;
import java.util.zip.*;

public class ByteDataReader
{
    byte[] data;
    int pos = 0;

    public ByteDataReader(byte[] d)
    {
        this.data = d;
    }

    public void setPos(int p)
    {
        pos = p;
    }

    public int readInt()
    {
        ByteBuffer buff = ByteBuffer.wrap(data, pos, 4);
        pos += 4;
        return buff.getInt();
    }

    public boolean hasNext()
    {
        return pos < data.length;
    }

    public double readDouble()
    {
        ByteBuffer buff = ByteBuffer.wrap(data, pos, 8);
        pos += 8;
        return buff.getDouble();
    }

    public long readLong()
    {
        ByteBuffer buff = ByteBuffer.wrap(data, pos, 8);
        pos += 8;
        return buff.getLong();
    }

    public int getPos()
    {
        return pos;
    }

    public String readString(int data_length, boolean gzipped)
    {
        StringBuffer sb = new StringBuffer();
        try
        {
            InputStream is = new ByteArrayInputStream(data, pos, data_length);
            if (gzipped)
            {
                is = new GZIPInputStream(is);
            }
            
            InputStreamReader isr = new InputStreamReader(is);

            int nc = -1;
            while (-1 != (nc = isr.read()))
            {
                char ch = (char)nc;
                sb.append(ch);
            }

            pos += data_length;
        }
        catch (IOException ie)
        {
            throw new FRuntimeException("i/o error: %s while reading string of length %d", ie, ie.getMessage(), data_length);
        }
        return sb.toString();
    }
}

