package lmu.utils;
import java.util.*;

public  class CollectionIterator<A> implements Iterator<A>
{
    Collection<A> c;
    A[] v;
    
    Iterator<A> it=null;
    int index=-1;
    int endindex=-1;

    
    public Set<A> toSet()
    {
        HashSet<A> rv = new HashSet<A>();
        while (hasNext())
        {
            rv.add(next());
        }
        return rv;
    }

    public CollectionIterator(Iterator<A> it)
    {
        this.it=it;
    }
    public CollectionIterator(Collection<A> c)
    {
        this.c=c;
        if (c==null)
        {
            it = new SingleIterator<A>(null);
            return;
        }
        it=c.iterator();
    }
    

    public static abstract class IndexedIterator<A> implements Iterator<A>
    {
        int index=-1;
        int length=0;
        
        IndexedIterator(int l)
        {
            length=l; index=0;

        }

        public void setIndex(int idx)
        {
            index = idx;
        }

        public boolean hasNext()
        {
            return index>=0 && index<length;
        }
        public void remove()
        {
            throw new RuntimeException("cannot remove on array");
        }
    }

    public static class IndexedVectorIterator<A> extends IndexedIterator<A>
    {
        Vector<A> v;

        public IndexedVectorIterator(Vector<A> c, int pos)
        {
            super(c.size());
            this.v = c;
            setIndex(pos);
        }

        public A next()
        {
            return v.get(index++);
        }
    }

    static class IntArrayIterator extends IndexedIterator<Integer>
    {
        int [] data;
        IntArrayIterator(int [] data)
        {
            super(data.length);
            this.data=data;
        }
        public Integer next()
        {
            return data[index++];
        }
        
    }
    
    public static class DoubleArrayIterator extends IndexedIterator<Double>
    {
        double [] data;
        public DoubleArrayIterator(double [] data)
        {
            super(data.length);
            this.data=data;
        }
        public Double next()
        {
            return data[index++];
        }
        
    }

    static class LongArrayIterator extends IndexedIterator<Long>
    {
        long [] data;
        LongArrayIterator(long [] data)
        {
            super(data.length);
            this.data=data;
        }
        public Long next()
        {
            return data[index++];
        }
        
    }

    static class FloatArrayIterator extends IndexedIterator<Float>
    {
        float [] data;
        FloatArrayIterator(float [] data)
        {
            super(data.length);
            this.data=data;
        }
        public Float next()
        {
            return data[index++];
        }
        
    }


    public static Iterator<Integer> getIterator(int[] v)
    {
        return new IntArrayIterator(v);
    }

    public static Iterator<Double> getIterator(double[] v)
    {
        return new DoubleArrayIterator(v);
    }
    public static Iterator<Long> getIterator(long[] v)
    {
        return new LongArrayIterator(v);
    }
    public static Iterator<Float> getIterator(float[] v)
    {
        return new FloatArrayIterator(v);
    }

    
    public CollectionIterator(A[] c)
    {
        this(c,0, (c==null)?0:c.length);
    }
    public CollectionIterator(A[] c, int startindex)
    {
        this(c,startindex, c.length);
    }
    public CollectionIterator(A[] c, int startindex, int endindex)
    {
        this.v=c;
        index=startindex;
        this.endindex=endindex;
    }

    public boolean hasNext()
    {
        if (it!=null)
        {
            return it.hasNext();
        }
        return v!=null && index<endindex;
    }

    public A next()
    {
        if (it!=null)
        {
            return it.next();
        }
        if (index>=endindex)
        {
            throw new RuntimeException("invoked next but no more element");
        }
        return v[index++];
    }

    public void remove()
    {
        throw new RuntimeException("not implemented");
    }
}

