
package lmu.utils;

import java.awt.Color;
import java.util.*;

public class ColorManager
{

    public static Color getGray(int depth)
    {
        return new Color(depth, depth, depth);
    }

    public static Iterator<Color> getGrayIterator(int maxnum)
    {
        final int base = 100;
        int step = (int) (150.0 / maxnum);
        Vector<Color> colors = new Vector<Color>();
        for (int i = 0; i < maxnum; i++)
        {
            int g = base + (i * step);
            colors.add(getGray(g));
        }
        return colors.iterator();
    }

    
    public static Iterator<Color> getColorIterator()
    {
        return new IteratorUtils.EndlessIterator(ObjectGetter.toVector(Color.BLACK, Color.BLUE, Color.RED, Color.GREEN, Color.MAGENTA, Color.CYAN,
                                                 Color.ORANGE, Color.PINK, Color.GRAY));
        
    }

    
    public enum PREDEFINED_COLOR
    {
        WHITE("white", Color.WHITE),
        LIGHTGRAY("lightgray", Color.LIGHT_GRAY),
        GRAY("gray", Color.GRAY),
        DARKGRAY("darkgray", Color.DARK_GRAY),
        BLACK("black", Color.BLACK),
        RED("red", Color.RED),
        PINK("pink", Color.PINK),
        ORANGE("orange", Color.ORANGE),
        YELLOW("yellow", Color.YELLOW),
        GREEN("green", Color.GREEN),
        MAGENTA("magenta", Color.MAGENTA),
        CYAN("cyan", Color.CYAN),
        BLUE("blue", Color.BLUE);

        String name;
        Color c;

        PREDEFINED_COLOR(String n, Color c)
        {
            this.name = n;
            this.c = c;
            EnumGetter.add(n, this);
        }

        public static Color get(String n)
        {
            PREDEFINED_COLOR pc = (PREDEFINED_COLOR) EnumGetter.get(n);
            if (pc == null)
                return null;

            return pc.c;
        }

        public Color getColor()
        {
            return c;
        }

        public String getName()
        {
            return name;
        }
        
        public static Vector<Color> getColors()
        {
            return ObjectGetter.map(PREDEFINED_COLOR.values(), (PREDEFINED_COLOR c) -> c.c);
        }
        
    }

    public static Color getColor(String s)
    {
        return getColor(s, true);
    }

    /** gets a color - if the string corresponds to a predefined color
     * it returns it. if it is an integer interpretable as an rgb value
     * it returns a color for the rgb */
    public static Color getColor(String s, boolean throw_exc)
    {
        if (s == null || s.length() == 0)
        {
            if (!throw_exc)
            {
                return null;
            }
            throw new RuntimeException("cannot convert a color from empty string!");

        }
        Color c = PREDEFINED_COLOR.get(s);
        if (c != null)
            return c;

        try
        {
            Integer rgb = Integer.parseInt(s);
            return new Color(rgb);
        }
        catch (NumberFormatException nfe)
        {
            // silent error
        }
        if (!throw_exc)
        {
            return null;
        }

        throw new RuntimeException("cannot convert a color from  string: " + s + "!");
    }

    /** if the given color is one of the predefined ones the name is returned, if
     * not color.toString() */
    public static String getColorName(Color color)
    {
        for (PREDEFINED_COLOR def : PREDEFINED_COLOR.values())
        {
            if (def.c.equals(color))
            {
                return def.name();
            }
        }
        return color.toString();
    }
    /**returns a color vector in sinus curve color waves -> rainbow colors */
    public static Vector<Color> makeColorGradient(double frequency1, double frequency2, double frequency3,
                                                  double phase1, double phase2, double phase3,
                                                  Integer center, Integer width, Integer len)
    {
        //idea from http://krazydad.com/tutorials/makecolors.php
        Vector<Color> colorGradient = new Vector<Color>();
        if (center == null)
        {
            center = 128;
        }
        if (width == null)
        {
            width = 127;
        }
        if (len == null)
        {
            len = 50;
        }

        for (int i = 0; i < len; ++i)
        {
            int red = (int) (Math.sin(frequency1 * i + phase1) * width + center);
            int grn = (int) (Math.sin(frequency2 * i + phase2) * width + center);
            int blu = (int) (Math.sin(frequency3 * i + phase3) * width + center);
            Color actColor = new Color(red, grn, blu);
            colorGradient.add(i, actColor);
        }
        return colorGradient;
    }

    public static Color changeLuminal(Color col, double lum)
    {
        float[] rgb = col.getRGBColorComponents(null);
        float[] modRgb = new float[3];
        for (int i = 0; i < 3; i++) {
            float c = rgb[i];
            c = (float) Math.min(Math.max(0, c + (c * lum)), 1);
            modRgb[i]=c;
        }
        return new Color(modRgb[0], modRgb[1], modRgb[2]);
    }
}
