package lmu.utils;

import java.util.*;

public class DataAggregator
{
    interface TypedAggregator
    {
        public void init(double[] data, int[] data_sizes);

        public void update(int p, double d, double[] data, int[] data_sizes);
        public void compile(double[] data, int[] data_sizes);
    }
    
    static class AvgTypeAggregator implements TypedAggregator
    {
        public void init(double[] data, int[] data_sizes)
        {

        }
        public void update(int p, double d, double[] data, int[] data_sizes)
        {
            data[p]+=d;
            data_sizes[p]++;
        }
        public void compile(double[] data, int[] data_sizes)
        {
            for (int i=0; i<data.length; i++)
            {
                if (data_sizes[i]==0)
                {
                    continue;
                }
                data[i]/=data_sizes[i];
            }
        }
    }


    static class MaxTypeAggregator implements TypedAggregator
    {
        public void init(double[] data, int[] data_sizes)
        {
            //set min value?
        }

        public void update(int p, double d, double[] data, int[] data_sizes)
        {
            data[p]=Math.max(d, data[p]);
            
        }
        public void compile(double[] data, int[] data_sizes)
        {
            
        }
    }
    
    public enum TYPE
    {
        AVG("avg"),
        MIN("min"),
        MAX("max"),
        MEDIAN("median")
        ;

        String n;
        TYPE(String n)
        {
            this.n=n;
            EnumGetter.add(n,this);
        }

        public static TYPE get(String n)
        {
            return (TYPE)EnumGetter.get(n);
        }

    }

    TypedAggregator aggr = null;
    TYPE _type;
    int width;

    double[] data;
    int [] data_n;

    public DataAggregator(TYPE type, int width)
    {
        this._type=type;
        this.width=width;

        data = new double[width];
        data_n = new int[width];

        switch(type)
        {
            case AVG:
                aggr = new AvgTypeAggregator();
                break;
            case MAX:
                aggr = new MaxTypeAggregator();
                break;
            default:
                throw new RuntimeException("aggregation type: "+type+" is not yet implemented!");
        }
        aggr.init(data,data_n);
    }
    
    public DataAggregator(String type, int width)
    {
        this(TYPE.get(type), width);
    }


    public void addData(int pos, double d)
    {
        aggr.update(pos,d,data,data_n);
    }

    public void addData(Iterator<Pair<Integer,Double> > dit)
    {
        setData(dit,false);
    }

    public void setData(Iterator<Pair<Integer,Double> > dit)
    {
        setData(dit,true);
    }
    private void setData(Iterator<Pair<Integer,Double> > dit, boolean compile_at_end)
    {
        while(dit.hasNext())
        {
            Pair<Integer, Double> p = dit.next();
            aggr.update(p.getFirst(), p.getSecond(), data, data_n);

        }
        if (compile_at_end)
        {
            aggr.compile(data, data_n);
        }
        
    }
    public void compile()
    {
        aggr.compile(data,data_n);
    }

    public double[] getData()
    {
        return data;
    }

}


