package lmu.utils;

import java.io.*;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.csvreader.CsvReader;

public class DataFile<A>
{

    Logger log = LogConfig.getLogger();

    static
    {
        Locale.setDefault(Locale.UK);
    }
    static Pattern BOOLE_PATTERN=Pattern.compile("(true|false)", Pattern.CASE_INSENSITIVE);
    static Pattern NUM_PATTERN=Pattern.compile("(-)?(?:([Ii][Nn][fF])|(NA)|(NaN)|(\\d+((?:\\.|,)\\d+)?([e|E][-|+]\\d+)?))");

    static Pattern EVAL_PATTERN=Pattern.compile("\\$\\{([^\\}]+)\\}");

    enum TYPES
    {
        INT,
        LONG,
        DOUBLE,
        BOOLEAN,
        STRING,
        STRINGSET,
        DOUBLESET,
        LONGSET
        ;
    
        public static Pair<TYPES, Vector >  autodetect(String header, Vector<String> v, String null_string)
        {
            Logger log = LogConfig.getLogger();
            boolean is_string = false;
            boolean gotnumber=false;
            boolean gotdecimal=false;
    
            boolean got_boolean=false;
            
            HashMap<Integer, Boolean> infinities = new HashMap<Integer, Boolean>();
            HashSet<Integer> unset = new HashSet<Integer>();
            HashSet<Integer> nan = new HashSet<Integer>();
    
            for (int i=0; i<v.size(); i++)
            {
                String s = v.get(i);
                if (s==null || s.length()==0 || (null_string!=null && null_string.equals(s)))
                {
                    unset.add(i);
                    continue;
                }
                Matcher m = BOOLE_PATTERN.matcher(s);
                if (m.matches())
                {
                    got_boolean=true;
                    continue;
                }
                m = NUM_PATTERN.matcher(s);
                if (!m.matches())
                {
                    gotnumber=false;
                    is_string=true;
                    log.trace(header+" is not a number due: "+s+" set is_string (i:"+i+")");
                    break;
                }
                if (i==0)
                {
                    log.trace("found number: "+m.group()+" for "+header+"\n");
                }
                gotnumber=true;
                if (m.group(2)!=null) //infinity
                {
                    infinities.put(i, null==m.group(1)); //true if non negative
                }
                if (null!=m.group(3)) //unset
                {
                    
                    unset.add(i);
                }
                if( null!=m.group(4)) //nan
                {
                    gotdecimal=true;
                    nan.add(i);
                }
                if (null!=m.group(6))
                {
                    gotdecimal=true;
                }
                
            }
            
            log.trace("header: %s after checking got number: %s decimal: %s got string: %s\n", header, ""+gotnumber, ""+gotdecimal, ""+is_string);
            if (is_string)
            {
                return new Pair<TYPES, Vector>(TYPES.STRING, v);
            }
            if (got_boolean)
            {
                Vector<Boolean> rd = new Vector<Boolean>();
                for (int i=0; i<v.size(); i++)
                {
                    rd.add(Boolean.parseBoolean(v.get(i).toLowerCase()));
                }
                return new Pair<TYPES, Vector>(TYPES.BOOLEAN, rd);
            }
            if (gotdecimal)
            {
                StringConverter sc = SimpleTypeConverter.getConverter("double", false);

                Vector<Double> rd = new Vector<Double>();
                for (int i=0; i<v.size(); i++)
                {
                    if (nan.contains(i))
                    {
                        rd.add(Double.NaN);
                        continue;
                    }
                    if (unset.contains(i))
                    {
                        rd.add(null);
                        continue;
                    }
                     rd.add((Double)sc.convert(v.get(i)));
                }
                return new Pair<TYPES, Vector>(TYPES.DOUBLE, rd);
            }
            //long or integer
            Vector<Integer> ri = new Vector<Integer>();
            Vector<Long> rl = new Vector<Long>();
            
            for (int longtest=0; longtest<2; longtest++)
            {
                try
                {
                    for (int i=0; i<v.size(); i++)
                    {
                        if (nan.contains(i))
                        {
                            ri.add(null);
                        }
                        if (unset.contains(i))
                        {
                            if(longtest==0)
                            {
                                ri.add(null);
                            }
                            else
                            {
                                rl.add(null);
                            }
                            
                            continue;
                        }
                        Boolean inftype = infinities.get(i);
                        if (inftype==null)
                        {
                            if(longtest==0)
                            {
                                ri.add(Integer.parseInt(v.get(i)));
                                continue;
                            }
                            
                            rl.add(Long.parseLong(v.get(i)));
                            continue;
                        }
                        if (longtest==0)
                        {
                            ri.add((inftype)?Integer.MAX_VALUE:Integer.MIN_VALUE);
                            continue;
                        }
                        rl.add((inftype)?Long.MAX_VALUE:Long.MIN_VALUE);
                        
                    }
                    if (longtest==0)
                    {
                        return new Pair<TYPES, Vector>(TYPES.INT, ri);
                    }
                    return new Pair<TYPES, Vector>(TYPES.LONG, rl);
                }
                catch(Exception e)
                {
                    log.trace(header+" is not int due to error: "+e.getMessage());
                }
                
            }
        
    
    
            return null;
        }
        public Object convert()
        {
            return null;
        }
        public static Object convert(TYPES t, String s)
        {
            switch(t)
            {
                case INT:
                    return Integer.parseInt(s);
                case DOUBLE:
                    return Double.parseDouble(s);
    
            }
            return null;
        }
    }





    String[] headers = null;
    Vector<TYPES>  types = new Vector<TYPES>();
    String separator = null;

    boolean iscsv=false;
    boolean contains_header=true;
    Vector<Pattern> pos2setsep = new Vector<Pattern>();
    int peaksize = -1;

    public void setCSV(boolean b)
    {
        iscsv=b;
    }
    public String[] getHeaders()
    {
    	return headers;
    }

    public void setSeparator(String s)
    {
        if (s!=null && s.equalsIgnoreCase("tab"))
        {
            s="\t";
        }
        separator=s;
    }

    public void setPeakSize(int num_lines)
    {
        peaksize = num_lines;
    }
    public void setHeaders(String[] headers)
    {
        this.headers=headers;
        
    }
    public void setContainsHeader(boolean b)
    {
        contains_header=b;
    }
    
    void initHeaders(int  l, Vector<Vector<String> > pos2data)
    {
        if (pos2data.size()>0)
        {
            return;
        }
        if (headers==null)
        {
            headers = new String[l];
            for (int i=0; i<l; i++)
            {
                headers[i]="field"+(i+1);
            }
        }
            
        for(int i=0; i<headers.length; i++)
        {
            pos2data.add(new Vector<String>());
        }
        
        
    }
        
    public static String auto_detect_separator(File f)
    {
        return auto_detect_separator(f, (Vector<String>)null);

    }

    public static String auto_detect_separator(File f, String ... comment_signs)
    {
        return auto_detect_separator(f, ObjectGetter.toVector(comment_signs));
    }

    public static String auto_detect_separator(File f, Collection<String> comment_signs)
    {
        String rv="\t";
        try
        {
            
            BufferedReader bfr = FileUtils.getReader(f);
            String header = readNextLine(bfr, comment_signs);
            String line = readNextLine(bfr, comment_signs);
            bfr.close();
            //String[] seps = Pattern.compile("\\w+").split(header);
            //String[] seps2 = (line==null)?new String[0]:Pattern.compile("\\w+").split(line);
            String[] seps = {"\t", ",", ";", " "};
            
            String[] seps2=seps;
    
            //System.out.printf("header line split: %d line2 split: %d\n", seps.length, seps2.length);
            if (line!=null && seps2.length<seps.length)
            {
                seps=seps2;
            }
            CountUpdater cu = new CountUpdater();
    
            
            for (int i=0; i<seps.length; i++)
            {
                cu.update(seps[i], 1);
                
            }
            Vector<Pair<Integer, String> > totest=cu.getSortedCounts();

            String separator=null;
            //check if it results the same number on header and line
            for (String s_test : seps )
            {
                
                //System.out.println("will test: >"+s_test+"<");
                String[] sp1 = null;
                try
                {
                    sp1 = header.split(s_test);
                }
                catch(java.util.regex.PatternSyntaxException pe)
                {
                    continue;
                }
                
                if (line==null)
                {
                    separator=s_test;
                    break;
                }
                String[] sp2 = line.split(s_test);
                //System.out.printf("separator: >%s< sp1: %d sp2: %d\n", s_test, sp1.length, sp2.length);
                if (Math.abs(sp1.length-sp2.length)>1)
                {
                    /*
                    System.out.printf("separator >%s< results different lengths on header(%d) and first line(%d), skipping...\n", 
                                      s_test, sp1.length, sp2.length);

                    int minmax=Math.min(sp1.length, sp2.length);
                    for (int i=0; i<Math.max(sp1.length, sp2.length); i++)
                    {
                        String fn=(i<sp1.length)?sp1[i]:null;
                        String v=(i<sp2.length)?sp2[i]:null;
                        System.out.printf("\t[%d] %s=%s\n", i, fn,v);
                        
                    }
                    */
                    continue;
                }
                separator=s_test;
                break;
            }
            //cu.printBestN(-1);
            //String separator=cu.getSortedCounts().get(0).getSecond();
            if (separator==null)
            {
                //System.out.printf("no valid separator found! assume \\s+");
                separator="\\s+";
            }
            String sepname=separator;
            if (Character.isWhitespace(separator.charAt(0)) &&  sepname.length()==1 && '\t'==sepname.charAt(0))
            {
                sepname="TAB";
            }
            //System.out.printf("selected: >%s< ", sepname);
            return separator;
        }
        catch(IOException e)
        {
            throw new RuntimeException("i/o error: "+e.getMessage()+" while auto-detecting separator from : "+f.getAbsolutePath(), e);
           
        }
        
    }
    public static void printHeaderOverview(File f, String separator)
    {
        if (separator==null)
        {
            separator = auto_detect_separator(f);
        }
        if (separator.equalsIgnoreCase("tab"))
        {
            separator="\t";
        }
        
        try
        {
            
            CsvReader csvreader = new CsvReader(new FileReader(f), separator.charAt(0));
            
                        
            if(!csvreader.readHeaders())
            {
                throw new RuntimeException("could not read csv headers from file: "+f.getAbsolutePath());
            }
            
            System.out.printf("read csv headers: %d\n",csvreader.getHeaderCount());
            String[] headers=new String[csvreader.getHeaderCount()];
            for (int i=0; i<csvreader.getHeaderCount(); i++)
            {
                
                headers[i]=csvreader.getHeader(i);
                System.out.printf("csv header[%d] = %s\n", i, headers[i]);
            }
            
        }
        catch(IOException e)
        {
            throw new RuntimeException("i/o error: "+e.getMessage()+" while creating header overview for "+f.getAbsolutePath());
        }

    }
    boolean fillup_with_null = true;

    public void setFillUpWithNull(boolean b)
    {
        fillup_with_null = b;
    }

    String nullstr = null;

    public void setNullString(String n)
    {
        nullstr = n;
    }

    HashSet<String> comment_signs = new HashSet<>();

    public void addCommentSigns(String ... s)
    {
        ObjectGetter.apply(s, (_s) -> comment_signs.add(_s));
    }

    public Vector<ObjectGetter<A>> read(File f)
    {
        return read(f, null);
    }

    static boolean isComment(String line, Collection<String> comment_signs)
    {
        if (comment_signs == null || line == null || comment_signs.size() == 0 )
            return false;
        
        for (String s : comment_signs)
        {
            if (line.startsWith(s))
                return true;
        }
        return false;
    }

    static String readNextLine(BufferedReader bfr, Collection<String> comment_signs) throws IOException
    {
        while (true)
        {
            String l = bfr.readLine();
            if (isComment(l, comment_signs))
                continue;
            return l;
        }
        
    }

    public Vector<ObjectGetter<A>> read(File f, Reader csv_reader)
    {   
        log.info("will read %s comments: %s", f, comment_signs);
        if (separator==null && csv_reader == null)
        {
            separator = auto_detect_separator(f, comment_signs);
        }
        CsvReader csvreader = null;
        
        Pattern sepP = (csv_reader != null) ? null : Pattern.compile(separator);

        Vector<Vector<String> > pos2data = new Vector<Vector<String> >();
        try
        {
            log.debug("will read from %s csv: %s headeR: %s sep: >%s<\n", (f==null) ? null : f.getAbsolutePath(),  iscsv,  contains_header, separator);

            if (csv_reader != null)
            {
                csvreader =new CsvReader(csv_reader, separator.charAt(0));
                csvreader.setCaptureRawRecord(false);
            }

            if (csvreader == null && iscsv)
            {
                csvreader =new CsvReader(FileUtils.getReader(f), separator.charAt(0));
                csvreader.setCaptureRawRecord(false);
                
            }
            String line = null;
            BufferedReader bfr = (csvreader==null) ? FileUtils.getReader(f):null;
            String headerline =  null;
            
            if (contains_header || csvreader!=null)
            {
                if (csvreader!=null)
                {
                
                    if(!csvreader.readHeaders())
                    {
                        throw new RuntimeException("could not read csv headers from file: "+f.getAbsolutePath());
                    }
                    //System.out.printf("read csv headers: %d\n",csvreader.getHeaderCount());
                    headers=new String[csvreader.getHeaderCount()];
                    for (int i=0; i<csvreader.getHeaderCount(); i++)
                    {
                        
                        headers[i]=csvreader.getHeader(i);
//                        System.out.printf("csv header[%d] = %s\n", i, headers[i]);
                    }
                    initHeaders(headers.length, pos2data);
                }
                else
                {
                    headerline = readNextLine(bfr, comment_signs);
                    System.out.printf("got headerline: >%s<\n", headerline);
                    final String HEADERDEF = "#Format: ";
                    if (headerline.startsWith(HEADERDEF))
                    {
                        int ep = headerline.indexOf('(');
                        headerline = headerline.substring(HEADERDEF.length(), (ep < 0) ? headerline.length() : ep);
                    }
                    System.out.printf("got corrected headerline: >%s<\n", headerline);
                    headers = sepP.split(headerline);
                    System.out.printf("got headers: %d: %s\n", headers.length, StringUtils.join(",", headers));
                    if (headers.length == 1)
                    {
                        
                        headers = headerline.split(" ");
                    }
                    
                }
                
            }
            

            
            FileUtils.RecordIterator rit = new FileUtils.RecordIterator(separator);
    
            
    
            String dummy[] = null;

            int ndatalines=0;
            while (true)
            {
                if (csvreader==null)
                {
                    if(null==(line=readNextLine(bfr, comment_signs)))
                        break;
                    
                    if (line.length()==0)
                        continue;
                    
                    
                    
                    String sp[] = rit.split(line);
                    initHeaders(sp.length, pos2data);
    
                    if (fillup_with_null && sp.length < headers.length)
                    {
                        dummy = sp;
                        sp = new String[headers.length];
                        for (int i=0; i<dummy.length; i++)
                        {
                            sp[i] = dummy[i];
                        }
                    }
                    if (sp.length!=headers.length)
                    {
                        BufferPrintWriter bpw = new BufferPrintWriter();
                        for (int i=0; i<Math.max(sp.length, headers.length); i++ )
                        {
                            String header = (headers.length>i) ? headers[i] : null;
                            String data = (sp.length>i) ? sp[i] : null;
                            bpw.printf("\t[%d] header: %s data: %s\n", i, header, data);
                        }
                        throw new RuntimeException("invalid line : "+line+" field num ("+sp.length+") does not match to header num ("+headers.length+"):\n"+headerline+"\n"+bpw.getBuffer());
                    }
                    for (int i=0; i<headers.length; i++)
                    {
                        pos2data.get(i).add(sp[i]);
                    }
                    ndatalines++;
                    if(peaksize > 0 && ndatalines >= peaksize)
                        break;
                    if (ndatalines % 10000 == 0)
                    {
                        log.info("read %d entries from %s", ndatalines, f.getName());
                    }
                    continue;
                }
                
                if(!csvreader.readRecord())
                {
                    break; 
                }
                ndatalines++;
                
                for (int i=0; i<headers.length; i++)
                {
                    pos2data.get(i).add(csvreader.get(i));
                }

                if (ndatalines % 10000 == 0)
                {
                    log.info("read %d entries from %s", ndatalines, f.getName());
                }

                
                
                
            }
    
            
    
            ObjectGetter.ObjectGetterHeader<A> oh = new ObjectGetter.ObjectGetterHeader<A>();
            oh.headers=headers;
    
            for (int i=0; i<ndatalines; i++)
            {
                oh.addData();
                
            }
            for (int i=0; i<headers.length; i++)
            {
                oh.addField(headers[i], i);
                
                Pair<TYPES, Vector > pt = TYPES.autodetect(headers[i], pos2data.get(i), nullstr);
                if (pt==null)
                {
                    throw new RuntimeException("could not detect type for "+headers[i]+" in file "+f.getAbsolutePath());
                }
    
                Vector v = pt.getSecond();
                for (int j=0; j<v.size(); j++)
                {
                    oh.getData(j).add(v.get(j));
                }
                
            }
            if (csvreader==null)
            {
                bfr.close();
            }
            else
            {
                csvreader.close();
            }
            
            return oh.getAllData();
        }
        catch(IOException e)
        {
            throw new RuntimeException("error: "+e.getMessage()+" by parsing "+f);
        }
    
    }


}

