package lmu.utils;
import java.util.*;
/** a memory list itself can operate on this structure, change the
     * order, set visibility, formatting, etc
     */
public class DataHeaderInfos
{
    public Vector<DataHeader> headers = new Vector<DataHeader>();
    HashMap<String, DataHeader> name2header = new HashMap<String, DataHeader>();


    public HashMap<String, ReflectedObject> getReflectedObjectLookup()
    {
        HashMap<String, ReflectedObject> rv = new HashMap<String, ReflectedObject>();

        for (Map.Entry<String,DataHeader> e: name2header.entrySet())
        {
            rv.put(e.getKey(), e.getValue().getRefObj());
        }
        return rv;
    }

    public Vector<DataHeader> getVisibleHeaders()
    {
        Vector<DataHeader> v = new Vector<DataHeader>();
        
        for (DataHeader dh : headers)
        {
            if (!dh.visible)
                continue;
            
            v.add(dh);
        }
        return v;

    }
    public int getVisibleHeaderSize()
    {
        return getVisibleHeaders().size();
    }
    public Collection<String> getHeaderNames()
    {
        return ObjectGetter.map(headers, (DataHeader dh) -> dh.getName());
    }
    public DataHeader addHeader(String label, String expression)
    {
        return addHeader(label);
    }
    public DataHeader addHeader(String label, ReflectedObject ro)
    {
        DataHeader dh = addHeader(label);
        dh.parent=this;
        dh.setRefObj(ro);
        return dh;
    }

    public void changeHeaderName(String label, String nlabel)
    {
        DataHeader dh = getHeader(label);
        name2header.remove(label);
        dh.name=nlabel;
        name2header.put(dh.name, dh);

    }

    /** defines order and sets all non specfied headers to invisible */
    public void setOrder(String ... labels)
    {
        HashSet<String> alllabels = new HashSet<String>();
        alllabels.addAll(name2header.keySet());

        headers.clear();

        for (String l : labels)
        {
            DataHeader dh = getHeader(l);
            headers.add(dh);
            alllabels.remove(dh.getName());
        }
        for (String left: alllabels)
        {
            DataHeader dh = getHeader(left);
            dh.setVisible(false);
            headers.add(dh);
        }
    }
    public DataHeader getHeader(int i)
    {
        return headers.get(i);
    }
    public DataHeader findHeader(String n)
    {
        return findHeader(n,false);
    }
    public DataHeader findHeader(String n, boolean contains)
    {
        if (!contains)
        {
            return name2header.get(n);
        }
        for (DataHeader dh : headers)
        {
            if (!dh.name.contains(n))
                continue;
            
            return dh;
        }
        return null;
        
    }
    
    public boolean checkForHeader(String n)
    {
        return name2header.containsKey(n);
    }
    
    public DataHeader getHeader(String n)
    {
        DataHeader dh = name2header.get(n);
        if (dh==null)
        {
            throw new RuntimeException("no header found: "+n);
        }
        return dh;

    }
    public void setVisible(boolean visible, Iterator<String> labels)
    {
        while (labels.hasNext())
        {
            DataHeader dh = getHeader(labels.next());
            dh.setVisible(false);
        }
    }
    public void setVisible(boolean visible, String ... labels)
    {
        setVisible(visible, new CollectionIterator(labels));
    }
    public DataHeader addHeader(String label)
    {
        DataHeader dh = new DataHeader();
        dh.col_index=headers.size();
        dh.name=label;
        name2header.put(dh.name, dh);
        headers.add(dh);
        return dh;
    }

    public void init(ObjectGetter.ObjectGetterHeader header)
    {
        clear();
        for (String h: header.headers)
        {
            if (h == null)
                continue;
            
            addHeader(h, header.getFieldInfo(h));
        }
    }

    public void clear()
    {
        headers.clear();
    }

}

