package lmu.utils;
import lmu.utils.db.*;
import java.sql.*;
import java.util.HashMap;

public class DbConnection implements ConnectionGetter
{
    long last_connection_retrieved=0;
    String constr="";
    Connection sql = null;
    Logger log = LogConfig.getLogger();
    static final long MAX_SQL_TIMEOUT=60*60*1000; //after 10 minutes  without getconnection we reopend the connection

    HashMap<PreparedStatement, String> stored_prep_statements = new HashMap<PreparedStatement, String>();
    
    String dbname="";

    public String getDbName()
    {
        return dbname;
    }

    public static boolean exec(PreparedStatement ps) 
    {
        try
        {
            return ps.execute();
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while executing: %s", se, se.getMessage(),  ps);
        }
    }
    public String create(DbDefinition def, String prefix) 
    {
        try
        {
            return def.create(prefix, getConnection().createStatement());
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while creating table: %s (prefix: %s)", se, se.getMessage(),  def.getTableName(),  prefix);
        }
        
    }
    public void exec(String q, Object ... data) 
    {
        execWithError(null,q, data);
    }

    public static void exec(Connection c, String q, Object ... data) 
    {
        execWithError(c, null,q, data);
    }
    
    public static void execWithError(Connection c , String error_text, String q, Object ... data) 
    {
        String cmd = (data.length == 0) ? q : String.format(q, data);
        try
        {
            Statement st = c.createStatement();
            st.execute(cmd);
            st.close();
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s %slast cmd was: %s", se, se.getMessage(),  (error_text == null) ? "": " "+error_text,  cmd);
        }
        
    }

    public void execWithError(String error_text, String q, Object ... data) 
    {
        execWithError(getConnection(), error_text, q, data);
    }
    
    public boolean tableExists(String tblname)
    {
        return SQLUtils.isEmpty(query("select 1 from "+tblname+" limit 1;"));
    }
    public ResultSet query(String q, Object ... fmt)
    {
        return query(getConnection(), q, fmt);
    }
    
    public static ResultSet query(Connection c, String q, Object ... fmt) 
    {
        return queryWithError(c, q, false,  null, fmt);
    }

    
    public boolean doesTableExist(String table)
    {
        try
        {
            return null != query("select 1 from "+table+" limit 1;") ;
            //return query("show tables like '"+table+"';").first();
        }
        catch(Exception se)
        {
            return false;
            //throw new FRuntimeException("sql error: %s while checking table: %s for existence", se, se.getMessage(), table);
        }
        
        
    }

    public ResultSet first(String q, Object ... fmt)
    {
        return first(getConnection(), q,  fmt);
    }

    public static ResultSet first(Connection c, String q, Object ... fmt)
    {
        return first(query(c,q,fmt), (fmt.length == 0) ? q : String.format(q, fmt));
    }

    public static ResultSet first(ResultSet rs, String info)
    {
        try
        {
            if(!rs.first())
                return null;

            return rs;
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while getting first record, info: %s", se, se.getMessage(), info);
        }
    }

    public static ResultSet first(PreparedStatement ps)
    {
        return first(query(ps), ""+ps);
    }

    public static ResultSet query(PreparedStatement ps)
    {
        try
        {
            return ps.executeQuery();
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while running query: %s", se,
                                        se.getMessage(), ps);
        }
    }

    public Object queryObject(String q, int fieldpos, SimpleTypeConverter.SIMPLETYPE type)
    {
        SimpleTypeConverter.SQLConverter sq = type.getSQLConverter();
        try
        {
            ResultSet rs = query(q);
            if (!rs.next())
                return null;
            
            return sq.get(rs, fieldpos);
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while getting simple type : %s at pos: %s from query: %s", se,
                                        se.getMessage(), type, fieldpos, q);
        }
    }



    static public DPreparedStatement setString(PreparedStatement ps, String v) 
    {
        return (new DPreparedStatement(ps)).setString(v);
    }

    static public DPreparedStatement setBytes(PreparedStatement ps, int i, byte[] data) 
    {
        try
        {
            if (data == null)
            {
                ps.setNull(i, java.sql.Types.BLOB);
            }
            ps.setBytes(i, data);
        }
        catch(SQLException e)
        {
            throw new FRuntimeException("sql error: %s setting binary data at idx %d to query: %s", e, e.getMessage(), i, ps);
        }
        return new DPreparedStatement(ps);
    }

    static public DPreparedStatement setString(PreparedStatement ps, int i, String v) 
    {
        try
        {
            if (v == null)
            {
                ps.setNull(i, java.sql.Types.CHAR);
            }
            ps.setString(i, v);
        }
        catch(SQLException e)
        {
            throw new FRuntimeException("sql error: %s setting param %d to %s query: %s", e, e.getMessage(), i, v, ps);
        }
        return new DPreparedStatement(ps);
    }

    static public DPreparedStatement setInt(PreparedStatement ps, Integer v) 
    {
        return (new DPreparedStatement(ps)).setInt(v);
    }

    static public DPreparedStatement setInt(PreparedStatement ps, int i, Integer v) 
    {
        try
        {
            if (v == null)
            {
                ps.setNull(i, java.sql.Types.INTEGER);
            }
            ps.setInt(i, v);
        }
        catch(SQLException e)
        {
            throw new FRuntimeException("sql error: %s setting param %d to %s query: %s", e, e.getMessage(), i, v, ps);
        }
        return new DPreparedStatement(ps);
    }
    
    static public DPreparedStatement setLong(PreparedStatement ps, long v) 
    {
        return (new DPreparedStatement(ps)).setLong(v);
    }

    static public DPreparedStatement setLong(PreparedStatement ps, int i, long v) 
    {
        try
        {
            ps.setLong(i, v);
        }
        catch(SQLException e)
        {
            throw new FRuntimeException("sql error: %s setting param %d to %s query: %s", e, e.getMessage(), i, v, ps);
        }
        return new DPreparedStatement(ps);
    }
    
    static public DPreparedStatement setDouble(PreparedStatement ps, Double v) 
    {
        return (new DPreparedStatement(ps)).setDouble(v);
    }

    static public DPreparedStatement setDouble(PreparedStatement ps, int i, Double v) 
    {
        try
        {
            if (v == null)
            {
                ps.setNull(i, java.sql.Types.DOUBLE);
            }
            ps.setDouble(i, v);
        }
        catch(SQLException e)
        {
            throw new FRuntimeException("sql error: %s setting param %d to %s query: %s", e, e.getMessage(), i, v, ps);
        }
        return new DPreparedStatement(ps);
    }

    static public DPreparedStatement setFloat(PreparedStatement ps, Float v) 
    {
        return (new DPreparedStatement(ps)).setFloat(v);
    }

    static public DPreparedStatement setFloat(PreparedStatement ps, int i, Float v) 
    {
        try
        {
            if (v == null)
            {
                ps.setNull(i, java.sql.Types.FLOAT);
            }
            ps.setDouble(i, v);
        }
        catch(SQLException e)
        {
            throw new FRuntimeException("sql error: %s setting param %d to %s query: %s", e, e.getMessage(), i, v, ps);
        }
        return new DPreparedStatement(ps);
    }
    
    /**
     * @param statement
     * @param i
     * @param bVisible
     * @return
     */
    static public DPreparedStatement setBoolean(PreparedStatement ps, int i, Boolean v) 
    {
        try
        {
            if (v == null)
            {
                ps.setNull(i, java.sql.Types.BOOLEAN);
            }
            ps.setBoolean(i, v);
        }
        catch(SQLException e)
        {
            throw new FRuntimeException("sql error: %s setting param %d to %s query: %s", e, e.getMessage(), i, v, ps);
        }
        return new DPreparedStatement(ps);
    }
    
    static public class DPreparedStatement 
    {
        Logger log = LogConfig.getLogger();
        PreparedStatement ps;
        int colidx = 0;
        DPreparedStatement(PreparedStatement ps)
        {
            this.ps = ps;
        }

        /**
         * @param v
         * @return
         */
        public DPreparedStatement setLong(long v)
        {
            return setLong(++colidx, v);
        }
        
        public DPreparedStatement setLong(int i, long v)
        {
            DbConnection.setLong(ps, i,v);
            return this;
        }

        public DPreparedStatement setInt(Integer v)
        {
            return setInt(++colidx, v);
        }

        public DPreparedStatement setInt(int i, Integer v)
        {
            DbConnection.setInt(ps, i,v);
            return this;
        }


        public DPreparedStatement setBytes(byte[] data)
        {
            return setBytes(++colidx, data);
        }

        public DPreparedStatement setBytes(int i, byte[] data)
        {
            DbConnection.setBytes(ps, i, data);
            return this;
        }
        
        public DPreparedStatement setString(String v)
        {
            return setString(++colidx, v);
        }

        public DPreparedStatement setString(int i, String v)
        {
            DbConnection.setString(ps,i,v);
            return this;
        }

        public DPreparedStatement setFloat(Float v)
        {
            return setFloat(++colidx, v);
        }

        public DPreparedStatement setFloat(int i, Float v)
        {
            DbConnection.setFloat(ps, i,v);
            return this;
        }

        public DPreparedStatement setDouble(Double v)
        {
            return setDouble(++colidx, v);
        }

        public DPreparedStatement setDouble(int i, Double v)
        {
            DbConnection.setDouble(ps, i,v);
            return this;
        }
        
        /**
         * @param i position
         * @param bValue Value to set
         * @return
         */
        public DPreparedStatement setBoolean(int i, Boolean bValue)
        {
            
            DbConnection.setBoolean(ps, i, bValue);
            
            return this;
        }

        public ResultSet first()
        {
            return DbConnection.first(ps);
        }
        public ResultSet query()
        {
            //log.info("query:  " + ps);
            return DbConnection.query(ps);
        }

        public Integer checkIndex()
        {
            if (isEmpty())
                return null;

            return queryInt();
        }
        public int queryInt()
        {
            //log.info("query:  " + ps);
            try
            {
                return first().getInt(1);
            }
            catch(SQLException se)
            {
                throw new FRuntimeException("sql error: %s while querying first int ps: %s", se, se.getMessage(), ps);
            }
            
        }

        public String queryString()
        {
            //log.info("query:  " + ps);
            try
            {
                return first().getString(1);
            }
            catch(SQLException se)
            {
                throw new FRuntimeException("sql error: %s while querying first string ps: %s", se, se.getMessage(), ps);
            }
            
        }
       
        public boolean isEmpty()
        {
            return DbConnection.isEmpty(ps);
        }

        /**
         * Executes prepared statement.
         *
         * @return
         */
        public boolean exec()
        {
            
            return DbConnection.exec(ps);
        }

        public DPreparedStatement addBatch()
        {
            try
            {
            
                ps.addBatch();
            }
            catch(SQLException se)
            {
                throw new FRuntimeException("sql error: %s while adding batch, ps: %s", se, se.getMessage(), ps);
            }
    
            return this;
        }

        public DPreparedStatement executeBatch()
        {
            try
            {
            
                ps.executeBatch();
            }
            catch(SQLException se)
            {
                throw new FRuntimeException("sql error: %s while executing batch, ps: %s", se, se.getMessage(), ps);
            }
    

            
            return this;
        }
        public int generateKey()
        {
            exec();
            return DbConnection.getGeneratedKey(ps);
        }

        /**
         * @return
         */
        public PreparedStatement getPreparedStatement()
        {
            return ps;
        }


    }
    public int getGeneratedKey(String insert_query)
    {
        try
        {
           PreparedStatement ps = sql.prepareStatement(insert_query, Statement.RETURN_GENERATED_KEYS);
           ps.execute();
           return getGeneratedKey(ps);
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while inserting new entry with: %s", se, se.getMessage(), insert_query);
        }
    }

    public static int getGeneratedKey(PreparedStatement ps)
    {
        try
        {
            
            ResultSet rs = ps.getGeneratedKeys();
            if (!rs.first())
            {
                throw new FRuntimeException("no generated db key after executing %s", ps);
            }
            return rs.getInt(1);
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while retreiving generated keys from: %s", se,
                                        se.getMessage(), ps);
        }
    }

    public Integer checkIndex(String q, Object ... fmt)
    {
        String fq = String.format(q, fmt);
        if (isEmpty(fq))
            return null;
        
        return queryInt(fq);
    }

    public static boolean isEmpty(PreparedStatement ps)
    {

        try
        {
            return !query(ps).first();
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while checking if empty query: %s", se,
                                        se.getMessage(), ps);
        }
        
    }


    public boolean isEmpty(String q, Object ... fmt)
    {
        try
        {
            return !query(q, fmt).first();
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while checking if empty query: %s", se,
                                        se.getMessage(), q);
        }
        
        

    }
    

    public int queryInt(String q)
    {
        return queryInt(q, 1);
    }
    public int queryInt(String q, int fieldpos)
    {
        return (Integer)queryObject(q, fieldpos, SimpleTypeConverter.SIMPLETYPE.INT);
    }
    
    public double queryDouble(String q)
    {
        return queryDouble(q, 1);
    }
    public double queryDouble(String q, int fieldpos)
    {
        return (Double)queryObject(q, fieldpos, SimpleTypeConverter.SIMPLETYPE.DOUBLE);
    }

    public Boolean queryBoolean(String q)
    {
        return queryBoolean(q,1);
    }
    public Boolean queryBoolean(String q, int fieldpos)
    {
        return (Boolean)queryObject(q, fieldpos, SimpleTypeConverter.SIMPLETYPE.BOOLEAN);
    }
    public String queryString(String q)
    {
        return queryString(q, 1);
    }
    public String queryString(String q, int fieldpos)
    {
        return (String)queryObject(q, fieldpos, SimpleTypeConverter.SIMPLETYPE.STRING);
    }


    public ResultSet queryWithError(String q, boolean check_first, String errormsg_if_empty, Object ... fmt_data) 
    {
        return queryWithError(getConnection(), q, check_first, errormsg_if_empty, fmt_data);
    }

    static public ResultSet queryWithError(Connection c, String q, boolean check_first, String errormsg_if_empty, Object ... fmt_data) 
    {
        String fq = String.format(q, fmt_data);
        try
        {
            
            ResultSet rs =  new AutoClosingResultSet(c.createStatement().executeQuery(fq));
            if (check_first && !rs.next())
            {
                throw new RuntimeException(String.format(errormsg_if_empty, fmt_data));
            }
            return rs;
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while executing query: %s", se, se.getMessage(), fq);
        }
    }
    public PreparedStatement getStatement(PreparedStatement ps) 
    {
        
        Connection con = getConnection();
        try
        {
            if(con==ps.getConnection())
                return ps;

        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while checking connection for prepared statement: %s", se, se.getMessage(), ps);
        }
        String def = stored_prep_statements.get(ps);
        if (def==null)
        {
            //cannot recreate
            return ps;
        }
        try
        {
            ps.close();
        }
        catch(Exception e)
        {
            //silent error
        }
        log.debug((new java.util.Date())+" will recreate statement: "+def);
        //have to recreate
        return prepareStatement(def);
    }
    
    static public PreparedStatement prepareStatement(Connection con, String statement, Object ... fmt) 
    {
        statement = (fmt.length == 0) ? statement : String.format(statement, fmt);
        
        try
        {
            return con.prepareStatement(statement, Statement.RETURN_GENERATED_KEYS);
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while prepareing statement: %s", se, se.getMessage(), statement);
        }
    }
    public PreparedStatement prepareStatement(String statement, Object ... fmt) 
    {
        PreparedStatement ps = prepareStatement(getConnection(), statement, fmt);
        stored_prep_statements.put(ps, statement);
        return ps;
    }

    static int DEFAULT_MYSQL_PORT = 3306;
    static String DEFAULT_TYPE = "mysql";
    static String DEFAULT_PARAMS = "";
    int port;
    String host;
    String database;
    String user;
    String pw;
    String dbtype;
    String params;

    public DbConnection(String host, String database, String user, String password)
    {
        this(host, DEFAULT_MYSQL_PORT, database, user, password);
    }
    
    public DbConnection(String host, int port, String database, String user, String password)
    {
        this(host, port, database, user, password, DEFAULT_TYPE, DEFAULT_PARAMS, null);
    }
    public DbConnection(String host, int port, String database, String user, String password, String dbtype, String params, String url)
    {
        this.host = host;
        this.port = port;
        this.database = database;
        this.user = user; 
        this.pw = password;
        this.dbtype = dbtype;
        this.params = params;
        reinit(url);

    }

    void reinit(String url)
    {
        constr = (url != null) ? url : String.format("jdbc:%s://%s:%d/%s?user=%s&password=%s%s", dbtype,  host, port, database, user, pw, params);

        try
        {
            Class.forName("com.mysql.jdbc.Driver");
            new com.mysql.jdbc.Driver();
            
        }
        catch(Exception e)
        {
            throw new RuntimeException(e);
        }
        
        

        String sp[]=constr.split("/");
        dbname=sp[sp.length-1].split("\\?")[0];
        log.debug("url: %s dbname: %s\n", constr, dbname);

    }
    
    public DbConnection resetDB(String db)
    {
        this.database = db;
        reinit(null);
        return this;
    }
    

    java.io.File h2file = null;
    //file based database
    public DbConnection(java.io.File f )
    {
        this(f, false);
    }

    public DbConnection(java.io.File f, boolean write)
    {
        h2file = f;
        try
        {
            Class.forName("org.h2.Driver");
            if (write)
            {
                sql = DriverManager.getConnection("jdbc:h2:"+h2file.getAbsolutePath()+";ACCESS_MODE_DATA=rw;FILE_LOCK=NO", "", "");
            }
            else
            {
                sql = DriverManager.getConnection("jdbc:h2:"+h2file.getAbsolutePath()+";ACCESS_MODE_DATA=r;FILE_LOCK=NO", "", "");
            }
            
        }
        catch(ClassNotFoundException nfe)
        {
            throw new RuntimeException("h2 driver not found, jar there?");
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("error; %s while initing h2 driver from file: %s", se , se.getMessage(), f.getName());
        }
    }

    public DbConnection(String url)
    {
        reinit(url);
    }

    public DbConnection clone()
    {
        return new DbConnection(host, port, database, user, pw, dbtype, params, constr);
    }

    public Connection getConnection()
    {
        if (h2file != null)
        {
            return sql;
        }
        try
        {
            long current = (new java.util.Date()).getTime();
            if (sql!=null && current<last_connection_retrieved+MAX_SQL_TIMEOUT)
            {
                //System.err.println("\treuse connection to "+constr+" diff: "+(current-last_connection_retrieved));
                last_connection_retrieved=current;
                return sql;
            }
            //System.err.println("\n\n\n\ncreate new connection to "+constr+" diff: "+(current-last_connection_retrieved)+"\n\n\n\n");
            if (sql!=null) //no connection in timeout range
            {
                try
                {
                    sql.close();
                }
                catch (Exception close_e)
                {
                    //silent error
                }


            }
            log.debug("will open: " + constr);
            sql = DriverManager.getConnection(constr);
            last_connection_retrieved=current;

            return sql;
        }
        catch (Exception e)
        {
            throw new RuntimeException(e);
        }



    }

    
    public static void main (String args[]) throws SQLException
    {
        SimpleOptionParser cmd = new SimpleOptionParser("url", "db", "host", "port", "user", "password", "query");
        cmd.setOptional("url", "db", "host", "port", "user", "password");
        cmd.setInt("port");
        cmd.setRequiredSubParamSets(ObjectGetter.toArray("url"), ObjectGetter.toArray("db", "user"));
        
        if (!OptionParser.parseParams(args, true, false, true, cmd))
            return;
        
        
        
        DbConnection dcon = null;
        
        if (cmd.isOptionSet("url"))
        {
             dcon = new DbConnection(cmd.getValue("url"));
        }
        else
        {
             dcon = new DbConnection(cmd.getOptionalValue("host", "localhost"), cmd.getOptionalInt("port", DEFAULT_MYSQL_PORT),
                                     cmd.getValue("db"), cmd.getValue("user"), cmd.getOptionalValue("password", ""));
        }

        Connection con = dcon.getConnection();

        String q = cmd.getValue("query");
                    
        ResultSet rs = con.createStatement().executeQuery(q);
            
        while (rs.next())
        {
            System.out.println(lmu.utils.SQLUtils.showResults(rs));
        }
        
    }




}
