package lmu.utils;

import java.sql.*;
import java.util.*;
import java.util.regex.Pattern;

public class DbDefinition implements java.io.Serializable
{
    Logger log = null;
    
    
    Logger getLogger()
    {
        if(log!=null)
            return log;
        return LogConfig.getLogger("SQL");
    }
    
    static Pattern WHITESPACE = Pattern.compile("\\s+");
    public static final String FIELDDEFS="FIELDDEFS";
    public static final String KEYDEFS="KEYDEFS";
    public static final String PRIMARY_KEYDEFS="PRIMARY_KEYDEFS";
    String tblname;
    Vector<String> fields = new Vector<String>();
    Vector<String> field_names = new Vector<String>();
    Vector<String> primary_keys = new Vector<String>();
    HashSet<String> keys = new HashSet<String>();
    
    String createstr = null;
    
    public EPreparedStatement prepareInsert(Connection con) 
    {
        return prepareInsert("", con);
    }

    /** prepares an insert from all fields */
    public EPreparedStatement prepareInsert(String prefix, Connection con) 
    {
        return prepareInsert(prefix, con, StringUtils.convert(field_names));
    }
    public EPreparedStatement prepareInsert(Connection con, String ... fields) 
    {
        return prepareInsert("", con, fields);
    }
    public EPreparedStatement prepareInsert(String prefix, Connection con, String ... fields) 
    {
        StringBuffer sb1 = new StringBuffer();
        StringBuffer sb2 = new StringBuffer();
        for (String s : fields)
        {
            if (sb1.length()>0)
            {
                sb1.append(',');
                sb2.append(',');
            }
            sb1.append(s);
            sb2.append('?');
        }
        return new EPreparedStatement(con, "insert into "+prefix+tblname+" ("+sb1.toString()+") values ("+sb2.toString()+");", fields);
    }
    
    public void setTableName(String tblname)
    {
        this.tblname = tblname;
    }

    public String getTableName()
    {
        return tblname;
    }

    public DbDefinition(String tblname,  String ... fielddefs)
    {
        this(tblname, true, fielddefs);
    }

    /*  @todo
    public class DbDefBuilder
    {
        public DbDefBuilder(String tablename)
        {

        }

        public DbDefBuilder addField(String fieldname, String fieldtype, boolean indexed)
        {

        }
    }

    public static DbDefinition build
    */
    void init(String ... fielddefs)
    {
        fields.clear();
        primary_keys.clear();
        keys.clear();
        field_names.clear();

        HashSet<String> known_fields = new HashSet<String>();
        Collection<String> target = fields;
        
        StringBuffer params = new StringBuffer("params:");
        for (String f : fielddefs)
        {
            //System.err.println(tblname+" :: next field: "+f);
            params.append(" :: ");
            params.append(f);

            if (f.equals(FIELDDEFS))
            {
                //System.err.println("found fielddef sep");
                target=fields;
                continue;
            }
            if (f.equals(KEYDEFS))
            {
                //System.err.println("found keydefs sep");
                target=keys;
                continue;
            }
            if (f.equals(PRIMARY_KEYDEFS))
            {
                //System.err.println("found primary key sep");
                target=primary_keys;
                continue;
            }
            if (target==fields)
            {
                String sp[]=WHITESPACE.split(f);
                if (sp.length<2)
                {
                    throw new RuntimeException("invalid field definition (got "+sp.length+" splitted elems) >"+f+"< for table: "+tblname+ " "+params.toString());
                }
                //System.err.println("add known field: >"+sp[0]+"<");
                known_fields.add(sp[0]);
            }
            target.add(f);
        }

        
        if (known_fields.size()==0)
        {
            throw new RuntimeException("table "+tblname+" does not have any fields definition "+params.toString());
        }
        field_names.addAll(known_fields);
        StringBuffer primkey_buff = new StringBuffer();
        for (String k : primary_keys)
        {
            if (!known_fields.contains(k))
                throw new RuntimeException("undefined field in primary key: "+k+" "+params.toString());
            
            if (primkey_buff.length()>0)
            {
                primkey_buff.append(',');
            }
            primkey_buff.append(k);
            
        }
        StringBuffer key_buff = new StringBuffer();
        for (String k : keys)
        {
            if (!known_fields.contains(k))
                throw new RuntimeException("undefined field in keydef: "+k+" "+params.toString());
            
            if (key_buff.length()>0)
            {
                key_buff.append(',');
            }
            key_buff.append("key(");
            key_buff.append(k);
            key_buff.append(')');
            
        }

        StringBuffer defbuff = new StringBuffer();
        for (String f : fields)
        {
            if (defbuff.length()>0)
            {
                defbuff.append(',');
            }
            defbuff.append(f);
        }

        if (primkey_buff.length()>0)
        {
            defbuff.append(',');
            defbuff.append(" primary key( ");
            defbuff.append(primkey_buff.toString());
            defbuff.append(") ");

        }
        
        if (key_buff.length()>0)
        {
            defbuff.append(',');
            defbuff.append(key_buff.toString());
        }
        
        StringBuffer cbuff = new StringBuffer();
        cbuff.append("create table ");
        if (check_if_exists)
        {
            cbuff.append(" if not exists ");
        }
        cbuff.append("%s");
        cbuff.append(tblname);
        cbuff.append('(');
        cbuff.append(defbuff.toString());
        cbuff.append(") ENGINE=MyISAM;");

        createstr=cbuff.toString();             
    }
    public void extend(DbDefinition def)
    {
        String[] concat = StringUtils.concatArrays(fielddefs, def.fielddefs);
        this.fielddefs=concat;
        init(fielddefs);
    }

    boolean check_if_exists=false;
    String[] fielddefs;
    public DbDefinition(String tblname, boolean checkifexists, String ... fielddefs)
    {
        this.tblname=tblname;
        this.fielddefs=fielddefs;
        this.check_if_exists=checkifexists;
        init(fielddefs);
        
    }
    /** returns the prefixed table name */
    public String create(String prefix, Statement st) 
    {
        String cmd = String.format(createstr, (prefix==null)?"":prefix);
        
        getLogger().debug(cmd);
        try
        {
            st.execute(cmd);
        }
        catch(SQLException se)
        {
            throw new FRuntimeException("sql error: %s while creating table: %s last cmd: %s", se, se.getMessage(), tblname,cmd);
        }
        
        return (prefix==null)?tblname:prefix+tblname;
    }
}

