package lmu.utils;

import java.util.*;

public class DenseRegionLayouter
{

    public static Vector<Vector<Region1D> > getDenseLayout(int min_space, Region1D ... r)
    {
        return getDenseLayout(min_space, new CollectionIterator(r));
    }
    public static Vector<Vector<Region1D> > getDenseLayout(int min_space, Collection<Region1D> r)
    {
        return getDenseLayout(min_space, new CollectionIterator(r));
    }
    /** returns the a dense layout, i.e. a (heruisticly) minimized
     * number of rows containing all regions so all regions are separated at least 
     * by min_space
     */
    public static Vector<Vector<Region1D> > getDenseLayout(int min_space, Iterator<Region1D> r)
    {
        Vector<Region1D> sorted = new Vector<Region1D>();
        while (r.hasNext())
        {
            sorted.add(r.next());
        }
        Collections.sort(sorted);

        Vector<Vector<Region1D> > rows = new Vector<Vector<Region1D> >();

        while (sorted.size()>0)
        {
            Vector<Region1D> row = new Vector<Region1D>();
            rows.add(row);

            Region1D last = null;
            Vector<Region1D> left = new Vector<Region1D>();

            for (Region1D r1d : sorted)
            {
                if (last==null)
                {
                    row.add(last=r1d);
                    continue;
                }

                if (last.getX2()+min_space>=r1d.getX1())
                {
                    left.add(r1d);
                    continue;
                }
                row.add(last=r1d);
            }

            sorted=left;
        }

        return rows;
    }
}

