package lmu.utils;

import java.util.*;
import java.sql.*;

import org.jdom.*;

public class EPreparedStatement
{
    Logger log = LogConfig.getLogger();
    HashMap<String,Integer> fieldmap = new HashMap<String,Integer>();
    HashMap<Integer, String> inverse_map = new HashMap<Integer, String>();
    HashMap<String, String> field2value = new HashMap<String, String>();
    String query;
    PreparedStatement ps;


    public EPreparedStatement(Connection con, String query, String ... fields) 
    {
        this.query=query;
        if (con!=null)
        {
            try
            {
                ps = con.prepareStatement(query, Statement.RETURN_GENERATED_KEYS);
            }
            catch(SQLException e)
            {
                throw new RuntimeException("error while preparing query: "+query, e);
            }
            
            
        }
        
        int i=0;
        for (String s :fields)
        {
            ++i;
            inverse_map.put(i, s);
            fieldmap.put(s, i);
        }
    }

    public PreparedStatement getPS()
    {
        return ps;
    }
    
        
    public int getGeneratedId()
    {
        return SQLUtils.getGeneratedId(ps);

    }
    public void readAttribs(Element elem)
    {
        for (String s: fieldmap.keySet())
        {
            String v = XMLUtils.decodeAttrib(elem, s);
            if (v==null)
                continue;
            
            field2value.put(s,v);
        }
    }
    public Integer getInt(String field)
    {
        String v = field2value.get(field);
        if (v==null)
            return null;
        
        return Integer.parseInt(v);
    }
    public Double getDouble(String field)
    {
        String v = field2value.get(field);
        if (v==null)
            return null;
        
        return Double.parseDouble(v);
    }
    public String getString(String field)
    {

        return field2value.get(field);
        
    }
    public void writeAttribs(Element elem)
    {
        for (Map.Entry<String,String> s : field2value.entrySet())
        {
            XMLUtils.encodeAttrib(elem, s.getKey(), s.getValue());
        }
    }
    public void close() throws SQLException
    {
        ps.close();
    }
    public boolean execute() throws SQLException
    {
        try
        {
            return ps.execute();
        }
        catch(SQLException ex)
        {
            //check inputs
            StringBuffer nonsetparams = new StringBuffer();
            for (String field: fieldmap.keySet())
            {
                String v = field2value.get(field);
                if (v!=null)
                    continue;
                
                if (nonsetparams.length()>0)
                {
                    nonsetparams.append(", ");
                }
                nonsetparams.append(field);

            }
            if (nonsetparams.length()>0)
            {
                log.error("error while executing "+ps+"\nfollowing params are not set: "+nonsetparams.toString());
            }
            throw ex;
        }
        
    }

    public ResultSet executeQuery() throws SQLException
    {
        return ps.executeQuery();
    }

    int n2i(String fieldname)
    {
        Integer v = fieldmap.get(fieldname);
        if (v==null)
        {
            throw new RuntimeException("unknown fieldname: "+v+" query: "+query+" fields: "+fieldmap);
        }
        return v;
    }
    public void setInt(int i, int v) 
    {

        try
        {
        
            ps.setInt(i, v);
        }
        catch(SQLException e)
        {
            ExceptionManager.handleSQLException(log, "while setting param: ("+inverse_map.get(i)+":"+i+") to "+v+" into query: "+query, e);
        }
    }
    public void setInt(String fieldname, int v) 
    {
        field2value.put(fieldname, ""+v);
        if (ps==null)
        {
            return;
        }
        setInt(n2i(fieldname), v);
    }


    public int getIndexForField(String fieldname)
    {
        return n2i(fieldname);
    }
    public void setString(String fieldname, String v)
    {

        field2value.put(fieldname, (v==null) ? null : ""+v);
        if (ps==null)
        {
            return;
        }

        setString(n2i(fieldname),v);
            
        
        
    }

    void ex(SQLException e, int i, String v)
    {
        ExceptionManager.handleSQLException(log, "while setting param: ("+inverse_map.get(i)+":"+i+") to "+v+" into query: "+query, e);
    }
    public void setString(int i, String v) 
    {
        try
        {
            if (v == null)
            {
                ps.setNull(i, java.sql.Types.CHAR);
            }
            ps.setString(i, v);
        }
        catch(SQLException e)
        {
            ExceptionManager.handleSQLException(log, "while setting param: ("+inverse_map.get(i)+":"+i+") to "+v+" into query: "+query, e);
        }
    }

    public void setFloat(String fieldname, float v) 
    {
        field2value.put(fieldname, ""+v);
        
        if (ps==null)
        {
            return;
        }

        setFloat(n2i(fieldname), v);
    }

    public void setFloat(int i, float v) 
    {
        try
        {
            ps.setFloat(i, v);
        }
        catch(SQLException e)
        {
            ex(e,i,""+v);
        }
        
    }

    public void setDouble(String fieldname, double v) 
    {
        field2value.put(fieldname, ""+v);
        
        if (ps==null)
        {
            return;
        }
        setDouble(n2i(fieldname), v);
    }

    public void setDouble(int i, double v) 
    {
        try
        {
        
            ps.setDouble(i, v);
        }
        catch(SQLException e)
        {
            ex(e,i,""+v);
        }
    }

    public String toString()
    {
        return ""+ps;
    }
}

