package lmu.utils;
import java.io.ByteArrayOutputStream;
import java.io.PrintStream;

import java.sql.*;

public class ExceptionManager 
{

  public static String getStackTrace(Throwable e) 
  {

    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    PrintStream ps = new PrintStream(baos);
    e.printStackTrace(ps);

    return baos.toString();
  }

  public static void handleSQLException(Logger log, String message, java.sql.SQLException e)
  {
      ExceptionManager.log(log, e);
      ExceptionManager.log(LogConfig.getSQLLogger(), e);
      throw new RuntimeException(message, e);
  }
  
   public static String getCallerInfo(int depth)
   {
           
      RuntimeException re = new RuntimeException("caller stack");
      StackTraceElement[] stea = re.getStackTrace();
      String cn = stea[depth].getClassName();
      String cnp[] = cn.split("\\.");
      return cnp[cnp.length-1]+":"+stea[depth].getLineNumber();
  }
  public static String logCallerStack()
   {
       RuntimeException re = new RuntimeException("caller stack");
       StackTraceElement[] stea = re.getStackTrace();
       StringBuffer sb = new StringBuffer();

       for (int i=1; i<stea.length; i++)
       {
           StackTraceElement ste = stea[i];
           sb.append("\t"+ste);
           sb.append('\n');
       }
       return sb.toString();
   }

  public static String info()
  {
      RuntimeException re = new RuntimeException("caller stack");
      StackTraceElement[] stea = re.getStackTrace();
      return   stea[1].getFileName()+":"+stea[1].getLineNumber();
  }
  public static int ln()
  {
      return getLineNo();
  }
  
  public static int getLineNo()
  {
      RuntimeException re = new RuntimeException("caller stack");
      StackTraceElement[] stea = re.getStackTrace();
      return stea[1].getLineNumber();
  }
    public static String getString(Throwable e)
    {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        PrintStream ps = new PrintStream(baos);
        e.printStackTrace(ps);
        return baos.toString();
    }
    public static void log(Logger where, Throwable e)
    {
        where.error(getString(e));
    }

    public static void throwRuntimeException(String fmt, Object[] params)
    {
        throwRuntimeException(fmt,params,null);
    }
    public static void throwRuntimeException(String fmt, Object[] params, Exception e)
    {
        throwRuntimeException(String.format(fmt, params), e);
    }
    public static void throwRuntimeException(String msg)
    {
        throwRuntimeException(msg,(Exception)null);
    }
    public static void throwRuntimeException(String msg, Exception e)
    {
        RuntimeException re = new RuntimeException("caller stack");
        int index=(e==null)?2:1;
        StackTraceElement[] stea = re.getStackTrace();
        String from=stea[index].getFileName()+":"+stea[index].getLineNumber();

        if (e==null)
        {
            throw new RuntimeException(from+"\n"+msg);
        }

        throw new RuntimeException(from+"\n"+msg, e);
        
    }

    public static String getResultSetInfo(ResultSet rs )
    {
        try
        {
            StringBuffer sb = new StringBuffer();
            ResultSetMetaData rsmd = rs.getMetaData();
            
            for (int i=1; i<=rsmd.getColumnCount(); i++)
            {
                sb.append('\t');
                sb.append(rsmd.getColumnName(i));
                sb.append(" => ");
                sb.append(rs.getString(i));
                sb.append('\n');
            }
            return sb.toString();
        }
        catch(SQLException e)
        {
            return "error while getting info about resultset: "+e.getMessage();
        }

        
    }

}