package lmu.utils;

import java.io.*;
import java.util.*;
import java.util.regex.*;
import lmu.utils.*;
import javax.script.*;

public class Expression 
{

    Expression root_expr;
    Logger log = LogConfig.getLogger();
    /*Priority 	Operators 	Operation 	Associativity
        1 	[ ] 	array index 	left
        () 	method call
        . 	member access
        
        --------------------------------------------------                                     
        2 	++ 	pre- or postfix increment 	right
        -- 	pre- or postfix decrement
        + - 	unary plus, minus
        ~ 	bitwise NOT
        ! 	boolean (logical) NOT
        (type) 	type cast
        new 	object creation
        
        
        --------------------------------------------------
        
        
        
        3 	* / % 	multiplication, division, remainder 	left
        
        
        --------------------------------------------------
        
        4 	+ - ^ 	addition, substraction 	left
        
        
        --------------------------------------------------
        + 	string concatenation
        
        --------------------------------------------------
        
        5 	<< 	signed bit shift left 	left
        >> 	signed bit shift right
        >>> 	unsigned bit shift right
        
        
        --------------------------------------------------
        
        
        6 	< <= 	less than, less than or equal to 	left
            > >= 	greater than, greater than or equal to
            instanceof 	reference test
            
            
        --------------------------------------------------
        
        
            
        7 	== 	equal to 	left
        != 	not equal to
        
        
        --------------------------------------------------
        
        
        
        8 	& 	bitwise AND 	left
        & 	boolean (logical) AND
        
        
        --------------------------------------------------
        
        9 	^ 	bitwise XOR 	left
        ^ 	boolean (logical) XOR
        
        
        --------------------------------------------------
        
        
        10 	| 	bitwise OR 	left
        | 	boolean (logical) OR
        
        
        --------------------------------------------------
        
        
        11 	&& 	boolean (logical) AND 	left
        
        --------------------------------------------------
        
        
        12 	|| 	boolean (logical) OR 	left
        
        --------------------------------------------------
        
        13 	? : 	conditional 	right
        
        --------------------------------------------------
        
        
        14 	= 	assignment 	right
        *= /= += -= %=
        <<= >>= >>>=
        &= ^= |= 	combinated assignment
        (operation and assignment)
        
        
        -------------------------------------------------
    */

    public interface EvalInterface 
    {
        public  Object eval(Object left, Object right);
    }


    HashMap<String, Vector<Operator>> overload_symbol2op = new HashMap<String, Vector<Operator>>();
    static HashMap<String, Operator> symbol2op = new HashMap<String, Operator>();
    static HashMap<String, AggregatorFunction> symbol2aggregator = new HashMap<String, AggregatorFunction>();
    HashMap<String, Vector<AggregatorFunction> > overload_aggregators = new HashMap<String, Vector<AggregatorFunction> >();

    public static class FakeOperator extends Operator
    {
        FakeOperator(String symbol)
        {
            super(0,0,0,symbol);
        }
        public Object eval(Object left, Object right)
        {
            return null;
        }
    }

    static class OverloadedOperator
    {
        Class leftOperandType;
        Class rightOperandType;
        Operator op;

    }
    public void registerOperator(Operator op)
    {
        Vector<Operator> v = overload_symbol2op.get(op.symbol);
        if (v==null)
        {
            overload_symbol2op.put(op.symbol, v=new Vector<Operator>());
        }
        v.add(op);

        log.debug("register "+op.symbol+" to operands: "+op.leftOperandType+" "+op.rightOperandType);
        log.debug("overload_symbol2op: "+overload_symbol2op);


    }
    static FakeOperator OPEN_PAR = new FakeOperator("(");
    static FakeOperator CLOSE_PAR = new FakeOperator(")");

    public static abstract class AggregatorFunction
    {
        String name;
        Class _type=null;

        public AggregatorFunction(String name)
        {
            this.name=name;
            symbol2aggregator.put(name, this);
        }

        public abstract Object eval(Collection data);
    }

    static Object getMinMax(Collection data, boolean min)
    {
        Comparable maxval = null;
        Comparable minval = null;
        for (Object o : data)
        {
            if (o==null)
                continue;


            if (! ( o instanceof Comparable ))
            {
                throw new RuntimeException("cannot apply "+((min)?"min":"max")+" on type: "+o.getClass().getName());
            }

            Comparable c = (Comparable)o;
            if (maxval==null || c.compareTo(maxval)>0)
            {
                maxval=c;
            }

            if (minval==null || c.compareTo(minval)<0)
            {
                minval=c;
            }

        }

        return(min)?minval:maxval;
    }

    public static class MaxAggregator extends AggregatorFunction
    {

        MaxAggregator()
        {
            super("max");
        }

        public Object eval(Collection data)
        {
            return getMinMax(data, false);
        }
    }
    static MaxAggregator maxA = new MaxAggregator();

    public static class MinAggregator extends AggregatorFunction
    {

        MinAggregator()
        {
            super("min");
        }

        public Object eval(Collection data)
        {
            return getMinMax(data, true);
        }
    }
    static MinAggregator minA = new MinAggregator();

    public static class CountAggregator extends AggregatorFunction
    {

        CountAggregator()
        {
            super("count");
        }

        public Object eval(Collection data)
        {
            return(data==null)?0:data.size();
        }
    }
    static CountAggregator countA=new CountAggregator();

    public static abstract class Operator implements EvalInterface
    {
        int priority;
        int nl;
        int nr;
        String symbol;
        Logger log = LogConfig.getLogger();

        Class leftOperandType;
        Class rightOperandType;


        public Operator(int prio, int nl, int nr, String symbol, Class leftOperandType, Class rightOperandType)
        {
            this.leftOperandType=leftOperandType;
            this.rightOperandType=rightOperandType;

            this.priority=prio;
            this.nl=nl; this.nr=nr;
            this.symbol=symbol;

        }

        private Operator(int prio, int nl, int nr, String symbol)
        {
            this.priority=prio;
            this.nl=nl; this.nr=nr;
            this.symbol=symbol;

            //System.out.println("put op: "+this.hashCode()+" to symbol : "+symbol);
            symbol2op.put(symbol,this);
        }

        public boolean canEvalShortCut(Object leftOperand)
        {
            return false;
        }
        public Object evalShortCut(Object leftOperand)
        {
            return null;
        }
        public int getPriority()
        {
            return priority;
        }
        public int getNumLeftOperands()
        {
            return nl;
        }
        public int getNumRightOperands()
        {
            return nr;
        }
        public abstract Object eval(Object left, Object right);


        public Object parse(String s)
        {
            return null;
        }

        static int getFirstIndex(String str)
        {
            if (str==null) return -1;


            int idx=str.length();

            for (String symbol: symbol2op.keySet())
            {
                int hiti=str.indexOf(symbol);
                if (hiti<0)
                    continue;

                idx=Math.min(hiti, idx);
            }

            return idx;
        }

        public static Operator getOperator(String symbol)
        {
            return symbol2op.get(symbol);

        }


        static Operator parseToken(String s)
        {
            Operator bestOp = null;
            for (Operator op: symbol2op.values())
            {
                if (!s.startsWith(op.symbol))
                    continue;


                if (bestOp==null || bestOp.symbol.length()<op.symbol.length())
                {
                    bestOp=op;
                }

            }

            return bestOp;
        }

        public String toString()
        {
            return "OP["+symbol+" prio: "+priority+"]";
        }
    }


    public static abstract  class BoolOperator extends Operator
    {

        public BoolOperator(int prio, int nl, int nr, String symbol)
        {
            super(prio,nl,nr,symbol);
        }
        public Object parse(String s)
        {
            if (s==null || s.equals("null"))
                return null;


            return Boolean.parseBoolean(s);
        }
    }


    public static class AndOperator extends BoolOperator
    {

        public AndOperator()
        {
            super(11,1,1, "&&");
        }

        public boolean canEvalShortCut(Object leftOperand)
        {
            return !(Boolean)leftOperand;
        }
        public Object evalShortCut(Object leftOperand)
        {
            //only if leftOperand false
            boolean b = (Boolean)leftOperand;
            if (b)
                throw new RuntimeException("program error, AND shortcut only works for false left operand!");

            return false;
        }


        public Object eval(Object left, Object right)
        {
            return((Boolean)left) && ((Boolean)right);
        }
    }
    static AndOperator andOp = new AndOperator();
    public static class OrOperator extends BoolOperator
    {

        public OrOperator()
        {
            super(12,1,1, "||");
        }

        public boolean canEvalShortCut(Object leftOperand)
        {
            return(Boolean)leftOperand;
        }
        public Object evalShortCut(Object leftOperand)
        {
            //only if leftOperand false
            boolean b = (Boolean)leftOperand;
            if (!b)
                throw new RuntimeException("program error, OR shortcut only works for true left operand!");

            return true;
        }


        public Object eval(Object left, Object right)
        {
            return((Boolean)left) || ((Boolean)right);
        }
    }
    static OrOperator orOp = new OrOperator();

    public static class NotOperator extends BoolOperator
    {
        public NotOperator()
        {
            super(2,0,1, "!");
        }

        public Object eval(Object left, Object right)
        {
            return !((Boolean)right);
        }
    }
    static NotOperator notOp = new NotOperator();

    public static abstract class NumberBasedOperator extends Operator
    {
        public NumberBasedOperator(int prio, int nl, int nr, String symbol)
        {
            super(prio,nl,nr,symbol);
        }

        boolean accept_string=false;

        public Object parse(String s)
        {
            if (s==null || s.equals("null"))
                return null;

            try
            {
                return Integer.parseInt(s);
            }
            catch (NumberFormatException nfe)
            {

            }
            try
            {
                return Short.parseShort(s);
            }
            catch (NumberFormatException nfe)
            {

            }


            try
            {
                return Long.parseLong(s);
            }
            catch (NumberFormatException nfe)
            {

            }

            String ls = s.replaceAll(",", ".");
            try
            {
                return Double.parseDouble(ls);
            }
            catch (NumberFormatException nfe)
            {

            }

            try
            {
                return Float.parseFloat(ls);
            }
            catch (NumberFormatException nfe)
            {

            }
            if (accept_string)
            {
                return s;
            }
            throw new RuntimeException("cannot parse numeric value from >"+s+"<");
        }
    }

    public static class MatchOperator extends Operator
    {
        public MatchOperator()
        {
            super(4,1,1,"~");

        }
        public Object parse(String s)
        {
            if (s==null || s.equals("null"))
                return null;

            return Pattern.compile(s);
        }

        public Object eval(Object left, Object right)
        {
            Pattern p = (Pattern)((left instanceof Pattern)?left:right);
            //String s = (String)((left instanceof String)?left:right);
            String s = (String)((left == p) ? right : left);
            if (s == null || s.length() == 0)
                return false;
            
            return p.matcher(s).find();
        }
    }

    static MatchOperator matchOperator = new MatchOperator();

    static Class getMaxDetailNumber(Class c1, Class c2)
    {
        if (c1 == Double.class || c2 == Double.class)
        {
            return Double.class;
        }

        if (c1 == Float.class || c2 == Float.class)
        {
            return Float.class;
        }

        if (c1 == Long.class || c2 == Long.class)
        {
            return Long.class;
        }

        if (c1 == Integer.class || c2 == Integer.class)
        {
            return Integer.class;
        }

        throw new RuntimeException("no number class: "+c1+" or: "+c2);

    }


    abstract static class CompareOperator extends NumberBasedOperator
    {
        public CompareOperator(String symbol)
        {
            this(symbol, 6);
        }

        public CompareOperator(String symbol,int prio)
        {
            super(prio,1,1, symbol);
        }
        abstract public Object eval(int comv_val);

        public Object eval(Object left, Object right)
        {
            //System.out.println(this.getClass().getName()+" compare: "+left+ " left is null: "+ (left==null)+" to "+right+" right is null: "+ (right==null));

            if (left==null && right==null)
            {
                //System.out.println("got compare null vs null return 0!");
                return eval(0);
            }
            if (left==null)
                return eval(-1);

            if (right==null)
                return eval(1);


            if (left.getClass() != right.getClass())
            {
                //check
                Class max =  getMaxDetailNumber(left.getClass(), right.getClass());
                left = NumUtils.convertNumber(left, max);
                right = NumUtils.convertNumber(right, max);

            }
            //System.out.printf("check %s vs %s\n", (Comparable)left, (Comparable)right);

            return eval(((Comparable)left).compareTo((Comparable)right));

        }

    }


    public static class GTOperator extends CompareOperator
    {

        public GTOperator()
        {
            super(">");
        }
        public Object eval(int c)
        {
            return c>0;
        }

    }
    static GTOperator gotOp = new GTOperator();


    public static class GTEOperator extends CompareOperator
    {
        public GTEOperator()
        {
            super(">=");
        }

        public Object eval(int c)
        {
            return c>=0;
        }

    }
    static GTEOperator goteOp = new GTEOperator();

    public static class LTOperator extends CompareOperator
    {

        public LTOperator()
        {
            super("<");
        }

        public Object eval(int c)
        {
            return c<0;
        }

    }
    static LTOperator ltOp = new LTOperator();

    public static class LTEOperator extends CompareOperator
    {

        public LTEOperator()
        {
            super("<=");
        }

        public Object eval(int c)
        {
            return c<=0;
        }

    }
    static LTEOperator lteOp = new LTEOperator();

    public static class NEQOperator extends CompareOperator
    {
        public NEQOperator()
        {
            super("!=", 7);
            accept_string=true;
        }
        public Object eval(int c)
        {
            //System.out.println("NEQ got : "+c+" will return: "+(c!=0));
            return c!=0;
        }

    }

    static NEQOperator neqOp = new NEQOperator();

    public static class EQOperator extends CompareOperator
    {
        public EQOperator()
        {
            super("==", 7);
            accept_string=true;
        }
        public Object eval(int c)
        {
            return c==0;
        }

    }


    static EQOperator eqOp = new EQOperator();

    static class IntPlus implements EvalInterface
    {
        public Object eval(Object left, Object right)
        {
            return((Integer)left)+((Integer)right);
        }
    }

    static class IntMinus implements EvalInterface
    {
        public Object eval(Object left, Object right)
        {
            return((Integer)left)-((Integer)right);
        }
    }

    static class DoubleMinus implements EvalInterface
    {
        public Object eval(Object left, Object right)
        {
            return((Double)left)-((Double)right);
        }
    }

    static class DoublePus implements EvalInterface
    {
        public Object eval(Object left, Object right)
        {

            return((Double)left)+((Double)right);
        }
    }




    static double toDouble(Object o)
    {
        try
        {
            return(Double)o;
        }
        catch (ClassCastException cce)
        {
            if (o instanceof Integer)
            {
                return(double)((Integer)o);
            }
        }

        throw new RuntimeException("cannot convert to double: "+o);
    }
    static double toFloat(Object o)
    {
        try
        {
            return(Float)o;
        }
        catch (ClassCastException cce)
        {
            if (o instanceof Integer)
            {
                return(float)((Integer)o);
            }
        }

        throw new RuntimeException("cannot convert to double: "+o);
    }


    public static class PlusOperator extends NumberBasedOperator
    {
        //EvalInterface eval_helper;
        boolean inited=false;
        public PlusOperator()
        {
            super(4, 1, 1, "+");
        }


        void init(Object left, Object right)
        {
            if (inited)
                return;


        }
        public Object eval(Object left, Object right)
        {
            if (left instanceof Set)
            {
                HashSet rv = new HashSet();
                rv.addAll((Set)left);
                rv.addAll((Set)right);
                return rv;
            }

            return((Number)left).doubleValue()+((Number)right).doubleValue();
            /*
            if (left instanceof Double || right instanceof Double)
            {
                return toDouble(left)+toDouble(right);
            }
            if (left instanceof Float || right instanceof Float)
            {
                return toFloat(left)+toFloat(right);
            }


            return (int)(left)+(int)(right);
            */

        }
    }


    static PlusOperator plusOp = new PlusOperator();

    public static class MinusOperator extends NumberBasedOperator
    {
        //EvalInterface eval_helper;
        boolean inited=false;
        public MinusOperator()
        {
            super(4, 1, 1, "-");
        }


        void init(Object left, Object right)
        {
            if (inited)
                return;


        }
        public Object eval(Object left, Object right)
        {
            if (left instanceof Set)
            {
                HashSet rv = new HashSet();
                Set s1 = (Set)left;
                Set s2 = (Set)right;

                rv.addAll(s1);
                for (Object o : s2)
                {
                    rv.remove(o);
                }
                return rv;
            }
            return((Number)left).doubleValue()-((Number)right).doubleValue();
            //return ((double)left)-(double)right;

        }
    }

    static MinusOperator minusOp = new MinusOperator();

    public static class DivideOperator extends NumberBasedOperator
    {
        //EvalInterface eval_helper;
        boolean inited=false;
        public DivideOperator()
        {
            super(3, 1, 1, "/");
        }


        void init(Object left, Object right)
        {
            if (inited)
                return;


        }
        public Object eval(Object left, Object right)
        {
            return((Double)left)/(Double)right;

        }
    }

    static DivideOperator divOp = new DivideOperator();


    public static class IntersectOperator extends Operator
    {
        //EvalInterface eval_helper;

        public IntersectOperator()
        {
            super(3, 1, 1, "^");
        }

        public Object eval(Object left, Object right)
        {
            Set s1=(Set)left;
            Set s2=(Set)right;

            Set smaller = (s1.size()<s2.size())?s1:s2;
            Set bigger = (s2.size()<s1.size())?s1:s2;

            HashSet rv = new HashSet();
            for (Object o : smaller)
            {
                if (!bigger.contains(o))
                    continue;
                rv.add(o);
            }
            //System.out.printf("intersect %d vs %d -> %d\n", smaller.size(), bigger.size(), rv.size());
            return rv;

        }
    }

    static IntersectOperator intersectOp = new IntersectOperator();


    public static class MultiplyOperator extends NumberBasedOperator
    {
        //EvalInterface eval_helper;
        boolean inited=false;
        public MultiplyOperator()
        {
            super(3, 1, 1, "*");
        }


        void init(Object left, Object right)
        {
            if (inited)
                return;


        }
        public Object eval(Object left, Object right)
        {
            return((Double)left)*(Double)right;

        }
    }

    static MultiplyOperator multOp = new MultiplyOperator();



    List tokens;
    ExprNode root = null;
    Vector expr_fields = new Vector(); //the elements can be ExpressionLeaf, Expression, Operator

    Tokenizator tokenizer=null;

    public void reset()
    {
        tokenizer=null;
        expr_fields.clear();
        tokens=null;
        root=null;
    }

    static public<A> A quickEval(String expr, Object ... variables)
    {
        Expression e = new Expression();
        return (A)e.evalExpression(expr, variables);
    }

    public<A> A evalExpression(String expr, Object ... variables)
    {
        Matcher ma = numPat.matcher(expr);
        
        StringBuffer exp=new StringBuffer();
        int start=0;
        HashMap<Integer, Object> resolved = new HashMap<>();
        while (ma.find(start))
        {
            String val=ma.group();
        
            int ref_idx = Integer.parseInt(val);
            if (resolved.get(ref_idx) == null)
            {
                if (ref_idx >= variables.length)
                {
                    throw new FRuntimeException("invalid expression: >%s< it references variable: %d but only %d variables are given", expr, ref_idx, variables.length);
                }
                resolved.put(ref_idx, variables[ref_idx]);
            }
            exp.append(expr.substring(start, ma.start()));
            exp.append("${"+ref_idx+"}");
            start=ma.end();
        
        }
        exp.append(expr.substring(start));
        
        HashMap<String, ReflectedObject> m = new HashMap<String, ReflectedObject>();
        for (int idx : resolved.keySet())
        {
            m.put(""+idx, new ReflectedObject(true));
        }
        
        setExpression(exp.toString(), m);
        
        for(Map.Entry<Integer, Object> e: resolved.entrySet())
        {
            assign(""+e.getKey()+"", e.getValue());
        }
        return (A)eval();
        

    }
    public void setExpression(String expr)
    {
        setExpression(expr,null);
    }

    public void setExpression(String expr, HashMap<String, ReflectedObject> namedvars)
    {
        tokenizer=new Tokenizator(expr, namedvars);
        Vector tokens = tokenizer.tokens;
        //System.out.println("tokens: "+tokens);
        //System.out.printf("got %d refobjs\n", tokenizer.default_refobjs.size());
        init(tokens, 0, tokens.size()-1);

    }

    public Expression()
    {
        this.root_expr=this;
        aggregator_root=true;
    }
    public Expression(String expr)
    {
        this(expr,null);

    }
    public Expression(String expr, HashMap<String, ReflectedObject> namedvars)
    {
        this(expr, namedvars, null, true);
    }
    String input_expression=null;

    public Expression(String expr, HashMap<String, ReflectedObject> namedvars, Expression root_expr, boolean aggregator_root)
    {
        input_expression=expr;
        this.root_expr=(root_expr==null)?this:root_expr;;
        this.aggregator_root=aggregator_root;
        setExpression(expr,namedvars);

    }
    public String getInputExpression()
    {
        return input_expression;
    }


    Vector<Vector<ExprNode> > ordered_nodes = null;

    public void setOrder(String ... names)
    {
        setOrder(new CollectionIterator(names));
    }
    public void setOrder(Vector<String>  names)
    {
        setOrder(names.iterator());
    }
    public void setOrder(Iterator<String> names)
    {
        ordered_nodes= new Vector<Vector<ExprNode> >();

        while (names.hasNext())
        {
            String n=names.next();

            if (n==null)
            {
                ordered_nodes.add(tokenizer.default_refobjs);
                continue;
            }
            Vector<ExprNode> v = tokenizer.named_refobjs.get(n);
            if (v==null)
            {
                ordered_nodes.add(null);
                continue;
                //throw new RuntimeException("unknown variable: "+n);
            }
            ordered_nodes.add(v);
        }
    }
    public void assign(String name, Object o)
    {
        Vector<ExprNode> v = tokenizer.named_refobjs.get(name);
        if (v==null)
        {
            //System.out.println("namedvars: "+tokenizer.namedvars+"  name: "+name);
            if (tokenizer.namedvars.get(name)!=null)
            {
                return;
            }
            throw new RuntimeException("unknown variable: "+name);
        }
        for (ExprNode en: v)
        {
            en.setConstant(en.ref_obj.get(o));
        }
    }

    public static Vector quickApply(String apply_expr, String filter_expr, Collection var2values, boolean uniq)
    {
        Collection toapplyon = (filter_expr==null)?var2values:quickFilter(filter_expr,var2values);
        Expression e = new Expression(apply_expr);
        Vector v = new Vector();
        v.addAll(toapplyon);
        v= e.apply(v, false);
        if (!uniq)
        {
            return v;
        }
        HashSet hs = new HashSet();
        hs.addAll(v);
        v.clear();
        v.addAll(hs);
        return v;
    }
    public static Vector quickFilter(String expression, Collection var2values)
    {
        Expression e = new Expression(expression);
        return e.filter(var2values);
    }
    /** filters by the expression assuming itworks if we assign only
     * values to the default variable 
     */
    public Vector filter(Collection var2values)
    {
        Vector v = new Vector();
        for (Object o : var2values)
        {
            if (!filter(o))
                continue;

            v.add(o);
        }
        return v;
    }

    public Vector filter(Vector default_values, HashMap<String, Vector> var2values, HashMap<String, Object> var2static_values)
    {
        return apply(default_values,var2values,var2static_values,true);
    }

    public Vector apply(Vector default_values, boolean filter)
    {
        return apply(default_values,null, null, false);
    }
    /** processes all entries in the vectors and results the */
    public Vector apply(Vector default_values, HashMap<String, Vector> var2values, HashMap<String, Object> var2static_values, boolean filter)
    {
        Vector rv = new Vector();
        if (var2static_values!=null)
        {
            for (Map.Entry<String, Object> e : var2static_values.entrySet())
            {
                assign(e.getKey(), e.getValue());
            }
        }
        for (int i=0; i<default_values.size(); i++)
        {
            Object o = default_values.get(i);
            for (ExprNode n : tokenizer.default_refobjs)
            {
                n.setConstant(n.ref_obj.get(o));
            }
            if (var2values!=null)
            {
                for (Map.Entry<String, Vector> e : var2values.entrySet())
                {
                    assign(e.getKey(),e.getValue().get(i));
                }
            }

            Object res = eval();
            if (filter )
            {
                if (!(Boolean)res)
                    continue;
                rv.add(o);
                continue;
            }
            rv.add(res);
        }

        return rv;

    }
    public boolean filter(Object ... applyOn)
    {
        return(Boolean)eval(applyOn);
    }

    public boolean check_null(Object ... applyOn)
    {
        if (applyOn==null || applyOn.length==0)
            return false;

        if (ordered_nodes!=null)
        {
            for (int i=0; i<applyOn.length; i++)
            {
                if (null==ordered_nodes.get(i))
                    continue;

                for (ExprNode n: ordered_nodes.get(i))
                {
                    if (n.ref_obj.get(applyOn[i]) == null)
                        return true;
                }

            }
            return false;
        }
        for (ExprNode n : tokenizer.default_refobjs)
        {
            if (n.ref_obj==null)
            {
                if (applyOn[0] == null)
                    return true;

                continue;
            }
            if (n.ref_obj.get(applyOn[0]) == null)
                return true;

            //System.out.println("set constant :"+n.constant+" for "+applyOn[0]);
        }
        return false;
    }
    public Object eval(Object ... applyOn)
    {

        if (applyOn==null || applyOn.length==0)
        {
            return root.getValue();
        }
        if (ordered_nodes!=null)
        {
            for (int i=0; i<applyOn.length; i++)
            {
                if (null==ordered_nodes.get(i))
                    continue;

                for (ExprNode n: ordered_nodes.get(i))
                {
                    n.setConstant(n.ref_obj.get(applyOn[i]));
                }
            }
            return root.getValue();
        }
        for (ExprNode n : tokenizer.default_refobjs)
        {
            if (n.ref_obj==null)
            {
                n.setConstant(applyOn[0]);
                continue;
            }
            n.setConstant(n.ref_obj.get(applyOn[0]));

            //System.out.println("set constant :"+n.constant+" for "+applyOn[0]);
        }
        return root.getValue();
    }
    public String toString()
    {
        return "\n\tEXPR: >> "+expr_fields+" <<";
    }

    boolean aggregator_root=false;

    public Expression(Vector tokens, int start, int end, Expression root, boolean aggregator_root)
    {
        this.root_expr=root;
        this.aggregator_root=aggregator_root;
        init(tokens,start,end);

    }

    void reinit(Expression e)
    {
        this.tokens=e.tokens;
        root=e.root;
        expr_fields=e.expr_fields;
    }
    void init(Vector tokens, int start, int end)
    {
        //System.out.printf("\n\ninit EXPR: %d %d-%d\n", tokens.size(),start,end);

        this.tokens = tokens.subList(start, end+1);

        //System.out.println("\n\n INIT tokens: "+this.tokens+"\n\n");

        for (int i=start; i<=end; i++)
        {
            if (tokens.get(i)==OPEN_PAR)
            {
                int j=i+1;
                int nopen=1;
                for (; j<=end && nopen>0; j++)
                {
                    if (tokens.get(j)==OPEN_PAR)
                    {
                        nopen++;
                        continue;
                    }
                    if (tokens.get(j)==CLOSE_PAR)
                    {
                        if (--nopen==0)
                        {
                            break;
                        }
                        continue;
                    }
                }
                if (nopen!=0)
                {
                    throw new RuntimeException("no close paranthesis found to expression: "+ tokens.subList(i,end));
                }
                expr_fields.add(new Expression(tokens, i+1,j-1, root_expr, false));
                i=j;
                continue;

            }

            if (tokens.get(i) instanceof Operator)
            {
                expr_fields.add(tokens.get(i));
                continue;
            }
            expr_fields.add(tokens.get(i));

        }


        //System.out.println("expr with tokens: "+this.tokens);
        log.debug("after parsing got "+expr_fields.size()+" expr fields: "+expr_fields);


        int iter=0;

        while (true)
        {
            iter++;


            //get the leftmost lowest prio
            Operator best_op=null;
            int best_op_index=-1;
            for (int i=0; i<expr_fields.size(); i++)
            {
                if (! (expr_fields.get(i) instanceof Operator))
                    continue;


                Operator op = (Operator)expr_fields.get(i);

                if (best_op==null || best_op.getPriority()>op.getPriority())
                {
                    best_op=op;
                    best_op_index=i;
                }

            }

            //System.out.println("expression: "+this.tokens+" iter: "+iter+" next op: "+best_op);
            if (best_op==null)
            {
                break;

            }

            //consume the params
            if (best_op_index<best_op.getNumLeftOperands())
            {
                throw new RuntimeException("op: "+best_op+" needs x params but got only");
            }
            if (expr_fields.size()-best_op_index-1<best_op.getNumRightOperands())
            {
                throw new RuntimeException("op: "+best_op+" needs x params but got only");
            }

            int toskip_start=best_op_index-best_op.getNumLeftOperands();
            int toskip_end=best_op_index+best_op.getNumRightOperands();


            Object left_op = null;
            if (best_op.getNumLeftOperands()>0)
            {
                left_op=expr_fields.get(best_op_index-1);
            }

            Object right_op = null;
            if (best_op.getNumRightOperands()>0)
            {
                right_op=expr_fields.get(best_op_index+1);
            }

            //System.out.printf("best op index %d all: %d toskip :%d-%d\n", best_op_index, expr_fields.size(), toskip_start, toskip_end);
            //System.out.println("leftr op: "+left_op);
            //System.out.println("rhight op: "+right_op);
            ExprNode en = null;
            
            try
            {
                en = new ExprNode(best_op, left_op, right_op);
            }
            catch(Exception _ex)
            {
                throw new FRuntimeException("error: %s at building node op: %s left: %s right: %s", _ex.getMessage(), 
                                            best_op, left_op, right_op);
            }
            //System.out.println("created new expr node: ");
            //System.out.println(en.print(""));
            root=en;
            Vector nl = new Vector();
            for (int j=0; j<expr_fields.size(); j++)
            {
                if (j==best_op_index)
                {
                    nl.add(en);

                }
                if (j>=toskip_start && j<=toskip_end)
                {
                    continue;
                }
                nl.add(expr_fields.get(j));

            }
            expr_fields=nl;

            //System.out.println("after iter : "+iter+" new expr fields: "+expr_fields);
        }



        if (root==null)
        {
            //System.out.println("no root found yet...");

            if (expr_fields.size()!=1)
            {
                throw new RuntimeException("no operand found, expected constant or ref obj as expression got: "+this.tokens);

            }

            if (expr_fields.get(0) instanceof Expression)
            {
                reinit( ((Expression)expr_fields.get(0)));
            }
            else
            {
                root = (ExprNode)expr_fields.get(0);
            }
            //System.out.println("only field type. "+expr_fields.get(0).getClass().getCanonicalName());

            if (root instanceof AggratorExpression)
            {
                AggratorExpression aex = (AggratorExpression)root;
                if (aex.expr==null)
                {
                    aex.buildExpression(root_expr);
                }
            }

        }

        if (!aggregator_root)
            return;



        HashSet<AggratorExpression> aggregators = new HashSet<AggratorExpression>();
        root.collectAggregatorExpressions(aggregators);

        //System.out.println("root got "+aggregators.size()+" aggregators");
        for (AggratorExpression ae : aggregators)
        {
            tokenizer.merge(ae.expr.tokenizer);
        }



    }




    class ExprNode
    {

        ExprNode leftOperand;
        ExprNode rightOperand;

        ReflectedObject ref_obj;
        boolean constant_set=false;
        Object constant;
        String constant_def;

        Operator op;

        void setConstant(Object o)
        {
            constant=o;
            constant_set=true;
        }
        boolean inited=false;

        public ExprNode(ReflectedObject ref_obj)
        {
            this.ref_obj=ref_obj;
        }

        public ExprNode(String constant_def)
        {
            this.constant_def=constant_def;
        }
        public ExprNode()
        {

        }

        public void collectAggregatorExpressions(HashSet<AggratorExpression> aggregators )
        {
            //System.out.println("collect agr: "+this);
            if (this instanceof AggratorExpression)
            {
                AggratorExpression ae = (AggratorExpression)this;
                if (ae.apply_refobj!=null)
                {
                    return;
                }
                aggregators.add(ae);
                ae.expr.root.collectAggregatorExpressions(aggregators);
                return;
            }
            if (leftOperand!=null)
            {
                leftOperand.collectAggregatorExpressions(aggregators);
            }
            if (rightOperand!=null)
            {
                rightOperand.collectAggregatorExpressions(aggregators);
            }
        }
        ExprNode createNode(Object operand)
        {
            if (operand==null)
            {
                return null;
            }
            if (operand instanceof AggratorExpression)
            {
                AggratorExpression ae = (AggratorExpression) operand;
                //System.out.println("will build expression from : "+ae.expr_def);
                ae.buildExpression(root_expr);

                return ae;
            }

            if (operand instanceof ExprNode)
            {
                return(ExprNode)operand;
            }
            if (operand instanceof ReflectedObject)
            {
                return new ExprNode((ReflectedObject)operand);
            }
            if (operand instanceof String)
            {
                return new ExprNode((String)operand);
            }
            if (operand instanceof ExprNode)
            {
                return(ExprNode)operand;

            }
            if (operand instanceof Expression)
            {
                return((Expression)operand).root;
            }
            throw new RuntimeException("unknown operand type: " + operand.getClass().getName());
        }

        public ExprNode(Operator op, Object left_op, Object right_op)
        {
            this.op=op;

            leftOperand = createNode(left_op);
            rightOperand = createNode(right_op);

        }

        void init()
        {
            if (inited)
                return;

            inited=true;

            //System.out.println("init : left "+leftOperand+"  right: "+rightOperand+" op:" +op);
            if (leftOperand==null && rightOperand==null)
            {
                return;
            }



            //let the operator help
            if (leftOperand!=null && leftOperand.constant_def!=null)
            {
                leftOperand.setConstant(op.parse(leftOperand.constant_def));

            }
            if (rightOperand!=null & rightOperand.constant_def!=null)
            {
                rightOperand.setConstant(op.parse(rightOperand.constant_def));
            }

            Vector<Operator> v = root_expr.overload_symbol2op.get(op.symbol);
            //System.out.println("init on op: "+op.hashCode()+":"+op+" overlaod:"+((v==null)?0:v.size()));
            //System.out.println("overload: "+root_expr.overload_symbol2op);

            //check if we have an overladed operator
            if (v==null || v.size()==0)
            {
                return;
            }
            //System.out.println("will test overload types on "+op);
            Class leftOperandType = (leftOperand==null)?null:leftOperand.getValue().getClass();
            Class rightOperandType= (rightOperand==null)?null:rightOperand.getValue().getClass();

            for (Operator oo: v)
            {
                /*  System.out.printf("%s test overlaod type: %s %s on %s %s\n",
                                    op.symbol,
                           (oo.leftOperandType==null)?null:oo.leftOperandType.getName(),
                           (oo.rightOperandType==null)?null:oo.rightOperandType.getName(),
  
                           (leftOperandType==null)?null:leftOperandType.getName(),
                           (rightOperandType==null)?null:rightOperandType.getName()
                           );
                  */

                if (leftOperandType!=null && 
                    ( oo.leftOperandType==null || !oo.leftOperandType.isAssignableFrom(leftOperandType))
                   )
                {
                    continue;
                }
                if (rightOperandType!=null && 
                    ( oo.rightOperandType==null || !oo.rightOperandType.isAssignableFrom(rightOperandType))
                   )
                {
                    continue;
                }


                //System.out.println("found!");
                op=oo;
                break;

            }
            return;

        }
        public Object getValue()
        {
            if (constant_set)
                return constant;

            init();

            if (op==null)
            {
                throw new RuntimeException("invalid expression operator is null on:" +this);
            }


            Object lo = (leftOperand==null)?null:leftOperand.getValue();
            if (op.canEvalShortCut(lo))
            {
                return op.evalShortCut(lo);
            }
            return op.eval(lo, (rightOperand==null)?null:rightOperand.getValue());
            //System.out.println("result "+rv);
            //return rv;
        }

        public String toString()
        {
            if (constant_set)
            {
                return "constant: "+constant;
            }
            if (ref_obj!=null)
            {
                return "ref obj: "+ref_obj.in_expr;
            }
            return "op:"+op+" on "+(leftOperand!=null)+" and "+(rightOperand!=null);
        }

        public String print(String prefix)
        {
            StringBuffer sb = new StringBuffer();

            if (constant_def!=null)
            {
                return prefix+" constant: "+constant_def+"\n";
            }
            if (ref_obj!=null)
            {
                return prefix+" ref obj: " +ref_obj.in_expr+"\n";
            }
            sb.append(prefix+" op: "+ op+"\n");
            if (leftOperand!=null)
            {
                sb.append(prefix+" left: \n"+leftOperand.print(prefix+"\t"));
            }

            if (rightOperand!=null)
            {
                sb.append(prefix+" right: \n"+rightOperand.print(prefix+"\t"));
            }
            return sb.toString();

        }
    }


    //static Pattern constantPattern = Pattern.compile("^(([a-z|A-Z][a-z|A-Z|0-9|\\.]*)|(\\-?[0-9]+(\\.[0-9]+)?))");
    static Pattern constantPattern = Pattern.compile("^((([a-z]|[A-Z])([a-z]|[A-Z]|[0-9]|[\\.\\-\\:\\;_\\ ])*)|(\\-?[0-9]+(\\.[0-9]+)?(?:(?:e|E)-?\\d+)?))");




    Vector parseOperators(String expr)
    {
        Vector rv = new Vector();
        expr=expr.trim();
        if (expr.length()==0)
        {
            return rv;
        }
        String[] sp = wsp.split(expr);

        for (int i=0; i<sp.length; i++)
        {
            String unconsumed=sp[i];
            while (unconsumed.length()>0)
            {
                Matcher m = constantPattern.matcher(unconsumed);
                boolean _found = m.find();
                log.trace("unconsumed: %s hit: %s", unconsumed, _found);
                if (_found)
                {
                    log.trace("found at: %d >%s<", m.start(), m.group());
                }
                if (_found && m.start() == 0)
                {
                    //System.out.println("found constant: "+unconsumed +" => "+ m.group()) ;
                    log.trace("add constant: %s", m.group());
                    rv.add(new ExprNode(unconsumed.substring(0, m.end())));
                    unconsumed=unconsumed.substring(m.end());
                    continue;
                }


                Operator op = Operator.parseToken(unconsumed);
                
                if (op!=null)
                {
                    log.trace("add operator: %s", op);
                    rv.add(op);
                    unconsumed=unconsumed.substring(op.symbol.length());
                    continue;
                }

                int fi=Operator.getFirstIndex(unconsumed);
                if (fi<0)
                {
                    fi=unconsumed.length();
                }

                log.trace("add new expr node: %s", unconsumed.substring(0, fi));
                rv.add(new ExprNode(unconsumed.substring(0, fi)));
                unconsumed = unconsumed.substring(fi);
            }

        }
        return rv;
    }


    class AggratorExpression extends ExprNode
    {
        boolean filter=false;
        String funcname;
        Expression expr = null;
        String expr_def;

        ReflectedObject apply_refobj=null;
        AggregatorFunction agr_function = null;

        AggratorExpression(AggrMatcher matcher)
        {
            funcname=matcher.getFuncName();
            expr_def=matcher.getExpression();

            ref_obj=null;
            if (funcname.equals("apply") || (filter=funcname.equals("filter")))
            {
                //System.out.println("got apply expr def:"+expr_def);
                int firstcomma=expr_def.indexOf(',');
                if (firstcomma<0)
                {
                    throw new RuntimeException("apply takes a ref object definition as param first!");
                }
                String refexpr=expr_def.substring(0,firstcomma).trim();
                if (refexpr.length()==0)
                {
                    apply_refobj=new ReflectedObject(true);
                }
                else
                {
                    //could be also named
                    apply_refobj=new ReflectedObject(refexpr);
                }
                expr_def=expr_def.substring(firstcomma+1);
                //System.out.println("expr def: "+expr_def);

            }



        }

        void buildExpression(Expression root)
        {
            //System.out.println("build EXPRESSION FROM "+expr_def);

            expr = new Expression(expr_def, root.tokenizer.namedvars, root, (apply_refobj!=null));
            //System.out.println("build EXPRESSION FROM "+expr_def+" READY riit"+ expr.root);

        }

        public String toString()
        {
            return String.format("AGGR[%s]=%s", funcname, expr_def);
        }


        public Object getValue()
        {
            Object o =null;
            if (apply_refobj==null)
            {
                o = expr.eval((Object[])null);
            }
            else
            {
                //System.out.println("AGGR will eval apply on :\n\n"+constant);
                Collection toapplyon = (filter)?(Collection)apply_refobj.get(constant):(Collection)constant;

                Vector result = new Vector();
                for (Object ao : toapplyon)
                {
                    if (ao==null)
                        continue;

                    //System.out.println("\ntAGGR next apply on :\n\n"+ao+"\nexpr:\n"+expr_def);
                    Object ao2=(filter)?ao:apply_refobj.get(ao);
                    Object ro = expr.eval(ao2);
                    if (filter && !(Boolean)ro)
                    {
                        continue;
                    }
                    //System.out.println("\tapply: "+ro);
                    result.add(ro);
                }
                o=result;
                //System.out.println("AGGR result eval apply on :\n\n"+o+"\n\n");
                return result;
            }


            if (o!=null && !Collection.class.isAssignableFrom(o.getClass()))
            {
                throw new RuntimeException("cannot apply "+funcname+" on type: "+o.getClass().getName());
            }

            Collection col = (Collection)o;
            if (agr_function!=null)
            {
                return agr_function.eval(col);
            }

            AggregatorFunction base = symbol2aggregator.get(funcname);
            Vector<AggregatorFunction> v = overload_aggregators.get(funcname);
            if (base==null && (v==null || v.size()==0))
            {
                throw new RuntimeException("unknown aggregator type: "+funcname);
            }
            if (col==null || col.size()==0)
            {
                if (base!=null)
                {
                    return base.eval(col);
                }
                return null;
            }
            if (v==null || v.size()==0)
            {
                agr_function=base;
                return base.eval(col);
            }
            Object item=col.iterator().next();

            for (AggregatorFunction af: v)
            {
                if (item.getClass().isAssignableFrom(af._type))
                {
                    agr_function=af;
                    return af.eval(col);
                }
            }

            if (base==null)
            {
                throw new RuntimeException("unknown aggregator type: "+funcname);
            }

            agr_function=base;
            return base.eval(col);
        }

        public String print(String prefix)
        {
            return prefix+this;
        }
    }

    class Tokenizator
    {


        Vector<ExprNode> default_refobjs = new Vector<ExprNode>();
        HashMap<String, Vector<ExprNode>> named_refobjs = new HashMap<String, Vector<ExprNode>>();


        HashMap<String, ReflectedObject> namedvars = new HashMap<String, ReflectedObject>();

        Vector tokens = new Vector();

        public void merge(Tokenizator t)
        {
            default_refobjs.addAll(t.default_refobjs);
            for (Map.Entry<String,Vector<ExprNode> > e : t.named_refobjs.entrySet() )
            {
                Vector<ExprNode> old = named_refobjs.get(e.getKey());
                if (old==null)
                {
                    named_refobjs.put(e.getKey(), e.getValue());
                }
                else
                {
                    old.addAll(e.getValue());
                }
            }
        }

        public Tokenizator(String expr, HashMap<String, ReflectedObject> namedvars)
        {

            if (namedvars!=null)
            {
                this.namedvars=namedvars;
            }

            Matcher m = variable_pattern.matcher(expr);
            Matcher im = invoke_pattern.matcher(expr);
            int start=0;

            AggrMatcher agrm = new AggrMatcher(expr);

            boolean var_pattern_found=false;
            boolean aggr_pattern_found=false;

            while (true)
            {
                var_pattern_found=m.find(start);
                aggr_pattern_found=agrm.find(start);

                if (!var_pattern_found && !aggr_pattern_found)
                {
                    break;
                }

                //System.out.printf("start: %d>%s<\nvar found: %s aggr found: %s\n", start, expr.substring(start), var_pattern_found, aggr_pattern_found);

                //check which to eval first
                boolean var_first=true;

                if (!var_pattern_found || (aggr_pattern_found && agrm.getMatchStart()<m.start()))
                {
                    var_first=false;
                }

                if (!var_first)
                {
                    //System.out.printf("agr match: %d-%d\n", agrm.getMatchStart(), agrm.getMatchEnd());
                    //got agrm pattern
                    tokens.addAll(parseOperators(expr.substring(start, agrm.getMatchStart())));

                    AggratorExpression agex = new AggratorExpression(agrm);
                    default_refobjs.add(agex);
                    start=agrm.getMatchEnd();

                    tokens.add(agex);
                    continue;
                }

                tokens.addAll(parseOperators(expr.substring(start, m.start())));

                start=m.end();

                boolean extendable=true;

                String subexpr=null;
                while (im.find(start))
                {
                    if (im.start()!=start)
                        break;

                    start=im.end();
                    int nopen=1;
                    for (;start<expr.length() && nopen>0; start++)
                    {
                        if (expr.charAt(start)==')')
                        {
                            nopen--;
                            continue;
                        }
                        if (expr.charAt(start)=='(')
                        {
                            nopen++;
                            continue;
                        }

                    }

                }
                ExprNode node = null;
                ReflectedObject ref_obj = null;
                if (namedvars==null || null == (ref_obj=namedvars.get(m.group(1))))
                {
                    default_refobjs.add(node=new ExprNode(ref_obj=new ReflectedObject(m.group(1))));

                }
                else
                {
                    Vector<ExprNode> v = named_refobjs.get(m.group(1));
                    if (v==null)
                    {
                        named_refobjs.put(m.group(1), v = new Vector<ExprNode>());
                    }
                    v.add(node=new ExprNode(ref_obj));
                }
                subexpr=expr.substring(m.end(),start);
                if (subexpr.length()>0)
                {
                    ref_obj.setSub(new ReflectedObject(subexpr));
                }
                //System.out.printf("next: %s -> %s\n", m.group(1), subexpr);
                tokens.add(node);
            }
            tokens.addAll(parseOperators(expr.substring(start)));
        }


    }

    static Pattern wsp = Pattern.compile("\\s+");
    public static Pattern aggr_pattern=Pattern.compile("\\$(\\w+)\\(");
    public static Pattern variable_pattern=Pattern.compile("\\$\\{([^\\}]+)\\}");
    public static Pattern invoke_pattern=Pattern.compile("\\s*\\.\\s*\\w+\\s*\\(");

    public static Vector<String> getVariableReferences(String s)
    {
        Vector<String> v = new Vector<>();
        int start = 0; 
        Matcher m = variable_pattern.matcher(s);
        while (m.find(start))
        {
            v.add(m.group(1));
            start = m.end();
        }
        return v;
    }
    static class AggrMatcher
    {
        String func_name=null;
        Integer match_start=null;
        Integer match_end=null;
        Integer subexpr_start=null;
        Integer subexpr_end=null;

        String expr;
        Matcher m = null;
        AggrMatcher(String expr)
        {
            m=aggr_pattern.matcher(this.expr=expr);
        }

        String getExpression()
        {
            if (match_start==null)
            {
                throw new RuntimeException("not in match state!");
            }
            return expr.substring(subexpr_start, subexpr_end);
        }
        String getFuncName()
        {
            if (match_start==null)
            {
                throw new RuntimeException("not in match state!");
            }
            return func_name;
        }
        int getMatchStart()
        {
            if (match_start==null)
            {
                throw new RuntimeException("not in match state!");
            }
            return match_start;
        }

        int getMatchEnd()
        {
            if (match_end==null)
            {
                throw new RuntimeException("not in match state!");
            }
            return match_end;
        }
        boolean find(int start)
        {
            func_name=null;
            match_start=null;
            match_end=null;
            subexpr_start=null;
            subexpr_end=null;

            if (!m.find(start))
                return false;


            System.out.println("aggregator got hit: "+m.group());
            func_name=m.group(1);
            subexpr_start=m.end();
            match_start=m.start();

            int nopen=1;
            for (match_end=m.end(); nopen>0 && match_end<expr.length(); match_end++)
            {

                switch (expr.charAt(match_end))
                {
                    case '(':
                        nopen++;
                        break;
                    case ')':
                        nopen--;
                        break;
                    default:
                        break;
                }

            }
            subexpr_end=match_end-1;
            //System.out.println("aggr got match : "+match_start+"-"+match_end+" "+expr.substring(match_start, match_end));
            if (nopen!=0)
            {
                throw new RuntimeException("unclosed paraenthesis: "+expr.substring(match_start));
            }

            return true;
        }
    }

    static HashSet<Integer> getSet(int maxn)
    {
        int ne = Math.min(maxn, 1+(int)(Math.random()*maxn));
        int base=(5*maxn);
        HashSet<Integer> hs = new HashSet<Integer>();
        while (hs.size()<ne)
        {
            hs.add((int)(Math.random()*base));
        }
        return hs;

    }
    static class TestO
    {
        public HashSet<Integer> s1 = new HashSet<Integer>();
        public HashSet<Integer> s2 = new HashSet<Integer>();
        int index;
        public TestO(int idnex)
        {
            index=idnex;
            s1=getSet(20);
            s2=getSet(20);
        }

        public String toString()
        {
            Vector<Integer> v1 = new Vector<Integer>();
            v1.addAll(s1);
            Collections.sort(v1);
            Vector<Integer> v2 = new Vector<Integer>();
            v2.addAll(s2);
            Collections.sort(v2);

            return String.format("testo %d set1(%d)=%s set2(%d)=%s\n", index, v1.size(), v1, v2.size(), v2);

        }
    }
    static void aggr_test()
    {
        HashMap<String, ReflectedObject> name2var = new HashMap<String, ReflectedObject>();

        Vector<HashSet<TestO>> testv = new Vector<HashSet<TestO>>();
        for (int i=0; i<10; i++)
        {
            HashSet<TestO> hs = new HashSet<TestO>();
            for (int j=0; j<4; j++)
            {
                hs.add(new TestO(i));
            }
            testv.add(hs);
        }


        Expression e = new Expression("$max($apply(,$count(${s1}))) + $min($apply(,$count(${s2})))");
        for (int i=0; i<testv.size(); i++)
        {
            System.out.println("will eval on :"+testv.get(i));
            System.out.println(e.eval(testv.get(i)));
        }






        System.out.println("eval: "+e.eval());
    }


    static Pattern numPat = Pattern.compile("\\d+");

    /** builds an expression expection numeric references in expr whereas i == data[i]
     */
    public static<A> Expression buildExpression(Expression e, String expr, A ... data)
    {
        Logger log = LogConfig.getLogger();
        if (e == null)
        {
            e = new Expression();
        }
        StringBuffer exp=new StringBuffer();
        int start=0;
        HashMap<Integer, A> resolved = new HashMap<>();
        
        Matcher ma = numPat.matcher(expr);

        while (ma.find(start))
        {
            String val=ma.group();
            int ref_idx = Integer.parseInt(val);
            if (ref_idx < 0 || ref_idx >= data.length)
            {
                throw new FRuntimeException("cannot resolve variable >%d< only %d input variables given", ref_idx, data.length);
            }
            if (resolved.get(ref_idx) == null)
            {
                resolved.put(ref_idx, data[ref_idx]);
            }
            exp.append(expr.substring(start, ma.start()));
            exp.append("${"+ref_idx+"}");
            start=ma.end();
        
        }
        exp.append(expr.substring(start));
        
        HashMap<String, ReflectedObject> m = new HashMap<String, ReflectedObject>();
        for (int idx : resolved.keySet())
        {
            m.put(""+idx, new ReflectedObject(true));
        }
        e.setExpression(exp.toString(), m);

        log.trace("converted expression: >%s<", exp.toString());
        for(Map.Entry<Integer, A> entry: resolved.entrySet())
        {
            e.assign(""+entry.getKey()+"", entry.getValue());
            //log.info("asign: %s = %s",  entry.getKey(), entry.getValue());
        }

        return e;
        
    }



    public static class TestS
    {
        public int n=4;
        Object o = new Object();
        public Integer getP(int p)
        {
            return n+p;
        }
    }


    public static void main (String args[]) throws Exception
    {
        ScriptEngineManager sem = new ScriptEngineManager();
        ScriptEngine se = sem.getEngineByName("nashorn");
        Bindings b = se.createBindings();
        ScriptContext sc = se.getContext();
        String expr = "${pvalue.Stratil}<-1.4820880790419225E-75";
        Matcher _m = Expression.constantPattern.matcher("-1.4820880790419225E-75");
        System.out.println("constant found: "+_m.find()+" at : " + _m.group());
        
        Expression eex = new Expression(expr);
        System.out.println(eex.root);

        expr = " (${cond1}<5 && ${cond1}> -5) && (${cond3}>3 || ${cond3}< -3)  ";
        String fn = "function ex(cond1, cond2, cond3){ return (Math.abs(cond1)<5) && Math.abs(cond3)<3; };";

        String fn2 = "function matches(s){ return p.matcher(s).find(); };";
        Pattern pat = java.util.regex.Pattern.compile("[0-9]+");
        b.put("p", pat);
        System.out.println("se eval fn2: "+se.eval(fn2+" matches('b123');", b));



        Invocable invocable = (Invocable)se;

        b.put("v", "a123");
        System.out.println(se.eval("p.matcher(v).find();", b));

        //System.out.println(invocable.invokeFunction("matches", "a123"));






        double[][] data = new double[100000][3];

        int acc=0;
        Object o[] = new Object[3];
        for (int i=0; i<data.length; i++)
        {
            for (int j=0; j<3; j++)
            {
                double v = Math.random()*10;
                if (Math.random()<0.5)
                {
                    v = -v;
                }
                data[i][j] = v;
                //System.out.printf("%d: %2.3f ", j, v);
            }
        }

        long t1=System.currentTimeMillis();
        se.eval(fn);

        for (int i=0; i<data.length; i++)
        {
            //System.out.println();
            o[0] = data[i][0];
            o[1] = data[i][1];
            o[2] = data[i][2];
            //System.out.println("invoke: "+invocable.invokeFunction("ex", data[i][0], data[i][1], data[i][2]));
            if ((Boolean)invocable.invokeFunction("ex", o))
            {
                acc++;
            }
            //System.out.println("invoke2: "+invocable.invokeFunction("ex", o));
        }

        long t2=System.currentTimeMillis();
        System.out.printf("acc: %d took: %2.3f sec\n", acc, (t2-t1)/1000.0);


        t1=System.currentTimeMillis();
        HashMap<String, ReflectedObject> vm = new HashMap<String, ReflectedObject>();

        vm.put("cond1", new ReflectedObject(true));
        vm.put("cond2", new ReflectedObject(true));
        vm.put("cond3", new ReflectedObject(true));
        Expression te = new Expression(expr, vm);
        te.setOrder("cond1", "cond2", "cond3");
        acc = 0;
        for (int i=0; i<data.length; i++)
        {
            //System.out.println();
            o[0] = data[i][0];
            o[1] = data[i][1];
            o[2] = data[i][2];
            //System.out.println("invoke: "+invocable.invokeFunction("ex", data[i][0], data[i][1], data[i][2]));
            if ((Boolean)te.eval(o))
            {
                acc++;
            }
            //System.out.println("invoke2: "+invocable.invokeFunction("ex", o));
        }
        t2=System.currentTimeMillis();
        System.out.printf("acc: %d took: %2.3f sec\n", acc, (t2-t1)/1000.0);


        if (true)
        {
            return;
        }



        int start=0;
        Pattern constantP = Pattern.compile("^((([a-z]|[A-Z])([a-z]|[A-Z]|[0-9]|[\\.\\-\\:\\;\\ ])*)|(\\-?[0-9]+(\\.[0-9]+)?))");
        //Pattern constantP = Pattern.compile("^((([a-z]|[A-Z])([a-z]|[A-Z]|[0-9])*))");
        for (int i=0; i<expr.length(); i++)
        {
            Matcher m = constantP.matcher(expr.substring(i));
            if (!m.find())
                continue;
            System.out.println("next match: "+m.group());

        }
        Expression ee = new Expression(expr);


        System.out.println(ee.root);
        if (true) return;

        Pair<TestS,TestS> p = new Pair<TestS,TestS>(new TestS(), new TestS());
        ReflectedObject ro = new ReflectedObject(new ReflectedObject("getFirst()"), new ReflectedObject("getP", 14));
        System.out.println(ro.get(p));

        HashSet<Integer> ts = new HashSet<Integer>();

        for (int i=0; i<21; i++)
        {
            if (i%2 == 1)
                continue;


            ts.add(i);
        }
        ReflectedObject ro2 = ReflectedObject.apply(ts, "contains", "this");
        for (int i=0; i<21; i++)
        {
            System.out.println("contains: "+ i +" : "+ro2.get(i));
        }




        TestO2 to = new TestO2();
        TestO2 to2 = new TestO2();
        to2.sub=new TestS();

        Expression e = new Expression("${sub}!=null && ${sub.n}>3");
        System.out.println("FINAL: "+e.eval(to));
        System.out.println(e.eval(to2));
        //System.out.println( (new Expression("((${1} + (${2}) ) ^  ((${3} + ${4}))) + $max(${1} - ${4})")).root.print(""));
        //aggr_test();
    }

    public static class TestO2
    {

        public TestS sub=null;
    }

}   
