package lmu.utils;

import java.io.*;
import java.util.*;
import java.util.regex.Pattern;

import lmu.utils.*;

public class FileLookup
{
	File f;
	public File indexF;
	public HashMap<String,Pair<Long,Long>> lookups = null;
	boolean changed=true;

	public Pattern SEP = Pattern.compile("\t");
	public int indexKey=0;
	
	public FileLookup(File toIndex, File index)
	{
		this.f=toIndex;
		this.indexF = index;
	}
	
	public FileLookup(File toIndex, File index,Pattern sep, int indexKey)
	{
		this(toIndex,index);
		this.SEP = sep;
		this.indexKey = indexKey;
	}

    /**
     * has to be called if something is written to the file, or this won't be in the lookup
     */
	public void fileChanged()
	{
		changed=true;
	}
	
	public void initLookups()
	{
		lookups = new HashMap<String,Pair<Long,Long>>();
		try
		{
			String line;
			if(!f.canRead())
				return;
			RandomAccessFile raf = new RandomAccessFile(f, "r");
			Long lastPos = (long) 0;
			while((line=raf.readLine())!=null)
			{
				String[] sp = SEP.split(line);
				if(sp.length<indexKey)
				{
					ExceptionManager.throwRuntimeException("less columns than index given");
				}
				String key = sp[indexKey];
				//file pointer is at position that will be read as next -> -2 to get last non-newline symbol in current line
	            Long end = raf.getFilePointer()-1;
//	            System.out.println(XmlIndexer.extractBuffer(f, lastPos, end).toString());
	            lookups.put(key, new Pair<Long,Long>(lastPos,end));
	            lastPos = end+1;
			}
			raf.close();
		}
		catch(IOException e)
		{
			ExceptionManager.throwRuntimeException("unable to index file "+f.getAbsolutePath());
		}
	}
	
	public String getBuffer(String key)
    {
    	if(lookups==null || changed)
    	{
    		initLookups();
    	}
        if(lookups==null)
        {
        	ExceptionManager.throwRuntimeException("no keys found at index "+indexKey+" separated by"+SEP+" in file "+f.getAbsolutePath());
        }
        LogConfig.getLogger().debug("after init of lookup it contains "+lookups.size()+" keys");
        Pair<Long,Long> p = lookups.get(key);
        if (p==null)
        {
            return null;
        }
        StringBuffer xbuff = XmlIndexer.extractBuffer(f, false, p.getFirst(), p.getSecond(), true);
        return xbuff.toString();
    }

	public static void main(String[] args)
	{
		SimpleOptionParser cmd = new SimpleOptionParser("file");
		OptionParser.parseParams(args, true, true, true, cmd);
		
		FileLookup look = new FileLookup(cmd.getFile("file"), new File("index.tmp"),Pattern.compile("\\$"),0);
		look.initLookups();
	}
}
