package lmu.utils;

import java.awt.*;
import java.awt.event.*;
//import lmu.utils.plotting.*;
import java.awt.geom.*;
import java.awt.image.BufferedImage;
import java.util.*;

import javax.swing.*;

import lmu.utils.plotting.*;

public class ImageUtils
{
    static int WHITE  = Color.WHITE.getRGB();

    static int getFirstNonWhite(boolean reverse, boolean x, BufferedImage bim)
    {
        int s = 0;
        int e = (x)?bim.getWidth():bim.getHeight();
        int start = (reverse)?e-1:s;
        int end = (reverse)?s:e-1;
        int inc = (reverse)?-1:1;


        boolean found=false;
        int dim1=(x)?bim.getHeight():bim.getWidth();


        //System.out.println("test x:"+x+" reverse: "+reverse+" dim: "+bim.getWidth()+":"+bim.getHeight());
        //System.out.printf(" start: %d end: %d dim1: %d inc: %d\n", start, end, dim1, inc);
        for (int i=start; !found && i!=end; i+=inc )
        {
            for (int j=0; !found && j<dim1; j++)
            {
                int check_x=(x)?i:j;
                int check_y=(x)?j:i;

                if (j==0 || j==dim1-1)
                {
                    //System.out.printf("\t check : %d,%d (dim %d,%d) i: %d j: %d\n", check_x, check_y, bim.getWidth(), bim.getHeight(), i, j);
                }

                if (! (found = (WHITE!=bim.getRGB(check_x, check_y))))
                    continue;

                //System.out.printf("\t found first non white at %s:%d : %d,%d \n", (x)?"row":"column", i, check_x, check_y);
                return Math.max(0, i-inc);
            }

        }
        return end;

    }
    static public BufferedImage autocrop(BufferedImage source) 
    {
        Point2D upper_left = new Point2D.Double(getFirstNonWhite(false, true, source), getFirstNonWhite(false, false, source) );
        Point2D lower_right = new Point2D.Double(getFirstNonWhite(true, true, source), getFirstNonWhite(true, false, source));

        System.out.println("upper left: "+upper_left);
        System.out.println("lower right: "+lower_right);

        BufferedImage destination = new BufferedImage((int)(lower_right.getX()-upper_left.getX()), (int)(lower_right.getY()-upper_left.getY()), BufferedImage.TYPE_INT_ARGB) ;
        return source.getSubimage((int)upper_left.getX(), (int)upper_left.getY(), (int)(lower_right.getX() - upper_left.getX()), (int)( lower_right.getY() - upper_left.getY()));
        /*
        //destination.createGraphics().drawImage(source, null, 
        for (int i=0, i2=(int)upper_left.getX(); i2<(int)lower_right.getX(); i++, i2++)
        {
            for (int j=0, j2=(int)upper_left.getY(); j2<(int)lower_right.getY(); j++, j2++)
            {
                destination.setRGB(i,j, source.getRGB(i2, j2));
            }
        }


        return destination ;
        */

    }


    static public BufferedImage getImage(int width, int height, Color bg)
    {
        BufferedImage bim = new BufferedImage(width, height, BufferedImage.TYPE_3BYTE_BGR);
        
        Graphics2D g2 = bim.createGraphics();
        g2.setColor(bg);
        
        g2.fillRect(0, 0,width,height);
        g2.dispose();
        
        return bim;
    }
    static public BufferedImage getInsetImage(int width, int height, BufferedImage inset_img)
    {
        BufferedImage bim = getWhiteImage(width, height);
        Graphics2D g2 =  bim.createGraphics();
        g2.drawImage(inset_img, null, 0,0);
        g2.dispose();
        return bim;
    }
    static public BufferedImage getWhiteImage(int width, int height)
    {
        return getImage(width, height, Color.WHITE);
    }
    static public Pair<Integer,Integer> getDrawnTextDimension(int font_size, String text)
    {
        return getDrawnTextDimension((Font)null,-1,text);
    }
    
    static public Pair<Integer,Integer> getDrawnTextDimension(Graphics2D g2, String text)
    {
        return getDrawnTextDimension(g2.getFont(),-1,text);
    }

    static public Pair<Integer,Integer> getDrawnTextDimension(Graphics2D g2, int font_size, String text)
    {
        return getDrawnTextDimension(g2.getFont(),font_size,text);
    }


    static public Pair<Integer,Integer> getDrawnTextDimension(Font f, int font_size, String ... texts)
    {
        return getDrawnTextDimension(f,font_size,new CollectionIterator<String>(texts));
    }

    static public Pair<Integer,Integer> getDrawnTextDimension(Font f, int font_size, Collection<String>  texts)
    {
        return getDrawnTextDimension(f,font_size,   texts.iterator());
    }



    static public Pair<Integer,Integer> getDrawnTextDimension(Font f, int font_size, String text)
    {
        return getDrawnTextDimension(f,font_size,new SingleIterator<String>(text));
    }

    static public Pair<Integer,Integer> getDrawnTextDimension(Font f, int font_size, Iterator<String> texts )
    {
        return getDrawnTextDimension(f,font_size,texts,0);
    }
    static public Pair<Integer,Integer> getDrawnTextDimension(Font f, int font_size, Iterator<String> texts, int style)
    {
        
        BufferedImage bim = new BufferedImage(10, 10, BufferedImage.TYPE_3BYTE_BGR);
        Graphics2D g2 = bim.createGraphics();
        
        if (f==null)
        {
            f=g2.getFont();
        }
        if (font_size<0)
        {
            font_size=f.getSize();
        }
        else
        {
            if (f.getSize()!=font_size)
            {
                f=f.deriveFont((float)font_size);
                f=f.deriveFont(style);
            }
        }
        
        
        int char_h=(int)(f.getMaxCharBounds(g2.getFontRenderContext()).getHeight());
        int width=0;
        while (texts.hasNext())
        {
            String text = texts.next();
            int w=(int)f.getStringBounds(text,g2.getFontRenderContext()).getWidth();
            width=Math.max(width,w);
        }
        
        //int width = g2.getFontMetrics(

        return new Pair<Integer,Integer>(width, char_h);
    }
    

    public static Color invertColor(Color c)
    {
        int red = 255-c.getRed();
        int green = 255-c.getGreen();
        int blue = 255-c.getBlue();

        return new Color(red, green, blue);
    }

    public static Color getColor(String c, Color def)
    {
        if (c==null)
            return def;

        Color c1 = ColorManager.PREDEFINED_COLOR.get(c);
        if (c1!=null)
        {
            return c1;
        }

        return new Color(Integer.parseInt(c, 16));
    }


    public static  BufferedImage scale(double factor, BufferedImage im)
    {
        return scale(factor,factor,im);
    }
    public static  BufferedImage transform(AffineTransform at, BufferedImage im)
    {
        Point2D p = new Point2D.Double(im.getWidth(), im.getHeight());

        Point2D tp = at.transform(p, null);

        BufferedImage nimg = new BufferedImage((int)tp.getX(), (int)tp.getY(), BufferedImage.TYPE_3BYTE_BGR);

        Graphics2D g2 = nimg.createGraphics();
        g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);

        g2.setTransform(at);
        g2.drawImage(im,0,0, null);

        return nimg;
    }
    public static  BufferedImage scale(double x, double y, BufferedImage im)
    {
        //if(1==1)return im;
        BufferedImage scaled = new BufferedImage((int)(im.getWidth()*x), (int)(im.getHeight()*y), BufferedImage.TYPE_INT_RGB);
        Graphics2D g2 = scaled.createGraphics();
        g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
        g2.drawImage(im, 0,0, scaled.getWidth(), scaled.getHeight(), null);
        g2.dispose();
        return scaled;

    }

    public static class ImagePanel extends javax.swing.JLabel 
    {
            
        BufferedImage img;

    
        int width_offset=30;
        int  height_offset=50;
        public ImagePanel(BufferedImage im)
        {
            this(im, null, null);
        }

        public void reloadPicture()
        {
            getGraphics().drawImage(img, 0, 0, img.getWidth(), img.getHeight(), this);
        }
        public ImagePanel(BufferedImage im, Double factor)
        {
            this(im, factor, null);
        }
        public ImagePanel(BufferedImage im, Double zoomfactor, MouseListener listener)
        {
            if (zoomfactor!=null)
            {
                im = scale(zoomfactor,im);
            }
            setPicture(im);
            if (listener!=null)
            {
                this.addMouseListener(listener);
            }
        }
         public void setPicture(BufferedImage im)
         {
             this.img=im;
             if (img!=null)
             {
                 Dimension d = new Dimension(im.getWidth()+width_offset, im.getHeight()+height_offset);
                 setPreferredSize(d);
                 setSize(im.getWidth(), im.getHeight());
    
             }
             setAutoscrolls(true);
             
         }

         public void paintComponent(Graphics g)
         {
             if (img==null)
                 return;
             

             Graphics2D graphics2D = (Graphics2D)g;

             graphics2D.setRenderingHint(RenderingHints.KEY_INTERPOLATION,RenderingHints.VALUE_INTERPOLATION_BILINEAR);
             
             g.drawImage(img, 0, 0, img.getWidth(), img.getHeight(), this);
         }
    

     
         public BufferedImage getImage()
         {
             return img;
         }
    }

    public static class LabelFrame extends JFrame
    {
           
        public LabelFrame(String title, JLabel ip)
        {
            this(title, ip, true);
        }
        public LabelFrame(String title, JLabel ip, boolean close)
        {
            super(title);

            this.setDefaultCloseOperation((close)?JFrame.EXIT_ON_CLOSE:JFrame.HIDE_ON_CLOSE);
            
            
            
            JScrollPane pictureScrollPane = new JScrollPane(ip);
            Dimension dim =ip.getPreferredSize();
            dim.setSize(dim.getWidth()+10, dim.getHeight()+10);
            pictureScrollPane.setPreferredSize(dim);
            pictureScrollPane.setOpaque(true); //content panes must be opaque
            setContentPane(pictureScrollPane);

            pack();
            setVisible(true);

        }
    }
    
    public interface ImageClickHandler
    {
        public void onClicked(int x, int y, MouseEvent e);
        
    } 
    
    static public BufferedImage vconcat(Iterator<BufferedImage> images)
    {
        int nwidth=0;
        int height=0;
        Vector<BufferedImage> v = new Vector<BufferedImage>();
        while (images.hasNext())
        {
            BufferedImage im = images.next();
            v.add(im);
            nwidth=Math.max(nwidth,im.getWidth());
            height+=im.getHeight();
        }
        BufferedImage res = getWhiteImage(nwidth, height);

        Graphics2D g2 = res.createGraphics();

        int start=0;
        for (BufferedImage im : v)
        {
            g2.drawImage(im, null, 0, start);
            start+=im.getHeight();
        }
        
        return res;
    }

    static public BufferedImage concat(Iterator<BufferedImage> images)
    {
        int nwidth=0;
        int height=0;
        Vector<BufferedImage> v = new Vector<BufferedImage>();
        while (images.hasNext())
        {
            BufferedImage im = images.next();
            v.add(im);
            nwidth+=im.getWidth();
            height=Math.max(height, im.getHeight());
        }
        BufferedImage res = getWhiteImage(nwidth,height);
        
        Graphics2D g2 = res.createGraphics();
        
        int start=0;
        for (BufferedImage im : v)
        {
            g2.drawImage(im, null, start,0);
            start+=im.getWidth();
        }
        
        return res;
    }
    static public BufferedImage concat(Collection<BufferedImage> images)
    {
        return concat(new CollectionIterator<BufferedImage>(images));
    }
    static public BufferedImage concat(BufferedImage ... images)
    {
        return concat(new CollectionIterator<BufferedImage>(images));
        
    }

    static public BufferedImage vconcat(Collection<BufferedImage> images)
    {
        return vconcat(new CollectionIterator<BufferedImage>(images));
    }
    static public BufferedImage vconcat(BufferedImage ... images)
    {
        return vconcat(new CollectionIterator<BufferedImage>(images));
        
    }

    public static void drawCenteredString(Graphics g, String string, int x, int y, int areaHeight)
    {
        FontMetrics font = g.getFontMetrics();
        System.out.println("area height: "+areaHeight+" font height: "+font.getAscent());
        g.drawString(string, x, y+areaHeight-(Math.min(0,areaHeight-font.getAscent()))/2);
    }
    
    

    public static class ImageFrame extends JFrame
    {
        //ImagePanel ip;

        Vector<ImageClickHandler> click_handlers = new Vector<ImageClickHandler>();

        ForwardMouseHandler mouse_handler = new ForwardMouseHandler();
        
        
        public void addHandler(ImageClickHandler ich)
        {
            click_handlers.add(ich);
            
        }
        class ForwardMouseHandler implements MouseListener
        {
            public void mouseClicked(MouseEvent e)
            {   
                //System.out.printf("clicked : %d,%d  ip: %d,%d\n", e.getX(), e.getY(), ip.getX(), ip.getY());
                for (ImageClickHandler ich : click_handlers)
                {
                    ich.onClicked(e.getX(),e.getY(), e);
                }
                
                
                
            }
            public void mousePressed(MouseEvent e)
            {
                
                //parentOrganizer.setReplySet();
                /*unnecessary*/
            }
    
            public void mouseReleased(MouseEvent e){/*unnecessary*/}
    
            public void mouseEntered(MouseEvent e){/*unnecessary*/}
    
            public void mouseExited(MouseEvent e){/*unnecessary*/}        
        }
    
        public ImageFrame(String title, BufferedImage im)
        {
            this(title, im, true);
        }

        ImageScrollPane pictureScrollPane = new ImageScrollPane();

        public ImageFrame(String title, SliceImageDrawer im, boolean close)
        {
            super(title);

            this.setDefaultCloseOperation((close)?JFrame.EXIT_ON_CLOSE:JFrame.HIDE_ON_CLOSE);
            
            
            //ip = new ImagePanel(im);
            pictureScrollPane.setSlicer(im);
            
            init();        
    
        }

        public void save()
        {
            javax.swing.JFileChooser jfc = new javax.swing.JFileChooser();
            if(JFileChooser.APPROVE_OPTION!=jfc.showDialog(this, "ok"))
            {
                return;
            }
            java.io.File out = jfc.getSelectedFile();
            String sp[]=out.getName().split("\\.");
            if (sp.length<2)
            {
                throw new RuntimeException("no extension detected...in file name:: "+out.getName());
            }
            String format = sp[sp.length-1];
            //System.out.println("will write file to "+out.getAbsolutePath());
            
            try
            {
                javax.imageio.ImageIO.write(pictureScrollPane.getFullImage(), format, out);
            }
            catch(java.io.IOException ie)
            {
                throw new RuntimeException("i/o error: "+ie.getMessage()+" while writing image file to : "+out.getName(), ie);
            }
            
        }

        public void init()
        {
            JMenuBar jb = new JMenuBar();
            setJMenuBar(jb);

            JMenu jm = new JMenu("file");
            jb.add(jm);

            JMenuItem jmi = new JMenuItem("save");
            jm.add(jmi);
            jmi.setAccelerator(KeyStroke.getKeyStroke("control S"));
            jmi.addActionListener(new lmu.utils.swing.ReflectedObjectListener(new ReflectedObject("save()"), this));

            setContentPane(pictureScrollPane);

            
            pictureScrollPane.supportScaling(true);
            //pictureScrollPane.addMouseListener(mouse_handler);
            pictureScrollPane.addListener(mouse_handler);
            pack();
            setVisible(true);
            
        }
        public ImageFrame(String title, BufferedImage im, boolean close)
        {
            super(title);

            //System.out.println(ExceptionManager.logCallerStack());
            this.setDefaultCloseOperation((close)?JFrame.EXIT_ON_CLOSE:JFrame.HIDE_ON_CLOSE);
            
            //System.out.println("close op: "+close);
            
            //ip = new ImagePanel(im);
            pictureScrollPane.setImage(im);
            
            init();
            /*JScrollPane pictureScrollPane = new JScrollPane(ip);
            Dimension dim =ip.getPreferredSize();
            dim.setSize(dim.getWidth()+10, dim.getHeight()+10);
            pictureScrollPane.setPreferredSize(dim);
            pictureScrollPane.setOpaque(true); //content panes must be opaque
            */
            

        }
        
        public void setImage(BufferedImage im)
        {
            pictureScrollPane.setImage(im);
        }
    }

    public static class TextImage
    {
        int font_width = 15;
        int font_height = 15;

        Vector<String> textes = new Vector<>();
        public TextImage()
        {

        }

        public TextImage(String n)
        {
            this();
            textes.add(n);
        }

        public TextImage add(String n)
        {
            textes.add(n);
            return this;
        }

        public BufferedImage getImage()
        {
            int width = font_width * NumUtils.max(ObjectGetter.map(textes, (t) -> t.length()));
            int height = font_height * textes.size();
            BufferedImage im = getWhiteImage(width, height);
            Graphics g = im.getGraphics();
            g.setColor(Color.BLACK);

            for (int i=0; i<textes.size(); i++)
            {
                g.drawString(textes.get(i), 4, (i+1) * font_height);
            }
            g.dispose();
            return im;
        }

    }

    static public TextImage getTextImage(String ... text)
    {
        TextImage ti = new TextImage();
        ObjectGetter.apply(text, (t) -> ti.add(t));
        return ti;
    }

    static public LabelFrame showImage(String title, JLabel jl)
    {
        return showImage(title,jl,true);
    }
    static public LabelFrame showImage(String title, JLabel jl, boolean close)
    {
        return new LabelFrame(title, jl, close);
    }
    static public ImageFrame showImage(String title, BufferedImage im)
    {
        return showImage(title, im, 1.0, true);
    }

    static public ImageFrame showImage(String title, BufferedImage im, boolean close)
    {
        return showImage(title,im,1.0,close);
    }

    static public ImageFrame showImage(String title, BufferedImage im, double factor,  boolean close)
    {
        return new ImageFrame(title, scale(factor,im), close);
    }

    static public ImageFrame showImage(String title, lmu.utils.plotting.SliceImageDrawer drawer)
    {
        return new ImageFrame(title, drawer, true);
    }
    static public ImageFrame showImage(String title, BufferedImage im, double factor)
    {
        return new ImageFrame(title, scale(factor,im), true);
    }

    static public BufferedImage readImage(java.io.File f)
    {
        try
        {
            return javax.imageio.ImageIO.read(f);
        }
        catch(java.io.IOException ie)
        {
            throw new RuntimeException("i/o error: "+ie.getMessage()+" while reading from file: "+f.getAbsolutePath(), ie);
        }
        
    }
    public static void saveImage(BufferedImage bim, java.io.File outf)
    {
        String sp[] = outf.getName().split("\\.");
        String fmt = (sp.length < 2) ? "png" : sp[sp.length-1];
        try
        {
            javax.imageio.ImageIO.write(bim, fmt, outf);
        }
        catch(java.io.IOException ie)
        {
            throw new RuntimeException("i/o error: "+ie.getMessage()+" while writing image to file: "+ outf.getAbsolutePath(), ie);
        }
    }
    public static void main (String args[]) throws Exception
    {

        
        final int L = 50_000;
        AbstractZoomedMultiPicPlane _plot = new AbstractZoomedMultiPicPlane(null, null, L);
         lmu.utils.plotting.RegionPicturePane rpp = _plot.createRegionPane("MSA");
         rpp.addRegion("test", null, new Region1D(100, L, null));
         _plot.update(true);

         BufferedImage im = _plot.getCombinedImage();
         ImageFrame _if = new ImageFrame("test", im);
         _if.addHandler((int x, int y, MouseEvent e) -> System.out.printf("clicked: %d,%d\n", x, y));
         //showImage("test", _if, true);
         /*
        javax.imageio.ImageIO.write(im, "png", new java.io.File("test.png"));
        */        
        //showImage("test", javax.imageio.ImageIO.read(new java.io.File(args[0])), true);
    }
}

