
package lmu.utils;

import java.util.*;

public class IntCountUpdater
{
    String name="";
    int[] distrib=null;

    int min_value;
    int max_value;
    int num_underflow=0;
    int num_overflow=0;

    int total_count=0;
    int update_count=0;

    public IntCountUpdater(String name, int min_value, int max_value)
    {
        this.name=name;
        this.min_value=min_value;
        this.max_value=max_value;
        distrib=new int[max_value-min_value+1];

    }
    
    public int getUpdateCount()
    {
        return update_count;
    }
    public int getTotalCount()
    {
        return total_count;
    }

    

    public void updateMulti(int increase, int ... labels)
    {
        int save_total=total_count;
        int save_update=update_count;

        for (int l : labels)
        {
            update(l,increase);
        }
        total_count=save_total+increase;
        update_count=save_update+1;
    }
    
    public void update(int label, int increase)
    {
        total_count+=increase;
        update_count++;

        if (label<min_value)
        {
            num_underflow++;
            return;
        }
        if (label>max_value)
        {
            num_overflow++;
            return;
        }
        distrib[label-min_value]++;
    }

    public int getCount(int l)
    {
        if (l<min_value)
        {
            return num_underflow;
        }
        if (l>max_value)
        {
            return num_overflow;
        }
        return distrib[l-min_value];
        
    }
    public Vector<Pair<Integer, Integer> > getSortedCounts()
    {
        return getSortedCounts(0);
    }
    public Vector<Pair<Integer, Integer> > getSortedCounts(int around)
    {

        Vector<Pair<Integer, Integer> > rv = new Vector<Pair<Integer, Integer>>();
        final int N=distrib.length-1;
        for (int i=0, l=min_value; i<distrib.length; i++, l++)
        {
            int v=0;
            for (int j=Math.max(i-around, 0); j<=Math.min(i+around,N); j++)
            {
                v+=distrib[j];
            }
            rv.add(new Pair<Integer,Integer>(v, l));
        }

        Collections.sort(rv);
        Collections.reverse(rv);
        return rv;
    }

    public void printBestN(int n, int around)
    {
        
        Vector<Pair<Integer,Integer>> v = getSortedCounts(around);
        if (n<0)
        {      
            n=v.size();
        }
        if (total_count==0)
        {
            return;
        }
        double norm=100.0/total_count;
        System.out.printf("%s total count: %d underflow: %d(%2.2f%%) overflow: %d(%2.2f%%)\n", 
                          name, total_count, num_underflow, norm*num_underflow, num_overflow, norm*num_overflow);
        
        String aroundinfo=(around==0)?"":"(+/- "+around+")";
        for (int i=0; i<v.size() && i<n; i++)
        {
            int c = v.get(i).getFirst();
            System.out.printf("\t%d%s: %d (%2.2f%% direct: %2.2f%%)\n", v.get(i).getSecond(), aroundinfo, c, c*norm, distrib[v.get(i).getSecond()-min_value]*norm);
        }
    }
}

