package lmu.utils;

import java.io.*;
import java.util.*;
import java.util.regex.Pattern;

import org.apache.log4j.*;
import org.jdom.*;
import org.jdom.input.SAXBuilder;

/** configures the logging hiearchically 
 *  the configuration is based on an xml elem named "logconfig"
 *  with 0-n subelements named "logger"
 * 
 *   the root elem has the the same optional attributes as
 *   all subelements:
 *          the options of loggerparser (see LoggerParser for that)
 *          
 *          and package|class 
 * 
 * 
 * , with the base using the
 *  elem: 
 * 
 
 */

public class LogConfig extends OptionParser<Logger>
{

    
    static String default_log_config="logconfig.xml";


    static public File getLogConfig()
    {
        try
        {
            File f = ConfigFileGetter.getConfigFile(default_log_config);
            //System.out.printf("logconfig: %s\n", (f == null) ? "null" : f.getAbsolutePath());
            return f;
        }
        catch (ConfigFileGetter.ConfigFileNotFound e)
        {
            return null;
        }


    }

    static public LogConfig configureLog(LogConfig lc)
    {
        return configureLog(lc,false);
    }

    static public LogConfig configureLog(LogConfig lc, boolean supress_long_help)
    {
        if (lc==null)
        {
            lc = new LogConfig();
        }
        File logconfig = getLogConfig();
        
        
        if (!lc.isLogConfigXmlSet() && null!=logconfig)
        {
            try
            {
                lc.init(logconfig);
            }
            catch (org.jdom.JDOMException e)
            {
                throw new RuntimeException("xml parsing error while initializing logconfig with default file: "+logconfig.getAbsolutePath(), e);
            }
            catch (IOException e)
            {
                throw new RuntimeException("i/o error while initializing logconfig with default file: "+logconfig.getAbsolutePath(), e);
            }

        }
        lc.suppressLongHelp(supress_long_help);
        
        return lc;
    }


    static boolean DEBUGMODE = false;
        
    static final String LOGCONFIGCLASSNAME = LogConfig.class.getCanonicalName();
    public static final String SQL_LOGGER_NAME ="SQL";
    public static final String LOGGER_LOGGER_NAME ="LOG";
    static Pattern dotp = Pattern.compile("\\.");
    
    public static String default_loglevel = "info";
    public final static String LOGGER_ELEM_NAME ="logger";

    static boolean package_logger_uninited=true;
    static PackageLogger package_logger = new PackageLogger();
    static Element default_config = null;
    static HashMap<String, LoggerParser> name2logger = new HashMap<String, LoggerParser>();
    static HashMap<Class, LoggerParser> class2logger = new HashMap<Class, LoggerParser>();
    static HashMap<String, LoggerParser> classname2logger = new HashMap<String, LoggerParser>();


    static public void reinit()
    {
        Logger.reloadFactory();
        name2logger.clear();
        class2logger.clear();
        classname2logger.clear();
        package_logger = new PackageLogger();
        package_logger_uninited=true;
        default_config=null;
        DEFAULT_LOGGER = getDefaultConfiguredLogger();

    }

    static String getLoggerConfigInfo(String prefix , LoggerParser logconfig)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix+"logconfig:"+logconfig+"\n");
        sb.append(prefix+"logger:"+logconfig.getLogger()+"\n");
        logconfig.logger=null;
        sb.append(prefix+"name:"+logconfig.logger_name+"\n");
        sb.append(prefix+"level: "+logconfig.loglevel+"\n");
        sb.append(prefix+"outfile: "+logconfig.outfile+"\n");
        sb.append(prefix+"immediate flush: "+logconfig.immediate_flush+"\n");
        sb.append(prefix+"append: "+logconfig.append+"\n");
        sb.append(prefix+"date pattern: "+logconfig.date_pattern+"\n");
        sb.append(prefix+"code pattern: "+logconfig.code_pattern+"\n");

        return sb.toString();
    }
    public static String getConfigInfo()
    {
        StringBuffer sb = new StringBuffer();
        sb.append("got "+name2logger.size()+" named logger and "+ class2logger.size()+" logger for classes\n");
        for (Map.Entry<String, LoggerParser> e: name2logger.entrySet())                      
        {
            sb.append("\t logger named: "+e.getKey()+ " is configured as follows:\n");
            sb.append(getLoggerConfigInfo("\t\t", e.getValue()));
            sb.append('\n');
        }
        sb.append('\n');

        for (Map.Entry<Class, LoggerParser> e: class2logger.entrySet())                      
        {
            sb.append("\t logger for class: "+e.getKey().getName()+ " is configured as follows:\n");
            sb.append(getLoggerConfigInfo("\t\t", e.getValue()));
            sb.append('\n');
        }

        sb.append("\npackage specific config: \n");
        package_logger.printInfos("\t", sb);

        return sb.toString();
        
    }

    public static void configureLogger(LoggerParser logconfig, Element elem, String info) throws JDOMException
    {
        for (Object o : elem.getAttributes())
        {
            Attribute a = (Attribute)o;
            if (null==logconfig.getOption(a.getName())) //unknown option for logger parser
                continue;
            
            //System.err.printf("set to %s %s = %s\n", info, a.getName(), a.getValue());
            if (!logconfig.parseOption(a.getName(), a.getValue()))
            {
                throw new RuntimeException("invalid logger config for "+info + " at attrib "+a.getName()+"="+a.getValue());
            }

        }

    }

    static LoggerParser getDefaultConfiguredLogger()
    {
        return getDefaultConfiguredLogger(null);
    }
    static LoggerParser getDefaultConfiguredLogger(ClassLoader cl)
    {

        LoggerParser parser = null;
        if (cl==null)
        {
            parser = new LoggerParser();
        }
        else
        {
            try
            {
                
                Class cz = cl.loadClass("lmu.utils.LoggerParser");
                parser=(LoggerParser)cz.newInstance();
                //java.lang.reflect.Constructor c = cz.getConstructor();
                
                
            }
            catch(Exception e)
            {
                throw new RuntimeException("error at instanciate loggerparser", e);
            }
            
        }
        

        if(default_config!=null)
        {
            try
            {
                configureLogger(parser, default_config, "configured per default");
            }
            catch(JDOMException ex)
            {
                //silent error
            }
            
        }
        else
        {
            parser.parseOption("loglevel","info");
        }
        
        
        return parser;
    }
    static class PackageLogger
    {
        String name;
        String packagename;
        LoggerParser logconfig;
        HashMap<String, PackageLogger> children = new HashMap<String, PackageLogger>();
        Element config_elem = null;

        public lmu.utils.Logger getLogger()
        {
            if (DEBUGMODE)
            {
                
                //System.out.println("LOGCONFIG: 116 getlogger for "+packagename+" is invoked: \n"+ExceptionManager.logCallerStack());
    
                Logger log = logconfig.getLogger(true);
                
                //System.out.println("created logger: "+log+" level: "+log.getLevel()+"\n");
    
                Enumeration en = log.getAllAppenders();
                LinkedList<String> appenders = new LinkedList<String>();
                while (en.hasMoreElements())
                {
                    Appender appender = (Appender)en.nextElement();
                    appenders.add(appender.getName());
                        
                }
                //System.out.println("got "+appenders.size()+" appenders: "+appenders);
                return log;
            }
            
            return logconfig.getLogger(true);
        }
        /** root constructor */
        public PackageLogger()
        {
            logconfig=getDefaultConfiguredLogger();
            
            
        }

        public PackageLogger(String packagename, String name) 
        {
            this();
            this.name=name;
            this.packagename=packagename;
            config_elem = new Element("e");
            config_elem.setAttribute("loggername", packagename);
            

           
        }

        public void configure(String key, String value)
        {
            if (config_elem==null)
            {
                if (DEBUGMODE)
                {
                    System.out.println("create config elem for: "+packagename);
                }
                
                config_elem = new Element("e");
            }
            if (DEBUGMODE)
            {
                System.out.println("configure "+packagename+" "+key+"="+value);
            }
            
            config_elem.setAttribute(key, value);
            
        }
        public void configure(Element e)
        {
            config_elem=e;
        }

        public void initLogger(Element parent_config) 
        {
            if (DEBUGMODE)
            {
                System.out.println("will init logger "+packagename+" got parent config: "+(parent_config!=null)+" got own config: "+(config_elem!=null));
            }
            
            Element config_e = new Element("config");
            try
            {
                
                if (parent_config!=null)
                {
                    XMLUtils.copyAttribs(parent_config, config_e);
                }
                
                if (config_elem!=null)
                {
                    //overwrite specific attirbs
                    XMLUtils.copyAttribs(config_elem, config_e);
                }
                //System.out.println("will configure package logger: " + packagename + " with config " + XMLUtils.prettyXML(config_e));
                config_e.setAttribute("loggername", (packagename==null)?"ROOT":packagename);
                configureLogger(logconfig, config_e, "pckg logger: "+packagename);
                for (PackageLogger pl: children.values())
                {
                    pl.initLogger(config_e);
                }
            }
            catch(JDOMException ex)
            {
                throw new RuntimeException("error while configuring package logger for "+packagename+" config elem: "+XMLUtils.getAttribs(config_e), ex);
            }
            
        }
        /** gets the  most specific logger */
        public LoggerParser getConfiguredLogger(String packagename) 
        {
            String sp[] = dotp.split(packagename);
            PackageLogger rv = this;

            for (int i=0; i<sp.length; i++)
            {
                PackageLogger next = rv.children.get(sp[i]);
                if (next==null)
                    return rv.logconfig;
                
                rv=next;
            }
            return rv.logconfig;
        }
        
        public PackageLogger addLogger(String packagename, Element config) 
        {
            String sp[] = dotp.split(packagename);
            StringBuffer pckgname = new StringBuffer();
            PackageLogger parent = this;

            //System.out.println("add package logger with name: " + packagename);
            for (int i=0; i<sp.length; i++)
            {
                if (pckgname.length()>0)
                {
                    pckgname.append('.');
                }
                pckgname.append(sp[i]);
                PackageLogger next = parent.children.get(sp[i]);

                if (next==null)
                {
                    parent.children.put(sp[i], next = new PackageLogger(pckgname.toString(), sp[i]));
                    
                    
                }
                parent=next;
            }
            //System.out.println("found parent: " + parent.packagename);
            if (config!=null)
            {
                parent.configure(config);
            }
            
            parent.logconfig.setName(packagename);
            return parent;
            
        }

        public void printInfos(String prefix, StringBuffer sb)
        {
            sb.append(prefix+"config for pckg: "+packagename+" name: "+logconfig.logger_name+"\n");
            sb.append("\t"+prefix+"-------------------------------------\n");
            sb.append(getLoggerConfigInfo("\t"+prefix,logconfig));
            sb.append("\t"+prefix+"-------------------------------------\n");
            for (PackageLogger child : children.values())
            {
                child.printInfos("\t"+prefix, sb);
            }
        }
        public String toString()
        {
            StringBuffer sb = new StringBuffer();
            printInfos("",sb);
            return sb.toString();
        }
    }

    
    /** has to invoked every time one changes something in the tree
     * to propagate the changes along the tree
     */
    public static void initPackageLogger()
    {
        package_logger.initLogger(null);

    }
    static public void setDefaultLogger(LoggerParser logger)
    {
        package_logger.logconfig=logger;
    }
    /** sets the default level*/
    static public void setLogLevel(String level)
    {
        package_logger.logconfig.setLevel(level);

    }

    

    static LoggerParser getLogger(String loggername, Class clazz, String pckg, boolean addifnotexists)
    {
        LoggerParser lp= null;
        if (loggername!=null)
        {
            lp=name2logger.get(loggername);
            if (addifnotexists && null==lp)
            {
                //System.out.println("adds logger by name: "+loggername);
                name2logger.put(loggername, lp = getDefaultConfiguredLogger());
            }
            if(lp!=null)return lp;
        }
        if (clazz!=null)
        {
            lp = class2logger.get(clazz);
            if(addifnotexists && null==lp)
            {
                class2logger.put(clazz, lp= getDefaultConfiguredLogger());
                lp.setName(clazz.getCanonicalName());
                classname2logger.put(clazz.getCanonicalName(), lp);
            }
            if(lp!=null)return lp;
        }
        if (pckg!=null)
        {
            if (addifnotexists)
            {
                return package_logger.addLogger(pckg, null).logconfig;
                
                
            }

            //package logger returns always a logger, if none specified the default one
            return package_logger.getConfiguredLogger(pckg);
            
        }
        
        return package_logger.logconfig;
    }
    static public void setLogLevel(String level, String loggername)
    {
        LoggerParser lp = getLogger(loggername, null, null, true);
        lp.setLevel(level);
    }

    static String getKey(int option_index)
    {
        Option o = DEFAULT_LOGGER.getOption(option_index);
        if (o==null)
        {
            throw new RuntimeException("error invlaid option index requested: "+option_index+" for logging...");
        }
        return o.option_name;
    }
    static public void setPackageLogLevel(String level, String packagename)
    {
        PackageLogger pl = package_logger.addLogger(packagename, null);
        pl.configure(getKey(LEVEL_OPTION), level);
        
    }

    static public void setLogLevel(String level, Class clazz)
    {
        //System.out.println("set loglevel invoked for "+clazz+" level: "+level);
        LoggerParser lp = getLogger(null, clazz, null, true);
        lp.setLevel(level);

    }

    static public void setOutFile(String loggername, File f)
    {
        LoggerParser lp = getLogger(loggername, null, null, true);
        lp.setOutFile(f);
    }
    static public void setOutFile(File f)
    {
        package_logger.logconfig.setOutFile(f);
    }

    static public void setPackageOutFile(String packagename, File f)
    {
        PackageLogger pl = package_logger.addLogger(packagename, null);
        pl.configure(getKey(OUT_OPTION), f.getAbsolutePath());
        
    }

    static public void setOutFile(File f, Class clazz)
    {
        LoggerParser lp = getLogger(null, clazz, null, true);
        lp.setOutFile(f);
    }

    static void updateFiles2del(HashSet<File> filestodel, Element e)
    {
        String outfile = XMLUtils.getAttrib(e, "logout", null, false);
        String append = XMLUtils.getAttrib(e, "append", null, false);

        if (outfile==null || append==null || Boolean.parseBoolean(append))
            return;
        

        
        filestodel.add(new File(outfile));
        

    }

    public static void init(Element root) throws JDOMException
    {
        root = XMLUtils.resolveXmlRefs(root);
        package_logger.configure(root);
        default_config = root;
        
        HashSet<File> filestodel = new HashSet<File>();
        updateFiles2del(filestodel,root);
        
        for(Object o : root.getChildren(LOGGER_ELEM_NAME))
        {
            if (!(o instanceof Element))
                continue;

            Element e = (Element)o;
            
            String name = XMLUtils.getAttrib(e, "loggername", null, false);
            String pckgname = XMLUtils.getAttrib(e, "package", null, false);
            String classname = XMLUtils.getAttrib(e, "class", null, false);

            //System.out.printf("add logger name: %s package: %s class: %s\n", name,  pckgname, classname);
            updateFiles2del(filestodel,e);
            LoggerParser lp = null;
            if (name!=null)
            {
                lp = getLogger(name, null, null, true);
                configureLogger(lp, e, "logger named: "+name);
                //System.out.println("will configure loggerparser for "+name+" lp: "+lp);
                name2logger.put(name, lp);
                //System.out.println(lp.getLogger().isDebugEnabled());
            }
            if (classname!=null)
            {
                lp=null;
                try
                {
                    lp = initClassLogger(classname);
                    
                }
                catch(Exception ex)
                {
                    package_logger.getLogger().warn("error while looking for class logger "+classname+": "+ex.getMessage()+"\n"+ExceptionManager.getString(ex));
                }
                if (lp!=null)
                {
                    configureLogger(lp, e, "logger for class: "+classname);
                }
               
                
            }
            if (pckgname!=null)
            {   
                package_logger.addLogger(pckgname, e);
            }
        }

        package_logger.initLogger(null);

        
        //delete the files that are not in append mode (log4j does not do it for some reason...)
        if (DEBUGMODE)
        {
            System.out.println("logconfig will delete following files: "+filestodel);
        }
        
        for (File f : filestodel)
        {
            if (!f.exists())
                continue;


            f.delete();
            
        }
        //System.out.println("got config: "+getConfigInfo());
    }


    public static void init(File xmlconfig) throws IOException, JDOMException
    {
        if (DEBUGMODE)
        {
            System.out.println("init logging with : "+xmlconfig.getAbsolutePath());
        }
        
        SAXBuilder sb = new SAXBuilder();
        Document doc = sb.build(new FileReader(xmlconfig));
        Element root = XMLUtils.resolvePaths(doc.getRootElement());
        init(root);
        
    }

    static public LoggerParser initClassLogger(String classname)
    {
        try
        {
            Class clazz = Class.forName(classname);
            return getLogger(null, clazz,null, true );
        }
        catch(ClassNotFoundException e)
        {
            throw new RuntimeException("cannot find class: "+classname);
        }
        

        
    }
    public static Logger getNamedLogger(String logger_class_name, String name)
    {
        return getLogger(name, (Class)null);
    }
    
    public static Logger getLogger(String name)
    {
        return getNamedLogger((String)null, name);
        
        
        
    }
    public static Logger getLogger(Class c)
    {
        return getLogger(null, c);
    }
    

    static Logger SQLLOGGER = null;

    public static Logger getSQLLogger()
    {
        if (SQLLOGGER!=null)
        {
            return SQLLOGGER;
        }
        return SQLLOGGER=getLogger(SQL_LOGGER_NAME);
    }
    public static Logger getLogger()
    {
        return getLogger((String)null, getInvoker());

    }

    public static Logger getNamedLogger(String name)
    {
        return getNamedLogger((String)null, null, name);
    }
    static Class  getInvoker()
    {
        RuntimeException re = new RuntimeException("caller stack");
        StackTraceElement[] stea = re.getStackTrace();
        
        boolean ownfound=false;
        for (int i=0; i<stea.length; i++)
        {
            String cname = stea[i].getClassName();
            if (cname.equals(LOGCONFIGCLASSNAME))
            {
                ownfound=true;
            }
            else
            {
                if (!ownfound)
                    continue;
                
                String testname=stea[i].getClassName();
                try
                {
                    return Class.forName(testname);
                }
                catch(ClassNotFoundException ex)
                {
                    //cannot happen - we derived the classname from stack, the class is known
                    System.err.println("cannot happen, but still... don't find class :"+testname+" found in stacktrace depth: "+i+" exception: "+ 
                                       ExceptionManager.getStackTrace(ex));
                }
                
            }

        }
        return null;                                                                                        
        
    }
    static HashMap<LoggerParser, Logger> lp2log = new HashMap<LoggerParser, Logger>();
    static HashMap<Logger, Level> logger2level = new HashMap<Logger, Level>();
    static HashMap<Logger, Vector<String>> logger2invoker = new HashMap<Logger, Vector<String>>();

    static String getOldInfos(Logger l)
    {
        Vector<String> oldinfos = logger2invoker.get(l);
        if (oldinfos==null)
        {
            return "NO INFOS YET STORED";
        }
        StringBuffer sb = new StringBuffer("GOT "+oldinfos.size()+" infos!:\n\n");

        for (int i=0; i<oldinfos.size(); i++)
        {
            sb.append("\n-------------------------- info "+(i+1)+" ---------------------------------\n");
            sb.append(oldinfos.get(i));
            sb.append("\n-------------------------- END "+(i+1)+" ---------------------------------\n\n\n");
        }
        return sb.toString();
    }
    private static Logger getLogger(LoggerParser lp)
    {
        return getNamedLogger(lp, null);
    }
    private static Logger getNamedLogger(LoggerParser lp, String additional_name)                                                        
    {

        String infos = "";
        

        Logger log = lp.getNamedLogger(additional_name);

        if (DEBUGMODE)
        {
            infos="caller stack: "+ ExceptionManager.logCallerStack()+"\n\n";
            infos+="\nconfig: "+getLoggerConfigInfo("\t", lp);                                                   
            System.out.println("logger appenders: ");
            Enumeration e = log.getAllAppenders();
            while (e.hasMoreElements())
            {
                Appender a = (Appender)e.nextElement();
                System.out.println("\t next appender :"+a.getName());
            }
            
            infos+="\n\nlogger:"+log+"\nloglevel: "+log.getLevel()+"\n";
        }
        

        Logger old = lp2log.get(lp);
        if (old!=null && old!=log)
        {
            //System.err.println("ERROR REPLCATE LOGGER: "+old+"! new infos:\n"+infos+"\nOLD:\n"+getOldInfos(old));
        }

        Level l =  logger2level.get(log);
        if (l!=null && l!=log.getLevel())
        {
            //System.err.println("ERROR OLD LOGLEVEL CHANGED ("+l+"=>"+log.getLevel()+" for"+log+":\nparser: "+" new infos:\n"+infos+"\nOLD:\n"+getOldInfos(old));
        }
        lp2log.put(lp, log);
        logger2level.put(log, log.getLevel());
        Vector<String> info_v = logger2invoker.get(log);
        if (info_v==null)
        {
            info_v = new Vector<String>();
        }
        info_v.add(infos);
        return log;
    }

    public static Logger getLogger(String name, Class c)
    {
        return getNamedLogger(name, c, null);
    }
    public static Logger getNamedLogger(String name, Class c, String additional_name)
    {
        //System.out.println("get named logger name: "+name+" c: "+c+" additional name "+additional_name);
        if (DEBUGMODE)
        {
            System.out.println("getLogger invoked with name: "+name+" and class: "+c);
            System.out.println("named loggers: "+name2logger.keySet());
            System.out.println("class loggers: "+class2logger.keySet());
        }
        
        if (c==null)
        {
            c=getInvoker();
        }
        if (name==null && c==null)
        {
            return  DEFAULT_LOGGER.getLogger();
        }
        if (DEBUGMODE)
        {
            System.out.println("getLogger determined the class name: "+c.getCanonicalName());
            System.out.println("caller stack: "+ ExceptionManager.logCallerStack());
        }
        
        
        LoggerParser lp =null;
        if (name!=null && null!=(lp=name2logger.get(name)))
        {
            if (DEBUGMODE)
            {
                System.out.println("get name logger, config: "+getLoggerConfigInfo("\t", lp));
            }
            
            return getNamedLogger(lp, additional_name);
        }

        if (c!=null && null!=(lp=class2logger.get(c)))
        {
            if (DEBUGMODE)
            {
                System.out.println("found class logger for "+c+" config: "+getLoggerConfigInfo("\t", lp));
            }
            
            return getNamedLogger(lp, additional_name);
        }
        
        if (DEBUGMODE)
        {
            System.out.println("package_logger_uninited: "+package_logger_uninited);
        }
        
        if (package_logger_uninited)
        {
            package_logger_uninited=false;
            initPackageLogger();
            if (DEBUGMODE)
            {
                System.out.println("package logging initalized config is: \n"+getConfigInfo());
            }
            
        }
        //System.out.println("logger for package: "+c);
        //System.out.println("logger for package: "+c.getPackage());
        String packagename=null;
        if (c.getPackage()==null)
        {
            packagename=c.getCanonicalName();
            packagename=packagename.substring(0,packagename.lastIndexOf('.'));
        }
        else
        {
            packagename=c.getPackage().getName();
        }
        lp=package_logger.getConfiguredLogger(packagename);
        class2logger.put(c, lp);
        if (DEBUGMODE)
        {
            System.out.println("package logger for "+c.getPackage().getName()+" is configured as:\n"+getLoggerConfigInfo("\t" ,lp));
        }
        
        Logger logger = getNamedLogger(lp, additional_name);
        if (DEBUGMODE)
        {
            
            Enumeration en = logger.getAllAppenders();
            LinkedList<String> appenders = new LinkedList<String>();
            while (en.hasMoreElements())
            {
                Appender appender = (Appender)en.nextElement();
                appenders.add(appender.getName());
                    
            }
            System.out.println("got "+appenders.size()+" appenders: "+appenders);
        }
        return logger;
        
    }

    public LogConfig()
    {
        this("");
    }

    static final int LEVEL_OPTION=LoggerParser.LEVEL_OPTION;
    static final int OUT_OPTION=LoggerParser.OUT_OPTION;
    static final int NODATE_OPTION=LoggerParser.NODATE_OPTION;
    static final int DATEFMT_OPTION=LoggerParser.DATEFMT_OPTION;
    static final int CODEFMT_OPTION=LoggerParser.CODEFMT_OPTION;
    static final int NAME_OPTION=LoggerParser.NAME_OPTION;
    static final int FLUSH_OPTION=LoggerParser.FLUSH_OPTION;
    static final int APPEND_OPTION=LoggerParser.APPEND_OPTION;


    static final int PACKAGE_LEVEL_START=APPEND_OPTION+1;
    static final int PACKAGE_LEVEL_OPTION=PACKAGE_LEVEL_START+LEVEL_OPTION;
    static final int PACKAGE_OUT_OPTION=PACKAGE_LEVEL_START+OUT_OPTION;
    static final int PACKAGE_NODATE_OPTION=PACKAGE_LEVEL_START+NODATE_OPTION;
    static final int PACKAGE_DATEFMT_OPTION=PACKAGE_LEVEL_START+DATEFMT_OPTION;
    static final int PACKAGE_CODEFMT_OPTION=PACKAGE_LEVEL_START+CODEFMT_OPTION;
    static final int PACKAGE_NAME_OPTION=PACKAGE_LEVEL_START+NAME_OPTION;
    static final int PACKAGE_FLUSH_OPTION=PACKAGE_LEVEL_START+FLUSH_OPTION;
    static final int PACKAGE_APPEND_OPTION=PACKAGE_LEVEL_START+APPEND_OPTION;

    static final int CLASS_LEVEL_START = PACKAGE_APPEND_OPTION+1;

    static final int CLASS_LEVEL_OPTION=CLASS_LEVEL_START+LEVEL_OPTION;
    static final int CLASS_OUT_OPTION=CLASS_LEVEL_START+OUT_OPTION;
    static final int CLASS_NODATE_OPTION=CLASS_LEVEL_START+NODATE_OPTION;
    static final int CLASS_DATEFMT_OPTION=CLASS_LEVEL_START+DATEFMT_OPTION;
    static final int CLASS_CODEFMT_OPTION=CLASS_LEVEL_START+CODEFMT_OPTION;
    static final int CLASS_NAME_OPTION=CLASS_LEVEL_START+NAME_OPTION;
    static final int CLASS_FLUSH_OPTION=CLASS_LEVEL_START+FLUSH_OPTION;
    static final int CLASS_APPEND_OPTION=CLASS_LEVEL_START+APPEND_OPTION;


    static final int NAME_LEVEL_START = CLASS_APPEND_OPTION+1;

    static final int NAME_LEVEL_OPTION=NAME_LEVEL_START+LEVEL_OPTION;
    static final int NAME_OUT_OPTION=NAME_LEVEL_START+OUT_OPTION;
    static final int NAME_NODATE_OPTION=NAME_LEVEL_START+NODATE_OPTION;
    static final int NAME_DATEFMT_OPTION=NAME_LEVEL_START+DATEFMT_OPTION;
    static final int NAME_CODEFMT_OPTION=NAME_LEVEL_START+CODEFMT_OPTION;
    static final int NAME_NAME_OPTION=NAME_LEVEL_START+NAME_OPTION;
    static final int NAME_FLUSH_OPTION=NAME_LEVEL_START+FLUSH_OPTION;
    static final int NAME_APPEND_OPTION=NAME_LEVEL_START+APPEND_OPTION;

        
    static final int LOGGER_CMDLINE_CONFIG_END=NAME_APPEND_OPTION+1;
        
    static final int CONFIGFILE_OPTION = LOGGER_CMDLINE_CONFIG_END;


    static LoggerParser DEFAULT_LOGGER = getDefaultConfiguredLogger();

    void addLogOptions(String prefix, int level_start)
    {
        int num_params=(prefix.length()==0)?1:2;
        LoggerParser dp = DEFAULT_LOGGER;
        //System.out.println("DEFUAULT LOGGER classloder: "+dp.getClass().getClassLoader());
        //System.out.println("current classloder: "+this.getClass().getClassLoader());
        /*if (dp.getClass().getClassLoader()!=this.getClass().getClassLoader())
        {
            dp=DEFAULT_LOGGER=getDefaultConfiguredLogger(this.getClass().getClassLoader());
            System.out.println("relaoded DEFUAULT LOGGER classloder: "+dp.getClass().getClassLoader());
        }
        */
        
        Option o = registerOption(level_start+LEVEL_OPTION, prefix+"loglevel", "loglevel", dp.loglevel, num_params);
        o.setEnumParam(num_params, LoggerParser.Log4JLevel.getLevelNames());
        registerOption(level_start+OUT_OPTION, prefix+"logout", "logfile", "stdout", num_params);
        
        registerOption(level_start+NODATE_OPTION, prefix+"nodate", " no date in log formatting", "<optional>", num_params-1);
        registerOption(level_start+DATEFMT_OPTION, prefix+"logdatefmt", "dateformat", dp.date_pattern, num_params);
        registerOption(level_start+CODEFMT_OPTION, prefix+"codefmt", "codeformat", dp.code_pattern, num_params);
        registerOption(level_start+NAME_OPTION, prefix+"loggername", "name of logger", dp.logger_name, num_params);
        o=registerOption(level_start+FLUSH_OPTION, prefix+"logflush", "boolean", ""+dp.immediate_flush, num_params);
        o.setBoolParam(1);
        

    }

    enum CONFIGTYPE
    {
        DEFAULT(0, 0),
        PACKAGE(PACKAGE_LEVEL_START,1),
        CLASS(CLASS_LEVEL_START,1),
        NAME(NAME_LEVEL_START,1)
        ;

        int option_start;
        int skip_param=0;

        CONFIGTYPE(int s, int skip)
        {
            option_start=s;
            skip_param=skip;
        }
    }
    
    public LogConfig(String prefix)
    {
        super(prefix);

        addLogOptions("",0);
        addLogOptions("pckg-", PACKAGE_LEVEL_START);
        addLogOptions("class-", CLASS_LEVEL_START);
        addLogOptions("name-", NAME_LEVEL_START);
        Option o = registerOption(CONFIGFILE_OPTION, "logconfigxml", "xml config for logging", "<optional>", 1);
        o.setFileParam(1);

        
        
    }

    boolean configureLogger(LoggerParser lp, int option_index, String[] args, int i, StringBuffer error)
    {
        return lp.parse(option_index,args,i,error);
    }
    boolean logconfigxmlset=false;

    public boolean isLogConfigXmlSet()
    {
        return logconfigxmlset;
    }
    public boolean parse(int option_index, String[] args, int i, StringBuffer error)
    {
        if (DEBUGMODE)
        {
            System.out.println("LOGCONFIG gets option: "+option_index+ " option: "+getOption(option_index));
        }
        
        if (option_index>=LOGGER_CMDLINE_CONFIG_END)
        {

            switch(option_index)
            {
                case CONFIGFILE_OPTION:
                    try
                    {
                        init(new File(args[i+1]));
                        logconfigxmlset=true;
                    }
                    catch(Exception e)
                    {
                        error.append("error while reading log config file: "+e.getMessage());
                        return false;
                    }
                    
                    return true;

            }
            return false;
        }
        
        CONFIGTYPE configtype = null;
        for (CONFIGTYPE ct : CONFIGTYPE.values())
        {
            if (ct.option_start>option_index)
                continue;
            
            if (configtype!=null && configtype.option_start>ct.option_start)
            {
                continue;
            }

            configtype=ct;
        }
        int cindex=option_index-configtype.option_start;
        int ci = i+configtype.skip_param;

        if (DEBUGMODE)
        {
            System.out.println("got param: "+args[i]+" will configure configtype: "+configtype);
        }
        
        Option o = DEFAULT_LOGGER.getOption(cindex);
        if (o==null)
        {
            throw new RuntimeException("PROGRAMMING BUG: option index: "+cindex+" not found for LoggerParser");
        }

        switch(configtype)
        {
            case DEFAULT:
                package_logger.configure(o.option_name, args[ci+1]);
                return true;
            case NAME:
                return configureLogger(getLogger(args[i+1], null, null, true), cindex, args, ci, error);
            case PACKAGE:
                PackageLogger plogger = package_logger.addLogger(args[i+1], null);
                plogger.configure(o.option_name, args[ci+1]);

                return true;


            case CLASS:
                try
                {
                    Class clazz = Class.forName(args[i+1]);
                    return configureLogger(getLogger(null, clazz,null, true), cindex, args, ci, error);
                }
                
                catch(ClassNotFoundException ex)
                {
                    error.append("cannot find class for name: "+args[i+1]+" to configure logging");
                    return false;
                }
                
        }
        return false;
    }

    public String getName()
    {
        return "LogConfig";
    }

    public Logger getObject()
    {
        return package_logger.getLogger();

    }

    public static void main (String args[]) 
    {
        System.out.println("variables: "+System.getenv()+"\n\n\n");

        System.out.println("properties: "+System.getProperties());
        LogConfig lc = new LogConfig();
        if (args.length==0)
        {
            System.out.println("USAGE:\n"+lc.getHelp());
            return;
        }
        if(!OptionParser.parseParams(args, true, true, true, lc))
        {
            return;
        }
        initPackageLogger();
        System.out.println("config info: "+LogConfig.getConfigInfo());
    }
}
