package lmu.utils;

import org.apache.log4j.*;


import lmu.utils.OptionParser.Option;
import java.io.*;
import java.util.regex.*;

public class LoggerParser extends OptionParser<Logger>
{
    
    public enum Log4JLevel
    {

        DEBUG(Level.DEBUG, "debug"),
        INFO(Level.INFO, "info"),
        FATAL(Level.FATAL, "fatal"),
        TRACE(Level.ALL, "trace"),
        WARN(Level.WARN, "warn"),
        ERROR(Level.ERROR, "error"),
        OFF(Level.OFF, "off")
        ;
        
        Level level;
        String name;

        Log4JLevel(Level l, String n)
        {
            level=l; name=n;
        }

        static String[] getLevelNames()
        {
            int nlevels=0;
            for (Log4JLevel l : Log4JLevel.values())
            {
                nlevels++;
            }
            String[] rv = new String[nlevels];
            int i=0;
            for (Log4JLevel l : Log4JLevel.values())
            {
                rv[i++]=l.name;
            }
            return rv;
        }
        static String getLevels()
        {
            StringBuffer names = new StringBuffer();
            
            for (Log4JLevel l : Log4JLevel.values())
            {
                if (names.length()>0)
                {
                    names.append('|');
                }
                names.append(l.name);
            }

            return names.toString();

        }
        static Level get(String n)
        {
            for (Log4JLevel l : Log4JLevel.values())
            {
                if (l.name.equals(n))
                {
                    return l.level;
                }
            }

            throw new RuntimeException("unknown log level "+n+ " valid level names are: "+ getLevels());
        }
    }

    String logger_name="";
    String date_pattern="yyyy-MM-dd HH:mm:ss";
    String code_pattern="%F:%L %c{1}- %m%n";
    String loglevel = "error";
    String default_conversion_pattern="%-5p: %d{yyyy-MM-dd HH:mm:ss} %F:%L %c{1}- %m%n";

    static Pattern varP = Pattern.compile("\\$\\{([^\\}]+)\\}");

    File outfile = null;
    boolean immediate_flush=true;
    boolean append=false;


    public String getName()
    {
        return "LoggerParser";
    }
    public void setAppend(boolean b)
    {
        append=b;
    }
    public void setFlush(boolean b)
    {
        immediate_flush=b;
    }

    public void setName(String s)
    {
        //System.out.println(this+" set LOGGER NAME "+s);
        logger_name=s;
    }
    public void setLevel(String s)
    {
        loglevel=s;
        logger = null;
    }
    enum OUTTYPE
    {
        STDOUT("stdout"),
        STDERR("stderr")
        ;

        String name=null;
        OUTTYPE(String n)
        {
            name=n;
        }

        static OUTTYPE get(String s)
        {
            if (s==null || s.length()==0)
            {
                return null;
            }
            return null;
        }

    }
    /*public void setOut(String outtype)
    {
        OUTTYPE ot = 
    }
    */
    public void setOutFile(String s)
    {
        if (s==null || s.length()==0 || s.equals("stdout"))
        {
            outfile=null;
            return;
        }
        String path = s;
        
        Matcher m = varP.matcher(path);
        int start=0;
        StringBuffer sb = new StringBuffer();
        boolean invalid=false;
        while (m.find(start))
        {
            sb.append(path.substring(start,m.start()));
            String evar = m.group(1);
            String env= System.getenv(evar);
            String prop = System.getProperty(evar);
            
            String part = (prop==null)?env:prop;
            
            if (part==null)
            {
                System.err.println("cannot find value for env variable: "+evar+" for logging, fall back to system out");
                invalid=true;
                break;
            }
            sb.append(part);
            start=m.end();

        }
        if (invalid)
        {
            outfile=null;
            return;
        }
        if (start<path.length())
        {
            sb.append(path.substring(start));
        }

        
        outfile = new File(sb.toString());

    }
    public void setOutFile(File f)
    {
        if (f==null || f.getName().length()==0 || f.getName().equals("stdout"))
        {
            outfile=null;
            return;
        }
        String path = f.getName();
        System.out.println("check for variables: "+f+" abs: "+path);
        Matcher m = varP.matcher(path);
        int start=0;
        StringBuffer sb = new StringBuffer();
        while (m.find(start))
        {
            sb.append(path.substring(start,m.start()));
            System.out.println("next: "+m.group());
            start=m.end()+1;

        }
        if (start<path.length())
        {
            sb.append(path.substring(start));
        }
        outfile=f;
    }
    public void setNoDate()
    {
        setDatePattern(null);
    }
    public void setDatePattern(String p)
    {
        date_pattern=p;
    }
    public void setCodePattern(String s)
    {
        code_pattern=s;
    }

    public PatternLayout getLayout()
    {
        PatternLayout layout = new PatternLayout();
        String pattern = "%-5p: ";
        if (date_pattern!=null && date_pattern.length()!=0)
        {
            pattern+="%d{"+date_pattern+"} ";
        }
        if (code_pattern!=null && code_pattern.length()!=0)
        {
            pattern+=code_pattern;
        }
        
        layout.setConversionPattern(pattern);
        return layout;
    }

    
    public static lmu.utils.Logger getFileLogger(String filename, String name) 
    {
        LoggerParser lp = new LoggerParser();
        lp.setOutFile(filename);
        lp.setName(name);
        return lp.getLogger();
    }

    public static lmu.utils.Logger getConsoleLogger(String name) 
    {
        LoggerParser lp = new LoggerParser();
        lp.setName(name);
        return lp.getLogger();
    }
    
    public lmu.utils.Logger getLogger(boolean nocache)
    {
        Logger rv = getObject();
        logger=null;
        return rv;
    }
    public lmu.utils.Logger getLogger()
    {
        return getObject();
    }

    lmu.utils.Logger logger = null;

    
    public Logger getNamedLogger(String additional_name)
    {
        
        if (additional_name == null && logger!=null)
        {
            return logger;
        }
        //System.out.println(this+" will get logger to : "+logger_name);
        //logger = org.apache.log4j.Logger.getLogger(logger_name);
        String ln = (additional_name == null) ? logger_name : additional_name;

        
        logger = (lmu.utils.Logger)lmu.utils.Logger.getLogger(ln);
        logger.setAdditivity(false);
        PatternLayout layout = getLayout();
        
        //we use only a single appender (for now - so let's check if therr
        //are others, if so remove them
        logger.removeAllAppenders();

        if (outfile==null)
        {
            ConsoleAppender ca=new ConsoleAppender(layout);
            ca.setName("console appender");
            logger.addAppender(ca);
        }
        else
        {
            try
            {
                FileAppender fa = new FileAppender(layout, outfile.getAbsolutePath());
                fa.setAppend(append);
                fa.setName("fileappender to "+outfile.getAbsolutePath()+" append: "+fa.getAppend());
                fa.setFile(outfile.getAbsolutePath());
                fa.setImmediateFlush(immediate_flush);
                logger.addAppender(fa);
                
            }
            catch(IOException e)
            {
                throw new RuntimeException(e);
            }
            
        }
        logger.setLevel(Log4JLevel.get(loglevel));
        return logger;
    }

    public Logger getObject()
    {
        if (logger!=null)
            return logger;
        
        return logger = getNamedLogger(null);
    }
    
    public void debug(Exception e)
    {
        logger.debug(ExceptionManager.getString(e));
    }

    public void debug(String msg)
    {
        logger.debug(msg);
    }
    public void debug(String fmt, Object ... params)
    {
        logger.debug(String.format(fmt, params));
    }
    

    public void trace(String msg)
    {
        logger.trace(msg);
    }
    public void trace(Exception e)
    {
        logger.trace(ExceptionManager.getString(e));
    }

    public void trace(String fmt, Object ... params)
    {
        logger.trace(String.format(fmt, params));
    }


    public void error(String msg)
    {
        logger.error(msg);
    }

    public void error(Exception e)
    {
        logger.error(ExceptionManager.getString(e));
    }

    public void error(String fmt, Object ... params)
    {
        logger.error(String.format(fmt, params));
    }

    public void warn(String msg)
    {
        logger.warn(msg);
    }


    public void warn(Exception e)
    {
        logger.warn(ExceptionManager.getString(e));
    }

    public void warn(String fmt, Object... params)
    {
        logger.warn(String.format(fmt, params));
    }

    public void info(String msg)
    {
        logger.info(msg);
    }



    public void info(Exception e)
    {
        logger.info(ExceptionManager.getString(e));
    }


    public void info(String fmt, Object... params)
    {
        logger.info(String.format(fmt, params));
        
    }


    public void fatal(String msg)
    {
        logger.fatal(msg);
    }


    public void fatal(Exception e)
    {
        logger.fatal(ExceptionManager.getString(e));
    }


    public void fatal(String fmt, Object... params)
    {
        logger.fatal(String.format(fmt, params));
        
    }


    static final int LEVEL_OPTION=0;
    static final int OUT_OPTION=1;
    static final int NODATE_OPTION=2;
    static final int DATEFMT_OPTION=3;
    static final int CODEFMT_OPTION=4;
    static final int NAME_OPTION=5;
    static final int FLUSH_OPTION=6;
    static final int APPEND_OPTION=7;

    
    public static lmu.utils.Logger getNoLogger()
    {
        LoggerParser lp = new LoggerParser();
        lp.setLevel("off");
        return lp.getLogger();
    }
    public LoggerParser()
    {
        this("");
    }
    public LoggerParser(String prefix)
    {
        super(prefix);
        Option o = registerOption(LEVEL_OPTION, "loglevel", "loglevel", loglevel, 1);
        o.setEnumParam(1, Log4JLevel.getLevelNames());
        registerOption(OUT_OPTION, "logout", "logfile", "stdout", 1);
        
        registerOption(NODATE_OPTION, "nodate", "", null, 0);
        registerOption(DATEFMT_OPTION, "logdatefmt", "dateformat", date_pattern, 1);
        registerOption(CODEFMT_OPTION, "codefmt", "codeformat", code_pattern, 1);
        registerOption(NAME_OPTION, "loggername", "name of logger", logger_name, 1);
        o=registerOption(FLUSH_OPTION, "logflush", "boolean", ""+immediate_flush, 1);
        o.setBoolParam(1);
        o=registerOption(FLUSH_OPTION, "logappend", "boolean", ""+immediate_flush, 1);
        o.setBoolParam(1);

    }
    public boolean parse(int option_index, String[] args, int i, StringBuffer error)
    {
        
        
        if (option_index==NODATE_OPTION)
        {
            setNoDate();
            return true;
        }
        
        String value = args[i+1];
        
        switch(option_index)
        {
            case LEVEL_OPTION:
                setLevel(value);
                
                break;
            case OUT_OPTION:
                setOutFile(value);
                break;
            case DATEFMT_OPTION:
                setDatePattern(value);
                break;

            case CODEFMT_OPTION:
                setCodePattern(value);
                break;
            case NAME_OPTION:
                setName(value);
                break;
            case FLUSH_OPTION:
                setFlush(Boolean.parseBoolean(value));
                break;
            case APPEND_OPTION:
                setAppend(Boolean.parseBoolean(value));
                break;

            default:
                //unexcpected
                throw new RuntimeException("implementation error! unknown option index: "+option_index+" for switch: "+args[i]);
        }
        return true;

    }
        
}


