package lmu.utils;

import java.util.*;
import java.io.*;

public class MapBuilder 
{

    public static class MB<A,B> extends HashMap<A,B>
    {
        public MB(A a, B b)
        {
            super();
            put(a, b);
        }

        public MB<A,B> add(A a, B b)
        {
            put(a,b);
            return this;
        }
    }

    public static<A,B> MB<A,B> build(A a, B b)
    {
        return new MB(a,b);
    }



    public static class MSB<A,B> extends HashMap<A, HashSet<B>>
    {
        public MSB(A a, B b)
        {
            super();
            add(a,b);
        }
    
        public MSB<A,B> add(A a, B b)
        {
            update(this, a, b);
            return this;
        }
    }
    
    public static<A,B> MSB<A,B> buildMultiSet(A a, B b)
    {
        return new MSB(a,b);
    }

/*    public static<A,B> HashMap<A,B> build(Object ... keyvalues)
    {
        HashMap<A,B> rv = new HashMap<A,B>();

        for (int i=0; i<keyvalues.length; i+=2)
        {
            rv.put((A)keyvalues[i], (B)keyvalues[i+1]);
        }
        return rv;
    }

  */

    
        
    public static<A extends Object> Set<A> buildSet(A ... array)
    {
        HashSet<A> rv = new HashSet<>();

        for (A a : array)
        {
            rv.add(a);
        }
        return rv;
    }
    
    public static<A,B> void update(Map<A, HashSet<B>> map, A key, B obj)
    {
        HashSet<B> h = map.get(key);
        if (h == null)
        {
            map.put(key, h = new HashSet<B>());
        }
        h.add(obj);
    }

    public static<A,B> void updateV(HashMap<A, Vector<B>> map, A key, B obj)
    {
        Vector<B> h = map.get(key);
        if (h == null)
        {
            map.put(key, h = new Vector<B>());
        }
        h.add(obj);
    }

    public static<A,B> void update(HashMap<A, HashSet<B>> map, A key, Collection<B> objs)
    {
        HashSet<B> h = map.get(key);
        if (h == null)
        {
            map.put(key, h = new HashSet<B>());
        }
        h.addAll(objs);
    }

    public static<A> void update(Map<A, Integer> counts, A key)
    {
        update(counts, key, 1);
    }

    public static<A> void update(Map<A, Integer> counts, A key, int inc)
    {
        Integer old = counts.get(key);
        if (old == null)
        {
            counts.put(key, inc);
            return;
        }
        counts.put(key, old + inc);
    }


    public static<A> void update(HashMap<A, Double> scores, A key, double inc)
    {
        Double old = scores.get(key);
        if (old == null)
        {
            scores.put(key, inc);
            return;
        }
        scores.put(key, old + inc);
    }

    public interface InfoGetter<A>
    {
        public String getInfo(A a);
    }


    public static<A> void printSortedInfos(PrintWriter pw, HashMap<A, Integer> m, Integer all_size, int limit)
    {
        printSortedInfos(pw, m, all_size, limit, (A a) -> ""+a);
    }

    public static<A> void printSortedInfos(PrintWriter pw, HashMap<A, Integer> m, Integer all_size, int limit, InfoGetter<A> ig)
    {
        Vector<Pair<Integer, A>> mappings =  Pair.convert_reverse_sorted(m, false);
        double tperc = 0;
        Double norm = (all_size == null) ? null : 100.0 / all_size;
        
        for (int i=0; i<mappings.size() && (limit <0 || i<limit); i++)
        {
         Pair<Integer, A> p = mappings.get(i);
        
         double perc = (norm == null) ? 0 : p.getFirst() * norm;
         tperc += perc;
        
         String info = (norm == null) ? "" : String.format(" (%2.2f %%) until now: %2.2f%%\n", perc, tperc);
        
         System.out.printf("\t%s mapped: %d%s", ig.getInfo(p.getSecond()), p.getFirst(), info);
        }
    }
    

    public static<A> void printSortedInfos(HashMap<A, Integer> m, Integer all_size, int limit)
    {
        printSortedInfos(m, all_size, limit, (A a) -> ""+a);
    }
    
    public static<A> void printSortedInfos(HashMap<A, Integer> m, Integer all_size, int limit, InfoGetter<A> ig)
    {
        BufferPrintWriter bpw = new BufferPrintWriter();
        printSortedInfos(bpw.getWriter(), m, all_size, limit, ig);
        System.out.println(bpw.getBuffer());
    }



}
