package lmu.utils;

import java.util.*;

/** inputs are two sorted iterators   */
public class MergeSortIterator<A,B> implements Iterator<Pair<A,B>>
{
    public interface DComparator<A,B>
    {
        public int compareTo(A a, B b);
    }

    Pair<A,B> p = new Pair<>(null, null);

    PeakableIterator<A> it1;
    PeakableIterator<B> it2;
    DComparator<A, B> comp;

    public<C extends Number, D extends Number> MergeSortIterator(Iterator<A> it1, Iterator<B> it2, NumUtils.NumberGetter<A, C> ng1, NumUtils.NumberGetter<B, D> ng2)
    {
        this.it1 = new PeakableIterator<A>(it1);
        this.it2 = new PeakableIterator<B>(it2);
        comp = (A a, B b) -> Double.compare(ng1.get(a).doubleValue(), ng2.get(b).doubleValue());
    }


    public MergeSortIterator(Vector<A> v1, Vector<B> v2, DComparator<A, B> comp)
    {
        this(v1.iterator(), v2.iterator(), comp);
    }

    public MergeSortIterator(Iterator<A> it1, Iterator<B> it2, DComparator<A, B> comp)
    {
        this.it1 = new PeakableIterator<A>(it1);
        this.it2 = new PeakableIterator<B>(it2);
        this.comp = comp;
    }

    public boolean hasNext()
    {
        return it1.hasNext() || it2.hasNext();
    }

    public Pair<A,B> next()
    {
        Pair<A,B> p = new Pair<>(null, null);
        p.set(null, null);
        A a = it1.peak();
        B b = it2.peak();

        //System.out.printf("next :: a: %s b : %s\n", a, b);
        if (a == null || b == null)
        {
            if (a != null)
            {
                p.setFirst(it1.next());
            }
            if (b != null)
            {
                p.setSecond(it2.next());
            }
            return p;
        }
        int c = comp.compareTo(a, b);
        

        //System.out.printf("v1: %2.4f v2: %2.4f\n", v1, v2);

        if (c <= 0)
        {
            p.setFirst(it1.next());
            return p;
        }
        p.setSecond(it2.next());
        return p;
    }

    public static void main (String args[]) 
    {
        for (int i=0; i<100; i++)
        {
            HashSet<Double> d = new HashSet<>();
            HashSet<Integer> si = new HashSet<>();

            int N = 1000;
            int size1 = (int)(10 * Math.random());
            int size2 = (int)(10 * Math.random());

            System.out.printf("check %d vs %d\n", size1, size2);
            while (d.size() < size1)
            {
                d.add(1000.0 * Math.random());
            }
            
            while (si.size() < size2)
            {
                si.add((int)(1000.0 * Math.random()));
            }

            Vector<Double> v1 = ObjectGetter.toVector(d);

            Vector<Integer> v2 = ObjectGetter.toVector(si);

            Collections.sort(v1);
            Collections.sort(v2);

            
            MergeSortIterator<Double, Integer> msi = new MergeSortIterator<>(v1, v2, (Double _d, Integer _i) -> Double.compare(_d, (double)_i));

            Vector<Pair<Double, Integer>> sorted = ObjectGetter.toVector(msi);

            double last = Double.NEGATIVE_INFINITY;

            boolean invalid = false;
            for (Pair<Double, Integer> p : sorted)
            {
                double v = (p.getFirst() != null) ? p.getFirst() : p.getSecond().doubleValue();
                if (v < last)
                {
                    invalid = true;
                    break;
                }
                last = v;
            }
            if (invalid)
            {
                throw new FRuntimeException("ERROR!\n");
            }
            System.out.printf("TEST: %d passed\n", i+1);

        }
    }
}


