package lmu.utils;

import java.util.*;

import org.apache.commons.math3.stat.correlation.PearsonsCorrelation;
import org.apache.commons.math3.stat.inference.TTest;

public class NumUtils 
{
    static TTest TTester = new TTest();

    public enum AGGREGATE
    {
        MIN("min"),
        MAX("max"),
        COUNT("count"),
        MEDIAN("median"),
        SUM("sum"),
        AVG("avg"),
        SD("SD"),
        FIRST("first")
        ;

        String n;

        AGGREGATE(String n)
        {
            this.n = n;
            EnumGetter.add(n, this);
        }

        public static AGGREGATE get(String n)
        {
            return (AGGREGATE)EnumGetter.get(n);
        }
    }

    
    public static<A extends Number> Number reduce(Collection<A> a, String s)
    {
        AGGREGATE ag = AGGREGATE.get(s);
        return reduce(a, (ag==null) ? AGGREGATE.FIRST : ag);
    }


    
    public static<A extends Number> Number reduce(Collection<A> a, AGGREGATE ag)
    {
        if (a==null || a.size() == 0)
            return null;
        
        
        boolean havetosort = false;
        switch(ag)
        {
            case FIRST:
                return a.iterator().next();
            case COUNT:
                return a.size();
            case MIN:
            case MAX:
            case MEDIAN:
                havetosort = true;
                break;
            default: 
                break;
        }


        BasicNumberInfo<A> bni = getNumInfo(a, !havetosort);

        if(bni==null)
        {
        	return null;
        }
        
        switch(ag)
        {
            case MIN:
                return bni.min;
            case MAX:
                return bni.max;
            case AVG:
                return bni.mean;
            case SD:
                return bni.SD;
            case MEDIAN:
                return bni.median;
            case SUM:
                return bni.sum;


            default:
                return null;
        }
    }
    static double[] quantiles = {0, 0.25, 0.5, 0.75, 1.0};

    public static class BasicNumberInfo<A extends Number>
    {
        public int num_elems;

        public Vector<A> quantiles = new Vector<>();
        public A min;
        public A max;
        public A median;
        public A sum;

        public double mean = 0;
        public double SD = 0;

        public double getZScore(A score)
        {
            return NumUtils.getZScore(score, this);
        }

        public String toString()
        {
            return String.format("%d elems min: %s max. %s sum: %s median: %s mean: %2.3f SD: %2.3f", num_elems, min, max, sum, median, mean, SD);
        }
    }

    
    public static double getZScore(double score, double mean, double SD)
    {
        return (score-mean) / SD;
    }

    public static<A extends Number> double getZScore(A score, BasicNumberInfo info)
    {
        return getZScore(score.doubleValue(), info.mean, info.SD);
    }

    
    public static<A extends Number, B extends Object> Vector<B> sort(Vector<B> v, NumberGetter<B,A> ng)
    {
        return sort(v, ng, false);
    }

    public static <A extends Number, B extends Object> boolean isSorted(Vector<B> v, NumberGetter<B,A> ng, boolean reverse)
    {
        if (v.size() == 0)
            return true;
        
        double last = ng.get(v.get(0)).doubleValue();

        for (int i = 1; i < v.size(); i++)
        {
            double n = ng.get(v.get(i)).doubleValue();
            int cv = compareTo(last, n);

            last = n;

            if (cv == 0 || cv > 0 == reverse)
                continue;

            return false;
        }
        return true;
    }

    public static<A extends Number, B extends Object> Vector<B> sort(Vector<B> v, NumberGetter<B,A> ng, boolean reverse)
    {
        Collections.sort(v, (B a1, B a2) -> compareTo(ng.get(a1).doubleValue(), ng.get(a2).doubleValue()));
        if (reverse)
        {
            Collections.reverse(v);
        }
        return v;
    }
    
    public static<A extends Number, B extends Object> Vector<A> getSorted(Iterator<B> it, NumberGetter<B,A> ng, boolean reverse)
    {
        Vector<A> v = ObjectGetter.map(it, (B b) -> ng.get(b));
        
        Collections.sort(v, (A a1, A a2) -> compareTo(a1.doubleValue(), a2.doubleValue()));
        if (reverse)
        {
            Collections.reverse(v);
        }
        return v;
    }

    public static BasicNumberInfo<Integer> getNumInfo(int[] a)
    {
        return getNumInfo(a, 0, a.length);
    }

    public static BasicNumberInfo<Integer> getNumInfo(int[] a, int start, int end)
    {
        return getNumInfo(IteratorUtils.getIntIterator(start, end), (Integer idx) -> a[idx]);
    }

    public static BasicNumberInfo<Double> getNumInfo(double[] a)
    {
        return getNumInfo(a, 0, a.length);
    }
    public static BasicNumberInfo<Double> getNumInfo(double[] a, int start, int end)
    {
        return getNumInfo(IteratorUtils.getIntIterator(start, end), (Integer idx) -> a[idx]);
    }

    public static BasicNumberInfo<Float> getNumInfo(float[] a)
    {
        return getNumInfo(CollectionIterator.getIterator(a));
    }

    public static<A extends Number, B extends Object> BasicNumberInfo<A> getNumInfo(Iterator<B> it, NumberGetter<B,A> ng)
    {
        return getNumInfo(it, false, ng);
    }


    public static<A extends Number> BasicNumberInfo<A> getNumInfo(Map<A, Integer> histogram)
    {
         BasicNumberInfo<A> rv = new BasicNumberInfo<>();

         Vector<A> sortedkeys = ObjectGetter.toVector(histogram.keySet());
         sort(sortedkeys, (_a) -> _a.doubleValue());

         rv.num_elems = sum(histogram.values(), 0);

         if (rv.num_elems == 0)
             return rv;
         
         

         

         rv.mean = 0.0;
         double sum = 0.0;
         rv.min = sortedkeys.get(0);
         rv.max = sortedkeys.get(sortedkeys.size() -1);
         rv.quantiles = ObjectGetter.mapIndex(quantiles.length, (_q) -> null);

         Vector<Integer> checkpos = ObjectGetter.mapIndex(quantiles.length, (_q) -> (int) (quantiles[_q] * rv.num_elems));
         A last = null;
         int pos = 0;
         for (A k : sortedkeys)
         {
             last = k;
             int count = histogram.get(k);
             int nextpos = pos + count;

             for(int i=0; i<checkpos.size(); i++)
             {                                            
                 if (pos > checkpos.get(i) || nextpos < checkpos.get(i))
                     continue;
                 
                 rv.quantiles.set(i, k);
             }
             sum += k.doubleValue() * count;
             pos = nextpos;
         }
         
         rv.median = rv.quantiles.get(2);
         rv.sum = (A)convertNumber(sum, last.getClass());
         
         

         rv.mean = sum  / rv.num_elems;

         if (rv.num_elems == 1)
             return rv;
         

         for (A k : sortedkeys)
         {
             double diff = k.doubleValue() - rv.mean;
             double ddiff = diff * diff;
             int count = histogram.get(k);

             rv.SD += (count * ddiff);
         }
         
         rv.SD = Math.sqrt(rv.SD/(rv.num_elems-1));
         
         return rv;
    }

    public static<A extends Number, B extends Object> BasicNumberInfo<A> getNumInfo(Iterator<B> it, boolean nosort, NumberGetter<B,A> ng)
    {
        BasicNumberInfo<A> rv = new BasicNumberInfo<>();

        Vector<Pair<Double, A>> data = new Vector<>();
        
        
        rv.mean = 0.0;
        A last = null;
        
        while (it.hasNext())
        {
            A a = ng.get(it.next());
            if(a==null)
            	continue;
            double dv = a.doubleValue();
            data.add(new Pair<Double, A>(dv,a));
            rv.mean += dv;
            last = a;
        }
        
        if(data.size()==0)
        	return null;

        if (last != null)
        {
            rv.sum = (A)convertNumber(rv.mean, last.getClass());
        }
        
        if (!nosort)
        {
            Collections.sort(data);
        }
        

        rv.num_elems = data.size();
        if (data.size() ==0 )
            return rv;
        
        
        rv.min = data.get(0).getSecond();
        rv.max = data.get(data.size()-1).getSecond();
        final int N = data.size()-1;
        rv.median = data.get((int)(0.5*N)).getSecond();

        for (int i=0; i<quantiles.length; i++)
        {
            rv.quantiles.add(data.get((int)(quantiles[i]*N)).getSecond());
        }

        rv.mean /= rv.num_elems;

        if (data.size() == 1)
            return rv; //cannot compute SD

        for (Pair<Double, A> a : data)
        {
            double diff = a.getFirst()-rv.mean;
            rv.SD += (diff * diff);
        }

        rv.SD = Math.sqrt(rv.SD/(rv.num_elems-1));
                                
        return rv;
    }

    


    public static<A extends Number, B extends Object> BasicNumberInfo<A> getNumInfo(Collection<B> c, boolean nosort, NumberGetter<B,A> ng)
    {
        return getNumInfo(c.iterator(), nosort, ng);
    }
    public static<A extends Number, B extends Object> BasicNumberInfo<A> getNumInfo(Collection<B> c, NumberGetter<B,A> ng)
    {
        return getNumInfo(c.iterator(), false, ng);
    }
    
    public static<A extends Number, B extends Object> BasicNumberInfo<A> getNumInfo(B[] array, NumberGetter<B,A> ng)
    {
        return getNumInfo(new CollectionIterator<B>(array), false, ng);
    }

    public static<A extends Number, B extends Object> BasicNumberInfo<A> getNumInfo(B[] array, boolean nosort, NumberGetter<B,A> ng)
    {
        return getNumInfo(new CollectionIterator<B>(array), nosort, ng);
    }


    public static<A extends Number> BasicNumberInfo<A> getNumInfo(Collection<A> c, boolean nosort)
    {
        return getNumInfo(c.iterator(), nosort);
    }
    public static<A extends Number> BasicNumberInfo<A> getNumInfo(Collection<A> c)
    {
        return getNumInfo(c.iterator(), false);
    }

    public static<A extends Number> BasicNumberInfo<A> getNumInfo(A[] a, boolean nosort)
    {
        return getNumInfo(new CollectionIterator<A>(a), nosort);
    }

    public static<A extends Number> BasicNumberInfo<A> getNumInfo(A[] a)
    {
        return getNumInfo(new CollectionIterator<A>(a), false);
    }


    public static<A extends Number> BasicNumberInfo<A> getNumInfo(Iterator<A> it)
    {
        return getNumInfo(it, false);
    }
    public static<A extends Number> BasicNumberInfo<A> getNumInfo(Iterator<A> it, boolean nosort)
    {
        return getNumInfo(it, nosort, (A o) -> o);
    }

    public static int toInt(Object o)
    {
        return (Integer)convertNumber(o, Integer.class);
    }

    public static Vector<Integer> toInts(Collection o)
    {
        return convertNumbers(o.iterator(), 0);
    }

    public static Vector<Float> toFloats(Collection o)
    {
        return convertNumbers(o.iterator(), 0.0f);
    }

    public static double[] toDoubleArray(int[] data)
    {
        double[] dv = new double[data.length];
        for (int i=0; i<data.length; dv[i] = (double)data[i++]);
        return dv;
    }

    public static double[] toDoubleArray(Double[] data)
    {
        double[] dv = new double[data.length];
        for (int i=0; i<data.length; dv[i] = data[i++]);
        return dv;
    }

    public static double[] toDoubleArray(Vector<Double> data)
    {
        double[] dv = new double[data.size()];
        for (int i=0; i<data.size(); dv[i] = data.get(i++));
        return dv;
    }

    static public<A>  double[] toDoubleArray(Vector<A> data, NumberGetter<A, Double> ng)
    {
        double[] dv = new double[data.size()];
        for (int i=0; i<data.size(); dv[i] = ng.get(data.get(i++)));
        return dv;
    }

    public static Vector<Double> toDoubles(int[] data)
    {
        return ObjectGetter.map(IteratorUtils.getIntIterator(0, data.length), (Integer idx) -> (double)data[idx]);
    }
    public static Vector<Double> toDoubles(Collection o, boolean filter_nulls)
    {
        return convertNumbers(o.iterator(), 0.0, filter_nulls);
    }
    public static Vector<Double> toDoubles(Collection o)
    {
        return toDoubles(o, false);
    }

    public static Vector<Integer> toInts(Object[] o)
    {
        return convertNumbers(new CollectionIterator(o), 0);
    }

    public static Vector<Float> toFloats(Object[] o)
    {
        return convertNumbers(new CollectionIterator(o), 0.0f);
    }

    
    public static Vector<Double> toDoubles(Object[] o)
    {
        return convertNumbers(new CollectionIterator(o), 0.0);
    }

    public static Vector<Integer> toInts(Iterator it)
    {
        return convertNumbers(it, 0);
    }

    public static Vector<Float> toFloats(Iterator it)
    {
        return convertNumbers(it, 0.0f);
    }

    public static Vector<Double> toDoubles(Iterator it)
    {
        return convertNumbers(it, 0.0);
    }

    public static<A extends Number> Vector<A> convertNumbers(Iterator it, A value)
    {
        return convertNumbers(it, value, false);
    }

    public static<A extends Number> Vector<A> convertNumbers(Iterator it, A value, boolean filter_nulls)
    {

        Vector<A> rv = new Vector<>();

        StringConverter sc = SimpleTypeConverter.getConverter(value.getClass(), false);

        while (it.hasNext())
        {
            Object o = it.next();
            if(o==null)
            {
                if (!filter_nulls)
                {
                    rv.add(null);
                }
                
                continue;
            }
            if (o.getClass() == String.class)
            {
                rv.add((A)sc.convert((String)o));
                continue;
            }
            rv.add((A)convertNumber(o, value.getClass()));
            
        }
        return rv;
    }
    
    
    public static Object convertNumber(Object o, Class convert)
    {
        if(o==null)
            return o;

        Class cz = o.getClass();
        if (cz == convert)
            return o;
        
        if (o instanceof String)
        {
            return SimpleTypeConverter.getConverter(convert, false).convert((String)o);
        }
        if (!(o instanceof Number))
        {
            throw new RuntimeException("cannot convert "+o.getClass()+" to number!");
        }
        Number n = (Number)o;

        if(convert == Double.class)
        {
            return n.doubleValue();
        }

        if(convert == Float.class) 
        {
            return n.floatValue();
        }

        if(convert == Long.class) 
        {
            return n.longValue();
        }
        if (convert == Integer.class)
        {
            return n.intValue();
        }
        return o;
    }

    public interface NumberGetter<A extends Object, B extends Number>
    {
        public B get(A a);
    }

    public static<B extends Number> B sum(Collection<B> c)
    {
        return sum(c, (B b) -> b, null);
    }

    public static<B extends Number> B sum(Collection<B> c, B defval)
    {
        return sum(c, (B b) -> b, defval);
    }

    public static<B extends Number> B sum(Iterator<B> c)
    {
        return sum(c, (B b) -> b);
    }

    public static<B extends Number> B sum(B[] c)
    {
        return sum(c, (B b) -> b);
    }

    public static<A extends Object, B extends Number> B sum(Collection<A> c, NumberGetter<A,B> ng)
    {
        return sum(c.iterator(), ng);
    }

    public static<A extends Object, B extends Number> B sum(Collection<A> c, NumberGetter<A,B> ng, B defval)
    {
        return sum(c.iterator(), ng, defval);
    }

    public static<A extends Object, B extends Number> B sum(A[] c, NumberGetter<A,B> ng)
    {
        return sum(new CollectionIterator<A>(c), ng);
    }

    public static<A extends Object, B extends Number> B sum(Iterator<A> it, NumberGetter<A,B> ng, B defval)
    {
        B v = sum(it, ng);
        return (v != null) ? v : defval;
    }

    public static<A extends Object, B extends Number> B sum(Iterator<A> it, NumberGetter<A,B> ng)
    {
        double sum = 0.0;
        B converter = null;

        boolean allNull = true;
        while (it.hasNext())
        {
            A o = it.next();
            if(o==null)
            	continue;
            
            B v = ng.get(o);
            if (v == null)
                continue;
            
            
            converter = v;
            allNull = false;
            sum += v.doubleValue(); //SimpleTypeConverter.doubleValue(ng.get(o));
        }

        if(allNull)
            return null;
        
        return (B) convertNumber(sum, converter.getClass());
    }

    public static<B extends Number> B mean(Collection<B> c)
    {
        return mean(c, (B b) -> b);
    }
    public static<A extends Object, B extends Number> B mean(Collection<A> c, NumberGetter<A,B> ng)
    {
        return mean(c.iterator(), ng);
    }
    
    public static<A extends Object, B extends Number> B mean(Iterator<A> it, NumberGetter<A,B> ng)
    {
        double sum = 0.0;
        int notNull = 0;
        B v = null;

        while (it.hasNext())
        {
            A o = it.next();
            if(o==null)
                continue;
            
            v = ng.get(o);
            if (v == null)
                continue;
            notNull++;
            
            sum += v.doubleValue(); //SimpleTypeConverter.doubleValue(ng.get(o));
        }
        if(notNull==0)
            return null;
        
        return (B) convertNumber(sum/notNull, v.getClass());
    }
    
    public static<A extends Object, B extends Number> B mean(Iterator<A> it, NumberGetter<A,B> ng, B defval)
    {
        B v = mean(it, ng);
        return (v != null) ? v : defval;
    }
    public static<A extends Object, B extends Number> B mean(Collection<A> it, NumberGetter<A,B> ng, B defval)
    {
        B v = mean(it, ng);
        return (v != null) ? v : defval;
    }

    public static int maxInt(int startx, int endx, int ... c)
    {
        return max(IteratorUtils.getIntIterator(startx, endx), (Integer idx) -> c[idx]);
    }

    public static int maxInt(int ... c)
    {
        return max(CollectionIterator.getIterator(c), (Integer n) -> n);
    }

    public static int minInt(int ... c)
    {
        return minObj(CollectionIterator.getIterator(c), (Integer n) -> n).getFirst();
    }

    public static int minInt(int startx, int endx, int ... c)
    {
        return min(IteratorUtils.getIntIterator(startx, endx), (Integer idx) -> c[idx]);
    }

    public static double maxDouble(int startx, int endx, double ... c)
    {
        return max(IteratorUtils.getIntIterator(startx, endx), (Integer idx) -> c[idx]);
    }

    public static double minDouble(int startx, int endx, double ... c)
    {
        return min(IteratorUtils.getIntIterator(startx, endx), (Integer idx) -> c[idx]);
    }
    public static double maxDouble(double ... c)
    {
        return max(CollectionIterator.getIterator(c), (Double n) -> n);
    }

    public static double minDouble(double ... c)
    {
        return min(CollectionIterator.getIterator(c), (Double n) -> n);
    }


    public static<B extends Number>  B max(B ... c)
    {
        return max(new CollectionIterator<B>(c), (B b) -> b);
    }

    public static<B extends Number>  B min(B ... c)
    {
        return min(new CollectionIterator<B>(c), (B b) -> b);
    }

    

    public static<B extends Number>  B max(Collection<B> c)
    {
        return max((c == null) ? null : c.iterator(), (B b) -> b);
    }
    
    /**
     * returns the value with the maximal absolute value, but with the original sign
     */
    public static<A extends Object, B extends Number>  B maxAbs(Iterator<A> it, NumberGetter<A,B> ng)
    {
        B max = null;
        if (it == null)
            return max;
        
        while (it.hasNext())
        {
            A o = it.next();
        
            //Double v = SimpleTypeConverter.doubleValue(ng.get(o));
            B v = ng.get(o);
            if (v == null)
                continue;
            
            
            if (max!=null && Math.abs(max.doubleValue()) >= Math.abs(v.doubleValue()))
                continue;
            
            max = v;
        }
        return max;
    }
    
    public static<A extends Object, B extends Number>  B max(Collection<A> c, NumberGetter<A,B> ng)
    {
        return max(c, ng, null);
    }

    public static<A extends Object, B extends Number>  B max(Collection<A> c, NumberGetter<A,B> ng, B def)
    {
        B rv = max(c.iterator(), ng);
        return (rv == null) ? def : rv;
    }
    
    public static<A extends Object, B extends Number>  B max(A[] a, NumberGetter<A,B> ng)
    {
        return maxObj(new CollectionIterator<A>(a), ng).getFirst();
    }

    public static<A extends Object, B extends Number>  Pair<B,A> maxObj(Collection<A> c, NumberGetter<A,B> ng)
    {
        return maxObj(c.iterator(), ng);
    }

    public static<A extends Object, B extends Number>  Pair<B,A> minObj(Collection<A> c, NumberGetter<A,B> ng)
    {
        return minObj(c.iterator(), ng);
    }

    public static<A extends Object, B extends Number>  B min(Iterator<A> it, NumberGetter<A,B> ng)
    {
        return min(it, ng, null);
    }

    public static<A extends Object, B extends Number>  B min(Iterator<A> it, NumberGetter<A,B> ng, B def)
    {
        B rv = minObj(it,ng).getFirst();
        return (rv == null) ? def : rv;
    }

    public static<A extends Object, B extends Number>  Pair<B,A> maxObj(Iterator<A> it, NumberGetter<A,B> ng)
    {
        B max = null;
        A maxObj = null;
        if (it == null)
            return Pair.create(max, maxObj);
        
        while (it.hasNext())
        {
            A o = it.next();
        
            //Double v = SimpleTypeConverter.doubleValue(ng.get(o));
            B v = ng.get(o);
            if (v == null)
                continue;
            
            
            if (Double.isNaN(v.doubleValue()))
                continue;
            
            
            
            if (max!=null && max.doubleValue() >= v.doubleValue())
                continue;
            
            max = v;
            maxObj = o;
        }
        return Pair.create(max, maxObj);
    }

    public static<A extends Object, B extends Number>  B max(Iterator<A> it, NumberGetter<A,B> ng, B def)
    {
        B rv = maxObj(it, ng).getFirst();
        return (rv == null) ? def : rv;
    }
    public static<A extends Object, B extends Number>  B max(Iterator<A> it, NumberGetter<A,B> ng)
    {
        return max(it, ng, null);
    }


    public static<A extends Number>  A min(Collection<A> c)
    {
        return min(c, (A a) -> a);
    }

    

    public static<A extends Object, B extends Number>  B min(Collection<A> c, NumberGetter<A,B> ng, B def)
    {
        B rv = min(c.iterator(), ng);
        return (rv == null) ? def : rv;
    }

    public static<A extends Object, B extends Number>  B min(Collection<A> c, NumberGetter<A,B> ng)
    {
        return min(c, ng, null);
    }
    
    public static<A extends Object, B extends Number>  B min(A[] a, NumberGetter<A,B> ng)
    {
        return minObj(new CollectionIterator<A>(a), ng).getFirst();
    }

    public static<A extends Object, B extends Number>  Pair<B, A> minObj(Iterator<A> it, NumberGetter<A,B> ng)
    {
        B min = null;
        A obj = null;

        while (it.hasNext())
        {
            A o = it.next();
            B v = ng.get(o);
            if (v == null)
                continue;
            
            
            if (min!=null && min.doubleValue() <= v.doubleValue())
                continue;
            
            min = v; 
            obj = o;
        }
        return Pair.create(min, obj);
    }
    
    



    public static<A extends Number> double getSquaredDistance(A[] co1, A[] co2)
    {
        double rv=0;
        final int L = Math.min(co1.length, co2.length);

        for (int i=0; i<L; rv+=Math.pow(co1[i].doubleValue()-co2[i].doubleValue(), 2), i++);

        return rv;
    }

    /** calculates the eucledian distance between two 3dim points */
    public static<A extends Number> double getDistance(A[] co1, A[] co2)
    {
        return Math.sqrt(getSquaredDistance(co1, co2));
    }
    
    public static double getSquaredDistance(float[] co1, float[] co2)
    {
        double rv=0;
        final int L = Math.min(co1.length, co2.length);

        for (int i=0; i<L; rv+=Math.pow((double)co1[i]-(double)co2[i], 2), i++);

        return rv;
    }

    public static double getDistance(float[] co1, float[] co2)
    {
    
        return Math.sqrt(getSquaredDistance(co1, co2));
    }


    static double[] convertToSimple(Collection<Double> dv)
    {
        double[] rv = new double[dv.size()];
        int idx=0;
        for (Double d : dv)
        {
            rv[idx++]=d;
        }
        return rv;
    }
    
    public static<A extends Number, B extends Number> double uncentered_corr(Vector<A> v1, Vector<B> v2)
    {
        double mult = 0.0;
        double i2 = 0.0;
        double j2 = 0.0;

        for (int i=0; i<v1.size(); i++)
        {
            mult += (v1.get(i).doubleValue() * v2.get(i).doubleValue());
            i2 += Math.pow(v1.get(i).doubleValue(), 2);
            j2 += Math.pow(v2.get(i).doubleValue(), 2);
        }

        double divident = Math.sqrt(i2 * j2);
        return mult / divident;
    }

    public static<A extends Number, B extends Number> double corr(Collection<A> ita, Collection<B> itb)
    {
        return corr(ita, itb, null);
    }
    public static<A extends Number, B extends Number> double corr(Collection<A> ita, Collection<B> itb, Integer minNotNA)
    {
        return corr(ita.iterator(), itb.iterator(),minNotNA);
    }
    
    public static<A extends Number, B extends Number> double corr(Iterator<A> ita, Iterator<B> itb)
    {
        return corr(ita,itb, null);
    }
    /**
     * @param minNotNA minimum number of valid numbers (not null, NaN or Infinity) in ita and itb, else exception is thrown
     * @return
     */
    public static<A extends Number, B extends Number> double corr(Iterator<A> ita, Iterator<B> itb, Integer minNotNA)
    {
        Vector<Double> v1 = ObjectGetter.map(ita, (A a) -> a.doubleValue());
        Vector<Double> v2 = ObjectGetter.map(itb, (B b) -> b.doubleValue());

        if (v1.size() != v2.size())
            throw new FRuntimeException("invalid inputs for correlation lengths (%d, %d) differ", v1.size(), v2.size());
        
        if(minNotNA==null)
        {
            minNotNA=v1.size();
        }
            
        Vector<Double> v1Clean = new Vector<>();
        Vector<Double> v2Clean = new Vector<>();
        ObjectGetter.apply(IteratorUtils.getIntIterator(v1.size()),(Integer a)->{
           Double d1 = v1.get(a);
           Double d2 = v2.get(a);
           if(d1!=null && !d1.isInfinite() &&!d1.isNaN() && d2!=null && !d2.isInfinite() && !d2.isNaN())
           {
               v1Clean.add(d1);
               v2Clean.add(d2);
           }
        });
        
        if(v1Clean.size()<minNotNA)
            ExceptionManager.throwRuntimeException("too many NA (null, NaN, Infinity) values");

        double[] dv1 = ObjectGetter.toDoubleArray(v1Clean);
        double[] dv2 = ObjectGetter.toDoubleArray(v2Clean);
        PearsonsCorrelation pc = new PearsonsCorrelation();
        return pc.correlation(dv1, dv2);
    }



    public static<A extends Number,B extends Number> double getTTestPvalue(A[] ca, B[] cb)
    {
        return getTTestPvalue(new CollectionIterator<A>(ca), new CollectionIterator<B>(cb));
    }

    public static<A extends Number,B extends Number> double getTTestPvalue(Collection<A> ca, Collection<B> cb)
    {
        return getTTestPvalue(ca.iterator(), cb.iterator());
    }

    public static<A extends Number,B extends Number> double getTTestPvalue(Iterator<A> ita, Iterator<B> itb)
    {
        return getTTestValue(ita, itb, true);
    }


    public static<A extends Number,B extends Number> double getTTestStat(A[] ca, B[] cb)
    {
        return getTTestStat(new CollectionIterator<A>(ca), new CollectionIterator<B>(cb));
    }

    public static<A extends Number,B extends Number> double getTTestStat(Collection<A> ca, Collection<B> cb)
    {
        return getTTestStat(ca.iterator(), cb.iterator());
    }

    static<A extends Number,B extends Number> double getTTestValue(Iterator<A> ita, Iterator<B> itb, boolean getpval)
    {
        double[] v1 = convertToSimple(ObjectGetter.map(ita, (A a) -> a.doubleValue()));
        double[] v2 = convertToSimple(ObjectGetter.map(itb, (B b) -> b.doubleValue()));
        if (v1.length < 2  || v2.length < 2)
        {
            return Double.NaN;
        }
        return (getpval) ? TTester.homoscedasticTTest(v1, v2) : TTester.homoscedasticT(v1, v2);
    }

    public static<A extends Number,B extends Number> double getTTestStat(Iterator<A> ita, Iterator<B> itb)
    {
        return getTTestValue(ita, itb, false);
    }
    
    static HashMap<Double, Double> v2lv = new HashMap<>();

    public static synchronized double logN(double value, double base)
    {
        Double lv = v2lv.get(base);
        if (lv ==null)
        {
            v2lv.put(base, lv = Math.log(base));
        }
        return Math.log(value)/lv;
    }


    static public<A extends Number>  Pair<Vector<A>, Vector<Integer>> calcBins(Vector<A> data, double tolerance, boolean sort_first)
    {
        if (sort_first)
        {
            Collections.sort(data, (A a1, A a2) -> compareTo(a1.doubleValue(), a2.doubleValue()));
        }
        Vector<A> rv1 = new Vector<>();
        Vector<Integer> counts = new Vector<>();

        A last = null;
        int count = 0;
        for (A a : data)
        {
            double v = a.doubleValue();
            if (last == null)
            {
                last = a;
                count=1;
                continue;
            }
            if (last.doubleValue()+tolerance > v)
            {
                count++;
                continue;
            }

            rv1.add(last);
            counts.add(count);
                
            last = a;
            count = 1;
        }
        if (last != null && count > 0)
        {
            rv1.add(last);
            counts.add(count);

        }
        return Pair.create(rv1, counts);
    }

 

    public static int compareTo(double d1, double d2)
    {
        double diff = d1 - d2;
        return (diff == 0.0) ? 0 : (diff > 0) ? 1 :  -1;
    }
    
       
    static void testt(int hit_start_field, int hit_length_field, int  hit_sentid_field)
    {
        final int checksize = 1 + NumUtils.max(hit_start_field, hit_length_field, hit_sentid_field);
        
    }

    public static int[] init_intarray(int length, int value)
    {
        int[] rv = new int[length];
        for (int i=0; i<length; rv[i++] = value);
        return rv;
    }



    public static double[] scale(int[] array, double scale)
    {
        double rv[] = new double[array.length];
        for (int i=0; i<array.length; rv[i] = (scale * array[i++]));
        return rv;
    }
    
    public static double[] scale(double[] array, double scale)
    {
        double rv[] = new double[array.length];
        for (int i=0; i<array.length; rv[i] = (scale * array[i++]));
        return rv;
    }

    static public<A extends Number>  void initVector(Vector<A> v, A value)
    {
        for (int i=0; i<v.size(); v.set(i++, value));
    }

    static public<A extends Number> A getDefault(A value, A def)
    {
        return (value == null) ? def : value;
    }

    static public<A extends Number> double getDoubleDefault(A value, double def)
    {
        return (value == null) ? def : value.doubleValue();
    }

    public static void main (String args[]) 
    {
        int a1 = 13;
        int a2 = 14;
        
        int v = 1 + max(a1, a2, 14);
        
        try
        {
            System.out.println(getNumInfo(args, (String s) -> Integer.parseInt(s)));
        }
        catch(NumberFormatException nfe)
        {
            System.err.println("int test failed");
        }
        try
        {
            
            System.out.println(getNumInfo(args, (String s) -> Float.parseFloat(s)));
        }
        catch(NumberFormatException nfe)
        {
            System.err.println("float test failed");
        }

        try
        {
            System.out.println(getNumInfo(args, (String s) -> Double.parseDouble(s)));
        }
        catch(NumberFormatException nfe)
        {
            System.err.println("double test failed");
        }
        
        
        System.out.println("string convert test:");
        System.out.println(getNumInfo(toDoubles(args)));

        System.out.printf("sorted: %s\n", isSorted(toDoubles(args), (Double d) -> d, false));

        System.out.println(getNumInfo(toDoubles(new Float[]{1.0f, 2f, 123f, -123f})));
    }
}
