package lmu.utils;

import java.util.*;

public class ObjectGetter<A>
{

    int idx=0;
    
    public interface Handler<A>
    {
        public void handle(A a);
    }


    public static class TransformIterator<A,B> implements Iterator<B>
    {
        Iterator<A> it;
        MapGetter<A,B> m;

        public TransformIterator(Iterator<A> it, MapGetter<A,B> m)
        {
            this.it = it;
            this.m = m;
        }

        public B next()
        {
            return m.get(it.next());
        }

        public boolean hasNext()
        {
            return it.hasNext();
        }
        public void remove()
        {
            it.remove();
        }
    }

        
    public static Vector<Integer> toInt(String[] sp)
    {
        return toInt(sp, 0, sp.length);
    }

    public static Vector<Integer> toInt(String[] sp, int start)
    {
        return toInt(sp,start, sp.length);
    }
    
    public static Vector<Integer> toInt(String[] sp, int start, int end)
    {
        return map(IteratorUtils.getIntIterator(start, end), (Integer idx) -> Integer.parseInt(sp[idx]));
    }
    static public<A> StringBuffer getLinedBuffer(A[] c)
    {
        return getLinedBuffer(new CollectionIterator<A>(c));
    }

    static public<A> StringBuffer getLinedBuffer(Collection<A> c)
    {
        return getLinedBuffer(c.iterator());
    }
    
    static public<A> StringBuffer getLinedBuffer(Iterator<A> it)
    {
        StringBuffer sb = new StringBuffer();
        while(it.hasNext())
        {
            sb.append("\t"+it.next()+"\n");
        }
        return sb;
    }
    public static<A, B> Iterator<B> getTransformedIterator(Iterator<A> it, MapGetter<A,B> m)
    {
        return new TransformIterator<A,B>(it, m);
    }

    static public int getSize(Collection c)
    {
        return (c==null) ? -1 : c.size();
    }
    static public<A> Collection<A> getNotNullCollection(Collection<A> c)
    {
        if(c != null)
            return c;
        return new Vector<A>(0);
    }

    public static HashSet getSet(Collection c, String expr)
    {
        HashSet rv = new HashSet();
        ReflectedObject ro = new ReflectedObject(expr);
        for (Object o : c)
        {
            Object so = ro.get(o);
            if (so==null)
            {
                continue;
            }
            rv.add(so);
        }

        return rv;

    }

    
    public static HashMap getMultiLookup(Collection c, ReflectedObject ro, ReflectedObject transformer)
    {
        HashMap rv = new HashMap();
        //ReflectedObject ro = new ReflectedObject(expr);
        //ReflectedObject transformer = (transform_expr==null)?null:new ReflectedObject(transform_expr);
        
        for (Object o : c)
        {
            Object so = ro.get(o);
            if (so==null)
            {
                continue;
            }
            Vector v = (Vector)rv.get(so);
            if (v==null)
            {
                rv.put(so, v= new Vector());
            }
            v.add((transformer==null)?o:transformer.get(o));
            
            
        }
        return rv;
    }
    public static<A, B, C>  A getMapInfo(Map<B, C> m, B key, MapGetter<C,A> mapper)
    {
        C c = m.get(key);
        if (c == null)
            return null;
        
        return mapper.get(c);
    }

    static public<A,B,C,D> HashMap<C,D> transformMap(Map<A,B> m, MapGetter<A,C> keyMapper, MapGetter<B,D> valueMapper)
    {
        HashMap<C,D> hm = new HashMap<>();
        for (Map.Entry<A,B> e: m.entrySet())
        {
            hm.put(keyMapper.get(e.getKey()), valueMapper.get(e.getValue()));
        }
        return hm;
    }

    public static<A,B> HashMap<A,B> getLookup(Collection<A> c, MapGetter<A,B> mapper)
    {
        HashMap<A,B> rv = new HashMap<>();
        for (A o : c)
        {
            B b = mapper.get(o);
            if (b==null)
                continue;
            rv.put(o,b);
            
        }
        return rv;
    }
    /** creates a lookup by expression */
    public static HashMap getLookup(Collection c, String expr)
    {
        HashMap rv = new HashMap();
        ReflectedObject ro = new ReflectedObject(expr);
        for (Object o : c)
        {
            Object so = ro.get(o);
            if (so==null)
            {
                continue;
            }
            rv.put(so,o);
            
        }
        return rv;
    }

    public static HashMap getMultiLookup(Collection c, String expr)
    {
        return getMultiLookup(c,new ReflectedObject(expr),(ReflectedObject)null);
    }
    /** creates a multi-value lookup by expression - ie the values in hashmap 
     * are vectors containing the objects for the same key
     */
    public static HashMap getMultiLookup(Collection c, String expr, String transform_expr)
    {
        return getMultiLookup(c,new ReflectedObject(expr), (transform_expr==null)?(ReflectedObject)null:new ReflectedObject(transform_expr));
        
    }
    public static<A> ObjectGetterHeader<A> getHeader(Vector<ObjectGetter<A>> v)
    {
        for (ObjectGetter<A> og : v)
        {
            if (og==null || og.header==null)
                continue;
            
            return og.header;
        }
        return null;
    }
    
    public interface Filterer<A>
    {
        public boolean filter(A o);

    }
    public interface MapGetter<A, B>
    {
        public B get(A o); 
    }
    
    public interface MultiMapGetter<A, B>
    {
        public Iterator<B> get(A o); 
    }

    public interface MultiCollectionGetter<A, B>
    {
        public Collection<B> get(A o); 
    }

    public interface Applyer<A>
    {
        public void apply(A o);
    }

        
    public static<A, B, C> Map<A,C> transformMapValues(Map<A, B> m, MapGetter<B, C> applyer)
    {
        HashMap<A,C> nm = new HashMap<>();

        for (Map.Entry<A,B> e : m.entrySet())
        {
            C c = applyer.get(e.getValue());
            nm.put(e.getKey(), c);
        }
        return nm;
    }
    
    public static<A, B> HashSet<A> collapse(Collection<B> c, MultiCollectionGetter<B,A> mc)
    {
        return collapse(c.iterator(), mc);
    }

    
    public static<A, B> HashSet<A> collapse(Iterator<B> c, MultiCollectionGetter<B,A> mc)
    {
        return collapse(ObjectGetter.map(c, (B b) -> mc.get(b)));
    }

    public static<A, B extends Collection<A>> HashSet<A> collapse(Collection<B> c)
    {
        return collapse(c.iterator());
    }
    public static<A, B extends Collection<A>> HashSet<A> collapse(Iterator<B> it)
    {
        HashSet<A> rv = new HashSet<>();
        while (it.hasNext())
        {                  
            rv.addAll(it.next());
        }
        return rv;
    }

    public static<A> void addAll(Collection<A> target, Iterator<A> it)
    {
        while (it.hasNext())
        {
            target.add(it.next());
        }
    }
    public static<A> void extend(Collection<A> c, Collection<A> toextendwith)
    {
        extend(c, toextendwith.iterator());
    }

    public static<A> void extend(Collection<A> c, A ... toextendwith)
    {
        extend(c,  new CollectionIterator<A>(toextendwith));
    }

    public static<A> void extend(Collection<A> c, Iterator<A> it)
    {
        while (it.hasNext())
        {
            c.add(it.next());
        }
        
    }
    public static<A > void apply(Applyer<A> applyer, A ... data)
    {
        apply(new CollectionIterator<A>(data), applyer);
    }
    
    public static<A > void apply(A[] array, Applyer<A> applyer)
    {
        apply(new CollectionIterator<A>(array), applyer);
    }


    public static<A > void apply(Iterator<A> it, Applyer<A> applyer)
    {
        if (it == null)
            return;
        
        
        while (it.hasNext())
        {
            A o = it.next();
            applyer.apply(o);
        }
    }

    public static<A > void apply(Collection<A> v, Applyer<A> applyer)
    {
        apply((v == null) ? null : v.iterator(), applyer);
    }
    
    public static<A, E> Vector<E> convert(Collection<A> c)
    {
        return map(c, (A a) -> (E)a);
    }

    public static<A, E> Vector<E> convert(Iterator<A> c)
    {
        return map(c, (A a) -> (E)a);
    }

    public static<A, E> Vector<E> convert(A[] c)
    {
        return map(c, (A a) -> (E)a);
    }

    public static<E , A > Vector<E> map(A[] v, MapGetter<A, E> mapper)
    {
        return map(new CollectionIterator<A>(v),mapper);
    }
    static public<A> MapGetter<A,A>  getIDMapper()
    {
        return (A a) -> a;
    }
    public static<E extends Comparable, A > Vector<E> mapSorted(A[] v, MapGetter<A, E> mapper, boolean asc)
    {
        return mapSorted((v==null)?(Iterator<A>)null:new CollectionIterator<A>(v), mapper, asc);
    }

    public static<E extends Comparable, A > Vector<E> mapSorted(Collection<A> v, MapGetter<A, E> mapper, boolean asc)
    {
        return mapSorted((v==null)?(Iterator<A>)null:v.iterator(), mapper,  asc);
    }

    public static HashSet<Character> toSet(char[] data)
    {
        HashSet<Character> rv = new HashSet<>();
        if (data == null)
            return rv;
        
        for (int i=0; i<data.length; i++)
        {
            rv.add(data[i]);
        }
        return rv;
    }
    public static HashSet<Integer> toSet(int[] data)
    {
        HashSet<Integer> rv = new HashSet<>();
        if (data == null)
            return rv;
        
        for (int i=0; i<data.length; i++)
        {
            rv.add(data[i]);
        }
        return rv;
    }
    public static<A> HashSet<A> toSet(Collection<A> data)
    {
        if (data instanceof HashSet)
        {
            return (HashSet<A>)data;
        }
        return toSet((data == null) ? null : data.iterator());
    }
    public static<A> HashSet<A> toSet(Iterator<A> dataIt)
    {
        HashSet<A> s = new HashSet<>();
        if (dataIt == null)
            return s;
        apply(dataIt, (A a) -> s.add(a));
        return s;
    }
    public static<A> HashSet<A> toSet(A ... data)
    {
        return toSet(new CollectionIterator<A>(data));
    }
    public static<A> Vector<A> toVector(A ... data)
    {
        return toVector(new CollectionIterator<A>(data));
    }

    public static<A extends Comparable> Vector<A> toSortedVector(Collection<A> c, boolean asc)
    {
        Vector<A> rv = toVector(c);
        Collections.sort(rv);
        if (asc)
            return rv;
        
        Collections.reverse(rv);
        return rv;
    }

    public static<B, A extends Comparable> Vector<B> toSortedVector(Collection<B> c, MapGetter<B,A> mg, boolean asc)
    {
        return toSortedVector(c.iterator(), mg, asc);
    }
    public static<B, A extends Comparable> Vector<B> toSortedVector(Iterator<B> c, MapGetter<B,A> mg, boolean asc)
    {
        Vector<B> rv = toVector(c);
        Collections.sort(rv, (c1, c2) -> mg.get(c1).compareTo(mg.get(c2)));
        if (asc)
            return rv;
        
        Collections.reverse(rv);
        return rv;
    }

    public static<A> Vector<A> toSortedVector(Collection<A> c, Comparator<A> comp, boolean asc)
    {
        Vector<A> rv = toVector(c);
        Collections.sort(rv, comp);
        if (asc)
            return rv;
        
        Collections.reverse(rv);
        return rv;
    }

    public static Vector<Character> toVector(char[] a)
    {
        Vector<Character> rv = new Vector<>();
        for (char ch : a)
        {
            rv.add(ch);
        }
        return rv;
    }
    public static Vector<Integer> toVector(int[] a)
    {
        Vector<Integer> v = new Vector<>();
        if (a == null)
            return v;
        
        for (int i : a)
        {
            v.add(i);
        }
        return v;
    }

    public static Vector<Double> toVector(double[] a)
    {
        Vector<Double> v = new Vector<>();
        if (a == null)
            return v;
        
        for (double i : a)
        {
            v.add(i);
        }
        return v;
    }

    public static<A> Vector<A> toVector(Collection<A> c)
    {
        return toVector((c==null) ? null : c.iterator());
    }
    
    public static <A extends Comparable> Vector<A> sort(Vector<A> v, boolean reverse)
    {
        Collections.sort(v);
        if (reverse)
        {
            Collections.reverse(v);
        }
        return v;
    }

    public static<A> Vector<A> toVector(Iterator<A> it)
    {
        Vector<A> rv = new Vector<>();
        if (it == null)
            return rv;
        
        while (it.hasNext())
        {
            rv.add(it.next());
        }
        return rv;
    }

    

    public static int[] toArray(Collection<Integer> data)
    {
        int[] rv = new int[data.size()];
        int idx = 0;
        Iterator<Integer> it = data.iterator();
        while (it.hasNext())
        {
            rv[idx++] = it.next();
        }
        return rv;
    }

    public static double[] toDoubleArray(Collection<Double> data)
    {
        double[] rv = new double[data.size()];
        int idx = 0;
        Iterator<Double> it = data.iterator();
        while (it.hasNext())
        {
            rv[idx++] = it.next();
        }
        return rv;
    }
    
    public static<A> A[] toArray(A ... data)
    {
        return (A[]) data;
    }

    
    public static<E, A> HashMap<A,HashSet<E>> reverseMapMulti(Map<E, A> m)
    {
        return buildReverseMultiMap(m.keySet(), (E k) -> m.get(k));
    }
        
    public static<E, A> HashMap<A,HashSet<E>> reverseMultiMap(Map<E, Set<A>> m)
    {
        HashMap<A,HashSet<E>> mm = new HashMap<>();
        for (Map.Entry<E, Set<A>> e : m.entrySet())
        {
            for (A a : e.getValue())
            {
                HashSet<E> hs = mm.get(a);
                if (hs == null)
                {
                    mm.put(a, hs = new HashSet<>());
                }
                hs.add(e.getKey());
            }
        }
        return mm;
    }

    public static<E, A> HashMap<A,HashSet<E>> reverseMultiMapH(Map<E, HashSet<A>> m)
    {
        HashMap<A,HashSet<E>> mm = new HashMap<>();
        for (Map.Entry<E, HashSet<A>> e : m.entrySet())
        {
            for (A a : e.getValue())
            {
                HashSet<E> hs = mm.get(a);
                if (hs == null)
                {
                    mm.put(a, hs = new HashSet<>());
                }
                hs.add(e.getKey());
            }
        }
        return mm;
    }

    public static<E, A> HashMap<A,HashSet<E>> buildMultiMap(MapGetter<A, E> mapper, A ... data)
    {
        return buildMultiMap(new CollectionIterator<A>(data), mapper);
    }

    public static<E, A> HashMap<A,HashSet<E>> buildMultiMap(Collection<A> c, MapGetter<A, E> mapper)
    {
        return buildMultiMap(c.iterator(), mapper);
    }

    
    public static<E, A> HashMap<E,HashSet<A>> buildReverseMultiMap(MapGetter<A, E> mapper, A ... data)
    {
        return buildReverseMultiMap(new CollectionIterator<A>(data), mapper);
    }

    public static<E, A> HashMap<E,HashSet<A>> buildReverseMultiMap(Collection<A> c, MapGetter<A, E> mapper)
    {
        return buildReverseMultiMap((c == null) ? null : c.iterator(), mapper);
    }


    public static<E, A> HashMap<E,HashSet<A>> buildReverseMultiMap(Iterator<A> it, MapGetter<A, E> mapper)
    {
        return buildReverseMultiMap(it, (A a) -> new SingleIterator<E>(mapper.get(a)));
    }


    public static<E, A> HashMap<E,HashSet<A>> buildReverseMultiMap(Collection<A> data, MultiCollectionGetter<A, E> mapper)
    {
        return buildReverseMultiMap(data.iterator(), mapper);
    }
    public static<E, A> HashMap<E,HashSet<A>> buildReverseMultiMap(Iterator<A> it, MultiCollectionGetter<A, E> mapper)
    {
        return buildReverseMultiMap(it, (A a) -> mapper.get(a).iterator());
    }

    public static<E, A> HashMap<E,HashSet<A>> buildReverseMultiMap(Iterator<A> it, MultiMapGetter<A, E> mapper)
    {
        HashMap<E,HashSet<A>> hm = new HashMap<>();
        if (it == null)
            return hm;
        
        
        while (it.hasNext())
        {
            A a = it.next();
            Iterator<E> ite = mapper.get(a);

            if (ite == null)
                continue;
            
            while (ite.hasNext())
            {
                E e = ite.next();
                if (e == null)
                    continue;
                
                
                
                HashSet<A> hs = hm.get(e);
                if (hs == null)
                {
                    hm.put(e, hs = new HashSet<>());
                }
                hs.add(a);
            }
        }
        return hm;
    }

    public static<A, K, V> HashMap<K,HashSet<V>> buildMultiMap(Collection<A> c, MapGetter<A, K> key_mapper, MapGetter<A, V> value_mapper)
    {
        return (c == null) ? null : buildMultiMap(c.iterator(), key_mapper, value_mapper);
    }
    public static<A, K, V> HashMap<K,HashSet<V>> buildMultiMap(Iterator<A> it, MapGetter<A, K> key_mapper, MapGetter<A, V> value_mapper)
    {
        HashMap<K,HashSet<V>> hm = new HashMap<>();

        while (it.hasNext())
        {
            A a = it.next();
            K key = key_mapper.get(a);
            HashSet<V> hs = hm.get(key);
            if (hs == null)
            {
                hm.put(key, hs = new HashSet<>());
            }
            hs.add(value_mapper.get(a));
        }
        return hm;
    }

    public static<E, A> HashMap<A,HashSet<E>> buildMultiMap(Iterator<A> it, MapGetter<A, E> mapper)
    {
        return buildMultiMap(it, (A a) -> a, mapper);
    }

    public static<K,V> HashMap<K,V> filterMapByValue(Map<K,V> m, Filterer<V> f)
    {

        HashMap<K,V> rv = new HashMap<>();
        for (Map.Entry<K,V> e: m.entrySet())
        {
            if (e.getValue() == null || !f.filter(e.getValue()))
                continue;
            
            rv.put(e.getKey(), e.getValue());
        }
        return rv;
    }
    public static<A, K, V> HashMap<K,V> buildMap(Collection<A> v, MapGetter<A, K> key_getter, MapGetter<A, V> value_getter)
    {
        return buildMap((v==null)?(Iterator<A>)null:v.iterator(), key_getter, value_getter);
    }

    public static<A, K, V> HashMap<K,V> buildMap(Iterator<A> v, MapGetter<A, K> key_getter, MapGetter<A, V> value_getter)
    {
        HashMap<K,V> rv = new HashMap<>();
        while (v.hasNext())
        {
            A o = v.next();
            if (o == null)
                continue;
            
            
            rv.put(key_getter.get(o), value_getter.get(o));

        }
        return rv;
    }

    public static<E , A > HashMap<A,E> buildMap(Collection<A> v, MapGetter<A, E> mapper)
    {
        return buildMap((v==null)?(Iterator<A>)null:v.iterator(), mapper);
    }

    public static<E , A > HashMap<A,E> buildMap(A[] v, MapGetter<A, E> mapper)
    {
        return buildMap((v==null)?(Iterator<A>)null:new CollectionIterator<A>(v), mapper);
    }

    public static<E , A > HashMap<A,E> buildMap(Iterator<A> it, MapGetter<A, E> mapper)
    {
        HashMap<A,E> hm = new HashMap<>();

        while (it.hasNext())
        {
            A a = it.next();
            hm.put(a, mapper.get(a));
        }
        return hm;
    }

    public static<E , A > HashMap<E,A> reverseMap(Map<A,E> m)
    {
        HashMap<E,A> hm = new HashMap<>();
        for (Map.Entry<A,E> e : m.entrySet())
        {
            hm.put(e.getValue(), e.getKey());
        }
        return hm;
    }

    public static<E , A > HashMap<E,A> buildReverseMap(Collection<A> v, MapGetter<A, E> mapper)
    {
        return buildReverseMap((v==null)?(Iterator<A>)null:v.iterator(), mapper);
    }

    public static<E , A > HashMap<E, A> buildReverseMap(A[] v, MapGetter<A, E> mapper)
    {
        return buildReverseMap((v==null)?(Iterator<A>)null:new CollectionIterator<A>(v), mapper);
    }

    public static<E , A > HashMap<E, A> buildReverseMap(Iterator<A> it, MapGetter<A, E> mapper)
    {
        HashMap<E, A> hm = new HashMap<>();

        if (it == null)
            return hm;
        
        
        while (it.hasNext())
        {
            A a = it.next();
            hm.put(mapper.get(a), a);
        }
        return hm;
    }

    public static<E , A > Vector<E> filter_and_map(Collection<A> v, Filterer<A> f, MapGetter<A, E> mapper)
    {
        return map(filter(v, f), mapper);
    }

    public static<E , A > Vector<E> filter_and_map(Collection<A> v, Filterer<A> f, MapGetter<A, E> mapper, Comparator<E> c, Boolean sort_dir)
    {
        return tosorted(map(filter(v, f), mapper), sort_dir, c);
    }

    public static<E , A > Vector<E> filter_and_map(Iterator<A> v, Filterer<A> f, MapGetter<A, E> mapper)
    {
        return map(filter(v, f), mapper);
    }

    public static<E , A > Vector<E> filter_and_map(Iterator<A> v, Filterer<A> f, MapGetter<A, E> mapper, Comparator<E> c, Boolean sort_dir)
    {
        return tosorted(map(filter(v, f), mapper), sort_dir, c);
    }

    public static<E , A > Vector<E> filter_and_map(A[] v, Filterer<A> f, MapGetter<A, E> mapper)
    {
        return filter_and_map(v, f, mapper, null, null);
    }

    public static<E , A > Vector<E> filter_and_map(A[] v, Filterer<A> f, MapGetter<A, E> mapper, Comparator<E> comp, Boolean sort_dir)
    {
        return tosorted(map(filter(v, f), mapper), sort_dir, comp);
    }

    static<E> Vector<E> tosorted(Vector<E> v, Boolean sort_dir, Comparator<E> c)
    {
        if (sort_dir == null && c == null)
            return v;

        if (sort_dir != null)
            return (Vector<E>)toSortedVector((Collection<Comparable>)v, sort_dir);

        return toSortedVector(v, c, (sort_dir == null) ? true : sort_dir);
    }

    public static<E , A > Vector<E> map(Collection<A> v, MapGetter<A, E> mapper)
    {
        return map((v==null)?(Iterator<A>)null:v.iterator(), mapper);
    }
    
        
    public static<E , A, C2 extends Collection<E>> C2 mapTo(Collection<A> a, MapGetter<A, E> mapper, C2 c2)
    {
        return mapTo((a == null) ? null : new CollectionIterator<A>(a), mapper,c2);
    }

    public static<E , A, C2 extends Collection<E>> C2 mapTo(A[] a, MapGetter<A, E> mapper, C2 c2)
    {
        return mapTo((a == null) ? null : new CollectionIterator<A>(a), mapper,c2);
    }

    public static<E , A, C2 extends Collection<E>> C2 mapTo(Iterator<A> it, MapGetter<A, E> mapper, C2 c2)
    {
        if (it == null)
            return c2;
        
        while (it.hasNext())
        {
            A o = it.next();
            c2.add(mapper.get(o));
        }
        return c2;
    }

    public static<E> Vector<E> map(double[] data, MapGetter<Double, E> mapper)
    {
        return mapTo(CollectionIterator.getIterator(data), mapper, new Vector<E>());
    }

    public static<E , A > Vector<E> map(Iterator<A> it, MapGetter<A, E> mapper)
    {
        return mapTo(it, mapper, new Vector<E>());
    }

    public static<E , A > HashSet<E> mapToSet( MapGetter<A, E> mapper, A ... data)
    {
        return mapToSet(new CollectionIterator<A>(data), mapper);
    }

    

    public static<E , A > HashSet<E> mapToSet(Iterator<A> it, MapGetter<A, E> mapper)
    {
        return mapTo(it, mapper, new HashSet<E>());
    }

    public static<E , A > HashSet<E> mapToSet(Collection<A> c, MapGetter<A, E> mapper)
    {
        return mapTo((c == null) ? null : c.iterator(), mapper, new HashSet<E>());
    }

    public static<E , A > HashSet<E> mapToSet(A[] c, MapGetter<A, E> mapper)
    {
        return mapTo((c == null) ? null : new CollectionIterator<A>(c), mapper, new HashSet<E>());
    }

    public static<A> Vector<A> cloneVector(Vector<A> c)
    {
        Vector<A> v = new Vector<>();
        v.addAll(c);
        return v;
    }

    public static<E extends Comparable, A > Vector<E> mapSorted(Iterator<A> it, MapGetter<A, E> mapper, boolean asc)
    {
        Vector<E> rv = new Vector<>();
        if (it == null)
            return rv;
        
        while (it.hasNext())
        {
            A o = it.next();
            rv.add(mapper.get(o));
        }
        Collections.sort(rv);
        if (!asc)
        {
            Collections.reverse(rv);
        }
        
        return rv;
    }

    public interface Creator<A >
    {
        public A create();
    }

    public static<A> Vector<A> create(int number, Creator<A> c)
    {
        Vector<A> v = new Vector<>();
        for (int i=0; i<number; i++)
        {
            v.add(c.create());
        }
        return v;
    }

    public interface Doer
    {
        public void doit();
    }

    public static void decide(boolean b, Doer d1, Doer d2)
    {
        if(b)
        {
            d1.doit();
            return;
        }
        d2.doit();
    }

    /** maps the objects to other objects */
    public static Vector map(Collection v, String expr)
    {
        Vector rv = new Vector();
        if (v == null)
            return rv;

        ReflectedObject ro = new ReflectedObject(expr);
        for (Object o : v)
        {
            rv.add(ro.get(o));
        }
        return rv;
    } 

    static public<A> Filterer<A> getNotNullFilter()
    {
        return (A a) -> a!=null;
    }
    /*public static Vector filter(Collection v, Filtererr ro)
    {
        Vector rv = new Vector();
        for (Object o : v)
        {
            Object so = ro.get(o);
            if (so==null || !((Boolean)so))
            {
                continue;
            }
            rv.add(o);
        }
        return rv;
    }
    */
    public static<A>  boolean contains(A[] v, Filterer<A> ro)
    {
        return contains((v==null)? null : new CollectionIterator<A>(v), ro);
    }
    
    public static<A>  boolean contains(Collection<A> v, Filterer<A> ro)
    {
        return contains((v==null)? null : v.iterator(), ro);
    }

    public static<A>  boolean contains(Iterator<A> it, Filterer<A> ro)
    {
        if (it == null)
            return false;
        
        while (it.hasNext())
        {
            if(ro.filter(it.next()))
                return true;
        }
        return false;
    }

    public static<A> A filterOne(A[] v, Filterer<A> ro)
    {
        return ObjectGetter.first(filter(v, ro));
    }

    public static<A > Vector<A> filter(A[] v, Filterer<A> ro)
    {
        return filter((v==null)?null:new CollectionIterator<A>(v), ro);
    }

    public static<A > Vector<A> filter(Collection<A> v, Filterer<A> ro)
    {
        return filter((v==null)?null:v.iterator(), ro);
    }

    public static<A > int filteredSize(Collection<A> v, Filterer<A> ro)
    {
        return (v == null) ? 0 : filter(v.iterator(), ro).size();
    }

    public static<A > int filteredSize(Iterator<A> it, Filterer<A> ro)
    {
        return filter(it, ro).size();
    }


    public static<A> A filterOne(Collection<A> v, Filterer<A> ro)
    {
        return ObjectGetter.first(filter(v, ro));
    }

    public static<A> A filterOne(Iterator<A> it, Filterer<A> ro)
    {
        return ObjectGetter.first(filter(it, ro));
    }

    public static<A, C2 extends Collection<A>> C2 filterTo(Iterator<A> it, Filterer<A> ro, C2 c)
    {
        if (it==null)
            return c;

        while (it.hasNext())
        {
            A o = it.next();
            
            boolean f = ro.filter(o);
            if (!f)
                continue;
            
            c.add(o);
        }
        return c;
    }
    public static<A > Vector<A> filter(Iterator<A> it, Filterer<A> ro)
    {
        return filterTo(it,ro, new Vector<A>());
    }

    public static<A > HashSet<A> filterToSet(Iterator<A> it, Filterer<A> ro)
    {
        return filterTo(it,ro, new HashSet<A>());
    }

    public static<A > HashSet<A> filterToSet(Collection<A> c, Filterer<A> ro)
    {
        return filterToSet((c==null) ? null : c.iterator(),ro);
    }

    public static Vector filter(Collection v, ReflectedObject ro)
    {
        Vector rv = new Vector();
        if (v==null)
            return rv;
        
        for (Object o : v)
        {
            Object so = ro.get(o);
            if (so==null || !((Boolean)so))
            {
                continue;
            }
            rv.add(o);
        }
        return rv;
    }
    /** filters by expression */
    public static Vector filter(Collection v, String expr)
    {
        return filter(v,new ReflectedObject(expr));
    }
    public static Vector<Object> getSortedObjects(Vector v, String expr)
    {
        ReflectedObject ro = new ReflectedObject(expr);
        Vector vp = new Vector();
        for (Object o : v)
        {
            vp.add(ro.get(o));
        }
        Collections.sort(vp);
        return vp;
    }
    
    public static Object max(Vector v, String expr)
    {
        Vector sv = getSortedObjects(v, expr);
        if (v.size()==0)
            return null;
        
        
        return sv.get(sv.size()-1);
    }
    
    public static Object min(Vector v, String expr)
    {
        Vector sv = getSortedObjects(v, expr);
        if (v.size()==0)
            return null;
        
        
        return sv.get(0);
    }

    public interface SortableGetter<A>
    {
        public Comparable get(A o);
    }

    public static<A > void sortBy(Vector<A> v, SortableGetter<A> ro, boolean asc)
    {
        Vector<Pair<Comparable, A>> vp = new Vector<>();
        Vector nulls = new Vector();
        for (A o : v)
        {
            if (o==null)
            {
                nulls.add(o);
                continue;
            }
            Comparable so = ro.get(o);
            //System.out.println("add: "+so+" type: "+so.getClass()+" and: "+o);
            if (so==null)
            {
                nulls.add(o);
                continue;
            }

            vp.add(new Pair<Comparable, A>(so, o));
        }
        Collections.sort(vp);
        v.clear();
        v.addAll(nulls);
        //int sni=0;
        for (Pair<Comparable, A> p : vp)
        {
            //System.out.println("sorted: "+(sni++)+" "+p);

            v.add(p.getSecond());
        }

        if(!asc)
        {
            Collections.reverse(v);
        }
    }

    public static void sortBy(Vector v, ReflectedObject ro, boolean asc)
    {
        Vector<Pair> vp = new Vector<Pair>();
        Vector nulls = new Vector();
        for (Object o : v)
        {
            if (o==null)
            {
                nulls.add(o);
                continue;
            }
            Object so = ro.get(o);
            //System.out.println("add: "+so+" type: "+so.getClass()+" and: "+o);
            if (so==null)
            {
                nulls.add(o);
                continue;
            }

            vp.add(new Pair(so, o));
        }
        Collections.sort(vp);
        v.clear();
        v.addAll(nulls);
        //int sni=0;
        for (Pair p : vp)
        {
            //System.out.println("sorted: "+(sni++)+" "+p);

            v.add(p.getSecond());
        }

        if(!asc)
        {
            Collections.reverse(v);
        }
    }
    public static void sortBy(Vector v, String expr, boolean asc)
    {
        
        ReflectedObject ro = new ReflectedObject(expr);
        sortBy(v,ro, asc);
        
        
    }
    public static void sortBy(Vector v, String expr)
    {
        sortBy(v,expr, true);
    }
    
    public static class ObjectGetterHeader<A>
    {
        public String headers[] = null;
        Vector<ObjectGetter<A>> data = new Vector<>();
        ReflectedObject[] _getter_vec=null;
        HashMap<String, ReflectedObject> getters = new HashMap<String, ReflectedObject>();
        HashMap<String, Integer> field2value = new HashMap<String, Integer>();
        Boolean[] is_number = null;

        public void updateHeaders()
        {
            Vector<Pair<Integer,String>> v = Pair.convert_reverse(field2value);
            Collections.sort(v);
            int maxi=v.get(v.size()-1).getFirst();
            headers = new String[maxi+1];
            _getter_vec=new ReflectedObject[headers.length];
            is_number=new Boolean[headers.length];


            for (Pair<Integer,String> p : v)
            {
                headers[p.getFirst()]=p.getSecond();
                ReflectedObject ro = getters.get(p.getSecond());
                _getter_vec[p.getFirst()]=ro;
            }
        }

        public boolean isNumericHeader(String fieldname)
        {
            return isNumericHeader(field2value.get(fieldname));
        }
        public boolean isNumericHeader(int idx)
        {
            if (is_number[idx] != null)
                return is_number[idx];
            
            boolean numeric = true;
            boolean onlynull = true;
            for (int i=0; i<data.size(); i++)
            {
                Object o = data.get(i).get(idx);
                if (o == null)
                    continue;
                
                onlynull = false;
                
                if (o instanceof Number)
                    continue;
                
                numeric = false;
                break;

            }
            return is_number[idx] = (!onlynull && numeric);
        }
        public ObjectGetterHeader clone(boolean copy_data_references)
        {
            ObjectGetterHeader ogh = new ObjectGetterHeader();
            ogh.headers= this.headers;
            ogh.getters.putAll(this.getters);
            ogh.field2value.putAll(field2value);
            ogh.is_number = (this.is_number == null) ? null : this.is_number.clone();

            if (copy_data_references)
            {
                ogh.data.addAll(this.data);

                
            }

            return ogh;
        }
        public HashMap<String, ReflectedObject> getGetters()
        {
            return getters;
        }
        public ReflectedObject addData(String nheader, ReflectedObject ro, Vector ndata)
        {
            if (ndata.size()!=data.size())
            {
                throw new RuntimeException("data sizes dont match!");
            }
            int index=-1;
            for (int i=0; i<ndata.size(); i++)
            {
                ObjectGetter og = data.get(i);
                index=og.data.size();
                Object o = ndata.get(i);
                if (ro!=null)
                {
                    o = ro.get(o);
                }
                og.data.add(o);
            }
            field2value.put(nheader, index);
            ReflectedObject rv = new ReflectedObject("get", index);
            getters.put(nheader, rv);
            updateHeaders();
            return rv;
        }
        public ReflectedObject addData(String nheader, Vector<Object> ndata)
        {
            return addData(nheader,null, ndata);
        }
        public ReflectedObject getFieldInfo(String header)
        {
            ReflectedObject ro =getters.get(header);

            if (ro==null)
            {
                throw new RuntimeException("cannot get header named: "+header+" known header names are: "+getters.keySet());
            }
            return ro;
        }

        public Object getFieldValue(String header, int i)
        {
            return getFieldInfo(header).get(data.get(i));
        }
        public HashMap<String, ReflectedObject> getFieldInfos()
        {
            return getters;
        }

        public ReflectedObject addField(String field, ReflectedObject getter)
        {
            int idx = getters.size();
            field2value.put(field, idx);
            getters.put(field, getter);
            return getter;
        }

        public ReflectedObject addField(String field, Integer value)
        {
            field2value.put(field, value);

            ReflectedObject ro = new ReflectedObject("get", value);
            getters.put(field, ro);
            return ro;
        }

        public TestExpression getExpression(String expr)
        {
            return new TestExpression(expr, getters);
        }


        public ObjectGetter<A> addData(A in_data)
        {
            ObjectGetter<A> og = new ObjectGetter<A>();
            og.header=this;
            og.idx=data.size();
            og.setInData(in_data);
            data.add(og);
            return og;
        }
        public ObjectGetter<A> addData()
        {
            ObjectGetter<A> og = new ObjectGetter<A>();
            og.data = new Vector();
            og.header=this;
            og.idx=data.size();
            data.add(og);
            return og;
        }

        public ObjectGetter<A> getData(int i)
        {
            return data.get(i);
        }

        public int getNumData()
        {
            return data.size();
        }
        
        public Vector<ObjectGetter<A>> getAllData()
        {
            return data;
        }
        public Vector<Iterator<Object> > classify(String expr, Collection<ObjectGetter> data )
        {
            TestExpression te = getExpression(expr);
            for (ObjectGetter o : data)
            {
                Object eo = te.eval(o);
                System.out.println(o+" => "+eo);
            }

            return null;
        }


        public Vector<ObjectGetter> filter(String expr)
        {
            TestExpression te = getExpression(expr);

            Vector<ObjectGetter> rv = new Vector<ObjectGetter>();
            for (ObjectGetter o : data)
            {
                if(!te.filter(o))
                    continue;

                rv.add(o);
            }
            return rv;
        }
    }
    public ObjectGetterHeader getHeader()
    {
        return header;
    }

    public static ObjectGetterHeader createHeader()
    {
        return new ObjectGetterHeader();
    }
    ObjectGetterHeader<A> header;
    //either data or _in_data must be set or both
    //if data is set, the data ist in a tabular form, so
    //that we can access the i-th field trough simple getting it
    //if _in_data is set, the reflected object to the i-th header in the header
    //object can access it
    Vector data = null; //new Vector();
    A _in_data = null;


    public void setInData(A o)
    {
        _in_data=o;
    }

    public A getInData()
    {
        return (_in_data==null)?null:_in_data;
    }
    public void add(Object o)
    {
        if(data==null)
            data=new Vector();
        data.add(o);
    }
    public Object get(Integer i)
    {
        if (data==null)
        {
            return header._getter_vec[i].get(_in_data);
        }
        return data.get(i);
    }
    
    public Boolean getBoolean(String n)
    {
        return (Boolean)get(n);
    }
    
    public String getString(String n)
    {
        return (get(n) == null) ? null :  ""+get(n);
    }

    public Integer getInt(String n)
    {
        return (Integer)get(n);
    }

    public Double getDouble(String n)
    {
        Object g = get(n);
        if (g instanceof Number)
        {
            return ((Number)g).doubleValue();
        }
        if (g instanceof String)
        {
            return Double.parseDouble((String)g);
        }
        //return (Double)get(n);
        throw new FRuntimeException("%s is not a double >%s<", n, g);
    }

    public Object get(int i)
    {
        if (data==null)
        {
            //System.out.println("getter: "+i+" on "+_in_data);
            return header._getter_vec[i].get(_in_data);
        }
        return data.get(i);
    }

    public void set(int i, Object o)
    {
        data.set(i, o);
    }
    
    public boolean gotHeader(String s)
    {
        return null != header.field2value.get(s);
    }

    public Object get(String s)
    {
        if (data!=null)
        {
            if (!gotHeader(s))
                throw new FRuntimeException("unknown header: %s", s);
            
//            if(he==null)
//            {
//                LogConfig.getLogger().warn("no field for header "+s);
//                return null;
//            }
            return data.get(header.field2value.get(s));
        }
        return header.getters.get(s).get(_in_data);

        
    }

    public Map<String, Object> getData()
    {
        HashMap<String, Object> rv = new HashMap<String, Object>();
        for (String h : header.headers)
        {
            rv.put(h, get(h));
        }
        return rv;
    }
    public static<A> Vector<A> buildVector(Iterator<A> it)
    {
        Vector<A> v = new Vector<>();
        while (it.hasNext())
        {
            v.add(it.next());
        }
        return v;
    }
    public static<A> Vector<A> buildVector(A ... data)
    {
        Vector<A> v = new Vector<>();
        for (A s : data)
        {
            v.add(s);
        }
        return v;
    }

    public int getDataLength()
    {
    	return (data!=null)?data.size():header.headers.length;
    }


    public static <A,B,C> HashMap<A, HashMap<B,C>> update(HashMap<A,HashMap<B,C>> map, A a, B b, C c)
    {
        HashMap<B,C> inner = map.get(a);
        if(inner==null)
        {
            inner = new HashMap<>();
            map.put(a, inner);
        }
        inner.put(b, c);
        return map;
    }
    public static<A,B> HashMap<A, HashSet<B>> update(HashMap<A, HashSet<B>> m, A a, B b)
    {
        return update(m,a,b, false);
    }
    public static<A,B> HashMap<A, HashSet<B>> update(HashMap<A, HashSet<B>> m, A a, B b, boolean synch)
    {
        HashSet<B> c = m.get(a);
        if (c == null)
        {
            if (synch)
            {
                synchronized(m)
                {
                    m.put(a, c = new HashSet<B>());
                }
            }
            else
            {
                m.put(a, c = new HashSet<B>());
            }
            
        }

        if (synch)
        {
            synchronized(c)
            {
                c.add(b);
            }
        }
        else
        {
            c.add(b);
        }
        return m;
    }


    public static<A,B> HashMap<A, Vector<B>> updateV(HashMap<A, Vector<B>> m, A a, B b)
    {
        return updateV(m,a,b, false);
    }
    public static<A,B> HashMap<A, Vector<B>> updateV(HashMap<A, Vector<B>> m, A a, B b, boolean synch)
    {
        Vector<B> c = m.get(a);
        if (c == null)
        {
            if (synch)
            {
                synchronized(m)
                {
                    m.put(a, c = new Vector<B>());
                }
            }
            else
            {
                m.put(a, c = new Vector<B>());
            }
            
        }

        if (synch)
        {
            synchronized(c)
            {
                c.add(b);
            }
        }
        else
        {
            c.add(b);
        }
        return m;
    }

    static public<A, C extends Collection<A>>  C merge(C target, Collection<A> ... inputs)
    {
        for (Collection<A> c : inputs)
        {
            if (c == null)
                continue;
            
            target.addAll(c);
        }
        return target;
    }

    static public<A>  Vector<A> mergeToVector(Collection<A> ... inputs)
    {
        return merge(new Vector<A>(), inputs);
    }

    static public<A>  HashSet<A> mergeToSet(Collection<A> ... inputs)
    {
        return merge(new HashSet<A>(), inputs);
    }

    public static int csize(Collection c)
    {
        return (c == null) ? 0 : c.size();
    }

    public static int csize(Iterator it)
    {
        if (it == null) 
            return 0;
        int rv = 0;
        while (it.hasNext())
        {
            it.next();
            rv++;
        }
        return rv;
    }

    public String toString()
    {
        return getInfos();
    }
    public String getInfos()
    {
        StringBuffer sb = new StringBuffer();
        for (String h : header.headers)
        {
            if (sb.length()>0)
            {
                sb.append(" ");
            }
            sb.append(String.format("%s=%s", h, get(h)));
            
        }
        return sb.toString();
    }

    static public<A> A first(A ... data)
    {
        return (data == null || data.length == 0) ? null : data[0];                                                                         
    }
    static public<A> A first(Iterator<A> it)
    {
        return (!it.hasNext()) ? null : it.next();
    }
    
    static public<A> A first(Collection<A> c)
    {
        return (c==null || c.size() == 0) ? null :first(c.iterator());
    }

    static public<A> Iterator<A> getIterator(Collection<A> c)
    {
        return (c == null) ? new SingleIterator(null) : c.iterator();
    }

    static public<A> Iterator<A> getIterator(A ... data)
    {
        return new CollectionIterator<A>(data);
    }

    static public<A> boolean containsElem(Iterator<A> it, A a)
    {
        if (it == null)
            return false;

        while (it.hasNext())
        {
            if (it.next() == a)
                return true;
        }
        return false;
    }


    public interface IndexMapper<A> extends MapGetter<Integer, A>
    {
    }

    static public<A> Vector<A> mapIndex(int length, IndexMapper<A> im)
    {
        return mapIndex(0, length, im);
    }

    static public<A> void applyIndex(int length, IndexMapper<A> im)
    {
        ObjectGetter.apply(IteratorUtils.getIntIterator(0, length), (_i) -> im.get(_i));
    }

    static public<A> Vector<A> mapIndex(Iterator<Integer> it, IndexMapper<A> im)
    {
        return map(it, im);
    }

    static public<A> Vector<A> mapIndex(Collection<Integer> c, IndexMapper<A> im)
    {
        return map((c==null) ? null : c.iterator(), im);
    }
    
    static public<A> Vector<A> mapIndex(int[] indeces, IndexMapper<A> im)
    {
        return mapIndex(CollectionIterator.getIterator(indeces), im);
    }

    static public<A> Vector<A> mapIndex(int start, int end, IndexMapper<A> im)
    {
        return mapIndex(IteratorUtils.getIntIterator(start, end), im);
    }

        
    static public<A> Vector<A> initializeN(int n, Creator<A> c)
    {
        return mapIndex(IteratorUtils.getIntIterator(0, n), (Integer idx) -> c.create());
    }

    public static<A extends Comparable, B extends Object> Vector<B> sort(Vector<B> v, MapGetter<B,A> ng)
    {
        return sort(v, ng, false);
    }

    public static<A extends Comparable, B extends Object> Vector<B> sort(Vector<B> v, MapGetter<B,A> ng, boolean reverse)
    {
        Vector<Pair<A, Integer>> _v = mapIndex(v.size(), (_i) -> Pair.create(ng.get(v.get(_i)), _i));

        Collections.sort(_v, (a1, a2) -> a1.getFirst().compareTo(a2.getFirst()));
        Vector<B> sorted = map(_v, (p) -> v.get(p.getSecond()));

        if (reverse)
        {
            Collections.reverse(sorted);
        }
        //return sort(v, ng, false);
        return sorted;
    }

    public static<A> Vector<A> select(Vector<A> data, Vector<Integer> selection)
    {
        Vector<A> rv = new Vector<>();
        for (int i=0; i<selection.size(); i++)
        {
            if (selection.get(i) == 0)
                continue;
            
            rv.add(data.get(i));
        }
        return rv;
    }
    
}

