package lmu.utils;

import java.io.File;
import java.util.*;

import org.jdom.Element;

public abstract class OptionParser<T>
{
    public static final String LONG_HELP_OPTION="--longhelp";
    public static final String PRINT_HELP_OPTION="--";
    public static final String CONFIG_XML_OPTION="--cmdconfigxml";

    String prefix="";
    public static class Option implements Comparable<Option>
    {
        int option_index;
        boolean multiple_settable=false;
        String switch_name;
        String option_name;
        String help_info;
        public String default_value=null;
        public String type;
        public boolean is_set=false;

        HashSet<Integer> outfile_params = new HashSet<Integer>();
        HashSet<Integer> existing_file_params = new HashSet<Integer>();
        HashSet<Integer> existing_dir_params = new HashSet<Integer>();
        HashSet<Integer> int_params = new HashSet<Integer>();
        HashSet<Integer> long_params = new HashSet<Integer>();
        HashSet<Integer> double_params = new HashSet<Integer>();
        HashSet<Integer> boolean_params = new HashSet<Integer>();

        HashMap<Integer, HashSet<String> > enum_params = new HashMap<Integer, HashSet<String> >();
        

        int nparams_needed=0;

        public void setNParams(int n)
        {
            nparams_needed=n;
        }
        public int compareTo(Option o)
        {
            return option_index-o.option_index;
        }
        public Option(int option_index, String name, String helpinfo, String default_value, int params_needed)
        {
            this.option_index=option_index; 
            this.option_name=name;
            this.switch_name="-"+name;
            this.help_info=helpinfo;
            this.default_value=default_value;
            setNParams(params_needed);
        }
        public void setType(String type)
        {
            this.type = type;
        }
        public void setDefaultValue(String default_value)
        {
            this.default_value = default_value;
        }
        private static int LEADING_WHITE_SPACES=3;
        public String getHelp(int maSwithName, int maxDefaultName, int maxType)
        {
            
            StringBuffer sb = new StringBuffer();
            
            sb.append(String.format("%-" + LEADING_WHITE_SPACES + "s","" ));
            if (default_value!=null)
            {
                sb.append('[');
                maSwithName-=1;
                if ( maSwithName <0)maSwithName=0;
            }
            maxType=(maxType>0?maxType:maxType+2);
            sb.append(String.format("%-"+(maSwithName+4) + "s",switch_name));
            sb.append(String.format("%-"+(maxType) + "s",type==null?"":type));
            sb.append(String.format("%-"+(maxDefaultName+2) + "s",default_value==null||default_value.trim().length()==0?"":"=" +default_value ));
            
            sb.append("" + help_info);
            /*
            if (default_value!=null)
            {
                sb.append((help_info.length()==0?"":" ") +"default:"+default_value);
            }
            sb.append(String.format("%s-" +maxDefaultName+3 , default_value));
*/
            if (default_value!=null)
            {
                sb.append(']');
            }
            String hstr=sb.toString();

            if (multiple_settable)
            {
                if (default_value!=null)
                    hstr= hstr+"+";

                hstr= "("+hstr+")+";
            }

            
            int offset = (maSwithName+4+maxDefaultName+2+LEADING_WHITE_SPACES+maxType+1);
            if (enum_params.size()>0)
            {
                hstr+=String.format("\n%" + (offset) + "s%s","","Options:") ;
            }
            for (Map.Entry<Integer, HashSet<String> > e: enum_params.entrySet())
            {
                //String s =;
                
                hstr+=String.format("\n%" + (offset+3) + "s%s","",(enum_params.size()==1?"":e.getKey() + ":") + StringUtils.join(", ", e.getValue())) ;
            }

            return hstr;

        }

        public boolean check(String[] args, int i, StringBuffer error)
        {
            //System.out.println("will check: "+args[i]+" file params: "+ existing_file_params);
            if (nparams_needed==0)
            {

                return true;
            }


            //check number params
            if (args.length <= i+nparams_needed)
            {

                error.append("switch "+option_name+" requires "+nparams_needed+" params, but there are only "+(args.length-i-1)+" left: ");
                for (int j=i+1; j<args.length; j++)
                {
                    error.append(" "+args[j]);
                }
                return false;
            }

            //check the file and dir params
            for (int fi : existing_file_params)
            {
                if (args.length-1<i+fi)
                {
                    if (error.length()>0)
                    {
                        error.append('\n');
                    }
                    error.append("switch "+option_name+" requires an existing file for parameter: "+fi+" (missing)");
                    continue;
                }
                File f = new File(args[i+fi]);
                if (f.exists())
                    continue;

                if (error.length()>0)
                {
                    error.append('\n');
                }
                error.append("switch "+option_name+" requires an existing file for parameter: "+fi+" ("+args[i+fi]+")");
            }
            
            //check the outfile params
            for (int fi : outfile_params)
            {
                if (args.length-1<i+fi)
                {
                    if (error.length()>0)
                    {
                        error.append('\n');
                    }
                    error.append("switch "+option_name+" requires a writeable filepath for parameter: "+fi+" (missing)");
                    continue;
                }
                File f = new File(args[i+fi]);
                
                if(f.isDirectory())
                {
                    error.append("switch " + option_name + " can not open output file for writing: "+ f.getAbsolutePath() +" reason: output file is a directory");
                    continue;
                }

                File fd = (f.exists()) ? f : (new File(f.getAbsolutePath())).getParentFile();
                
                if(fd.canWrite())
                    continue;
                
                String reason = (f == fd || fd.exists())  ? " no write permission " :  "parent directory not existing";
                
                error.append("switch " + option_name + " can not open output file for writing: "+ f.getAbsolutePath() +" reason:" + reason);
           
                
            }
            
            for (int di : existing_dir_params)
            {
                File f = new File(args[i+di]);
                if (f.isDirectory())
                    continue;

                if (error.length()>0)
                {
                    error.append('\n');
                }
                error.append("switch "+option_name+" requires an existing directory for parameter: "+di+" ("+args[i+di]+")");
            }

            for (int ii : int_params)
            {
                try
                {
                    Integer.parseInt(args[i+ii]);
                }
                catch (NumberFormatException e)
                {
                    error.append("swith "+option_name+" requires an integer parameter: "+ii+" ("+args[i+ii]+")");
                }
            }

            for (int ii: long_params)
            {
                try
                {
                    Long.parseLong(args[i+ii]);
                }
                catch (NumberFormatException e)
                {
                    error.append("switch "+option_name+" requires a long parameter: "+ii+" ("+args[i+ii]+")");
                }
            }

            for (int ii : double_params)
            {
                try
                {
                    Double.parseDouble(args[i+ii]);
                }
                catch (NumberFormatException e)
                {
                    error.append("switch "+option_name+" requires a double parameter: "+ii+" ("+args[i+ii]+")");
                }
            }

            for (Map.Entry<Integer, HashSet<String> > e : enum_params.entrySet())
            {
                String v = args[i+e.getKey()];
                HashSet<String> hs = e.getValue();
                if (hs.contains(v))
                {
                    continue;
                };
                error.append("switch "+option_name+" requires for parameter "+e.getKey()+" a value from set: "+hs+" got "+args[i+e.getKey()]);

            }


            if (error.length()==0)
            {
                is_set=true;
                return true;
            }
            //System.out.println("resulting error: "+error.toString());
            return false;
        }

        void check_param_index(int param_index)
        {
            if (param_index>nparams_needed)
            {
                throw new RuntimeException("error option "+option_name+" is registered with only "+nparams_needed+" params cannot set property for param index: "+param_index);
            }
        }

        public void setEnumParam(int param_index, Collection<String> values)
        {
            check_param_index(param_index);

            HashSet<String> hs = enum_params.get(param_index);
            if (hs==null)
            {
                enum_params.put(param_index, hs = new HashSet<String>());
            }
            hs.addAll(values);
        }
        public void setEnumParam(int param_index, String ... values)
        {
            setEnumParam(param_index, (Object[])values);
        }

        void setEnumParam(int param_index, Object ... values)
        {
            check_param_index(param_index);

            HashSet<String> hs = enum_params.get(param_index);
            if (hs==null)
            {
                enum_params.put(param_index, hs = new HashSet<String>());
            }

            for (Object v : values)
            {
                hs.add((String)v);
            }
        }
        public void setLongParam(int param_index)
        {
            check_param_index(param_index);
            long_params.add(param_index);
        }
        public void setIntParam(int param_index)
        {
            check_param_index(param_index);
            int_params.add(param_index);
        }
        public void setDoubleParam(int param_index)
        {
            check_param_index(param_index);
            double_params.add(param_index);
        }
        public void setBoolParam(int param_index)
        {
            setEnumParam(param_index, "false", "true"); 

        }
        public void setDirParam(int param_index)
        {

            check_param_index(param_index);
            existing_dir_params.add(param_index);
        }

        public void setFileParam(int param_index)
        {
            check_param_index(param_index);
            existing_file_params.add(param_index);
        }
        
        public void setOutFileParam(int param_index)
        {
            check_param_index(param_index);
            outfile_params.add(param_index);
        }

        public String getOptionName()
        {
            return option_name;
        }
        public String toString()
        {
            return String.format("option: "+option_name+" param index: "+option_index+" needed params: "+nparams_needed);
        }
        public void setHelp(String help)
        {
            help_info = help;

        }
    }

    public static final int OPTION_NOT_FOUND = -1;
    public static final int OPTION_FOUND_BUT_ERROR = -100;
    HashMap<Option, Boolean> options = new HashMap<Option, Boolean>();

    StringBuffer error = new StringBuffer();

    /** returns if parsing was successful, if not error is set */
    abstract protected boolean parse(int option_index, String[] args, int i, StringBuffer error);
    abstract public T getObject();
    abstract protected String getName();

    public Collection<Option> getOptions()
    {
        return options.keySet();
    }
    
    public Option getOption(int index)
    {
        for (Option o : options.keySet())
        {
            if (o.option_index==index)
                return o;


        }
        return null;
    }
    public Option getOption(String name)
    {
        for (Option o : options.keySet())
        {
            if (o.option_name.equals(name) || o.switch_name.equals(name))
                return o;
        }
        return null;
    }
    public Option registerOneParamBoolOption(int option_index, String name, Boolean default_value)
    {
        Option o= registerOption(option_index, name, "true or false "+((default_value==null)?"":" defaults to"+default_value),
                                 (default_value==null)?null:""+default_value, 1);
        o.setBoolParam(1);
        return o;
    }
    Option registerOneParamOption(int option_index, String name, String helpinfo, Object default_value)
    {
        return registerOption(option_index, name, helpinfo, (default_value==null)?null:""+default_value, 1);
    }
    public Option registerOneParamIntOption(int option_index, String name, String helpinfo, Integer default_value)
    {
        Option o= registerOneParamOption(option_index,name,helpinfo, default_value);
        o.setIntParam(1);
        return o;
    }

    public Option registerOneParamDoubleOption(int option_index, String name, String helpinfo, Double default_value)
    {
        Option o= registerOption(option_index,name,helpinfo, (default_value==null)?null:String.format("%2.3f",default_value), 1);
        o.setDoubleParam(1);
        return o;
    }
    public Option registerOption(int option_index, String name, String helpinfo, String default_value, int params_needed, boolean multiple)
    {
        Option o= registerOption(new Option(option_index, prefix+name, helpinfo, default_value, params_needed));
        o.multiple_settable=multiple;
        return o;
    }

    public Option registerOneParamInputFileOption(int option_index, String name, String helpinfo )
    {
        Option o= registerOption(option_index, name, helpinfo, null, 1);
        o.setFileParam(1);
        return o;
    }
    public Option registerOneParamEnumOption(int option_index, String name, String helpinfo, String default_value, String ... values)
    {
        return registerOneParamEnumOption(option_index,name,helpinfo,default_value, (Object[]) values);
    }
    private Option registerOneParamEnumOption(int option_index, String name, String helpinfo, String default_value, Object[] values)
    {
        Option o= registerOption(option_index, name, helpinfo, default_value, 1);
        o.setEnumParam(1, values);

        return o;
    }
    public Option registerOneParamEnumOption(int option_index, String name, String helpinfo, String default_value, Collection<String> values)
    {
        return registerOneParamEnumOption(option_index, name, helpinfo, default_value, values.toArray());
    }

    public Option registerOption(int option_index, String name, String helpinfo, String default_value, int params_needed)
    {
        return registerOption(option_index,name,helpinfo,default_value,params_needed,false);

    }
    public Option registerOption(Option option)
    {
        options.put(option, false);
        return option;
    }


    boolean suppressLongHelp=false;

    public void suppressLongHelp(boolean b)
    {
        suppressLongHelp=b;
    }


    public String getHelp()
    {
        if (suppressLongHelp)
        {
            return String.format("( %d parameters for %s, use %s to get all options)\n", options.size(), getName(), LONG_HELP_OPTION);
        }
        StringBuffer sb = new StringBuffer(getName()+" parameters:\n");
        Vector<Option> params = new Vector<Option>();
        params.addAll(options.keySet());
        Collections.sort(params);
        
        int maxName = NumUtils.max(params, (Option o) -> o.switch_name.length());
        int maxDefault = NumUtils.max(params, (Option o) -> o.default_value==null?0:o.default_value.length());
        int maxType = NumUtils.max(params, (Option o) -> o.type==null?0:o.type.length());
        
        for (Option o : params)
        {
            sb.append(o.getHelp(maxName,maxDefault,maxType)+"\n");
        }
        return sb.toString();
    }


    public boolean isInited()
    {
        return isInited(null);
    }
    /** the class is inited if all non-default parameters are set */
    public boolean isInited(StringBuffer errorstr)
    {
        for (Map.Entry<Option, Boolean> e : options.entrySet())
        {
            if (e.getValue() || e.getKey().nparams_needed==0 || e.getKey().default_value!=null)
                continue;

            if (errorstr==null)
                return false;

            if (errorstr.length()>0)
                errorstr.append(' ');

            errorstr.append(e.getKey().switch_name);
        }
        if (errorstr!=null && errorstr.length()>0)
        {
            return false;
        }
        return true;
    }


    public void parseOptions(Element e)
    {

        Iterator it = e.getDescendants();
        while (it.hasNext())
        {
            Object o = it.next();
            if (!(o instanceof Element))
                continue;



            Element che=(Element)o;

            String key = che.getName();

            if (key==null)
            {
                XMLUtils.getAttrib(che, "param", null, false);
            }

            if (key==null)
                key = XMLUtils.getAttrib(che, "name", null, false);

            String value = XMLUtils.getAttrib(che, "value", null, false);
            if (value==null)
            {
                value = XMLUtils.getAttrib(che, "path", null, false);
            }

            if (key==null || value==null)
            {
                continue;
            }

            //System.out.println("try to parse by "+getName()+" : "+key+"=>"+value);
            parseOption(key,value);

        }
    }

    public boolean parseOption(String key, String value)
    {
        return parseOption(key,value,true, false);
    }

    public boolean parseOption(String key, String value, boolean throw_exc, boolean print_err)
    {
        String [] args = {key, value};
        int np=parseOption(args,0, throw_exc, print_err);

        return np>=0;

    }

    public int parseOption(String[] args, int i)
    {
        return parseOption(args,i,false,true);
    }

    /** returns the number of args has to be skipped additionally 
     * if return value <0 it is one of
     * OPTION_NOT_FOUND
     * or OPTION_FOUND_BUT_ERROR
     */
    public int parseOption(String[] args, int i, boolean throw_exc, boolean print_err)
    {

        error.setLength(0);
        Option option_found = parseOption(args,i,error);
        //System.err.println(getName()+" will read now argument: "+args[i]+" error: "+error.toString()+" option: "+option_found);


        if (error.length()==0 && option_found==null)
        {

            return OPTION_NOT_FOUND;
        }


        if (error.length()==0)
        {
            options.put(option_found, true);
            return option_found.nparams_needed;
        }

        if (!print_err && !throw_exc)
        {
            return option_found.nparams_needed;
        }

        String errstr ="error at parsing param: "+args[i]+" for "+getName()+" :" +error.toString();
        if (print_err)
        {
            System.err.println(errstr);
        }
        if (throw_exc)
        {
            throw new RuntimeException(errstr);
        }
        return OPTION_FOUND_BUT_ERROR; //found but erroneous
    }
    /** returns if parsing was successful if switch found  but error, error is set*/
    Option parseOption(String[] args, int i, StringBuffer error)
    {
        for (Option o : options.keySet())
        {
            if (!o.switch_name.equals(args[i]) && !o.option_name.equals(args[i]))
                continue;


            try
            {
                if (o.check(args,i,error) && parse(o.option_index,args ,i,error))
                {
                    //successfully parsed
                    return o;
                }
            }
            catch (Exception e)
            {
                if (error.length()!=0)
                {
                    error.append('\n');
                }
                error.append("error at parameter evaluation of switch "+args[i]+" :" + e.getMessage());
                error.append("\nstacktrace:\n"+ ExceptionManager.getStackTrace(e));
            }
            if (error.length()==0)
            {
                error.append("unspecified error at reading parameter: "+o.option_name);
            }
            return o;
        }
        return null;
    }


    /** if prefix given all options will be prefixed with it */
    public OptionParser(String prefix)
    {
        this.prefix=prefix;
    }
    public OptionParser()
    {
        this("");
    }


    public static void parseConfigXML(File f, OptionParser ... parsers)
    {

        if (!f.exists())
        {
            throw new RuntimeException(String.format("config file %s does not exist.",f.getAbsolutePath()));

        }
        try
        {
            for (OptionParser op: parsers)
            {
                if (op==null)
                {
                    continue;
                }
                op.parseOptions(XMLUtils.getResolvedRootElement(f));
            }


        }
        catch (Exception e)
        {

            throw new RuntimeException("error  "+e.getMessage()+" at parsing config xml: "+f.getAbsolutePath(), e);
        }

    }
    public static void parseConfigXML(String[] args, OptionParser ... parsers)
    {
        for (int i=0; i<args.length; i++)
        {
            if (args[i].equals(CONFIG_XML_OPTION))
            {
                if (args.length<i+2)
                {
                    System.err.println("not enough arguments: switch config "+CONFIG_XML_OPTION+" needs an existing xml file as param");
                    System.exit(-1);
                }
                parseConfigXML(new File(args[++i]), parsers);
                return;
            }
        }

    }

    static public boolean parseConfigFileAndParams(File config, String[] args, OptionParser ... parsers)
    {
        if (config!=null)
        {
            args = StringUtils.concatArrays(args, CONFIG_XML_OPTION, config.getAbsolutePath());
        }

        if (!OptionParser.parseParams(args,true, false, true, parsers))
        {
            return false;
        }
        return true;
    }

    static public boolean parseParams(String[] args, boolean writerror, boolean throw_exception, boolean write_help_if_not_inited, OptionParser ... parsers)
    {
        return parseParams(args,writerror,throw_exception,write_help_if_not_inited,true, parsers);
    }

    static public boolean parseParams(String[] args, boolean writerror, boolean throw_exception, boolean write_help_if_not_inited, boolean write_help_if_no_args, OptionParser ... parsers)
    {
        int toskip=-1;
        StringBuffer errbuff = new StringBuffer();
        boolean USE_LONG_HELP=false;

        boolean force_longhelp = false;

        boolean print_longhelp = false;
        parseConfigXML(args, parsers);
        for (int i=0; i<args.length; i++)
        {
            boolean found=false;
            if (args[i].equals(PRINT_HELP_OPTION))
            {
                force_longhelp = true;
                continue;
            }
            if (args[i].equals(LONG_HELP_OPTION))
            {
                USE_LONG_HELP=true;
                continue;
            }

            if (args[i].equals(CONFIG_XML_OPTION))
            {
                i+=1;
                continue;
            }

            //System.out.println("will check param: "+args[i]);
            for (OptionParser op : parsers)
            {

                if (op==null)
                {
                    continue;
                }
                if (0<=(toskip=op.parseOption(args,i, throw_exception, writerror)))
                {
                    //System.out.println("read param, (by:"+op.getName()+") skip "+toskip+" params got "+i+"/"+args.length);
                    i+=toskip;

                    found=true;
                    break;
                }
                if (toskip==OPTION_FOUND_BUT_ERROR)
                {
                    if (writerror)
                    {
                        System.err.println(errbuff.toString());
                    }
                    return false;

                }
            }
            if (found)
                continue;


            if (writerror)
            {
                System.err.println("unknown parameter: "+args[i]);
            }

        }

        boolean inited=true;

        //read all params to the registered parsers... check if all inited
        StringBuffer notinited=new StringBuffer();

        for (OptionParser op : parsers)
        {
            if (op==null)
            {
                continue;
            }
            StringBuffer err = new StringBuffer();
            if (op.isOptional() || op.isInited(err))
            {
                //System.out.println(op.getName()+" is inited!");
                continue;
            }


            notinited.append(op.getName()+" requires the following non-optional parameters:\n   "+err.toString()+"\n");
            inited=false;

        }
        if (force_longhelp || (inited && args.length==0 && write_help_if_no_args))
        {
            for (OptionParser op : parsers)
            {
                if (USE_LONG_HELP)
                {
                    op.suppressLongHelp(false);
                }
                System.out.println(op.getHelp());
            }
            if (force_longhelp)
            {
                System.exit(0);
            }
            System.out.printf("to set multiple parameters from an xml file use the switch: %s\n", CONFIG_XML_OPTION);

        }
        if (inited || !write_help_if_not_inited)
        {
            return inited;
        }
        System.out.println(notinited.toString());
        for (OptionParser op : parsers)
        {
            if (USE_LONG_HELP)
            {
                op.suppressLongHelp(false);
            }
            System.out.println(op.getHelp());
        }

        System.out.printf("to set multiple parameters from an xml file use the switch: %s\n", CONFIG_XML_OPTION);


        return inited;
    }

    boolean optional=false;

    public void setOptional(boolean b)
    {
        optional=b;
    }
    public boolean isOptional()
    {
        return optional;
    }
}

