package lmu.utils;

import java.util.*;
import com.vividsolutions.jts.index.strtree.STRtree;
import com.vividsolutions.jts.geom.Envelope;


public class Overlap2DDetector
{
    Logger log = LogConfig.getLogger();
    STRtree tree = new STRtree();
    HashMap<Object, HashSet<Object> > overlaps = new HashMap<Object, HashSet<Object> >();

    HashMap<Object, Envelope> envelopes = new HashMap<Object, Envelope>();

    boolean compiled=false;

    

    class OverlapIterator implements Iterator<Pair>
    {
        
        boolean save;

        Iterator queryIt;
        Iterator<Pair> pair_it;

        HashSet<Object> analyzed = new HashSet<Object>();
        
        OverlapIterator(boolean save)
        {
            this.save=save;
            //log.debug("start overlap iterator, got %d init envelopes", envelopes.size());
            queryIt = envelopes.keySet().iterator();
        }
        public void remove()
        {
            throw new UnsupportedOperationException();
        }
        boolean precalcNext()
        {
            pair_it=null;
            
            while (queryIt.hasNext() && pair_it==null)
            {
                Object q = queryIt.next();

                
                HashSet ov = getOverlaps(q, save);

                Vector<Pair> pairs = new Vector<Pair>();
                for (Object o : ov)
                {
                    if (analyzed.contains(o))
                        continue;
                    
                    pairs.add(new Pair(q, o));
                }
                analyzed.add(q);

                //log.debug("got %d overlaps to next query: %s\n", pairs.size(), q);
                if (pairs.size()==0)
                    continue;
                
                pair_it=pairs.iterator();

            }

            return pair_it!=null;
            
        }
        public boolean hasNext()
        {
            if (pair_it!=null && pair_it.hasNext())
                return true;

            return precalcNext();
        }

        public Pair next()
        {
            if (pair_it!=null && pair_it.hasNext())
            {
                return pair_it.next();
            }
            
            //precalc next
            if (!queryIt.hasNext())
                return null;
            
            
            if (!precalcNext())
                return null;
            
            return pair_it.next();
            
        }

        

    }

    public Iterator<Pair> getOverlaps(boolean save)
    {
        return new OverlapIterator(save);
    }
    public void add(Pair dim1, Pair dim2,  Object o)
    {
        add((Double)dim1.getFirst(),(Double)dim1.getSecond(),(Double)dim2.getFirst(),(Double)dim2.getSecond(), o);
    }

    public void add(Region1D dim1, Region1D dim2,  Object o)
    {
        add((double)dim1.getX1(),(double)dim1.getX2(),(double)dim2.getX1(),(double)dim2.getX2(), o);
    }
    public void add(java.awt.Rectangle r,  Object o)
    {
        add(r.getMinX(), r.getMaxX(), r.getMinY(), r.getMaxY(),o );
    }
    
    public void add(java.awt.geom.Rectangle2D box, Object o)
    {
        add(box.getX(), box.getX()+box.getWidth(), box.getY(), box.getY()+box.getHeight(),o);
    }
    public void add(double x1, double x2, double y1, double y2, Object o)
    {
        Envelope env = new Envelope(x1,x2,y1,y2);
        envelopes.put(o, env);
        tree.insert(env,o);
        //log.debug("inserted : "+env+" to "+o);
        compiled=false;
    }

    public HashSet getOverlaps(double x, double y, double around)
    {
        Envelope env = new Envelope(x-around, x-around, y-around, y+around);
        return getOverlaps(env, null);
    }

    public HashSet getOverlaps(double x1, double x2, double y1, double y2, Object o)
    {
        return getOverlaps(new Envelope(x1,x2,y1,y2), o);
    }
    HashSet getOverlaps(Envelope e, Object o)
    {
        //log.debug("query "+e+" to object: "+o);
        HashSet rv = new HashSet();
        for(Object to: tree.query(e))
        {
            if (o!=null && to==o)
                continue;
            
            rv.add(to);
        }
        //log.debug("found "+rv.size()+" overlaps");
        return rv;

    }
    public HashMap<Class, HashSet> getTypedOverlapMap(Object o)
    {
        HashMap<Class, HashSet> rv = new HashMap<Class, HashSet>();

        for (Object ho : getOverlaps(o,true))
        {
            Class clazz = ho.getClass();
            HashSet hs = rv.get(clazz);
            if (hs==null)
            {
                rv.put(clazz, hs= new HashSet());
            }
            hs.add(ho);
        }
        return rv;
    }
    public HashSet getOverlaps(Object o)
    {
        return getOverlaps(o,true);
    }
    public HashSet getOverlaps(Object o, boolean save)
    {
        
        compile();
        HashSet rv = overlaps.get(o);
        if (rv!=null)
        {
            return rv;
        }
        Envelope e = envelopes.get(o);
        if (e==null)
        {
            throw new RuntimeException("object: "+o+" is not yet added (no envelope found");
        }

        rv=getOverlaps(e,o);
        if (save)
        {
            overlaps.put(o, rv);
        }
        
        return rv;
    }

    void compile()
    {
        if (compiled)
        {
            return;
        }
        log.debug("start compilining tree with "+envelopes.size()+" envelopes");
        overlaps.clear();
        tree.build();
        log.debug("compiling ready!");
        compiled=true;
    }
    

    public static void main (String args[]) 
    {
        LogConfig.configureLog(null, true);
        Overlap2DDetector od = new Overlap2DDetector();
        for (int i=0; i<100; i++)
        {
            double x1 = Math.random()*10;
            double x2 = x1+Math.random()*10;

            double y1 = Math.random()*10;
            double y2 = y1+Math.random()*10;

            java.awt.geom.Rectangle2D r = new java.awt.geom.Rectangle2D.Double(x1, y1, x2-x1, y2-y1);

            od.add(x1,x2,y1,y2, r);

        }
        Iterator<Pair> it = od.getOverlaps(false);

        while (it.hasNext())
        {
            Pair p = it.next();
            java.awt.geom.Rectangle2D r1 = (java.awt.geom.Rectangle2D)p.getFirst();
            java.awt.geom.Rectangle2D r2 = (java.awt.geom.Rectangle2D)p.getSecond();

            System.out.println("overlap: "+r1+" vs "+ r2+" : "+r1.createIntersection(r2));
        }
    }
}

