
package lmu.utils;

import java.util.*;

public class OverlapDetector
{
    enum RegTagType
    {
        OPEN,
        CLOSE;
    }

    static class RegTag implements Comparable<RegTag>
    {
        RegTagType _type;
        Object o;
        int p;

        public RegTag(RegTagType _t, int x, Object o)
        {
            this.o=o; _type=_t;
            p=x;
        }

        public String toString()
        {
            return String.format("%s:%d %s", ""+_type, p, o);}
        public int compareTo(RegTag rt)
        {
            int d = p-rt.p;
            if (d!=0)
                return d;
            
            if(_type==rt._type) return 0;

            return (_type==RegTagType.CLOSE)?-1:1;
            
        }
    }
    
    public interface OverlapFilter
    {
        public boolean filter(HashSet s);
    }
    static class OpenRegionHandler
    {
        int min_overlap_size=1;
        Vector<Region1D<HashSet>>  overlaps = new Vector<Region1D<HashSet>>();
        HashSet open = new HashSet();
        OverlapFilter filter = null;

        int start;
        //int current_end=-1;
        

        /** closes a current region if neccesarry
         */
        public void closeRegion(RegTag rt)
        {
            if (open.size()==0)
                return;
            
            addCurrentRegion(rt.p);
            open.remove(rt.o);
            start=rt.p;
        }

        void addCurrentRegion(int end)
        {
            if (start==end)
                return;
            
            if (open.size()<min_overlap_size || (filter!=null && filter.filter(open)))
            {
                return;
            }
            HashSet oset = new HashSet<Region1D>();
            oset.addAll(open);
            overlaps.add(new Region1D<HashSet>(start,end,oset));
        }
        public void openRegion(RegTag rt)
        {

            if (open.size()!=0)
            {
                addCurrentRegion(rt.p);
            }
            start=rt.p;
            open.add(rt.o);
        }
    }

    
    public static Vector<Region1D<HashSet>> getOverlaps(Collection<Region1D> regions)
    {
        return getOverlaps(regions,1, null);
    }
    public static Vector<Region1D<HashSet>> getOverlaps(Collection<Region1D> regions, int minoverlap, OverlapFilter filter)
    {
        OpenRegionHandler orh = new OpenRegionHandler();
        orh.min_overlap_size=minoverlap;
        orh.filter=filter;
        
        Vector<RegTag> regtags = new Vector<RegTag>();

        for (Region1D r : regions)
        {
            regtags.add(new RegTag(RegTagType.OPEN, r.getX1(),r.getObject()));
            regtags.add(new RegTag(RegTagType.CLOSE, r.getX2(), r.getObject()));
        }
        //System.out.println("sorting "+regtags.size()+" regtags");

        Collections.sort(regtags);

        //System.out.println("start building overlapping regions");
        

        for (RegTag rt : regtags)
        {
            //System.out.println("next rt: "+rt);
            switch (rt._type)
            {
                case OPEN:
                    orh.openRegion(rt);
                    break;
                case CLOSE:
                    orh.closeRegion(rt);
                    break;
            }
        }
        return orh.overlaps;
        
    }
    
        
    public interface AnnotOverlapFilter
    {
        public boolean filter(Region1D query, Region1D target);
        
        
    }


    public static Vector<Vector<Region1D>> annotateOverlapping(Vector<Region1D> q_regions,  Vector<Region1D> t_regions)
    {
        return annotateOverlapping(q_regions,t_regions,null);
    }
    public static Vector<Vector<Region1D>> annotateOverlapping(Vector<Region1D> q_regions,  Vector<Region1D> t_regions, AnnotOverlapFilter of)
    {
        Vector<Vector<Region1D> > rv = new Vector<Vector<Region1D>>();
        int last_start=0;
        for (int qi=0; qi<q_regions.size(); qi++)
        {
        
            Region1D r = q_regions.get(qi);
            
            //System.out.println("next query: "+r+ "last start: "+last_start+"/"+t_regions.size());

            final int x1 = r.getX1();
            while (last_start<t_regions.size())
            {
                Region1D testr = t_regions.get(last_start);
                if (testr.getX2()>=x1)
                {
                    break;
                    
                }
                last_start++;
            }
            if (last_start>=t_regions.size())
            {
                rv.add(null);
                continue;
            }
            
            Region1D start_ov = t_regions.get(last_start);

            //System.out.println("found "+start_ov+" as first overlapping at "+last_start);

            Vector<Region1D> overlaps = new Vector<Region1D>();
            final int x2=r.getX2();
            int last_ov=last_start;
            while (last_ov<t_regions.size())
            {
                Region1D testr = t_regions.get(last_ov);
                if (testr.getX1()>=x2)
                {
                    break;
                }
                if (of==null || of.filter(r, testr))
                {
                    //System.out.println("\tov ("+qi+"): "+testr.getObject()+" to ("+last_ov+"): "+r.getObject());
                    overlaps.add(testr);
                }
                
                last_ov++;
            }
            rv.add(overlaps);
        }
        return rv;
    }
        

}

