package lmu.utils;

import java.util.*;

public class Pair<A,B> implements java.io.Serializable, Comparable<Pair>
{
    public A first;
    public B second;

    static int[] primes=PrimeGenerator.generatePrimes(3, 100);
    public Pair()
    {

    }
    public Pair(A first, B second)
    {
        this.first=first; this.second=second;
    }

    static public<X extends Number, Y extends Number> void updateLimits(Pair<X,Y> toupdate, Pair<X,Y> updatewith)
    {
        if (updatewith == null)
            return;
        toupdate.setFirst(NumUtils.min(toupdate.getFirst(), updatewith.getFirst()));
        toupdate.setSecond(NumUtils.max(toupdate.getSecond(), updatewith.getSecond()));
    }

    static public<X extends Number, Y extends Number> Pair<X,Y> max(Pair<X,Y> p1, Pair<X,Y> p2)
    {
        if (p1 == null || p2 == null)
            return null;
        
        return create(NumUtils.max(p1.getFirst(), p2.getFirst()), NumUtils.max(p1.getSecond(), p2.getSecond()));
    }
    
    static public<X extends Number, Y extends Number> Pair<X,Y> min(Pair<X,Y> p1, Pair<X,Y> p2)
    {
        if (p1 == null || p2 == null)
            return null;
        
        return create(NumUtils.min(p1.getFirst(), p2.getFirst()), NumUtils.min(p1.getSecond(), p2.getSecond()));
    }


    static public<A,B> Vector<A> reduceFirst(Vector<Pair<A,B>> v)
    {
        return ObjectGetter.map(v, (Pair<A,B> p) -> p.getFirst());
    }

    static public<A,B> Vector<B> reduceSecond(Vector<Pair<A,B>> v)
    {
        return ObjectGetter.map(v, (Pair<A,B> p) -> p.getSecond());
    }
        

    public static<A,B> Pair<A,B> create(A a, B b)
    {
        return new Pair<A,B>(a,b);
    }


    
    public static class PairBuilder<A,B> extends Vector<Pair<A,B>>
    {
        public PairBuilder(A a, B b)
        {
            super();
            super.add(Pair.create(a,b));
        }

        public PairBuilder<A,B> add(A a, B b)
        {
            super.add(Pair.create(a,b));
            return this;
        }
    }
    public static<A,B> PairBuilder<A,B> build(A a, B b)
    {
        return new PairBuilder<A,B>(a,b);
        
    }
    /*
    public static<A,B> Pair<A,B> buildTuple(A a, B b)
    {
        
    }
    */

    public Pair<A,B> set(Pair<A,B> p)
    {
        setFirst(p.getFirst());
        setSecond(p.getSecond());
        return this;
    }
    public Pair<A,B> clone()
    {
        return new Pair<A,B>(first,second);
    }
    public static<A, B> Vector<A> getProjectionOnFirst(Vector<Pair<A,B> > c)
    {
        Vector<A> v = new Vector<A>();

        for (Pair<A,B> p : c)
        {
            v.add(p.getFirst());
        }
        return v;
    }
    public static<A extends Object, B extends Object> Vector<B> getProjectionOnSecond(Vector<Pair<A,B> > c)
    {
        Vector<B> v = new Vector<B>();

        for (Pair<A,B> p : c)
        {
            v.add(p.getSecond());
        }
        return v;
    }
        
    
    
    public static<A extends Object, B extends Collection> Vector<Pair<Integer, A> > getSizes(Map<A, B> c)
    {
        return ObjectGetter.map(c.entrySet(), (Map.Entry<A,B> e) -> new Pair<Integer,A>(e.getValue().size(), e.getKey()));
    }

    public static<A, B> HashMap<A,B> toMap(Collection<Pair<A,B>> c)
    {
        return ObjectGetter.buildMap(c, (Pair<A,B> p1) -> p1.getFirst(), (Pair<A,B> p2) -> p2.getSecond());
    }

    public static<A, B> Vector<Pair<A,B> > convert(Map<A,B> c)
    {
        Vector<Pair<A,B> > rv = new Vector<Pair<A,B> >();
        for(Map.Entry<A,B> e : c.entrySet())
        {
            rv.add(new Pair<A,B>(e.getKey(),e.getValue()));
        }
        return rv;
    }

    public static<A, B> Vector<Pair<A,B> > convert_sorted(Map<A,B> c, boolean asc, Comparator<A> comp)
    {
        Vector<Pair<A,B> > rv = convert(c);
        Collections.sort(rv, (Pair<A,B> a1, Pair<A,B> a2) -> comp.compare(a1.getFirst(), a2.getFirst()));
        if (asc)
            return rv;
        
        Collections.reverse(rv);
        return rv;
    }
    public static<A extends Comparable, B> Vector<Pair<A,B> > convert_sorted(Map<A,B> c, boolean asc)
    {
        return convert_sorted(c, asc, (A a1, A a2) -> a1.compareTo(a2));
    }

    public static<A,B> Pair<B,A> swap(Pair<A,B> p)
    {
        return Pair.create(p.getSecond(), p.getFirst());
    }
    public static<A , B > Vector<Pair<B,A> > swap(Vector<Pair<A,B>> v)
    {
        return swap(v, false, false);
    }
    public static<A, B> Vector<Pair<B,A> > swap(Vector<Pair<A,B>> v, boolean sort, boolean reverse)
    {
        if (sort)
        {
            Collections.sort(v);
        }
        if (reverse)
        {
            Collections.reverse(v);
        }
        Vector<Pair<B,A> > rv = new Vector<Pair<B,A> >();
        for (Pair<A,B> p : v)
        {
            rv.add(new Pair<B,A>(p.getSecond(),p.getFirst()));
        }
        return rv;
    }

    
    public static<A, B extends Comparable> Vector<Pair<B,A> > convert_reverse_sorted(Map<A,B> c, boolean asc)
    {
        return convert_reverse_sorted(c, asc, (B b1, B b2) -> b1.compareTo(b2));
    }

    public static<A, B> Vector<Pair<B,A> > convert_reverse_sorted(Map<A,B> m, boolean asc, Comparator<B> c)
    {
        Vector<Pair<B,A> > v = convert_reverse(m);
        Collections.sort(v, (Pair<B,A> a1, Pair<B,A> a2) -> c.compare(a1.getFirst(), a2.getFirst()));
        if (!asc)
        {
            Collections.reverse(v);
        }
        return v;
    }

    public static<A , B > Vector<Pair<B,A> > convert_reverse(Map<A,B> c)
    {
        Vector<Pair<B,A> > rv = new Vector<Pair<B,A> >();
        for(Map.Entry<A,B> e : c.entrySet())
        {
            rv.add(new Pair<B,A>(e.getValue(), e.getKey()));
        }
        return rv;
    }
    
    
    public Pair<A,B> set(A a, B b)
    {
        first=a; second=b;
        hash_code=-1;
        return this;
    }
    public void setFirst(A a)
    {
        first=a;
        hash_code=-1;
    }
    public void setSecond(B b)
    {
        second=b;
        hash_code=-1;
    }
    public A getFirst()
    {
        return first;
    }

    public B getSecond()
    {
        return second;
    }

    public static Pair<Integer,Integer> parseIntPair(String s)
    {
        Pair<String,String> p1 = parse(s);
        return new Pair<Integer,Integer>(Integer.parseInt(p1.getFirst()), Integer.parseInt(p1.getSecond()));
    }
    public static Pair<String,String> parse(String s)
    {
        final int sl=s.length();
        if (sl==0)
        {
            throw new RuntimeException("invalid pair string: "+s+":: 0-long");
        }
        if (s.charAt(0)!='(' || s.charAt(sl-1)!=')')
        {
            throw new RuntimeException("invalid pair string: "+s+":: missing () at start and end!");
        }
        String sp[]=s.substring(1, sl-1).split(",");
        if (sp.length!=2)
        {
            throw new RuntimeException("cannot parse pair string: "+s+" split by ',' resulted in "+sp.length+" fields, expected 2");
        }
        return new Pair<String,String>(sp[0], sp[1]);
    }
    public String toString()
    {
        return "("+first+","+second+")";
    }

    public boolean eq(Object o1, Object o2)
    {
        int num_null=((o1==null)?1:0)+((o2==null)?1:0);
        if (num_null==1) return false;
        if (num_null==2) return true;
        return o1.equals(o2);
    }
    public boolean equals(Object o)
    {
        Pair p = (Pair)o;
        return eq(p.getFirst(), first) && eq(p.getSecond(),second);
    }
    int hash_code=-1;
    

    public int hashCode()
    {
        if(hash_code>=0)return hash_code;

        return hash_code=((first==null)?0:first.hashCode()%primes[0])+(primes[2]*((second==null)?0:second.hashCode()%primes[1]));
    }
    public int compareTo(Pair o)
    {
        int d = 0;
        if (first instanceof Comparable)
        {
            d = ((Comparable)first).compareTo(((Comparable)o.first));
        }
        else
        {
            d = first.hashCode()-o.first.hashCode();
        }
        
        if(d!=0) return d;
        
        if (second instanceof Comparable)
        {
            d = ((Comparable)second).compareTo(((Comparable)o.second));
        }
        else
        {
            d = second.hashCode()-o.second.hashCode();
        }
        
        return d;
        
    }

    public static<A , B > HashMap<A, HashSet<B>> getMapOnFirst(Collection<Pair<A,B>> c)
    {
        return getMapOnFirst(c, null);
    }

    public static<A , B > HashMap<A, HashSet<B>> getMapOnFirst(Collection<Pair<A,B>> c, HashMap<A, HashSet<B>> toupdate)
    {
        HashMap<A, HashSet<B>> rv = (toupdate!=null) ? toupdate : new HashMap<>();
        for (Pair<A, B> p : c)
        {
            MapBuilder.update(rv, p.getFirst(), p.getSecond());
        }
        return rv;
    }

    

    public static<A , B > HashMap<A, HashSet<B>> getMapOnSecond(Collection<Pair<B,A>> c)
    {
        return getMapOnSecond(c, null);
    }

    public static<A , B > HashMap<A, HashSet<B>> getMapOnSecond(Collection<Pair<B,A>> c, HashMap<A, HashSet<B>> toupdate)
    {
        HashMap<A, HashSet<B>> rv = (toupdate!=null) ? toupdate : new HashMap<>();
        for (Pair<B, A> p : c)
        {
            MapBuilder.update(rv, p.getSecond(), p.getFirst());
            
        }
        return rv;
    }


    public static<A , B > HashMap<A, Integer> getCountMapOnFirst(Collection<Pair<A,B>> c)
    {
        return getCountMapOnFirst(c, null);
    }

    public static<A , B > HashMap<A, Integer> getCountMapOnFirst(Collection<Pair<A,B>> c, HashMap<A, Integer> toupdate)
    {
        HashMap<A, Integer> rv = (toupdate!=null) ? toupdate : new HashMap<>();
        for (Pair<A, B> p : c)
        {
            Integer old = rv.get(p.getFirst());
            rv.put(p.getFirst(), (old == null) ? 1 : 1+old);
        }
        return rv;
    }

    public static<A , B > HashMap<A, Integer> getCountMapOnSecond(Collection<Pair<B,A>> c)
    {
        return getCountMapOnSecond(c, null);
    }

    public static<A , B > HashMap<A, Integer> getCountMapOnSecond(Collection<Pair<B,A>> c, HashMap<A, Integer> toupdate)
    {
        HashMap<A, Integer> rv = (toupdate!=null) ? toupdate : new HashMap<>();
        for (Pair<B, A> p : c)
        {
            Integer old = rv.get(p.getSecond());
            rv.put(p.getSecond(), (old == null) ? 1 : 1+old);
        }
        return rv;
    }

    public static<A , B > HashSet<A> getSetOnFirst(Collection<Pair<A,B>> c)
    {
        HashSet<A> rv = new HashSet<>();
        for (Pair<A, B> p : c)
        {
            rv.add(p.getFirst());
        }
        return rv;
    }

    public static<A , B > HashSet<B> getSetOnSecond(Collection<Pair<A,B>> c)
    {
        HashSet<B> rv = new HashSet<>();
        for (Pair<A, B> p : c)
        {
            rv.add(p.getSecond());
        }
        return rv;
    }


    public static Vector<Pair<Integer,Integer>> getSorted(int[] counts, boolean asc)
    {
        Vector<Pair<Integer,Integer>> v = new Vector<>();
        for (int i = 0; i < counts.length; i++)
        {
            v.add(new Pair<Integer,Integer>(counts[i], i));
        }
        Collections.sort(v);
        if (!asc)
        {
            Collections.reverse(v);
        }
        return v;
    }


    public static Vector<Pair<Double,Integer>> getSortedDouble(double[] counts, boolean asc)
    {
        Vector<Pair<Double,Integer>> v = new Vector<>();
        for (int i = 0; i < counts.length; i++)
        {
            v.add(new Pair<Double,Integer>(counts[i], i));
        }
        Collections.sort(v);
        if (!asc)
        {
            Collections.reverse(v);
        }
        return v;
    }


    public static <A extends Comparable, B> void sortByFirst(Vector<Pair<A, B>> v, boolean desc)
    {
        Collections.sort(v, (p1, p2) -> p1.getFirst().compareTo(p2.getFirst()));
        if (!desc)
            return;
        
        Collections.reverse(v);
    }



    
    public static <A extends Comparable, B> void sortBySecond(Vector<Pair<B, A>> v, boolean desc)
    {
        Collections.sort(v, (p1, p2) -> p1.getSecond().compareTo(p2.getSecond()));
        if (!desc)
            return;
        
        Collections.reverse(v);
    }
}
