package lmu.utils;
import java.lang.reflect.*;
import java.util.*;
import java.util.regex.*;

import javax.xml.stream.*;

import org.jdom.Element;

/** extracts an object from another object via field access or parameter-less method invoke 
 */
public class ReflectedObject
{
    public static<A> void initer(A a, java.io.File attribs)
    {
        Class cz = a.getClass();
        Iterator<String[]> it = FileUtils.getFieldSetIterator(attribs, "\t");
        while (it.hasNext())
        {
            String[] sp = it.next();
            try
            {
                Field f = cz.getField(sp[0]);
                f.set(a, SimpleTypeConverter.getConverter(f.getType(), true).convert(sp[1]));
                System.out.printf("set %s -> %s\n", sp[0], f.get(a));
            }
            catch(NoSuchFieldException ex)
            {
                throw new RuntimeException("error at finding field: "+sp[0]+":"+ex.getMessage()+" at prepare writing attribs on class: "+cz.getCanonicalName(), ex);
            } 
            catch(IllegalAccessException aex)
            {
                throw new RuntimeException("error at accessing field: "+sp[0]+":"+aex.getMessage()+" at setting attrib "+sp[1]);
            }
        }

    }
    public static class XMLAttribHandler
    {
        Vector<Field> fields = new Vector<Field>();
        Vector<ReflectedObject> ref_objs = null;
        Vector<StringConverter> converter = new Vector<StringConverter>();
        boolean error_in_not_set=false;

        public int getNumFields()
        {
            return fields.size();
        }

        public String getFieldName(int i)
        {
            return fields.get(i).getName();
        }


        public Class getType(int i)
        {
            return fields.get(i).getType();
        }
        public ReflectedObject getRefObj(int i)
        {
            if (ref_objs!=null)
            {
                return ref_objs.get(i);
            }
            ref_objs = new Vector<ReflectedObject>();
            for (Field f : fields)
            {
                
                ref_objs.add(new ReflectedObject(f.getName()));
            }
            return ref_objs.get(i);
        }
        public XMLAttribHandler(Class clazz, boolean error_in_not_set,  String ... fields)
        {
            try
            {
            
                this.error_in_not_set=error_in_not_set;

                for (String fn : fields)
                {
                    Field f = clazz.getField(fn);
                    this.fields.add(f);
                    converter.add(SimpleTypeConverter.getConverter(f.getType(), true));
                    
                }
            }
            catch(NoSuchFieldException ex)
            {
                throw new RuntimeException("error at finding field: "+ex.getMessage()+" at prepare writing attribs on class: "+clazz.getCanonicalName(), ex);
            }
            
        }

        public void copy(Object from, Object to)
        {
            try
            {

                for (Field f : fields)
                {
                    f.set(to, f.get(from));
                }
                
            }
            catch(IllegalAccessException aex)
            {
                throw new RuntimeException("error at accessing field: "+aex.getMessage()+" at copying attribs from "+from +" "+to);
            }

        }
        public void write(Element e, Object o)
        {
            try
            {
                for (Field f : fields)
                {
                    e.setAttribute(f.getName(), ""+f.get(o));
                    
                }
            }
            catch(IllegalAccessException aex)
            {
                throw new RuntimeException("error at accessing field: "+aex.getMessage()+" at writing attribs on "+o+" into xml elem", aex);
            }
        }

        public void write(XMLStreamWriter xsw, Object o) throws XMLStreamException
        {
            try
            {
                for (Field f : fields)
                {
                    xsw.writeAttribute(f.getName(), ""+f.get(o));
                    
                }
            }
            catch(IllegalAccessException aex)
            {
                throw new RuntimeException("error at accessing field: "+aex.getMessage()+" at writing attribs on "+o+" into xml elem", aex);
            }
        }
        public void read(Element e, Object o)
        {
                    
            try
            {
            
                for (int i=0; i<fields.size(); i++)
                {
                    Field f = fields.get(i);
                    StringConverter conv = converter.get(i);
                    String value = e.getAttributeValue(f.getName());
    
                    if (error_in_not_set && value==null)
                    {
                        throw new RuntimeException("attribute for field: "+f.getName()+" not set in "+e.getName());
                    }
    
                    f.set(o, conv.convert(value));
                    
                }
            }
            catch(IllegalAccessException aex)
            {
                throw new RuntimeException("error at accessing field: "+aex.getMessage()+" at writing attribs on "+o+" into xml elem", aex);
            }
        }
        public void read(XMLStreamReader xsr, Object o)
        {
            try
            {
            
                
                for (int i=0; i<fields.size(); i++)
                {
                    Field f = fields.get(i);
                    StringConverter conv = converter.get(i);
                    String value = xsr.getAttributeValue(null, f.getName());

                    if (error_in_not_set && value==null)
                    {
                        throw new RuntimeException("attribute for field: "+f.getName()+" not set in "+xsr.getName());
                    }
                    if(value==null)
                        continue;

                    f.set(o, conv.convert(value));
                    
                }
            }
            catch(IllegalAccessException aex)
            {
                throw new RuntimeException("error at accessing field: "+aex.getMessage()+" at writing attribs on "+o+" into "+xsr.getName(), aex);
            }
        }

    }

    /*public static void readAttribs(XMLStreamReader xsr, Object o, boolean error_in_not_set,  String ... fields) throws XMLStreamException
    {
        try
        {
        
            Class clazz = o.getClass();
            for (String fn : fields)
            {
                Field f = clazz.getField(fn);
                String value = xsr.getAttributeValue(null, fn);
                
                if (error_in_not_set && value==null)
                {
                    throw new RuntimeException("attribute for field: "+fn+" not set in "+xsr.getName());
                }
                
                
                StringConverter conv = SimpleTypeConverter.getConverter(f.getType(), true);
                f.set(o, conv.convert(value));
                
            }
        }
        catch(NoSuchFieldException ex)
        {
            throw new RuntimeException("error at finding field: "+ex.getMessage()+" at writing attribs on "+o+" into "+xsr.getName(), ex);
        }
        catch(IllegalAccessException aex)
        {
            throw new RuntimeException("error at accessing field: "+aex.getMessage()+" at writing attribs on "+o+" into "+xsr.getName(), aex);
        }

    }

    public static void writeAttribs(XMLStreamWriter xsw, Object o, String ... fields) throws XMLStreamException
    {
        try
        {
            if (fields==null)
            {
                o.getClass();
            }
            
            Class clazz = o.getClass();
            
            for (String fn : fields)
            {
                Field f = clazz.getField(fn);
                xsw.writeAttribute(fn, ""+f.get(o));
                
            }
        }
        catch(NoSuchFieldException ex)
        {
            throw new RuntimeException("error at finding field: "+ex.getMessage()+" at writing attribs on "+o+" into xml elem", ex);
        }
        catch(IllegalAccessException aex)
        {
            throw new RuntimeException("error at accessing field: "+aex.getMessage()+" at writing attribs on "+o+" into xml elem", aex);
        }

    }
    public static void writeAttribs(Element e, Object o, String ... fields)
    {
        try
        {
            if (fields==null)
            {
                o.getClass();
            }
            
            Class clazz = o.getClass();
            
            for (String fn : fields)
            {
                Field f = clazz.getField(fn);
                e.setAttribute(fn, ""+f.get(o));
            }
        }
        catch(NoSuchFieldException ex)
        {
            throw new RuntimeException("error at finding field: "+ex.getMessage()+" at writing attribs on "+o+" into "+e.getName(), ex);
        }
        catch(IllegalAccessException aex)
        {
            throw new RuntimeException("error at accessing field: "+aex.getMessage()+" at writing attribs on "+o+" into "+e.getName(), aex);
        }
        
    }

    public static void readAttribs(Element e, Object o, boolean error_in_not_set,  String ... fields)
    {
        try
        {
        
            Class clazz = o.getClass();
            for (String fn : fields)
            {
                Field f = clazz.getField(fn);
                Attribute a = e.getAttribute(fn);
                if (error_in_not_set && a==null)
                {
                    throw new RuntimeException("attribute for field: "+fn+" not set in "+e.getName());
                }
                
                
                StringConverter conv = SimpleTypeConverter.getConverter(f.getType(), true);
                f.set(o, conv.convert(a.getValue()));
                
            }
        }
        catch(NoSuchFieldException ex)
        {
            throw new RuntimeException("error at finding field: "+ex.getMessage()+" at writing attribs on "+o+" into "+e.getName(), ex);
        }
        catch(IllegalAccessException aex)
        {
            throw new RuntimeException("error at accessing field: "+aex.getMessage()+" at writing attribs on "+o+" into "+e.getName(), aex);
        }

    }
    */
    static String VARP ="([a-z|A-Z|0-9|_]+)(\\(([^\\)]*)\\))?";
    static String SUBP="\\."+VARP;

    static Pattern methodP = Pattern.compile(VARP+"(("+SUBP+")*)");
    static Pattern subP = Pattern.compile(SUBP);
    static Pattern commaP = Pattern.compile("\\s*,\\s*");

    Field f;
    Method m;
    String fieldname;
    String methodname;
    String getter_methodname;
    Object getter_params[];
    boolean got_refobj_param = false;
    

    boolean inited=false;
    boolean ident=true; //if nothing is specified the ReflectedObject will return the input object (applying the sub_obj on it if needed)
    Getter getter = null;

    Object staticObject; //if specified all getter invoke will result this object

    Method getter_m;
    
    String autodetect_method=null;
    String autodetect_params[];

    public ReflectedObject sub_obj = null;

    
    void updateIdent()
    {
        ident=(fieldname==null && methodname==null && getter_methodname==null && autodetect_method==null);
        if (fieldname!=null && fieldname.equals("this"))
        {
            ident=true;
        }
        
    }
    
    static Class convertClass(Class inc)
    {
        //System.out.println("will convert class: "+inc.getName());
        if (inc == Integer.class)
        {
            int o = (int)0;
            
            
            
            return Integer.TYPE;

        }
        if(inc == Double.class)
        {
            double o = (double)0.0;
            return Double.TYPE;
        }
        if(inc == Boolean.class)
        {
            boolean o = false;
            return Boolean.TYPE;
        }
        return inc;
    }

    static HashMap<String, Class> simpleTypeName2Class = new HashMap<String, Class>();

    static
    {
        simpleTypeName2Class.put("int", Integer.class);
        simpleTypeName2Class.put("double", Double.class);
        simpleTypeName2Class.put("float", Float.class);
        simpleTypeName2Class.put("long", Long.class);
        simpleTypeName2Class.put("boolean", Boolean.class);
    }
    
    public static Class getSimpleClass(Class c)
    {
        return simpleTypeName2Class.get(c.getName());
    }

    Object convert(Class c, String p)
    {
        String cn = c.getName();
        Class simpleClass = simpleTypeName2Class.get(c.getName());
        if (simpleClass!=null)
        {
            c=simpleClass;
        }
        try
        {
            if (c == Integer.class)
                return Integer.parseInt(p);
            if (c == Double.class)
                return Double.parseDouble(p);
            
            if (c == Float.class)
                return Float.parseFloat(p);
            
            if (c == Long.class)
                return Long.parseLong(p);
            
            return p;
        }
        catch(NumberFormatException e)
        {
            return null;
        }
    }

    Method initAutodetectMethod(Object o)
    {
        //System.out.println("init autodetect with "+o);
        if (o==null)
        {
            return null;
        }
        Method[] ma = o.getClass().getMethods();
        if (ma==null)
        {
            throw new RuntimeException("cannot find method: "+ autodetect_method);
        }
        //System.out.println("found "+ma.length+" methods for "+o.getClass().getName());

        
        for (int i=0; i<ma.length; i++)
        {
            //System.out.println("check :" +ma[i].getName()+" vs "+autodetect_method);
            if(!ma[i].getName().equals(autodetect_method))
            {
                continue;
            }
            

            Class[] pt = ma[i].getParameterTypes();
            Type[] t = ma[i].getGenericParameterTypes();
            //System.out.println("detected method "+autodetect_method+" with "+t.length+" params (look for "+autodetect_params.length+")");
            
            if (t.length!=autodetect_params.length)
            {
                continue;
            }
            boolean invalid=false;
            Object mparams[]=new Object[t.length];
            for (int j=0; j<t.length; j++)
            {
                //System.out.println("\t param "+(j+1)+": "+t[j]+" te: " + pt[j].getName()+" int? : "+(pt[j] == Integer.class));
                if (null==(mparams[j]=convert(pt[j], autodetect_params[j])))
                {
                    invalid=true;
                    break;
                }
                
                
            }
            if (invalid)
                continue;
            
            
            getter_params = mparams;
            return getter_m = ma[i];
        }
        StringBuffer params = new StringBuffer();
        for (int i=0; i<autodetect_params.length; i++)
        {
            if (params.length()>0)
            {
                params.append(',');
            }
            params.append(autodetect_params[i]);
        }
        throw new RuntimeException("could not autodetect method: "+autodetect_method+"("+params.toString()+")");
    }

    static ReflectedObject create(Matcher m)
    {
        return create(m,null);
    }
    static ReflectedObject create(Matcher m, ReflectedObject ro)
    {
        return create(m,ro,false);
    }
    static ReflectedObject create(Matcher m, ReflectedObject ro, boolean verbose)
    {
        String paramstr = m.group(3);
        String params[] = (paramstr==null || paramstr.length()==0)?new String[0]:commaP.split(paramstr);
        String method = m.group(1);


        if (ro==null)
        {
            ro = new ReflectedObject();
            
        }
        if (verbose)
        {
            System.out.println("create::match: "+m.group());
        }
        

        

        if (m.group(2)==null)
        {
            ro.fieldname=method;
            if (verbose)
            {
                System.out.println("create::init with field "+ro.fieldname);
            }
            ro.updateIdent();
            return ro;
        }
        if (params.length!=0)
        {
            ro.autodetect_method=method;
            ro.autodetect_params=params;
            if (verbose)
            {
                System.out.println("create::init with autodetect method "+ro.autodetect_method+" params: "+paramstr);
            }
            ro.updateIdent();
            return ro;
        }
        ro.methodname = method;
        if (verbose)
        {
            System.out.println("create::init with method "+ro.methodname);
        }
        ro.updateIdent();
        return ro;

    }
    static ReflectedObject autoDetect(String expr)
    {
        return autoDetect(expr,null);
    }
    static ReflectedObject autoDetect(String expr, ReflectedObject rv)
    {
        return autoDetect(expr,rv,false);
    }
    static ReflectedObject autoDetect(String expr, ReflectedObject rv, boolean verbose)
    {
        if(rv==null)
        {
            rv=new ReflectedObject();
        }

        if (verbose)
        {
            System.out.println("will autodected: >"+expr+"<");
        }
        
        Matcher m = methodP.matcher(expr);
        
        if (!m.matches())
        {
            throw new RuntimeException("cannot parse: "+expr);
            
        }

        String field = m.group(1);
        if (verbose)
        {
        
            System.out.println("got field: "+field+" remaining: "+m.group(2)+" total:"+m.group());
            for(int i=0; i<m.groupCount(); i++)
            {
                System.out.println("\t"+i+"=>"+m.group(i+1));
            }
        }
        
        int start=0;
        create(m, rv, verbose);
        ReflectedObject current = rv;
        m=subP.matcher(m.group(4));
        while (m.find(start))
        {
            current.sub_obj=create(m, null, verbose);            
            if (verbose)
            {
                System.out.println("add sub: for matched part: "+m.group()+" current:"+current.getInfo());
            }

            current = current.sub_obj;

            start=m.end();
        }
        rv.updateIdent();
        if (verbose)
        {
            System.out.println("returning: "+rv.getInfo());
        }
        return rv;
    }
    private ReflectedObject()
    {
        
    }

    public interface Getter<B>
    {
        public Object get(B o);
    }
    public<B> ReflectedObject(Getter<B> g)
    {
        this.getter = g;
        inited = true;
    }
    public ReflectedObject(boolean ident)
    {
        this.ident=ident;
    }
    /** copy constructor */
    public  ReflectedObject(ReflectedObject obj)
    {
        f=obj.f; m=obj.m; 
        fieldname=obj.fieldname; methodname=obj.methodname;
        getter_methodname = obj.getter_methodname;
        getter_params = obj.getter_params;
        inited = obj.inited;
        getter_m = obj.getter_m;
        autodetect_method = obj.autodetect_method;
        autodetect_params = obj.autodetect_params;

        sub_obj = obj.sub_obj;

        updateIdent();
    }

    public ReflectedObject setSub(ReflectedObject obj)
    {
        sub_obj=obj;
        return this;
    }
    /**
     * to generate sequence of method calls, e.g. to call x.get(0).toString() do new ReflectedObject(new ReflectedObject("get(0)"),new ReflectedObject("toString()"))
     * @param obj
     * @param subobj will be applied to result of obj.get()
     */
    public  ReflectedObject(ReflectedObject obj, ReflectedObject subobj)
    {
        this(obj);
        sub_obj=subobj;
        updateIdent();
    }

    
    public String in_expr;
    public ReflectedObject(String expr)
    {
        in_expr=expr;
        if (expr!=null && expr.equals("this"))
        {
            this.ident=true;
            return;
        }
        autoDetect(expr,this);
    }
    /*public ReflectedObject(String fieldname)
    {
        this(fieldname, (String)null);
    }
    */
    public ReflectedObject(String fieldname, String methodname)
    {
        this.fieldname=fieldname;
        if(null==(this.methodname=methodname))
        {
            this.methodname="get"+Character.toUpperCase(fieldname.charAt(0))+fieldname.substring(1);
        }
        updateIdent();
    }
    public ReflectedObject(String methodname, Object  ... getter_params)
    {
        getter_methodname = methodname;
        this.getter_params = getter_params;
        updateIdent();
    }

    public String toString()
    {
        return String.format("ref object field: %s method: %s expression: %s inited: %s", fieldname, methodname, in_expr,""+inited);
    }
    void init(Object o)
    {
                
        if (inited || o==null)
        {
            return;
        }

        if (ident)
        {
            inited=true;
            return;
        }
        Object inObject=o;
        if (staticObject!=null)
        {
            o=staticObject;
        }
        //System.out.println("will init with: "+o);
        Exception e1 = null;
        if (fieldname!=null)
        {
        
            try
            {
                f = o.getClass().getField(fieldname);
                inited=true;
                return;
            }
            catch(Exception e)
            {
                e1=e;
            }
        }
        if (methodname!=null)
        {
            //System.out.println("init with: method: "+methodname);
            try
            {
                m= o.getClass().getMethod(methodname);
                inited=true;
                return;
            }
            catch(Exception e)
            {
                throw new RuntimeException("could not get field "+fieldname+" or method " + methodname+  " for class "+ o.getClass().getName()+" error for field: "
                                           + ((e1==null)?"none":e1.getMessage())+" error for method: "+e.getMessage(), e);
            }
        }

        if (getter_methodname!=null)
        {

            //System.out.println("init with: getter: "+getter_methodname);
            try
            {
                if(getter_params==null)
                {
                    getter_m = o.getClass().getMethod(getter_methodname);
                    inited=true;
                    return;
                }
                Class[] param_types = new Class[getter_params.length];
                StringBuffer classtypes = new StringBuffer();
                for (int i=0; i<getter_params.length; i++)
                {
                    classtypes.append(' ');
                    Object param=getter_params[i];
                    if (param instanceof ReflectedObject )
                    {
                        ReflectedObject paramObj = (ReflectedObject)param;
                        //System.out.printf("arg: "+i+" will get param "+paramObj.getUnInitedInfo());
                        param = paramObj.get(inObject);
                        //System.out.println("param objecT: "+param);
                        got_refobj_param=true;
                    }
                    param_types[i]=(param==null)?null:convertClass(param.getClass());
                    classtypes.append((param==null)?null:param_types[i].getName());
                }
                

                try
                {
                    getter_m = o.getClass().getMethod(getter_methodname, param_types);
                }
                catch(NoSuchMethodException nsme)
                {
                    System.out.println("will get "+getter_methodname+" on "+o.getClass().getName()+" params: "+classtypes);
                    for (Method m : o.getClass().getMethods())
                    {
                        if (!m.getName().equals(getter_methodname))
                        {
                            continue;
                        }
                        Class[] params = m.getParameterTypes();
                        if (params.length!=param_types.length)
                        {
                            continue;
                        }
                        boolean valid=true;
                        System.out.printf("check %s\n",m.getName());
                        for (int pi=0; valid && pi<params.length; pi++)
                        {
                            if (param_types[pi]==null)
                            {
                                continue;
                            }
                            if (params[pi] == Object.class)
                            {
                                valid = true;
                                break;
                            }
                            valid=params[pi].isAssignableFrom(param_types[pi]) ;

                            System.out.printf("\t[%d] %s vs %s eq: %s assignable: %s\n", pi, param_types[pi].getCanonicalName(),
                                              params[pi].getCanonicalName(), param_types[pi].equals(params[pi]), params[pi].isAssignableFrom(param_types[pi]));
                        }
                        if (valid)
                        {
                            getter_m=m;
                            break;
                        }

                    }

                }
                if (getter_m==null)
                {
                    getter_m = o.getClass().getMethod(getter_methodname, param_types);
                }
                
                inited=true;
                return;
            
                
            }
            catch(Exception e)
            {
                throw new RuntimeException("could not get getter method "+getter_methodname +" for class "+ o.getClass().getName()+" error: "+e.getMessage(), e);
            }
        }
        if (autodetect_method!=null)
        {
            //System.out.println("init with: autodetec method: "+autodetect_method);
            try
            {
            
                getter_m = initAutodetectMethod(o);
                inited = true;
                return;
            }
            catch(Exception e)
            {
                throw new RuntimeException("could not autodetect method "+autodetect_method +" on " +o.getClass().getName()+" with "+autodetect_params.length+" params !");
            }
        }
        

        throw new RuntimeException("could not init "+this+" on "+o);
        
    }
            
    public String getUnInitedInfo()
    {
        String prefix=null;;
        if (fieldname!=null)
        {
                prefix=" (field: "+fieldname+ " )";
        }
        if (prefix==null && methodname !=null)
        {
            prefix=" (method: "+methodname+ " )";
        }
        if (prefix==null && getter_methodname!=null)
        {
            StringBuffer gp = new StringBuffer();
            for (int i=0; i<getter_params.length; i++)
            {
                if (gp.length()>0)
                {
                    gp.append(' ');
                }
                
                gp.append(""+getter_params[i]);
            }
            prefix=" (method: "+getter_methodname+" params: "+gp.toString()+ ")";
        }
        if (prefix==null && autodetect_method!=null)
        {
            StringBuffer gp = new StringBuffer();
            for (int i=0; i<autodetect_params.length; i++)
            {
                if (gp.length()>0)
                {
                    gp.append(' ');
                }
                
                gp.append(""+autodetect_params[i]);
            }
            prefix=" (autodetect method: "+autodetect_method+" params: "+gp.toString()+ ")";
        }
        return "<uninited>: "+prefix;
    }

    public String getInfo()
    {
        String prefix="";

        if (!inited)
        {
            prefix=getUnInitedInfo();
        }
        else
        {
            
            
            if (f!=null)
            {
                prefix=" (field: "+f.getName() + " )";
            }
            if (prefix==null && m !=null)
            {
                prefix=" (method: "+m.getName()+ " )";
            }
            if (prefix==null && getter_m!=null)
            {
                StringBuffer gp = new StringBuffer();
                for (int i=0; i<getter_params.length; i++)
                {
                    if (gp.length()>0)
                    {
                        gp.append(' ');
                    }
                    
                    gp.append(""+getter_params[i]);
                }
                prefix=" (method: "+getter_m.getName()+" params: "+gp.toString()+ ")";
            }
            
        }
        if (sub_obj==null)
        {
            return prefix;
        }
            
        return "ident: "+ident+" "+prefix+" :: "+sub_obj.getInfo();
        
    }
    /** returns the expression value of the reflected object 
     * on the object or null if somewhere in the evlaution 
     * encounters a null value. if null values should lead 
     * to exceptions invoke get(o, false);
     */
    public Object get(Object o)
    {
        return get(o,true);
    }
    public Object get(Object o, boolean tolerate_null)
    {
        if (getter!=null)
            return getter.get(o);
        
        
        if (!inited)
        {
            init(o);
        }

        
        Object inObject = o;

        if (staticObject!=null)
        {
            o=staticObject;
        }

        //System.out.println("will invoke on "+o);
        Object fo = null;
                     
        try
        {
            if (f!=null)
            {
                //System.out.println("got field: "+f.getName());
                fo = f.get(o);
            }
    
            if (m!=null)
            {
                //System.out.println("got method: "+m.getName());
                fo=m.invoke(o);
            }
            if (getter_m != null)
            {
                /*System.out.println("got getter: "+getter_m.getName()+" on >"+o+"< class: "+o.getClass()+" getter params: "+getter_params.length);
                for (int i=0; i<getter_params.length; i++)
                {
                    System.out.printf("\tgetter: %d %s param class: %s\n", i, getter_params[i], getter_params[i].getClass());
                }
                */
                
                if(!got_refobj_param)
                {
                    
                    /*
                    System.out.println("got !got_refobj_param");
                    if (getter_params.length==1 && getter_params[0].getClass()==String.class)
                    {
                        
                        System.out.println("invoke >"+o+"<."+getter_m.getName()+"(>"+getter_params[0]+"<) ");
                        System.out.println("single param class string?: "+(getter_params[0].getClass()==String.class));
                        Object to[]=new Object[1];

                        
                        to[0]=new String(""+getter_params[0]);
                        try
                        {

                            System.out.println("test invoke: "+getter_m.invoke(o,to));
                            
                        }
                        catch(Exception e)
                        {
                            System.out.println("invoke "+getter_m.getName()+" on itself: "+o+" lead to error: "+e.getMessage());
                        }
                        
                    }
                    */
                    
                    
                    fo = getter_m.invoke(o, getter_params);
                    //System.out.println("result: "+fo);
                }
                else
                {
                    //System.out.println("calc real params");
                    Object real_params[] = new Object[getter_params.length];
                    for (int i=0; i<getter_params.length; i++)
                    {
                        if (! (getter_params[i] instanceof ReflectedObject) )
                        {
                            real_params[i]=getter_params[i];
                            continue;
                        }

                        ReflectedObject ro = (ReflectedObject)getter_params[i];
                        //System.out.println("will invoke param ro: "+ro.getUnInitedInfo()+" with object: "+inObject);
                        real_params[i]=ro.get(inObject);
                    }

                    fo=getter_m.invoke(o, real_params);
                }
                
                
            }
        }
        catch(Exception e)
        {
            throw new RuntimeException(e);
        }
        
        if (ident)
        {
            fo=o;
        }

        if (tolerate_null && fo==null)
        {
            return null;
        }
        if (sub_obj!=null)
        {
            return sub_obj.get(fo);
        }
        return fo;
    }

    
    public static class TestC
    {
        public HashSet<String> ts = new HashSet<String>();
        public HashMap<String,Integer> tsi = new HashMap<String,Integer>();

        public TestC()
        {
            ts.add("what");
            ts.add("w2");
            ts.add("w3");
        }
        public String testT="what ever whatever";

        public int sum(Integer a, int b)
        {
            return a+b;
        }

        public double sum(double a, double b)
        {
            return a+b;
        }

        public String testS(String s, int a)
        {
            return s+":"+a;
        }

    }

    public static ReflectedObject apply(Object toapplyon, String methodname, String expr)
    {
        return apply(toapplyon,methodname, new ReflectedObject(expr));
    }
    public static ReflectedObject apply(Object toapplyon, String methodname, Object ... params)
    {
        ReflectedObject ro = new ReflectedObject(methodname, params);
        ro.staticObject=toapplyon;
        return ro;
    }
    public static class TestC2
    {
        public int index=0;
        public TestC2(int i)
        {
            index=i;
        }
    }

    public static class TestName
    {
        String name;
        public TestName(String n)
        {
            name=n;
        }
        public String getName()
        {
            return name;
        }
    }
    public static class TestVector
    {
        Vector<TestName> names = new Vector<TestName>();


        public TestName getName(int index)
        {
            return names.get(index);
        }
    }

    /** Vector<TestC2>  TestVector
     * TestVector.getName( refob(test2.index));
     */
    
    
    public static class TestS
   {
       public int n=4;
       Object o = new Object();
       public Integer getP(int p)
       {
           return n+p;
       }
   }


    static void testNESTED()
    {
        Pair<TestS,TestS> p = new Pair<TestS,TestS>(new TestS(), new TestS());
      ReflectedObject ro = new ReflectedObject(new ReflectedObject("getFirst()"), new ReflectedObject("getP", 14));
      System.out.println(ro.get(p));
    }
    
    static void testAPPLY()
    {
        HashSet<Integer> ts = new HashSet<Integer>();

          for (int i=0; i<21; i++)
          {
              if (i%2 == 1)
                  continue;
        
        
              ts.add(i);
          }
          ReflectedObject ro2 = ReflectedObject.apply(ts, "contains", "this");
          for (int i=0; i<21; i++)
          {
              System.out.println("contains: "+ i +" : "+ro2.get(i));
          }

    }

  
    public static class TestDynamic
    {
        int index;
        public TestDynamic(int index)
        {
            this.index=index;
        }
        public String get(String s)
        {
            return s+"::"+index;
        }
        public int getIndex()
        {
            return index;
        }
    }
    public static class TestFactory
    {
        Vector<TestDynamic> v = new Vector<TestDynamic>();

        public TestFactory(int num_dyn)
        {
            for (int i=0; i< num_dyn; i++)
            {
                v.add(new TestDynamic(i));
            }
        }
        
        public TestDynamic getIdx(int i)
        {
            return v.get(i);
        }
    }

    static void testStatic2Dyn()
    {
        TestFactory tf = new TestFactory(10);
        Object obj = new String("hello");
        
        ReflectedObject getter = new ReflectedObject("getIndex()").setSub(apply(tf, "getIdx", "this").setSub(new ReflectedObject("get", obj)));
        for (TestDynamic td : tf.v)
        {
            System.out.println(getter.get(td));
        }
        
        
    }

    public static void main (String args[]) 
    {
        testStatic2Dyn();
        if(true)return;
        TestVector tv = new TestVector();
        Vector<TestC2> tv2 = new Vector<TestC2>();

        for (int i=0; i<100; i++)
        {
            tv.names.add(new TestName("test"+i));
            tv2.add(new TestC2(i));
        }

        ReflectedObject ro2 = apply(tv, "getName", "index");
        ro2.setSub(new ReflectedObject("getName()"));

        for (TestC2 t : tv2)
        {
            System.out.println("test: "+t.index+" result: "+ro2.get(t));
        }

        //ReflectedObject ro = ReflectedObject.autoDetect(args[0]);

        //System.out.println("result: "+ ro.get(new TestC()));
    }
}


