package lmu.utils;

import java.util.*;
import java.util.regex.*;

public class RegexUtils
{

    static void check(Matcher m, int gi, String type)
    {
        if (gi>0 && m.groupCount()>=gi)
            return;
        
            
        throw new RuntimeException("invalid "+type+" group requested in "+m.group());
        
    }
    static RuntimeException report_error(Matcher m, int gi, String type, NumberFormatException nfe)
    {
        return new RuntimeException("cannot convert group "+gi+" >"+m.group(gi)+"< to "+type, nfe);

    }
    public static Pair<Integer,Integer> getIntPair(Matcher m, int gi1, int gi2)
    {
        return new Pair<Integer,Integer>(getInt(m,gi1),getInt(m,gi2));
    }

    public static boolean getBoolean(Matcher m, int gi)
    {
        check(m,gi,"<boolean>");
        try
        {
            return Boolean.parseBoolean(m.group(gi));
        }
        catch(NumberFormatException nfe)
        {
            throw report_error(m,gi,"boolean",nfe);
        }
    }
    public static String getString(Matcher m, int gi)
    {
        check(m,gi,"<string>");
        return m.group(gi);
    }

    public static boolean isEmpty(Matcher m, int gi)
    {
        return m.group(gi) == null;
    }

    public static double getDouble(Matcher m, int gi)
    {
        check(m,gi,"<double>");
        try
        {
            return Double.parseDouble(m.group(gi));
        }
        catch(NumberFormatException nfe)
        {
            throw report_error(m,gi,"double",nfe);
        }
    }
    public static int getInt(Matcher m, int gi)
    {
        check(m,gi,"<int>");
        try
        {
            return Integer.parseInt(m.group(gi));
        }
        catch(NumberFormatException nfe)
        {
            throw report_error(m,gi,"int",nfe);
        }
    }

    public static String[] getGroups(Matcher m)
    {
        final int N = m.groupCount();
        String[] sp = new String[N];
        for (int i=1; i<=N; i++)
        {
            sp[i-1] = m.group(i);
        }

        return sp;
    }

    public interface MatchConverter<A>
    {
        public A convert(Matcher m);
    }


    static public<A> Vector<A> findAll(Pattern p, String text, MatchConverter<A> mc)
    {
        Matcher m = p.matcher(text);
        int start = 0;
        Vector<A> v = new Vector<>();
        final int N = m.groupCount();
        while (m.find(start))
        {
            v.add(mc.convert(m));
            start = m.end();
        }
        return v;
    }
    public static Vector<String[]> findAll(Pattern p, String text)
    {
        return findAll(p,text, (Matcher m) -> getGroups(m));
    }

    public static Pattern compile(String fmt, Object ... data)
    {
        return Pattern.compile(String.format(fmt, data));
    }
}

