package lmu.utils;

import lmu.utils.*;
import java.util.*;

public class RegionMapper
{

    Vector<Region1D> regions = new Vector<Region1D>();

    Vector<Region1D<HashSet>> nonoverlapping_regions = new Vector<Region1D<HashSet>>() ;
    
    
    public Integer getFirst()
    {
        if (!compiled)
        {
            throw new RuntimeException("compile first!");
        }
        if (nonoverlapping_regions.size()==0)
            return null;
                       
        return nonoverlapping_regions.get(0).getX1();
            
        
    }

    public Integer getLast()
    {
        if (!compiled)
        {
            throw new RuntimeException("compile first!");
        }
        if (nonoverlapping_regions.size()==0)
            return null;

        return nonoverlapping_regions.get(nonoverlapping_regions.size()-1).getX2();
    }
    public RegionMapper()
    {

    }
    public RegionMapper(RegionMapper rm)
    {
        regions.addAll(rm.regions);

    }
    boolean compiled=false;

    enum RegTagType
    {
        OPEN,
        CLOSE;
    }

    class RegTag implements Comparable<RegTag>
    {
        RegTagType _type;
        Region1D region;
        int p;

        public RegTag(RegTagType _t, Region1D r)
        {
            region=r; _type=_t;
            switch(_t)
            {
                case OPEN:
                    p=r.getX1();
                    break;
                case CLOSE:
                    p=r.getX2();
                    break;
            }
        }

        public String toString()
        {
            return String.format("%s:%d %s", ""+_type, p, region);
        }
        public int compareTo(RegTag rt)
        {
            int d = p-rt.p;
            if (d!=0)
                return d;
            
            return (_type==RegTagType.CLOSE)?-1:1;
            
        }
    }
    
    class RegionComparer implements java.util.Comparator<Region1D>
    {
        boolean x1=true;

        RegionComparer(boolean firstx)
        {
            x1=firstx;
        }
        public int compare(Region1D o1, Region1D o2)
        {
            return (x1)?o1.getX1()-o2.getX1():o1.getX2()-o2.getX2();
        }

        public boolean equals(Region1D obj)
        {
            return false;
        }

        public int getX(Region1D obj)
        {
            return (x1)?obj.getX1():obj.getX2();
        }

    }

    RegionComparer x1_sorter = new RegionComparer(true);
    RegionComparer x2_sorter = new RegionComparer(false);

    public void addRegions(Collection<Region1D>  regions)
    {
        for (Region1D r : regions)
        {
            this.regions.add(r);
        }
        compiled=false;
    }
    public void addRegion(Region1D r)
    {
        regions.add(r);
        compiled=false;
    }

    class OpenRegionHandler
    {
        Vector<Region1D<HashSet<Region1D> > > regs = new Vector<Region1D<HashSet<Region1D> > >();
        HashSet<Region1D> open_regions = new HashSet<Region1D>();

        int start;
        //int current_end=-1;
        

        /** closes a current region if neccesarry
         */
        public void closeRegion(RegTag rt)
        {
            if (open_regions.size()==0)
                return;
            
            
            addCurrentRegion(rt.p);

            open_regions.remove(rt.region);
                
            start=rt.p;
        }

        void addCurrentRegion(int end)
        {
            if (start==end)
                return;
            HashSet<Region1D> regset = new HashSet<Region1D>();
            regset.addAll(open_regions);
            regs.add(new Region1D(start, end, regset));
                //current_end=end;
            //System.out.printf("\t add region: %d-%d (%d overlapping regions)\n", start, end, regset.size());

            
            
        }
        public void openRegion(RegTag rt)
        {

            
            if (open_regions.size()!=0)
            {
                addCurrentRegion(rt.p);
            }
            start=rt.p;
            open_regions.add(rt.region);
        }
    }
    public Vector<Vector<Region1D>> annotate(Vector<Region1D> regions)
    {
        return annotate(regions,null);
    }
    Vector<Region1D> convert(Vector<Region1D<HashSet>> v)
    {
        Vector<Region1D> cv = new Vector<Region1D>();
        for (Region1D<HashSet> r : v)
        {
            cv.add((Region1D)r);
        }
        return cv;
    }
    public Vector<Vector<Region1D>> annotate(Vector<Region1D> regions, OverlapDetector.AnnotOverlapFilter of)
    {
        //Vector<Region1D> v = compile();
        return OverlapDetector.annotateOverlapping(regions, convert(compile()), of);
    }
    public synchronized Vector<Region1D<HashSet>>  compile()
    {
        if(compiled) return nonoverlapping_regions;

        if (regions.size()==0)
        {
            compiled=true;
            return nonoverlapping_regions;
        }
        
        nonoverlapping_regions.clear();

        //nonoverlapping_regions = (Vector<Region1D<HashSet<Region1D>>> )
        Vector<Region1D > regregs = new Vector<Region1D >();
        for (Region1D r: regions)
        {
            regregs.add(new Region1D(r.getX1(), r.getX2(),r));
        }
        nonoverlapping_regions=OverlapDetector.getOverlaps(regregs);

        
        
        /*Vector<RegTag> regtags = new Vector<RegTag>();

        for (Region1D r : regions)
        {
            regtags.add(new RegTag(RegTagType.OPEN, r));
            regtags.add(new RegTag(RegTagType.CLOSE, r));
        }
        System.out.println("sorting "+regtags.size()+" regtags");

        Collections.sort(regtags);

        System.out.println("start building overlapping regions");
        OpenRegionHandler orh = new OpenRegionHandler();

        for (RegTag rt : regtags)
        {
            //System.out.println("next rt: "+rt);
            switch (rt._type)
            {
                case OPEN:
                    orh.openRegion(rt);
                    break;
                case CLOSE:
                    orh.closeRegion(rt);
                    break;
            }
        }
        System.out.println("got "+orh.regs.size()+" nonoverlapping regions");
        nonoverlapping_regions=orh.regs;
        */
        compiled=true;
        
        return nonoverlapping_regions;
    }

    public HashSet<Region1D> getOverlapping(int from, int to)
    {
        HashSet<Region1D> rv = new HashSet<Region1D>();

        Vector<Region1D<HashSet>> v = nonoverlapping_regions;

        if (v.size()==0)
        {
            return rv;
        }
        int start=getFirstNotSmallerV(from, nonoverlapping_regions, false);
        System.out.printf("look for %d got first after: %d all: %d \n", from, start, nonoverlapping_regions.size());
        while (start>0 && v.get(start-1).getX2()>from)
        {
        }
        if (start>=0 && start<nonoverlapping_regions.size())
        {
            System.out.println("got to start: "+nonoverlapping_regions.get(start));
        }


        int end=getFirstNotSmallerV(to, nonoverlapping_regions, false);
        System.out.printf("look for %d got first after: %d \n", to, end);
        if (end>=0 && end<nonoverlapping_regions.size())
        {
            System.out.println("got to start: "+nonoverlapping_regions.get(end));
        }
        for (int i=start; i<=end; i++)
        {
            Region1D<HashSet> r = v.get(i);
            if (r.getX2()>from && r.getX1()<to)
            {
                rv.addAll((HashSet<Region1D>)r.getObject());
            }
        }
        return rv;
        /*
        while (st)
        {
        }
        HashSet<Region1D> r1 = getMatching(from);
        if (r1==null)
            return null;
        
        HashSet<Region1D> r2 = getMatching(to);
        if (r2==null)
            return null;
        
        HashSet<Region1D> rv = new HashSet<Region1D>();

        for (Region1D r: r1)
        {
            if (!r2.contains(r))
                continue;
            
            rv.add(r);
        }
        return rv;
        */
    }
    public HashSet<Region1D> getMatching(int x)
    {
        return getMatching(x,false);
    }
    

    public static int getFirstNotSmaller(int x, Vector<Region1D> v)
    {
        return getFirstNotSmaller(x,v,false);
    }
    
    public static int getFirstNotSmaller(int x, Vector<Region1D> v, boolean debug)
    {
        return getFirstNotSmallerV(x,v,debug);
    }
    

    

    public static int getFirstNotSmallerV(int x, Vector iv, boolean debug)
    {
        int lower=0;
    
        Vector<Region1D> v = (Vector<Region1D>)iv;
        

        int upper=v.size()-1;
       
        int c=0;
        int mi=-1;
        while (mi<0)
        {
            if (lower>upper)
            {
                if (debug)
                {
                    System.out.printf("\tlower>upper, break\n");
                }
                return lower;
                //break;
            }
            if (++c>v.size())
            {
                throw new RuntimeException("ERROR");
            }
            int middle = (int)(0.5*(lower+upper));
            Region1D r = v.get(middle);
            if (debug)
            {
                System.out.printf("\t x: %d binsearch next:  l: %d u: %d m: %d r: %s\n", x, lower, upper, middle, r);
            }

            if (r.contains(x))
            {
                    
                if (debug)
                {
                    System.out.printf("\t found! got "+v.get(middle).getObject()+" regions\n");
                }
                mi=middle;
                break;
            }
            
            if (r.getX1()>x)
            {
                upper=middle-1;
                continue;
            }
             
            lower=middle+1;
             
            
                
        }
        if (mi<0)
            return -1;

        return mi;
    }
    public int getFirstNotSmaller(int x, boolean debug)
    {
        return getFirstNotSmallerV(x,nonoverlapping_regions, debug);
        
    }

    public Vector<Region1D> getAllOverlapping(int x, int y, boolean verbose)
    {
        final Vector<Region1D<HashSet>> v = nonoverlapping_regions;
        
        Vector<Region1D> rv = new Vector<Region1D>();
        HashSet<Region1D> visited = new HashSet<Region1D>();

        int first=getFirstNotSmaller(x, verbose);

        if (verbose)
        {
            if(first<=0)
            {
                System.out.printf("request: %d is smaller then first %d-%d\n", x, v.get(0).getX1(),v.get(0).getX2());
            }
            else
            {
                System.out.printf("request: %d: before: %d-%d first not smaller: %d-%d\n", x, 
                                  v.get(first-1).getX1(),v.get(first-1).getX2(),
                                  v.get(first).getX1(),v.get(first).getX2());
            }
        }
        
        first=Math.max(0,first);
        for (int i=first; i<v.size(); i++ )
        {
            Region1D<HashSet> r = v.get(i);

            if (r.getX1()>y)
            {
                break;
            }
            for (Object o : r.getObject())
            {
                Region1D subr = (Region1D)o;
                if (visited.contains(subr))
                    continue;
                
                if (verbose)
                {
                    System.out.printf("found overlap to %d-%d : %d-%d: %s\n", x ,y, subr.getX1(), subr.getX2(), ""+subr.getObject());
                }
                

                visited.add(subr);
                rv.add(subr);
            }
        }
        

        return rv;
    }
        
    public HashSet<Region1D> getMatching(int x, boolean debug)
    {
        if (debug)
        {
            System.out.printf("start search for %d v: %d\n", x , nonoverlapping_regions.size());
        }
        
        int mi = getFirstNotSmaller(x,debug);

        if (mi<0 || mi>=nonoverlapping_regions.size())
        {
            return null;
        }
            
        
        
        return nonoverlapping_regions.get(mi).getObject();
        
        

    }

    public void getMatchedRegions(int x1, int x2,  HashSet<Region1D> matching_regions)
    {
        compile();


        if (regions.size()==0)
        {
            return;
        }

        
        //getMatchedRegions(x1,matching_regions);
        //getMatchedRegions(x2,matching_regions);
        
    }
    public HashSet<Region1D> getMatchedRegions(int x)
    {
        return getMatchedRegions(x,false);
    }
    public HashSet<Region1D> getMatchedRegions(int x, boolean debug)
    {
        compile();

        return getMatching(x,debug);
        
        
        /*int lower=getMatching(x,true, debug);
        if (lower<0)
            return;
        
        
        int upper=getMatching(x,false, debug);

        for (int i=lower; i<=upper; i++)
        {
            matching_regions.add(regions.get(i));
        }
        */
    }
    

    static class Test
    {
        int num_regions;
        int length=0;
        int num_tests = 0;

        RegionMapper rm = new RegionMapper();
         
        Vector<Region1D> rlist = new Vector<Region1D>();

        long d1=0;
        long d2=0;


        HashSet<Region1D> naive_res = new HashSet<Region1D>();
        HashSet<Region1D> bs_res = new HashSet<Region1D>();

        public Test(int nr, int length, int nt)
        {
            System.out.printf("test with %d regions l: %d nt: %d\n", nr, length, nt);
            num_regions=nr;
            this.length=length;
            this.num_tests=nt;

            int maxstart=length-10;
            for (int i=0; i<num_regions; i++)
            {
                int start= (int)(Math.random()*maxstart);
                int maxl = length-start-1;

                int rlength = Math.min(10000, Math.max(10, (int)(Math.random()*maxl)));

                Region1D rd = new Region1D(start, start+rlength, this);
                rm.addRegion(rd);
                rlist.add(rd);
            }
        }

        public void naive(int x)
        {
            naive_res.clear();
            for (Region1D r : rlist)
            {
                if (!r.contains(x))
                    continue;
                
                naive_res.add(r);
            }
        }

       public boolean check(int x)
       {
           HashSet<Region1D> all = new HashSet<Region1D>();
           
           if (naive_res.size()==0 && bs_res==null)
           {
               return true;
           }


           if (bs_res==null)
           {
               bs_res= new HashSet<Region1D>();
               
           }
           
           
           all.addAll(naive_res);
           if (all.size()==bs_res.size() && all.size()==naive_res.size())
           {
               return true;
           }
           System.err.println("error at test: "+x+" naive: "+naive_res.size()+" bs size: "+bs_res.size());

           
           bs_res = rm.getMatchedRegions(x, true);
           for (Region1D r : all)
           {
               if (naive_res.contains(r) && bs_res.contains(r))
               {
                   continue;
               }
               System.err.println(r+" found in naive: "+naive_res.contains(r)+" in bs: "+bs_res.contains(r));
           }
           return false;
       }
        public void test()
        {

            /*HashSet<Integer> posset = new HashSet<Integer>();
            for (Region1D r : rlist)
            {
                for (int i=r.getX1(); i<r.getX2(); i++)
                {
                    posset.add(i);
                }
            }
            
            Vector<Integer> poslist =new Vector<Integer>();
            poslist.addAll(posset);
            Collections.sort(poslist);
            System.out.println(rlist.size()+" regions induced "+poslist.size()+" positions from "+length);
            */
            long s,e;
            
            int num_pos=0;
            int num_neg=0;

            s=new java.util.Date().getTime();
            rm.compile();
            e=new java.util.Date().getTime();
            long comptime=(e-s);
            System.out.printf("compile took: %2.3f s\n", (e-s)/1000.0);
            for (int i=0; i<num_tests; i++)
            {
                int testp = (int)(Math.random()*length);
                s=new java.util.Date().getTime();
                naive(testp);
                e=new java.util.Date().getTime();
                d1+=(e-s);

                
                
                s=new java.util.Date().getTime();
                //System.out.println("start bs");
                try
                {
                    bs_res = rm.getMatchedRegions(testp, false);
                }
                catch(Exception ex)
                {
                    bs_res = rm.getMatchedRegions(testp, true);
                }
                
                //System.out.println("end bs"); System.out.flush();
                e=new java.util.Date().getTime();
                d2+=(e-s);

                //check
                if (!check(testp))
                {
                    throw new RuntimeException("error in test: "+(i+1)+"/"+num_tests+"!");
                }
                if (bs_res==null)
                {
                    num_neg++;
                }
                else
                {
                    num_pos++;
                }
                if ((i+1)%100==0)
                {
                    System.err.printf("test: %d/%d passed...\n", i+1, num_tests);
                    System.out.printf("%d pos tests, %d neg tests runtime naive: %2.3f ms bs: %2.3f ms (%2.3f)\n", num_pos, num_neg, d1/1000.0, d2/1000.0, (d2+comptime)/1000.0);
                }
                
            }
            System.out.printf("%d pos tests, %d neg tests runtime naive: %2.3f ms bs: %2.3f ms (%2.3f)\n", num_pos, num_neg, d1/1000.0, d2/1000.0, (d2+comptime)/1000.0);
        }
    }
    public static void main (String args[]) 
    {
        int nr=100;
        int length=1000;
        int num_tests = 100;


        for (int i=0; i<args.length; i++)
        {
            if (args[i].equals("-nr"))
            {
                nr=Integer.parseInt(args[++i]);
                continue;
            }
            if (args[i].equals("-l"))
            {
                length=Integer.parseInt(args[++i]);
                continue;
            }
            if (args[i].equals("-tests"))
            {
                num_tests=Integer.parseInt(args[++i]);
                continue;
            }
        }
        Test tr = new Test(nr, length, num_tests);
        tr.test();
    }
}
