package lmu.utils;

import lmu.utils.tuple.*;
import java.util.*;

import javax.xml.stream.*;

import org.jdom.*;

public class RegionVector extends Region1D implements Comparable 
{
    public int index=-1;

    static FileUtils.StringSplitter regSplitter = FileUtils.getRecordSplitter("|");
    static FileUtils.StringSplitter XSplitter = FileUtils.getRecordSplitter("-");

    public void setIndex(int index)
    {
        this.index=index;
    }
    public int getIndex()
    {
        return index;
    }
    public final static String ELEM_NAME="regvec";
    Object vec_obj;

    public RegionVector()
    {
        super(0,1, null);
        setObject(this);
    }

    public RegionVector(int start, int end)
    {
        super(start, end, null);
        addRegion(start, end);
        setObject(this);
    }
    public void initFromXML(Element e)
    {
        index = XMLUtils.getIntAttrib(e, "index", -1);
        for(Element r1e : XMLUtils.getChildren(e, Region1D.ELEM_NAME))
        {
            addRegion(new Region1D(r1e));
        }
    }
    public RegionVector(Element e)
    {
        this();
        initFromXML(e);
    }
    public List<Region1D> getIntersectingRegions(Region1D other)
    {
        return getIntersectingRegions(other,false);
    }
   
    public List<Region1D> getIntersectingRegions(Region1D other, boolean trim)
    {
    	List<Region1D> ret = new ArrayList<Region1D>();
    	for(Region1D region : this.regions)
    	{
    		if ( other.getOverlap(region) > 0) 
    		{
    		    if ( trim)
    		        ret.add( region.getIntersectedRegion(other));
    		    else
    		        ret.add(region);
    		}
    	}
    	return ret;
    }
    public RegionVector(XMLStreamReader xsr) throws XMLStreamException
    {
        super(0,1, null);
        String en = xsr.getLocalName();
        index = XMLUtils.getIntAttrib(xsr, "index", -1);
        while (xsr.hasNext())
        {
            int event = xsr.next();
            if (event == XMLStreamConstants.END_ELEMENT && xsr.getLocalName().equals(en))
            {
                return;
            }
            if (event != XMLStreamConstants.START_ELEMENT)
                continue;
            
            
            if (xsr.getLocalName().equals(Region1D.ELEM_NAME))
            {
                addRegion(new Region1D(xsr));
            }
        }
    }

    
    public RegionVector(Vector<Pair<Integer,Integer>> regions)
    {
        super(0,1,null);
        for (Pair<Integer,Integer> p : regions)
        {
            addRegion(new Region1D(p.getFirst(),p.getSecond(), null));
        }
    }

    public RegionVector(Vector<Region1D> regions, boolean from_regions)
    {
        super(0,1,null);
        ObjectGetter.apply(regions, (_r) -> addRegion(_r.getX1(), _r.getX2(), null));
    }
    
    public String getSimpleRepresentation()
    {
        if (regions.size() == 0)
            return "";
        
        
        BufferPrintWriter bpw = new BufferPrintWriter();
        bpw.printf("%d-%d", regions.get(0).getX1(),  regions.get(0).getX2());
        for (int i=1; i<regions.size(); i++)
        {
            bpw.printf("|%d-%d", regions.get(i).getX1(),  regions.get(i).getX2());
        }
        return bpw.getBuffer();

    }

    public static RegionVector parseSimpleRepresentation(String s)
    {
        RegionVector rv = new RegionVector();

        for(String rs : regSplitter.split(s))
        {
            String[] sp = XSplitter.split(rs);
            if (sp.length != 2)
                throw new FRuntimeException("invalid region spec: %s in regvec simple representation: %s", rs, s);
            
            try
            {
                rv.addRegion(new Region1D(Integer.parseInt(sp[0]), Integer.parseInt(sp[1])));
            }
            catch(NumberFormatException nfe)
            {
                throw new FRuntimeException("invalid region spec: %s in regvec simple representation: %s", rs, s);
            }

            
        }
        return rv;
    }

    public boolean containsRV(RegionVector rv)
    {
        for (Region1D r : rv.getRegions())
        {
            if(!contains(r))
                return false;
        }
        return true;

    }
    public boolean containsInRegion(int pos)
    {
        for (Region1D r : this.getRegions())
        {
            if ( r.contains(pos))
                return true;
        }
        return false;
    }
    
    
    public boolean contains(Region1D r)
    {
        if (r instanceof RegionVector)
            return containsRV((RegionVector)r);
        
        
        for (int i=0; i<getNumRegions(); i++)
        {
            if (!getRegion(i).contains(r))
                continue;
            return true;
        }
        return false;
    }

    public boolean isSubVector(RegionVector rv)
    {
        if (rv.getNumRegions()==0)
            return true;
        
        int idx = 0;
        int startmatch = -1;
        
        for (int i=0; i<getNumRegions(); i++)
        {
            if (!getRegion(i).contains(rv.getRegion(idx)))
                continue;
            
            startmatch = i;
            break;
        }
        
        if (startmatch < 0 || rv.getRegion(idx).getX2() != getRegion(startmatch).getX2())
            return false;
        
        if (getNumRegions()-startmatch < rv.getNumRegions())
            return false;
        
        
        for (int i=1, j=startmatch; i<rv.getNumRegions()-1; i++, j++)
        {
            if (!getRegion(j).equals(rv.getRegion(i)))
                return false;
            
        }
        Region1D last1 = rv.getRegion(rv.getNumRegions()-1);
        Region1D last2 = getRegion(startmatch+rv.getNumRegions()-1);

        return last2.contains(last1) && (startmatch == 0 ||  (last1.getX1() == last2.getX1()));
        

    }

    public boolean isSubVector2(RegionVector rv)
    {
        if (rv.getNumRegions()==0)
            return true;

        int idx = 0;
        int startmatch = -1;

        for (int i=0; i<getNumRegions(); i++)
        {
            if (!getRegion(i).contains(rv.getRegion(idx)))
                continue;

            startmatch = i;
            break;
        }

        System.out.printf("IS2:\n%s\n%s\n", this, rv);
        System.out.printf("startmatch: %d\n", startmatch);
        if (startmatch < 0 || rv.getRegion(idx).getX2() != getRegion(startmatch).getX2())
            return false;

        if (getNumRegions()-startmatch < rv.getNumRegions())
            return false;

        
        for (int i=1, j=startmatch; i<rv.getNumRegions()-1; i++, j++)
        {
            System.out.printf("%d, %d check %s vs %s\n", j, i, getRegion(j), rv.getRegion(i));
            if (!getRegion(j).equals(rv.getRegion(i)))
                return false;

        }
        Region1D last1 = rv.getRegion(rv.getNumRegions()-1);
        Region1D last2 = getRegion(startmatch + rv.getNumRegions()-1);

        return last2.contains(last1) && (startmatch == 0 || (last1.getX1() == last2.getX1()));


    }


    public RegionVector translate(int offset)
    {
        return translate(this, offset);
    }

    public static RegionVector translate(RegionVector rv, int offset)
    {
        RegionVector rrv = new RegionVector();
        rrv.setObject(rv.getObject());
        for (Region1D r : rv.getRegions())
        {
            rrv.addRegion(new Region1D(r.getX1()+offset, r.getX2()+offset, r.getObject()));
        }
        return rrv;
    }
    public int size()
    {
        return regions.size();
    }

    public Region1D get(int i)
    {
        return regions.get(i);
    }

    public void writeXML(XMLStreamWriter xsw) throws XMLStreamException
    {
        writeXML(xsw, null);
    }
    public void writeXML(XMLStreamWriter xsw, String elem_name) throws XMLStreamException
    {
        if (elem_name==null)
        {
            elem_name = ELEM_NAME;
        }
        xsw.writeStartElement(elem_name);
        xsw.writeAttribute("index", ""+index);
        for (Region1D r :regions)
        {
            r.writeXML(xsw);
            
        }
        xsw.writeEndElement();
    }
    public Element getXML()
    {
        return getXML(ELEM_NAME);
    }
    public Element getXML(String elem_name)
    {
        Element e = new Element(elem_name);
        e.setAttribute("index", ""+index);
        for (Region1D r :regions)
        {
            e.addContent(r.getXML());
        }

        return e;
    }
    public RegionVector(Object o)
    {
        this();
        vec_obj=o;
    }
    public RegionVector(Region1D r)
    {
        this();
        this.addRegion(r);
    }
    public RegionVector(RegionVector rv)
    {
        this();
        this.addRegions(rv.getRegions());
    }
    public void addRegions(Collection<Region1D> c)
    {
        for (Region1D r1d: c)
        {
            addRegion(r1d);
        }
    }

    public boolean isIn(Integer pos) {
        for(Region1D r : regions)
        {
            if ( r.contains(pos))
                return true;
        }
        return false;
    }
    public boolean isIn(int x1, int x2)
    {
        Pair<Boolean, Integer> p = VectorUtils.binarySearch(regions, x1, (Region1D r, Integer i) -> r.getX1() - i);

        return false;
    }
                     
    public boolean isIn(Region1D r)
    {
        return isIn(r.getX1(), r.getX2());
    }
    public Region1D addRegion(int x1, int x2)
    {
        return addRegion(x1,x2,null);

    }
    public Region1D addRegion(int x1, int x2, Object o)
    {
        return addRegion(new Region1D(x1,x2,o));
    }
    
    public double getCoverage(int[] coverage)
    {
        double cov = 0.0;
        double covlength=0;
        final int N=coverage.length;
        for (Region1D r : getRegions())
        {
            for (int i=r.getX1(); i<r.getX2() && i<N; i++)
            {
                cov+=coverage[i];
            }
            covlength+=r.getLength();
        }
        cov/=covlength;
        return cov;


    }
    public Region1D addRegion(Region1D r)
    {
        if (regions.size()==0)
        {
            setX1(r.getX1());
        }
        else 
        {
            if (r.getX1() < getX2())
            {
                throw new RuntimeException("cannot add overlapping region: "+r+" to: "+this);
            }
        }
        
        setX2(r.getX2());
        regions.add(r);
        _hashcode = -1;
        return r;
    }

    /** returns the inverted region pairs: the region vector defined by the non-covered part of the region vector*/
    public RegionVector getInverseRegionVector()
    {
        RegionVector rv = new RegionVector();
        for (int i=1; i<regions.size(); i++)
        {
            Region1D last = regions.get(i-1);
            Region1D next = regions.get(i);
            rv.addRegion(new Region1D(last.getX2(),next.getX1(),null));
        }
        return rv;
    }
    public Vector<Pair<Integer,Integer> > getRegionPairs()
    {
        Vector<Pair<Integer,Integer> > rv = new Vector<Pair<Integer,Integer> >();

        for (Region1D r: regions)
        {
            rv.add(r.toPair());
        }
        return rv;
    }

    public int getNumRegions()
    {
        return regions.size();
    }
    public Region1D getRegion(int idx)
    {
        return regions.get(idx);
    }
    public Vector<Region1D> getRegions()
    {
        return regions;
    }
    Vector<Region1D> regions =  new Vector<Region1D>();

    public String toString()
    {
        return "RegVec:: "+regions;
    }
    
    int _hashcode = -1;


    static int[] key_primes = PrimeGenerator.generatePrimes(100, 50_000);
    static int[] mod_primes = PrimeGenerator.generatePrimes(101, 10_000);
    

    
    static int[] old_key_primes = PrimeGenerator.getPrimes(2);
    static int[] old_mod_primes = PrimeGenerator.getPrimes(3);
    
    static boolean use_old = false;
    
    //static Tuple keyTuple = new Tuple(2);

    public int hashCode()
    {
        if (_hashcode>=0)
            return _hashcode;
        
        
        if (use_old)
            return _hashcode = (old_key_primes[0]*x1)%old_mod_primes[1]+(old_key_primes[1]*x2)%old_mod_primes[2];
        
        

        _hashcode = (key_primes[0]*x1)%mod_primes[1]+(key_primes[1]*x2)%mod_primes[2];

        int pidx = 1;

        for (Region1D r : getRegions())
        {
            pidx++;
            pidx = pidx % 100;
            _hashcode += ((key_primes[pidx]* r.x1) % mod_primes[pidx + 1]+(key_primes[pidx]* r.x2) % mod_primes[pidx + 1]);
                
        }
        return _hashcode;
        
    }

    public RegionVector cut(int x1, int x2)
    {
        RegionVector rv = new RegionVector();
        for (Region1D r : regions)
        {
            if (r.getX2() < x1 || r.getX1() >= x2)
                continue;
            
            
            int start = Math.max(x1, r.getX1());
            int end = Math.min(x2, r.getX2());

            if (start >= end)
                continue;
            
            rv.addRegion(start, end);
        }
        return rv;
    }

    public RegionVector substract(RegionVector rv)
    {
        Vector<Tuple3<Integer, Boolean, Boolean>> infos = new Vector<>();
        
        for (int i=0; i<2; i++)
        {
            RegionVector r = (i == 0) ? this : rv;
            Boolean v = (i == 0) ? false : true;
            for (int j = 0; j < 2; j++)
            {
                infos.addAll(ObjectGetter.map(r.getRegions(), 
                                              (j == 0) ? (_r) -> (Tuple3<Integer, Boolean, Boolean>)Tuple3.create(_r.getX1(), true, v) 
                                                       : (_r) -> (Tuple3<Integer, Boolean, Boolean>)Tuple3.create(_r.getX2(), false, v) 
                                              )
                             );

            }
        }


        NumUtils.sort(infos, (_i) -> _i.get0() + ((!_i.get1()) ? 0.5 : 0.0));
        Integer del_open = null;
        Integer reg_open = null;
        Integer last_deleted = null;

        RegionVector rrv = new RegionVector();

        //System.out.println("infos: " +infos);
        for (Tuple3<Integer, Boolean, Boolean> p : infos)
        {
            if (p.get2()) //to substract
            {
                if (!p.get1()) //end of region
                {
                    //System.out.printf("end of del region: %d\n", p.get0());
                    if (reg_open != null)
                    {
                        reg_open = p.get0();
                        //System.out.printf("got reg open: %d\n", reg_open);
                    }
                    last_deleted = p.get0();
                    del_open = null;
                    continue;
                }
                if (reg_open != null && reg_open < p.get0())
                {
                   
                    int ro = (last_deleted == null) ? reg_open : Math.max(reg_open, last_deleted);
                    if (ro < p.get0())
                    {
                        rrv.addRegion(ro, p.get0());
                    }
                    //System.out.printf("close reg open %d - %d del open: %d\n", reg_open, p.get0(), del_open);
                    
                    
                }
                //System.out.printf("set del start to%d\n", p.get0());
                del_open = p.get0();
                continue;
            }
            if (p.get1()) //starting a region
            {
                reg_open = p.get0();// : Math.max(p.get0(),del_end);
                //System.out.printf("set reg open  to %d (%d)\n", reg_open, p.get0());
                continue;
            }
            if (reg_open == null || p.get0() <= reg_open) // empty region
            {
                //System.out.printf("cancel open region ending at %d as empty (end : %d)\n", p.get0(), reg_open);
                reg_open = null;
                continue;
            }
                
            if (del_open == null)
            {
                int ro = (last_deleted == null) ? reg_open : Math.max(reg_open, last_deleted);
                if (ro < p.get0())
                {
                    rrv.addRegion(ro, p.get0());
                }
                
                //System.out.printf("close region %d-%d\n", reg_open, p.get0());
            }
            
            
            
            reg_open = null;
            
        }
        return rrv;
    }

    public static RegionVector merge(RegionVector ... rvs)
    {
        return merge(new CollectionIterator<RegionVector>(rvs));
    }

    public static RegionVector merge(Collection<RegionVector> c)
    {
        return merge(c.iterator()); 
    }
    
    public static RegionVector merge(Iterator<RegionVector> it)
    {
        Vector<Region1D> r = new Vector<>();
        ObjectGetter.apply(it, (RegionVector rrv) -> r.addAll(rrv.getRegions()));


        
        
        RegionVector rv = new RegionVector();
        
        if (r.size() == 0)
            return rv;
        
        Collections.sort(r);
        Region1D open = null;
        
        for (int i=0; i <= r.size(); i++)
        {
            Region1D r1d = (i == r.size()) ? null : r.get(i);
            
            if (r1d == null || (open != null && (open.getX2() < r1d.getX1())))
            {
                rv.addRegion(open);
                open = null;
            }
            if (r1d == null)
                continue;
            
            if (open == null)
            {
                open = new Region1D(r1d);
                continue;
            }
            open.setX2(Math.max(open.getX2(), r1d.getX2()));
        }

        return rv;
    }



    public int getCoveredLength()
    {
        return (regions.size() == 0) ? 0 : NumUtils.sum(regions, (Region1D r) -> r.getLength());
    }
    
    public boolean equals(Object o)
    {
        return 0 == compareTo(o);
    }

    public int compareTo(Object o)
    {
        if(!(o instanceof RegionVector))
            return super.compareTo(o);

        //System.out.printf(" compare %s to %s\n", this, o);
        RegionVector rv = (RegionVector)o;
        int diff = getX1() - rv.getX1();
        if (diff != 0)
            return diff;

        if(0 != (diff = getX2() - rv.getX2()))
            return diff;
        
        if(0!= (diff = rv.regions.size() - regions.size()))
            return diff;
        

        
        final int ml = regions.size();

        for (int i=0; i<ml; i++)
        {
            Region1D r1 = regions.get(i);
            Region1D r2 = rv.regions.get(i);
            //System.out.printf("\t%s vs %s: %d vs %d\n", r1, r2, r1.hashCode(), r2.hashCode());
            if (0!= (diff = r1.getX1() - r2.getX1()))
                return diff;

            if (0!= (diff = r1.getX2() - r2.getX2()))
                return diff;
            
        }
             
        //System.out.println("EQUALS!");

        return 0;
    }
	
    public Region1D getLastRegion() 
    {
        return this.get(this.size()-1);
    }


    
    public void serialize(java.io.DataOutput out) throws java.io.IOException
    {
        out.writeInt(getNumRegions());
        for (Region1D r : getRegions())
        {
             out.writeInt(r.getX1());
             out.writeInt(r.getX2());

        }
    }
         
    static public RegionVector deserialize(java.io.DataInput in, int available) throws java.io.IOException
    {
        int nreg = in.readInt();
        RegionVector _rv = new RegionVector();

         for (int ri = 0; ri < nreg; ri++)
         {
             int x1 = in.readInt();
             int x2 =  in.readInt();
             _rv.addRegion(x1, x2);
         }
         
         return _rv;
    }


    public static class RVMSerializer extends org.mapdb.Serializer<HashMap<RegionVector, Integer>> implements java.io.Serializable
    {
         public void serialize(java.io.DataOutput out, HashMap<RegionVector, Integer> m) throws java.io.IOException
         {
             out.writeInt(m.size());
             for (Map.Entry<RegionVector, Integer> e : m.entrySet() )
             {
                 e.getKey().serialize(out);
                 out.writeInt(e.getValue());
             }
             
         }


         public HashMap<RegionVector, Integer> deserialize(java.io.DataInput in, int available) throws java.io.IOException
         {
             HashMap<RegionVector, Integer> rv = new HashMap<>();
             int toread = in.readInt();
             
             for (int i=0; i<toread; i++)
             {
                 rv.put(RegionVector.deserialize(in, available), in.readInt());
             }
             return rv;
             
         }
    }
    
    public static RVMSerializer SERIALIZER = new RVMSerializer();

    public static void main (String args[]) 
    {
        /*String s1 = "11869-12228|12595-12722|12975-13053|13221-14413";
        String s2 = "11900-11950";

        RegionVector rv1 = RegionVector.parseSimpleRepresentation(s1);
        RegionVector rv2 = RegionVector.parseSimpleRepresentation(s2);

        System.out.printf("%s\nvs\n%s\nsub: %s\n", rv1, rv2,
                          rv1.containsRV(rv2)
                          );

        */
        int maxpos = 10_000;
        HashMap<Integer, HashSet<RegionVector>> hashtoset = new HashMap<>();
        for (int i=0; i<100_000; i++)
        {
            int lastpos = 0;
            RegionVector rv = new RegionVector();
            while (lastpos < maxpos)
            {
                int start = lastpos + (int) (Math.random() * (maxpos - lastpos));
                int end = start + (int)(Math.random() * (maxpos - start));

                if (end <= start + 2)
                    break;
                
                
                rv.addRegion(start, end);
                lastpos = end;
            }
            String rep = rv.getSimpleRepresentation();
            RegionVector rv2 = RegionVector.parseSimpleRepresentation(rep);
            if (rv2.hashCode() != rv.hashCode())
            {
                throw new RuntimeException("hashcode !=");
            }
            if (!rv2.equals(rv))
            {
                throw new RuntimeException("noneq");
            }
            //System.out.printf("%s HASHCODE: %d, %d\n", rv, rv.hashCode(), rv2.hashCode());
            MapBuilder.update(hashtoset, rv.hashCode(), rv);
            if (i > 0 && i % 10_000 == 0)
            {
                System.out.printf("after %d: %s\n", i, NumUtils.getNumInfo(hashtoset.values(), (_e) -> _e.size()));
            }
            
        }
        
        System.out.printf("final: %s\n", NumUtils.getNumInfo(hashtoset.values(), (_e) -> _e.size()));
    }
}
