package lmu.utils;

import java.util.*;
import lmu.utils.*;

public class RegionVectorChecker<A>
{

    Overlap2DDetector od = new Overlap2DDetector() ;

    static class RVRegion<A>
    {
        Region1D region;
        Pair<RegionVector, A> info;
        
        public RVRegion(Region1D r, RegionVector p, A data)
        {
            this.region = r; info = Pair.create(p, data);
        }

        public String toString()
        {
            return ""+region;
        }
    }


    public RegionVectorChecker()
    {

    }

    public RegionVectorChecker(RegionVector rv, A a)
    {
        this();
        add(rv,a);
    }

    boolean sorted = false;
    Vector<RVRegion<A>> regions = new Vector<>();
    static Pair<Double, Double> y = Pair.create(0.0, 1.0);

    public void add(RegionVector rv, A a)
    {
        ObjectGetter.apply(rv.getRegions(), (Region1D r) -> od.add(Pair.create((double)r.getX1(), (double)r.getX2()), y, new RVRegion(r, rv, a)));
        sorted = false;
    }

    void compile()
    {
        if (sorted)
            return;
        
        Collections.sort(regions, (RVRegion r1, RVRegion r2) -> r1.region.getX1() - r2.region.getX1());
    }


    public interface CoExtractor<B>
    {
        public int get(B b);
    }

    public<B> HashSet<Pair<RegionVector, A>> getOverlapping(Collection<B> query, CoExtractor<B> start_extractor, CoExtractor<B> end_extractor, boolean contained)
    {
        HashSet<Pair<RegionVector,A>> hits = new HashSet<Pair<RegionVector,A>>();

        HashSet<Pair<RegionVector,A>> filtered = null;

        for (B b : query)
        {
            int start = start_extractor.get(b);
            int end = end_extractor.get(b);

            Collection<RVRegion<A>> h = (HashSet<RVRegion<A>>)od.getOverlaps((double)start, (double)end, y.getFirst(), y.getSecond(), null);

            //System.out.printf("%s hits: %s\n", r, h);
            if (!contained)
            {
                hits.addAll(ObjectGetter.map(h, (RVRegion<A> r2) -> r2.info));
                continue;
            }

            if (contained)
            {
                h = ObjectGetter.filter(h, (RVRegion<A> r2) -> r2.region.getX1() <= start &&  r2.region.getX2() >= end);
            }

            //System.out.printf("coneined %s hits: %s\n", r, h);
            

            HashSet<Pair<RegionVector,A>> ns = ObjectGetter.mapToSet(h, (RVRegion<A> r2) -> r2.info);

            if (filtered == null)
            {
                filtered = ns;
                continue;
            }
            
            filtered = (HashSet<Pair<RegionVector,A>>)SetInfo.intersect((HashSet)filtered, (HashSet)ns);
            //Pair<Boolean, Integer> p = VectorUtils.binarySearch(regions, r, (RVRegion<A> a, Region1D b) -> a.region.getX1() - b.getX1());
            //System.out.printf("looking for %s resulted : %s\n", r, h);
        }
        
        return (contained) ?  filtered : hits;
    }
    
    public boolean isIn(int pos)
    {
        Collection<RVRegion<A>> h = (HashSet<RVRegion<A>>)od.getOverlaps((double)pos, (double)pos, y.getFirst(), y.getSecond(), null);
        return (h != null && h.size() > 0);
    }

    

    public HashSet<Pair<RegionVector, A>> getOverlapping(RegionVector rv, boolean contained)
    {
        return getOverlapping(rv.getRegions(), (Region1D r1) -> r1.getX1(), (Region1D r2) -> r2.getX2(), contained);
    }

    public static void main(String[] args)
    {
        RegionVectorChecker<String> rvc = new RegionVectorChecker<>();

        RegionVector rv = new RegionVector(ObjectGetter.toVector(Pair.create(10, 20), Pair.create(30, 40), Pair.create(50, 60)));
        RegionVector rv2 = new RegionVector(ObjectGetter.toVector(Pair.create(17, 24), Pair.create(35, 47), Pair.create(48, 78)));
        RegionVector rv3 = new RegionVector(ObjectGetter.toVector(Pair.create(4, 53)));

        System.out.println("mergeD: " + RegionVector.merge(rv, rv2,rv3));
        

        rvc.add(rv, "a");
        rvc.add(rv2, "b");


        RegionVector test = new RegionVector(ObjectGetter.toVector(Pair.create(15, 19), Pair.create(37, 40)));
        RegionVector test2 = new RegionVector(ObjectGetter.toVector(Pair.create(18, 20), Pair.create(35, 40)));
        //RegionVector test2 = new RegionVector(ObjectGetter.toVector(Pair.create(15, 19), Pair.create(37, 40)));
        System.out.println("overlappings: " + rvc.getOverlapping(test, false));
        System.out.println("containing: " + rvc.getOverlapping(test, true));



        System.out.println("test2  : " +test2);
        System.out.println("overlappings: " + rvc.getOverlapping(test2, false));
        System.out.println("containing: " + rvc.getOverlapping(test2, true));
          //lmu.dv.MemoryList.redraw
    }
}

