package lmu.utils;

import java.sql.*;
import org.jdom.*;
import org.jdom.input.*;
import java.util.List;

import lmu.utils.OptionParser.Option;
import lmu.utils.DbConnection;
import java.io.*;

public class SQLConnectionOptionParser extends OptionParser<Connection>
{
    
    static final int HOST_OPTION=0;
    static final int PW_OPTION=1;
    static final int DB_OPTION=2;
    static final int PORT_OPTION=3;
    static final int DBPARAM_OPTION=4;
    static final int USER_OPTION=5;
    static final int DBTYPE_OPTION=6;
    static final int XMLCONFIG_OPTION=7;
    static final int XMLCONFIG_ELEMNAME_OPTION=8;


    
    public void configure(Element e) 
    {
        //parseOptions(e);
        for (Object o : e.getAttributes())
        {
            Attribute a = (Attribute)o;
            parseOption(prefix+a.getName(),a.getValue());

        } 
        
        
    }
    void parseFromXML()
    {
        if (xmlconfig==null || xmlelem==null)
            return;
        
        try
        {
            SAXBuilder sb = new SAXBuilder();
            Document doc = sb.build(new FileReader(xmlconfig));
            Element root = XMLUtils.resolvePaths(doc.getRootElement());
            List l = root.getChildren(xmlelem);
            if (l.size()==0)
            {
                throw new RuntimeException("cannot find config elem");
            }
            if (l.size()>1)
            {
                throw new RuntimeException("ambigous config elem");
            }
            configure((Element)l.get(0));

        }
        catch(Exception e)
        {
            throw new RuntimeException("error at configuring db connection from elem "+xmlelem +" in "+xmlconfig.getAbsolutePath()+":" +e.getMessage(), e);
        }
        
    }

    public boolean parse(int option_index, String[] args, int i, StringBuffer error)
    {
        String value = args[i+1];
        switch(option_index)
        {
            case XMLCONFIG_OPTION:
                xmlconfig = new File(value);
                parseFromXML();
                break;
            case XMLCONFIG_ELEMNAME_OPTION:
                xmlelem = value;
                parseFromXML();
                break;
            case DBTYPE_OPTION:
                dbtype=value;
                break;
        
            case HOST_OPTION:
                host=value;
                break;
            case PW_OPTION:
                password=value;
                break;
            case USER_OPTION:
                user=value;
                break;

            case DB_OPTION:
                db=value;
                break;
            case PORT_OPTION:
                port=Integer.parseInt(value);
                break;
            case DBPARAM_OPTION:
                String sp[]=value.split(":");
                if(sp.length!=2)
                {
                    String err=args[i]+" expects a db parameter in format key:value";
                    throw new RuntimeException(err);
                    
                }
                if (params.length()>0)
                {
                    params.append('&');
                }
                params.append(sp[0]);
                params.append('=');
                params.append(sp[1]);
                break;
            default:
                //unexcpected
                throw new RuntimeException("implementation error! unknown option index: "+option_index+" for switch: "+args[i]);
        }
        return true;

    }
    public String getName()
    {
        return "SQLConnection";
    }

    public String db = null;
    String dbtype="mysql";
    String user="root";
    String host="localhost";
    String password="";
    int port=3306;
    String driver="com.mysql.jdbc.Driver";
    String xmlelem = null;
    File xmlconfig = null;
    StringBuffer params = new StringBuffer();

    
    public void configure(String db, String user, String host, String password)
    {
        configure(db,user,host,password, port);
    }
    public void configure(String db, String user, String host, String password, int port)
    {
        if (db != null)
        {
            this.db = db;
        }
        if (user != null)
        {
            this.user = user;
        }
        if (host != null)
        {
            this.host = host;
        }
        if (password != null)
        {
            this.password = password;
        }
        this.port = port;
    }

    public SQLConnectionOptionParser()
    {
        this("");
    }
    public SQLConnectionOptionParser(String prefix)
    {
        this(prefix, true);
    }
    public SQLConnectionOptionParser(String prefix, boolean needdb)
    {
        super(prefix);
        registerOption(HOST_OPTION, "host", "host", host, 1);
        registerOption(USER_OPTION, "user", "user", user, 1);
        registerOption(PW_OPTION, "pw", "password", "emtpy", 1);
        registerOption(DB_OPTION, "db", "dbname", (needdb)?null:"either this or -xmldbconfig  with -xmldbconfigelem must be set" , 1);
        registerOption(PORT_OPTION, "port", "port", ""+port, 1);
        registerOption(DBTYPE_OPTION, "dbtype", "dbtype", ""+dbtype, 1);
        Option o= registerOption(XMLCONFIG_OPTION, "xml", "xml config file with connection params", "either this and -xml or -db and the other params must be set", 1);
        o.setFileParam(1);
        registerOption(XMLCONFIG_ELEMNAME_OPTION, "xmlelem", "elem name in config xml with the connection params", "either this with -xmldbconfig or -db and the other params must be set", 1);
        registerOption(DBPARAM_OPTION, "dbparam", "dbparam", "emtpy", 1);

    }


    DbConnection keepalive_con = null;
    public DbConnection getKeepAliveConnection()
    {
        if(keepalive_con!=null)
            return keepalive_con;

        //System.out.printf("db: %s\n", db);
        if (db==null)
        {
            return null;
        }
        return keepalive_con = new DbConnection(host, port,  db, user, password, dbtype, params.toString(), null);
    }
    public Connection getObject()
    {
        try
        {
            return getConnection();
        }
        catch(Exception e)
        {
            throw new RuntimeException(e);
        }
    }
    
    public Connection getConnection() throws SQLException
    {
        return getKeepAliveConnection().getConnection();
    }

    
}


