package lmu.utils;

import java.util.HashMap;
import java.util.regex.Pattern;

/**
 * superclass for serializer, gets the corresponding serializer for the given class name
 */
public abstract class Serializer
{

	static HashMap<String, Serializer> cname2serializer = new HashMap<String, Serializer>();
    abstract public boolean check(Object o);

	public static Serializer _getSerializer(String classname)
    {
        Serializer ser = cname2serializer.get(classname);
        if (ser!=null)
            return ser;
        
        try
        {
            ser = (Serializer)Class.forName(classname).newInstance();
            cname2serializer.put(classname, ser);
        }
        catch(Exception ex)
        {
            throw new RuntimeException("error:"+ex.getMessage()+" while looking for xml serializer for "+classname, ex);
        }
        return ser;
    }
	/**
	 * wrapper for serializers that store the serialized information in a csv-like string. It ensures that the serializer's class
	 * is appended to the end of the serialized string (separated by a special String) and that the right serializer is used for 
	 * deserialization
	 * 
	 */
	public static abstract class LinewiseSerializer extends Serializer
	{
		public static String SERIALIZER_SEP="$$";
		public static Pattern SERIALIZER_SEP_PATTERN=Pattern.compile("\\$\\$");

	    abstract protected String serialize_object(Object o);
	    
	    public static void setSep(String sep, Pattern sep_pattern)
	    {
	    	SERIALIZER_SEP = sep;
	    	SERIALIZER_SEP_PATTERN = sep_pattern;
	    }

	    public static String serialize(Class serializer, Object o)
	    {
	        return getSerializer(serializer).serialize(o)+SERIALIZER_SEP+serializer.getName();
	    }

	    public static LinewiseSerializer getSerializer(Class clazz)
	    {
	        return (LinewiseSerializer)_getSerializer(clazz.getName());
	    }
	    
	    public String serialize(Object o)
	    {
	        return serialize_object(o);
	    }
	    abstract protected Object deserialize_object(String in);

	    public static Object deserialize(String in)
	    {
	    	in = in.replaceAll("\n", "");
	    	String[] sp = SERIALIZER_SEP_PATTERN.split(in);
	    	if(sp.length!=2)
	    	{
	    		ExceptionManager.throwRuntimeException("serialized object contains serializer seperator "+SERIALIZER_SEP);
	    	}
	        String classname = sp[1];
	        LinewiseSerializer deser = (LinewiseSerializer)cname2serializer.get(classname);
	        if (deser==null)
	        {
	            try
	            {
	            	Class blub = Class.forName(classname);
	        
	                deser = (LinewiseSerializer)Class.forName(classname).newInstance();
	                cname2serializer.put(classname, deser);
	            }
	            catch(Exception ex)
	            {
	                throw new RuntimeException("error:"+ex.getMessage()+" while looking for deserializer for "+classname, ex);
	            }
	        }
	        return deser.deserialize_object(sp[0]);
	    }
	}

}
