
package lmu.utils;

import java.util.*;

public class SetInfo<A>
{
    
    //static class SortedIntArraySetUnionOperator
    public static Vector mergeSort(Vector v1, Vector v2)
    {
        PeakableIterator<Comparable> pit1 = new PeakableIterator<Comparable>(v1.iterator());
        PeakableIterator<Comparable> pit2 = new PeakableIterator<Comparable>(v2.iterator());
        
        Vector out = new Vector();
        
        
        while (pit1.hasNext() || pit2.hasNext())
        {
            Comparable front1 = pit1.peak();
            Comparable front2 = pit2.peak();

             //System.out.printf("next fronts: %s %s\n", front1, front2);
             if (front1!=null && front2!=null && 0 == front1.compareTo(front2))
             {
                 out.add(front1);
                 out.add(front2);
                 
                 pit1.next();
                 pit2.next();
                 continue;
             }
        
             while ( pit1.peak()!=null && (front2==null || pit1.peak().compareTo(front2) < 0))
             {
                 out.add(pit1.peak());
                 pit1.next();
             }
        
             while ( pit2.peak()!=null && (front1==null || pit2.peak().compareTo(front1) < 0))
             {
                 out.add(pit2.peak());
                 pit2.next();
             }
        
         }
         return out;

    }

    public static class BitSetInvertOperator extends Expression.Operator
    {
        int length;

        BitSetInvertOperator(int max_num_bits)
        {
            super(2, 0, 1, "!", null, BitSet.class);
            length = max_num_bits;
        }

        public Object eval(Object left, Object right)
        {
            BitSet toinvert = (BitSet)right;
            BitSet rv = new BitSet(length);
            ObjectGetter.apply(AutoResizeBitSet.iterator(toinvert, true, length), (Integer idx) -> rv.set(idx));
            return rv;
        }
    }

    public static class SortedIntSetInvertOperator extends Expression.Operator
    {
        Vector<Integer> v;
        int[] a;

        Iterator<Integer> getAllIt()
        {
            return (v!=null)?v.iterator(): CollectionIterator.getIterator(a);
        }
    
        private SortedIntSetInvertOperator()
        {
            super(2, 0, 1, "!", null, Vector.class);
        }
        public SortedIntSetInvertOperator(int[] all_entity_idx)
        {
            this();
            a=all_entity_idx;
        }
        public SortedIntSetInvertOperator(Vector<Integer> all_entity_idx)
        {
            this();
            v=all_entity_idx;
        
        }
        
        public Object eval(Object left, Object right)
        {
            Vector<Integer> v= (Vector<Integer>)right;
            Vector<Integer> excludeIdx = new Vector<Integer>();
            SetInfo.getSortedIntSetInfo(getAllIt(), v.iterator(), excludeIdx, null);
        
            return excludeIdx;
        }
    
    }
    
    static class SortedIntSetUnionOperator extends Expression.Operator
    {
    
    
        public SortedIntSetUnionOperator()
        {
            super(2, 1, 1, "+", Vector.class, Vector.class);
        
        
        
        }
    
        public Object eval(Object left, Object right)
        {
            Vector<Integer> v1= (Vector<Integer>)left;
            Vector<Integer> v2= (Vector<Integer>)right;
            Vector<Integer> unionIdx = new Vector<Integer>();
            SetInfo.getSortedIntSetInfo(v1,v2, null, null, unionIdx);
        
            return unionIdx;
        }
    
    }


    static class BitSetUnionOperator extends Expression.Operator
    {
    
    
        public BitSetUnionOperator()
        {
            super(2, 1, 1, "+", BitSet.class, BitSet.class);
        
        
        
        }
    
        public Object eval(Object left, Object right)
        {
            BitSet v1= (BitSet)left;
            BitSet v2= (BitSet)right;
            BitSet rv = new BitSet(Math.max(v1.length(), v2.length()));
            ObjectGetter.apply(AutoResizeBitSet.iterator(v1), (Integer idx) -> rv.set(idx));
            ObjectGetter.apply(AutoResizeBitSet.iterator(v2), (Integer idx) -> rv.set(idx));

            return rv;
            /*Vector<Integer> union = new Vector<>();
            getSortedIntSetInfo(new BitSetIndexOperator(v1),new BitSetIndexOperator(v2), null, null, union);
            return sortedInttoBitSet(union);
            */
        }
    
    }

    static class SortedIntSetIntersectOperator extends Expression.Operator
    {
    
    
        public SortedIntSetIntersectOperator()
        {
            super(2, 1, 1, "^", Vector.class, Vector.class);
        
        
        
        }
    
        public Object eval(Object left, Object right)
        {

            Vector<Integer> v1= (Vector<Integer>)left;
            Vector<Integer> v2= (Vector<Integer>)right;
            return SetInfo.getSortedIntSetInfo(v1, v2, null, null);
        
        }
    
    }

    static class BitSetIntersectOperator extends Expression.Operator
    {
    
        Logger log = LogConfig.getLogger();
        public BitSetIntersectOperator()
        {
            super(2, 1, 1, "^", BitSet.class, BitSet.class);
        
        
        
        }
    
        public Object eval(Object left, Object right)
        {

            BitSet v1= (BitSet)left;
            BitSet v2= (BitSet)right;
            
            //log.info("start eval v1: %d v2: %d\n", v1.length(), v2.length());
            return getBitSetInfo(v1, v2, null, null, null);
            /*
            //log.info("got result %d\n", rv.length());
            if (true)
                return rv;
            
            
            
            Vector<Integer>  vv1 = ObjectGetter.toVector(AutoResizeBitSet.iterator(v1));
            log.info("vv1: %d\n", vv1.size());
            Vector<Integer>  vv2 = ObjectGetter.toVector(AutoResizeBitSet.iterator(v2));
            log.info("vv2: %d\n", vv2.size());

            Vector<Integer> is = getSortedIntSetInfo(vv1, vv2);
            log.info("insersction: %d\n", is.size());
            return sortedInttoBitSet(getSortedIntSetInfo(AutoResizeBitSet.iterator(v1), AutoResizeBitSet.iterator(v2), null, null, null));
            */
            
        }
    
    }

    static class SortedIntSetMinusOperator extends Expression.Operator
    {
    
    
        public SortedIntSetMinusOperator()
        {
            super(2, 1, 1, "-", Vector.class, Vector.class);
        
        
        
        }
    
        public Object eval(Object left, Object right)
        {
            Vector<Integer> v1= (Vector<Integer>)left;
            Vector<Integer> v2= (Vector<Integer>)right;

            
            Vector<Integer> minusIdx = new Vector<Integer>();
            SetInfo.getSortedIntSetInfo(v1, v2,  minusIdx, null, null);
        
            return minusIdx;
        }
    
    }


    static class BitSetMinusOperator extends Expression.Operator
    {
    
    
        public BitSetMinusOperator()
        {
            super(2, 1, 1, "-", BitSet.class, BitSet.class);
        
        
        
        }
    
        public Object eval(Object left, Object right)
        {
            BitSet v1= (BitSet)left;
            BitSet v2= (BitSet)right;

            final int v1n = v1.cardinality();

            BitSet rv = new BitSet(v1n);
            ObjectGetter.apply(AutoResizeBitSet.iterator(v1), (Integer idx) -> rv.set(idx));
            ObjectGetter.apply(AutoResizeBitSet.iterator(v2), (Integer idx) -> rv.clear(idx));
            return rv;
        }
    
    }
    
    public static class SetInvertOperator  extends Expression.Operator
    {
        Set all_elem;
        public SetInvertOperator(Set all_elem)
        {
            super(2, 0, 1, "!", Set.class, Set.class);
            this.all_elem=all_elem;
        }
        public Object eval(Object left, Object right)
        {
            if (all_elem == null)
            {
                throw new RuntimeException("cannot use invert operator on set, the whole set is undefined");
            }
            Set s = (Set)right;
            HashSet rv = new HashSet();
            for (Object o : all_elem)
            {
                if (s.contains(o))
                    continue;
                
                rv.add(o);
            }

            return rv;
        }
    }

    public static Expression buildSetExpression(Set all_elems, String expr, Set ... data)
    {
        return Expression.buildExpression(SetInfo.getSetExpression(all_elems), expr, data);
    }

    public static Expression getSetExpression(Set all_elem)
    {
        Expression expr = new Expression();
        expr.registerOperator(new SetInvertOperator(all_elem));
        return expr;
    }


    public static BitSet evalBitSetExpression(int max_num_bits, String expression, Object ... data)
    {
        Expression e = getBitSetExpression(max_num_bits);
        return e.evalExpression(expression, data);

    }
    public static Expression getBitSetExpression(int max_num_bits)
    {
        Expression expr = new Expression();
        expr.registerOperator(new BitSetInvertOperator(max_num_bits));
        expr.registerOperator(new BitSetUnionOperator());
        expr.registerOperator(new BitSetIntersectOperator());
        expr.registerOperator(new BitSetMinusOperator());

        return expr;
    }

    public static Expression getSortedIntSetExpression(SortedIntSetInvertOperator invertOp)
    {
        Expression expr = new Expression();
        expr.registerOperator(invertOp);
        expr.registerOperator(new SortedIntSetUnionOperator());
        expr.registerOperator(new SortedIntSetIntersectOperator());
        expr.registerOperator(new SortedIntSetMinusOperator());

        return expr;
    }
    
    public static Expression getSortedIntSetExpression(int[] all_index)
    {
        return getSortedIntSetExpression(new SortedIntSetInvertOperator(all_index));
    }
    public static Expression getSortedIntSetExpression(Vector<Integer> all_index)
    {
        return getSortedIntSetExpression(new SortedIntSetInvertOperator(all_index));
    }

    public static Vector<Integer> getFullIntVector(int n)
    {
        Vector<Integer> v = new Vector<Integer>(n+1);
        for (int i=0; i<=n; i++)
        {
            v.add(i);
        }
        return v;
    }
    public final static int INTERSECTION_FLAG=1;
    public final static int EXL_SET_1_FLAG=2;
    public final static int EXL_SET_2_FLAG=4;
    public final static int UNION_FLAG=8;
    public final static int ALL_FLAGS=INTERSECTION_FLAG+EXL_SET_1_FLAG+EXL_SET_2_FLAG+UNION_FLAG;
    Set set1;
    Set set2;

    int n_excl_set1=0;
    int n_excl_set2=0;
    int n_intersection=0;
    int n_union = 0;

    public HashSet<A> exl_set1 = null;
    public HashSet<A> exl_set2 = null;
    public HashSet<A> intersection_set = null;
    public HashSet<A> union_set = null;
    
    
    public enum SetType
    {
        MERGED(0),
        OVERLAP(1),
        DISTINCT1(2),
        DISTINCT2(3)
        ;

        public int idx;
        SetType(int idx)
        {
            this.idx = idx;
        }

        public static SetType get(int idx)
        {
            switch (idx)
            {
                case 0:
                    return MERGED;
                case 1:
                    return OVERLAP;
                case 2:
                    return DISTINCT1;
                case 3:
                    return DISTINCT2;
            }
            return null;
        }

        public static int length()
        {
            return 4;
        }

        static public<A> Vector<Set<A>> getSets(Set<A> s1, Set<A> s2)
        {
            Vector<Set<A>> rv = ObjectGetter.mapIndex(4, (_idx) -> null);

            SetInfo<A> si = new SetInfo<>(s1, s2, SetInfo.ALL_FLAGS);

            for (SetType st : SetType.values())
            {
                switch (st)
                {
                    case MERGED:
                        rv.set(st.idx, si.union_set);
                        break;
                    case OVERLAP:
                        rv.set(st.idx, si.intersection_set);
                        break;
                    case DISTINCT1:
                        rv.set(st.idx, si.exl_set1);
                        break;
                    case DISTINCT2:
                        rv.set(st.idx, si.exl_set2);
                        break;
                }
            }
            System.out.println("will return "+ ObjectGetter.map(rv, (_s) -> (_s == null) ? -1 :  _s.size()));
            return rv;
        }
    }


    public static<A> Set<A> union(Collection<Set<A>> sets) 
    {
        return union(sets.iterator());
    }

    public static<A> Set<A> union(Set<A> ... sets) 
    {
        return union(new CollectionIterator<Set<A>>(sets));
    }

    public static<A> Set<A> union(Iterator<Set<A>> setIt) 
    {
        HashSet<A> h = new HashSet<>();
        ObjectGetter.apply(setIt, (Set<A> hs) -> h.addAll(hs));
        return h;
    }

    public static BitSet bs_intersect(Collection<BitSet> sets)
    {
        return bs_intersect(sets.iterator());
    }

    public static BitSet bs_intersect(Iterator<BitSet> setit)
    {
        if (!setit.hasNext())
            return null;
        
        BitSet bs = setit.next();

        BitSetIntersectOperator bsi = new BitSetIntersectOperator();

        while (setit.hasNext())
        {
            bs = (BitSet)bsi.eval(bs, setit.next());
        }
        return bs;
    }


    public static BitSet intersect(BitSet ... sets)
    {   
        return bs_intersect(new CollectionIterator<BitSet>(sets));
    }

    public static BitSet bs_union(Collection<BitSet> sets)
    {
        if (sets.size() < 2)
            return (sets.size() == 0) ? null : ObjectGetter.first(sets);
            //throw new RuntimeException("need at least to sets");

        return bs_union(sets.iterator());
    }


    public static BitSet bs_union(Iterator<BitSet> sets)
    {
        if (!sets.hasNext())
            return null;
        
        BitSet bs1 = sets.next();
        
        if (!sets.hasNext())
            return null;

        BitSet bs2 = sets.next();
        
        
        BitSetUnionOperator os = new BitSetUnionOperator();
        
        BitSet bs = (BitSet)os.eval(bs1, bs2);

        while (sets.hasNext())
        {
            bs = (BitSet)os.eval(sets.next(), bs);
        }

        return bs;
    }

    public static BitSet union(BitSet ... sets)
    {   
        if (sets.length < 2)
            throw new RuntimeException("need at least to sets");
        
        return bs_union(new CollectionIterator<>(sets));
        
    }

    public static BitSet minus(BitSet ... sets)
    {   
        if (sets.length < 2)
        {
            throw new RuntimeException("need at least to sets");
        }
        BitSetMinusOperator ms = new BitSetMinusOperator();
        BitSet bs = (BitSet)ms.eval(sets[0], sets[1]);

        for (int i=2; i<sets.length; i++)
        {
            bs = (BitSet)ms.eval(bs, sets[i]);
        }

        return bs;
    }


    public static<A> Set<A> intersect(Set<A> ... sets) 
    {
        return intersect(new CollectionIterator<Set<A>>(sets));

    }
    public static<A> Set<A> intersect(Collection<Set<A>> sets) 
    {
        return intersect(sets.iterator());
    }

    public static<A> Set<A> intersect(Iterator<Set<A>>  setIt) 
    {
        Set<A> result = setIt.next();
        while (setIt.hasNext())
        {
            Set<A> s = setIt.next();
            if (result == null || s == null)
            {
                return null;
            }
            result = (new SetInfo<A>(result, s, INTERSECTION_FLAG)).intersection_set;
        }
        return result;
    }

    public static<A> Set<A> minus(Set<A> ... sets) 
    {
        Set<A> result = sets[0];
        for (int i=1; i<sets.length; i++)
        {
            result = (new SetInfo<A>(result, sets[i], EXL_SET_1_FLAG)).exl_set1;
        }
        return result;
    }

    
    public static Vector<Integer> intersect(Vector ... sets)
    {
        Vector<Integer> result = intersect2(sets[0], sets[1]);

        for (int i=2; i<sets.length; i++ )
        {
            result=intersect2(result, sets[i]);
        }

        return result;
    }

    public static Vector<Integer> intersectV(Collection<Vector<Integer>> c)
    {
        Vector<Vector<Integer>> vv = ObjectGetter.toVector(c);
        if (vv.size() == 0)
            return new Vector<>();

        if (vv.size() == 1)
            return vv.get(0);

        Vector<Integer> result = intersect2(vv.get(0), vv.get(1));

        for (int i=2; i<vv.size(); i++ )
        {
            result = intersect2(result, vv.get(i));
        }

        return result;
    }

    
    private static Vector<Integer> intersect2(Vector<Integer> set1, Vector<Integer> set2)
    {
        return getSortedIntSetInfo(set1,set2, null, null);
    }
    public static Vector<Integer> union(Vector ... sets)
    {
        Vector<Integer> result = union2(sets[0], sets[1]);

        for (int i=2; i<sets.length; i++ )
        {
            result=union2(result, sets[i]);
        }

        return result;
    }

    private static Vector<Integer> union2(Vector<Integer> set1, Vector<Integer> set2)
    {
        Vector<Integer> rv = new Vector<Integer>();
        getSortedIntSetInfo(set1,set2, null, null, rv);
        return rv;
    }

    
    public static Vector<Integer> minus(Vector ... sets)
    {
        Vector<Integer> result = minus2(sets[0], sets[1]);

        for (int i=2; i<sets.length; i++ )
        {
            result=minus2(result, sets[i]);
        }

        return result;
    }


    private static Vector<Integer> minus2(Vector<Integer> set1, Vector<Integer> set2)
    {
        Vector<Integer> rv = new Vector<Integer>();
        getSortedIntSetInfo(set1,set2, rv, null);
        return rv;
    }

    public static HashMap<Integer,Integer> count(Vector<Integer>...sets)
    {
        HashMap<Integer,Integer> counts = new HashMap<>();
        for(Vector<Integer> ids : sets)
        {
            for(int id : ids)
            {
                Integer count = counts.get(id);
                if(count==null)
                {
                    count = 0;
                }
                counts.put(id, count+1);
            }
        }
        return counts;
    }

    public static Vector<Integer> getCountFiltered(boolean above, int cutoff, Vector<Integer>...sets)
    {
        HashMap<Integer,Integer> counts = count(sets);
        Vector<Integer> filtered = new Vector<>();
        for(int id : counts.keySet())
        {
            int count = counts.get(id);
//            System.out.println("count for "+id+" is "+count);
            if((above && count>cutoff) || (!above && count<cutoff))
            {
                filtered.add(id);
            }
        }
        return filtered;
    }
    public static Vector<Integer> getSortedIntSetInfo(Vector<Integer> set1, Vector<Integer> set2)
    {
        return getSortedIntSetInfo(set1,set2, null, null);
    }
    public static Vector<Integer> getSortedIntSetInfo(Vector<Integer> set1, Vector<Integer> set2, Vector<Integer> only_in_set1, Vector<Integer> only_in_set2)
    {
        return getSortedIntSetInfo(set1, set2,only_in_set1,only_in_set2, null);
    }
    public static Vector<Integer> getSortedIntSetInfo(Vector<Integer> set1, Vector<Integer> set2, Vector<Integer> only_in_set1, Vector<Integer> only_in_set2, Vector<Integer> union)
    {
        return getSortedIntSetInfo(set1.iterator(),set2.iterator(),only_in_set1,only_in_set2, union);
    }

    public static Vector<Integer> getSortedIntSetInfo(int[] set1, int[] set2)
    {
        return getSortedIntSetInfo(CollectionIterator.getIterator(set1),CollectionIterator.getIterator(set2),null, null);
    }

    public static Vector<Integer> getSortedIntSetInfo(int[] set1, int[] set2, Vector<Integer> only_in_set1, Vector<Integer> only_in_set2)
    {
        return getSortedIntSetInfo(CollectionIterator.getIterator(set1),CollectionIterator.getIterator(set2),only_in_set1,only_in_set2);
    }


    public static Vector<Integer> getSortedIntSetInfo(Iterator<Integer> set1, Iterator<Integer> set2)
    {
        return getSortedIntSetInfo(set1,set2,null, null);
    }

    public static Vector<Integer> getSortedIntSetInfo(Iterator<Integer> set1, Iterator<Integer> set2, Vector<Integer> only_in_set1, Vector<Integer> only_in_set2)
    {
        return getSortedIntSetInfo(set1,set2,only_in_set1,only_in_set2,null);
    }
    
    public static BitSet getBitSetInfo(BitSet set1, BitSet set2, BitSet only_in_set1, BitSet only_in_set2, BitSet union)
    {
        
        PeakableIterator<Integer> pit1 = new PeakableIterator<Integer>(AutoResizeBitSet.iterator(set1));
        PeakableIterator<Integer> pit2 = new PeakableIterator<Integer>(AutoResizeBitSet.iterator(set2));
        

        BitSet overlaps = new BitSet(Math.min(set1.length(), set2.length()));
        
         while (pit1.hasNext() || pit2.hasNext())
         {
             Integer front1 =pit1.peak();
             Integer front2 =pit2.peak();

             //System.out.printf("next fronts: %s %s\n", front1, front2);
             if (front1!=null && front2!=null && ((int)front1)==((int)front2))
             {
                 overlaps.set(front1);
                 if (union!=null)
                 {
                     union.set(front1);
                 }
                 pit1.next();
                 pit2.next();
                 continue;
             }
             
             while ( pit1.peak()!=null && (front2==null || pit1.peak()<front2))
             {
                 if (only_in_set1!=null)
                 {
                     only_in_set1.set(pit1.peak());
                 }
                 if (union!=null)
                 {
                     union.set(pit1.peak());
                 }
                 pit1.next();
             }

             while ( pit2.peak()!=null && (front1==null || pit2.peak()<front1))
             {
                 if (only_in_set2!=null)
                 {
                     only_in_set2.set(pit2.peak());
                 }
                 if (union!=null)
                 {
                     union.set(pit2.peak());
                 }

                 pit2.next();
             }
             //System.out.printf("next fronts after inc %s %s\n", front1, front2);
         }
         return overlaps;
    }



    public static Vector<Integer> getSortedIntSetInfo(Iterator<Integer> set1, Iterator<Integer> set2, Vector<Integer> only_in_set1, Vector<Integer> only_in_set2, Vector<Integer> union)
    {
        PeakableIterator<Integer> pit1 = new PeakableIterator<Integer>(set1);
        PeakableIterator<Integer> pit2 = new PeakableIterator<Integer>(set2);
        

         Vector<Integer> overlaps = new Vector<Integer>();


         while (pit1.hasNext() || pit2.hasNext())
         {
             Integer front1 =pit1.peak();
             Integer front2 =pit2.peak();

             //System.out.printf("next fronts: %s %s\n", front1, front2);
             if (front1!=null && front2!=null && ((int)front1)==((int)front2))
             {
                 overlaps.add(front1);
                 if (union!=null)
                 {
                     union.add(front1);
                 }
                 pit1.next();
                 pit2.next();
                 continue;
             }
             
             while ( pit1.peak()!=null && (front2==null || pit1.peak()<front2))
             {
                 if (only_in_set1!=null)
                 {
                     only_in_set1.add(pit1.peak());
                 }
                 if (union!=null)
                 {
                     union.add(pit1.peak());
                 }
                 pit1.next();
             }

             while ( pit2.peak()!=null && (front1==null || pit2.peak()<front1))
             {
                 if (only_in_set2!=null)
                 {
                     only_in_set2.add(pit2.peak());
                 }
                 if (union!=null)
                 {
                     union.add(pit2.peak());
                 }

                 pit2.next();
             }
             //System.out.printf("next fronts after inc %s %s\n", front1, front2);
         }
         return overlaps;
    }

    public int getNumOnlyInSet1()
    {
        return n_excl_set1;
    }

    public int getNumOnlyInSet2()
    {
        return n_excl_set2;
    }

    public static boolean isDisjunct(Set s1, Set s2)
    {
        return 0==getIntersectionSize(s1,s2);
    }
    public static int getIntersectionSize(Set s1, Set s2)
    {
        if(s1==null || s2==null) return 0;
        SetInfo si = new SetInfo(s1,s2);
        //System.out.println("check intersection : "+s1+"\nvs: "+s2+"\nresult: "+si);
        return si.getIntersectionSize();
    }
    public int getIntersectionSize()
    {
        return n_intersection;
    }
    public SetInfo(Set<A> s1, Set<A> s2)
    {
        this(s1,s2, 0);
    }


    public SetInfo(Set<A> s1, Set<A> s2, int save_set_flags) //1 -> save intersaction, 2 -> save exl set 1, 4-> save intersection 2
    {
        if ((save_set_flags & INTERSECTION_FLAG) > 0)
        {
            intersection_set = new HashSet<A>();
        }
        if ((save_set_flags & EXL_SET_1_FLAG) > 0)
        {
            exl_set1= new HashSet<A>();
        }

        if ((save_set_flags & EXL_SET_2_FLAG) > 0)
        {
            exl_set2= new HashSet<A>();
        }

        if ((save_set_flags & UNION_FLAG) > 0)
        {
            union_set= new HashSet<A>();
        }
    

        this.set1=s1; this.set2=s2;
        for (A o : s1)
        {
            if (union_set != null)
            {
                union_set.add(o);
            }
            if (s2.contains(o))
            {
                if (intersection_set!=null)
                {
                    intersection_set.add(o);
                }
                n_intersection++;
                continue;
            }
            n_excl_set1++;
            if (exl_set1!=null)
            {
                exl_set1.add(o);
            }
        }
        for (A o : s2)
        {
                        
            if (s1.contains(o))
                continue;
            
            if (union_set != null)
            {
                union_set.add(o);
            }


            n_excl_set2++;
            if (exl_set2!=null)
            {
                exl_set2.add(o);
            }
        }
    }
    
    public static<A> int[][] getContingencyTable(Vector<HashSet<A>> sets)
    {
        int length = sets.size();
        int[][] table = new int[length][length];
        for(int i=0; i<length; i++)
        {
            for(int j=i; j<length; j++)
            {
                int intersect = getIntersectionSize(sets.get(i), sets.get(j));
                table[i][j] = intersect;
                table[j][i] = intersect;
            }
        }
        return table;
    }
    
    public String toString()
    {
        return String.format("S1: %d(exl: %d) S2: %d (excl:%d) intersection: %d", 
                             set1.size(), n_excl_set1,
                             set2.size(), n_excl_set2,
                             n_intersection);
    }

    
    public static void main (String args[]) 
    {
       
        if (true)
        {
            BitSet bs = new BitSet(10);
            bs.set(0);
            bs.set(7);
            ObjectGetter.apply(AutoResizeBitSet.iterator(bs), (Integer idx) -> System.out.println(" next set: "+idx));

            ObjectGetter.apply(AutoResizeBitSet.iterator(bs, true, 10), (Integer idx) -> System.out.println(" next unset: "+idx));
            return;
        }
        Vector<Vector<Integer>> vsets = new Vector<>();
        Vector<HashSet<Integer>> hsets = new Vector<>();

        final int MAXN = 10_000;
        final int MAX_SETSIZE = 1000;

        for (int i=0; i<1000; i++)
        {
            HashSet<Integer> h = new HashSet<>();
            int size = 1+(int)(Math.random()*MAX_SETSIZE);
            while (h.size() != size)
            {
                h.add((int)(Math.random()*MAXN));
            }
            hsets.add(h);
            System.out.printf("test %d size: %d\n", i, h.size());
            Vector<Integer> v = new Vector<>();
            v.addAll(h);
            Collections.sort(v);
            vsets.add(v);
        }

        System.out.println("start set test");
        long t1 = System.currentTimeMillis();
        int ntests=0;
        for (int i=0; i<hsets.size(); i++)
        {
            for (int j=i+1; j<hsets.size(); j++)
            {
                SetInfo.intersect(hsets.get(i), hsets.get(j));
                ntests++;
            }
        }
        long t2 = System.currentTimeMillis();
        System.out.printf("sets took: %2.3f sec (%d tests)\n", (t2-t1)/1000.0, ntests);


        System.out.println("start vector test");
        t1 = System.currentTimeMillis();
        for (int i=0; i<hsets.size(); i++)
        {
            for (int j=i+1; j<hsets.size(); j++)
            {
                SetInfo.intersect(vsets.get(i), vsets.get(j));
            }
        }
        t2 = System.currentTimeMillis();
        System.out.printf("vectors took: %2.3f sec (%d tests)\n", (t2-t1)/1000.0, ntests);
    }
}


