package lmu.utils;

import java.io.*;
import java.util.*;

/** class to read one-param options */
public class SimpleOptionParser extends OptionParser<HashMap<String,String> >
{
    String name="cmdline";
    HashMap<String, String> key2value = new HashMap<String, String>();
    HashMap<String, Option> options = new HashMap<String, Option>();
    HashMap<String, Boolean> switches = new HashMap<String, Boolean>();

    Vector<Option> options_v = new  Vector<Option>();

    Vector<Vector<SubParamSet>> optional_required_parameter_sets = new Vector<>();

    class SubParamSet
    {
        String name;
        String[] params;
        boolean is_set = false;

        public SubParamSet(String name, String[] params)
        {
            this.name = name; this.params = params;
        }

        void check()
        {
            is_set = true;
            for (String p : params)
            {
                if (switches.containsKey(p))
                {
                    if (is_set = isSet(p))
                        continue;
                    
                    return;
                }
                if(is_set = isOptionSet(p))
                    continue;

                return;
            }
        }

        public String toString()
        {
            return String.format("(%s: %s)", name, StringUtils.buildSet(params));
        }
    }
    
    HashMap<String, SubParamSet> subparamsets = new HashMap<>();

    int option_index=0;

    
    public SimpleOptionParser(String ... keys)
    {
         addSimpleOptions(keys);
    }

    public boolean isSubParamsetSet(String name)
    {
        if (subparamsets.get(name) == null)
        {
            throw new RuntimeException("unknown subparam name: "+name);
        }
        return subparamsets.get(name).is_set;
    }
    public void addSubParameterSet(String name, String ... params)
    {
        subparamsets.put(name, new SubParamSet(name, params));
    }

    public void setRequiredSubParamSets(String[] ... paramnames)
    {
        String[] nn = new String[paramnames.length];
        int idx = 0;
        for (String[] s: paramnames)
        {
            String n = StringUtils.join("::", s);
            nn[idx++] = n;
            addSubParameterSet(n, s);
        }
        
        setRequiredSubParamSets(nn);
    }

    public void setRequiredSubParamSets(String ... names)
    {
        Vector<SubParamSet> v = new Vector<>();
        for (String n : names)
        {
            SubParamSet s = subparamsets.get(n);
            if (s==null)
            {
                throw new RuntimeException("unknown subparam name: "+n);
            }
            v.add(s);
        }
        optional_required_parameter_sets.add(v);
    }

    public boolean isInited(StringBuffer errorstr)
    {

        if(!super.isInited(errorstr))
            return false;

        if (optional_required_parameter_sets.size() == 0)
            return true;
        
        for (SubParamSet sp : subparamsets.values())
        {
            sp.check();
        }
        for (Vector<SubParamSet> v : optional_required_parameter_sets)
        {
            if(ObjectGetter.filter(v, (o) -> ((SubParamSet)o).is_set).size() > 0)
                continue;

            if (errorstr == null)
                return false;
            
            
            if (errorstr.length() > 0)
            {
                errorstr.append('\n');
            }
            errorstr.append("one of the parameter sets must be set:\n\t"+StringUtils.join("\n\t", (Collection<String>)(ObjectGetter.map(v, (o) -> ""+ o))));
            
        }

        return errorstr == null || errorstr.length() == 0;
    }
    
   public Option addSimpleOption(String k)
   {
       addSimpleOptions(false, k);
       return getOption(k);
   }

   public void addSimpleOptions(String  ... keys)
    {
        addSimpleOptions(false, keys);
    }

    public void addSimpleOptions(boolean addonlynew, String  ... keys)
    {
        for (String k : keys)
        {
            
            Option o = getOption(k);
            if (o!=null)
            {
                if (addonlynew)
                    continue;
                
                throw new RuntimeException("parameter: "+k+" is already  set!");
            }
            o = registerOption(option_index++, k, "", null, 1);
            options_v.add(o);
            options.put(k,o);
        }
    }

    public DataTable getDataTable(String key)
    {
        return getDataTable(key, true, true, "\t");
    }
    public DataTable getDataTable(String key, boolean csv, boolean gotheader, String sep)
    {
        DataTable dt = new DataTable();
        dt.readDataFile(getFile(key), true, true, "\t");
        return dt;
    }
    public Vector<Integer> getIntVect(String key, String sep)
    {
        return StringUtils.convert_int_vector(getValue(key),sep);
    }
    
    public void setDouble(String ... keys)
    {
        for (String k: keys)
        {
            Option o = options.get(k);
            o.setDoubleParam(1);
        }
    }
    public double getDouble(String key)
    {
        try
        {
            return Double.parseDouble(getValue(key));
        }
        catch(NumberFormatException nfe)
        {
            throw new RuntimeException("error for parameter: "+key+ " excepted double got: "+getValue(key));
        }
        
    }
    public void setLong(String ... keys)
    {
           
        for (String k: keys)
        {
            Option o = options.get(k);
            o.setLongParam(1);
        }
    }
    public Long getLong(String key)
    {
        try
        {
            return Long.parseLong(getValue(key));
        }
        catch(NumberFormatException nfe)
        {
            throw new RuntimeException("error for parameter: "+key+ " excepted long got: "+getValue(key));
        }
    }
    Option getCheckedOption(String k)
    {
        Option o = options.get(k);
        if (o != null)
            return o;
            
        throw new FRuntimeException("switch: %s is unknown", k);
        
    }
    public void setInt(String ... keys)
    {
        for (String k: keys)
        {
            getCheckedOption(k).setIntParam(1);
            
        }
    }
    
    public void setFile(String ... keys)
    {
        for (String k: keys)
        {
            getCheckedOption(k).setFileParam(1);
        }
    }
    
    public void setOutFile(String ... keys)
    {
        for (String k: keys)
        {
            getCheckedOption(k).setOutFileParam(1);
        }
    }

    public void setDir(String ... keys)
    {
        for (String k: keys)
        {
            getCheckedOption(k).setDirParam(1);
        }
    }

    
    public boolean isOptionSet(String k)
    {
        return key2value.containsKey(k);
    }
    public boolean isSet(String k)
    {
        Boolean b = switches.get(k);
        return b!=null && b;
    }
    public void setSwitches(String ... keys)
    {
        for (String k : keys )
        {
            Option o = options.get(k);
            o.setNParams(0);
            switches.put(k, false);
        }
    }
    public void setDefault(String k, String v)
    {
        Option o = options.get(k);
        o.default_value=v;
    }
    
    public void setPreset(String k, String v)
    {
        key2value.put(k,  v);
    }
    public void setOptional(String ... keys)
    {
        for (String k: keys)
        {
            Option o = options.get(k);
            if (o==null)
            {
                throw new RuntimeException("cannot set "+k+" optional, it is unknown! known values are: "+options.keySet());
            }
            o.default_value="<optional>";
        }
    }
    
    protected boolean parse(int option_index, String[] args, int i, StringBuffer error)
    {
        if (options_v.size()<=option_index)
            return false;
        
        Option o = options_v.get(option_index);
        String n = o.getOptionName();

        if (switches.get(n)!=null)
        {
            switches.put(n, true);
            return true;
        }
        key2value.put(n, args[i+1]);
        return true;
    }

    public HashMap<String, String> getObject()
    {
        return key2value;
    }
    protected String getName()
    {
        return name;
    }

    public void setName(String name)
    {
        this.name=name;
    }

    public String getValue(String key)
    {
        String rv = key2value.get(key);
        if (rv!=null)
            return rv;
        
        Option o = options.get(key);
        return o.default_value;
    }

    public Double getOptionalDouble(String key, Double default_value)
    {
        if (!isOptionSet(key))
            return default_value;
        
        return getDouble(key);
    }

    public static void checkKeys(HashMap<String, String> tocheck,boolean check_if_one_exists , String ... keys)
    {
        checkKeys((tocheck==null)?null:tocheck.keySet(), check_if_one_exists, keys);
    }
    public static void checkKeys(Collection<String> tocheck, boolean check_if_one_exists, String ... keys)
    {
        HashSet<String> hs = new HashSet<String>();
        if (tocheck!=null)
        {
            hs.addAll(tocheck);
        }
        

        
        HashSet<String> missing = new HashSet<String>();
        for (String k : keys)
        {
            if (hs.contains(k))
            {
                if (check_if_one_exists)
                {
                    return;
                }
                continue;
            }
                
            
            missing.add(k);
            
        }
        if (missing.size()==0)
        {
            return;
        }
        if (check_if_one_exists)
        {
            throw new RuntimeException("pls. specify at least one of the follwoing keys: "+missing);
        }
        throw new RuntimeException("missing keys: "+missing);
    }

    public HashMap<String, String> getKeyValueMap(String key, String sep)
    {
        if (!isOptionSet(key))
        {
            return null;
        }
        String rv = getValue(key);
        if (rv==null)
        {
            return null;
        }
        HashMap<String, String> m = new HashMap<String, String>();
        String[] sp = rv.split(sep);
        for (String s: sp)
        {
            String sp2[]=s.split("=");
            if (sp2.length!=2)
            {
                throw new RuntimeException("cannot parse key-value map from "+rv+" specified to "+key+" param: "+s+" is not in form x=y");
            }
            m.put(sp2[0], sp2[1]);
        }
        return m;

    }
    public Integer getOptionalInt(String key, Integer default_value)
    {
        if (!isOptionSet(key))
            return default_value;
        
        return getInt(key);
    }

    public HashSet<String> getOptionalSetFromFile(String key, HashSet<String> defval)
    {
        if(!isOptionSet(key))
            return defval;

        return getSetFromFile(key);
    }
    public HashSet<String> getSetFromFile(String key)
    {
        File f = getFile(key);
        if(!f.exists())
        {
            return null;
        }
        try
        {
            
            BufferedReader bfr  = new BufferedReader(new FileReader(f));
            HashSet<String> hs = new HashSet<String>();
            String line;
            while (null!=(line=bfr.readLine()))
            {
                hs.add(line);
            }
    
            return hs;
        }
        catch(IOException e)
        {
            throw new FRuntimeException("error at reading set from file: %s to key: %s", e, (f==null) ? null :  f.getAbsolutePath(), key);
        }

    }
    public HashSet<String> getSet(String key, String separator)
    {
        String[] sp = getValue(key).split(separator);
        HashSet<String> h = new HashSet<String>();
        for (String s :sp)
        {
            h.add(s);
        }
        return h;
    }

    public HashSet<String> getOptionalSet(String key, String separator, HashSet<String> optional_value)
    {
        if (!isOptionSet(key))
        {
            return optional_value;
        }
        return getSet(key, separator);

    }
    public Integer getInt(String key)
    {
        try
        {
        
            return Integer.parseInt(getValue(key));
        }
        catch(NumberFormatException nfe)
        {
            throw new RuntimeException("error for parameter: "+key+ " excepted int got: "+getValue(key));
        }
    }

    public String getOptionalValue(String key, String default_value)
    {
        String v = key2value.get(key);
        return (v==null)?default_value:v;
    }

    public PrintWriter getOptionalWriter(String key, OutputStream os)
    {
        if (isOptionSet(key))
            return getWriter(key);
        
        if (os == null) 
            return null;

        return new PrintWriter(os);
    }
    public PrintWriter getOptionalWriter(String key)
    {
        return getOptionalWriter(key, null);
    }
    public File getOptionalFile(String key)
    {
        return getOptionalFile(key, null);
    }
    public File getOptionalFile(String key, File default_file)
    {
        if (!isOptionSet(key))
            return default_file;
        
        return getFile(key);
    }

    /** reads from the file specified by key and returns a file list */
    public File[] getFileArray(String key)
    {
        Vector<File> v = getFiles(key);
        File[] rv = new File[v.size()];
        for (int i=0; i<v.size(); i++)
        {
            rv[i]=v.get(i);
        }
        return rv;
        
    }
    /** */
    public Vector<File> getFiles(String key)
    {
        File f = getFile(key);
        Vector<File> rv = new Vector<File>();
        for(String s: FileUtils.readSet(f))
        {
            rv.add(new File(s));
        }
        return rv;

    }
    public File getFile(String key)
    {
        String fn=getValue(key);
        if (fn==null)
        {
            throw new RuntimeException("unknown file switch: "+key+" requested!");
        }
        return new File(fn);
    }
    
    public File getOutFile(String key)
    {
        String fn=getValue(key);
        if (fn==null)
        {
            throw new RuntimeException("unknown outfile switch: "+key+" requested!");
        }
        return new File(fn);
    }

    public File getOptionalOutFile(String key, File defaultF)
    {
        if(!isOptionSet(key))
            return defaultF;
        return getOutFile(key);
    }

    public PrintWriter getWriter(String key)
    {
        if (!isOptionSet(key))
        {
            return null;
        }
        try
        {
            return new PrintWriter(getFile(key));
        }
        catch(IOException e)
        {
            throw new RuntimeException("error:"+e.getMessage()+" while openinng "+getFile(key).getAbsolutePath()+" set by option: "+key+" for writing!", e);
        }
        
    }
            
    public Iterator<String[]> getFieldSetIterator(String key, String pat)
    {
        return FileUtils.getFieldSetIterator(getFile(key),pat);
        

    }
    public BufferedReader getReader(String key)
    {
        return FileUtils.getReader(getFile(key));
        
        
    }
    
}


