package lmu.utils;

import org.jdom.*;
import java.util.*;
public class SimpleXmlSerializer extends XmlSerializer
{
    
    
    static HashMap<Class, TYPES> c2t = new HashMap<Class, TYPES>();
    static HashMap<String, TYPES> n2t = new HashMap<String, TYPES>();
    public enum TYPES
    {
        STRING("string", new String()),
        INT("int", new Integer(1)),
        BOOL("bool", new Boolean(true)),
        DOUBLE("double", new Double(2.3)),
        INTARRAY("intarray", new int[1]),
        INTVECTOR("intvector", new Vector<Integer>()),
        DOUBLEARRAY("doublearray", new double[1]),
        DOUBLEVECTOR("doublevector", new Vector<Integer>()),
        STRINGARRAY("stringarray", new String[1]),
        STRINGVECTOR("stringvector", new Vector<String>())
        ;

        Object o; 
        String name=null;
        TYPES(String name, Object o)
        {
            this.name=name;
            this.o=o;
            
        }

    }
    static
    {
        for (TYPES t : TYPES.values())
        {
            c2t.put(t.o.getClass(),t);
            n2t.put(t.name, t);
        }
    }

    public boolean check(Object o)
    {
        return null!=c2t.get(o.getClass());
    }
    public void serialize_object(Element e, Object o)
    {
        TYPES t = c2t.get(o.getClass());
        if (t==null)
        {
            throw new RuntimeException("no converter defined for class: "+o.getClass().getCanonicalName());
        }
        e.setAttribute("type", t.name);
        e.setAttribute("value", getAttribString(o,t));
    }
    

    public Object deserialize_object(Element e)
    {
        String v = XMLUtils.getAttrib(e, "value", null, true);
        String type = XMLUtils.getAttrib(e, "type", null, true);
        TYPES t = n2t.get(type);
        if (t==null)
        {
            throw new RuntimeException("no converter defined for type: "+type);
        }

        
        switch (t)
        {
            case INT:
                return Integer.parseInt(v);
            case DOUBLE:
                return Double.parseDouble(v);
            case STRING:
                return v;
            default:
                break;
        }

        String[] sp = v.split(" ");
        
        switch(t)
        {
            case INTARRAY:
                int rvi[] = new int[sp.length];
                for (int i=0; i<sp.length; i++)
                {
                    rvi[i]=Integer.parseInt(sp[i]);
                }
                return rvi;
            case DOUBLEARRAY:
                double rvd[] = new double[sp.length];
                for (int i=0; i<sp.length; i++)
                {
                    rvd[i]=Double.parseDouble(sp[i]);
                }
                return rvd;
            case STRINGARRAY:
                return sp;
            case INTVECTOR:
                Vector<Integer> rvvi=new Vector<Integer>();
                for (String s : sp)
                {
                    rvvi.add(Integer.parseInt(s));
                }
                return rvvi;
            case DOUBLEVECTOR:
                Vector<Double> rvvd=new Vector<Double>();
                for (String s : sp)
                {
                    rvvd.add(Double.parseDouble(s));
                }
                return rvvd;
            case STRINGVECTOR:
                Vector<String> rvvs=new Vector<String>();
                for (String s : sp)
                {
                    rvvs.add(s);
                }
                return rvvs;
        }
        return null;
        
    }
    
    public String getAttribString(Object o, TYPES t)
    {
        String rv = null;
        StringBuffer sb = new StringBuffer();
        switch(t)
        {
            case INTARRAY:
                int [] iv = (int[])o;
                 for (int oo: iv)
                 {
                     if (sb.length()>0)
                     {
                         sb.append(' ');
                     }
                     sb.append(""+oo);
                 }
                 rv=sb.toString();
                return rv;

            case DOUBLEARRAY:
                double [] dv = (double[])o;
                 for (double oo: dv)
                 {
                     if (sb.length()>0)
                     {
                         sb.append(' ');
                     }
                     sb.append(""+oo);
                 }
                 rv=sb.toString();
                return rv;


            case STRINGARRAY:
                String[] sa = (String[])o;
                for (String oo: sa)
                {
                    if (sb.length()>0)
                    {
                        sb.append(' ');
                    }
                    sb.append(""+oo);
                }
                return sb.toString();
            case DOUBLE:
            case STRING:
            case INT:
                return ""+o;

            case DOUBLEVECTOR:
            case INTVECTOR:
            case STRINGVECTOR:
                Vector ov = (Vector)o;
                for (Object oo: ov)
                {
                    if (sb.length()>0)
                    {
                        sb.append(' ');
                    }
                    sb.append(""+oo);
                }
                return sb.toString();
        }
        return null;
    }

    void writeAttrib(Object o)
    {
        Element e = new Element("attrib");
        serialize(e, o);

        System.out.println("input: "+o);
        System.out.println(XMLUtils.prettyXML(e));

        Object o2 = deserialize(e);

        System.out.println("output: "+o2);
    }



    public static void main (String args[]) 
    {
        SimpleXmlSerializer sxs = new SimpleXmlSerializer();
        sxs.writeAttrib(2.33);
        sxs.writeAttrib(23);
        sxs.writeAttrib("whatever");
        int[] testv = new int[12];
        sxs.writeAttrib(testv);
        String[] testsv = new String[]{"dafs", "asfd"};
        
        Vector<String> v = new Vector<String>();
        v.add("hehe"); v.add("hoho");

        sxs.writeAttrib(v);
        
        sxs.writeAttrib(testsv);

    }
}
