package lmu.utils;

import java.util.*;

public class SparseDynMatrixIndex 
{


    AutoResizeVector<Vector<Integer> > x_matchings = new AutoResizeVector<Vector<Integer> >();
    
    class DynIterator implements Iterator<Pair<Integer,Integer>>
    {
        int idx=0;
        Iterator<Integer> it;

        Pair<Integer,Integer> p = new Pair<Integer,Integer>();


        public DynIterator()
        {
            calcNextIt();
        }
        public void remove()
        {
        }
        public boolean hasNext()
        {
            return it!=null && it.hasNext();
        }
        public Pair<Integer,Integer> next()
        {
            if (it==null)
            {
                return null;
            }
            int y=it.next();
            p.set(idx,y);
            if(!it.hasNext())
            {
                idx++;
                calcNextIt();
            }
            return p;

        }

        void calcNextIt()
        {
            it=null;
            while (idx<x_matchings.size() &&  (x_matchings.get(idx)==null || x_matchings.get(idx).size()==0))
            {
                idx++;
            }
            if (idx>=x_matchings.size())
            {
                return;
            }
            it = x_matchings.get(idx).iterator();
        }
    }
    public Iterator<Pair<Integer,Integer>> getDynMatrixIterator()
    {
        sort();
        return new DynIterator();
    }

    void add(AutoResizeVector<Vector<Integer>> v, int pos, int value)
    {
        Vector<Integer> vv = v.get(pos);
        if (vv==null)
        {
            v.set(pos, vv=new Vector<Integer>());
        }
        vv.add(value);
    }

    boolean _sorted=false;
    void sort()
    {
        if (_sorted)
        {
            return;
        }
        for (Vector<Integer> vv: x_matchings)
        {
            if(vv==null)
                continue;
            Collections.sort(vv);
        }
        _sorted=true;
    }
    
    public void addMatch(int i, int j)
    {
        _sorted=false;
        add(x_matchings, i, j);
        
    }

    public void addMatch(Pair<Integer,Integer> p)
    {
        _sorted=false;
        add(x_matchings, p.getFirst(), p.getSecond());
        
    }

    class PrefIterator implements Iterator<Pair<Integer,Integer> >
    {

        int stop_y=-1;
        int target_y=0;
        int target_x=0;
        int startx=-1;
        Pair<Integer,Integer> n = new Pair<Integer,Integer>(0,0);
        //Pair<Integer,Integer> rv = new Pair<Integer,Integer>(0,0);
        
        boolean verbose=false;

        PrefIterator(int x, int y, boolean verbose)
        {
            sort();
            startx=x;
            this.verbose=verbose;
            target_x=x;
            target_y=y;

            if(verbose)
            {
                System.out.printf("start at: x: %d y: %d\n", x, y);
            }
            calcNext();

        }

        public void remove()
        {
            throw new UnsupportedOperationException("remove not supported on PrefIterator");
        }
        public Pair<Integer,Integer> next()
        {
            Pair<Integer,Integer> rv = n.clone();
            calcNext();
            return rv;
        }
        public boolean hasNext()
        {
            return n.getSecond()>=0;
        }

        int get_y(int x, int target_y, int min_y)
        {
            Vector<Integer> v = x_matchings.get(x);

            if (v==null || v.size()==0)
                return -1; //no relevant y

            //binary search for target_y in sorted v
            int low_y=0;
            int high_y=v.size()-1;

            while (low_y!=high_y)
            {
                int middle = (int)((0.5)*(low_y+high_y));
                int test = v.get(middle);
                if (test==target_y)
                {
                    low_y=high_y=middle;
                    break;
                }
                if (test<target_y)
                {
                    low_y=Math.min(high_y, middle+1);
                    continue;
                }
                high_y=Math.max(low_y, middle-1);

            }

            //look for the first entry <=target_y and >min y
            int y_index=(v.get(low_y)>target_y)?low_y-1:low_y;
            if(y_index<0 || v.get(y_index)<=min_y)
                return -1;

            return v.get(y_index);

        }
        void calcNext()
        {
            int ny=-1;
            
            for (; target_x>=0; target_x--)
            {

                ny=get_y(target_x, (target_x==startx)?target_y-1:target_y, stop_y);

                if(verbose)
                {
                    System.out.printf("test: x: %d (y: %d-%d) res: %d\n", target_x, (target_x==startx)?target_y-1:target_y, stop_y, ny);
                }
                if(0<=(ny))
                {
                    break;
                }
            }
            if(ny>=0)
            {
                //System.out.printf("got next: %d,%d targety: %d\n", target_x, ny, target_y);
                stop_y=ny;
                
            }
            if(verbose)
            {
                System.out.printf("got new child: %d,%d\n", target_x, ny);
            }
            n.set(target_x,ny);
            target_x--;
            
        }


    }

    public Iterator<Pair<Integer,Integer> > getPrefixes(int x, int y)
    {
        return new PrefIterator(x,y, false);
    }
    public Iterator<Pair<Integer,Integer> > getPrefixes(int x, int y, boolean verbose)
    {
        return new PrefIterator(x,y, verbose);
    }


    public boolean removeChildren(HashSet<Pair<Integer,Integer>> h, Pair<Integer,Integer> p)
    {
        Iterator<Pair<Integer,Integer>> it = h.iterator();

        boolean removed=false;
        while (it.hasNext())
        {
            Pair<Integer,Integer> t = it.next();
            if (t.getFirst()==p.getFirst() && t.getSecond()==p.getSecond())
            {
                continue;
            }

            if (t.getFirst()<=p.getFirst() && t.getSecond()<=p.getSecond())
            {
                //System.out.printf("remove: %s due to %s\n", t, p);
                it.remove();
                removed=true;
            }
        }
        return removed;
    }
    static class Cell
    {
        
        char c=' ';

        public void setQuery()
        {
            c='0';
        }
        public void setUsed()
        {
            c='x';
        }
        public void setChild()
        {
            c='*';
        }
    }
    
    static public String getMatrix(int maxx, int maxy, HashSet<Pair<Integer,Integer> > used, int qx, int qy, HashSet<Pair<Integer,Integer>> children)
    {
        Cell[][] m = new Cell[maxx][maxy];
        StringBuffer rb = new StringBuffer();
        rb.append('+');
        for (int j=0; j<maxy; j++)
        {
            rb.append("---+");
        }
        for (int i=0; i<maxx; i++)
        {
            for (int j=0; j<maxy; j++)
            {
                m[i][j]=new Cell();
            }
        }
        for (Pair<Integer,Integer> p : used)
        {
            m[p.getFirst()][p.getSecond()].setUsed();
        }
        for (Pair<Integer,Integer> p : children)
        {
            m[p.getFirst()][p.getSecond()].setChild();
        }
        m[qx][qy].setQuery();

        
        rb.append('\n');
        String rowsep = rb.toString();
        StringBuffer sb = new StringBuffer();
        sb.append(rowsep);

        for (int i=0; i<maxx; i++)
        {
            StringBuffer rcb  = new StringBuffer();
            rcb.append('|');

            for (int j=0; j<maxy; j++)
            {
                rcb.append(String.format(" %c |", m[i][j].c));
            }
            rcb.append('\n');
            sb.append(rcb.toString());
            sb.append(rowsep);
        }
        return sb.toString();
    }
    public HashSet<Pair<Integer,Integer>> getPrefixSetBF(int tx,int ty)
    {
        HashSet<Pair<Integer,Integer>> hs = new HashSet<Pair<Integer,Integer>>();
        HashSet<Pair<Integer,Integer>> all = new HashSet<Pair<Integer,Integer>>();

        HashSet<Pair<Integer,Integer>> children = new HashSet<Pair<Integer,Integer>>();

        for (int x=0; x<=tx; x++)
        {
            Vector<Integer> v = x_matchings.get(x);
            if (v==null)
                continue;
            
            
            for (int y : v)
            {
                Pair<Integer,Integer> p = new Pair<Integer,Integer>(x,y);
                if(x==tx && y==ty)
                    continue;

                if (x<=tx && y<=ty)
                {
                    children.add(p);
                }
                all.add(p);
            }
        }
        Vector<Pair<Integer,Integer> > v = new Vector<Pair<Integer,Integer> >();
        v.addAll(children);
        boolean removed=true;
        int iter=0;

        while (removed)
        {
            //System.out.printf("iter: %d children: %s\n",iter, children);
            iter++;
            removed=false;
            //v.clear();
            //v.addAll(children);
            for (Pair<Integer,Integer> p : v)
            {
                if(!removeChildren(children, p))
                    continue;
                removed=true;
            }
        }

        return children;

    }
    public HashSet<Pair<Integer,Integer>> getPrefixSet(int x,int y, boolean verbose)
    {
        HashSet<Pair<Integer,Integer>> hs = new HashSet<Pair<Integer,Integer>>();
        Iterator<Pair<Integer,Integer> > it = getPrefixes(x,y, verbose);
        while (it.hasNext())
        {
            Pair<Integer,Integer> p = it.next();
            //System.out.printf("prefset next : %s\n",p);
            hs.add(p);
        }
        return hs;
    }

    static void ptest(int maxx, int maxy, int npoints)
    {
        SparseDynMatrixIndex sdmi = new SparseDynMatrixIndex();

            HashSet<Pair<Integer,Integer>> hs = new HashSet<Pair<Integer,Integer>>();
            while (hs.size()<npoints)
            {
                int x = (int)(Math.random()*maxx);
                int y = (int)(Math.random()*maxy);

                hs.add(new Pair<Integer,Integer>(x,y));
                
            }
            for (Pair<Integer,Integer> p : hs)
            {
                sdmi.addMatch(p.getFirst(), p.getSecond());
            }

            long t1=new java.util.Date().getTime();
            sdmi.sort();
            long t2=new java.util.Date().getTime();

            long ps = (t2-t1);


            //System.out.println("hs: "+hs);
            double childpropos=0.0;

            for (Pair<Integer,Integer> p : hs)
            {
                t1=new java.util.Date().getTime();
                HashSet<Pair<Integer,Integer>> cset = sdmi.getPrefixSet(p.getFirst(), p.getSecond(), false);
                t2=new java.util.Date().getTime();
                ps+=(t2-t1);
                childpropos+=cset.size();
            }

            System.out.printf("iterator: %2.2f sec child/pos: %2.2f\n", ps/1000.0, childpropos/hs.size());

    }
    static void test(int maxx, int maxy, int npoints)
    {
        SparseDynMatrixIndex sdmi = new SparseDynMatrixIndex();

        HashSet<Pair<Integer,Integer>> hs = new HashSet<Pair<Integer,Integer>>();
        while (hs.size()<npoints)
        {
            int x = (int)(Math.random()*maxx);
            int y = (int)(Math.random()*maxy);

            hs.add(new Pair<Integer,Integer>(x,y));
        }
        for (Pair<Integer,Integer> p : hs)
        {
            sdmi.addMatch(p.getFirst(), p.getSecond());
        }

        long t1=new java.util.Date().getTime();
        sdmi.sort();
        long t2=new java.util.Date().getTime();

        long ps = (t2-t1);

        long bf = 0l;


        //System.out.println("hs: "+hs);
        
        for (Pair<Integer,Integer> p : hs)
        {
            t1=new java.util.Date().getTime();
            HashSet<Pair<Integer,Integer>> bfset= sdmi.getPrefixSetBF(p.getFirst(), p.getSecond());
            t2=new java.util.Date().getTime();
            bf+=(t2-t1);

            t1=new java.util.Date().getTime();
            HashSet<Pair<Integer,Integer>> cset = sdmi.getPrefixSet(p.getFirst(), p.getSecond(), false);
            t2=new java.util.Date().getTime();
            ps+=(t2-t1);
            SetInfo si = new SetInfo(bfset,cset);
            if (si.getNumOnlyInSet1()==0 && si.getNumOnlyInSet2()==0)
            {
                continue;
            }

            cset = sdmi.getPrefixSet(p.getFirst(), p.getSecond(), true);
            System.out.println("\n\n\n=========================================================\n");
                        System.out.printf("query: %s children: %s\n", p, bfset);
            System.out.println(getMatrix(maxx, maxy, hs, p.getFirst(), p.getSecond(), bfset));

            

            System.out.printf("p query: %s children: %s\n", p, cset);
            System.out.println(getMatrix(maxx, maxy, hs, p.getFirst(), p.getSecond(), cset));



        }

        System.out.printf("bf: %2.2f sec iterator: %2.2f sec\n", bf/1000.0, ps/1000.0);
        //System.out.println(getMatrix(maxx, maxy, hs, maxx-1, maxy-1, new HashSet<Pair<Integer,Integer>>()));

        Iterator<Pair<Integer,Integer>> it = sdmi.getDynMatrixIterator();
        Vector<Pair<Integer,Integer> > v = new Vector<Pair<Integer,Integer> >();
        v.addAll(hs);
        Collections.sort(v);
        Vector<Pair<Integer,Integer> > tv = new Vector<Pair<Integer,Integer> >();

        //System.out.println("points: "+v);
        while (it.hasNext())
        {
            tv.add(it.next().clone());
            //System.out.println("next dyn it: "+it.next());
        }

        if (tv.size()!=v.size())
        {
            System.out.println(v);
            System.out.println(tv);
            throw new RuntimeException("ERROR at tv!=v");
        }
        for (int i=0; i< v.size(); i++)
        {
            if (!v.get(i).equals(tv.get(i)))
            {
                System.out.println(v);
                System.out.println(tv);
                throw new RuntimeException("expected "+v.get(i)+" == "+tv.get(i));
            }
        }
        
    }
    public static void main (String args[]) 
    {
        for (int i=0; i<1000; i++)
        {
             test(100, 100, 2000);
        }
        
    }
}



