package lmu.utils;

import java.util.*;
import java.util.regex.*;
import bsh.*;

public class TestExpression
{
    static final String VARIABLE_PATTERN="\\$\\{([^\\}]+)\\}";

    static Pattern EVAL_PATTERN=Pattern.compile("\\$\\{([^\\}]+)\\}");
    static Pattern CALL_PATTERN = Pattern.compile(VARIABLE_PATTERN+"(("+ReflectedObject.SUBP+")+)?");
    static Interpreter interpreter = new Interpreter();
    Vector<String> nonvars = new Vector<String>();
    String suffix = "";
    Vector<ReflectedObject> variables = new Vector<ReflectedObject>();
    

    /** constructor using no getters: the variables referenced in the object are assumed to
     * be accesible from the object itself
     */
    public TestExpression(String expr)
    {
        this(expr, false);
    }
    
    public TestExpression(String expr, boolean verbose)
    {
        this(expr, null, verbose);
    }
    
    /** the variables will be replaced by the values derived by the getters
     */
    public TestExpression(String expr, HashMap<String, ReflectedObject> field2getter)
    {
        this(expr, field2getter, false);
    }
    public TestExpression(String expr, HashMap<String, ReflectedObject> field2getter, boolean verbose)
    {
        
        
        
        Matcher m = CALL_PATTERN.matcher(expr);
        int start=0;
        while (m.find(start))
        {
            if (verbose)
            {
            
                System.out.println("hit: "+m.group());
                for (int i=0; i<m.groupCount(); i++)
                {
                    System.out.printf("\t%d=%s\n", i, m.group(i+1));
                }
            }

            ReflectedObject ref_obj=null;
            String field = m.group(1);
            
            if (field2getter==null || null==(ref_obj=field2getter.get(field)))
            {

                //no fieldgetter added, in this case we assume that the variable itself is a reflected Object
                //ref_obj = ReflectedObject.autoDetect(field);
                if (verbose)
                {
                    System.out.println("WILL AUTODETECT FIELD: "+field);
                }
                ref_obj = new ReflectedObject(field);
                
            }
             

            

            ReflectedObject ro = null;
            ReflectedObject toadd = ref_obj;
            
            

            if (m.group(2)!=null)
            {
                String rexpr="field";
                rexpr+=m.group(2);
                ro = ReflectedObject.autoDetect(rexpr,null, verbose);

                if (verbose)
                {
                    System.out.println("gut subexpression on :"+m.group(1)+":"+m.group(2)+" created from "+rexpr+" : "+ro);
                }
                toadd=new ReflectedObject(ref_obj, ro.sub_obj);
            }
            
            
            
            
            //thest if idx is in valid range
            nonvars.add(expr.substring(start, m.start()));
            

            if (verbose)
            {
                System.out.println("will add reflected object: "+toadd.getInfo()+ " for " + m.group());
            }
            variables.add(toadd);
            
            start=m.end();
        }

        if (start<expr.length())
        {
            suffix = expr.substring(start);
        }
        
        
        
    }


    public static Vector<Pair<String, Vector> > multi_filter(Vector data, boolean tolerate_null, boolean verbose, String ... expressions)
    {
        Vector<Pair<String, Vector> > rv = new Vector<Pair<String, Vector> >();
        for (String e : expressions)
        {
            Vector v = filter(e,data,tolerate_null);

            if (verbose)
            {
                System.out.printf("%s: %d\n", e, v.size());
            }
            Pair<String,Vector> p = new Pair<String,Vector>(e, v);
            rv.add(p);
        }
        return rv;
    }
    public static Vector filter(String expression, Vector data, boolean tolerate_null)
    {
        TestExpression te = new TestExpression(expression);
        
        return te.filter(data,tolerate_null);
    }
    public boolean filter(Object o)
    {
        return filter(o, true);
    }

    public Vector filter(Vector v, boolean tolerate_null)
    {
        Vector rv = new Vector();
        for (Object o : v)
        {
            if (!filter(o,tolerate_null))
                continue;
            
            rv.add(o);
        }
        return rv;
    }
    public boolean filter(Object o, boolean tolerate_null)
    {
        Object ro = eval(o);
        if (ro==null && tolerate_null)
        {
            return false;
        }
        if (! (ro instanceof Boolean) )
        {
            throw new RuntimeException("expression >"+getExpression(o)+"< lead to non-boolean result: "+ro+" on " +o);
        }
        return (Boolean)ro;
    }
    public Object eval(Object o)
    {
        return eval(o,true);
    }
    public Object eval(Object o, boolean tolerate_null)
    {
        String expr = getExpression(o);
        //System.out.println("will eval: "+expr);
        if (expr==null)
        {
            return null;
        }
        try
        {
            //System.out.println("result: "+interpreter.eval(expr));
            return interpreter.eval(expr);
        }
        catch(bsh.EvalError e)
        {
            throw new RuntimeException("error at evaluating expression: >"+ expr+"<: "+e.getMessage(), e);
        }
        
    }
    
    
    /** builds the expression where the variable names are replaced by their
     * values 
     * 
     * see getExpression(o, boolean) for details
     */
    public String getExpression(Object o)
    {
        return getExpression(o,true);
    }


    /** builds the expression where the variable names are replaced by their
     * values.
     * 
     * if tolerate_null is set it returns null  if some of the values are null. 
     *  otherwise it sets the null values also into the expression, 
     * 
     * Remark: using null values might be neccesarry if one would like to 
     * explicitely check for null values, otherwise setting tolerate_null
     * makes the writing of expressions easier: one can write
     * 
     *  o.getMethod().whatever.getMethod(12,1)
     *  
     * without checking if o is null, o.getMethod() is null o.getMethod().whatever
     * is null etc...
     * 
     */
    
    public String getExpression(Object o, boolean tolerate_null_in_expr)
    {
        StringBuffer sb = new StringBuffer();
        for (int i=0; i<variables.size(); i++)
        {
            sb.append(nonvars.get(i));
            Object ro = variables.get(i).get(o);
            if (ro==null && tolerate_null_in_expr)
            {
                return null;
            
            }
            
            if (ro instanceof String)
            {
                //sb.append("\""+ro+"\"");
                sb.append(ro);
                
                continue;
            }
            if (ro instanceof Double)
            {
                //System.out.println("got double object: " + ro);
                double d = (Double)ro;
                //System.out.println("d: "+d+" isNan:"+ Double.isNaN(d));
                if (d == Double.NEGATIVE_INFINITY)
                {
                    sb.append("Double.NEGATIVE_INFINITY");
                }
                else if (d == Double.POSITIVE_INFINITY)
                {
                    sb.append("Double.POSITIVE_INFINITY");
                }
                else if( Double.isNaN(d))
                {
                    sb.append("Double.NaN");
                }
                else
                {
                    sb.append(""+d);
                }

            }
            else
            {
                sb.append(""+ro);
            }
            

        }
        sb.append(suffix);
        //System.out.println("expression: "+sb.toString());;
        return sb.toString();
    }
  

    public String toString()
    {
        StringBuffer rv = new StringBuffer();
        rv.append("expression: \n");
        for (int i=0; i<variables.size(); i++)
        {
            String nv = nonvars.get(i);
            if (nv.length()!=0)
            {
                rv.append("\t "+ nv+"\n");
            }
            rv.append("\t "+variables.get(i)+"\n");
        
        }
        rv.append("\t"+suffix+"\n");
        return rv.toString();
    }
        
    public static void main (String args[]) 
    {
        StringBuffer sb = new StringBuffer();
        for (int i=0; i<args.length; i++)
        {
            if (sb.length()>0)
            {
                sb.append(' ');
            }
            sb.append(args[i]);
        }
        String exp = sb.toString();
        System.out.println("exp: "+exp);
        TestExpression te = new TestExpression(exp,true);
        System.out.println("input: "+exp);
        System.out.println(te);
        System.out.println(te.eval(null));
    }

}

