package lmu.utils;

import java.util.*;

public class Tuple<A> implements Comparable<Tuple<A>>
{

    
    A value;
    int hashcode=-1;
    Object[] values;
    int [] key_primes;
    int [] mod_primes;
    
    public void setValue(A a)
    {
        value=a;
    }

    public A getValue()
    {
        return value;
    }

    public Object[] values()
    {
        return values;
    }

    public int cardinality()
    {
        return values.length;
    }
    public Object get(int i)
    {
        return values[i];
    }

    
    public java.io.File getAsFile(int i)
    {
        Object o = get(i);
        if (o instanceof java.io.File)
            return (java.io.File)o;

        if (o instanceof String)
        {
            return new java.io.File((String)o);
        }
        
        throw new RuntimeException("dunno how to convert "+o+" to file!");
    }
    public double getAsDouble(int i)
    {
        Double d = (Double)NumUtils.convertNumber(get(i), Double.class);
        if(d==null)
            d = Double.NaN;
        return (double)d;
    }
    public int getAsInt(int i)
    {
        try
        {
            return (int)NumUtils.convertNumber(get(i), Integer.class);
        }
        catch(Exception e)
        {
            throw new RuntimeException("error while converting "+get(i)+" to int! tuple was:" +this, e);
        }
        
        
    }

    public Boolean getAsBool(int i)
    {
        return (Boolean)get(i);
    }

    public String getAsString(int i)
    {
        return (String)get(i);
    }
    /** constructor defines only the cardinality */
    public Tuple(int size)
    {
        key_primes=PrimeGenerator.getPrimes(size);
        mod_primes=PrimeGenerator.getPrimes(size+1);
        values=new Object[size];

    }

    public Tuple(Object ... values)
    {
        this(values.length);
        for (int i=0; i<values.length; i++)
        {
            set(i, values[i]);
        }
    }
    
    public<B> B getConverted(int i)
    {
        return (B)get(i);
    }
    
    
    

    public static<A extends Object, B extends Collection<Tuple>> Vector<Pair<Integer, A>> getCountMap(B c, boolean sorted, boolean asc, int ... fields)
    {
        int idx = 0;
        HashMap<A,  HashSet<Integer> > m = new HashMap<>();
        for (Tuple t : c)
        {
            for (int fidx: fields)
            {
                A o = (A)t.get(fidx);
                //System.out.printf("put %s : %d\n", o, idx);
                HashSet<Integer> hs = m.get(o);
                if (hs == null)
                {
                    m.put(o, hs  = new HashSet<>());
                }
                hs.add(idx);
            }
            idx++;
        }
        
                                                
        Vector<Pair<Integer, A>> v = Pair.getSizes(m);
        if (sorted)
        {
            Collections.sort(v);
            if (!asc)
            {
                Collections.reverse(v);
            }
        }
        return v;
    }
    

    public static void sortBy(Vector<Tuple> v, int field, boolean asc)
    {
        ObjectGetter.sortBy(v, (Tuple t) -> (Comparable)t.get(field), asc);
    }
    public int compareTo(Tuple<A> o)
    {
        
        int diff = cardinality() - o.cardinality();
        if (diff!=0) 
            return diff;
        
        

        for (int i=0; i<values.length; i++)
        {
            if (values[i] instanceof Comparable)
            {
                diff = ((Comparable)values[i]).compareTo((Comparable)o.values[i]);
            }
            else
            {
                if (values[i] == null && o.values[i] == null)
                {
                    return 0;
                }
                if (values[i] == null || o.values[i] == null)
                {
                    return (values[i]==null) ? -1 : 1;
                }
                diff = values[i].hashCode()- o.values[i].hashCode();
            }
            if (diff!=0)
                return diff;
            
            
        }
        return diff;
    }
    
    public Tuple<A> clone()
    {
        Tuple<A> t = new Tuple<A>(values.length);

        for (int i=0; i<values.length; i++)
        {
            t.set(i,values[i]);
        }
        return t;
    }

    public Object set(int i, Object o)
    {
        values[i]=o;
        hashcode=-1;
        return o;
    }
    public String toString()
    {
        StringBuffer sb = new StringBuffer();

        for (int i=0; i<values.length; i++)
        {
            if (i>0)
            {
                sb.append(',');
            }
            sb.append(""+values[i]);
        }
        return "("+sb.toString()+")";
    }

    public boolean equalsUnOrdered(Tuple t)
    {
        if (t.values.length!=values.length)
        {
            return false;
        }
        int num_nulls=0;
        int num_nulls_t=0;

        for (int i=0; i<values.length; i++)
        {
            if (t.values[i]==null)
            {
                num_nulls_t++;
            }
            if (values[i]==null)
            {
                num_nulls++;
            }

        }
        if (num_nulls!=num_nulls_t)
        {
            return false;
        }
        
        for (int i=0; i<values.length; i++)
        {
            if(values[i]==null)
                continue;

            boolean found=false;
            for (int j=0; !found && j<t.values.length; j++)
            {
                if (t.values[j]==null)
                    continue;
                
                found=t.values[j].equals(values[i]);

            }
            if (!found)
                return false;
            
            
        
        }
        return true;
    }
    public boolean equals(Object o )
    {

        if (!(o instanceof Tuple))
        {
            return false;
        }
        Tuple t = (Tuple)o;
        //System.out.println("EQUALS invoked on : "+this);

        if (t.values.length!=values.length)
        {
            return false;
        }
        
        for (int i=0; i<values.length; i++)
        {
            if(values[i]==null && t.values[i]==null)
                continue;

            if(values[i]==null || t.values[i]==null)
            {
                return false;
            }
                

            if (!values[i].equals(t.values[i]))
            {
                return false;
            }
                
            
            
        }
        return true;
    }

    public int hashCode()
    {
        if (hashcode>=0)
        {
            return hashcode;
        }
            
        
        hashcode=0;
        for (int i=0; i<values.length; i++)
        {
            if (values[i]==null)
                continue;
            
            hashcode+=(key_primes[i]*(values[i].hashCode()%mod_primes[i+1]));
        }
        
        return hashcode;
    }
    
    static public<A> Tuple toTuple(int[] keys)
    {
        Tuple t = new Tuple(keys.length);
        for (int i=0; i<keys.length; i++)
        {
            t.set(i, keys[i]);
        }
        return t;
    }

    static public<A> Tuple toTuple(A ... keys)
    {
        Tuple t = new Tuple(keys.length);
        for (int i=0; i<keys.length; i++)
        {
            t.set(i, keys[i]);
        }
        return t;
    }

    
    static public<A> Tuple tupleFromCollection(Collection<A> c)
    {
        Tuple t = new Tuple(c.size());
        Iterator<A> it = c.iterator();
        int idx = 0;
        while (it.hasNext())
        {
            t.set(idx++, it.next());
        }
        return t;
    }
}


