package lmu.utils;

import java.util.*;

public class VectorUtils
{

    static public<A> void add(Vector<A> v, A ... elems)
    {
        for (A a : elems)
        {
            v.add(a);
        }
    }

    static public<A> Vector<A> join(Vector<A> ... va)
    {
        Vector<A> v = new Vector<>();
        for (Vector<A> vv : va)
        {
            v.addAll(vv);
        }
        return v;
    }

    static public<A> Vector<A> join(Collection<Vector<A>> cva)
    {
        Vector<A> v = new Vector<>();
        for (Vector<A> vv : cva)
        {
            v.addAll(vv);
        }
        return v;
    }

    static public<A> A get(Vector<A> v, int idx)
    {
        if (v.size() == 0)
            return null;
        
        
        if (idx > 0 || idx < -v.size())
            return v.get(idx);
        
        return v.get(v.size() + idx);
    }

    static public<A> Vector<A> slice(Vector<A> v, int start, int end)
    {
        end = (end > 0 ) ? end : v.size() + end;
        start = (start >= 0 ) ? start : v.size() + start;
        end = Math.min(end, v.size());
        start = Math.max(0, start);
        
        return ObjectGetter.map(IteratorUtils.getIntIterator(start, end), (Integer idx) -> v.get(idx));
    }

    /** returns if found, and at which position (or at which position it would be inserted */
    static public<A> Pair<Boolean,Integer> binarySearch(Vector<A> v, A key, Comparator<A> c) 
    {
        return binarySearch(v, 0, v.size(), key, (A k1, A k2) -> c.compare(k1, k2));
    }


    static public<A,B> Pair<Boolean,Integer> binarySearch(Vector<A> v, B tosearch, Comparator2<A, B> c) 
    {
        return binarySearch(v, 0, v.size(), tosearch, c);
    }
    
    static public<A, B> Pair<Integer,Integer>  getRange(Vector<A> v, B tosearch, Comparator2<A, B> c, B lower, B higher) 
    {
        Pair<Boolean, Integer> p = binarySearch(v, 0, v.size(), tosearch, c);

        if (p.getSecond() >= v.size())
            return null;

        int start = p.getSecond();
        for (int hit_idx = p.getSecond() -1; hit_idx >= 0; hit_idx--)
        {
            if (c.compare(v.get(hit_idx), lower) < 0)
                break;
            
            start = hit_idx;
        }
    
        if (start == p.getSecond() && (start >= v.size() || c.compare(v.get(start), higher) >0))
            return null;
        
        int end = p.getSecond();

        for (int hit_idx = p.getSecond(); hit_idx < v.size(); hit_idx++)
        {
            if (c.compare(v.get(hit_idx), higher) > 0)
                break;
            
            end = hit_idx;
        }

        return Pair.create(start, end);
        
    }     

    /** returns the pos it was found/inserted into */
    static public<A extends Comparable> int insertSorted(Vector<A> v, A key) 
    {
        return insertSorted(v, key, (A a1, A a2) -> a1.compareTo(a2));
    }


    /** returns the pos it was found/inserted into */
    static public<A> int insertSorted(Vector<A> v, A key, Comparator<A> c) 
    {
        Pair<Boolean,Integer> pos = binarySearch(v, 0, v.size(), key, (A k1, A k2) -> c.compare(k1, k2));
        if (!pos.getFirst())
        {
            v.insertElementAt(key, pos.getSecond());
        }
            
        return pos.getSecond();
        
    }

    static int toInsertionPoint(int pos)
    {
        return (-pos) - 1;
    }
    static public<A extends Comparable> Pair<Boolean,Integer>  binarySearch(Vector<A> v, A key) 
    {
        return binarySearch(v, 0, v.size(), key, (A a1, A a2) -> a1.compareTo(a2));
    }


    


    static public<A extends Comparable> Pair<Boolean,Integer>  binarySearch(Vector<A> v, A key, int fromIndex, int toIndex) 
    {
        return binarySearch(v, fromIndex, toIndex, key, (A a1, A a2) -> a1.compareTo(a2));
    }


    public interface Comparator2<A,B>
    {
        public int compare(A a, B b);
    }

    static public<A,B> Pair<Boolean,Integer> binarySearch(Vector<A> v, int fromIndex, int toIndex, B tosearch, Comparator2<A,B> c) 
    {
        int low = fromIndex;
        int high = toIndex - 1;
        
        while (low <= high) 
        {
            int mid = (low + high) >>> 1;
            A midVal = v.get(mid);
            
            int relation = c.compare(midVal, tosearch);
            if (relation < 0)
            {
                low = mid + 1;
                continue;
            }
               
            if (relation > 0)
            {
                high = mid - 1;
                continue;
            }
                
            return Pair.create(true, mid); // key found
        }
        return Pair.create(false, low);  // key not found.
    }

    /*
    static public<A> Pair<Boolean,Integer> binarySearch(Vector<A> v, int fromIndex, int toIndex, A key, Comparator<A> c) 
    {
        int low = fromIndex;
        int high = toIndex - 1;
        
        while (low <= high) 
        {
            int mid = (low + high) >>> 1;
            A midVal = v.get(mid);
            
            int relation = c.compare(midVal, key);
            if (relation < 0)
            {
                low = mid + 1;
                continue;
            }
               
            if (relation > 0)
            {
                high = mid - 1;
                continue;
            }
                
            return Pair.create(true, mid); // key found
        }
        return Pair.create(false, low);  // key not found.
    }
    */

}
