package lmu.utils;

import java.awt.Color;
import java.io.*;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.namespace.NamespaceContext;
import javax.xml.namespace.QName;
import javax.xml.stream.*;

import org.jdom.*;
import org.jdom.input.SAXBuilder;

public class XMLUtils
{
    static java.nio.charset.Charset DEFAULT_CHARSET = java.nio.charset.StandardCharsets.UTF_8;
    static final String COLOR_RGB_ATTRIB_NAME="color_rgb";
    static final String COLOR_HASH_ATTRIB_NAME="color_hash";
    static final String BASEPATH_ELEM_NAME="basepath";
    static final String NAME_ATTRIB_NAME="name";
    public static final String PATH_ATTRIB_NAME="path";
    static final String RELATIVE_ATTRIB_NAME="relative";
    static final String XMLREF_ATTRIB_NAME="xmlref";
    static final String VARDEF_ELEM_NAME="vardef";

    static final String UTF8 = "UTF-8";
    static ExceptionManager EM = new ExceptionManager();

    static public XMLInputFactory infactory = XMLInputFactory.newInstance();
    static public XMLOutputFactory outfactory = XMLOutputFactory.newInstance();

    public static class PrettyXMLStreamWriter implements XMLStreamWriter
    {
        XMLStreamWriter xsw;

        public void writeEmptyElement(String namespaceURI, String localName) throws XMLStreamException
        {
            xsw.writeEmptyElement(namespaceURI,localName);
        }

        public void writeEmptyElement(String prefix, String localName, String namespaceURI) throws XMLStreamException
        {
            xsw.writeEmptyElement( prefix,  localName,  namespaceURI);
        }


        public void writeEmptyElement(String localName) throws XMLStreamException
        {
            identbuff.append('\t');
            xsw.writeCharacters(identbuff.toString());
            //System.out.println("EMPTY, ident size: "+identbuff.length());
            xsw.writeEmptyElement(localName);
            xsw.writeCharacters("\n");
        }
        
        public void flush() throws XMLStreamException
        {
            xsw.flush();
        }
        public void writeAttribute(String namespaceURI,
                             String localName,
                             String value) throws XMLStreamException
        {
            xsw.writeAttribute(namespaceURI,localName,value);
        }
        public void writeAttribute(String prefix,
                             String namespaceURI,
                             String localName,
                             String value) throws XMLStreamException
        {
            xsw.writeAttribute(prefix,namespaceURI,localName,value);
        }
        public void writeAttribute(String localName, String value) throws XMLStreamException
        {
            xsw.writeAttribute(localName,value);
        }
        /**
         * Closes any start tags and writes corresponding end tags.
         * @throws XMLStreamException
         */
        public void writeEndDocument() throws XMLStreamException
        {
            xsw.writeEndDocument();
        }
        
        public void close() throws XMLStreamException
        {
            xsw.close();
        }
        
        public void writeNamespace(String prefix, String namespaceURI) throws XMLStreamException
        {
            xsw.writeNamespace(prefix,namespaceURI);
        }
        public void writeDefaultNamespace(String namespaceURI) throws XMLStreamException
        {
            xsw.writeDefaultNamespace(namespaceURI);
        }
         
        public void writeProcessingInstruction(String target, String data) throws XMLStreamException
        {
            xsw.writeProcessingInstruction(target,data);
        }

        public void writeProcessingInstruction(String data) throws XMLStreamException
        {
            xsw.writeProcessingInstruction(data);
        }
        
        public void writeCData(String data) throws XMLStreamException
        {
            xsw.writeCData(data);
        }
        public void writeComment(String data) throws XMLStreamException
        {
            xsw.writeComment(data);
        }
        public void writeStartDocument() throws XMLStreamException
        {
            xsw.writeStartDocument();
        }
        public void writeStartDocument(String version) throws XMLStreamException
        {
            xsw.writeStartDocument(version);
        }

        public void writeStartDocument(String version, String whatever) throws XMLStreamException
        {
            xsw.writeStartDocument(version, whatever);
        }
        public void writeDTD(String dtd) throws XMLStreamException
        {
            xsw.writeDTD(dtd);
        }
        public void writeEntityRef(String name) throws XMLStreamException
        {
            xsw.writeEntityRef(name);
        }
        public void writeCharacters(char[] text, int start, int len) throws XMLStreamException
        {
            xsw.writeCharacters(text,start,len);
        }
        public String getPrefix(String uri) throws XMLStreamException
        {
            return xsw.getPrefix(uri);
        }
        public void writeCharacters(String text) throws XMLStreamException
        {
            xsw.writeCharacters(text);
        }
        public void setNamespaceContext(NamespaceContext context) throws XMLStreamException
        {
            xsw.setNamespaceContext(context);
        }
        public NamespaceContext getNamespaceContext()
        {
            return xsw.getNamespaceContext();
        }
        public void setPrefix(String prefix, String uri) throws XMLStreamException
        {
            xsw.setPrefix(prefix,uri);
        }
        public void setDefaultNamespace(String uri) throws XMLStreamException
        {
            xsw.setDefaultNamespace(uri);
        }
        public Object getProperty(java.lang.String name) throws IllegalArgumentException
        {
            return xsw.getProperty(name);
        }
        
        StringBuffer identbuff = new StringBuffer();
        
        public void writeStartElement(String prefix, String localName, String namespaceURI) throws XMLStreamException
        {
            identbuff.append('\t');
            xsw.writeCharacters(identbuff.toString());
            
            xsw.writeStartElement(prefix, localName, namespaceURI);
        }
        public void writeStartElement(String namespaceURI, String localName) throws XMLStreamException
        {
            identbuff.append('\t');
            xsw.writeCharacters(identbuff.toString());
            xsw.writeStartElement(namespaceURI, localName);
        }
        public void writeStartElement(String localName) throws XMLStreamException
        {
            identbuff.append('\t');
            xsw.writeCharacters(identbuff.toString());
            //System.out.println("start: "+localName+", ident size: "+identbuff.length());
            //xsw.flush();
            xsw.writeStartElement(localName);
        }
        
        public void writeEndElement() throws XMLStreamException
        {
            xsw.writeEndElement();
            writeCharacters("\n");

            if (identbuff.length()>0)
            {
                identbuff.setLength(identbuff.length()-1);
            }
            
            
        }
        
            
        public PrettyXMLStreamWriter(XMLStreamWriter xsw)
        {
            this.xsw=xsw;
        }
    }
    public static PrettyXMLStreamWriter getXMLStreamWriter(OutputStream os) throws  XMLStreamException 
    {

        return new PrettyXMLStreamWriter(outfactory.createXMLStreamWriter(os));
    }
    static public Vector<Element> getElementsByAttribFilter(Element root, String elemname, String attrib, String value)
    {
        Vector<Element> v = new Vector<Element>();
        List l = root.getChildren(elemname);
        if (l==null)
            return v;
        
        for(Object o : l)
        {
            Element e = (Element)o;
            if (null==e.getAttribute(attrib))
                continue;
            
            
            if (!value.equals(e.getAttributeValue(attrib)))
                continue;
            
            v.add(e);
        }
        
        return v;
    }
    
    static public String getContent(Element root, String ... path)
    {
        Element e = root;

        if (path!=null)
        {
            Vector<String> testpath = new Vector<String>();
            testpath.add(root.getName());
            for (String p : path)
            {
                testpath.add(p);
                if(null == (e=e.getChild(p)))
                {
                    throw new RuntimeException("no elem found at path: "+testpath);
                }
            }
        }
        return ((Content)e.getContent().get(0)).getValue();
        
    }
    static public Element resolveAll(Element e)
    {
        return resolveAll(e, null);
    }
    
    static public Element resolveAll(Element e, File infile)
    {
        e=resolveVariables(e);
        e=resolveXmlRefs(e);
        e=resolvePaths(e, infile);

        return e;
    }
    static public Element getResolvedRootElement(File xml)
    {
        return resolveAll(getRootElement(xml), xml);
    }

    

    static public HashSet<Element> getDescentantsByName(Element e, String ... names)
    {
        HashSet<Element> rv = new HashSet<Element>();
        for (String name: names)
        {
            org.jdom.filter.ElementFilter f = new org.jdom.filter.ElementFilter(name);
            Iterator it = e.getDescendants(f);
            LinkedList<Element> l = new LinkedList<Element>();
    
            while (it.hasNext())
            {
                rv.add((Element)it.next());
            }
        }
        return rv;
    }

    static public String getAlternativeAttrib(Element e, String default_value, String ... alternatives)
    {
        String rv = null;
        for (String a : alternatives)
        {
            if(null!=(rv = e.getAttributeValue(a)))
            {
                return rv;
            }
        }

        
        return default_value;
    }
    
    static public Element getRootElement(File xml)
    {
        return getRootElement(FileUtils.getReader(xml, DEFAULT_CHARSET), " file: "+xml.getAbsolutePath());
    }

    static public Element getRootElement(byte[] data, String info)
    {
        return getRootElement(new InputStreamReader(new ByteArrayInputStream(data), DEFAULT_CHARSET),info);
    }

    static public Element getRootElement(Reader reader, String info)
    {
        try
        {
            SAXBuilder sb = new SAXBuilder();
            
            Document doc = sb.build(reader);
            reader.close();
            return doc.getRootElement();
        }
        catch(java.io.IOException e)
        {
            throw new FRuntimeException("i/o: %s error while reading from xml: %s ", e, e.getMessage(), info);
        }
        catch(JDOMException je)
        {
            throw new FRuntimeException("xml error: %s while trying to reading from xml: %s ", je,  je.getMessage(), info);
        }
        

    }
    public static void encodeAttrib(Element e, String key, String value)
    {
        try
        {
            e.setAttribute(key, URLEncoder.encode(value, UTF8));
        }
        catch(java.io.UnsupportedEncodingException ex)
        {
            throw new RuntimeException("while encodeing "+value+" with "+UTF8, ex);
        }
        
    }

    public static String decodeAttrib(Element e, String key)
    {
        Attribute a = e.getAttribute(key);
        if (a==null)
            return null;
        String value = a.getValue();

        try
        {

            return URLDecoder.decode(value, UTF8);
        }
        catch(java.io.UnsupportedEncodingException ex)
        {
            throw new RuntimeException("while encodeing "+value+" with "+UTF8, ex);
        }
        
            
    }
    public static void copyAttribs(Element src, Element dst)
    {
        for (Object o : src.getAttributes())
        {
            if (!(o instanceof Attribute))
            {
                continue;
            }
            Attribute a = (Attribute)o;
            dst.setAttribute(a.getName(), a.getValue());

        }
    }
        
            
     public static File readOptionalPath(Element e, File f)
     {
         return readOptionalPath(e,f, true);
     }


     public static File readChildPath(Element e, String ch)
     {
         return readPath(e, ch);
     }

     public static File readPath(Element e, String ... xml_path)
     {
         Element sube = e;
         for (String s : xml_path)
         {
             sube = sube.getChild(s);
             if (sube == null)
                 throw new FRuntimeException("xml elem: %s not found in path %s", s, StringUtils.join(",", xml_path));
         }
         if (sube == null)
             throw new FRuntimeException("no xml elem given to read path: %s", StringUtils.join(",", xml_path));

         
         return readPath(sube);
     }

     public static File readPath(Element e)
     {
         return readOptionalPath(e, null, false);
     }

     public static File readRelativePath(Element e, File dir)
     {
         return readOptionalPath(e, null, false, dir);
     }

     static File readOptionalPath(Element e, File def_file, boolean optional)
     {
         return readOptionalPath(e,def_file,optional,null);
     }
     static File readOptionalPath(Element e, File def_file, boolean optional, File parentdir)
     {
         String s = getAttrib(e, PATH_ATTRIB_NAME, null, false);
         if (!optional && s==null)
         {
             throw new RuntimeException("expected attrib: "+PATH_ATTRIB_NAME+" is not set for elem:"+e.getName()+" attribs: "+getAttribs(e));
         }
         if (s==null)
             return def_file;
         
         return (parentdir != null) ? new File(parentdir,s) : new File(s);
     }


     static Pattern VAR_REF_PATTERN=Pattern.compile("\\$\\{([^\\}]+)\\}");

     static String resolveVars(String value, HashMap<String,String> vardefs)
     {
         //System.out.println("will resolve: >"+value+"< vardefs: "+vardefs);
         Matcher m = VAR_REF_PATTERN.matcher(value);
         
         int start=0;
         StringBuffer resolved = new StringBuffer();

         while (m.find(start))
         {
             //System.out.printf("got var pattern hit:>%s<\n", m.group(1));
             resolved.append(value.substring(start, m.start()));

             String key = m.group(1);
             String varvalue = vardefs.get(key);
             if (varvalue==null)
             {
                 //System.out.println("no resolution to "+key+" known resols: "+vardefs);
                 return null;
             }
             resolved.append(varvalue);
             start=m.end();
         }

         resolved.append(value.substring(start));
         //System.out.printf("resolved >%s< to >%s<\n", value, resolved.toString());
         return resolved.toString();
     }
     /** resolves variables 
      * defined by <vardef name="" value=""/>
      * 
      * any variable can referenced by ${varname} in any attribute
      */
     public static Element resolveVariables(Element rootElem)
     {
         HashMap<String, String> varname2value = new HashMap<String, String>();

         HashMap<String, String> var2def = new HashMap<String, String>();
         Iterator it = rootElem.getDescendants();
         
         while (it.hasNext())
         {
           Object o = it.next();
           if (! (o instanceof Element) )
               continue;
           
           
           Element e = (Element)o;
           if (!e.getName().equals(VARDEF_ELEM_NAME))
           {
               continue;
           }
           String name = getAttrib(e,"name", null, true);
           String value = getAttrib(e,"value", null, true);

           String resolved = resolveVars(value, varname2value);
           if (resolved==null)
           {
               //not yet resolved
               var2def.put(name, value);
               
               continue;
           }
           //System.out.println("add resolved: "+name+"->"+resolved);
           varname2value.put(name, resolved);
        
         }

         //System.out.println("start resolving not yet resolved vars:\n"+var2def+"\n\nresolved so far:\n"+varname2value+"\n");
         while (var2def.size()>0)
         {
             HashSet<String> resolved = new HashSet<String>();
             for (Map.Entry<String,String> e : var2def.entrySet())
             {
                 String res = resolveVars(e.getValue(), varname2value);
                 if (res==null)
                 {
                     //still not resolved
                     continue;
                 }
                 varname2value.put(e.getKey(), res);
                 resolved.add(e.getKey());
                 
             }
             if (resolved.size()==0)
             {
                 throw new RuntimeException("cannot resolve variable referencing variables: "+var2def+" resolved: "+varname2value); 
             }
             for (String s : resolved)
             {
                 var2def.remove(s);
             }
         }

         //no resolve all attribs referencing something
         
         it=rootElem.getDescendants();

         while (it.hasNext())
         {
           Object o = it.next();
           if (! (o instanceof Element) )
               continue;
           
           
           Element e = (Element)o;
           List<Attribute> al = (List<Attribute>)e.getAttributes();
           for (Attribute a: al)
           {
               String res = resolveVars(a.getValue(),varname2value);
               if (res==null)
               {
                   throw new RuntimeException("cannot resolve variable reference: : "+a.getValue()+" resolved: "+varname2value); 
               }
               a.setValue(res);
           }
         }
         return rootElem;
     }
     /** resolves xmlreferences defined in the xml - i.e. references of attributes
      */

     public static Element resolveXmlRefs(Element rootElem)
     {
         HashMap<String, Element> elemname2elem = new HashMap<String, Element>();

         LinkedList<Element> xref_elems = new LinkedList<Element>();

         Iterator it = rootElem.getDescendants();
         while (it.hasNext())
         {
           Object o = it.next();
           if (! (o instanceof Element) )
               continue;
           
           
           Element e = (Element)o;
           if (e.getAttribute(XMLREF_ATTRIB_NAME)!=null)
           {
               xref_elems.add(e);
               continue;
           }
           elemname2elem.put(e.getName(),e);
        
         }

         while (!xref_elems.isEmpty())
         {
             Element e = xref_elems.poll();
             String ename = getAttrib(e,XMLREF_ATTRIB_NAME,  null, true);
             Element base = elemname2elem.get(ename);
             if (base==null)
             {
                 throw new RuntimeException("referenced elem name "+ename+" not found!");
             }
             //System.out.println("before extending: "+getAttribs(e));
             for (Object o : base.getAttributes())
             {
                 Attribute a = (Attribute)o;
                 if (e.getAttribute(a.getName()) != null) //don't overwrite explicitely defined attributes
                     continue;
                 e.setAttribute(a.getName(), a.getValue());
             }
             //System.out.println("after extending: "+getAttribs(e));
         }

         return rootElem;

     }
    /** resolves relative pathes defined in the xml. 
     * to do so it relieas on two assumptions:
     *  (1) _if_ there are general pathes they are defined with BASEPATH_ELEM_NAME
     *  having two attributes "name" and "path"  if any of them is missing,
     * the method will skip this definition
     *  (2) the resolution of the path is needed iff any elem has 
     * the attribs "relative" and "path" whereas realtive refers to a "name" defined
     *  in (1)
     * 
     * if (1) and (2) are fulfilled, the method will replace the value of "path" for
     * elems fulfilling (2) by the absolute path
     * 
     */
     public static Element resolvePaths(Element rootElem)
     {
         return resolvePaths(rootElem, null);
     }
     
    public static Element resolvePaths(Element rootElem, File infile)
    {
       HashMap<String, File> name2root = new HashMap<String, File>();
       HashMap<String, String> name2path = new HashMap<String, String>();
       HashMap<String, String> name2rel = new HashMap<String, String>();

       //round 1, read the basepaths
       Iterator it = rootElem.getDescendants();
       String name=null;
       String path = null;

       if (infile != null)
       {
           File pardir = (infile.getParentFile() != null) ? infile.getParentFile() : new File(".");
           name2path.put(".", pardir.getAbsolutePath());
           name2root.put(".", pardir);
       }
       

       while (it.hasNext())
       {
           Object o = it.next();
           if (! (o instanceof Element) )
               continue;
           
           
           Element e = (Element)o;
           if (!e.getName().equals(BASEPATH_ELEM_NAME))
               continue;
                                             
           if ( null==(name=getAttrib(e,NAME_ATTRIB_NAME, null,false))
                ||
                null==(path=getAttrib(e,PATH_ATTRIB_NAME, null,false))
               )
           {
               continue; //not properly set
           }

           
           String relative = getAttrib(e, RELATIVE_ATTRIB_NAME, null, false);

           if (relative==null)
           {
               //System.out.println("found realtive elem: "+e.getName()+ " attribs: "+getAttribs(e));
               name2root.put(name, new File(path));
               continue;
           }
               
           name2path.put(name, path);
           
           
           //the basepath itself refers to an other basepath
           name2rel.put(name, relative);
           
       }

       /*if (name2path.size()==0)
           return rootElem; //nothing to do here...
           */
           
       //System.out.println("got relative bases to resolve: "+name2path+" rel: "+name2rel);
       
       /** resolve the dependencies of basepathes */
       while (name2rel.size()>0)
       {
           
           String relname=null;
           File base = null;
           for (Map.Entry<String,String> e: name2rel.entrySet())
           {
               if (name2rel.get(e.getValue())!=null) //the base path itself is not yet resolved
                   continue;
               
               //base is resolved (if set at all), so we can resolve it now
               if (null == (base = name2root.get(e.getValue())))
                   continue; //not set at all

               //we found base 
               relname=e.getKey();
           }
           if (relname==null)
           {
               throw new RuntimeException("missing path references: "+ name2path +" rel: "+ name2rel );
           }
               
           
           
           String relpath = name2path.get(relname);
           name2root.put(relname, new File(base, relpath));
           name2rel.remove(relname);
           //debug
           //System.out.println("resolve dependent pathname "+relname+": "+relpath+" to "+name2root.get(relname).getAbsolutePath());
           
       }

       
       //second round, change the elems needed
       File base = null;
       it = rootElem.getDescendants();
       while (it.hasNext())
       {
           Object o = it.next();
           if (! (o instanceof Element) )
               continue;

           Element e = (Element)o;

           if ( null==(name=getAttrib(e,RELATIVE_ATTRIB_NAME, null,false))
             ||
             null==(path=getAttrib(e,PATH_ATTRIB_NAME, null,false))
              )
            {
                continue; //nothing to replace
            }


           if (null == (base=name2root.get(name)))
           {
               if (infile != null && name != null && name.equals("."))
               {
                   base = infile.getParentFile();
               }
           }

           if (base == null)
           {
               System.err.println("no root found to "+name);
               continue;
           }
               
           Attribute a = e.getAttribute(PATH_ATTRIB_NAME);
           File absolute = new File(base, path);
           a.setValue(absolute.getAbsolutePath());
           //debug 
           //System.out.println("set "+name+" and "+path+" to "+absolute.getAbsolutePath());
       }
       return rootElem; 
    }
    public static String getAttribs(Element e)
    {
        StringBuffer sb = new StringBuffer();
        for (Object o : e.getAttributes())
        {
            Attribute a = (Attribute)o;
            sb.append(String.format(" %s=%s", a.getName(), a.getValue()));

        }
        return sb.toString();
    }

    public static String getAttribs(XMLStreamReader xsr)
    {
        BufferPrintWriter bpw = new BufferPrintWriter();
        bpw.printf("elem: %s ", xsr.getName());
        for (int i=0; i<xsr.getAttributeCount(); i++)
        {
            bpw.printf(" %s = %s ", xsr.getAttributeLocalName(i), xsr.getAttributeValue(i));
        }
        return bpw.getBuffer();
    
    }

    public static String getTextValueByPath(Element e, String ... elems)
    {
        Element sub = getByPath(e,elems);
        if (sub==null)
            return null;
        
        
        return sub.getText();
    }

    public interface ElementFilter
    {
        public boolean accept(Element e);
    }
    
    public static Vector<Element> getDescendantsByName(Element e, String name)
    {
        return getDescendantsByName(e,name, null);
    }

    public static Vector<Element> getDescendantsByName(Element e, String name, ElementFilter filter)
    {

        Vector<Element> v = new Vector<>();

        Iterator it = e.getDescendants();

        while (it.hasNext())
         {
           Object o = it.next();
           if (! (o instanceof Element) )
               continue;
           
           
           Element ee = (Element)o;
           if (!ee.getName().equals(name))
               continue;
           
           if (filter != null && ! filter.accept(ee))
               continue;
           
           v.add(ee);
        }

        return v;

    }
    public static Element getByPath(Element e, String ... elems)
    {
        if (elems == null)
            return e;
        
        Element p = e;
        for (int i=0; i<elems.length; i++)
        {
            p=p.getChild(elems[i]);
            if (p==null)
                return null;
            
            
        }
        return p;
    }

    public static Character getChrAttrib(Element e, String name, Character def, boolean throw_exc)
    {
        String s = getAttrib(e, name, ""+def,throw_exc);
        if (s==null || s.length() == 0)
            return def;
        
        return s.charAt(0);
    }
    public static Character getChrAttrib(XMLStreamReader xsr, String name, Character def, boolean throw_exc)
    {
        String s = getAttrib(xsr,name, ""+def,throw_exc);
        if (s==null || s.length() == 0)
            return def;
        
        return s.charAt(0);
    }

    
    public static String getAttrib(XMLStreamReader xsr, String name, String def, boolean throw_exc)
    {
        String v = xsr.getAttributeValue(null, name);
        if (v==null && throw_exc)
        {
            EM.throwRuntimeException("cannot read attrib: "+name+" from element: "+xsr.getName());
        }
        if (v==null)
            return def;
        return v;
    }

    public static String getFromMultipleAttrib(Element e, String def, boolean throw_exc, String ... optional_names)
    {
        for (String on : optional_names)
        {
            String rv = getAttrib(e, on, null , false);
            if (rv !=null)
                return rv;
        }
        if (throw_exc)
            EM.throwRuntimeException("cannot read attrib with any of the names: "+StringUtils.join(",", optional_names)+" from element: "+e.getName());
        
        
        return def;
    }


    public static String getFromMultipleAttrib(XMLStreamReader xsr, String def, boolean throw_exc, String ... optional_names)
    {
        for (String on : optional_names)
        {
            String rv = getAttrib(xsr, on, null , false);
            if (rv !=null)
                return rv;
        }
        if (throw_exc)
            EM.throwRuntimeException("cannot read attrib with any of the names: "+StringUtils.join(",", optional_names)+" from element: "+xsr.getName());
        
        
        return def;
    }
    
    public static Vector<String> getVectorAttrib(Element e, String name, Vector<String> def, boolean throw_exc, String splitter)
    {
        String v = getAttrib(e,name, null, throw_exc);
        if (v == null)
            return def;
        
        return ObjectGetter.toVector(FileUtils.RecordIterator.split(splitter, v));
    }
    
    public static Vector<String> getVectorAttrib(Element e, String name, Vector<String> def, boolean throw_exc, String splitter, boolean removeEnclosingParenthesis)
    {
        String v = getAttrib(e,name, null, throw_exc);

        if (v == null)
            return def;

        if (removeEnclosingParenthesis)
            v = v.substring(1,v.length()-2);
        
        return ObjectGetter.toVector(FileUtils.RecordIterator.split(splitter, v));
    }

    public static String getAttrib(Element e, String name, String def, boolean throw_exc)
    {
        Attribute a = e.getAttribute(name);
        if (a==null && throw_exc)
        {
            EM.throwRuntimeException("cannot read attrib: "+name+" from element: "+e.getName());
        }
        if (a==null)
            return def;
        
        
        return a.getValue();
    }

    public static Boolean getOptionalBooleanAttrib(Element e, String bname)
    {
        String bn = getAttrib(e, bname, null, false);
        
        return (bn == null) ? null : Boolean.parseBoolean(bn);
    }

    public static boolean getBooleanAttrib(Element e, String bname, Boolean def)
    {
        return Boolean.parseBoolean(getAttrib(e, bname, ""+def, def == null));
    }

    public static boolean getBooleanAttrib(XMLStreamReader xsr, String bname, Boolean def)
    {
        return Boolean.parseBoolean(getAttrib(xsr, bname, ""+def, def == null));
    }

    public static String toColorAttrib(java.awt.Color c)
    {
       return String.format("%X", c.getRGB()).substring(2);
    }
    public static java.awt.Color getColorAttrib(Element e, String colattribname, java.awt.Color def)
    {
        String  c = getAttrib(e,colattribname, null, false);
        return ImageUtils.getColor(c,def);
    }

    public static java.awt.Color getColorAttrib(XMLStreamReader e, String colattribname, java.awt.Color def)
    {
        String  c = e.getAttributeValue(null, colattribname);
        if (c==null)
        {
            return def;
        }
        
        return ImageUtils.getColor(c,def);
    }

    public static Integer getIntAttrib(Element e, String intattrib)
    {
        return getIntAttrib(e,intattrib, null, true);
    }

    public static Integer getIntAttrib(Element e, String intattrib, Integer defvalue)
    {
        return getIntAttrib(e,intattrib,defvalue, false);
    }

    public static Integer getIntAttrib(Element e, String intattrib, Integer defvalue, boolean error_if_not_set)
    {
        Attribute a = e.getAttribute(intattrib);
        if (a==null)
        {
            if (error_if_not_set)
            {
                throw new FRuntimeException("missing expected int attrib : %s attribs set : %s", intattrib, getAttribs(e));
            }
            return defvalue;
        }
            
        try
        {
             return Integer.parseInt(a.getValue());
        }
        catch(Exception ex)
        {
            EM.throwRuntimeException("could not parse integer value for attrib: "+intattrib+" value: >"+a.getValue()+"<");
        }
        
        return null;
    }

    public static double getDoubleAttrib(Element e, String dattrib)
    {
        Double d = getDoubleAttrib(e,dattrib,null);
        if (d==null)
        {
            throw new RuntimeException("no double value found for attrib: "+dattrib);
        }
        return d;
    }

    public static double getDoubleAttrib(XMLStreamReader xsr, String dattrib)
    {
        Double d = getDoubleAttrib(xsr,dattrib,null);
        if (d==null)
        {
            throw new RuntimeException("no double value foudn for attrib: "+dattrib);
        }
        return d;
    }
    public static Double getDoubleAttrib(XMLStreamReader xsr, String dattrib, Double defvalue)
    {
        String v = xsr.getAttributeValue(null, dattrib);
        if (v==null)
        {
            return defvalue;
        }
        try
        {
             return Double.parseDouble(v);
        }
        catch(Exception ex)
        {
            EM.throwRuntimeException("could not parse double value for attrib: "+dattrib+" value: >"+v+"<");
        }
        return null;
    }


        
    public static int getIntAttrib(XMLStreamReader xsr, String dattrib)
    {
        return getIntAttrib(xsr, dattrib, null, true);
    }
    
    public static Integer getIntAttrib(XMLStreamReader xsr, String dattrib, Integer defvalue)
    {
        return getIntAttrib(xsr, dattrib, defvalue, false);
    }
    public static Integer getIntAttrib(XMLStreamReader xsr, String dattrib, Integer defvalue, boolean error_if_not_there)
    {
        String v = xsr.getAttributeValue(null, dattrib);
        if (v==null)
        {
            if (error_if_not_there)
            {
                throw new FRuntimeException("missing expected int attrib : %s attribs set : %s", dattrib, getAttribs(xsr));
            }
            return defvalue;
        }
        try
        {
             return Integer.parseInt(v);
        }
        catch(Exception ex)
        {
            EM.throwRuntimeException("could not parse int value for attrib: "+dattrib+" value: >"+v+"<");
        }
        return null;
    }



    public static Long getLongAttrib(XMLStreamReader xsr, String attrib)
    {
        Long d = getLongAttrib(xsr,attrib,null);
        if (d==null)
        {
            throw new RuntimeException("no long value foudn for attrib: "+attrib);
        }
        return d;
    }
    
    public static Long getLongAttrib(XMLStreamReader xsr, String dattrib, Long defvalue)
    {
        String v = xsr.getAttributeValue(null, dattrib);
        if (v==null)
        {
            return defvalue;
        }
        try
        {
             return Long.parseLong(v);
        }
        catch(Exception ex)
        {
            EM.throwRuntimeException("could not parse long value for attrib: "+dattrib+" value: >"+v+"<");
        }
        return null;
    }

    public static Double getDoubleAttrib(Element e, String dattrib, Double defvalue)
    {
        Attribute a = e.getAttribute(dattrib);
        if (a==null)
            return defvalue;
        
        try
        {
             return Double.parseDouble(a.getValue());
        }
        catch(Exception ex)
        {
            EM.throwRuntimeException("could not parse double value for attrib: "+dattrib+" value: >"+a.getValue()+"<");
        }
        
        return null;
    }

    public static Long getLongAttrib(Element e, String dattrib, Long defvalue)
    {
        Attribute a = e.getAttribute(dattrib);
        if (a==null)
            return defvalue;
        
        try
        {
             return Long.parseLong(a.getValue());
        }
        catch(Exception ex)
        {
            EM.throwRuntimeException("could not parse long value for attrib: "+dattrib+" value: >"+a.getValue()+"<");
        }
        
        return null;
    }

    public static void writeElem2Stream(XMLStreamWriter xsw, Element e) throws XMLStreamException
    {
        List ch = e.getChildren();
        boolean empty=ch.size()==0;
        if (empty)
        {
            xsw.writeEmptyElement(e.getName());
        }
        else
        {
            xsw.writeStartElement(e.getName());
        }
        for (Object o : e.getAttributes())
        {
            Attribute a = (Attribute) o;
            xsw.writeAttribute(a.getName(), a.getValue());
        }

        for (Object o : ch)
        {
            if (o instanceof Element)
            {
                writeElem2Stream(xsw, (Element)o);
                continue;
            }
            if (o instanceof CDATA)
            {
                xsw.writeCData(((CDATA)o).getText());
                continue;
            }                
            
            if (o instanceof Text)
            {
                xsw.writeCharacters(((Text)o).getText());
                continue;
            }

            throw new RuntimeException("unsupported child type: "+o.getClass().getCanonicalName());
        }
        if (!empty)
        {
            xsw.writeEndElement();
        }
        
        
    }
    public static String getElemInfo(Element e)
    {
        BufferPrintWriter bpw = new BufferPrintWriter();
        bpw.println("elem="+ e.getName());
        bpw.println(" attribs: ");
        for (Attribute a : (List<Attribute>)e.getAttributes())
        {
            bpw.printf("\t%s = %s\n", a.getName(), a.getValue());
        }
        return bpw.getBuffer();
    }
    public static String getElemInfo(XMLStreamReader xsr) throws XMLStreamException
    {
        StringBuffer sb = new StringBuffer();
        sb.append("elem="+xsr.getLocalName());
        sb.append(" attribs: ");
        for (int i=0; i<xsr.getAttributeCount(); i++)
        {
            if (i>0)
            {
                sb.append(", ");
            }
            sb.append(String.format("%s=%s",xsr.getAttributeLocalName(i),xsr.getAttributeValue(i)));
        }
        return sb.toString();
    }

    static Element prepareElem(XMLStreamReader xsr) throws XMLStreamException
    {
        
        Element e = new Element(xsr.getLocalName());
        
        for (int i=0; i<xsr.getAttributeCount(); i++)
        {
            e.setAttribute(xsr.getAttributeLocalName(i),xsr.getAttributeValue(i));
        }
        return e;
    }
    
    
    public static Element readElemFromStream(XMLStreamReader xsr) throws XMLStreamException
    {
        return readElemFromStream(xsr,false);
    }
    public static Element readElemFromStream(XMLStreamReader xsr, boolean verbose) throws XMLStreamException
    {
        if (verbose)
        {
            System.out.println("PREPARE ROOT: "+xsr.getName()+"\n");
        }
        
        Element root = prepareElem(xsr);
        
        int nopen=1;
        LinkedList<Element> q = new LinkedList<Element>();
        q.add(root);

        while (xsr.hasNext())
        {
            int event = xsr.next();
            
            if (verbose)
            {
                System.out.println("next event: "+event);
            }
            

            if (verbose && (event == XMLStreamConstants.END_ELEMENT || event == XMLStreamConstants.START_ELEMENT))
            {
                System.out.printf("\tevent: %s name: %s q: %s peek: %s\n", 
                                  (event == XMLStreamConstants.END_ELEMENT)?"END":"OPEN",
                                  xsr.getName(), q, q.peek());
                System.out.println("PEEK pretty:\n"+prettyXML(q.peek())+"\n\n");

            }
            if (event == XMLStreamConstants.END_ELEMENT)
            {
                Element te = q.pop();

                if (q.size()==0)
                {
                    if (verbose)
                    {
                    
                        System.out.println("WILL RETURN ROOT\n\n\n"+prettyXML(root)+"\n\nlast popped:\n"+prettyXML(te)
                                           +"\n\n");
                    }
                    return root;
                }
                continue;
            }
            switch (event)
            {
                case XMLStreamConstants.START_ELEMENT:
                    Element sub = prepareElem(xsr);
                    q.peek().addContent(sub);
                    if (verbose)
                    {
                        System.out.println("added "+sub.getName()+" to "+ q.peek().getName()+" q.peek(): "+ prettyXML(q.peek()));
                    }

                    q.push(sub);
                    continue;
                case XMLStreamConstants.CHARACTERS:
                    if (xsr.getText().trim().length()==0)
                    {
                        continue;
                    }
                    if (verbose)
                    {
                        System.out.println("add chars: >"+xsr.getText()+"< to "+q.peek().getName());
                    }
                    
                    q.peek().setText(xsr.getText());
                    continue;
                case XMLStreamConstants.CDATA:
                    if (verbose)
                    {
                        System.out.println("add cdata: "+xsr.getText()+" to "+q.peek().getName());
                    }
                    
                    q.peek().addContent(new CDATA(xsr.getText()));
                    continue;
                default:
                    continue;

            }
            
        }
        if (verbose)
        {
            System.out.println("WILL RETURN ROOT\n\n\n"+prettyXML(root)+"\n\n\n");
        }
        

        return root;
    }

    final static org.jdom.output.XMLOutputter pretty_out = new org.jdom.output.XMLOutputter();
    
    static
    {
        org.jdom.output.Format.getPrettyFormat().setLineSeparator("\n");
        pretty_out.setFormat(org.jdom.output.Format.getPrettyFormat());

    }
    public static InputStream prettyXMLStream(Element e)
    {
        
        String pretty = prettyXML(e);
        return new ByteArrayInputStream(pretty.getBytes());
        
    }
    public static void writeXML(Element e, File f)
    {
        try
        {
            PrintWriter pw = new PrintWriter(f);
            writeXML(e,pw);
            pw.close();
        }
        catch(IOException ex)
        {
            throw new RuntimeException("error:"+ex.getMessage()+" while writing " +e.getName()+" to "+f.getAbsolutePath(), ex);
        }
    }
    public static void writeXML(Element e, PrintWriter pw)
    {
        pw.print(prettyXML(e));
    }
    
    public static String prettyXML(Element e)
    {
        StringWriter sw = new StringWriter();
        try
        {
            pretty_out.output(e, sw);
        }
        catch(IOException ex)
        {
            throw new RuntimeException("error:"+ex.getMessage()+" while pretty formatting "+e.getName(), ex);
        }
        
        return sw.toString();
    }

    public static class Stream2ElemWriter
    {
        StringWriter sw = new StringWriter();
    
        XMLStreamWriter xsw = null;
    
        public Stream2ElemWriter(boolean create_doc)
        {
            try
            {
                xsw = XMLUtils.outfactory.createXMLStreamWriter(sw);
                if (create_doc)
                {
                    xsw.writeStartDocument();
                }
            }
            catch(XMLStreamException xse)
            {
                throw new RuntimeException("xml error while creating bufferxml stream writer", xse);
            }
            
        }
    
        public XMLStreamWriter getWriter()
        {
            return xsw;
        }
    
        public Element getElem()
        {
            try
            {
                xsw.close();
            }
            catch(XMLStreamException xse)
            {
                throw new RuntimeException("xml error while closing bufferxml stream writer", xse);
            }
            Element e = XMLUtils.makeElement(sw.getBuffer().toString());
            e.detach();
            return e;
        }
    }

    static public XMLStreamWriter getWriter(Writer wr) 
    {
        try
        {
            return outfactory.createXMLStreamWriter(wr);
        }
        catch(XMLStreamException xse)
        {
            throw new FRuntimeException("xml error: %s while opening writer", xse, xse.getMessage());
        }
        
    }

    static public XMLStreamWriter getWriter(File f) 
    {
        try
        {
            return outfactory.createXMLStreamWriter(FileUtils.getWriter(f));
        }
        catch(XMLStreamException xse)
        {
            throw new FRuntimeException("xml error: %s while opening writer", xse, xse.getMessage());
        }
        
    }

    static public XMLStreamReader getReader(Reader r) 
    {
        try
        {
            return infactory.createXMLStreamReader(r);
        }
        catch(XMLStreamException xse)
        {
            throw new FRuntimeException("xml error: %s while opening reader", xse, xse.getMessage());
        }
        
    }

    public interface XMLElemReader<A>
    {
        public A read(XMLStreamReader xsr) throws XMLStreamException;
    }

    public interface XMLDomElemReader<A>
    {
        public A read(Element e);
    }

    static class XMLElemIterator<A> implements Iterator<A>
    {
        String srcname;
        XMLStreamReader xsr;
        XMLElemReader<A> converter;
        String elem_name;
        String namespace;
        boolean got_next = false;
        

        public XMLElemIterator(String srcname, XMLStreamReader xsr, XMLElemReader<A> converter, String elem_name, String namespace)
        {
            this.srcname = srcname;
            this.xsr = xsr;
            this.converter = converter;
            this.elem_name = elem_name;
            this.namespace = namespace;
            got_next = searchNext();
        }

        public boolean hasNext()
        {
            return got_next;
        }

        public A next()
        {
            if(!got_next)
                return null;


            try
            {
                A rv = converter.read(xsr);
                got_next = searchNext();
                return rv;
            }
            catch(XMLStreamException xse)
            {
                throw new FRuntimeException("xml error: %s while converting next elem: %s (ns: %s)", xse, xse.getMessage(), elem_name, namespace);
            }
            
        }

        public boolean searchNext()
        {
            if (xsr == null)
                return false;
            
            
            try
            {

                //@TODO -make it namespace aware 
                QName qn = (namespace == null) ? null : new QName(namespace, elem_name);
                boolean got = false;
                
                while (xsr.hasNext())
                {
                    int e = xsr.next();
                    if (e != XMLStreamConstants.START_ELEMENT)
                        continue;
                    
                    if (!xsr.getLocalName().equals(elem_name))
                        continue;
                    
                    if (qn != null && !qn.equals(xsr.getName()))
                        continue;
                    
                    
                    return true;
                }
                xsr = null;
                return false;
            }
            catch(XMLStreamException xse)
            {
                throw new FRuntimeException("xml error: %s while searching for next elem: %s (ns: %s) source: %s", xse, xse.getMessage(), elem_name, namespace, (srcname == null) ? "unknown" : srcname);
            }
            catch(Exception e)
            {
                throw new FRuntimeException("error: %s while searching for next elem: %s (ns: %s) source: %s ", e, e.getMessage(), elem_name, namespace, (srcname == null) ? "unknown" : srcname);
            }
        }
        
    }

    static public<A> Iterator<A> getConvertedIterator(String srcname, XMLStreamReader xsr, String name, XMLElemReader<A> converter)
    {
        return getConvertedIterator(srcname, xsr, name, null, converter);
    }

    static public<A> Iterator<A> getConvertedIterator(String srcname, XMLStreamReader xsr, String name, String namespace, XMLElemReader<A> converter)
    {
        return new XMLElemIterator<A>(srcname, xsr, converter, name, namespace);
    }


    static public<A> Iterator<A> getConvertedIterator(File f, String name, String namespace, XMLElemReader<A> converter)
    {
        return getConvertedIterator(f.getName(), getReader(FileUtils.getReader(f, DEFAULT_CHARSET)), name, namespace, converter);
    }

    static public<A> Iterator<A> getConvertedIterator(File f, String name, XMLElemReader<A> converter)
    {
        return getConvertedIterator(f, name, null, converter);
    }

    static public<A> Iterator<A> getConvertedIterator(File f, String name, XMLDomElemReader<A> converter)
    {
        return getConvertedIterator(f, name, null, (XMLStreamReader xsr) -> converter.read(readElemFromStream(xsr)));
    }

    static public Iterator<Element> getElementIterator(File f, String name)
    {
        return getConvertedIterator(f, name, null, (XMLStreamReader xsr) -> readElemFromStream(xsr));
    }

    static public XMLStreamReader getReader(Element e) throws XMLStreamException
    {
        XMLStreamReader xsr = infactory.createXMLStreamReader(new StringReader(XMLUtils.prettyXML(e)));

        while (xsr.hasNext())
        {
            int event = xsr.next();
            if (event != XMLStreamConstants.START_ELEMENT)
                continue;
            
            return xsr;
        }
        return null;
    }
    
    static public Element makeElement(String s)
    {
        return makeElement(new StringReader(s));
    }

    

    static public Element makeElement(InputStream stream)
    {
        return makeElement(new InputStreamReader(stream, DEFAULT_CHARSET));
    }
    
    static public Element makeElement(Reader r)
    {
        try
        {
            SAXBuilder sb = new SAXBuilder();
            Document doc = sb.build(r);
            return doc.getRootElement();
        }
        catch(IOException e)
        {
            throw new RuntimeException("i/o error while reading from stream...: ", e);
        }
        catch(JDOMException e)
        {
            throw new RuntimeException("error while trying to read xml from stream", e);
        }
    }

    public static LinkedList<Element> getChildren(Element e)
    {
        return getChildren(e,null);
    }
    public static LinkedList<Element> getChildren(Element e, String tagname)
    {
        java.util.List chl = (tagname==null)?e.getChildren():e.getChildren(tagname);
        LinkedList<Element> rv = new LinkedList<Element>();
        for (Object o : chl)
        {
            if (!(o instanceof Element))
                continue;
            
            rv.add((Element)o);
        }
        return rv;
    }

    

    /** checks for following:
     *  color=xxx -> name of color
     *  name=xxx -> name of color
     * 
     * COLOR_RGB_ATTRIB_NAME -> rgb
     * COLOR_HASH_ATTRIB_NAME-> hashcode in form #FF00FF
     * 
     */
    public static Color getColor(Element e, Color default_color, boolean throw_exc)
    {
        String rgb =  getAttrib(e,COLOR_RGB_ATTRIB_NAME, null, false);
        if (rgb!=null)
        {
            return new java.awt.Color(Integer.parseInt(rgb));
        }
    
    
        String hash =  getAttrib(e,COLOR_HASH_ATTRIB_NAME, null, false);
    
        if (hash!=null)
        {
            throw new RuntimeException("color has decoding not yet implemented");
        }
    
    
        String cn = getAttrib(e, "color", null, false);
        if (cn!=null)
        {
            return ColorManager.getColor(cn);
        }

        cn = getAttrib(e, "name", null, false);
        if (cn!=null)
        {
            return ColorManager.getColor(cn);
        }

        if (throw_exc)
        {
            throw new RuntimeException("cannot decode color from element: "+e.getName()+" attribs "+getAttribs(e));
        }
        return default_color;
    }

    public static Vector<Integer> getIntAttribVector(Element e , String attrib, String split_pattern)
    {
        return (Vector<Integer>)getAttribVector(e,attrib,split_pattern,SimpleTypeConverter.getConverter("int", true));
    }
    public static Vector getAttribVector(Element e , String attrib, String split_pattern, StringConverter sc)
    {
        Attribute a = e.getAttribute(attrib);
        if (a==null)
        {
            throw new RuntimeException("attribute: "+attrib+" not found in "+e.getName());
        }
        if (a.getValue().length()==0)
        {
            return new Vector();
        }
        //System.out.printf("will read vecto attrib: "+attrib+" on "+e.getName()+"\n");
        String[] sp = a.getValue().split(split_pattern);
        Vector rv = new Vector();
        for (int i=0; i<sp.length; i++)
        {
            if (sc==null)
            {
                rv.add(sp[i]);
                continue;
            }
            rv.add(sc.convert(sp[i]));
        }
        //System.out.printf("ready vecto attrib: "+attrib+" on "+e.getName()+"\n");
        return rv;
    }

    public static void writeAttrib(Element e , String name, String separator, Collection data)
    {
        writeAttrib(e,name,separator,data.iterator());
    }
    public static void writeAttrib(Element e , String name, String separator, Object[] data)
    {
        writeAttrib(e,name,separator,new CollectionIterator(data));
    }

    public static void writeAttrib(Element e , String name, String separator, int[] data)
    {
        writeAttrib(e,name,separator, CollectionIterator.getIterator(data));
    }

    public static void writeAttrib(Element e , String name, String separator, Iterator it)
    {
        StringBuffer sb = new StringBuffer();
        while (it.hasNext())
        {
            if (sb.length()>0)
            {
                sb.append(separator);
            }
            sb.append(""+it.next());
        }
        e.setAttribute(name, sb.toString());
    }
    public static double[] toDoubleArray(CDATA data)
    {
        String[] sp = data.getText().split(",");
        double[] rv = new double[sp.length];
        for (int i=0; i<sp.length; i++)
        {
            rv[i]=Double.parseDouble(sp[i]);
        }
        return rv;
    }
    public static CDATA writeDoubleArray(double[] array)
    {
        
        StringBuffer sb = new StringBuffer();
        for (int i=0; i<array.length; i++)
        {
            if (i>0)
            {
                sb.append(',');
            }
            sb.append(String.format(Locale.UK, "%2.8f", array[i]));
        }
        return new CDATA(sb.toString());
    }
    public static void writeColorAttrib(Element e, java.awt.Color c)
    {                                         
        e.setAttribute(COLOR_RGB_ATTRIB_NAME, ""+c.getRGB());
    }

    
    public static Element fromGZipped(byte[] data)
    {
        return fromGZipped(data, false);
    }

    public static Element fromGZipped(byte[] data, boolean null_if_error)
    {
        return fromGZipped(data, 0, data.length, null_if_error);
    }

    public static Element fromGZipped(byte[] data, int offset, int length, boolean null_if_error)
    {
        try
        {
            return makeElement(new java.util.zip.GZIPInputStream(new ByteArrayInputStream(data, offset, length)));
        }
        catch(IOException ie)
        {
            if (null_if_error)
                return null;
            
            
            throw new FRuntimeException("i/o error %s while unzipping data of length %d (offset: %d element length: %d)", ie, ie.getMessage(), data.length, offset, length);
        }
        
    }
    

    public static byte[] getGZipped(Element e)
    {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        java.util.zip.GZIPOutputStream go = null;
        try
        {
            go = new java.util.zip.GZIPOutputStream(baos);
            PrintWriter pw = new PrintWriter(go);
            pw.print(prettyXML(e));
            pw.close();
            baos.close();
        }
        catch(IOException ie)
        {
            throw new FRuntimeException("i/o error %s while zipping element: %s", ie, ie.getMessage(), e.getName());
        }
        
         
        
        return baos.toByteArray();
    }


    static public<A> A decodeObject(String text)
    {
        try
        {
            
            byte[] dec = java.util.Base64.getDecoder().decode(text);
            InputStream is = new java.util.zip.GZIPInputStream(new ByteArrayInputStream(dec));
            ObjectInputStream ois = new ObjectInputStream(is);
            return (A) ois.readObject();
        }
        catch(IOException ie)
        {
            throw new FRuntimeException("i/o error: %s while decoding bitset", ie, ie.getMessage());
        }
        catch(ClassNotFoundException cnfe)
        {
            throw new FRuntimeException("internal error (decoding class not found): %s while decoding bitset", cnfe, cnfe.getMessage());
        }
    }

    public static BitSet decodeBitSet(String text)
    {
        return (BitSet)decodeObject(text);
        /*
        try
        {
            
            byte[] dec = java.util.Base64.getDecoder().decode(text);
            InputStream is = new java.util.zip.GZIPInputStream(new ByteArrayInputStream(dec));
            ObjectInputStream ois = new ObjectInputStream(is);
            return (BitSet) ois.readObject();
        }
        catch(IOException ie)
        {
            throw new FRuntimeException("i/o error: %s while decoding bitset", ie, ie.getMessage());
        }
        catch(ClassNotFoundException cnfe)
        {
            throw new FRuntimeException("internal error (bitset class not found): %s while decoding bitset", cnfe, cnfe.getMessage());
        }
        */
    }
    
    static public<A> String encodeObject(A a)
    {
        try
        {
            
            ByteArrayOutputStream baos2 = new ByteArrayOutputStream();
            java.util.zip.GZIPOutputStream go = new java.util.zip.GZIPOutputStream(baos2);
            ObjectOutputStream oos2 = new ObjectOutputStream(go);
            oos2.writeObject(a);
            oos2.close();
    
            return java.util.Base64.getEncoder().encodeToString(baos2.toByteArray());
       }
        catch(IOException ie)
        {
            throw new FRuntimeException("i/o error: %s while encoding object", ie, ie.getMessage());
        }
        
    }

    static public<A> void writeEncodedCData(XMLStreamWriter xsw, A a)
    {
        try
        {
            xsw.writeCData(encodeObject(a));
        }
        catch(XMLStreamException xse)
        {
            throw new FRuntimeException("xml error: %s while writing encoded cdata", xse, xse.getMessage());
        }
    }

    static public<A> A readEncodedCData(XMLStreamReader xsr)
    {
        try
        {
            while (xsr.hasNext())
            {
                int event = xsr.next();
                if (event == XMLStreamConstants.END_ELEMENT)
                    return null; //end of element, no encoded cdata

                if (event != XMLStreamConstants.CHARACTERS)
                    continue;
                
                return decodeObject(xsr.getText());
            }
            return null;
        }
        catch(XMLStreamException xse)
        {
            throw new FRuntimeException("xml error: %s while reading encoded cdata", xse, xse.getMessage());
        }
    }

    public static String encodeBitSet(BitSet bs)
    {
        return encodeObject(bs);
        /*
        try
        {
            
            ByteArrayOutputStream baos2 = new ByteArrayOutputStream();
            java.util.zip.GZIPOutputStream go = new java.util.zip.GZIPOutputStream(baos2);
            ObjectOutputStream oos2 = new ObjectOutputStream(go);
            oos2.writeObject(bs);
            oos2.close();
    
            return java.util.Base64.getEncoder().encodeToString(baos2.toByteArray());
       }
        catch(IOException ie)
        {
            throw new FRuntimeException("i/o error: %s while encoding bitset", ie, ie.getMessage());
        }
        */
    }

    public static String getParam(File f, String  param, String path_str)
    {
        return getParam(getResolvedRootElement(f), param, path_str);
    }

    public static String getParam(Element rootElem, String  param, String path_str)
    {
        
        if (path_str != null)
        {
            String[] path = path_str.split("\\.");

            //look for the n-1 level
            for (int i=0; i<path.length-1; i++)
            {
                Element nroot = rootElem.getChild(path[i]);

                if (nroot==null)
                {
                    System.err.println("cannot find config for "+path[i]+" within path: "+ path_str);
                    return null;
                }
                rootElem=nroot;

            }
            param = path[path.length-1];
        }
        if (param == null)
            return null;
        
        
        Iterator it = rootElem.getDescendants();
         
        
        String keynames[]={"key", "param", "name"};
        String values[]={"value", "path"};
        
        while (it.hasNext())
        {
           Object o = it.next();
           if (! (o instanceof Element) )
               continue;
    
           Element ch = (Element)o;
           boolean gotit=false;
           if (ch.getName().equals(param))
           {
               gotit=true;
           }
           for (int i=0; !gotit && i<keynames.length; i++)
           {
               String totest = getAttrib(ch, keynames[i], null, false);
               if (totest==null)
                   continue;
               
               gotit=totest.equals(param);
               
               
           }
           if (!gotit)
               continue;
           
           
           for (String v: values)
           {
               String totest = getAttrib(ch, v, null, false);
               if (totest==null)
                   continue;
               
               return totest;
               
           }
           return ch.getText(); //multiple output??
           //System.out.println(ObjectGetter.map(ch.getText().trim().split("\n"), (String n) -> n.trim()));
           
        }
        return null;
    }

    public static void main (String args[]) 
    {
        Locale.setDefault(Locale.UK);
        SimpleOptionParser cmd = new SimpleOptionParser("i", "param", "path", "topretty");

        cmd.setOptional("path", "param");
        cmd.setSwitches("topretty");
        cmd.setFile("i");
            
        if(!OptionParser.parseParams(args,true, false, true, cmd))
        {
            return;
        }

        if (cmd.isSet("topretty"))
        {
            Element e = getRootElement(cmd.getFile("i"));
            System.out.println(prettyXML(e));
            return;
        }
        


        String rv = getParam(cmd.getFile("i"), cmd.getOptionalValue("param", null), cmd.getOptionalValue("path", null));
        if (rv == null)
            return;
        
        System.out.println(rv);
    }
}



