package lmu.utils.dag;

import java.util.*;

import lmu.utils.*;

import org.jdom.Element;

public class DAGEntry implements java.io.Serializable 
{
    Logger log = LogConfig.getLogger();
    public final static String ELEM_NAME = "de";
    public int id = -1;

    boolean is_root = false;
    private String name = null;
    private String term = null;

    public DAG dag;

    public Object userObject = null;

    HashSet<Object> labels = new HashSet<Object>();
    public String description = null;
    public Vector<String> alt_names = new Vector<String>();
    public Vector<String> xrefs = new Vector<String>();

    public HashMap<String, Object> attribs = new HashMap<>();

    protected HashSet<DAGEntry> parents = new HashSet<>();

    protected HashSet<DAGEntry> children = new HashSet<>();
    

    public void cloneInfos(DAGEntry to)
    {
        
        to.setObject(getObject());
        ObjectGetter.apply(labels, (Object o) -> to.addLabel(o));
        to.description = description;
        to.alt_names.addAll(alt_names);
        to.xrefs.addAll(xrefs);
        to.attribs.putAll(attribs);

    }
    public void makeRoot()
    {
        //remove all parents
        if (!dag.lookup_in_db)
        {
            for (DAGEntry de : parents)
            {
                de.removeChild(this);
            }
            return;
        }
            
        throw new RuntimeException("not yet implemented!");
    }
    
    public HashSet<DAGEntry> getParents()
    {
        if (!dag.lookup_in_db)
            return parents;

        return ObjectGetter.toSet(ObjectGetter.map(getParentIndeces(), (Integer idx) -> dag.getByIdx(idx)));
    }

    public Collection<Integer> getChildrenIndeces()
    {
        if (!dag.lookup_in_db) 
            return ObjectGetter.map(children, (DAGEntry de) -> de.id);

        return dag.getLinks(id, false);
        
    }


    
    public Collection<Integer> getParentIndeces()
    {
        if (!dag.lookup_in_db)
            return ObjectGetter.filter(ObjectGetter.map(parents, (DAGEntry de) -> de==null ? null : de.id),(Integer i)->i!=null);

        return dag.getLinks(id, true);
        
    }


    public HashSet<DAGEntry> getChildren()
    {
        if (!dag.lookup_in_db)
            return children;
        return ObjectGetter.toSet(ObjectGetter.map(getChildrenIndeces(), (Integer idx) -> dag.getByIdx(idx)));
        
    }
    public String toString()
    {
        return (this == dag.getRoot()) ? "ROOT" :  (name == null) ? term : name;
    }

    public void addLabel(Object l)
    {
        dag.addLabels(this, l);
    }

    public void setObject(Object o)
    {
        dag.object2entry.put(o, this);
        userObject = o;
    }

    public Object getObject()
    {
        return userObject;
    }

    
    protected DAGEntry(Element e, Vector<Integer> parents, Vector<Integer> children) 
    {
        id = XMLUtils.getIntAttrib(e, "id", null);
        is_root = XMLUtils.getBooleanAttrib(e, "isroot", false);
        name = XMLUtils.getAttrib(e, "name", null, false);
        term = XMLUtils.getAttrib(e, "term", null, false);
        description = XMLUtils.getAttrib(e, "description", null, false);
        alt_names = XMLUtils.getVectorAttrib(e, "alt_names", new Vector<>(), true, "|");
        xrefs = XMLUtils.getVectorAttrib(e, "xrefs", new Vector<>(), true, "|");

        if (parents != null)
        {
            parents.addAll(XMLUtils.getIntAttribVector(e, "parents", ","));
        }
        if (children != null)
        {
            children.addAll(XMLUtils.getIntAttribVector(e, "children", ","));
        }
        
        

        for (Element ate : XMLUtils.getChildren(e, "attrib"))
        {
            String t = XMLUtils.getAttrib(ate, "type", null, true);
            String name = XMLUtils.getAttrib(ate, "name", null, true);

            
            if (SimpleTypeConverter.XML_TYPE.equals(t))
            {
                
                LinkedList<Element> ve = XMLUtils.getChildren(ate);
                if (ve.size() == 0)
                {
                    throw new FRuntimeException("no xml child given for xml based attribute: %s", name);
                }
                
                attribs.put(name, ve.get(0));
                continue;
            }

            Vector v = SimpleTypeConverter.parseVector(ate);
            if (v != null)
            {
                attribs.put(name, v);
                continue;
            }

            StringConverter sc = SimpleTypeConverter.getConverter(t, true);
            attribs.put(name,sc.convert(XMLUtils.getAttrib(ate, "value", null, true)));

        }

    }

    
    public Element toXML()
    {
        return toXML(getParentIndeces(), getChildrenIndeces());
    }
    
    protected Element toXML(Collection<Integer> parentids, Collection<Integer> childrenids)
    {
        Element e = new Element(ELEM_NAME);
        e.setAttribute("id", "" + id);
        e.setAttribute("isroot", "" + is_root);
        if (name != null)
        {
            e.setAttribute("name", name);
        }
        if (term != null)
        {
            e.setAttribute("term", term);
        }
        if (description != null)
        {
            e.setAttribute("dec", description);         
        }
        e.setAttribute("parents",StringUtils.joinObjects(",", parentids));
        e.setAttribute("children", StringUtils.joinObjects(",",childrenids));
        e.setAttribute("alt_names", StringUtils.join("|", alt_names));
        e.setAttribute("xrefs", StringUtils.join("|", xrefs));

        for (Map.Entry<String, Object> a : attribs.entrySet())
        {
            if (a.getValue() instanceof Element)
            {
                Element attrib = new Element("attrib");
                attrib.setAttribute("name", a.getKey());
                attrib.setAttribute("type", SimpleTypeConverter.XML_TYPE);
                attrib.addContent((Element)(((Element)a.getValue()).clone()));
                e.addContent(attrib);
                continue;
            }

            Element vattrib = SimpleTypeConverter.createVectorElement("attrib", a.getValue());
            if (vattrib != null)
            {
                vattrib.setAttribute("name", a.getKey());
                e.addContent(vattrib);
                continue;
            }

            Element attrib = new Element("attrib");
            Object o = a.getValue();

            SimpleTypeConverter.SIMPLETYPE st = SimpleTypeConverter.get(o,
                                                                        false);
            if (st == null)
            {
                throw new RuntimeException("no converter found to "
                                           + o.getClass().getCanonicalName() + " for key:" + a
                                           + " value:" + o
                                           + " while saving attributes to DAG entry: " + getName());

            }

            attrib.setAttribute("name", a.getKey());
            attrib.setAttribute("type", st.getName());
            attrib.setAttribute("value", "" + a.getValue());
            e.addContent(attrib);
        }
        return e;
    }

    public boolean isRoot()
    {
        return is_root;
    }

    
    public void setRoot(boolean rootValue)
    {
        this.is_root = rootValue;
    }

    protected DAGEntry()
    {

    }

    public String getName()
    {
        return name;
    }

    public String getTerm()
    {
        return term;
    }

    public void saveChanges()
    {
        if (!dag.lookup_in_db)
            return;
        
        dag.updateDAGEntryInDB(this, null, null, false);
    }

    public void setName(String name)
    {
        dag.addNameReference(this.name, name, this);
        this.name = name;
        if (!dag.lookup_in_db)
            return;
                        
        dag.updateDAGEntryInDB(this, null, null, false);
    }

    public void setTerm(String term)
    {
        dag.addTermReference(this.term, term, this);
        this.term = term;
        if (!dag.lookup_in_db)
            return;
                        
        dag.updateDAGEntryInDB(this, null, null, false);
    }

    public DAGEntry addChild(DAGEntry entry)
    {
        if (id < 0) //the root
        {
            entry.is_root = true;
        }
        children.add(entry);
        entry.parents.add(this);
        
        if (!dag.lookup_in_db)
            return entry;
                        
        dag.updateDAGEntryInDB(this, null, entry.id, true);
        dag.updateDAGEntryInDB(entry, this.id, null, true);
        
        return entry;
        
    }

    public void removeChild(DAGEntry entry)
    {
        children.remove(entry);
        entry.parents.remove(this);
        if (dag.lookup_in_db)
        {
            dag.updateDAGEntryInDB(this, null, entry.id, false);
            dag.updateDAGEntryInDB(entry, this.id, null, false);
        }
    }

    public static boolean areLinearRelated(DAGEntry de1, DAGEntry de2)
    {
        if (de1 == de2) return true;

        if (de1.isAscendentOf(de2)) return true;

        if (de2.isAscendentOf(de1)) return true;

        return false;
    }

    boolean isRelated(int tocheck, boolean up)
    {
        HashSet<Integer> visited = new HashSet<>();
        LinkedList<Integer> q = new LinkedList<>();
        q.add(id);
        

        while (!q.isEmpty())
        {
            int n = q.poll();
            if (n == tocheck)
                return true;
            
            visited.add(n);

            DAGEntry de = (n == id) ? this : dag.getByIdx(n);

            q.addAll((up) ? de.getParentIndeces() : de.getChildrenIndeces());
        }

        return false;
    }
        
    public boolean isDescendantOf(DAGEntry de)
    {
        if (dag.isInDbLookupMode())
        {
            //iterate over the database entries
            return isRelated(de.id, true);
        }

        if (parents.contains(de))
        {
            return true;
        }

        for (DAGEntry p : parents)
        {
            if (p.isDescendantOf(de)) return true;
        }
        return false;
    }

    public boolean isAscendentOf(DAGEntry de)
    {
        if (dag.isInDbLookupMode())
        {
            return isRelated(de.id, false);
        }
        if (children.contains(de))
        {
            return true;
        }

        for (DAGEntry ch : children)
        {
            if (ch.isAscendentOf(de)) return true;
        }
        return false;
    }

    public void addLabels(Iterator it)
    {
        Vector<Object> nl = new Vector<Object>();

        while (it.hasNext())
        {
            Object l = it.next();
            if (labels.contains(l)) continue;

            nl.add(l);
            labels.add(l);

        }
        dag.addLabels(this, nl);
    }

    public void addLabels(Collection labels)
    {
        addLabels(labels.iterator());
    }

    public void addLabels(Object... labels)
    {
        addLabels(new CollectionIterator(labels));

    }

    public class DAGEntryIndexIterator implements Iterator<Integer>
    {
        boolean child_iterator = true;
        LinkedList<Integer> q = new LinkedList<>();
        HashMap<Integer, Integer> visited = new HashMap<>();
        int maxlevel = -1;

        void addAll(Collection<Integer> d, int level)
        {
            if (maxlevel >= 0 && level > maxlevel)
            {
                return;
            }
            for (Integer e : d)
            {
                if(e < 0 || null != visited.get(e))
                    continue;

                visited.put(e, level);
                q.add(e);
            }
        }

        DAGEntryIndexIterator(boolean b, int maxlevel)
        {
            this.maxlevel = maxlevel;
            if (child_iterator = b)
            {
                addAll(getChildrenIndeces(), 1);
            }
            else
            {
                addAll(getParentIndeces(), 1);
            }

        }

        public Integer next()
        {
            if (!hasNext()) return null;

            Integer next = q.poll();

            
            int level = visited.get(next);
            DAGEntry de = dag.getByIdx(next);
            addAll((child_iterator) ? de.getChildrenIndeces() : de.getParentIndeces(), level + 1);
            return next;
        }

        public boolean hasNext()
        {
            return q.size() > 0;
        }
    }

    public class DAGEntryIterator implements Iterator<DAGEntry>
    {
        DAGEntryIndexIterator it;

        DAGEntryIterator(boolean b, int maxlevel)
        {
            it = new DAGEntryIndexIterator(b, maxlevel);
        }

        public boolean hasNext()
        {
            return it.hasNext();
        }
        public DAGEntry next()
        {
            if (!it.hasNext()) return null;

            return dag.getByIdx(it.next());
        }

    }

    /** set maxlevel to <0 if all descendents are wanted */
    public Iterator<DAGEntry> getDescendatsIterator(int maxlevel)
    {
        return new DAGEntryIterator(true, maxlevel);
    }

    /** set maxlevel to <0 if all parents are wanted */
    public Iterator<DAGEntry> getAscendetsIterator(int maxlevel)
    {
        return new DAGEntryIterator(false, maxlevel);
    }


    /** set maxlevel to <0 if all descendents are wanted */
    public Iterator<Integer> getDescendatsIndexIterator(int maxlevel)
    {
        return new DAGEntryIndexIterator(true, maxlevel);
    }

    /** set maxlevel to <0 if all parents are wanted */
    public Iterator<Integer> getAscendetsIndexIterator(int maxlevel)
    {
        return new DAGEntryIndexIterator(false, maxlevel);
    }


    public HashSet<DAGEntry> findByAttrib(String attrib, Object value, int maxlevel,boolean up)
    {
        if (attrib == null)
            return ObjectGetter.toSet(this);
        
        
        HashSet<DAGEntry> rv = new HashSet<>();
        Iterator<DAGEntry> it =  new DAGEntryIterator(!up, maxlevel);

        
        while (it.hasNext())
        {
            DAGEntry ch = it.next();
            
            Object av = ch.attribs.get(attrib);
            if (av == null)
                continue;
            
            if (value == null)
            {
                rv.add(ch);
                continue;
            }
                
            
            if (!value.equals(av))
                continue;
                
            
            rv.add(ch);
            
        }
        return rv;
    }

    public HashSet<DAGEntry> findChildByAttrib(String attrib, Object value)
    {
        return findChildByAttrib(attrib,value,-1);
    }

    public HashSet<DAGEntry> findChildByAttrib(String attrib, Object value, int maxlevel)
    {
        return findByAttrib(attrib,value,maxlevel, false);
    }


    public HashSet<DAGEntry> findParentByAttrib(String attrib, Object value, int maxlevel)
    {
        return findByAttrib(attrib,value,maxlevel, true);
    }

    public HashSet<DAGEntry> findParentByAttrib(String attrib, Object value)
    {
        return findByAttrib(attrib, value, -1, true);
    }

}
