package lmu.utils.index;

import java.io.*;
import java.util.*;
import lmu.utils.*;

public class DoubleRAFIndex<A> implements Indexer<A>
{

    public static final String FPLOOKUP_SUFFIX = ".fplookup";

    public interface RecordReader<A>
    {
        public A read(int idx, RandomAccessFile raf) throws IOException;
    }

    public interface RecordWriter<A>
    {
        public void write(int idx, RandomAccessFile raf, A data) throws IOException;
    }

    File outfile = null;
    RandomAccessFile raf = null;
    RandomAccessFile index = null;
    int LONGSIZE = 8;
    RecordReader<A> rr;
    RecordWriter<A> rw;
    int size = 0;

    public DoubleRAFIndex(File outfile, RecordReader<A> rr, RecordWriter<A> rw)
    {
        try
        {
            this.outfile = outfile;
            
            this.rr = rr;
            this.rw = rw;
            raf = new RandomAccessFile(outfile, "rw");
            size = (int)(raf.length() / (double)LONGSIZE);
            index = new RandomAccessFile(outfile.getAbsolutePath() + FPLOOKUP_SUFFIX, "rw");
            
        }
        catch(IOException ie)
        {
            throw new FRuntimeException("i/o error: %s while initing double raf from/to file: %s", ie, ie.getMessage(), outfile.getAbsolutePath());
        }
        
    }

    public synchronized void write(A data, int idx)
    {
        try
        {
            
            long pos = LONGSIZE * idx;
            long sizediff = pos - index.length();
            if (sizediff > 0)
            {
                index.write(new byte[(int)sizediff]);
            }
            else
            {
                index.seek(pos);
            }
            long info = raf.getFilePointer();
            index.writeLong(info);
            rw.write(idx, raf,data);

            size = Math.max(size, idx);
        }
        catch(IOException ie)
        {
            throw new FRuntimeException("i/o error: %s while writing data %d to double raf to file: %s", ie, ie.getMessage(), idx, outfile.getAbsolutePath());
        }
    }


    public int size()
    {
        return size;
    }

    Logger log = LogConfig.getLogger();

    public synchronized A read(int idx)
    {
        long pos = LONGSIZE * idx;
        long fp2 = -1;
        try
        {
            if (index.length() < pos + LONGSIZE)
                return null;

            index.seek(pos);
            fp2 = index.readLong();
            if (fp2 < 0) //not contained
            {
                return null;
            }
            raf.seek(fp2);
            A rv = rr.read(idx, raf);

            return rv;
        }
        catch(IOException ie)
        {
            throw new FRuntimeException("i/o error: %s while reading data %d from raf to file: %s (fp2: %d)", ie, ie.getMessage(), idx, outfile.getAbsolutePath(), fp2);
        }
        
        
        
    }
    
}
