package lmu.utils.obo;

import java.io.*;
import java.util.*;
import java.util.regex.*;

public class OboElement {
	public Vector<OboElement> parents = null;
	public Vector<OboElement> children = null;

	HashMap<OboField, Vector<String>> entries = new HashMap<OboField, Vector<String>>();
	// tags that are not specified in the original obo format
	HashMap<String, Vector<String>> additionalTags = new HashMap<String, Vector<String>>();
	String name;
	String id;
	OboElementType type;

	private Pattern idvalue = Pattern.compile("(.+?): (.+?)($|( \\!))");

	public String toString() {
		return name + ":" + id + "::" + type;
	}

	public boolean isAncestorOf(OboElement e) {
		if (e.parents == null)
			return false;

		LinkedList<OboElement> q = new LinkedList<OboElement>();
		q.add(e);
		while (!q.isEmpty()) {
			OboElement test = q.poll();
			if (test == this) {
				return true;
			}
			if (test.parents == null || test.parents.size() == 0) {
				continue;
			}
			q.addAll(test.parents);
		}
		return false;
	}

	public OboElement(String id, String name, OboElementType type) {
		this.id = id;
		this.name = name;
		this.type = type;
	}

	public void addEntry(OboField field, String entry) {
		Vector<String> v = entries.get(field);
		if (v == null) {
			entries.put(field, v = new Vector<String>());
		}
		v.add(entry);
	}

	public OboElement(Vector<String> stanza_lines, String type) {
		this.type = OboElementType.get(type);
		for (String line : stanza_lines) {
			Matcher m = idvalue.matcher(line);
			if (m.find()) {
				OboField field = OboField.get(m.group(1));
				// additional tag that is not part of obo format specification
				if (field == null) {
					String key = m.group(1);
					if (!additionalTags.containsKey(key)) {
						additionalTags.put(key, new Vector<String>());
					}
					additionalTags.get(key).add(m.group(2));
					continue;
				}
				String value = m.group(2);
				if (field == OboField.NAME) {
					this.name = value;
					continue;
				}
				if (field == OboField.ID) {
					this.id = value;
					continue;
				}
				if (entries.get(field) == null) {
					entries.put(field, new Vector<String>());
				}
				entries.get(field).add(value);
			}
		}
	}

	public void write(PrintWriter pw) {
		pw.printf("[%s]\nid: %s\nname: %s\n", type.name, id, name);
		for (Map.Entry<OboField, Vector<String>> o : entries.entrySet()) {
			final String on = o.getKey().name;
			for (String sub : o.getValue()) {
				pw.printf("%s: %s\n", on, sub);
			}
			// lmu.utils.StringUtils.join("|"

		}
		pw.println();

	}

	public void setName(String name) {
		this.name = name;
	}

	public String getName() {
		return name;
	}

	public String getTerm() {
		return id;
	}
	public Vector<OboElement> getParents()
	{
	    if ( parents == null)
	        parents = new Vector<OboElement>();
	    
	    return parents;
	}
	public Vector<OboElement> getChildren()
    {
	    if ( children == null)
	        children = new Vector<OboElement>();
        
        return children;
    }
    
	public Vector<String> getEntries(OboField f) {
		return entries.get(f);
	}

	public HashMap<OboField, Vector<String>> getEntries(){
		return entries;
	}

        public Set<String> getAdditonalTagTypes() {
            return additionalTags.keySet();
        }
	public Vector<String> getAdditionalEntry(String key) {
		return additionalTags.get(key);
	}

	public OboElementType getType() {
		return type;
	}

	public String getId() {
		// TODO Auto-generated method stub
		return id;
	}
}
