package lmu.utils.obo;

import java.io.*;
import java.util.*;
import java.util.regex.*;

import lmu.utils.*;

public class OboReader implements Iterator<OboElement>
{
    Logger log = LogConfig.getLogger();
    BufferedReader bf;
    Vector<String> next;
    String nextTag;
    String lastTag;
    Pattern tag = Pattern.compile("\\[(.+?)\\]");

    public OboReader(File input, Reader r)
    {
        init(r, input);
    }

    public OboReader(File input)
    {
        init(null, input);

    }

    public OboReader(Reader input)
    {
        init(input, null);
    }

    public void init(Reader input, File f)
    {

        this.next = null;
        this.nextTag = null;
        this.lastTag = null;
        String line;
        try
        {
            if (input == null)
            {
                input = new FileReader(f);
            }
            if (f == null)
            {
                f = new File("unknown");
            }
            bf = new BufferedReader(input);
            boolean firstTagSeen = false;
            Vector<String> savedLines = new Vector<String>();
            while (next == null && (line = bf.readLine()) != null)
            {
                // if(line.startsWith("[") && line.endsWith("]"))
                // {
                Matcher m = tag.matcher(line);
                if (m.matches())
                {
                    if (!firstTagSeen)
                    {
                        firstTagSeen = true;
                        savedLines.add(line);
                        // lastTag=line.substring(1,line.length()-1);
                        lastTag = m.group(1);
                    } else
                    {
                        lastheader = line;
                        nextTag = m.group(1);
                        // nextTag=line.substring(1,line.length()-1);
                        next = savedLines;
                    }
                } else
                {
                    if (!firstTagSeen) continue;

                    savedLines.add(line);
                }
            }
        } catch (IOException e)
        {
            throw new RuntimeException("exception:" + e.getMessage()
                                       + " while reading from " + f.getAbsolutePath(), e);
        }
    }

    public boolean hasNext()
    {
        return next != null;
    }

    boolean lastElement = false;
    String lastheader = null;
    int nidx = 0;
    public OboElement next()
    {
        OboElement el = new OboElement(next, lastTag);
        
        boolean noTagYet = false;
        String line;
        try
        {
            Vector<String> savedLines = new Vector<String>();
            if (lastheader != null)
            {
                savedLines.add(lastheader);
            }
            while (!noTagYet && (line = bf.readLine()) != null)
            {
                // if(line.startsWith("[") && line.endsWith("]"))
                // {
                Matcher m = tag.matcher(line);
                if (m.matches())
                {
                    lastTag = nextTag;
                    noTagYet = true;
                    nextTag = m.group(1);
                    // nextTag=line.substring(1,line.length()-1);
                    next = savedLines;
                } else
                {
                    savedLines.add(line);
                }

            }
            // no more lines
            if (!noTagYet)
            {
                if (!lastElement)
                {
                    lastElement = true;
                    next = savedLines;
                    lastTag = nextTag;
                } else
                {
                    next = null;
                }
            }
        } catch (IOException e)
        {
            ExceptionManager.throwRuntimeException("error while reading obo file");
        }
        log.trace("read " + el.getTerm());
        return el;
    }

    public void remove()
    {
        throw new RuntimeException("not yet implemented");
    }

    static LinkedList<Pair<Integer, OboElement>> getRootPaths(OboElement e)
    {

        LinkedList<Pair<Integer, OboElement>> rv = new LinkedList<Pair<Integer, OboElement>>();
        LinkedList<Pair<Integer, OboElement>> q = new LinkedList<Pair<Integer, OboElement>>();

        q.add(new Pair<Integer, OboElement>(0, e));
        while (!q.isEmpty())
        {
            Pair<Integer, OboElement> p = q.poll();
            rv.add(p);
            if (p.getSecond().parents == null
                || p.getSecond().parents.size() == 0)
            {
                continue;
            }
            int l = p.getFirst() + 1;
            for (OboElement pe: p.getSecond().parents)
            {
                q.add(new Pair<Integer, OboElement>(l, pe));
            }
        }

        return rv;
    }

    static public OboElement lca(OboElement e1, OboElement e2)
    {
        LinkedList<Pair<Integer, OboElement>> p1 = getRootPaths(e1);
        LinkedList<Pair<Integer, OboElement>> p2 = getRootPaths(e2);
        int best = p1.size() + p2.size();
        OboElement rv = null;
        for (int i = 0; i < p1.size(); i++)
        {
            int l1 = p1.get(i).getFirst();
            OboElement o1 = p1.get(i).getSecond();

            for (int j = 0; j < p2.size(); j++)
            {
                int l2 = p2.get(j).getFirst();
                OboElement o2 = p2.get(j).getSecond();

                if (o1 != o2 || Math.max(l1, l2) >= best)
                {
                    continue;
                }
                best = Math.max(l1, l2);
                rv = o1;
            }
        }

        return rv;
    }

    public static Iterator<OboElement> getTreeIterator(OboElement root)
    {
        return IteratorUtils.getDAGIterator(root, (OboElement r) -> r.children);
    }

    public static Collection<OboElement> getTreeElems(OboElement root)
    {
        Vector<OboElement> rv = new Vector<>();
        ObjectGetter.apply(getTreeIterator(root), (OboElement oe) -> rv.add(oe));
        return rv;

    }

    /** returns all values for a given OboField for a collection of elements */
    public static HashSet<String> getTagValues(HashSet<OboElement> TreeElems,
                                               OboField type)
    {
        HashSet<String> values = new HashSet<String>();
        for (OboElement act_ele: TreeElems)
        {
            if (act_ele.entries.containsKey(type))
            {
                Vector<String> val = act_ele.entries.get(type);
                values.addAll(val);
            }

        }

        return values;
    }

    /** returns the OboElement matching id or name for the given name */

    public OboElement getElement(String name)
    {
        // @TODO check if obo can have mulitple elements with same name
        while (hasNext())
        {
            OboElement ele = next();
            if (ele.id.equalsIgnoreCase(name)
                || ele.name.equalsIgnoreCase(name))
            {
                return ele;
            }

        }
        return null;
    }

    public static void extractSubTree(File f, PrintWriter pw, OboElement ... nodes)
    {
        extractSubTree(f, pw, new CollectionIterator<OboElement>(nodes));
    }

    public static void extractSubTree(File f, PrintWriter pw, Collection<OboElement> nodes)
    {
        extractSubTree(f, pw, nodes.iterator());
    }

    public static void extractSubTree(File f, PrintWriter pw, Iterator<OboElement> nodes)
    {
        LinkedList<OboElement> q = new LinkedList<OboElement>();
        while (nodes.hasNext())
        {
            q.add(nodes.next());
        }

        HashSet<OboElement> toextract = new HashSet<OboElement>();
        while (!q.isEmpty())
        {
            OboElement oe = q.poll();
            toextract.add(oe);
            if (oe.parents == null) continue;

            for (OboElement p: oe.parents)
            {
                if (toextract.contains(p)) continue;

                q.add(p);
            }
        }

        HashSet<String> ids = new HashSet<String>();
        ids.add("root");
        for (OboElement oe: toextract)
        {
            ids.add(oe.getId());
        }


        extractIds(f, ids, pw);

    }
    public static void extractIds(File f, HashSet<String> ids, PrintWriter pw)
    {

        OboReader or = new OboReader(f);
        Vector<String> v = new Vector<String>();
        while (or.hasNext())
        {
            v.clear();
            v.addAll(or.next);

            OboElement ele = or.next();
            if (!ids.contains(ele.id))
            {
                continue;
            }
            for (String l: v)
            {
                pw.println(l);
            }
        }

    }


    /** returns all roots */
    public Vector<OboElement> readTree()
    {
        HashMap<String, OboElement> id2elem = new HashMap<String, OboElement>();
        ObjectGetter.apply(this, (OboElement ele) -> id2elem.put(ele.id, ele));

        int missing_refs = 0;
        for (OboElement ele: id2elem.values())
        {

            Vector<String> parents = ele.getEntries(OboField.IS_A);
            // for material stuff the relationship is set as parent
            Vector<String> relations = ele.getEntries(OboField.RELATIONSHIP);
            if (parents == null && relations == null)
                continue;
            
            if (parents == null && relations != null)
            {
                parents = relations;
            }

            if (parents != null && relations != null)
            {
                parents.addAll(relations);
            }

            ele.parents = new Vector<OboElement>();

            for (String p: parents)
            {
                // bei den relationships steht part_of noch davor
                if (p.contains(" "))
                {
                    p = p.split(" ")[1].trim();
                }

                OboElement oe = id2elem.get(p);
                if (oe == null)
                {
                    String test = unescape(p);
                    System.out.println("missing: " + p + " escaped: " + test);
                    if (null == (oe = id2elem.get(test)))
                    {
                        missing_refs++;
                        continue;
                    }

                    // throw new
                    // RuntimeException("invalid reference to parent "+
                    // p+" in elem "+ele.id+" test: "+test);
                    // continue;
                }
                if (oe.children == null)
                {
                    oe.children = new Vector<OboElement>();
                }
                oe.children.add(ele);
                ele.parents.add(oe);
            }
        }
        log.warn("got " + missing_refs + " missing references");
        Vector<OboElement> rv = new Vector<OboElement>();
        for (OboElement oe: id2elem.values())
        {
            if (oe.parents != null) continue;
            rv.add(oe);
        }
        return rv;
    }

    static String unescape(String s)
    {
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < s.length(); i++)
        {
            if (s.charAt(i) == '\\')
            {
                continue;
            }
            sb.append(s.charAt(i));
        }
        return sb.toString();
    }

    public static void main(String[] args)
    {

        SimpleOptionParser cmdline = new SimpleOptionParser("obo", "o");
        cmdline.setFile("obo");
        cmdline.setOptional("o");

        String p = "test \"wjat\"";
        String test = unescape(p);
        if (!OptionParser.parseParams(args, true, false, true, cmdline)) return;
        Logger log = LogConfig.getLogger();
        log.info("start reading obo");
        PrintWriter pw = cmdline.getWriter("o");
        OboReader read = new OboReader(new File(cmdline.getValue("obo")));
        Vector<OboElement> v = read.readTree();
        System.out.printf("got %d roots\n", v.size());
        for (OboElement oe: v)
        {
            System.out.printf("read from %s\n", oe.getName());
            read.getTreeElems(oe);
        }

        /*
         * while (read.hasNext()) { OboElement ele = read.next(); if (pw!=null)
         * { ele.write(pw); }
         * //System.out.printf("id: %s type: %s name: %s entries: %s\n" ,
         * ele.id, ele.getType(), ele.getName(), ""+ele.entries); } if
         * (pw==null) { pw.close(); }
         */
        log.info("finished reading obo");
    }
}
