package lmu.utils.partialorder;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import lmu.utils.BufferPrintWriter;
import lmu.utils.IteratorUtils;
import lmu.utils.ObjectGetter;
import lmu.utils.Pair;
import lmu.utils.graph.GraphUtils;

public class MergablePartialOrderGraph<A> extends PartialOrderGraph<HashSet<A>>
{
    HashMap<A, PartialNode<HashSet<A>>> lookup = new HashMap<>();

    public void setMRelation(A a1, A a2)
    {
        super.addRelation(lookup.get(a1).data, lookup.get(a2).data);
    }
    public void addMergable(A a)
    {
        PartialNode<HashSet<A>> n = super.add(ObjectGetter.toSet(a));  
        lookup.put(a, n);    
    }

    public void addMergable(A  ... data)
    {
        for (A a : data)
        {
            addMergable(a);
        }

    }

    public class CompositionElement<A>
    {
        public Vector<A> q1_elems = new Vector<>();
        public Vector<A> q2_elems = new Vector<>();

        public A a1;
        public A a2;

        CompositionElement()
        {
        }


        void add(A q1a, A q2a)
        {
            Vector<A> v = (q1a != null) ? q1_elems : q2_elems;
            v.add((q1a!=null) ? q1a : q2a);
        }

        CompositionElement(A a1, A a2)
        {
            this.a1 = a1; this.a2 = a2;
        }
        public boolean isMatched()
        {
            return a1 != null && a2 != null;
        }

        public int getQSize(boolean first)
        {
            return (first) ? q1_elems.size() : q2_elems.size();
        }
        public boolean isOneSide()
        {
            return getQSize(true) == 0 || getQSize(false) == 0;
        }

        public String toString()
        {
            if (isMatched())
            {
                return "matched: " + a1 + " VS " + a2;
            }
            return "unmatched: " + q1_elems + " VS "+ q2_elems;
        }
    }


    public UnorderedDecompositor getUnorderedDecompositor()
    {
        return new UnorderedDecompositor();
    }

    public class UnorderedDecompositor
    {

        Vector<HashSet<A>> onodes = new Vector<>();
        HashMap<A, Integer> a2pos = new HashMap<>();

        HashMap<HashSet<A>, HashSet<HashSet<A>>> after = new HashMap<>();


        UnorderedDecompositor()
        {
            Iterator<HashSet<A>> it = iterator();
            int pos = -1;

            HashMap<HashSet<A>, HashSet<HashSet<A>>> edges = new HashMap<>();


            while (it.hasNext())
            {
                pos++;
                final int idx = pos;
                HashSet<A> h = it.next();
                PartialNode<HashSet<A>> pn = nodes.get(h);
                edges.put(h, ObjectGetter.mapToSet(pn.post, (PartialNode<HashSet<A>> pn2)  -> pn2.data));

                onodes.add(h);
                ObjectGetter.apply(h, (A a) -> a2pos.put(a, idx));
            }

            BufferPrintWriter bpw = new BufferPrintWriter();

            if (log.isDebugEnabled())
            {
                bpw.println("edges");
                for (HashSet<A> h : edges.keySet())
                {
                    for (HashSet<A> h2 : edges.get(h))
                    {
                        bpw.printf("\t EDGE: %s -> %s\n", h, h2);
                    }
                }
            }
            
            after = GraphUtils.getTransitiveEdges(edges);

            if (log.isDebugEnabled())
            {
                bpw.println("trans edges");
                for (HashSet<A> h : edges.keySet())
                {
                    for (HashSet<A> h2 : edges.get(h))
                    {
                        bpw.printf("\t TRANS EDGE: %s -> %s\n", h, h2);
                    }
                }
                log.debug(bpw.getBuffer());
            }
            
        }

        boolean isAfter(CompositionElement<A> c, HashSet<A> after)
        {
            for (A a : c.q1_elems)
            {
                if(isAfter(a, after))
                    return true;
            }

            for (A a : c.q2_elems)
            {
                if(isAfter(a, after))
                    return true;
            }
            return isAfter(c.a1, after) || isAfter(c.a2, after);
        }

        boolean isAfter(A n, HashSet<A> after)
        {
            if (n == null)
                return false;
            
            return isAfter(lookup.get(n).data, after);

        }
        boolean isAfter(HashSet<A> n, HashSet<A> after)
        {
            if (n == null)
                return false;
            
            
            HashSet<HashSet<A>> h = this.after.get(n);
            return(h != null && h.contains(after));

        }

        int o(A a)
        {
            return a2pos.get(a);
        }

        public class PairCompositionGetter
        {
            HashMap<HashSet<A>, A> q1info;
            HashMap<HashSet<A>, A> q2info;
            Vector<A> q1;
            Vector<A> q2;
              
            PairCompositionGetter(Vector<A> q1, Vector<A> q2)
            {
                this.q1 = q1;
                this.q2 = q2;

                q1info = ObjectGetter.buildReverseMap(q1, (A a) -> lookup.get(a).data);
                q2info = ObjectGetter.buildReverseMap(q2, (A a) -> lookup.get(a).data);
            }
        
            public Vector<CompositionElement<A>> getOrderedDecomposition(int start, int end)
            {
                Vector<CompositionElement<A>> rv = new Vector<>();
    
                CompositionElement<A> lastE  = null;
                HashSet<A> last = null;
    
                BufferPrintWriter bpw = new BufferPrintWriter();

                bpw.setSkip(!log.isDebugEnabled());

                bpw.printf("ordered decomp: %d-%d\n", start, end);
                bpw.println("onodes " + onodes);

                for (int i=start; i<=end; i++)
                {
                    HashSet<A> n = onodes.get(i);
                    if (n.size() == 1) //unmatched node, skipping
                    {
                        continue;
                    }
                    A q1a = q1info.get(n);
                    A q2a = q2info.get(n);
                    
                    bpw.printf("next[%d] %s q1a: %s q2a: %s\n", i, n, q1a, q2a);
                    if (q1a == null && q2a ==null)
                        continue;
    
                    if (q1a != null && q2a!=null)
                    {
                        bpw.printf("add match\n");
                        rv.add(new CompositionElement<A>(q1a, q2a));
                        last = null;
                        lastE = null;
    
                        continue;
                    }
                    if (lastE != null)
                    {
                        bpw.printf("check n: %s after %s: %s\n", n, last, isAfter(lastE, n));
                    }
                    boolean is_after = (lastE == null ||  isAfter(lastE, n));

                    if (lastE !=null && is_after && q1a != null && lastE.q2_elems.size() ==0)
                    {
                        is_after = false;
                    }

                    if (lastE !=null && is_after && q2a != null && lastE.q1_elems.size() ==0)
                    {
                        is_after = false;
                    }
                    if (is_after)
                    {
                        //check if we only extend one dimension
                        bpw.printf("create new elem\n");
                        rv.add(lastE = new CompositionElement<A>());
                    }
    
                    bpw.printf("update unmatched\n");
                    lastE.add(q1a, q2a);
                    last = n;
                }
                log.debug(bpw.getBuffer());
    
                return rv; //ObjectGetter.filter(rv, (CompositionElement<A> c) -> c.isMatched() || c.q1_elems.size() > 0 && c.q2_elems.size() > 0);
            }
            
            public Vector<CompositionElement<A>> getOrderedDecomposition(A startE, A endE)
            {                                                  
                return getOrderedDecomposition((startE == null) ? 0 : o(startE), (endE == null) ? onodes.size()-1: o(endE));
            }

            public Vector<CompositionElement<A>> getOrderedDecomposition()
            {
                int start = Math.min(o(q1.firstElement()), o(q2.firstElement()));
                int end = Math.max(o(q1.lastElement()), o(q2.lastElement()));
    
                return getOrderedDecomposition(start, end);
            }
        }

        public PairCompositionGetter decompose(Vector<A> q1, Vector<A> q2)
        {
            return new PairCompositionGetter(q1, q2);
        }
        
    }


    public MergablePartialOrderGraph()
    {
        super();
    }

    class MergeInfos
    {
        PartialNode<HashSet<A>> n1;
        PartialNode<HashSet<A>> n2;

        PartialNode<HashSet<A>> prefix_stop; 
        PartialNode<HashSet<A>> suffix_stop; 

        HashSet<PartialNode<HashSet<A>>> b_prefixes;
        HashSet<PartialNode<HashSet<A>>> b_suffixes;

        HashSet<PartialNode<HashSet<A>>> a_prefixes;
        HashSet<PartialNode<HashSet<A>>> a_suffixes;

        MergeInfos(A a, A b, PartialNode<HashSet<A>> prefix_stop, PartialNode<HashSet<A>> suffix_stop)
        {
            n1 = lookup.get(a);
            n2 = lookup.get(b);
            this.prefix_stop = prefix_stop;
            this.suffix_stop = suffix_stop;
        }

        public boolean isContained()
        {
            return n1 == n2;
        }

        public boolean checkValid()
        {
            
            IteratorUtils.Filter<PartialNode<HashSet<A>>> f = (PartialNode<HashSet<A>> pn) -> pn == n1;
            IteratorUtils.Filter<PartialNode<HashSet<A>>> f2 = (PartialNode<HashSet<A>> pn) -> pn == n2;
            /*
            Vector<PartialNode<HashSet<A>>> vv = ObjectGetter.toVector(getTransitivePrefixIterator(n2, true, prefix_stop));

            
            if (vv.contains(n1))
            {
                System.out.printf("prefixes of %s (check for %s): %s\n",
                              n2, n1, vv);


            }


            vv = ObjectGetter.toVector(getTransitiveSuffixIterator(n2, true, suffix_stop));

            if (vv.contains(n1))
            {
                System.out.printf("suffixs of %s (check for %s): %s\n",
                              n2, n1, vv);

                System.out.printf("prefixes of %s (check for %s): %s\n",
                                              n2, n1, vv);

              
            }
            */

                
            return( !IteratorUtils.containsFiltered(getTransitivePrefixIterator(n2, true, prefix_stop), f)
                    &&
                    !IteratorUtils.containsFiltered(getTransitiveSuffixIterator(n2, true, suffix_stop), f)
                    );
            /*
                    &&
                    !IteratorUtils.containsFiltered(getTransitivePrefixIterator(n1, false), f2)
                    &&
                    !IteratorUtils.containsFiltered(getTransitiveSuffixIterator(n1, false), f2)
                  );
              */

        }

        public boolean calcPrefixes()
        {
            b_prefixes = getTransitivePrefixes(n2, false, prefix_stop);
            b_suffixes = getTransitiveSuffixes(n2, false, suffix_stop);

            return(!b_prefixes.contains(n1) && !b_suffixes.contains(n2));
        }


        public void calcSuffixes()
        {
            a_prefixes  = getTransitivePrefixes(n1, false, prefix_stop);
            a_suffixes = getTransitiveSuffixes(n1, false, suffix_stop);
        }

    }


    public boolean contains(A a)
    {
        return lookup.get(a) != null;
    }
    public ACCEPT_STATE checkMerge(A a, A b)
    {
        return checkMerge(a,b, null, null);
    }

    public ACCEPT_STATE checkMerge(A a, A b, PartialNode<HashSet<A>> prefix_stop, PartialNode<HashSet<A>> suffix_stop)
    {
        return merge(a,b, prefix_stop, suffix_stop, true);
    }

    public ACCEPT_STATE merge(A a, A b)
    {
        return merge(a, b, null, null);
    }

    public ACCEPT_STATE merge(A a, A b, PartialNode<HashSet<A>> prefix_stop, PartialNode<HashSet<A>> suffix_stop)
    {
        return merge(a,b, prefix_stop, suffix_stop, false);
    }

    

    ACCEPT_STATE merge(A a, A b, PartialNode<HashSet<A>> prefix_stop, PartialNode<HashSet<A>> suffix_stop, boolean only_check)
    {
        MergeInfos mi = new MergeInfos(a, b, prefix_stop, suffix_stop);
        if (mi.isContained())
            return ACCEPT_STATE.CONTAINED;

        if(!mi.checkValid())
        {
            //log.info("invlaid here");
            return ACCEPT_STATE.INVALID;
        }
            

        if (only_check)
            return ACCEPT_STATE.ACCEPT;
        

        
        if (!mi.calcPrefixes())
            return ACCEPT_STATE.INVALID;


        /*log.debug("merged nodeds:\n1:  %s\n\tpre: %s\n\tpost: %s\n2:  %s\n\tpre: %s\n\tpost: %s\n",  
                  mi.n1 , mi.n1.pre, mi.n1.post,
                  mi.n2 , mi.n2.pre, mi.n2.post);
        

        log.debug("valid: " + mi.checkValid());
        */
        mi.calcSuffixes();
        //crossing relations: pre1 -> post2 and pre2 -> post2

        HashSet<Pair<PartialNode<HashSet<A>>, PartialNode<HashSet<A>>>> crossing_relations = new HashSet<>();
        searchCrossingRelations(crossing_relations, mi.b_prefixes, mi.a_suffixes);
        searchCrossingRelations(crossing_relations, mi.a_prefixes, mi.b_suffixes);
        fixCrossingRelations(crossing_relations, mi.n1, mi.n1); //all crossing are fixed by putting n1 in the middle

        //now transfer all prefixes/suffixes and data of b to a
        mi.n1.pre.addAll(mi.n2.pre);
        mi.n1.post.addAll(mi.n2.post);


        // remove links pointing to n2

        //log.debug("will remove references for  " + mi.n2.data +" pre :" + mi.n2.pre +   " post: " + mi.n2.post);
        ObjectGetter.apply(mi.n2.pre, (PartialNode<HashSet<A>> pn) -> pn.post.remove(mi.n2));
        ObjectGetter.apply(mi.n2.post, (PartialNode<HashSet<A>> pn) -> pn.pre.remove(mi.n2));

        //add references to the 
        ObjectGetter.apply(mi.n2.pre, (PartialNode<HashSet<A>> pn) -> pn.post.add(mi.n1));
        ObjectGetter.apply(mi.n2.post, (PartialNode<HashSet<A>> pn) -> pn.pre.add(mi.n1));



        //update the super's lookup
        super.nodes.remove(mi.n2.data);
        super.nodes.remove(mi.n1.data);


        //check if there are any other nodes referencing n1 or n2 data
        for (Map.Entry<HashSet<A>, PartialNode<HashSet<A>>> e : super.nodes.entrySet())
        {
            PartialNode<HashSet<A>> pn = e.getValue();
            HashSet<A> hk = e.getKey();
            if (pn.data != mi.n1.data && pn.data != mi.n2.data)
                continue;

            log.warn("invlaid reference : " +pn+ " data: " +pn.data+ " key: " +hk);
            log.warn("check: 1: " + (hk.equals(mi.n1.data))+" 2: " + (hk == mi.n2.data)+" nodes test 1 : " + nodes.get(mi.n1.data)+" nodes test 2 : " +nodes.get(mi.n2.data));
        }

        mi.n1.data.addAll(mi.n2.data);
        super.nodes.put(mi.n1.data, mi.n1);

        //log.debug("merged node:  %s\n\tpre: %s\n\tpost: %s\n",  mi.n1 , mi.n1.pre, mi.n1.post);
        //update n2 lookup
        ObjectGetter.apply(mi.n2.data, (A ob) -> lookup.put(ob, mi.n1));
        return ACCEPT_STATE.ACCEPT;

    }

    static Vector<Pair<String,Integer>> generate(String label, int length)
    {
        return ObjectGetter.map(IteratorUtils.getIntIterator(0,length), (Integer idx) -> Pair.create(label, idx));
    }

    public static void main (String args[]) 
    {

        final int L = 10;
        Vector<Pair<String, Integer>> n1 = generate("a", L);
        Vector<Pair<String, Integer>> n2 = generate("b", L);

        MergablePartialOrderGraph<Pair<String,Integer>> mpog = new MergablePartialOrderGraph<>();
        ObjectGetter.apply(n1, (Pair<String,Integer> p) -> mpog.addMergable(p));
        ObjectGetter.apply(n2, (Pair<String,Integer> p) -> mpog.addMergable(p));

        for (int i = 1; i < n1.size(); i++)
        {
            mpog.setMRelation(n1.get(i-1), n1.get(i));
        }
        for (int i = 1; i < n2.size(); i++)
        {
            mpog.setMRelation(n2.get(i-1), n2.get(i));
        }

        System.out.println("V1:" +ObjectGetter.toVector(mpog.iterator()));
        System.out.println("V2:" +ObjectGetter.toVector(mpog.iterator()));

        mpog.printGraph();

        System.out.println("merge result: " + mpog.merge(n1.get(3), n2.get(3)));
        mpog.printGraph();
        System.out.println("merge result: " + mpog.merge(n1.get(5), n2.get(7)));
        mpog.printGraph();
        System.out.println("merge result: " + mpog.merge(n1.get(1), n2.get(2)));


        System.out.println(ObjectGetter.toVector(mpog.iterator()));

        System.out.println("merge 4,8 result: " + mpog.merge(n1.get(4), n2.get(8)));


        System.out.println(ObjectGetter.toVector(mpog.iterator()));
    }

}