package lmu.utils.partialorder;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import gnu.trove.map.hash.TIntIntHashMap;
import gnu.trove.map.hash.TIntObjectHashMap;
import gnu.trove.map.hash.TObjectIntHashMap;
import lmu.utils.FRuntimeException;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.ObjectGetter;
import lmu.utils.Pair;
import lmu.utils.Region1D;
import lmu.utils.UPair;
import lmu.utils.VectorUtils;

public class PartialOrderIntGraph<E>  extends PartialOrderSetGraph<E>{
    private static int NOT_MATCHED_POSITION=-1;

    Logger log = LogConfig.getLogger();
    
    //position to node object
    TIntObjectHashMap<Node>[] lookup;
    
    //sorted added points
    Vector<Integer>[] points;
    
    //looup up from position to sorted points
    TIntIntHashMap[] pointLookup;;
    
    Node start;
    Node end;
    
    //E to internal id
    final TObjectIntHashMap<E> idMapping = new TObjectIntHashMap<>();
    
    //internal id to E
    final TIntObjectHashMap<E> revIdMapping = new TIntObjectHashMap<>();

    //final HashMap<UPair<Integer>,Vector<UPair<Region1D>>> ids2mapped_regions = new HashMap<>();;
    
    //internal id to length
    final List<Integer> lengths = new ArrayList<>();
    
    private int numEntities;
    
    private int nodeIdCounter = 0;
    
    private boolean init = false;
    
    class Node
    {
        int[] matches;
        int nodeId;
        
        Set<Node> preNodes;
        Set<Node> postNodes;
        
        public void addPosition(int sequenceIndex, int pos)
        {
            matches[sequenceIndex] = pos;
        }
        public int matches()
        {
            int ret =0;
            for(int i=0 ;i< numEntities; i++)
            {
                if ( matches[i]!=NOT_MATCHED_POSITION)
                    ret++;
            }
            return ret;
        }
        
        public HashSet<Pair<E,Integer>> getData()
        {
            HashSet<Pair<E,Integer>> ret = new HashSet<>();
            for(int i = 0 ; i< matches.length; i++)
            {
                if ( matches[i] == NOT_MATCHED_POSITION )
                    continue;

                ret.add(Pair.create(revIdMapping.get(i), matches[i]-1));
            }
            return ret;
        }
        
        public Node(int nodeId)
        {
            this.nodeId = nodeId;
            matches = new int[numEntities];
            for(int i = 0 ; i< numEntities; i++)
                matches[i] = NOT_MATCHED_POSITION;
            //System.out.println("NEW node:" + nodeId + " " + matches.length + " " + matches);
            preNodes = new HashSet<Node>();
            postNodes = new HashSet<Node>();
        }
        public String toString()
        {
            String ret ="";
            for(int i=0 ;i< numEntities; i++)
            {
                if ( matches[i]!=NOT_MATCHED_POSITION){
                    E obj = revIdMapping.get(i);
                    ret+=obj.toString() + ":" + (matches[i]-1) + " ";
                }
            }
            return ret.trim();
        }
    }
    
    @Override
    public RegionMatchInfoGetter<E> getRegionMatchInfo(boolean basic) {
        if ( basic == false) 
            throw new UnsupportedOperationException("Not yet implemented");
        
        return new BasicRegionMatchInfoGetter();
    }
    /*
    class BasicRegionMatchInfoGetter implements RegionMatchInfoGetter<E>
    {
        HashMap<UPair<E>, Vector<UPair<ARegion<E>>>> ids2mapped_regions = new HashMap<>();
        
        public BasicRegionMatchInfoGetter()
        {
            
            for (int i=0; i<numEntities; i++)
            {                             
                for (int j = 0; j<numEntities; j++)
                {
                    if (i == j)
                        continue;
                    
                    
                    UPair<E> p = UPair.createU(revIdMapping.get(i) ,revIdMapping.get(j) );
                    ids2mapped_regions.put(p, new Vector<UPair<ARegion<E>>>());
                }
            }
            
            for(Map.Entry<ARegion<E>, HashSet<ARegion<E>>>  e: region_mappings.entrySet())
            {
                ARegion<E> ar1 = e.getKey();
                for (ARegion<E> ar2 : e.getValue())
                {
                    ids2mapped_regions.get(UPair.createU(ar1.get0(), ar2.get0())).add(UPair.createU(ar1, ar2));
                }

            }
        
            //add also all matchings from iterator
            Iterator<HashSet<Pair<E,Integer>>> it = getGraphIterator();
            while (it.hasNext())
            {
                HashSet<Pair<E,Integer>> n = it.next();
                Vector<Pair<E,Integer>> v  = ObjectGetter.toVector(n);
                for (int i=0; i<v.size(); i++)
                {
                    Pair<E,Integer> p1 = v.get(i);
                    for (int j=0; j<v.size(); j++)
                    {
                        if (i == j)
                            continue;
                        
                        Pair<E,Integer> p2 = v.get(j);
                        ids2mapped_regions.get(UPair.createU(p1.getFirst(), p2.getFirst())).add(UPair.createU(new ARegion<E>(p1),new ARegion<E>(p2)));
                    }
                }
            }
            
            for (Vector<UPair<ARegion<E>>> v : ids2mapped_regions.values())
            {
                if (v.size() == 0)
                    continue;

                Collections.sort(v, (Pair<ARegion<E>,ARegion<E>> ar1, Pair<ARegion<E>,ARegion<E>> ar2) -> 
                                 { 
                                        int diff = ar1.getFirst().get1() - ar2.getFirst().get1(); 
                                        if (diff != 0) 
                                            return diff;

                                        return ar2.getSecond().get2() - ar1.getSecond().get2();
                                 }
                                );

                
                log.debug("%s vs %s: %s\n", v.get(0).getFirst().get0(), v.get(0).getSecond().get0(),v);

            }
        }
        @Override
        public int getOrder(HashSet<Pair<E, Integer>> n) {
            throw new UnsupportedOperationException("Not yet implemented");
        }

        @Override
        public Vector<UPair<Region1D>> getUnMatchedRegions(E obj1, E obj2) {
            if ( !idMapping.containsKey(obj1))
                throw new RuntimeException("Object:" + obj1 + " not registred");
            
            if ( !idMapping.containsKey(obj2))
                throw new RuntimeException("Object:" + obj2 + " not registred");

            int seq1 = idMapping.get(obj1);
            int seq2 = idMapping.get(obj2);

            return getUnmatched(seq1,seq2);
            
        }

        @Override
        public Vector<UPair<Region1D>> getMatchedRegions(E obj1, E obj2) {
  
            if ( !idMapping.containsKey(obj1))
                throw new RuntimeException("Object:" + obj1 + " not registred");
            
            if ( !idMapping.containsKey(obj2))
                throw new RuntimeException("Object:" + obj2 + " not registred");

            int seq1 = idMapping.get(obj1);
            int seq2 = idMapping.get(obj2);

            Vector<UPair<Region1D>> v = new Vector<>();
            UPair<Region1D> lastmatch = null;
            Vector<UPair<ARegion<E>>> matches = ids2mapped_regions.get(UPair.createU(obj1,obj2));
            if (matches == null)
                return v;
            
            System.out.println(matches.size());
            Iterator<UPair<ARegion<E>>> matchIt = matches.iterator();
            
            while (matchIt.hasNext())
            {
                UPair<ARegion<E>> p = matchIt.next();
                System.out.println("H:" + p + " " +lastmatch);
                if (lastmatch == null)
                {
                    v.add(lastmatch = UPair.createU(p.getFirst().toRegion(), p.getSecond().toRegion()));
                    continue;
                }

                if (p.getFirst().get1() <= lastmatch.getFirst().getX2() || p.getSecond().get1() <= lastmatch.getSecond().getX2() )
                {
                    lastmatch.getFirst().setX2(Math.max(p.getFirst().get2(), lastmatch.getFirst().getX2()));
                    lastmatch.getSecond().setX2(Math.max(p.getSecond().get2(), lastmatch.getSecond().getX2()));
                    continue;
                }

                v.add(lastmatch = UPair.createU(p.getFirst().toRegion(), p.getSecond().toRegion()));
                                    
            }
            return v;
            //return getMatched(seq1,seq2);
        }
        
    }
/*
    private Vector<UPair<Region1D>> getMatched(int sequenceIndex1, int sequenceIndex2)
    {
        Vector<UPair<Region1D>> ret = new Vector<>();
        
        UPair<Integer> key = null;
        if ( sequenceIndex1 < sequenceIndex2)
            key = UPair.createU(sequenceIndex1, sequenceIndex2);
        else
            key = UPair.createU(sequenceIndex2, sequenceIndex1);
        
        Vector<UPair<Region1D>> matchedRegions = this.ids2mapped_regions.get(key);
                
        
        //add also all matchings from iterator
        Iterator<HashSet<Pair<E,Integer>>> it = this.getGraphIterator();
        while (it.hasNext())
        {
            HashSet<Pair<E,Integer>> n = it.next();
            
            Vector<Pair<E,Integer>> v  = ObjectGetter.toVector(n);
           
            for (int i=0; i<v.size(); i++)
            {
                Pair<E,Integer> p1 = v.get(i);
                for (int j=i+1; j<v.size(); j++)
                {
                    if (i == j)
                        continue;
                    
                    Pair<E,Integer> p2 = v.get(j);
                    
                    //System.out.println(ids2mapped_regions.get(UPair.createU(p1.getFirst(), p2.getFirst())));
                    ids2mapped_regions.get(UPair.createU(p1.getFirst(), p2.getFirst())).add(UPair.createU(new Region1D(p1.getSecond(),p2.getSecond()),new Region1D(p2.getSecond(),p2.getSecond())));
                }
            }
        }

        
        Collections.sort(matchedRegions, (UPair<Region1D> r1, UPair<Region1D> r2) -> r1.getFirst().getX1()-r2.getFirst().getX1());
        //System.out.println("TMP:" + matchedRegions);
        
        if (matchedRegions == null)
            return ret;
        
        UPair<Region1D> lastmatch = null;
        
        Iterator<UPair<Region1D>> matchIt = matchedRegions.iterator();
       System.out.println(matchedRegions.size());
        while (matchIt.hasNext())
        {
            UPair<Region1D> p = matchIt.next();
            System.out.println("a:" + p + " "  + lastmatch);
            if (lastmatch == null)
            {
                ret.add(lastmatch = UPair.createU(p.getFirst(), p.getSecond()));
                continue;
            }

            if (p.getFirst().getX2() <= lastmatch.getFirst().getX2() || p.getSecond().getX2() <= lastmatch.getSecond().getX2() )
            {
                lastmatch.getFirst().setX2(Math.max(p.getFirst().getX2(), lastmatch.getFirst().getX2()));
                lastmatch.getSecond().setX2(Math.max(p.getSecond().getX2(), lastmatch.getSecond().getX2()));
                continue;
            }

            ret.add(lastmatch = UPair.createU(p.getFirst(), p.getSecond()));
                                
        }
        if ( sequenceIndex1 > sequenceIndex2)
        {
            //inverse
            Vector<UPair<Region1D>> tmp = new Vector<>();
            
            
            for(UPair<Region1D> u : ret)
                tmp.add(UPair.createU(u.getSecond(), u.getFirst()));
            
            ret = tmp;
        }
        return ret;
    }
    */
    public Vector<UPair<Region1D>> getUnmatched(int sequenceIndex1, int sequenceIndex2)
    {
        if ( !init){
            init();
        }
        
        Vector<UPair<Region1D>> ret = new Vector<>();
        /*
        Vector<UPair<Region1D>> mapped = this.ids2mapped_regions.get(UPair.createU(sequenceIndex1, sequenceIndex2));
        System.out.println("Mapped:" + mapped);
        
        Node lastMatch = start;
        for(int i = 0 ;i <points[sequenceIndex1].size();i++)
        {
            int index = points[sequenceIndex1].get(i);
            Node node = lookup[sequenceIndex1].get(index);
            
            if ( node.matches[sequenceIndex2] ==NOT_MATCHED_POSITION )
                continue;
            
            int d1 = node.matches[sequenceIndex1]-lastMatch.matches[sequenceIndex1];
            int d2 = node.matches[sequenceIndex2]-lastMatch.matches[sequenceIndex2];
            if ( d1 <= 1 || d2 <= 1) {
                lastMatch = node;
                continue;
            }
            
            ret.add(UPair.createU(new Region1D(lastMatch.matches[sequenceIndex1], node.matches[sequenceIndex1]-2),new Region1D(lastMatch.matches[sequenceIndex2], node.matches[sequenceIndex2]-2)));
            lastMatch = node;
        }
        */
        return ret;
    }
    @Override
    protected int getLastMatchedPoint(E obj) {
        int id = this.idMapping.get(obj);
        
        return points[id].get(points[id].size()-1)-1;
        
    }
    private void merge(Node n1, Node n2)
    {
        for(int i = 0 ; i< numEntities; i++)
        {
            //both not yet set
            if ( n2.matches[i] == NOT_MATCHED_POSITION && n1.matches[i] == NOT_MATCHED_POSITION )
                continue;
            
            //n2 not set
            if ( n2.matches[i] == NOT_MATCHED_POSITION && n1.matches[i] != NOT_MATCHED_POSITION )
                continue;
            
            //n2 set, n1 not => copy
            if ( (n2.matches[i] != NOT_MATCHED_POSITION && n1.matches[i] == NOT_MATCHED_POSITION) || (n1.matches[i] == n2.matches[i] && n1.matches[i] !=NOT_MATCHED_POSITION ) ){
               
                n1.matches[i] = n2.matches[i];
                //update lookup
                lookup[i].put(n1.matches[i], n1);
                
                continue;
            }
            
            //invalid situation
            throw new RuntimeException("Not possible:" + n1.matches[i]  + " " + n2.matches[i]);
        }
        n1.preNodes.addAll(n2.preNodes);
        n1.postNodes.addAll(n2.postNodes);
       
        
        ObjectGetter.apply(n2.preNodes, (Node pre) -> pre.postNodes.add(n1));
        ObjectGetter.apply(n2.postNodes, (Node post) -> post.preNodes.add(n1));
        
        ObjectGetter.apply(n2.preNodes, (Node pre) -> pre.postNodes.remove(n2));
        ObjectGetter.apply(n2.postNodes, (Node post) -> post.preNodes.remove(n2));
        
        
        for(Node pre : n1.preNodes)
        {
            pre.postNodes.removeAll(n1.postNodes);
        }
        for(Node post : n1.postNodes)
        {
            post.preNodes.removeAll(n1.preNodes);
        }
        
        
        log.debug("Merged node: %s (pre:%s) (post:%s)", n1,n1.preNodes,n2.postNodes);
     
        //delete all references
        n2.postNodes = null;
        n2.preNodes = null;
        n2.matches = null;
        
    }

    
    private Pair<Node,Integer> addPoint(int sequenceIndex, int pos)
    {
        if ( lookup[sequenceIndex].containsKey(pos)){
            return Pair.create(lookup[sequenceIndex].get(pos),pointLookup[sequenceIndex].get(pos));
        }
        
        Node node = new Node(this.nodeIdCounter++);
        node.addPosition(sequenceIndex, pos);
        lookup[sequenceIndex].put(pos, node);
        
        
        int index =  VectorUtils.insertSorted(points[sequenceIndex], pos, (Integer i, Integer j) -> i-j );
       
        
        //update indexes!
        for(int i = index+1; i< points[sequenceIndex].size();i++){
            int point = points[sequenceIndex].get(i);
            pointLookup[sequenceIndex].put(point, i);
            
        }
          
                    
        Integer preIndex = points[sequenceIndex].get(index-1);
        Integer postIndex = points[sequenceIndex].get(index+1);
        
        Node preNode = lookup[sequenceIndex].get(preIndex);
        Node postNode = lookup[sequenceIndex].get(postIndex);
        
        
        preNode.postNodes.add(node);
        postNode.preNodes.add(node);
        
        node.postNodes.add(postNode);
        node.preNodes.add(preNode);
        
        pointLookup[sequenceIndex].put(pos, index);
        
        log.debug("Add point %d for seqIndex: %d (pre:%s) (post:%s)",(pos-1),sequenceIndex,node.preNodes,node.postNodes);
        
        return Pair.create(node, index);
    }
    
    @Override
    public boolean match(E obj1, Region1D r1, E obj2, Region1D r2) 
    {
        
        if (r1.getLength() == 1 || r2.getLength() ==1) // 1:n matching
        {
            throw new FRuntimeException("invlaid matching %s:%s vs %s:%s one region is one long, the otherone >1 -> implies 1:n matching",obj1,r1,obj2,r2);
            
        }
        
        if ( !init){
            init();
        }
        if ( !idMapping.containsKey(obj1))
            throw new RuntimeException("Object:" + obj1 + " not registred");
        
        if ( !idMapping.containsKey(obj2))
            throw new RuntimeException("Object:" + obj2 + " not registred");

        int seq1 = idMapping.get(obj1);
        int seq2 = idMapping.get(obj2);

        boolean valid = this.match(seq1, r1.getX1(), seq2, r2.getX1(),false,true) && this.match(seq1, r1.getX2()-1, seq2, r2.getX2()-1,false,true);
        
        if ( !valid)
            return false;
        
        boolean ret = this.match(seq1, r1.getX1(), seq2, r2.getX1(),true,false) && this.match(seq1, r1.getX2()-1, seq2, r2.getX2()-1,true,false);
        
        if ( ret)
        {
            UPair<Integer> key = null;
            Region1D r1Tmp = new Region1D(r1.getX1(),r1.getX2()-1);
            Region1D r2Tmp = new Region1D(r2.getX1(),r2.getX2()-1);
            /*
            if ( seq1<seq2)
                ids2mapped_regions.get(UPair.createU(seq1, seq2)).add(UPair.createU(r1Tmp, r2Tmp));
            else    
                ids2mapped_regions.get(UPair.createU(seq2, seq1)).add(UPair.createU(r2Tmp, r1Tmp));
            */
            
            HashSet<ARegion<E>> h = region_mappings.get(new ARegion<E>(obj1, r1));
            if (h != null && h.contains(new ARegion<E>(obj2, r2)))
                return true; //mappings are knonw, region is knonw
            
            ARegion<E> ar1 = new ARegion(obj1, r1);
            ARegion<E> ar2 = new ARegion(obj2, r2);
            
            //log.debug("add " + ar1 +" <=> " +ar2);
            
            
            
            
            ObjectGetter.apply(region_mappings.get(ar1), (ARegion<E> mapped) -> { addMapping(mapped, ar2); addMapping(ar2, mapped); });
            ObjectGetter.apply(region_mappings.get(ar2), (ARegion<E> mapped) -> { addMapping(mapped, ar1); addMapping(ar1, mapped); });
            
        
            
            addMapping(ar1, ar2);
            addMapping(ar2, ar1);
            
         
            
        }
        return ret;
    }
   
    @Override
    public Collection<E> getRegionSources() 
    {
        return this.revIdMapping.valueCollection();
    }
    @Override
    public int getNumRegionSources() 
    {
        return this.numEntities;
    }
    
    public boolean match(E obj1,  int i, E obj2, int j)
    {
        if ( !init){
            init();
        }
        if ( !idMapping.containsKey(obj1))
            throw new RuntimeException("Object:" + obj1 + " not registred");
        
        if ( !idMapping.containsKey(obj2))
            throw new RuntimeException("Object:" + obj2 + " not registred");

        int seq1 = idMapping.get(obj1);
        int seq2 = idMapping.get(obj2);
        
        return match(seq1,i,seq2,j);
    }
    
    
    /**
     * Match positions i and j in sequenceIndex1 and sequenceIndex2 
     */
    public boolean  match(int sequenceIndex1,  int i, int sequenceIndex2, int j)
    {
        return match(sequenceIndex1,i,sequenceIndex2,j,false,false);
    }
    private boolean  match(int sequenceIndex1,  int i, int sequenceIndex2, int j,boolean isChecked, boolean onlyCheck)
    {
        if ( !init){
            init();
        }
        
        //translate positions; add one
        i = i+1;
        j = j+1;
        
        Pair<Node, Integer> n1 = addPoint(sequenceIndex1,i);
        Pair<Node, Integer> n2 = addPoint(sequenceIndex2,j);
        
        
        if ( n1.getFirst() == n2.getFirst()) return true;
        
        boolean consistent = isChecked;
        
        if (! isChecked)
        {
            
            Node preStop = null;
            for(int index = n1.getSecond();index > 0;index--)
            {
                int pointBefore = points[sequenceIndex1].get(index);
                Node nodeBefore = lookup[sequenceIndex1].get(pointBefore);
                
    
                if ( nodeBefore.matches[sequenceIndex2] != NOT_MATCHED_POSITION && nodeBefore.matches[sequenceIndex2]  < j)
                {
                    preStop = nodeBefore;
                    break;
                }
            }
            
            Node postStop = null;
            for(int index = n1.getSecond();index <- points[sequenceIndex1].size();index++)
            {
                int pointAfter = points[sequenceIndex1].get(index);
                Node nodeAfter = lookup[sequenceIndex1].get(pointAfter);
                
    
                if ( nodeAfter.matches[sequenceIndex2] != NOT_MATCHED_POSITION && nodeAfter.matches[sequenceIndex2]  > j)
                {
                    postStop = nodeAfter;
                    break;
                }
            }
            consistent = isConsistent(n1.getFirst(),n2.getFirst(),preStop,postStop);
            
            if ( !consistent) 
                return false;
        }
        
      //  System.out.println(preStop + " -- "  + sequenceIndex1 + ":" +i + "," + sequenceIndex2 + ":" + j);
        
        if ( onlyCheck)
            return true;
        
        log.debug("%d:%d and %d:%d => %b", sequenceIndex1,i-1,sequenceIndex2,j-1,consistent);
        
        merge(n1.getFirst(),n2.getFirst());
        
        return true;
    }
    
    /**
     * Check whether positions i and j in sequenceIndex1 and sequenceIndex2 can be matched 
     */
    private boolean isConsistent(Node node1, Node node2, Node preStop, Node postStop)
    {
        boolean c1 = preSufCheck(node1,node2,preStop,postStop);
        if ( !c1) return false;
        boolean c2 = preSufCheck(node2,node1,preStop,postStop);
        return c2;
    }
    
    private boolean  preSufCheck(Node node1, Node node2, Node preStop, Node postStop)
    {
        BitSet pre1 = new BitSet();
        BitSet suf2 = new BitSet();
        
        Stack<Node> preStack = new Stack<>();
        Stack<Node> sufStack = new Stack<>();
        
        preStack.add(node1);
        sufStack.add(node2);

        while(!preStack.isEmpty() || !sufStack.isEmpty())
        {
            if (! preStack.isEmpty())
            {
                Node preTop = preStack.pop();
                if ( pre1.get(preTop.nodeId)) continue;
                
                pre1.set(preTop.nodeId);
                
                if ( preTop == preStop) 
                {
                    preStack.clear();
                }else
                {
                    if ( suf2.get(preTop.nodeId)) return false;
                    preStack.addAll(preTop.preNodes);
                }
                
            }
          
            if (! sufStack.isEmpty())
            {
                Node postTop = sufStack.pop();
                if ( suf2.get(postTop.nodeId)) continue;
                
                suf2.set(postTop.nodeId);
                if ( postTop == postStop) 
                {
                    sufStack.clear();
                }else
                {
                    if ( pre1.get(postTop.nodeId)) return false;
                    sufStack.addAll(postTop.postNodes);
                }
                
            }
          
        }
        return true;
    }
    
   
    private void init()
    {
        if ( init)
        {
            log.warn("init method already called");
            return;
        }
        
     //   log.setLevel("info");
        numEntities = lengths.size();
        
        points = new Vector[numEntities];
        lookup = new TIntObjectHashMap[numEntities];
        pointLookup = new TIntIntHashMap[numEntities];
        
        start = new Node(nodeIdCounter++);
        end = new Node(nodeIdCounter++);
        
        start.postNodes.add(end);
        end.preNodes.add(start);

        
        for(int i = 0 ; i< numEntities;i++)
        {
         /*
            for(int j = i+1 ; j< numEntities;j++)
            {
                this.ids2mapped_regions.put(UPair.createU(i, j), new Vector<>());
            }
            */    
            points[i] = new Vector<Integer>();
            lookup[i] = new TIntObjectHashMap<Node>();
            pointLookup[i] = new TIntIntHashMap();
            
            start.addPosition(i,0);
            end.addPosition(i,lengths.get(i)+1);
            
            points[i].add(0);
            points[i].add(lengths.get(i)+1);
            
            lookup[i].put(0, start);
            lookup[i].put(lengths.get(i)+1, end);
        }
        this.init = true;
    }
    
    public void addRegionSource(E obj, int length) 
    {
        if ( init)
        {
            throw new RuntimeException("Cannot add objects after initialization");
        }
        super.objects.add(obj);
        this.idMapping.put(obj, lengths.size());
        this.revIdMapping.put(lengths.size(), obj);
        lengths.add(length);
        
    }
    public PartialOrderIntGraph( )
    {
        //empty
    }

    public GraphIterator getGraph()
    {
        return new GraphIterator();
    }
    
    @Override
    public Iterator<HashSet<Pair<E, Integer>>> getGraphIterator() {
        return new GraphIterator();
    }
   
    
    class GraphIterator implements Iterator<HashSet<Pair<E,Integer>>>
    {

        HashSet<Node> visited = new HashSet<>();
        
        LinkedList<Node> q = new LinkedList<Node>();
        
        int[] totalCurrentMatches = null;
        
        Comparator<Node> cmp =   (Node p1, Node p2) -> 
        {
            if ( p1 == p2)
                return 0;
            
            int pre1 =  0;
            for(Node pre : p1.preNodes) pre1+=visited.contains(pre)?0:1;
            int pre2 = 0;
            for(Node pre : p2.preNodes) pre2+=visited.contains(pre)?0:1;
            
            int diff= pre1-pre2;
            
            if ( diff != 0) return diff;
            
            for(int i = 0 ; i < p1.matches.length; i++)
            {
                if ( p1.matches[i] != NOT_MATCHED_POSITION ) pre1+=totalCurrentMatches[i];
                if ( p2.matches[i] != NOT_MATCHED_POSITION ) pre2+=totalCurrentMatches[i];
            }
            //System.out.println(diff + " " + pre1  + " " + pre2);
            
            return pre2-pre1;
        }
        ;
        
        HashSet<Pair<E,Integer>> next = null;
        Node current = null;
        
        public Node getCurrentNode()
        {
            return current;
        }
        
        public GraphIterator()
        {
            if (!init) init();
            q.add(start);
            totalCurrentMatches = new int[numEntities];
            int order = 0;
        }
        
        @Override
        public boolean hasNext() {
            while(!q.isEmpty())
            {
                Node node = q.poll();
                if ( visited.contains(node)) {
                    continue;
                } 
                //check
                for(Node pre : node.preNodes)
                {
                    if ( !visited.contains(pre)){
                        log.fatal(node + " q:" + q + " pre:" + pre);
                        for(Node o : q)
                        {
                            int pre2 = 0;
                            for(Node n : o.preNodes) pre2+=visited.contains(n)?0:1;
                            log.fatal("\t" + pre2 + "\n");
                        }
                        throw new RuntimeException("Inconsistent data");
                    }
                }
                visited.add(node);

                
                for(int i = 0 ; i< node.matches.length; i++)
                {
                    if ( node.matches[i]!=NOT_MATCHED_POSITION) 
                        totalCurrentMatches[i]++;
                }
               
               
                next = node.getData();
                current = node;
                
                for(Node post : node.postNodes)
                {
                    q.add(post);
                }
                Collections.sort(q, cmp);
                
                if ( node != start && node != end && next.size() > 1) {
                    return true;
                }
            }
            return false;
        }

        @Override
        public HashSet<Pair<E,Integer>> next() {
            return next;
        }
        
    }
    
    private void printGraph(PrintWriter pw)
    {
        Iterator<HashSet<Pair<E, Integer>>> it = this.getGraph();
        while(it.hasNext())
        {
           pw.println(it.next());
        }
        
        pw.flush();
    }
    public static void allMatchTest()
    {
        int n = 1000;

        List<Long> times = new ArrayList<>();
        
        long t1 = System.currentTimeMillis();

        PartialOrderRegionGraph<Integer> porg = new PartialOrderRegionGraph<Integer>();
        PartialOrderIntGraph<Integer> gs = new PartialOrderIntGraph<Integer>();
/*        
        porg.addRegionSource(1, n);
        porg.addRegionSource(2, n);
        porg.addRegionSource(3, n);
  */      
        int maxSeqs = 10;
        
        int[] seqs = new int[maxSeqs];
        
        for(int i = 0 ; i<seqs.length;i++){
            porg.addRegionSource(i, n);
            //seqs[i] = n;
            gs.addRegionSource(i, n);
        }
        gs.init();
        
        
        for(int l = 0 ; l< n; l++){
            for(int i = 0 ; i< maxSeqs ; i++)
            {
                for(int j = i+1; j<maxSeqs;j++)
                {
                  //  boolean m1 = porg.match(i, l, j, l);
                    boolean m2 = gs.match((Integer)i, l,(Integer)j,l);
                    boolean m1 = m2;
                  //  boolean m2 = m1;
                    if (m1 != m2)
                    {
                        throw new RuntimeException("ERROR");
                    }
                    
                }
                
            }
        }
        long total = System.currentTimeMillis()-t1;
        times.add(total);
        System.out.println(total);
                
                
             //   if ( i1 == i2) continue;
                
                
          
            //    boolean m2 = gs.match(i1, i2, rnd1, rnd2,false);
            /*
                if ( m1 != m2)
                {
                    System.err.println("Inconsistent:" + i1 + " " + rnd1  + " " + i2 + " " + rnd2) ;
                    System.out.println(m1 + " " + m2);
                    
                    gs.printGraph(new PrintWriter(System.out));
                    
                  //  m2 = gs.match(i1, i2, rnd1, rnd2,true);
                    
                  //  gs.printGraph(new PrintWriter(System.err));
                    
                    Iterator<HashSet<Pair<Integer, Integer>>> it = porg.getGraphIterator();
                    while(it.hasNext())
                    {
                        HashSet<Pair<Integer, Integer>> n = it.next();
                        if ( n.size() > 1) 
                            System.out.println(n);
                    }
                    
                    System.exit(1);
                }
             */   
            
      
         
           
        
        long mean  = lmu.utils.NumUtils.mean(times);
        
        System.out.println("Mean::"  + mean);
      
   
    }
    public static void testUnmatched()
    {
        PartialOrderRegionGraph<String> porg = new PartialOrderRegionGraph<String>();
        PartialOrderIntGraph<String> gs = new PartialOrderIntGraph<String>();
     
        porg.addRegionSource("A", 100);
        gs.addRegionSource("A",100);
        
        porg.addRegionSource("B", 100);
        gs.addRegionSource("B",100);
    
        porg.addRegionSource("C", 100);
        gs.addRegionSource("C",100);
    
        gs.init();

        porg.match("B", 20, "C", 50);
        porg.match("A", 10, "B", 40);
        //porg.match("A", 50, "B", 80);
        /*
        Iterator<HashSet<Pair<String, Integer>>> it = porg.getGraphIterator();
        System.out.println("PORG");
        while(it.hasNext())
        {
            System.out.println(it.next());
        }
        System.out.println("--");
        */
        gs.match("B", 20, "C", 50);
        gs.match("A", 10, "B", 40);
       // gs.match("A", 50, "B", 80);
    /*    
        List<Pair<Region1D<?>, Region1D<?>>> unmatch = gs.getUnmatched("A","C");
        System.out.println(unmatch);
        RegionMatchInfoGetter<String> basic = porg.getRegionMatchInfo(false,true);
        System.out.println(basic.getUnMatchedRegions("A","C"));;
        System.out.println("--");
    */
    }
    public static void testRandomRegionUnmatched()
    {
        Random rnd = new Random(96);
        
        int k = 3;
        
        int m = 100;
        
        
        //int n = 10000;
        int w = 50;
        
        long t1 = System.currentTimeMillis();

        PartialOrderIntGraph<Integer> intPorg = new PartialOrderIntGraph<>() ;
        PartialOrderRegionGraph<Integer> porg2 = new PartialOrderRegionGraph<>();
        
        for(int j = 0 ; j<k; j++)
        {
            intPorg.addRegionSource(j, m);
            porg2.addRegionSource(j, m);
        }
        /*
        for(int i = 0 ; i< 10;i++)
        {
            
            int rnd1S =rnd.nextInt(m-w)+1;
            int rnd2S =rnd.nextInt(m-w)+1;
            
            int rnd1E =rnd1S+5; //rnd1S + rnd.nextInt(w-1);
            int rnd2E =rnd2S+5; //rnd2S + rnd.nextInt(w-1);
            
            
            int i1 =rnd.nextInt(k);
            int i2 =rnd.nextInt(k);
            
            if ( i1 == i2) continue;
            Region1D r1 = new Region1D(rnd1S,rnd1E);
            Region1D r2 = new Region1D(rnd2S,rnd2E);
            
            if( r1.getLength() < 2 || r2.getLength() < 2) continue;
            //System.out.println(rnd1S + " " +rnd1E );
            //System.out.println(rnd2S + " " +rnd2E );
           
            boolean m2 =   porg2.match(i1, r1, i2, r2);
            
            boolean m1 = intPorg.match(i1, r1, i2, r2);
           // boolean m2 = m1;
            System.out.println(r1 + " " + r2);
            if ( m1 != m2)
            {
                System.out.println(m1 + " " + m2);
                System.out.println("ERROR");
            }
            
            //System.out.println(porg.getRegionMatchInfo(true).getMatchedRegions("1", "2"));
            
        }
        */
        //transitive
        intPorg.match(0, new Region1D(55,60), 2, new Region1D(55,60));
        porg2.match(0, new Region1D(55,60), 2, new Region1D(55,60));
        
        intPorg.match(1, new Region1D(55,60), 2, new Region1D(55,60));
        porg2.match(1, new Region1D(55,60), 2, new Region1D(55,60));
        
        intPorg.match(0, new Region1D(0,20), 1, new Region1D(0,20));
        porg2.match(0, new Region1D(0,20), 1, new Region1D(0,20));
        
        long time = System.currentTimeMillis()-t1;
        System.out.println(time);
        System.out.println("This (1,2)");
        System.out.println(intPorg.getRegionMatchInfo(true).getUnMatchedRegions(0, 1));
        System.out.println("Orig (1,2)");
        System.out.println(porg2.getRegionMatchInfo(true).getUnMatchedRegions(0, 1));
        
        /*
        System.out.println("Orig (2,1)");
        System.out.println(intPorg.getRegionMatchInfo(true).getUnMatchedRegions(2, 1));
        System.out.println("This (2,1)");
        System.out.println(porg2.getRegionMatchInfo(true).getUnMatchedRegions(2, 1));
        */
        
    }
   
    public static void  testRandom()
    {
        Random rnd = new Random(0);
        System.out.println("Match random");
        
        List<Long> times = new ArrayList<>();
        int  k = 10;
        
        for(int l = 0 ; l< 20; l++){
            
            PartialOrderRegionGraph<Integer> porg = new PartialOrderRegionGraph<Integer>();
            PartialOrderIntGraph<Integer> gs = new PartialOrderIntGraph<Integer>();
            
            
            for(int i = 0 ; i< k;i++){
                 gs.addRegionSource(i,1000);
                 porg.addRegionSource(i, 1000);
            }
            
            gs.init();
            long t1 = System.currentTimeMillis();
            for(int i = 0 ; i< 1000000;i++)
            {
                int rnd1 =rnd.nextInt(100);
                int rnd2 =rnd.nextInt(100);
                
                int i1 =rnd.nextInt(k);
                int i2 =rnd.nextInt(k);
                
                if ( i1 == i2) continue;
                
                
                boolean m1 = porg.match(i1, rnd1, i2, rnd2);
                boolean m2 = gs.match(i1,rnd1, i2 , rnd2);
                
                if ( m1 != m2)
                {
                    System.err.println("Inconsistent:" + i1 + " " + rnd1  + " " + i2 + " " + rnd2) ;
                    System.out.println(m1 + " " + m2);
                    
                    gs.printGraph(new PrintWriter(System.out));
                    
                  //  m2 = gs.match(i1, i2, rnd1, rnd2,true);
                    
                  //  gs.printGraph(new PrintWriter(System.err));
                    
                    Iterator<HashSet<Pair<Integer, Integer>>> it = porg.getGraphIterator();
                    while(it.hasNext())
                    {
                        HashSet<Pair<Integer, Integer>> n = it.next();
                        if ( n.size() > 1) 
                            System.out.println(n);
                    }
                    
                    System.exit(1);
                }
            }

         
            RegionMatchInfoGetter<Integer> basic = porg.getRegionMatchInfo(true);
             /*
            for(int i = 0 ; i< k; i++)
            {
                for(int j = i+1; j< k;j++)
                {
                    List<Pair<Region1D<?>, Region1D<?>>> unmatch = gs.getUnmatched((Integer)i,(Integer) j);
               //     System.out.println(unmatch);
                    Vector<UPair<Region1D>> unmatch2 = basic.getUnMatchedRegions((Integer)i, (Integer)j);
                 
                    if (unmatch.size() != unmatch2.size())
                    {
                        System.out.println(unmatch);
                        System.out.println(unmatch2);
                        
                        throw new RuntimeException("ERROR");
                        
                    }
                    
                    for(int w = 0 ; w< unmatch.size(); w++)
                    {
                        Pair<Region1D<?>, Region1D<?>> p1 = unmatch.get(w);
                        UPair<Region1D> p2 = unmatch2.get(w);
                        
                        if(p1.getFirst().getX1() != p2.getFirst().getX1() || p1.getFirst().getX2() != p2.getFirst().getX2() ){
                            System.out.println(p1 + " " + p2);
                            throw new RuntimeException("ERROR");
                        }
                        
                        if(p1.getSecond().getX1() != p2.getSecond().getX1() || p1.getSecond().getX2() != p2.getSecond().getX2() ){
                            System.out.println(p1 + " " + p2);
                            throw new RuntimeException("ERROR");
                        }
                        
                    }
                    //System.out.println(unmatch2);
                }
            }
            */
            long total = System.currentTimeMillis()-t1;
            
            times.add(total);
            System.out.println((total));
         
           
        }
        long mean  = lmu.utils.NumUtils.mean(times);
        
        System.out.println("Time::"  + (mean));
    }
    /*
    private static void completeMatch(Region r, GraphStructure<POMAGene> gs)
    {
        for (Pair<Integer,Integer> p: r.ra.ali.alignment)
        {
            int base1 = r.getSrcRegion().getX1()+(p.getFirst()*3);
            int base2 = r.getDstRegion().getX1()+(p.getSecond()*3);
            
            for (int i=0; i<3; i++)
            {
          
                //
                
                if ( base1< r.getSrcRegion().getX2() && base2 < r.getDstRegion().getX2())
                {
                   
                    boolean c_ret = gs.match(r.g1, base1, r.g2, base2);
                    
                    
                    
             //       System.out.println(i + " " + j + " " + c_ret);
                }
                base1++;
                base2++;
            }
        
        }
    }

    private static void completeMatch(Region r, PartialOrderRegionGraph<POMAGene> porg)
    {
        for (Pair<Integer,Integer> p: r.ra.ali.alignment)
        {
            int base1 = r.getSrcRegion().getX1()+(p.getFirst()*3);
            int base2 = r.getDstRegion().getX1()+(p.getSecond()*3);
            
            for (int i=0; i<3; i++)
            {
          
                //
                
                if ( base1< r.getSrcRegion().getX2() && base2 < r.getDstRegion().getX2())
                {
                   
                    boolean c_ret = porg.match(r.g1, new Region1D(base1,base1+1), r.g2, new Region1D(base2,base2+1));
                        
                        
                   
             //       System.out.println(i + " " + j + " " + c_ret);
                }
                base1++;
                base2++;
            }
        
        }
    }

   */

    public static void main(String[] args) 
    {
        testRandomRegionUnmatched();
        //testRandom();
      /*
        System.exit(1);
        
        SimpleOptionParser cmdLine = new SimpleOptionParser("isar_dir","chunk_file","out_dir","threads","new_imp");
        cmdLine.setDir("isar_dir","out_dir");
        cmdLine.setFile("chunk_file");
        cmdLine.setInt("threads");
        cmdLine.setSwitches("new_imp");
        cmdLine.setDefault("threads", "4");

        LogConfig lc = LogConfig.configureLog(null, true);

        if (!OptionParser.parseParams(args,true, false, true,  cmdLine,lc))
            return;
        
        //File xmlFile = new File("/home/proj/projekte/expressionlab/isar/msa-10/SPAC6F12.12.xml.gz");
        
        File baseDir = cmdLine.getFile("isar_dir");//new File("/home/proj/projekte/expressionlab/isar/msa-10/");
        File outDir = cmdLine.getFile("out_dir");
        
        System.out.println("Chunk file:" + cmdLine.getFile("chunk_file") + " New gs structure:" + cmdLine.isSet("new_imp"));
        Vector<String> lines = FileUtils.getLines(cmdLine.getFile("chunk_file"));
        
        ThreadUtils.runLambdas(lines, (String line)->
        {
            File f = new File(baseDir + "/" + line + ".gz");
            System.out.println("Process:" + f);
            if (! f.exists())
            {
                throw new RuntimeException(f + " does not exist");
            }
            
            if (! f.getName().endsWith(".xml.gz"))
                return;
           // if ( k.incrementAndGet() > 1000) return;
            
 //           System.out.println(f);
           
            BufferedReader reader = FileUtils.getReader(f);
            try {
                File out = new File(outDir + "/" + f.getName() + ".complete");
                initFromXML(XMLUtils.getReader(reader),out,new PrintWriter(System.out),cmdLine.isSet("new_imp"));
                reader.close();
            } catch (Exception e) {
                e.printStackTrace();
            }
        //    initFromXML(element,false);
        },cmdLine.getInt("threads"));
   //     gs.printGraph();
        */
       //
        //System.out.println(gs.isConsistent(0,1,8, 12));
     //   System.out.println(gs.isConsistent(0, 1, 8, 12));
        
        
      //  System.out.println(isConsistent(0,0,));
        
       // System.out.println(gs.match(2, 3, 50, 60));
        
      //  System.out.println(gs.match(1, 2, 15, 50));
        
    }
 
  
}
