package lmu.utils.partialorder;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;

public class PartialOrderIterator<A> implements Iterator<A> 
{

    LinkedList<DigraphNode<A>> zeroList = new LinkedList<>();
    Map<DigraphNode<A>, Integer> inDegrees = new HashMap<>(); // DigraphNode -> Integer

    public PartialOrderIterator(Iterator<DigraphNode<A>> iter)
    {
        // Initialize scratch in-degree values, zero list
        while (iter.hasNext())
        {
            DigraphNode<A> node = iter.next();
            int inDegree = node.getInDegree();
            inDegrees.put(node, inDegree);

            // Add nodes with zero in-degree to the zero list
            if (inDegree != 0)
                continue;
            
            zeroList.add(node);
        }
    }

    public boolean hasNext()
    {
        return !zeroList.isEmpty();
    }

    public A next()
    {
        DigraphNode<A> first = zeroList.removeFirst();

        // For each out node of the output node, decrement its in-degree
        Iterator<DigraphNode<A>> outNodes = first.getOutNodes();
        while (outNodes.hasNext())
        {
            DigraphNode<A> node = outNodes.next();
            int inDegree = inDegrees.get(node) - 1;
            inDegrees.put(node, inDegree);

            // If the in-degree has fallen to 0, place the node on the list
            if (inDegree != 0)
                continue;
            
            zeroList.add(node);
            
        }

        return first.getData();
    }
}

