package lmu.utils.partialorder;

import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import lmu.utils.BufferPrintWriter;
import lmu.utils.FRuntimeException;
import lmu.utils.IteratorUtils;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.NumUtils;
import lmu.utils.ObjectGetter;
import lmu.utils.Pair;
import lmu.utils.Region1D;
import lmu.utils.SetInfo;
import lmu.utils.UPair;
import lmu.utils.VectorUtils;
import lmu.utils.partialorder.PartialOrderGraph.GraphSorter;
import lmu.utils.tuple.Tuple3;

/**
 */
public class PartialOrderRegionGraph<A> extends PartialOrderSetGraph<A>
{
    /** the region matchings are stored as point matchings
     *  and additionally the linked regions are saved
     */
    Logger log = LogConfig.getLogger();
    //HashSet<Pair<A, Integer>> positions = new HashSet<Pair<A, Integer>>();
    protected HashMap<A, Vector<Pair<A,Integer>>> points = new HashMap<>();
    

    Comparator<Pair<A,Integer>> posC = (Pair<A,Integer> p1, Pair<A,Integer> p2) -> p1.getSecond().compareTo(p2.getSecond());
 
    MergablePartialOrderGraph<Pair<A,Integer>> graph = new MergablePartialOrderGraph<>();

   // HashMap<ARegion<A>, HashSet<ARegion<A>>> region_mappings = new HashMap<>();

    public int getNumRegionSources()
    {
        return points.size();
    }

    public Collection<A> getRegionSources()
    {
        return points.keySet();
    }

    public void addRegionSource(A obj, int length)
    {
        Vector<Pair<A,Integer>> v = points.get(obj);
        if (v != null)
        {
            if (v.get(v.size()-1).getSecond() == length-1)
                return; //already there
            
            throw new FRuntimeException("object %s already registered with different length, old: %d new: %d", obj, v.get(v.size()-1).getSecond()+1, length);
        }
        objects.add(obj);
        points.put(obj, v = new Vector<>());
        
        Pair<A,Integer> fp = Pair.create(obj, 0);
        Pair<A,Integer> lp = Pair.create(obj, length-1);
        VectorUtils.add(v, fp, lp);
        graph.addMergable(fp, lp);
        graph.setMRelation(fp, lp);
        
        //log.debug("lookup :  " +graph.lookup.keySet());
    }


    public Pair<A,Integer> addPoint(A a, int i)
    {
        Pair<A,Integer> x = Pair.create(a, i);
        //log.debug("add point: %s contains: %s", x, graph.contains(x));
        if (graph.contains(x))
            return x;

        //have to insert it, and add the corresponding relations
        Vector<Pair<A,Integer>> apoints = points.get(a);
        
        //log.debug("points :  " +apoints);
        //log.debug("lookup :  " +graph.lookup.keySet());
        int npos = VectorUtils.insertSorted(apoints, x);
        
        graph.addMergable(x);
        //log.debug("inserted x : " +x +" at pos : "+npos+ " get: " + apoints.get(npos)+ " length : "+ (VectorUtils.get(apoints, -1).getSecond()+1));
        graph.setMRelation(apoints.get(npos-1), x);     
        graph.setMRelation(x, apoints.get(npos+1));

        //log.debug("new point x: " +x +" node: " + graph.lookup.get(x) + " post: " + graph.lookup.get(x).post);
        return x;
    }


   

    public boolean match(A a1, int start1, int end1, A a2, int start2, int end2)
    {
        return match(a1, new Region1D(start1, end1, null), a2, new Region1D(start2, end2, null));
    }

    
    public boolean match(A a1, int pos1, A a2, int pos2)
    {
        return PartialOrderGraph.ACCEPT_STATE.INVALID != _match(a1, new Region1D(pos1, pos1+1, null), a2, new Region1D(pos2, pos2+1, null));
    }

    public boolean match(Tuple3<A, Integer, Integer> t1, Tuple3<A, Integer, Integer> t2)
    {
        return match(t1.get0(), t1.get1(), t1.get2(), t2.get0(), t2.get1(), t2.get2());
    }
    
    public boolean match(A a1, Region1D r1, A a2, Region1D r2)
    {
        switch(_match(a1,r1,a2,r2))
        {
            case INVALID:
                return false;
            case CONTAINED:
                //check if mapping known
                HashSet<ARegion<A>> h = region_mappings.get(new ARegion<A>(a1, r1));
                if (h != null && h.contains(new ARegion<A>(a2, r2)))
                    return true; //mappings are knonw, region is knonw
                
            default: 
                break;
        }
        
        
        ARegion<A> ar1 = new ARegion(a1, r1);
        ARegion<A> ar2 = new ARegion(a2, r2);
        
        //log.debug("add " + ar1 +" <=> " +ar2);
        ObjectGetter.apply(region_mappings.get(ar1), (ARegion<A> mapped) -> { addMapping(mapped, ar2); addMapping(ar2, mapped); });
        ObjectGetter.apply(region_mappings.get(ar2), (ARegion<A> mapped) -> { addMapping(mapped, ar1); addMapping(ar1, mapped); });
        
    
        
        addMapping(ar1, ar2);
        addMapping(ar2, ar1);
        

        
        return true;
    }
    
    boolean use_stop = false;

    public void setUseStop(boolean b)
    {
        use_stop = b;
    }

    Pair<PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>>, PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>>> 
        getStopNodes(Vector<Pair<A,Integer>> checkset, Vector<Pair<A,Integer>> v, Pair<A,Integer> center)
    {
        HashSet<PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>>> known = ObjectGetter.toSet(ObjectGetter.map(checkset, (Pair<A,Integer> p) -> graph.lookup.get(p)));

        int idx = v.indexOf(center);
            
        //System.out.printf("known: %d idx: %d\n", known.size(), idx);
        PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>> preStop = null;
        PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>> suffStop = null;

        for (int i=idx; i>=0; i--)
        {
            PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>> pn = graph.lookup.get(v.get(i));
            if (!known.contains(pn))
                continue;
            
            preStop = pn;
            break;
        }


        for (int i=idx; i<v.size(); i++)
        {
            PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>> pn = graph.lookup.get(v.get(i));
            if (!known.contains(pn))
                continue;
            
            suffStop = pn;
            break;
        }

        return Pair.create(preStop, suffStop);
    }


    Pair<PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>>, PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>>> getStopNodes(Pair<A,Integer> x1, Pair<A,Integer> x2)
    {
        PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>> preStop = null;
        PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>> suffStop = null;
        
        if (!use_stop)
            return Pair.create(preStop, suffStop);


        Vector<Pair<A,Integer>> v1 = points.get(x1.getFirst());
        Vector<Pair<A,Integer>> v2 = points.get(x2.getFirst());

        return (v1.size() < v2.size()) ? getStopNodes(v2, v1, x1) : getStopNodes(v1, v2, x2);
    }

    PartialOrderGraph.ACCEPT_STATE _match(A a1, Region1D r1, A a2, Region1D r2)
    {
        //log.debug("MATCDH: %s, %s, %s, %s", a1,r1, a2, r2);
        Pair<A,Integer> x11 = addPoint(a1, r1.getX1());
        Pair<A,Integer> x21 = addPoint(a2, r2.getX1());

        Pair<PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>>, PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>>> stop1 = getStopNodes(x11, x21);

        //System.out.println("stopnodes: " + stop1);

        if (r1.getLength() == 1 && r2.getLength() ==1) // 1:1 matching
        {
            return graph.merge(x11, x21, stop1.getFirst(), stop1.getSecond());
            
        }

        if (r1.getLength() == 1 || r2.getLength() ==1) // 1:n matching
        {
            throw new FRuntimeException("invlaid matching %s:%s vs %s:%s one region is one long, the otherone >1 -> implies 1:n matching",a1,r1,a2,r2);
            
        }

        Pair<A,Integer> x12 = addPoint(a1, r1.getX2()-1);
        Pair<A,Integer> x22 = addPoint(a2, r2.getX2()-1);

        //log.debug("point merge: %s vs %s mergable: %s", x11, x21, graph.checkMerge(x11, x21));
        
        
        
        PartialOrderGraph.ACCEPT_STATE state1  = graph.checkMerge(x11,x21, stop1.getFirst(), stop1.getSecond());
        if (state1 == PartialOrderGraph.ACCEPT_STATE.INVALID)
        {
            //log.info("invalid merge: %s vs %s",x11,x21);
            return state1;
        }
        
        
        Pair<PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>>, PartialOrderGraph.PartialNode<HashSet<Pair<A,Integer>>>> stop2 = getStopNodes(x12, x22);
        PartialOrderGraph.ACCEPT_STATE state2  = graph.checkMerge(x12,x22, stop2.getFirst(), stop2.getSecond());

        if (state2 == PartialOrderGraph.ACCEPT_STATE.INVALID)
        {
            //log.info("invalid merge: %s vs %s",x12,x22);
            return state2;
        }
        
        


        if (state1 == state2 && state1 == PartialOrderGraph.ACCEPT_STATE.CONTAINED)
            return state1;
        
        if (PartialOrderGraph.ACCEPT_STATE.INVALID == (state1 = graph.merge(x11, x21, stop1.getFirst(), stop1.getSecond())))
            return state1; //should not happen
        
        stop2 = getStopNodes(x12, x22);
        return graph.merge(x12, x22, stop2.getFirst(), stop2.getSecond());
        
        
    }
/*
    public class BasicRegionMatchInfoGetter implements RegionMatchInfoGetter<A>
    {
        HashMap<UPair<A>, Vector<UPair<ARegion<A>>>> ids2mapped_regions = new HashMap<>();
        HashMap<A, Integer> a2i = new HashMap<>();
        Vector<HashSet<Pair<A,Integer>>> onodes = new Vector<>();

        
        public int getOrder(HashSet<Pair<A,Integer>> n)
        {
            return onodes.indexOf(n);
        }

        public BasicRegionMatchInfoGetter()
        {
            
            Vector<A> as = ObjectGetter.toVector(points.keySet());
            for (int i=0; i<as.size(); i++)
            {                             
                a2i.put(as.get(i), i);
                for (int j = 0; j<as.size(); j++)
                {
                    if (i == j)
                        continue;
                    
                    
                    UPair<A> p = UPair.createU(as.get(i), as.get(j));
                    ids2mapped_regions.put(p, new Vector<UPair<ARegion<A>>>());
                }
            }
            
            for(Map.Entry<ARegion<A>, HashSet<ARegion<A>>>  e: region_mappings.entrySet())
            {
                ARegion ar1 = e.getKey();
                for (ARegion ar2 : e.getValue())
                {
                    ids2mapped_regions.get(UPair.createU(ar1.get0(), ar2.get0())).add(UPair.createU(ar1, ar2));
                }

            }
        
            //add also all matchings from iterator
            Iterator<HashSet<Pair<A,Integer>>> it = graph.iterator();
            while (it.hasNext())
            {
                HashSet<Pair<A,Integer>> n = it.next();
                onodes.add(n);
                Vector<Pair<A,Integer>> v  = ObjectGetter.toVector(n);
                for (int i=0; i<v.size(); i++)
                {
                    Pair<A,Integer> p1 = v.get(i);
                    for (int j=0; j<v.size(); j++)
                    {
                        if (i == j)
                            continue;
                        
                        Pair<A,Integer> p2 = v.get(j);
                        ids2mapped_regions.get(UPair.createU(p1.getFirst(), p2.getFirst())).add(UPair.createU(new ARegion(p1),new ARegion(p2)));
                    }
                }
            }
            
            for (Vector<UPair<ARegion<A>>> v : ids2mapped_regions.values())
            {
                if (v.size() == 0)
                    continue;

                Collections.sort(v, (Pair<ARegion<A>,ARegion<A>> ar1, Pair<ARegion<A>,ARegion<A>> ar2) -> 
                                 { 
                                        int diff = ar1.getFirst().get1() - ar2.getFirst().get1(); 
                                        if (diff != 0) 
                                            return diff;

                                        return ar2.getSecond().get2() - ar1.getSecond().get2();
                                 }
                                );

                
                log.debug("%s vs %s: %s\n", v.get(0).getFirst().get0(), v.get(0).getSecond().get0(),v);

            }
        }

        public Vector<UPair<Region1D>> getMatchedBoxes(A a1, A a2)
        {
            return ObjectGetter.map(ids2mapped_regions.get(UPair.createU(a1,a2)), (UPair<ARegion<A>> p ) -> UPair.createU(p.getFirst().toRegion(), p.getSecond().toRegion()));
        }

        public Vector<UPair<Region1D>> getUnMatchedRegions(A a1, A a2)
        {
            
            int l1 = 0;
            int l2 = 0;
            int L1 = VectorUtils.get(points.get(a1),-1).getSecond()+1;
            int L2 = VectorUtils.get(points.get(a2),-1).getSecond()+1;

            Vector<UPair<Region1D>> v = new Vector<>();
            Vector<UPair<Region1D>> matches = getMatchedRegions(a1,a2);
            matches.add(UPair.createU(new Region1D(L1, L1), new Region1D(L2, L2)));

            for (Pair<Region1D, Region1D> p : matches)
            {
                int e1 = p.getFirst().getX1()-1;
                int e2 = p.getSecond().getX1()-1;

                if (e1-l1 >= 0 && e2-l2 >= 0)
                {
                    v.add(UPair.createU(new Region1D(l1, e1), new Region1D(l2, e2)));
                }
                
                l1 = p.getFirst().getX2()+1;
                l2 = p.getSecond().getX2()+1;
            }
            return v;
        }

        public Vector<UPair<Region1D>> getMatchedRegions(A a1, A a2)
        {
            Vector<UPair<Region1D>> v = new Vector<>();
            UPair<Region1D> lastmatch = null;
            Vector<UPair<ARegion<A>>> matches = ids2mapped_regions.get(UPair.createU(a1,a2));
            if (matches == null)
                return v;
            
            System.out.println(matches.size());
            Iterator<UPair<ARegion<A>>> matchIt = matches.iterator();
            
            while (matchIt.hasNext())
            {
                UPair<ARegion<A>> p = matchIt.next();
                System.out.println("H:" + p + " " +lastmatch);
                if (lastmatch == null)
                {
                    v.add(lastmatch = UPair.createU(p.getFirst().toRegion(), p.getSecond().toRegion()));
                    continue;
                }

                if (p.getFirst().get1() <= lastmatch.getFirst().getX2() || p.getSecond().get1() <= lastmatch.getSecond().getX2() )
                {
                    lastmatch.getFirst().setX2(Math.max(p.getFirst().get2(), lastmatch.getFirst().getX2()));
                    lastmatch.getSecond().setX2(Math.max(p.getSecond().get2(), lastmatch.getSecond().getX2()));
                    continue;
                }

                v.add(lastmatch = UPair.createU(p.getFirst().toRegion(), p.getSecond().toRegion()));
                                    
            }
            return v;
        }

    }

*/
    
    
    
    class TransTester extends BasicRegionMatchInfoGetter
    {

        class APoints
        {
            A obj;
            Vector<Integer> points = new Vector<>();
            Vector<Integer> order = new Vector<>();

            HashMap<Integer, Integer> order_map = new HashMap<>();

            APoints(A o)
            {
                obj = o;
            }

            void addPoint(int a_point, int order_pos)
            {
                points.add(a_point);
                order.add(order_pos);
                order_map.put(order_pos, a_point);
            }

            int getPointPos(int point, int lower_limit)
            {
                return VectorUtils.binarySearch(points, point, lower_limit, points.size()).getSecond();
            }

            Integer nextPoint(int point, Integer default_next)
            {
                int idx = getPointPos(point, 0);
                if (idx == points.size() -1)
                    return default_next;
                
                return points.get(idx+1)-1;
            }

            Integer lastPoint(int point, Integer default_last)
            {
                int idx = getPointPos(point, 0);
                if (idx == 0)
                    return default_last;
                
                return points.get(idx-1)+1;
            }
            Vector<Integer> getFilteredEdges(int pos, int maxpos)
            {
                return ObjectGetter.filter(o_edges.get(pos), (Integer idx) -> maxpos > idx && ! order_map.containsKey(idx));
            }

            
            void addTransEdges(int pos, int maxpos)
            {
                LinkedList<Integer> q = new LinkedList<>();
                q.addAll(getFilteredEdges(pos, maxpos));
                HashSet<Integer> visited = new HashSet<>();

                while (!q.isEmpty())
                {
                    Integer n = q.poll();
                    if (visited.contains(n))
                        continue;
                    
                    visited.add(n);
                    //System.out.printf("\n next n: %d edges: %s\n", n, o_edges.get(n));
                    q.addAll(getFilteredEdges(n, maxpos));
                }
                t_edges.get(pos).addAll(visited);
            }

            

            HashMap<Integer,Integer> convert2PointMap(Map<Integer,Integer> ordermap, APoints ap)
            {
                return ObjectGetter.transformMap(ordermap, (Integer k) -> order_map.get(k), (Integer v) -> ap.order_map.get(v));
            }

            HashMap<Integer,HashSet<Integer>> convert2PointMapSet(Map<Integer,HashSet<Integer>> ordermap, APoints ap)
            {
                return ObjectGetter.transformMap(ordermap, (Integer k) -> order_map.get(k), (HashSet<Integer> h) -> ObjectGetter.mapToSet(h, (Integer v) -> ap.order_map.get(v)));
            }
            
            String getTransEdgeInfos(Vector<Integer> v)
            {
                BufferPrintWriter bpw = new BufferPrintWriter();

                for (int n : v)
                {
                    
                    HashSet<Integer> h = t_edges.get(n);
                    bpw.printf("node: %s:%d got %d edges (order: %s full: %s)\n", obj, order_map.get(n), h.size(),  n, ordered_nodes.get(n));
                    for (int e : h)
                    {
                        bpw.printf("\t\n", ordered_nodes.get(e));
                    }
                }
                return bpw.getBuffer();
            }

            /** returns the relevant order poisitions */
            Vector<Integer> updateTransEdges(int pre_match_pos, int pos_match_pos)
            {
                if (pre_match_pos >= points.size())
                    return null;
                
                
                //System.out.printf("%s updates trans edges: %d-%d (points: %d order: %d oedges size: %d)\n", obj, pre_match_pos, pos_match_pos, points.size(), order.size(), o_edges.size());
                int maxpos = (pos_match_pos < 0 || pos_match_pos >= order.size() ) ? ordered_nodes.size() : order.get(pos_match_pos);
                int end_point = (pos_match_pos < 0 || pos_match_pos >= order.size()) ? points.size() : pos_match_pos;


                //System.out.printf("it: %s\n", ObjectGetter.toVector(IteratorUtils.getIntIterator(pre_match_pos, end_point)));
                //System.out.printf("points: %s\n", ObjectGetter.toVector(ObjectGetter.getTransformedIterator(IteratorUtils.getIntIterator(pre_match_pos, end_point), (Integer idx) -> points.get(idx))));
                //System.out.printf("orders: %s\n", ObjectGetter.toVector(ObjectGetter.getTransformedIterator(IteratorUtils.getIntIterator(pre_match_pos, end_point), (Integer idx) -> order.get(idx))));
                Vector<Integer> v = ObjectGetter.toVector(getUnmatchedOrderPositions(pre_match_pos, end_point));
                /*log.info("%s update trans edges for unmatched points: %s (order positions: %s)", obj,
                         ObjectGetter.map(v, (Integer idx) -> order_map.get(idx)),
                         v);


                log.info("edges before: \n%s\n", getTransEdgeInfos(v));
                */
                ObjectGetter.apply(v, (Integer o) -> addTransEdges(o, maxpos));

                //log.info("edges after: \n%s\n", getTransEdgeInfos(v));
                return v;
            }

            Iterator<Integer> getUnmatchedOrderPositions(int pre_match_pos, int post_match_pos)
            {
                return ObjectGetter.getTransformedIterator(IteratorUtils.getIntIterator(pre_match_pos, post_match_pos), (Integer idx) -> order.get(idx));
            }

            Vector<Integer> getUnmatchedPositions(int pre_match_pos, int post_match_pos)
            {
                return ObjectGetter.map(IteratorUtils.getIntIterator(pre_match_pos, post_match_pos), (Integer idx) -> points.get(idx));
            }

        }


        HashMap<A, APoints> infos;
        HashMap<HashSet<A>, HashSet<HashSet<A>>> after = new HashMap<>();

        
        HashMap<HashSet<Pair<A,Integer>>, Integer> k2i = new HashMap<>();

        HashMap<Integer, HashSet<Integer>> o_edges = new HashMap<Integer, HashSet<Integer>>();
        HashMap<Integer, HashSet<Integer>> t_edges = new HashMap<Integer, HashSet<Integer>>();

        Vector<HashSet<Pair<A, Integer>>> ordered_nodes;

        HashMap<A, HashSet<Integer>> a2on = new HashMap<>();

        HashMap<UPair<A>, Vector<UPair<Region1D>> > unmappeds = new HashMap<>();

        public int getOrder(HashSet<Pair<A,Integer>> n)
        {
            return ordered_nodes.indexOf(n);
        }


        TransTester()
        {
            ordered_nodes = ObjectGetter.toVector(graph.iterator());
            final int L = ordered_nodes.size();

            infos = ObjectGetter.buildMap(points.keySet(), (A a) -> new APoints(a));

            HashMap<HashSet<Pair<A, Integer>>, Integer> data2order = new HashMap<>();
            for (int i = 0; i < L; i++)
            {
                HashSet<Pair<A, Integer>> op = ordered_nodes.get(i);

                data2order.put(op, i);
                for (Pair<A,Integer> p : op)
                {
                    infos.get(p.getFirst()).addPoint(p.getSecond(),i);
                }
            }
            
            //map edges to ordered points
            for (int i = 0; i < L; i++)
            {
                HashSet<Pair<A, Integer>> op = ordered_nodes.get(i);
                HashSet<Integer> successors = new HashSet<>();
                
                ObjectGetter.apply(graph.nodes.get(op).post, (PartialOrderGraph.PartialNode<HashSet<Pair<A, Integer>>> n) -> { Integer oi = data2order.get(n.data); if (oi ==null) return; successors.add(oi);});
                o_edges.put(i, successors);
            }
            for (Map.Entry<Integer, HashSet<Integer>> e : o_edges.entrySet())
            {
                HashSet<Integer> c = new HashSet<>();
                c.addAll(e.getValue());
                t_edges.put(e.getKey(), c);
            }
        }




                    
        
        public Vector<UPair<Region1D>> getUnMatchedRegions(A a1, A a2)
        {
            Vector<UPair<Region1D>> rv = unmappeds.get(UPair.create(a1, a2));
            if (rv != null)
            {
                return rv;
            }
            

            APoints ap1 = infos.get(a1);
            APoints ap2 = infos.get(a2);
                
            

            Vector<UPair<Region1D>>  um_regions = new Vector<>();
            
            //System.out.printf("%s vs %s\n", ap1.obj, ap2.obj);
            UPair<Integer> front = UPair.createU(0,0);

            Vector<UPair<Region1D>> v = getMatchedRegions(ap1.obj, ap2.obj);
            for (int ri = 0; ri <= v.size(); ri++)
            {

                //log.info("next match region: %s vs %s: %s\n", ap1.obj, ap2.obj, (ri >= v.size()) ? null : v.get(ri));
                front = getRegion(ap1, ap2, front, (ri==0) ? null : v.get(ri-1), (ri == v.size()) ? null : v.get(ri), um_regions);
            }

            unmappeds.put(UPair.createU(ap1.obj, ap2.obj), um_regions);
            unmappeds.put(UPair.createU(ap2.obj, ap1.obj), ObjectGetter.map(um_regions, (UPair<Region1D> p) -> UPair.swapU(p)));

            return um_regions;
                
        }
        
        void addMapping(Collection<UPair<Region1D>> v, int x11, int x12, int y11, int y12, boolean use_second)
        {
            if (x12 - x11 <= 0 ||  y12 - y11 <= 0)
                return;
            
            Region1D r1 = new Region1D(x11, x12);
            Region1D r2 = new Region1D(y11, y12);

            UPair<Region1D> up = UPair.createU((use_second) ? r2 : r1 , (use_second) ? r1 : r2);

            //log.info("add region pair: %s", up);
            v.add(up);

            

        }
                        
        void getUMRegions(HashSet<UPair<Region1D>> result, APoints own, APoints other, Vector<Integer> points, 
                          Map<Integer, Integer> pre, 
                          Map<Integer, Integer> post,
                          boolean use_second_in_pair,
                          UPair<Region1D> last, 
                          UPair<Region1D> next)
        {

            
            /*log.info("um: %s vs %s\n last: %s next: %s unmatched: %s\nother unmatched: %s\npre: %s post: %s", own.obj, other.obj, last, next, points, 
                     other.points, pre, post);
                     */

            int r1_start  = (last == null) ? 0 : last.get(!use_second_in_pair).getX2() +1;
            int r2_start  = (last == null) ? 0 : last.get(use_second_in_pair).getX2() +1;

            int r1_last = (next == null) ? VectorUtils.get(own.points, -1) : next.get(!use_second_in_pair).getX1() -1;
            int r2_last = (next == null) ? VectorUtils.get(other.points, -1) : next.get(use_second_in_pair).getX1() -1;

            //log.info("starts: %d, %d ends: %d, %d points: %s\n", r1_start, r2_start, r1_last, r2_last, points);
            int x1_start = r1_start;
            int x2_start = r2_start;

            Integer preP = null;
            for (int up : points)
            {
                //log.info("%s check: %d pre: %s post. %s preP: %s\n", own.obj, up, pre.get(up), post.get(up), preP);
                Integer npreP = pre.get(up);
                boolean new_pre = false;
                if (npreP != null && (preP == null || npreP > preP))
                {
                    preP = npreP;
                    new_pre= true;
                }
                     

                if (new_pre)
                {
                    //log.info("%d got new pre: %d", up, preP);
                    int x2_end = other.nextPoint(preP, r2_last);
                    //log.info("mapping: (%d-%d) -> (%d, %d)", x1_start, up-1, x2_start, x2_end);

                    addMapping(result, x1_start, up-1, x2_start, x2_end, use_second_in_pair);
                    //x2_start = x2_end+1;
                    x1_start = up;
                    x2_start = preP+1; //other.lastPoint(preP, x2_start);
                    //log.info("new open region: %s:%d-  %s:%d-", own.obj, x1_start, other.obj, x2_start);
                
                }
                
                Integer postP = post.get(up);

                if (postP == null)
                    continue;

                
                int x1_end = own.nextPoint(up, r1_last);
                //log.info("%d got post: %s x1end: %d open region: %s:%d- %s:%d-", up, postP, x1_end, own.obj, x1_start, other.obj, x2_start);


                //int x2_back = 
                //log.info("mapping: (%d-%d) -> (%d, %d)", x1_start, x1_end, x2_start, postP-1);
                
                
                addMapping(result, x1_start, x1_end, x2_start, postP-1, use_second_in_pair);

                x1_start = up+1;
                x2_start = postP;
                preP = null;
            }

            //log.info("%s %s mapping: (%d-%d) -> (%d, %d)", own.obj, other.obj, x1_start, r1_last, x2_start, r2_last);
            addMapping(result, x1_start, r1_last, x2_start, r2_last, use_second_in_pair);

        }

        Vector<HashSet<Pair<A,Integer>>> toNodes(Collection<Integer> c)
        {
            return ObjectGetter.map(c, (Integer idx) -> ordered_nodes.get(idx));
        }

        HashMap<Integer, HashSet<Integer>> getEdges(HashMap<Integer, HashSet<Integer>> edges, Collection<Integer> keys, HashSet<Integer> valset, 
                                                    APoints keypoints, APoints valpoints)
        {
            HashMap<Integer,HashSet<Integer>> hm = new HashMap<>();
            for (int p1 : keys)
            {
                Set<Integer> after_p1 = (valset == null) ? edges.get(p1) : SetInfo.intersect(edges.get(p1), valset);
                
                if (after_p1 == null || after_p1.size() == 0)
                    continue;
                

                hm.put(keypoints.order_map.get(p1), ObjectGetter.mapToSet(after_p1, (Integer idx) -> valpoints.order_map.get(idx)));
            }
            return hm;
        }

        Vector<Pair<HashSet<Pair<A,Integer>>, HashSet<Pair<A,Integer>>>> convertEdges(Map<Integer,Integer> m)
        {
            return ObjectGetter.map(m.entrySet(), (Map.Entry<Integer,Integer> e1) -> Pair.create(ordered_nodes.get(e1.getKey()), ordered_nodes.get(e1.getValue())));
        }

        HashMap<Integer, Integer> getRelevants(HashMap<Integer, HashSet<Integer>> pos)
        {
            HashMap<Integer, HashSet<Integer>> reverse_pre = ObjectGetter.reverseMultiMapH(pos);
            HashMap<Integer, Integer> rv = new HashMap<>();

            for (Map.Entry<Integer, HashSet<Integer>>  e : pos.entrySet())
            {
                int center = e.getKey();
                Integer post = NumUtils.min(e.getValue());
                Integer pre = NumUtils.max(reverse_pre.get(post));
                //log.info("test pair: %d-%d check: %d\n", up1, post, pre);
                if (center != pre)
                    continue;
                
                rv.put(center, post);
                //log.info("found valid pair: %s:%d->%s:%d\n", ap1.obj, up1, ap2.obj, post);
                
            }
            return rv;
        }
        UPair<Integer> getRegion(APoints ap1, APoints ap2, UPair<Integer> front, UPair<Region1D> last, UPair<Region1D> next, 
                                         Vector<UPair<Region1D>> result)
        {
            int pidx1 = 0;
            int pidx2 = 0;
            

            
            if (last != null)
            {
                pidx1 = ap1.getPointPos(last.getFirst().getX2(), front.getFirst()) +1;
                pidx2 = ap2.getPointPos(last.getSecond().getX2(), front.getSecond()) + 1;
            }

            int lidx1 = ap1.points.size();
            int lidx2 = ap2.points.size();

            if (next != null)
            {
                lidx1 = ap1.getPointPos(next.getFirst().getX1(), pidx1);
                lidx2 = ap2.getPointPos(next.getSecond().getX1(), pidx2);
            }
            
            Set<Integer> um1 = ObjectGetter.toSet(ap1.updateTransEdges(pidx1, lidx1));
            Set<Integer> um2 = ObjectGetter.toSet(ap2.updateTransEdges(pidx2, lidx2));
            
            if (lidx1 - pidx1 <= 0 || lidx2 - pidx2 <= 0)
            {
                return UPair.createU(lidx1, lidx2);
            }

            if (ObjectGetter.getSize(um1) <= 0 || ObjectGetter.getSize(um2) <= 0)
            {
                addMapping(result, ap1.points.get(pidx1), ap1.points.get(lidx1-1), ap2.points.get(pidx2), ap2.points.get(lidx2-1), false);
                
                return UPair.createU(lidx1, lidx2);
            }
            HashSet<UPair<Region1D>> regSet = new HashSet<>();
            
            //System.out.printf("next match: %s \n", p);
            //System.out.printf("unmatched points: %s\n",unmatched);
            //check in/out edges to the nodes
            //Vector<Integer> points = VectorUtils.join(um1, um2);
            //Set<Integer> points 

            
            Vector<Integer> um1v = ObjectGetter.toVector(ap1.getUnmatchedOrderPositions(pidx1, lidx1));
            
            Vector<Integer> um2v = ObjectGetter.toVector(ap2.getUnmatchedOrderPositions(pidx2, lidx2));


            /*
            //log.info("um1v: %s\n%s (order: %s)\num2v: %s\n%s (order: %s)\n", 

             */
            HashMap<Integer, HashSet<Integer>> um1_cpos = getEdges(t_edges, um1v, ObjectGetter.toSet(um2v), ap1, ap2);
            HashMap<Integer, HashSet<Integer>> um2_cpos = getEdges(t_edges, um2v, ObjectGetter.toSet(um1v), ap2, ap1);
            /*
            HashMap<Integer, HashSet<Integer>> um2_cpre = ObjectGetter.reverseMultiMapH(um1_cpos);
            HashMap<Integer, HashSet<Integer>> um1_cpre = ObjectGetter.reverseMultiMapH(um2_cpos);

            /*
            for (int up1 : ap1.getUnmatchedPositions(pidx1, lidx1))
            {

                HashSet<Integer> posts = um1_cpos.get(up1);
                log.info("test: %d post: %s\n", up1, posts);
                if (posts == null)
                    continue;
                
                Integer post = NumUtils.min(posts);
                log.info("reverse pre: %d:%s ", post, um2_cpre.get(post));
                Integer pre = NumUtils.max(um2_cpre.get(post));
                log.info("test pair: %d-%d check: %d\n", up1, post, pre);
                if (up1 != pre)
                    continue;
                
                
                log.info("found valid pair: %s:%d->%s:%d\n", ap1.obj, up1, ap2.obj, post);
                
            }

            
            log.info("relevants \n%s->%s: %s\n%s->%s: %s\n", 
                     ap1.obj, ap2.obj, getRelevants(um1_cpos), 
                     ap2.obj, ap1.obj, getRelevants(um2_cpos));
                     */

            HashMap<Integer, Integer> um1_post = getRelevants(um1_cpos);
            HashMap<Integer, Integer> um2_post = getRelevants(um2_cpos);
            
            HashMap<Integer, Integer> um1_pre = ObjectGetter.reverseMap(um2_post);
            //HashMap<Integer, Integer> um2_pre = ObjectGetter.reverseMap(um1_post);

            /*
            System.out.println("last: " +last);
            System.out.printf("prematch: %d,%d\n", ap1.points.get(pidx1), ap2.points.get(pidx2));
            System.out.printf("all um1 edges: %s\n", getEdges(t_edges, um1v, ObjectGetter.toSet(um2v), ap1, ap2));
            System.out.printf("all um2 edges: %s\n", getEdges(t_edges, um2v, ObjectGetter.toSet(um1v), ap2, ap1));
            System.out.printf("um1: %s pre: %s post: %s\n", um1v, um1_pre, um1_post);
            System.out.printf("um2: %s pre: %s post: %s\n", um2v, um2_pre, um2_post);
            //System.out.printf("postmatch: %d,%d\n", ap1.points.get(lidx1), ap2.points.get(lidx2));
            System.out.println("next: " +next);
            */

            getUMRegions(regSet, ap1, ap2, ObjectGetter.map(um1v, (Integer idx) -> ap1.order_map.get(idx)), um1_pre, um1_post, false, last, next);
            
            
            result.addAll(regSet);
            return UPair.createU(lidx1, lidx2);
            
        }
        
    }
/*
    public interface RegionMatchInfoGetter<A>
    {
        public int getOrder(HashSet<Pair<A,Integer>> n);
        public Vector<UPair<Region1D>> getUnMatchedRegions(A a1, A a2);
        public Vector<UPair<Region1D>> getMatchedRegions(A a1, A a2);
    }
    */
    public class ERegionMatchInfoGetter extends BasicRegionMatchInfoGetter
    {
        MergablePartialOrderGraph.UnorderedDecompositor dc;

        
        ERegionMatchInfoGetter(boolean basic)
        {
            if (basic)
                return;
            
            dc = graph.getUnorderedDecompositor();
            
            Vector<A> ids = ObjectGetter.toVector(points.keySet());
            
            for (int i=0; i<ids.size(); i++)
            {
                A a1 = ids.get(i);
                Vector<HashMap<A, Integer>> afterv = new Vector<>();
                Vector<Pair<A,Integer>> vp = points.get(a1);
                
                for(Pair<A,Integer> p : vp)
                {
                    HashSet<Pair<A,Integer>> h = graph.lookup.get(p).data;

                    HashMap<A,Integer> af = new HashMap<>();
                    afterv.add(af);

                    for (Pair<A, Integer> n : h)
                    {
                        if (n.getFirst() == a1)
                            continue;
                        
                        af.put(n.getFirst(), n.getSecond());
                    }

                    HashSet<HashSet<Pair<A, Integer>>> afterset = (HashSet<HashSet<Pair<A, Integer>>>)dc.after.get(h);
                    
                    for (A a2: ids)
                    {
                        if (a2 == a1 || af.containsKey(a2))
                            continue;
                        
                        Integer minafter = null;
                        for (HashSet<Pair<A,Integer>> hhs : afterset)
                        {
                            if (hhs == null)
                            {
                                continue;
                            }
                            for (Pair<A,Integer> pp : hhs)
                            {
                                if (pp.getFirst() != a2)
                                    continue;
                                
                                if (minafter != null && minafter < pp.getSecond())
                                    continue;
                                
                                minafter = pp.getSecond();
                            }
                        }
                        

                        af.put(a2, minafter);
                    }
                   
                }
                /*
                for (int j=0; j<ids.size(); j++)
                {
                    if (j == i)
                        continue;
                    
                    A a2 = ids.get(j);
                    System.out.printf("%s vs %s\n", a1, a2);
                    for (int k = 0; k < vp.size(); k++)
                    {
                        Integer afterP = afterv.get(k).get(a2);

                        if (afterP == null)
                            continue;
                        
                        System.out.printf("%d < %s ", vp.get(k).getSecond(), afterP);
                    }
                    System.out.println();
                }
                */
            }
            
                       
        }

        
            
        Region1D getQRegion(MergablePartialOrderGraph.CompositionElement ce, boolean q1)
        {

            Vector<Pair<A,Integer>> q_elems = (q1) ? ce.q1_elems : ce.q2_elems;

            if (q_elems.size() == 0)
                return null;
            
            return new Region1D(q_elems.get(0).getSecond(), q_elems.get(q_elems.size()-1).getSecond(), null);
        }


        int getLimit(Vector<Pair<Region1D, Region1D>>  limits, int pos, boolean reverse, boolean first)
        {
            int start = pos;
            int inc = (reverse) ? -1 : 1;
            int stop = (reverse) ? -1 : limits.size();
            Integer rv = null;
               
            for(int k=start + inc; k != stop; k += inc)
            {
                Pair<Region1D, Region1D> p = limits.get(k);
                Region1D r = (first) ? p.getFirst() : p.getSecond();
                if (r == null)
                    continue;
                
                return (reverse) ? r.getX2() + 1 : r.getX1() -1;
            }
            throw new FRuntimeException("invalid limit definition, or query (%d / %d) no limit found reverse: %s first: %s! (def was: %s)", 
                                        pos, limits.size(), 
                                        reverse, first,
                                        limits);
            
        }
        
        
        /** returns the regions where a1 and a2 are not matched, and for each region the partions
         * where corresponding regions may be matched onto each others
         */
        public Vector<UPair<Region1D>> getUnMatchedRegions(A a1, A a2)
        {
            if (dc == null)
            {
                return super.getUnMatchedRegions(a1,a2);
                //throw new RuntimeException("cannot retrieve unmatched region in basic MatchInfoGetter");
            }
            Vector<Pair<A, Integer>> a1elems = points.get(a1);
            Vector<Pair<A, Integer>> a2elems = points.get(a2);
            
            Vector<UPair<Region1D>> matches = getMatchedRegions(a1,a2);

            MergablePartialOrderGraph.UnorderedDecompositor.PairCompositionGetter pc = dc.decompose(a1elems, a2elems);

            Vector<UPair<Region1D>> unmatched_regions = new Vector<>();

            log.debug("matched regions %s vs %s: %s", a1, a2, matches);
            log.debug("unmatched %s vs %s\n", a1,a2);
            matches.insertElementAt(null, 0);
            matches.add(null);

            int L1 = VectorUtils.get(points.get(a1),-1).getSecond();
            int L2 = VectorUtils.get(points.get(a2),-1).getSecond();

            for (int i = 1; i < matches.size() ; i++)
            {
                UPair<Region1D> last = matches.get(i-1);
                UPair<Region1D> m = matches.get(i);

                Pair<A, Integer> lastE = (last == null) ? null : Pair.create(a1, last.getFirst().getX2());
                Pair<A, Integer> nextE = (m == null) ? null : Pair.create(a1, m.getFirst().getX1());

                Vector<MergablePartialOrderGraph.CompositionElement> v = ObjectGetter.filter(pc.getOrderedDecomposition(lastE, nextE), 
                                                                                             (MergablePartialOrderGraph.CompositionElement ce) -> ! ce.isMatched());

                //int start_i = (lastE == null) ? 0 : 1; //if lastE given, the first info is a match
                //int end_i = (nextE == null) ? v.size() : v.size()-1; //if nextE given, the last info is a match
                

                int reg_start_1 = (last == null) ? 0 : last.getFirst().getX2()+1;
                int reg_start_2 = (last == null) ? 0 : last.getSecond().getX2()+1;


                int reg_end_1 = (m == null) ? L1 : m.getFirst().getX1()-1;
                int reg_end_2 = (m == null) ? L2 : m.getSecond().getX1()-1;
                
                log.debug("match[%d]: %s last: %s reg1: %d-%d reg2: %d-%d\ndecopm:\n%s", i, m, last, reg_start_1, reg_end_1, reg_start_2, reg_end_2, v);
                //log.debug("start_i:  %d endi: %d diff: %d", start_i, end_i, end_i - start_i);

                if (v.size() < 1) //if the region is not splitted we have only one region
                {
                    if (reg_end_1 > reg_start_1 && reg_end_2 > reg_start_2)
                    {
                        //otherwise empty region
                        unmatched_regions.add(UPair.createU(new Region1D(reg_start_1, reg_end_1), new Region1D(reg_start_2, reg_end_2)));
                    }
                    
                    continue;
                }
                log.debug("got multi split region\n" + v);

                //for multi-split regions we now that  it is not allowed 
                //to match a pointin v[i] to a region in v[i+1]
                
                Integer reg1_end = null;
                Integer reg2_end = null;
                Vector<Pair<Region1D, Region1D>>  limits = new Vector<>();
                limits.add(Pair.create(new Region1D(reg_start_1-1, reg_start_1-1, null), new Region1D(reg_start_2-1, reg_start_2-1, null)));

                Pair<Region1D, Region1D> current_limit = Pair.create(null, null);

                /*the parts mapping only one of the elements on the beginning/end imply no constraint -> WRONG! :)

                int startJ = 0;
                if (v.get(startJ).isOneSide())
                {
                    boolean side1 = v.get(startJ).q1_elems.size() > 0;
                    while (startJ < v.size() && v.get(startJ).getQSize(!side1) == 0)
                    {
                        startJ++;
                    }
                }
                int endJ = v.size()-1;

                log.debug("endj: %d q1: %d q2: %d", endJ, v.get(endJ).getQSize(true), v.get(endJ).getQSize(false));
                if (v.get(endJ).isOneSide())
                {
                    
                    boolean side1 = v.get(endJ).q1_elems.size() > 0;
                    log.debug("endJ oneside: side1: %s", side1);
                    while (endJ >= 0 && v.get(endJ).getQSize(!side1) == 0)
                    {
                        endJ--;
                    }
                }
                log.debug("after fixing stale constraints check constraints in %d-%d: %s", startJ, endJ, VectorUtils.slice(v, startJ, endJ+1));
                */
                final int startJ = 0; final int endJ = v.size()-1;

                
                for (int j = startJ; j <= endJ; j++)
                {
                    MergablePartialOrderGraph.CompositionElement ce = v.get(j);
                    Region1D q1r = getQRegion(ce, true);
                    Region1D q2r = getQRegion(ce, false);
                    limits.add(Pair.create(q1r, q2r));
                }
                    
                limits.add(Pair.create(new Region1D(reg_end_1+1, reg_end_1+1, null), new Region1D(reg_end_2+1, reg_end_2+1, null)));

                log.debug("limits: %s", limits);
                //limits.add(Pair.create(new Region1D(reg_end_1, reg_end_1, null), new Region1D(reg_end_2, reg_end_2, null)));

                for (int li=1; li<limits.size()-1; li++)
                {
                    int x11 = getLimit(limits, li, true, true);
                    int x12 = getLimit(limits, li, false, true);

                    int x21 = getLimit(limits, li, true, false);
                    int x22 = getLimit(limits, li, false, false);

                    log.debug("query limit :  " +limits.get(li));
                    log.debug("next limits: (%d-%d) vs (%d, %d)", x11, x12, x21, x22);
                    if (x11 > x12 || x21 > x22)
                    {
                        //empty region
                        continue;
                    }
                    unmatched_regions.add(UPair.createU(new Region1D(x11, x12), new Region1D(x21, x22)));
                    
                }
            }

            return unmatched_regions;
        }

    }
    
    public RegionMatchInfoGetter<A> getRegionMatchInfo()
    {
        return getRegionMatchInfo(true);
    }

    /*public RegionMatchInfoGetter getRegionMatchInfo(boolean basic)
    {
        //return (basic) ?  new ERegionMatchInfoGetter(basic) : new TransTester();
    }
    */

    public RegionMatchInfoGetter<A> getRegionMatchInfo(boolean basic)
    {
        return getRegionMatchInfo(basic, false);
    }

    public RegionMatchInfoGetter getRegionMatchInfo(boolean basic, boolean slower)
    {
        
        return (basic || slower) ? new ERegionMatchInfoGetter(basic) : new TransTester();
    } 
    



    public Iterator<HashSet<Pair<A, Integer>>> getGraphIterator(GraphSorter gs) 
    {

        
        return graph.iterator(gs);
    }

    public Iterator<HashSet<Pair<A, Integer>>> getGraphIterator() 
    {
        return graph.iterator();
    }

    @Override
    protected int getLastMatchedPoint(A obj) {
       return VectorUtils.get(points.get(obj),-1).getSecond()+1;
    }

	
}
