package lmu.utils.partialorder;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import lmu.utils.ObjectGetter;
import lmu.utils.Pair;
import lmu.utils.Region1D;
import lmu.utils.UPair;

public abstract class PartialOrderSetGraph<E> {
    
    protected HashMap<ARegion<E>, HashSet<ARegion<E>>> region_mappings = new HashMap<>();
    
    //TODO: adapted PartialOrderIntGraph points approach
    //protected Vector<Integer>[] points;
    protected List<E> objects = new ArrayList<>();
    
    void addMapping(ARegion<E> ar1, ARegion<E> ar2)
    {
        HashSet<ARegion<E>> h = region_mappings.get(ar1);
        if (h == null)
        {
            region_mappings.put(ar1, h = new HashSet<>());
        }
        h.add(ar2);
    }
    
    public abstract boolean match(E obj1,  int i, E obj2, int j);
    
    public abstract boolean match(E obj1,  Region1D r1, E obj2, Region1D r2);
    
    public abstract void addRegionSource(E obj, int length);
    
    public abstract Collection<E> getRegionSources();
    
    public abstract int getNumRegionSources();
    
    public abstract RegionMatchInfoGetter<E> getRegionMatchInfo(boolean basic);
    
    public abstract Iterator<HashSet<Pair<E, Integer>>> getGraphIterator();
    
    
    public class BasicRegionMatchInfoGetter implements RegionMatchInfoGetter<E>
    {
        HashMap<UPair<E>, Vector<UPair<ARegion<E>>>> ids2mapped_regions = new HashMap<>();
        //HashMap<E, Integer> a2i = new HashMap<>();
        Vector<HashSet<Pair<E,Integer>>> onodes = new Vector<>();

        
        public int getOrder(HashSet<Pair<E,Integer>> n)
        {
            return onodes.indexOf(n);
        }
       
        public BasicRegionMatchInfoGetter()
        {
            for (int i=0; i<objects.size(); i++)
            {                             
             //   a2i.put(objs.get(i), i);
                for (int j = 0; j<objects.size(); j++)
                {
                    if (i == j)
                        continue;
                    
                    
                    UPair<E> p = UPair.createU(objects.get(i), objects.get(j));
                    ids2mapped_regions.put(p, new Vector<UPair<ARegion<E>>>());
                }
            }
            for(Map.Entry<ARegion<E>, HashSet<ARegion<E>>>  e: region_mappings.entrySet())
            {
                ARegion ar1 = e.getKey();
                for (ARegion ar2 : e.getValue())
                {
                    ids2mapped_regions.get(UPair.createU(ar1.get0(), ar2.get0())).add(UPair.createU(ar1, ar2));
                }

            }
        
            //add also all matchings from iterator
            Iterator<HashSet<Pair<E,Integer>>> it = getGraphIterator();
            while (it.hasNext())
            {
                HashSet<Pair<E,Integer>> n = it.next();
                onodes.add(n);
                Vector<Pair<E,Integer>> v  = ObjectGetter.toVector(n);
                for (int i=0; i<v.size(); i++)
                {
                    Pair<E,Integer> p1 = v.get(i);
                    for (int j=0; j<v.size(); j++)
                    {
                        if (i == j)
                            continue;
                        
                        Pair<E,Integer> p2 = v.get(j);
                        ids2mapped_regions.get(UPair.createU(p1.getFirst(), p2.getFirst())).add(UPair.createU(new ARegion(p1),new ARegion(p2)));
                    }
                }
            }
            
            for (Vector<UPair<ARegion<E>>> v : ids2mapped_regions.values())
            {
                if (v.size() == 0)
                    continue;

                Collections.sort(v, (Pair<ARegion<E>,ARegion<E>> ar1, Pair<ARegion<E>,ARegion<E>> ar2) -> 
                                 { 
                                        int diff = ar1.getFirst().get1() - ar2.getFirst().get1(); 
                                        if (diff != 0) 
                                            return diff;

                                        return ar2.getSecond().get2() - ar1.getSecond().get2();
                                 }
                                );

                
                //log.debug("%s vs %s: %s\n", v.get(0).getFirst().get0(), v.get(0).getSecond().get0(),v);

            }
        }

        public Vector<UPair<Region1D>> getMatchedBoxes(E a1, E a2)
        {
            return ObjectGetter.map(ids2mapped_regions.get(UPair.createU(a1,a2)), (UPair<ARegion<E>> p ) -> UPair.createU(p.getFirst().toRegion(), p.getSecond().toRegion()));
        }

        public Vector<UPair<Region1D>> getUnMatchedRegions(E a1, E a2)
        {
            
            int l1 = 0;
            int l2 = 0;
            int L1 = getLastMatchedPoint(a1);//VectorUtils.get(points.get(a1),-1).getSecond()+1;
            int L2 = getLastMatchedPoint(a2); //VectorUtils.get(points.get(a2),-1).getSecond()+1;

            Vector<UPair<Region1D>> v = new Vector<>();
            Vector<UPair<Region1D>> matches = getMatchedRegions(a1,a2);
            
            matches.add(UPair.createU(new Region1D(L1, L1), new Region1D(L2, L2)));

            for (Pair<Region1D, Region1D> p : matches)
            {
                int e1 = p.getFirst().getX1()-1;
                int e2 = p.getSecond().getX1()-1;

                if (e1-l1 >= 0 && e2-l2 >= 0)
                {
                    v.add(UPair.createU(new Region1D(l1, e1), new Region1D(l2, e2)));
                }
                
                l1 = p.getFirst().getX2()+1;
                l2 = p.getSecond().getX2()+1;
            }
            return v;
        }

        public Vector<UPair<Region1D>> getMatchedRegions(E a1, E a2)
        {
            Vector<UPair<Region1D>> v = new Vector<>();
            UPair<Region1D> lastmatch = null;
            Vector<UPair<ARegion<E>>> matches = ids2mapped_regions.get(UPair.createU(a1,a2));
            if (matches == null)
                return v;
            
            Iterator<UPair<ARegion<E>>> matchIt = matches.iterator();
            
            while (matchIt.hasNext())
            {
                UPair<ARegion<E>> p = matchIt.next();
                if (lastmatch == null)
                {
                    v.add(lastmatch = UPair.createU(p.getFirst().toRegion(), p.getSecond().toRegion()));
                    continue;
                }

                if (p.getFirst().get1() <= lastmatch.getFirst().getX2() || p.getSecond().get1() <= lastmatch.getSecond().getX2() )
                {
                    lastmatch.getFirst().setX2(Math.max(p.getFirst().get2(), lastmatch.getFirst().getX2()));
                    lastmatch.getSecond().setX2(Math.max(p.getSecond().get2(), lastmatch.getSecond().getX2()));
                    continue;
                }

                v.add(lastmatch = UPair.createU(p.getFirst().toRegion(), p.getSecond().toRegion()));
                                    
            }
            return v;
        }

    }
    protected abstract int getLastMatchedPoint(E obj);
}
