package lmu.utils.plotting;


import lmu.utils.*;
import java.beans.*;
import lmu.utils.ImageUtils.ImagePanel;
import java.awt.image.*;
import java.util.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.event.*;
import org.jdom.*;

public class AbstractZoomedMultiPicPlane extends AbstractMultiPicPlane implements ZoomedMultiPicPlane
{
    Logger log = LogConfig.getLogger();
    int data_width=0;
    int zoomed_data_width=-1;

    ZoomedCoordinateConverter zoom_converter = null;

    Vector<RegionPicturePane> regions = new Vector<RegionPicturePane>();
    Vector<DataPicturePane> datas = new Vector<DataPicturePane>();
    Vector<HeaderedPicturePane> order_save = new Vector<HeaderedPicturePane>();
    
    DataAggregator.TYPE default_aggreg_type = DataAggregator.TYPE.AVG;


    Vector<RegionPicturePane> zoomed_regions = new Vector<RegionPicturePane>();
    Vector<DataPicturePane> zoomed_datas = new Vector<DataPicturePane>();


    public AbstractZoomedMultiPicPlane(JInternalFrame jif, JFrame jframe, int data_width)
    {
        super(jframe, jif);
        this.data_width=data_width;
        setZoomedConverter(new ZoomedCoordinateConverter(data_width));
    }

    public void reset(int newlength)
    {
        super.clear();
        data_width=newlength;
    }

    public void setZoomedConverter(ZoomedCoordinateConverter zoom_converter)
    {
        this.zoom_converter=zoom_converter;
    }

    Vector<HeaderedPicturePane> getCurrentInfos()
    {
        if (zoom_converter==null)
        {
            return order_save;
        }
        Vector<HeaderedPicturePane> rv = new Vector<HeaderedPicturePane>();
        int idx=0;
        while (idx<order_save.size())
        {
            Object o = order_save.get(idx++);
            
            boolean added=false;
            for (int i=0; !added && i<regions.size(); i++)
            {
                if(o != regions.get(i))
                    continue;
                
                rv.add(zoomed_regions.get(i));
                added=true;
                break;
            }


            for (int i=0; !added && i<datas.size(); i++)
            {
                if(o != datas.get(i))
                    continue;
                
                rv.add(zoomed_datas.get(i));
                added=true;
                break;
            }
        }
        return rv;

    }
    public void update()
    {
        update(false);
    }
    public void update(boolean force)
    {
     //   System.out.println("zoom converter: "+ zoom_converter);
        if (!force && ( zoom_converter==null || zoom_converter.isUp2Date()))
        {
            super.update();
            return;
        }
        //System.out.printf("will update , got %d regions and %d data\n", regions.size(), datas.size()); 
        if (zoom_converter != null)
        {
            zoom_converter.update();
        }
        
        //Vector<HeaderedPicturePane> order_save = new Vector<HeaderedPicturePane>();
        //order_save.addAll(panes);

        Vector<Boolean> hiddens = new Vector<Boolean>();
        hiddens.addAll(super.hidden);

        super.clear();

        
        

        
        //create zooomed infos
        zoomed_regions.clear();
        zoomed_datas.clear();
        

        zoomed_data_width = (zoom_converter == null) ? data_width : zoom_converter.getZoomedLength();

        
        log.debug("update zooming...in length: %d zoomed lenght: %d\n", data_width, zoomed_data_width);
        for (RegionPicturePane rpp:regions)
        {
            zoomed_regions.add( new RegionPicturePane(rpp, zoom_converter) );
        }

        for (DataPicturePane dpp: datas)
        {
            zoomed_datas.add(new DataPicturePane(dpp, zoom_converter, default_aggreg_type));
        }

        for (RegionPicturePane rpp: zoomed_regions)
        {
            rpp.update();
            log.debug("updated : "+rpp.name+" dim: "+rpp.getPictureDimension());
        }
        for (DataPicturePane dpp: zoomed_datas)
        {
            dpp.update();
        }
        int missing = zoomed_regions.size()+zoomed_datas.size();

        Vector<HeaderedPicturePane> vhpp = getCurrentInfos();
        
        int idx=0;
        for (HeaderedPicturePane hpp: vhpp)
        {
            addInfo(hpp);
            if (hiddens.get(idx++))
            {
                setHidden(hpp,true);
            }
        }
        

        super.update();
    }
    public RegionPicturePane addRegionPane(RegionPicturePane rp)
    {
        if (rp.getDataWidth()!=data_width)
        {
            throw new RuntimeException("invalid length of region plane, expected "+data_width+" got "+rp.getDataWidth());
        }
        addInfo(rp);
        regions.add(rp);
        return rp;

    }
    public RegionPicturePane createRegionPane(String name)
    {
        return createRegionPane(name, null);
    }
    public RegionPicturePane createRegionPane(String name, Object userObject)
    {
        return addRegionPane(new RegionPicturePane(name, data_width, userObject));
        
    }
    
    public DataPicturePane createDataPane(String name)
    {
        return createDataPane(name,null);
    }

    public DataPicturePane addDataPane(DataPicturePane dp)
    {
        if (dp.getDataWidth()!=data_width)
        {
            throw new RuntimeException("invalid length of data plane, expected "+data_width+" got "+dp.getDataWidth());
        }
        addInfo(dp);
        datas.add(dp);
        return dp;
    }



    public DataPicturePane createDataPane(String name, Object userObject)
    {
        DataPicturePane dp = new DataPicturePane(name, data_width);
        dp.setObject(userObject);
        return addDataPane(dp);
        
    }

    public void addInfo(HeaderedPicturePane hpp)
    {
        /*if (hpp.getDataWidth()!=data_width)
        {
            throw new RuntimeException("in "+ this.getClass().getName()+" all data sections have to be the same width! configured: "+data_width+
                                       " got from: "+hpp.getName()+":"+hpp.getDataWidth());
        }
        */
        order_save.add(hpp);
        super.addInfo(hpp);
    }


    /** creates a simplified serialized xml from the 
     * data shown - no infos left from zooming etc, but
     * it is thought for the server-client architecture, so
     * the server has a copy of this pane, having all needed informations
     */
    public Element getSerializedXML()
    {
        Element root = new Element("frame");
        root.setAttribute("width", ""+((zoom_converter==null)?data_width:zoomed_data_width));

        int idx=0;
        for(HeaderedPicturePane hpp: getCurrentInfos())
        {
            if(hidden.get(idx++))
                continue;
            
            if (hpp instanceof RegionPicturePane)
            {
                root.addContent(((RegionPicturePane)hpp).getSerializedXML());
            }
            if (hpp instanceof DataPicturePane)
            {
                root.addContent(((DataPicturePane)hpp).getSerializedXML());
            }
        }
        
        return root;
    }
    public void init(Element root)
    {
        int width = XMLUtils.getIntAttrib(root, "width", null);
        data_width=width;
        for (Element e : XMLUtils.getChildren(root))
        {
            if (e.getName().equals(RegionPicturePane.ELEM_NAME))
            {

                RegionPicturePane rpp = new RegionPicturePane(e);
                addInfo(rpp);
                regions.add(rpp);
                continue;
            }
            if (e.getName().equals(DataPicturePane.ELEM_NAME))
            {
                DataPicturePane dpp = new DataPicturePane(e);
                addInfo(dpp);
                datas.add(dpp);
                continue;
            }
            log.warn("unknown elem: "+e.getName());
        }
        
        update();
    }

    public ZoomedCoordinateConverter getZoomedConverter()
    {
        return zoom_converter;
    }
}

