package lmu.utils.plotting;

import java.awt.image.*;
import java.awt.*;
import java.util.*;
import lmu.utils.*;
import org.jdom.*;

public class DataPlot extends MixedDataIterator
{
    public final static String ELEM_NAME="dataplot";
    String name="unset";
    Logger log = LogConfig.getLogger();

    PlotYAxisImage header;

    Object userObject=null;
    Color color = Color.BLACK;
    BufferedImage img = null;

    public DataPlot(Element xml)
    {
        super(XMLUtils.getIntAttrib(xml, "length", null));
        setName(XMLUtils.getAttrib(xml, "name", null, true));
        color = XMLUtils.getColor(xml, null, true);
        CDATA data = null;
        for (Object o : xml.getContent())
        {
            if (!(o instanceof CDATA))
            {
                continue;
            }
            data = (CDATA)o;
        }
        if (data==null)
        {
            throw new RuntimeException("no data set to data plot "+name+" length: "+getDataLength());
        }
        setDenseData(XMLUtils.toDoubleArray(data));
        
    }

    public DataPlot(PlotYAxisImage header, int length)
    {
        super(length);
        this.header = header;
    }

    public interface ZoomedDataProvider
    {
        public double getValue(int start, int end);
    }
    
    public static class ZoomAggregator
    {
        double[] aggregated;
        ZoomedCoordinateConverter zcc;
        Logger log = LogConfig.getLogger();

        public ZoomAggregator(ZoomedCoordinateConverter zcc, Iterator<Pair<Integer,Double>> data, DataAggregator.TYPE aggreg_type)
        {
            this.zcc = zcc;
            DataAggregator aggr = new DataAggregator(aggreg_type, zcc.getZoomedLength());

            log.info("get zoomed aggregator type: %s", aggreg_type);
            //System.out.printf("will 
            Iterator<Pair<Integer,Double> > it = data;
            while (it.hasNext())
            {
                Pair<Integer,Double> p = it.next();
                int zoomed_pos = zcc.getConvertedCoordinate(p.getFirst());
                //log.info("next pos: %s zoomed: %d", p, zoomed_pos);
                aggr.addData(zoomed_pos, p.getSecond());
            }

            aggr.compile();

            aggregated = aggr.getData();
        }

        public double getValue(int start, int end)
        {
            int zpp1 = zcc.getConvertedCoordinate(start);
            int zpp2 = zcc.getConvertedCoordinate(end);
            if (zpp1 != zpp2)
            {
                throw new FRuntimeException("invalid coord request (%d,%d) map to : (%d, %d)", start, end, zpp1, zpp2);
            }
                
            return aggregated[zpp1];
            
        }
    }

    public DataPlot(DataPlot parent, ZoomedCoordinateConverter zcc, ZoomedDataProvider zdp)
    {
        super(zcc.getZoomedLength());
        name=parent.name;
        header=parent.header;
        userObject = parent.userObject;
        color = parent.color;

        //get the converted cos
        HashSet<Integer> zp = new HashSet<>();
        Pair<int[],int[]> zoomed = zcc.getZoomMappings();
        final int[] starts = zoomed.getFirst();
        final int[] ends = zoomed.getSecond();

        double[] data = new double[zcc.getZoomedLength()];

        
        for (int i=0; i<zcc.getZoomedLength(); i++)
        {
            data[i] = zdp.getValue(starts[i], ends[i]);
        }
        setDenseData(data);
    }


    public DataPlot(DataPlot parent, ZoomedCoordinateConverter zcc, DataAggregator.TYPE aggreg_type)
    {
        super(zcc.getZoomedLength());
        name=parent.name;
        header=parent.header;
        userObject = parent.userObject;
        color = parent.color;
        

        DataAggregator aggr = new DataAggregator(aggreg_type, getDataLength());

        
        //System.out.printf("will 
        Iterator<Pair<Integer,Double> > it = parent.getIterator();
        int lastpos = -1;
        while (it.hasNext())
        {
            Pair<Integer,Double> p = it.next();
            int zoomed_pos = zcc.getConvertedCoordinate(p.getFirst());
            for (int i= lastpos + 1; i <= zoomed_pos; i++)
            {
                aggr.addData(i, p.getSecond());
            }
            lastpos = zoomed_pos;
            
            
        }
        
        aggr.compile();
        
        setDenseData(aggr.getData());

    }

    
    public void setHeader(PlotYAxisImage h)
    {
        header=h;
    }

    public PlotYAxisImage getHeader()
    {
        return header;
    }
    public void setObject(Object o)
    {
        this.userObject=o;
    }
    public void setName(String name)
    {
        this.name=name;
    }
    public String getName()
    {
        return name;
    }
    public void setColor(Color c)
    {
        this.color=c;
    }
    

    public Element getSerializedXML()
    {
        Element root = new Element(ELEM_NAME);
        XMLUtils.writeColorAttrib(root, color);
        root.setAttribute("name", name);
        root.setAttribute("length", ""+getDataLength());


        double densed[] = new double[getDataLength()];

        Iterator<Pair<Integer,Double>> it = getIterator();
        while (it.hasNext())
        {
            Pair<Integer,Double> p = it.next();
            densed[p.getFirst()]=p.getSecond();
        }
        root.addContent(XMLUtils.writeDoubleArray(densed));
        

        return root;
    }
    
    boolean data_sorted=false;
    
    public void plot(PlotYAxisImage header, DGraphics g2, int x_start, int x_end)
    {
        
        //img = ImageUtils.getWhiteImage(data_length, header.getHeight());
        
        //System.out.println("will plot with color: "+color);
        g2.setColor(color);

        //System.out.println("plotting with "+g2);
        if (data_full!=null)
        {
            log.debug("will draw full data for %s", name);
            header.drawCurve(g2, 0, 1, CollectionIterator.getIterator(data_full), x_start, x_end);
        }

        for(Pair<Integer, double[]> rd: data_arrays)
        {
            header.drawCurve(g2, rd.getFirst(), 1, CollectionIterator.getIterator(rd.getSecond()), x_start, x_end);
        }
        if (!data_sorted)
        {
            Collections.sort(data);
            data_sorted=true;
        }
        

        
        header.drawCurve(g2, data.iterator(), x_start, x_end);
    }

    public boolean isDense()
    {
        return data_full != null;
    }

    public Vector<Integer> getDenseDataY()
    {
        return getDenseDataY(0, getDataLength());
    }

    public Vector<Integer> getDenseDataY(int x_start, int x_end)
    {
        if (data_full ==null)
            throw new FRuntimeException("can't get fetch dense data if no dense data there...");

        return header.getDenseLineData(CollectionIterator.getIterator(data_full), x_start, x_end);
    }

    public Vector<Pair<Vector<Integer>, Vector<Integer>>> getLineData()
    {
        return getLineData(0, getDataLength());
    }

    public Vector<Pair<Vector<Integer>, Vector<Integer>>> getLineData(int x_start, int x_end)
    {
        if (data_full != null)
        {
            Vector<Integer> x = ObjectGetter.toVector(IteratorUtils.getIntIterator(x_start, x_end));
            return ObjectGetter.toVector(Pair.create(x, getDenseDataY(x_start, x_end)));
        }


        Vector<Pair<Vector<Integer>, Vector<Integer>>> rv = new Vector<>();
        
        for(Pair<Integer, double[]> rd: data_arrays)
        {
            rv.add(header.getLineData(rd.getFirst(), 1, CollectionIterator.getIterator(rd.getSecond()), x_start, x_end));
        }
        return rv;
    }


    public void addDataPoint(int x, double d)
    {
        data_sorted=false;
        super.addDataPoint(x,d);
    }
    public Pair<Double,Double> getDataRange()
    {
        
        
        Iterator<Pair<Integer,Double>> it = getIterator();
        if(!it.hasNext())
            return null;

        double maxvalue=it.next().getSecond();
        double minvalue=maxvalue;
        
        while (it.hasNext())
        {
            Pair<Integer,Double> p = it.next();
            maxvalue=Math.max(maxvalue,p.getSecond());
            minvalue=Math.min(minvalue,p.getSecond());
        }
        log.debug("data range: %2.4f-%2.4f", minvalue, maxvalue);
        return Pair.create(minvalue, maxvalue);

    }
}

