package lmu.utils.plotting;

import lmu.utils.*;
import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.color.*;

       
public class PlotYAxisImage
{
    
    public interface PlotConfigChangeListener
    {
        public void configChanged();
    }

    void addListener(PlotConfigChangeListener l)
    {
        listeners.add(l);
    }

    void removeListener(PlotConfigChangeListener l)
    {
        listeners.remove(l);
    }

    
    Vector<PlotConfigChangeListener> listeners = new Vector<PlotConfigChangeListener>();
    Logger logger = LogConfig.getLogger(); 
    boolean log;
    double logbase=2.0;
    double minvalue = 0;
    double maxvalue;
    

    int minvalue_y=-1;
    int maxvalue_y=-1;
    int value_range_y=-1;
    double range=0;
    double logrange=0;
    public int height=300;
    int char_height=0;
    String format="%2.2f";
    Vector<String> labels = null;
    Font f;
    
    BufferedImage image;

    public PlotYAxisImage()
    {
        this(100, 0,100, false);
    }
    public PlotYAxisImage(double minvalue, double maxvalue)
    {
        this(minvalue, maxvalue, false);
    }
    public PlotYAxisImage(double minvalue, double maxvalue, boolean log)
    {
        this(300, minvalue, maxvalue, log);
    }
    public PlotYAxisImage(int height, double minvalue, double maxvalue, boolean log)
    {
        this.height=height;
        this.log=log;
        this.minvalue=minvalue; 
        this.maxvalue=maxvalue;

        BufferedImage bim = new BufferedImage(10, 10, BufferedImage.TYPE_3BYTE_BGR);

        Graphics2D g2 = bim.createGraphics();

        this.f = g2.getFont();
    }

    public int getWidth()
    {
        return getImage().getWidth();
    }
    public void setHeight(int height)
    {
        this.height=height;
        image=null;
        notifyListeners();
    }
    public int getHeight()
    {
        return height+30;
    }
    void notifyListeners()
    {
           
        for(PlotConfigChangeListener l: listeners)
        {
            l.configChanged();
        }
    }

    
    public double getMinValue()
    {
        return minvalue;
    }

    public double getMaxValue()
    {
        return maxvalue;
    }
    public void setMaxValue(double maxv)
    {
        maxvalue=maxv;
        image=null;
        notifyListeners();
    }

    public void setMinValue(double minv)
    {
        minvalue=minv;
        image=null;
        notifyListeners();
    }

    public boolean isLogScale()
    {
        return log;
    }
    public void setLogScale(boolean b)
    {
        if (log==b)
            return;
        
        log=b;
        image=null;
        notifyListeners();
    }
    public void setLogbase(double logbase)
    {
        this.logbase=logbase;
        labels=null;
        notifyListeners();
     
    }
    public void setFormat(String format)
    {
        this.format=format;
        labels=null;
    }
    public void setFont(Font f)
    {
        this.f=f;
        image=null;
    }

    public BufferedImage getImage()
    {
        return getImage(false);
    }
    public BufferedImage getImage(boolean force)
    {
        if (force || (image==null || labels==null))
        {
            calcImage();
        }
        return image;
    }

    void calcImage()
    {

        int char_h = ImageUtils.getDrawnTextDimension(f, -1, "test").getSecond();

        char_height = 2+char_h;
        
        int n_labels=(int)(((height)/(double)char_height));

        
        logger.debug("plotyaxis image charh: %d nlabels: %d val range: %2.3f-%2.3f", char_h, n_labels, minvalue, maxvalue);
        range = maxvalue - minvalue;
        labels = new Vector<String>();
        double logstep = 0;
        double step=(range)/(double)(n_labels-1);
        
        if (log)
        {
            logrange = Math.log(maxvalue-minvalue);
            logrange = logrange/Math.log(logbase);
            logstep = logrange/(n_labels-1);

        }
        //System.out.printf("step: %2.3f logstep: %2.3f max relog: %2.3f\n",step, logstep, Math.pow(logbase,  logstep*n_labels));
        
        double v=minvalue;
        int x=0;//char_height;
        final double h = height;
        int char_half=(int)(0.5*char_h);
        
        for (int i=0; i<n_labels; i++)
        {
            
            //String l = String.format(format, getClicked((int)(h-x), h));
            String l = String.format(Locale.UK,format , v);
            labels.add(l);
            
            if (log)
            {       
                v=minvalue+Math.pow(logbase, (i+1)*logstep);
            }
            else
            {
                v+=step;
            }
            
            x+= char_height;

        }
        int longest_label=ImageUtils.getDrawnTextDimension(f, -1, labels).getFirst();
        
        BufferedImage bim = ImageUtils.getWhiteImage(longest_label, height+20);
        Graphics2D g2 = bim.createGraphics();
        g2.setColor(Color.BLACK);

        x=(int)(char_height);
        int corr=-char_half;
        for (int i=labels.size()-1; i>=0; i--)
        {
            if (i==labels.size()-1)
            {
                maxvalue_y=x+corr;
            }
            if (i==0)
            {
                minvalue_y=x+corr;
            }
            String l = labels.get(i);
            int y=0;
            g2.drawString(l, y, x);
            
            //System.out.printf("draw at %d, %d: %s\n",x,y,l);
            x+= char_height;
        }
        g2.dispose();
        
        value_range_y=minvalue_y-maxvalue_y;
        image=bim;
        /*
        System.out.printf("test:  value at 0: %2.2f at %d: %2.2f middle: %2.2f: %2.2f\n", 
                          getClicked(0), 
                          image.getHeight(), getClicked(image.getHeight()), 
                          image.getHeight()*0.5, 
                          getClicked((int)(image.getHeight()*0.5))
                          );
          */
        
    }

    public int getY(double value)
    {
        if(value > maxvalue) return maxvalue_y;
        if (value <= minvalue) return minvalue_y;

        double d= value-minvalue;
        if (!log)
        {
            return minvalue_y-(int)(value_range_y*(d/range));
        }
        d = NumUtils.logN(d, logbase);
        

        return minvalue_y-(int)(value_range_y*(d/logrange));

    }

    public void drawCurve(DGraphics g2,  Iterator<Pair<Integer, Double>> values, Integer x_start, Integer x_end)
    {
        drawCurve(g2,1, values, x_start, x_end);
    }

    public void drawCurve(DGraphics g2,  Iterator<Pair<Integer, Double>> values)
    {
        drawCurve(g2,values, null, null);
    }
    
    public void drawCurve(DGraphics g2,  int xstep, Iterator<Pair<Integer, Double>> values)
    {
        drawCurve(g2, xstep,values, null, null);
    }

    public void drawCurve(DGraphics g2,  int xstep, Iterator<Pair<Integer, Double>> values, Integer x_start, Integer x_end)
    {
        calcCurve(xstep, values, x_start, x_end, (int lastx, int lasty, int nextx, int nexty) -> g2.drawLine(lastx, lasty, nextx, nexty));
    }

    static class LineHolder implements CurveUpdater
    {
        Vector<Integer> x = new Vector<>();
        Vector<Integer> y = new Vector<>();
        
        boolean dense;

        public LineHolder(boolean dense)
        {
            this.dense = dense;
        }

        public void update(int lastx, int lasty, int nextx, int nexty)
        {
            if (!dense || x.size() == 0)
            {
                x.add(lastx); 
                y.add(lasty); 
            }
            

            x.add(nextx);
            y.add(nexty);

        }

        public Pair<Vector<Integer>,Vector<Integer>> getData()
        {
            return Pair.create(x, y);
        }
    }

    public Vector<Integer> getDenseLineData(Iterator<Double> values, Integer x_start, Integer x_end)
    {
        LineHolder lh = new LineHolder(true);
        calcCurve(0, 1, values , x_start, x_end, lh);
        //logger.info("dense x: %d y: %d", lh.x.size(), lh.y.size());
        return lh.getData().getSecond();
    }

    public Pair<Vector<Integer>, Vector<Integer>> getLineData(int xstep, Iterator<Pair<Integer, Double>> values, Integer x_start, Integer x_end)
    {
        LineHolder lh = new LineHolder(false);
        calcCurve(xstep, values, x_start, x_end, lh);
        return lh.getData();
    }

    public void calcCurve(int xstep, Iterator<Pair<Integer, Double>> values, Integer x_start, Integer x_end, CurveUpdater cu)
    {
        Integer lastY=null;
        
        Integer lastX=null;
        while (values.hasNext())
        {
            Pair<Integer,Double> v = values.next();
            int x = (lastX==null) ? v.getFirst() : lastX+(v.getFirst()-lastX)*xstep;

            
            int nY=getY(v.getSecond());
            
            
            if (lastY!=null)
            {
                //System.out.printf("will draw (iterator): %d,%d -> %d, %d\n", lastX, lastY, x,nY);
                if (x_start==null || (lastX>=x_start && lastX<=x_end) || (x>=x_start && x<=x_end))
                {
                    cu.update(lastX, lastY, x,nY);
                    
                }
                
            }
            lastX=x;
            lastY=nY;
        }
        }

    public void drawCurve(DGraphics g2, int startx, int xstep, Iterator<Double> values)
    {
        drawCurve(g2,startx,xstep,values,null,null);
    }
    

    interface CurveUpdater
    {
        public void update(int lastx, int lasty, int nextx, int nexty);
        
    }
    

    void calcCurve(int startx, int xstep, Iterator<Double> values, Integer x_start, Integer x_end, CurveUpdater cu)
    {
        
        //logger.info("calc curve start");
        //System.out.println("will draw iterator values with "+g2);
        int x=startx;
        Integer lastY=null;
        
        while (values.hasNext())
        {
            double n = values.next();
            int nY=getY(n);
            //logger.info("drawCurve: value: %2.2f x: %d y: %d", n, x, nY);
            //System.out.println("g2: "+g2);
            if (lastY!=null)
            {
                if (x_start==null || (x>=x_start && x<=x_end) || (x+xstep>=x_start && x+xstep<=x_end))
                {
                    cu.update(x, lastY, x+xstep, nY);
                }
                
            }
            x+=xstep;
            lastY=nY;
        }
    }
    
    public Pair<Vector<Integer>, Vector<Integer>> getLineData(int startx, int xstep, Iterator<Double> values, Integer x_start, Integer x_end)
    {
        LineHolder lh = new LineHolder(true);
        calcCurve(startx, xstep, values, x_start, x_end, lh);
        return lh.getData();
    }

    public void drawCurve(DGraphics g2, int startx, int xstep, Iterator<Double> values, Integer x_start, Integer x_end)
    {
        calcCurve(startx, xstep, values, x_start, x_end, (int lastx, int lasty, int nextx, int nexty) -> g2.drawLine(lastx, lasty, nextx, nexty));
    
    }

    public double getClicked(int y)
    {
        //System.out.printf("y: %d minvalue_y: %d: %2.3f maxvvalue: %d: %2.3f\n", y, minvalue_y, maxvalue, maxvalue_y, minvalue);
        
        if (y>=minvalue_y)
        {
            return minvalue;
        }
        if (y<=maxvalue_y)
        {
            return maxvalue;
        }
        int cy=minvalue_y-y;
        double diff=(cy)/((double)value_range_y);
        //System.out.printf("diff: %2.3f y: %d value rane: %d\n", diff, y, value_range_y);
        if (!log)
        {
            return minvalue+(range*diff);
        }
        
        return minvalue+Math.pow(logbase, logrange*diff);
    }
}

