package lmu.utils.plotting;


import lmu.utils.*;
import lmu.utils.swing.*;
import java.awt.image.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import org.jdom.*;


public class RegionPicturePane extends HeaderedPicturePane  implements java.awt.event.MouseListener, java.awt.event.MouseMotionListener,
                                                SliceImageDrawer
{
    Logger log = LogConfig.getLogger();
    int FONT_SIZE = 12;
    int REGION_ROW_WIDTH = 20;
    int REGION_Y_SPACE = 2;

    public final static String ELEM_NAME="regionpane";
    Object userObject;
    int width;
    Vector<RegionPlot> region_plots = new Vector<RegionPlot>();


    public Vector<RegionPlot> getRegionPlots()
    {
        return region_plots;
    }
    JMenuBar menubar = null;
    JMenu visibility = new JMenu("show/hide");
    Vector<JMenuItem> visible_items = new Vector<JMenuItem>();

    void showAt(MouseEvent e)
    {
        int x=e.getX();
        int y=e.getY();
    
        
        Region1D r = null;

        for (RegionPlot rp : region_plots)
        {
            if(null==(r=rp.getRegion(x,y)))
                continue;

            break;
        }
    
        if(r==null || null==r.getObject())
        {
            data_pane.getPanel().setToolTipText("");
            return;
        }
            

        //double value = header.getClicked(y);
        //System.out.println("x: "+x+"y: "+y+": "+value);
        data_pane.getPanel().setToolTipText(""+r.getObject());
        //data_pane.getPanel()
    }
    public void mouseDragged(MouseEvent e)
    {
    }
    
    public void mouseMoved(MouseEvent e)
    {
        showAt(e);
    }
    
    public void mouseClicked(MouseEvent e)
    {
        showAt(e);
        notifyListeners(e);
    }
    
    /**
     * Invoked when a mouse button has been pressed on a component.
     */
    public void mousePressed(MouseEvent e)
    {
    }
    
    /**
     * Invoked when a mouse button has been released on a component.
     */
    public void mouseReleased(MouseEvent e)
    {
    }
    
    /**
     * Invoked when the mouse enters a component.
     */
    public void mouseEntered(MouseEvent e)
    {
        showAt(e);
    
    
    }
    
    /**
     * Invoked when the mouse exits a component.
     */
    public void mouseExited(MouseEvent e)
    {
    }

    public void setVisible(int dpi)
    {
    
        if (dpi<0)
        {
            for (int i=0; i<region_plots.size(); i++)
            {
                region_plots.get(i).setVisible(true);
                visible_items.get(i).setText("hide"+" "+region_plots.get(i).getName());
            }
    
        }
        else
        {
    
    
            boolean nv = !region_plots.get(dpi).isVisible();
    
            region_plots.get(dpi).setVisible(nv);

            
    
            visible_items.get(dpi).setText(((nv)?"hide":"show")+" "+region_plots.get(dpi).getName());
        }
        int divloc=getDividerLocation();
        update();
        setDividerLocation(divloc);
    }
    public void updateMenu()
    {
        
        visibility.removeAll();
        

        JMenuItem jmi = new JMenuItem("show all");
        jmi.addActionListener(new ReflectedObjectListener(new ReflectedObject("setVisible", -1), this));
        visibility.add(jmi);
        for (int i=0; i<region_plots.size(); i++)
        {
            RegionPlot rp = region_plots.get(i);
            jmi = new JMenuItem("hide "+rp.getName());
            jmi.addActionListener(new ReflectedObjectListener(new ReflectedObject("setVisible", i), this));
            visibility.add(jmi);
            visible_items.add(jmi);
    
        }
    }
    public void setMenuBar(JMenuBar menubar)
    {
        
        this.menubar=menubar;
    
        if (visibility!=null)
        {
            menubar.add(visibility);
            return;
        }
        
    }

    public RegionPicturePane(String name, int width , Object o)
    {
        super(name,o);
        this.width=width;
    }

    /** constructor to initialize from serialized object */
    public RegionPicturePane(Element root)
    {
        super("");
        
        width=XMLUtils.getIntAttrib(root, "width", null);
        for (Element e : XMLUtils.getChildren(root, RegionPlot.ELEM_NAME))
        {
            region_plots.add(new RegionPlot(width,e));
        }
        //System.out.println("got region plot: "+getName()+" plots: "+region_plots.size());
        update();
    }

    public int getDataWidth()
    {
        return width;
    }
    public Element getSerializedXML()
    {
        Element root = new Element(ELEM_NAME);
        root.setAttribute("width", ""+width);
        for (RegionPlot rp : region_plots)
        {
            if (!rp.visible)
            {
                continue;
            }
            root.addContent(rp.getSerializedXML());
        }
        return root;
        
    }

    HashMap<RegionPlot, RegionPlot> parent_mapping = new HashMap<>();


    public RegionPicturePane(RegionPicturePane parent, ZoomedCoordinateConverter zcc)
    {
        super(parent.name, parent.userObject);
        width= zcc.getZoomedLength();
        for (RegionPlot rp: parent.region_plots)
        {
            RegionPlot converted = new RegionPlot(rp, zcc);
            region_plots.add(converted);
            parent_mapping.put(converted, rp);
        }
    }


    public Pair<RegionPlot, Region1D> getClicked(int x, int y, boolean original)
    {
        for (RegionPlot rp : region_plots)
        {
            if (rp.y_start > y || rp.y_end < y)
                continue;
            
            
            RegionPlot orp = parent_mapping.get(rp);
            RegionPlot rvp = (!original || orp == null) ? rp : orp;

            Region1D reg = rp.getRegion(x, y, original);
            
            return Pair.create(rvp, reg);
        }
        return null;
    }

    Vector<RegionPlot.RegionImageDrawer> region_drawers = new Vector<RegionPlot.RegionImageDrawer>();

        
    int annot_height=0;

    public int getImageWidth()
    {
        return getDataWidth();
    }
    public int getImageHeight()
    {
        return annot_height;
    }
    
    AwtGraphics awt_helper = new AwtGraphics(null);

    static class Connection
    {
        RegionPlot src;
        RegionPlot dst;
        int x1;
        int x2;
        Color c;
        Stroke stroke;
        
        public Connection(RegionPlot src, int src_x, RegionPlot dst, int dst_x, Color c, Stroke stroke)
        {
            this.src = src; x1 = src_x;
            this.dst = dst; x2 = dst_x;
            this.c = c;
            this.stroke = stroke;
        }
    }
    Vector<Connection> connections = new Vector<Connection>();

    public void connect(RegionPlot rp_src, int x, RegionPlot rp_dst, int x2, Color c)
    {
        connections.add(new Connection(rp_src, x, rp_dst, x2, c,null));
    }

    public void connect(RegionPlot rp_src, int x, RegionPlot rp_dst, int x2, Color c, Stroke stroke)
    {
        connections.add(new Connection(rp_src, x, rp_dst, x2, c,stroke));
    }


    public void paintSlice(DGraphics g2, int x1, int y1, int x2, int y2)
    {
        log.trace("got %d region drawers and %d connections", region_drawers.size(), connections.size());

        for (RegionPlot.RegionImageDrawer rimd : region_drawers)
        {
            if (rimd.y_end<y1 || rimd.y_start>y2)
                continue;
            
            log.debug("will paint next slice for  %d-%d", x1, x2);
            rimd.paintSlice(g2,x1,x2);
        }
        for (Connection c : connections)
        {
            RegionPlot.RegionImageDrawer src = null;
            RegionPlot.RegionImageDrawer dst = null;
            for (RegionPlot.RegionImageDrawer rimd : region_drawers)
            {
                if (c.src == rimd.getPlot())
                {
                    src = rimd;
                }
                if (c.dst == rimd.getPlot())
                {
                    dst = rimd;
                }                                                                                                                                             
                if (src!=null && dst!=null)
                    break;
                
                
            }
            if (src==null || dst == null)
            {
                continue;
            }
            boolean srcup = src.y_start < dst.y_start;
            RegionPlot.RegionImageDrawer r1 = (srcup)? src: dst;
            RegionPlot.RegionImageDrawer r2 = (srcup)? dst: src;

            int r1x = (srcup) ? c.x1 : c.x2;
            int r2x = (srcup) ? c.x2 : c.x1;
            
            g2.setColor(c.c);
            
            if ( c.stroke != null)
                g2.setStroke(c.stroke);
            
            g2.drawLine(r1x, r1.y_end, //+r1.getPlot().Y_SPACER,
                         r2x, r2.y_start+r2.getPlot().Y_SPACER);
        }
    }
    public void paintSlice(java.awt.Graphics g, int x1, int y1, int x2, int y2, java.awt.image.ImageObserver obs)
    {
        awt_helper.g2 = (Graphics2D)g;
        paintSlice(awt_helper,x1,y1,x2,y2);
        
    }

    public BufferedImage drawHeader()
    {
        int height = 0;
        int label_width=0;


        region_drawers.clear();

        
        //Vector<Pair<BufferedImage,BufferedImage> > region_images = new Vector<Pair<BufferedImage,BufferedImage> >();

        Vector<BufferedImage> header_images = new Vector<BufferedImage>();

        for (RegionPlot rp : region_plots)
        {
            if (!rp.isVisible())
                continue;
            
            RegionPlot.RegionImageDrawer  rimd = rp.getDrawer();

            region_drawers.add(rimd);

            //System.out.printf("rp: %s height: %d\n ", rp.getName(), height);
            rp.setYRegion(height, height+rimd.getHeight());
            //Pair<BufferedImage,BufferedImage> rp_im = rp.getImages();

            //rp.setYRegion(height, height+rimd.getHeight());

            //height += rp_im.getFirst().getHeight();

            rimd.y_start =height;
            
            rimd.y_end= height += rimd.getHeight();
            
            //System.out.printf(" rp : %s y start> %d end> %s\n", rp.getName(), rp.y_start, rp.y_end);

            BufferedImage rp_im = rimd.getLabelImage();

            label_width = Math.max(label_width, rp_im.getWidth());

            header_images.add(rp_im);
            //region_images.add(rp_im);
        }
      //  System.out.printf("got %d region plots height: %d\n", region_plots.size(), height);
        if (label_width==0)
        {
            label_width=10;
            height=10;
        }
        annot_height=height+50;
        BufferedImage label_im = ImageUtils.getWhiteImage(label_width,annot_height);
        //BufferedImage annot_im = ImageUtils.getWhiteImage(width,height);

        Graphics2D label_g2 = label_im.createGraphics();
        //Graphics2D annot_g2 = annot_im.createGraphics();
        int y=0;
        for(BufferedImage l_im : header_images)
        {
            label_g2.drawImage(l_im, null, 0, y);
            y+=l_im.getHeight();
        }
        /*for (Pair<BufferedImage, BufferedImage> rp_im  : region_images)
        {
            BufferedImage l_im = rp_im.getFirst();
            BufferedImage a_im = rp_im.getSecond();
                        
            label_g2.drawImage(l_im, null, 0, y);
            annot_g2.drawImage(a_im, null, 0, y);

            y+=l_im.getHeight();
        }
        */

        label_g2.dispose();

        return label_im;
    }

    public void update()
    {
        
        //annot_g2.dispose();
        BufferedImage label_im = drawHeader();
        updateMenu();
        
        setSlicers(ImageScrollPane.getSlicer(label_im), this);
        //setImages(label_im, annot_im);
        //setDividerLocation(label_im.getWidth());
        data_pane.getPanel().addMouseListener(this);
        data_pane.getPanel().addMouseMotionListener(this);
    }
    public RegionPlot addRegion(String name)
    {
        return addRegion(name,null);
    }
    /*public RegionPlot addRegion(String name, Object object)
    {
        return addRegion(name,null);
    }
    */
    public RegionPlot addRegionPlot(RegionPlot rp)
    {
        region_plots.add(rp);
        rp.region_ywidth = REGION_ROW_WIDTH;
        rp.Y_SPACER = REGION_Y_SPACE;
        rp.setFontSize(FONT_SIZE);
        return rp;
    }

    public void changeRegionSettings(int font_size, int region_width, int region_spacer)
    {
        REGION_ROW_WIDTH = region_width;
        REGION_Y_SPACE = region_spacer;
        FONT_SIZE = font_size;

        for (RegionPlot rp : region_plots)
        {
            rp.region_ywidth = REGION_ROW_WIDTH;
            rp.Y_SPACER = REGION_Y_SPACE;
            rp.setFontSize(font_size);
        }
        update();
    }

    public RegionPlot addRegion(String name, Object object, Region1D ... regions)
    {
      
        return addRegion(name,object, new CollectionIterator(regions));
    }
    public RegionPlot addRegion(String name, Object object, Vector<Region1D> regions)
    {
        return addRegion(name,object, regions.iterator());
    }
    public RegionPlot addRegion(String name, Object object, Iterator<Region1D> regions)
    {
        RegionPlot rp = new RegionPlot();
        rp.width=width;
        rp.name=name;
        rp.userObject=object;
        if (regions!=null)
        {
            while (regions.hasNext())
            {
                rp.addRegion(regions.next());
            }
        }
        
        
        return addRegionPlot(rp);
        
    }
}


