package lmu.utils.plotting;


import lmu.utils.*;
import java.beans.*;
import lmu.utils.ImageUtils.ImagePanel;
import java.awt.image.*;
import java.util.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.event.*;
import org.jdom.*;

public class RegionPlot 
{
    public int Y_SPACER = 2;
    public final static String ELEM_NAME="regionplot";
    public final static String DECORATOR_NAME="decorator";
    public String name;
    public Object userObject;
    public boolean visible=true;
    public int width=-1;
    Font f=null;
    int font_size=-1;

    public int y_start=0;
    public int y_end=0;

    public void setYRegion(int start, int end)
    {
        y_start=start; y_end=end;
    }
    
    int min_space_between_regions=2;
    public String getName()
    {
        return name;
    }

    public boolean isVisible()
    {
        return visible;
    }

    public void setVisible(boolean b)
    {
        visible=b;
    }
    public void  setMinSeparatorSpace(int minsep_space)
    {
        min_space_between_regions=minsep_space;
    }
    public static class RegionDecorator
    {
        public boolean filled=true;
        public boolean connect = true;
        public Color color = Color.BLACK;
        public String label="";

        public RegionDecorator()
        {

        }
        public RegionDecorator(Color c, boolean filled)
        {
            this(c,filled,true);
        }
        public RegionDecorator(Color c, boolean filled, boolean connect)
        {
            color=c; this.filled=filled;this.connect = connect;
        }
        public RegionDecorator(Element e)
        {
            label=XMLUtils.getAttrib(e, "label", "", false);
            color = XMLUtils.getColor(e, null, true);
            filled = XMLUtils.getBooleanAttrib(e, "filled", true);
        }
        public Element getXML(String name)
        {
            Element e = new Element(name);
            e.setAttribute("filled", ""+filled);
            XMLUtils.writeColorAttrib(e, color);
            e.setAttribute("label", label);
            return e;
        }

        //public Rectangle drawRegions(Graphics2D g2, int y_shift, Vector<Rectangle> multi_regions)
        public Rectangle drawRegions(DGraphics g2, int y_shift, Vector<Rectangle> multi_regions)
        {
            if (multi_regions.size() == 0)
                return null;
            
            
            //System.out.println("will draw regions: "+multi_regions);
            final int N=multi_regions.size()-1;
            int x=(int)multi_regions.get(0).getX();
            int y=y_shift+(int)multi_regions.get(0).getY();
            int height=(int)multi_regions.get(0).getHeight();

            int x2=(int)multi_regions.get(N).getX();
            

            int width= x2 - x;
            
            for (Rectangle box : multi_regions)
            {
                draw(g2, y_shift, box);
            }

            //connect the boxes
            int middle_h=(int)(y+(0.5*height));

            for (int i=1; i<multi_regions.size(); i++)
            {
                Rectangle prev = multi_regions.get(i-1);
                Rectangle next = multi_regions.get(i);

                int start=(int)(prev.getX()+prev.getWidth());
                int end = (int)next.getX();

                if ( connect) 
                    g2.drawLine(start, middle_h, end, middle_h);
            }

            return new Rectangle(x,y,width,height);
        }

        public Rectangle draw(DGraphics g, int y_shift, Rectangle box)
        {
            int x=(int)box.getX();
            int y=y_shift+(int)box.getY();
            int height=(int)box.getHeight();
            int width=(int)box.getWidth();
            g.setColor(color);
            //System.out.printf("%s: will draw filled: %s (%s) (%d,%d,%d,%d)\n", name, g.getColor(), filled, x1,x2,y1,y2);
            if (filled)
            {
                g.fillRect(x,y,width,height);
            }
            else
            {
                g.drawRect(x,y,width,height);
            }
            return box;
        }
    }

    
    RegionDecorator default_decorator = new RegionDecorator();
    //default decoration
    

    Vector<Region1D> regions = new Vector<Region1D>();
    Vector<RegionDecorator> decorators  = new Vector<RegionDecorator>(); //if a region has a as decorator, the default decorator used

    Vector<Rectangle> drawnPositions = new Vector<Rectangle>();

    int region_ywidth=20;

    public Element getSerializedXML()
    {
        Element e = new Element(ELEM_NAME);
        e.setAttribute("name", name);
        e.addContent(default_decorator.getXML(DECORATOR_NAME));
        
        for (int i=0; i<regions.size(); i++)
        {
            Element rege = new Element("r");
            Region1D r = regions.get(i);
            rege.setAttribute("x1", ""+r.getX1());
            rege.setAttribute("x2", ""+r.getX2());

            RegionDecorator rd = decorators.get(i);
            if (rd!=null)
            {
                rege.addContent(rd.getXML(DECORATOR_NAME));
            }
            
            
            e.addContent(rege);
        }
        return e;
    }
    public RegionPlot(int width, Element root)
    {
        this.width=width;
        name = XMLUtils.getAttrib(root, "name", null, true);
        default_decorator = new RegionDecorator(root.getChild(DECORATOR_NAME));

        for(Element e : XMLUtils.getChildren(root, "r"))
        {
            int x1 = XMLUtils.getIntAttrib(e, "x1", null);
            int x2 = XMLUtils.getIntAttrib(e, "x2", null);

            Element dece = e.getChild(DECORATOR_NAME);
            RegionDecorator dec = null;
            if (dece!=null)
            {
                dec = new RegionDecorator(dece);
            }
            addRegion(new Region1D(x1,x2, null), dec);
            

        }
    }
    public RegionPlot()
    {

    }

    HashMap<Region1D, Region1D> converted_regions = new HashMap<>();

    public RegionPlot(RegionPlot parent, ZoomedCoordinateConverter zcc)
    {
        name=parent.name;
        userObject=parent.userObject;
        visible=parent.visible;
        f=parent.f;
        font_size=parent.font_size;
        min_space_between_regions=parent.min_space_between_regions;
        default_decorator=parent.default_decorator;
        //decorators = parent.decorators;

        //convert the regions
        width = zcc.getZoomedLength();
        for (int i=0; i<parent.regions.size(); i++)
        {
            Region1D r = parent.regions.get(i);
            if (r==null)
            {
                addRegion(null,null);
                continue;
            }

            if (r instanceof RegionVector)
            {
                RegionVector rv = (RegionVector)r;
                RegionVector zoomed = new RegionVector();
                zoomed.setObject(rv.getObject());

                for (Region1D rv_r : rv.getRegions())
                {
                    zoomed.addRegion(zcc.getConvertedRegion(rv_r));
                }

                converted_regions.put(zoomed, rv);

                addRegion(zoomed, parent.decorators.get(i));
            }
            else
            {
                
                Region1D zoomed = zcc.getConvertedRegion(r);
                converted_regions.put(zoomed, r);
                addRegion(zoomed, parent.decorators.get(i));
            }
            
            
        }

    }
    public void setWidth(int width)
    {
        this.width=width;
    }

    public int getDataWidth()
    {
        return width;
    }
    public RegionDecorator getDefaultDecorator()
    {
        return default_decorator;
    }

    public RegionPlot setDefaultDecorator(RegionDecorator rd)
    {
        default_decorator = rd;
        return this;
    }

    int getRegionIndex(Region1D r , Object object)
    {
        for (int i=0; i<regions.size(); i++)
        {

            Region1D totest = regions.get(i);

            if ((object!=null && totest.getObject()==object) || (r!=null && r.getObject()!=null && totest.getObject()==r.getObject()))
            {
                return i;
            }
            
            if(r==null)
                continue;


            if (r.getX1()!=totest.getX1() || r.getX2()!=totest.getX2())
            {
                continue;
            }

            return i;
        }
        throw new RuntimeException("cannot find region : "+r+" to object: "+object+" in "+name);
        
    }
    /** returns the decorator associated to this region if any
     */
    public RegionDecorator getDecorator(Region1D r, Object object)
    {
        return decorators.get(getRegionIndex(r, object));
    }

    public RegionDecorator addDecorator(Region1D r, Object o)
    {
        int idx = getRegionIndex(r, o);
        RegionDecorator rd = decorators.get(idx);

        if (rd==null)
        {
            decorators.set(idx, rd = new RegionDecorator());
        }
        return rd;
    }
    
    public RegionDecorator addDecorator(Region1D r)
    {
        return addDecorator(r, r.getObject());
    }

    public RegionDecorator addDecorator(int x, int y)
    {
        return addDecorator(new Region1D(x,y,null),null);
    }


    public RegionDecorator addDecorator(Object o)
    {
        return addDecorator(null,o);
    }

    public RegionDecorator getDecorator(Object o)
    {
        return getDecorator(null, o);
    }

    public RegionDecorator getDecorator(Region1D r)
    {
        return getDecorator(r, r.getObject());
    }
    public RegionDecorator getDecorator(int x, int y)
    {
        return getDecorator(new Region1D(x,y,null), null);
    }
    public void addRegion(Region1D r)
    {
        addRegion(r,null);
    }

    public void addRegion(Region1D r, RegionDecorator rd)
    {
        //System.out.println(name+" adding region: "+r+" decorator: "+rd);
        if (r==null)
        {
            throw new RuntimeException("region should be not null for "+name+" invoker: "+ExceptionManager.logCallerStack());
        }
        regions.add(r);
        decorators.add(rd);
        drawnPositions.add(null);
    }



    public Region1D getRegion(int x, int y, boolean original)
    {
        Region1D r = getRegion(x,y);
        if (!original || r == null)
            return r;
        
        Region1D p = converted_regions.get(r);

        return (p == null) ? r : p;
    }

    public Region1D getRegion(int x, int y)
    {
        
        if (y>y_end || y<y_start)
        {
            return null;
        }
  
        y=y-y_start;
        for (int i=0; i<regions.size(); i++)
        {
            Rectangle r = drawnPositions.get(i);
            if (r==null || !r.contains(x,y))
                continue;
            
            return regions.get(i);
        }
        return null;
    }

    void drawRect(Graphics2D g, Rectangle r, boolean filled)
    {
        int x1=(int)r.getMinX();
        int x2=(int)r.getMaxX();
        x2-=x1;

        int y1=(int)r.getMinY();
        int y2=(int)r.getMaxY();
        y2-=y1;

        if (filled)
        {
            
            
            g.fillRect(x1,y1,x2,y2);;
        }
        else
        {
            g.drawRect(x1,y1,x2,y2);;
        }
    }


    public RegionImageDrawer getDrawer()
    {
        return new RegionImageDrawer();
    }
    public class RegionImageDrawer
    {

        public RegionPlot getPlot()
        {
            return RegionPlot.this;
        }
        BufferedImage label_im;
        Vector<Region1D> indexed_regions = new Vector<Region1D>();

        public int y_start=0;
        public int y_end=0;
        int height=0;
        
        public int getHeight()
        {
            return height;
        }
        public BufferedImage getLabelImage()
        {
            return label_im;
        }
        
        
        //public void paintSlice(java.awt.Graphics2D g, int x1, int x2)
        void paintSlice(DGraphics g, int x1, int x2)
        {
            for (int i=0; i<regions.size(); i++)
            {
                Rectangle r = drawnPositions.get(i);
                if (r==null)
                    continue;
                
                if (r.getMaxX()<x1 || r.getX()>x2)
                    continue;
                //nothing to draw

                RegionDecorator rd = decorators.get(i);
                if (rd==null)
                {
                    rd=default_decorator;
                }
                Vector<Rectangle> boxes = region2boxes.get(i);

                if (boxes.size()==1)
                {
                    rd.draw(g, y_start, boxes.get(0));
                }
                else
                {
                    rd.drawRegions(g, y_start, boxes);
                }
                
                
            }
        }

        

        Vector<Vector<Rectangle> > region2boxes = new Vector<Vector<Rectangle> >();

        
        public RegionImageDrawer()
        {
            
            for (int i=0; i<regions.size(); i++)
            {
                region2boxes.add(new Vector<Rectangle>());
                Region1D r = regions.get(i);
                indexed_regions.add(new Region1D(r.getX1(), r.getX2(), i));
            }
            Vector<Vector<Region1D> > rows = DenseRegionLayouter.getDenseLayout(min_space_between_regions, indexed_regions);

            int nrows=(rows.size()>0)?rows.size():1;
            height = region_ywidth*nrows;

            //calculate header picture: this is simply the name centered on it */
            Pair<Integer,Integer> label_dim = ImageUtils.getDrawnTextDimension(f, font_size, new SingleIterator<String>(name),Font.BOLD);


            //System.out.println(name+" :: LABEL DIM: "+label_dim);

            label_im = ImageUtils.getWhiteImage(label_dim.getFirst()+30, height);
            Graphics2D g2 = label_im.createGraphics();
            g2.setColor(Color.BLACK);
            int label_y_pos=(int)((height-label_dim.getSecond())*0.5);
            
            //g2.drawString(name, 0, label_y_pos);
            
            if (font_size>0)
            {
                Font font = g2.getFont().deriveFont((float)font_size);
                font = font.deriveFont(Font.BOLD);
                
                g2.setFont(font);

                
            }
            
            //System.out.println("WILL DRAW STRING : "+name+" size: "+font_size);
           // g2.setFont(new Font("monospaced", 12,Font.BOLD));
            
            g2.drawString(name, 0, label_y_pos+label_dim.getSecond());

            g2.dispose();

            //BufferedImage region_im = ImageUtils.getWhiteImage(width, height);

            //g2 = region_im.createGraphics();

            //draw regions
            //System.out.printf("compiling %d regions with %d regorators to %d rows\n", regions.size(), decorators.size(), rows.size());
            for (int r_i=0; r_i<rows.size(); r_i++)
            {
                int y_start = r_i*region_ywidth;

                Vector<Region1D> v = rows.get(r_i);
                for (Region1D r : v)
                {
                    int region_index=(Integer)r.getObject();
                    Vector<Rectangle> boxes = region2boxes.get(region_index);
                    Region1D orig = regions.get(region_index);

                    Rectangle drawn = new Rectangle(orig.getX1(), y_start+Y_SPACER, orig.getLength(),region_ywidth-Y_SPACER);
                                        
                    
                    if (orig instanceof RegionVector)
                    {
                        RegionVector rv = (RegionVector)orig;
                        
                        for (Region1D rr : rv.getRegions())
                        {
                            boxes.add(new Rectangle(rr.getX1(), y_start+Y_SPACER, rr.getLength(),region_ywidth-Y_SPACER));
                        }
                        
                    }
                    else
                    {
                        boxes.add(drawn);
                    }


                                        
                    drawnPositions.set(region_index, drawn);

                }
            }

        }
    }

    protected Rectangle getDrawn(Region1D r)
    {
        for (int i=0; i<regions.size(); i++)
        {
            if (r != regions.get(i))
                continue;
            
            
            return drawnPositions.get(i);
        }
         
        return null;
    }

    public void setFontSize(int fontsize)
    {
        font_size =fontsize;
    }
    public void drawImages(DGraphics header_g, DGraphics region_g)
    {
        
        Vector<Region1D> indexed_regions = new Vector<Region1D>();
        for (int i=0; i<regions.size(); i++)
        {
            Region1D r = regions.get(i);
            //System.out.println(name+" region: "+i+":"+r);
            indexed_regions.add(new Region1D(r.getX1(), r.getX2(), i));
        }
        Vector<Vector<Region1D> > rows = DenseRegionLayouter.getDenseLayout(min_space_between_regions, indexed_regions);

        int nrows=(rows.size()>0)?rows.size():1;
        int height = region_ywidth*nrows;

        //calculate header picture: this is simply the name centered on it */
        //Pair<Integer,Integer> label_dim = ImageUtils.getDrawnTextDimension(f, font_size, name);
        int label_height = header_g.getFontHeight();
        int label_width = header_g.getTextWidth(name);                          


        //System.out.println(name+" :: LABEL DIM: "+label_dim);
        header_g.setSize(label_width +30, height);

        //BufferedImage label_im = ImageUtils.getWhiteImage(label_dim.getFirst()+30, height);
        //Graphics2D g2 = label_im.createGraphics();
        header_g.setColor(Color.BLACK);
        int label_y_pos=(int)((height-label_height)*0.5);
        //g2.drawString(name, 0, label_y_pos);
        if (font_size>0)
        {
            header_g.setFont(null,font_size );
        }
        
      //  System.out.println("WILL DRAW STRING : "+name+" size: "+font_size);
        header_g.drawString(name, 0, label_y_pos+label_height);

        header_g.dispose();

        region_g.setSize(width, height);
        //BufferedImage region_im = ImageUtils.getWhiteImage(width, height);

        //g2 = region_im.createGraphics();
        DGraphics g2 = region_g;

        //draw regions
        //System.out.printf("compiling %d regions with %d regorators to %d rows\n", regions.size(), decorators.size(), rows.size());
        for (int r_i=0; r_i<rows.size(); r_i++)
        {
            int y_start = r_i*region_ywidth;

            Vector<Region1D> v = rows.get(r_i);
            for (Region1D r : v)
            {
                int region_index=(Integer)r.getObject();
                RegionDecorator rd = decorators.get(region_index);
                if (rd==null)
                {
                    rd=default_decorator;
                }

                Region1D orig = regions.get(region_index);
                Rectangle drawn = null;
                //System.out.println("will draw: "+orig);
                if (orig instanceof RegionVector)
                {
                    RegionVector rv = (RegionVector)orig;
                    Vector<Rectangle> boxes = new Vector<Rectangle>();
                    for (Region1D rr : rv.getRegions())
                    {
                        boxes.add(new Rectangle(rr.getX1(), y_start+2, rr.getLength(),region_ywidth-2));
                    }
                    drawn=rd.drawRegions(g2, 0, boxes);
                }
                else
                {
                    drawn = rd.draw(g2, 0, new Rectangle(orig.getX1(), y_start+2, orig.getLength(),region_ywidth-2));
                }

                drawnPositions.set(region_index, drawn);

            }
        }
        g2.dispose();


    }
    /** returns the header part and the annot part 
     * 
     * first it compiles the regions so that no annotations overlap
     * (and have some space between them)
     */
    public Pair<BufferedImage, BufferedImage> getImages()
    {
        ImageGraphics label_g = new ImageGraphics();
        ImageGraphics data_g = new ImageGraphics();

        drawImages(label_g,data_g);

        return new Pair<BufferedImage,BufferedImage>(label_g.getImage(), data_g.getImage());
        

            
    }
}

