package lmu.utils.plotting;

import lmu.utils.*;
import java.util.*;

public class ZoomedCoordinateConverter
{

    Logger log = LogConfig.getLogger();

    Vector<ZoomChangeListener> listeners = new Vector<ZoomChangeListener>();
    int length=-1;
    int zoomed_length=-1;
    int[] zoomed2_start = null;
    int[] zoomed2_end = null;
    

    ZoomSettings default_zoom = new ZoomSettings(this);

    HashMap<Object, ZoomSettings> category2zoom = new HashMap<Object, ZoomSettings>();
    
    Vector<Region1D> registered_regions = new Vector<Region1D>();
    Vector<Region1D> indexed_regions = new Vector<Region1D>();
    Vector<Object[]> region2_categories = new Vector<Object[]>();
    
    
    Vector<ZoomSettings> region2_zoom = new Vector<ZoomSettings>();

    Vector<ZoomRegionSet> zoomed_regions = new Vector<ZoomRegionSet>();
    Vector<Region1D<ZoomRegionSet>> original2zoomed_regions = new Vector<Region1D<ZoomRegionSet>>();
    
    public String getCategoryInfo()
    {
        return "" + category2zoom;
    }
    

    public int getZoomedLength()
    {
        if (!up2date)
        {
            throw new RuntimeException("not up2date, update() first!");
        }
        //log.debug("get zoomed length ("+zoomed_length+") invoked by:\n"+ExceptionManager.logCallerStack());
        return zoomed_length;
    }
    
    
    public ZoomedCoordinateConverter(int original_length)
    {
        length = original_length;
        log.debug("length set to %d",length);
    }


    public int getOriginalLength()
    {
        return length;
    }
    
    /** sets the default values*/
    public ZoomedCoordinateConverter(int original_length, Double default_factor, Integer default_fixed_width)
    {
        this(original_length);
        setDefaults(default_factor,default_fixed_width);
    }

    public void setDefaults(Double default_factor, Integer default_fixed_width)
    {
        default_zoom.set(default_fixed_width,default_factor);
        invalidate();
    }
    public void reset(int length)
    {
        up2date=false;
        this.length=length;
        log.debug("length set ot : %d", length);
        registered_regions.clear();
        indexed_regions.clear();
        region2_categories.clear();
        invalidate();

    }
    public interface ZoomChangeListener 
    {
        public void zoomSettingsChanged();
        

        
    }

    void invalidate()
    {
        up2date=false;
        for (ZoomChangeListener zcl: listeners)
        {
            zcl.zoomSettingsChanged();
        }
    }

    public boolean isUp2Date()
    {
        return up2date;
    }
    boolean up2date=false;

    
    public ZoomSettings addZoomSettings(Integer width, Double factor)
    {
        return new ZoomSettings(this, width, factor);
    }

    public void addListener(ZoomChangeListener zcl)
    {
        listeners.add(zcl);
    }

    public void removeListener(ZoomChangeListener zcl)
    {
        listeners.remove(zcl);
    }
    public static class ZoomSettings
    {
        ZoomedCoordinateConverter parent= null;
        Integer fixed_width;
        Double factor = null;

        protected ZoomSettings(ZoomedCoordinateConverter parent)
        {
            this(parent, null, 1.0);
        }
        protected ZoomSettings(ZoomedCoordinateConverter parent, Integer width, Double factor)
        {
            this.parent=parent;
            set(width,factor);
            
        }

        double getConversionFactor(int length)
        {
            if (fixed_width!=null)
            {
                return ((double)fixed_width)/length;
            }
            return factor;
        }
        /** width will overrule! */
        public void set(Integer width, Double factor)
        {
            
            fixed_width=width;
            this.factor=(fixed_width==null)?factor:null;
            
            parent.invalidate();
        }
        public void setFixedWidth(int w)
        {
            set(w,null);
            
        }

        public void setFactor(double d)
        {
            set(null,factor);
            
        }

        public String toString()
        {
            return String.format("fixed: %s factor: %s", fixed_width, factor);
        }
    }


    public ZoomSettings addCategory(Object o, Double factor, Integer fixed_width)
    {
        ZoomSettings rv = category2zoom.get(o);
        if (rv==null)
        {
            category2zoom.put(o, rv = new ZoomSettings(this, fixed_width, factor));
        }
        rv.set(fixed_width,factor);

        invalidate();
        return rv;
        
        
    }
    public void addRegion(Region1D region, Object ... categories)
    {
        addRegion(region, null, categories);
    }                          

    int num_removed=0;

    public void removeRegion(Region1D region)
    {
        int idx = Region1D.getIndex(region, null, registered_regions,true);

        registered_regions.set(idx,null);
        if (++num_removed<100)
        {
            invalidate();
            return;
        }
            
        
        //make the lists shorter
        //@TODO
        System.err.printf("GERGELY wanted to make the lists shorter here: %s!!!\n", ExceptionManager.getCallerInfo(1));
        return;
        
    }
    public void addRegion(Region1D region, ZoomSettings zoomsetting, Object ... categories)
    {
        indexed_regions.add(new Region1D(region.getX1(), region.getX2(),indexed_regions.size()));
        registered_regions.add(region);
        region2_zoom.add(zoomsetting);
        region2_categories.add(categories);
        
        invalidate();
    }                                                                                           


    
    void check_zoomed_coordinate(int zoomed_coordinate)
    {

       if(!up2date)
        {
            throw new RuntimeException("not up2date, update first!");
        }
        if (zoomed_coordinate<0 || zoomed_coordinate>zoomed_length)
        {                      
            throw new RuntimeException("invalid zoomed coordinate requested: "+zoomed_coordinate+" zoomed length is : "+zoomed_length);
        }
        
    }
        
    
        
    public Pair<Integer,Integer> getOriginalCoordinates(int zoomed_coordinate)
    {
        check_zoomed_coordinate(zoomed_coordinate);

        return Pair.create(zoomed2_start[zoomed_coordinate],zoomed2_end[zoomed_coordinate]);
    }

    public HashSet<Region1D> getRegionsAtZoomedPosition(int zoomed_coordinate)
    {
        check_zoomed_coordinate(zoomed_coordinate);
        
        int idx = RegionMapper.getFirstNotSmallerV(zoomed_coordinate, zoomed_regions, false);
        return (HashSet<Region1D>)zoomed_regions.get(idx).getObject();
        
    }


    int addZoomRegionSet(int lastpos, Region1D<HashSet> hr)
    {
        if (hr==null)
        {
            //no annots, last region
            
            addZoomRegionSet(new ZoomRegionSet(zoomed_length, lastpos, length, null));
            return length;
        }
        HashSet<Region1D> annots = hr.getObject();
        
        if (hr.getX1()>lastpos)
        {                  
            addZoomRegionSet(new ZoomRegionSet(zoomed_length, lastpos, hr.getX1(), null));
        }

        addZoomRegionSet(new ZoomRegionSet(zoomed_length, hr.getX1(), hr.getX2(), annots));
        return hr.getX2();
            
            
    }
    void addZoomRegionSet(ZoomRegionSet zrs)
    {
        log.debug("will add zoom region set: "+zrs);
        zoomed_length = zrs.getZoomedEnd();
        zoomed_regions.add(zrs);
        Region1D<ZoomRegionSet> r  = new Region1D<ZoomRegionSet>(zrs.in_x1, zrs.in_x2, zrs);
        original2zoomed_regions.add(r);
    }

    public RegionVector getConvertedRegVec(RegionVector irv)
    {
        RegionVector rv = new RegionVector(irv.getObject());
        for (Region1D r : irv.getRegions())
        {
            rv.addRegion(getConvertedRegion(r));;
        }
        return rv;
    }
    public Region1D getConvertedRegion(Region1D r)
    {
        /*log.info("will convert input region : %s => %d, %d ", r, 
                    getConvertedCoordinate(r.getX1()),
                            getConvertedCoordinate(r.getX2())
                                                   );
          */
        return new Region1D(getConvertedCoordinate(r.getX1()),
                            getConvertedCoordinate(r.getX2()),
                            r.getObject());
                             
    }

    public Pair<int[],int[]> getZoomMappings()
    {
        return Pair.create(zoomed2_start, zoomed2_end);
    }

    public int getConvertedCoordinate(int x)
    {
        
        int idx = RegionMapper.getFirstNotSmallerV(x, original2zoomed_regions, false);
        idx = Math.min(idx, original2zoomed_regions.size()-1);
        if (false && idx>=original2zoomed_regions.size())
        {
        
            log.debug("x: "+x+" got idx: "+idx+" size: "+ original2zoomed_regions.size());
            log.debug("caller: \n"+ExceptionManager.logCallerStack());
        }
        //log.debug("convert x: " + x + "  length: " + length);
        ZoomRegionSet zrs = original2zoomed_regions.get(idx).getObject();
        int rv = zrs.getConvertedCoordinate(x);

        if (rv >= zoomed_length)
        {
            log.debug("x: "+x+" got idx: "+idx+" size: "+ original2zoomed_regions.size()+" out: "+rv);
            throw new FRuntimeException("programming bug x: %d is converted to zooming co: %d but zoomed length is: %d!", x, rv, zoomed_length);
        }
        return rv;

    }
    public void update()
    {
        if (up2date)
            return; 
        
        
        zoomed_regions.clear();
        original2zoomed_regions.clear();
        
        log.debug("will setup zooom cordinate converter defualt converter: "+default_zoom+" got "+registered_regions.size()+" regions, length: %d", length);
        RegionMapper rm =  new RegionMapper();
        for (int i=0; i<registered_regions.size(); i++)
        {
            if (registered_regions.get(i)==null)
            {
                continue;
            }
            log.debug("add next region at "+i+" :" +registered_regions.get(i));
            rm.addRegion(indexed_regions.get(i));
        }
        Vector<Region1D<HashSet>> v = rm.compile();

        log.debug("got "+v.size()+" non overlapping regions");

        if (log.isDebugEnabled())
        {
            for (Region1D r : v)
            {
                log.debug("next: "+r);
            }
        }
        zoomed_length = 0;
        

        
        
        int last_zoomed_pos = 0;
        int lastpos=0;
        

        for (Region1D<HashSet> hr : v )
        {

            addZoomRegionSet(lastpos, hr);

            lastpos = hr.getX2();
        }
        if (lastpos<length)
        {
            addZoomRegionSet(lastpos, null);
        }
        
        zoomed_length++;
        
        zoomed2_start = new int[zoomed_length];
        zoomed2_end = new int[zoomed_length];

        for (int i=0; i<zoomed_length; i++)
        {
            zoomed2_start[i]=length;
            zoomed2_end[i]=0;
        }
        for (ZoomRegionSet zrs: zoomed_regions)
        {
            log.debug("next zoomed region: "+zrs);
            //int getConvertedCoordinate(int x)
            int z_start= zrs.img_start;
            int z_end= zrs.img_start+zrs.normed_length;

            for (int i=zrs.in_x1; i<zrs.in_x2; i++)
            {
                int z_i = zrs.getConvertedCoordinate(i);
                zoomed2_start[z_i]=Math.min(zoomed2_start[z_i], i);
                zoomed2_end[z_i]=Math.max(zoomed2_end[z_i], i);
            }
            
        }

        log.debug("original 2 zoomed infos");

        for(Region1D<ZoomRegionSet> r:  original2zoomed_regions)
        {
            ZoomRegionSet zrs = r.getObject();
            //log.info("next orig region: %d-%d: %s",r.getX1(), r.getX2(), zrs);
                       
        }

       
        up2date=true;
        



    }


    class ZoomRegionSet extends Region1D
    {
        int in_x1;
        int in_x2;
        int in_length=0;
        
        int img_start;
        
        double factor;
        int normed_length;

        
        public String toString()
        {
            return String.format("in: %d-%d out: %d-%d factor: %2.3f annots: %d\n",
                                 in_x1, in_x2, img_start, img_start+normed_length, factor, 
                                 (getObject()==null)?0:((HashSet)getObject()).size());
                
        }
        int getZoomedEnd()
        {
            return img_start + normed_length;
        }

        double getRelevantFactor(HashSet<Region1D> regions)
        {
            if (regions==null)
            {
                return default_zoom.getConversionFactor(in_length);
            }
            HashSet<ZoomSettings> zooms = new HashSet<ZoomSettings>();

            double conv_factor=0.0;
                               
            ZoomSettings zs = null;
            Object[] cats = null;
            for (Region1D r : regions)
            {
                int idx = (Integer)r.getObject();
                if((zs=region2_zoom.get(idx))!=null)
                {
                    conv_factor = Math.max(conv_factor, zs.getConversionFactor(r.getLength()));

                    continue;
                }
                if (null==(cats=region2_categories.get(idx)))
                {
                    continue;
                }
                
                for (Object o : cats)
                {
                    if(null==(zs = category2zoom.get(o)))
                        continue;
                   
                    conv_factor = Math.max(conv_factor, zs.getConversionFactor(r.getLength()));
                    
                }

            }
            
            if (conv_factor==0.0)
            {
                return default_zoom.getConversionFactor(in_length);

            }
            return conv_factor;
            
        }
        public ZoomRegionSet(int img_start, int x1, int x2, HashSet<Region1D> regions)
        {
            //calc the relevant zoom setting
            super(img_start, img_start+x2-x1, regions);
            this.img_start=img_start;
            this.in_x1=x1;
            this.in_x2=x2;
            in_length=in_x2-in_x1;

            factor= getRelevantFactor(regions);

            normed_length = (int)Math.ceil(factor*in_length);
            factor = ((double)normed_length)/in_length;
            reset(img_start, img_start+normed_length);

        }

        int getConvertedCoordinate(int x)
        {
            return img_start+(int)((x-in_x1)*factor);
        }
    }
    

}
