package lmu.utils.swing;

import java.awt.*;
import java.awt.event.ActionListener;
import java.io.*;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;

import lmu.utils.*;
import lmu.utils.plotting.*;

public class PagedDataTable extends JPanel 
{
    public class PDSorter extends TableRowSorter 
    {

        public PDSorter()
        {
            super(table);
        }
        public void toggleSortOrder(int column) 
        {

            //table.toggleSortOrder(column);
            //super.toggleSortOrder(column);
        }

        public void sort()
        {
        }
    }


    PagingHeader header;
    DataTable data;
    PagedDataTableModel table;
    JTable jtable;
    //DTHeaderRenderer hrend = new DTHeaderRenderer();

    static Icon SORTED_UP = UIManager.getIcon("Table.ascendingSortIcon");
    static Icon SORTED_DOWN = UIManager.getIcon("Table.descendingSortIcon");
    LinkedList<DataDetailHandler > detail_handlers = new LinkedList<DataDetailHandler >();

    public void addDetailHandler(DataDetailHandler dh)
    {
        detail_handlers.add(dh);
    }
    public void removeDetailHandler(DataDetailHandler dh)
    {
        detail_handlers.remove(dh);
    }
    class DTHeaderRenderer extends DefaultTableCellRenderer
    {
        Boolean sorted=null;
        int index=0;

        public DTHeaderRenderer(int index)
        {
            this.index=index;
            setHorizontalAlignment(CENTER);
            setHorizontalTextPosition(LEFT);
            setVerticalAlignment(BOTTOM);
            setOpaque(false);
        }

        public Component getTableCellRendererComponent(JTable table, Object value,
                                                       boolean isSelected, boolean hasFocus, int row, int column) 
        {
            super.getTableCellRendererComponent(table, value,
                                                isSelected, hasFocus, row, column);
            JTableHeader tableHeader = table.getTableHeader();
            if (tableHeader != null)
            {
                setForeground(tableHeader.getForeground());
            }

            setIcon(getIcon(table, column));
            setBorder(UIManager.getBorder("TableHeader.cellBorder"));
            return this;
        }

        protected Icon getIcon(JTable table, int column) 
        {
            if (sorted==null)
            {
                return null;
            }
            return (sorted)?SORTED_UP:SORTED_DOWN;
            /*
            if (sorted)
            {
                return UIManager.getIcon("Table.ascendingSortIcon");
            }
            return UIManager.getIcon("Table.descendingSortIcon");
            */

        }

        public void sort()
        {
            sorted=table.toggleSortOrder(index);
        }
    }
    class DTCellRenderer extends javax.swing.table.DefaultTableCellRenderer
    {

        int index=0;
        public DTCellRenderer(int index) { super(); this.index=index;}


        public void setValue(Object value) 
        {
            //System.out.println("header renderer for index: "+jtable.getTableHeader().getColumnModel().getColumn(index).getHeaderRenderer());
            //System.out.println("index: "+index+" vlaue: "+value);
            setText(data.header.headers.get(index).getFormatted(value));

        }


    }

    JLabel filterInfo = new JLabel("unfiltered");


    JTextField filterText = new JTextField();
    JCheckBox clearCB = new JCheckBox();

    JButton filtButton = new JButton("filter");
    JButton clearButton = new JButton("clear");
    JButton exportButton = new JButton("exportCSV");

    Vector<DTHeaderRenderer> headers = new Vector<DTHeaderRenderer>();



    ObjectGetter _selected = null;

    public void exportCSV()
    {
        JFileChooser chooser  = new JFileChooser();
        int option = chooser.showSaveDialog(this);
        if (option != JFileChooser.APPROVE_OPTION)
            return;
        

        data.writeCSV(chooser.getSelectedFile());
    }
    public void clear()
    {
        data.clearFilter();
        filterInfo.setText("unfiltered");
    }
    public void filter()
    {
        //System.out.println("filter invoked: "+filterText.getText()+" clear: "+clearCB.isSelected());
        String trimmed = filterText.getText().trim();
        if (trimmed.length()==0)
        {
            data.clearFilter();
            filterInfo.setText("unfiltered");
            return;
        }

        try
        {
            data.quickFilter(trimmed, clearCB.isSelected());
        }
        catch (Exception e)
        {
            e.printStackTrace();
            JOptionPane.showMessageDialog(this, "error at filter: "+ e.getMessage());
            
        }

    


        filterInfo.setText(String.format("filtered: %d/%d", data.getFilteredSize(), data.getSize()));
    }

    void insertIntoFilter(String s)
    {
        String prefix = filterText.getText().substring(0,filterText.getCaretPosition());
        String suffix = filterText.getText().substring(filterText.getCaretPosition());

        prefix+=s;
        filterText.setText(prefix+suffix);
        filterText.setCaretPosition(prefix.length());
        filterText.requestFocus();
    }
    class MouseHandler extends java.awt.event.MouseAdapter
    {

        public void headerClicked(java.awt.event.MouseEvent evt)
        {
            JTable table = ((JTableHeader) evt.getSource()).getTable();
            TableColumnModel colModel = table.getColumnModel();

            int index = colModel.getColumnIndexAtX(evt.getX());
            if (index == -1)
                return;

            String name = visible_headers.get(index).getName();
            if (evt.getClickCount()==2 && evt.isControlDown())
            {
                //write it into filter
                insertIntoFilter("${"+name+"}");
                
            }

            if (evt.getClickCount()==1  && !evt.isControlDown())
            {
                for (DTHeaderRenderer dth : headers)
                {
                    dth.sorted=null;
                }
                DTHeaderRenderer dth = headers.get(index);
                dth.sort();
                return;
            }




        }
        public void mouseClicked(java.awt.event.MouseEvent evt) 
        {
            if (evt.getSource() instanceof JTableHeader)
            {
                headerClicked(evt);
                return;
            }

            
            int row = jtable.rowAtPoint(evt.getPoint());
            int col = jtable.columnAtPoint(evt.getPoint());

            

            if (row<0 || row>data.getFilteredSize())
            {
                return;
            }
            
            Object obj = visible_headers.get(col).get(table.getValueAt(row, col));

            String headername = visible_headers.get(col).getName();
            
            
            //System.out.printf("selected: %d:%d = %s\n", row,col, obj);

            
            if (evt.isControlDown())
            {
                if (obj instanceof String)
                {
                    //insertIntoFilter("${"+headername+"}.indexOf("+obj+")>=0");
                    insertIntoFilter("${"+headername+"} ~ " + obj);
                }
                else
                {
                    insertIntoFilter("${"+headername+"}=="+obj);
                }
            }
        }
        
    }

    Vector<DataHeader> visible_headers = null;
    MouseHandler mouseHandler = new MouseHandler();
    JMenuBar menu = new JMenuBar();

    public PagedDataTable(DataTable dt)
    {
        super();
        this.data=dt;
        table = new PagedDataTableModel(dt);


        header = new PagingHeader(table);
        jtable = new JTable(table);




        jtable.setAutoCreateRowSorter(true);


        jtable.setCellSelectionEnabled(true);
        jtable.setRowSelectionAllowed(true);


        ListSelectionModel cellSelectionModel = jtable.getSelectionModel();

        cellSelectionModel.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        cellSelectionModel.addListSelectionListener(new ListSelectionListener() 
                                                    {
                                                        public void valueChanged(ListSelectionEvent e) 
                                                        {

                                                            
                                                            int[] selectedRow = jtable.getSelectedRows();
                                                            if (selectedRow==null || selectedRow.length==0)
                                                            {
                                                                _selected=null;
                                                                return;

                                                            }
                                                            if (_selected == (ObjectGetter)table.getRowObject(selectedRow[0]))
                                                            {
                                                                return;
                                                            }
                                                            System.out.println(e);
                                                            System.out.printf("selected : %d\n",selectedRow[0]);
                                                            _selected = (ObjectGetter)table.getRowObject(selectedRow[0]);
                                                            System.out.println("infos: "+_selected.getInfos());
                                                            for (DataDetailHandler ddh: detail_handlers)
                                                            {
                                                                ddh.selectData(_selected);
                                                            }
                                                        }
                                                    }
                                                   );


        PDSorter pds = new PDSorter();
        jtable.setRowSorter(pds);


        setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
        
        
        //

        JPanel menuPanel = new JPanel();
        add(menuPanel);
        menuPanel.add(menu);
        add(header);
        JPanel filterPanel = new JPanel();
        filterText.setColumns(80);
        ReflectedObjectListener filter_listener = new ReflectedObjectListener(new ReflectedObject("filter()"), this);
        filterText.addActionListener(filter_listener);
        filtButton.addActionListener(filter_listener);
        clearButton.addActionListener(new ReflectedObjectListener(new ReflectedObject("clear()"), this));
        exportButton.addActionListener(new ReflectedObjectListener(new ReflectedObject("exportCSV()"), this));
        filterPanel.add(filterInfo);


        filterPanel.add(exportButton);
        filterPanel.add(filtButton);
        filterPanel.add(clearButton);

        clearCB.setSelected(true);
        //filterPanel.add(new JLabel("clear"));
        clearCB.setText("clear");
        filterPanel.add(clearCB);
        filterPanel.add(filterText);




        JScrollPane jsp = new JScrollPane(filterPanel);
        Dimension d = filterPanel.getPreferredSize();
        jsp.setPreferredSize(new Dimension((int)d.getWidth()+20, (int)d.getHeight()+20));

        add(jsp);

        jsp=new JScrollPane(jtable);
        add(jsp);




        //jtable.getTableHeader().set
        TableColumnModel tcm  = jtable.getTableHeader().getColumnModel();

        visible_headers = dt.header.getVisibleHeaders();
        int total_width=0;
        int font_width = 11;
        for (int i=0; i<tcm.getColumnCount(); i++)
        {
            TableColumn tc = tcm.getColumn(i);



            
            DataHeader dh = visible_headers.get(i);
            
            
            int prefwidth = Math.max(40, dh.getName().length()*font_width);

            //System.out.printf("visible headers: %d -> %d init prefwidth: %d\n", i, dh.col_index, prefwidth);
            
            //check the first 100 entries (if there are enough) to set the real width

            for (int j=0; j<100 && j < dt.getSize(); j++)
            {
                 int cl = dh.getFormatted(dt.getDataAt(j)).length();
                 prefwidth = Math.max(prefwidth, cl*font_width);
            }
            //System.out.printf("set prefwidth: %d\n", prefwidth);
            tc.setPreferredWidth(prefwidth);
            total_width+=tc.getWidth();

            DTCellRenderer dtc = new DTCellRenderer(dh.col_index);

            tc.setCellRenderer(dtc);
            //tc.setHeader

            DTHeaderRenderer dth=new DTHeaderRenderer(dh.col_index);

            headers.add(dth);


            tc.setHeaderRenderer(dth);
            tc.setHeaderValue(dh.getName());


        }


        jtable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        //getViewport().setPreferredSize(panel_d);
        //jtable.setMinimumSize(new Dimension(total_width, (int)jtable.getPreferredSize().getHeight()));
        //jsp.getViewport().setPreferredSize(new Dimension(total_width, 600));
        //jsp.setPreferredSize(new Dimension(total_width, (int)jsp.getPreferredSize().getHeight()));
        //System.out.println("jtable prefsize: "+jtable.getPreferredSize());
        //System.out.println("jtable maxsize: "+jtable.getPreferredSize());
        jtable.getTableHeader().addMouseListener(mouseHandler);
        jtable.addMouseListener(mouseHandler);
    }

    public static interface DataDetailHandler 
    {
        public void selectData(ObjectGetter data);
    }

    public static abstract class DataDetailView extends JPanel implements DataDetailHandler
    {
        public abstract void selectData(ObjectGetter data);
    }

    public static class DefaultDetailView extends DataDetailView
    {
        //JPanel panel = new JPanel();
        JLabel info = new JLabel("no selection");
        DataTable dt;
        Vector<JLabel> comps = new Vector<>();
        BoxLayout layout = null;

        public DefaultDetailView(DataTable dt)
        {
            super();
            this.dt = dt;
            //add(panel);
            setLayout(layout = new BoxLayout(this, BoxLayout.Y_AXIS));
            addLabel("no selection");
            
        }
        void addLabel(String l, Object ... data)
        {
            JLabel jl = new JLabel(String.format(l, data));
            comps.add(jl);
            //System.out.println("add label : " +jl);
            add(jl);
        }
        public void selectData(ObjectGetter data)
        {
            
            ObjectGetter.apply(comps, (JLabel jl) -> remove(jl));
            comps.clear();
            setLayout(layout = new BoxLayout(this, BoxLayout.Y_AXIS));
            for (DataHeader dh : dt.header.headers)
            {
                addLabel("%s = %s\n", dh.getName(), dh.getFormatted(data));

                //sb.printf("\t%s = %s\n", dh.getName(), dh.getFormatted(data));
            }
            
            //info.setText("selected: \n"+sb.getBuffer());
            revalidate();
        }
    }

    public static class MJFrame extends MultiFrameWindow
    {
        PagedDataTable pdt;
        public MJFrame(PagedDataTable pdt)
        {
            super();
            this.pdt = pdt;
        }

        public MJFrame addMenu(String name, DataDetailHandler handler)
        {
            JMenuItem jmi = new JMenuItem(name);
            pdt.menu.add(jmi);
            
            MenuDetailHandler mdh = new MenuDetailHandler(handler);
            pdt.addDetailHandler(mdh);
            jmi.addActionListener((e) -> mdh.run());
            this.update();
            return this;
        }
    }

    public static MJFrame getDetailViewFrame(File csv, Container c, DataDetailHandler handler)
    {
        DataTable dt = new DataTable();
        dt.readDataFile(csv,true, true, null);
        return getDetailViewFrame(dt, c, handler);
    }
    public static MJFrame getDetailViewFrame(File csv, String separator, Container c, DataDetailHandler handler)
    {
        DataTable dt = new DataTable();
        dt.readDataFile(csv,true, true, separator);
        return getDetailViewFrame(dt, c, handler);
    }

    public static MJFrame getDetailViewFrame(DataTable data, Container c, DataDetailHandler handler)
    {
        return getDetailViewFrame(null, data, c, handler);
    }

    static class MenuDetailHandler implements DataDetailHandler
    {
        ObjectGetter data;
        DataDetailHandler action_handler;

        public 
            
            
            
            
            
            
            MenuDetailHandler(DataDetailHandler dh)
        {
            action_handler = dh;
        }

        public void selectData(ObjectGetter data)
        {
            this.data = data;
        }

        public void run()
        {
            if (data == null)
                return;
            action_handler.selectData(data);
        }
    }
        
    public static MJFrame getDetailViewFrame(String title, DataTable data, Container c, DataDetailHandler handler)
    {

        PagedDataTable pdt = new PagedDataTable(data);
        MJFrame mfw = new MJFrame(pdt);
        
        if (title != null)
        {
            mfw.setTitle(title);
        }
        

        mfw.addFrameComponent(pdt, "data table");
        if (c==null)
        {
            c = new DefaultDetailView(data);
            pdt.addDetailHandler((DefaultDetailView)c);
        }
        if (handler!=null)
        {
            pdt.addDetailHandler(handler);
        }

        if (c instanceof JInternalFrame)
        {
            mfw.addFrame((JInternalFrame)c, "details");
        }
        else
        {
            mfw.addFrameComponent(c, "details");
        }

        mfw.update();
        mfw.setPreferredSize(new Dimension(600, 800));
        mfw.setVisible(true);
        return mfw;

    }


    
    public static void main (String args[]) 
    {
        Locale.setDefault(Locale.UK);
        LogConfig lc = LogConfig.configureLog(null, true);

        
        SimpleOptionParser cmd = new SimpleOptionParser("i", "sep", "headers", "h", "csv", "peak", "noheader", "comment");
        cmd.setFile("i");
        cmd.setSwitches("h", "csv", "noheader");
        cmd.setInt("peak");
        cmd.setOptional("headers", "sep", "peak", "comment");
        if (!OptionParser.parseParams(args,true,false, true,cmd, lc))
        {
            return;
        }

        Logger log = lc.getLogger();
        //getDetailViewFrame(cmd.getFile("i"), "\t", null, null);

        /*ZoomedMultiPicPlaneIFrame mppi = new ZoomedMultiPicPlaneIFrame(5000);
        RegionPicturePane rpp = mppi.createRegionPane("gene");
        RegionPlot rp = rpp.addRegion("transcript1");
        rp.addRegion(new Region1D(200, 1200, null));
        rp.addRegion(new Region1D(2200, 4200, null));
        rpp.update();
        mppi.update();
        mppi.pack();
        mppi.setPreferredSize(new Dimension(600, 500));
        //mppi.setVisible(true);
        */
        String com = cmd.getOptionalValue("comment", null);
        
        HashSet<String> headers = cmd.getOptionalSet("headers", ",", null);
        DataTable dt = new DataTable();
        
        if (com != null)
        {
            dt.setCommentSigns(com);
        }
        String sep = cmd.getOptionalValue("sep", null);
        if (sep!=null && sep.equalsIgnoreCase("tab"))
        {
            sep="\t";
        }
        if (sep!=null && sep.equalsIgnoreCase("space"))
         {
             sep=" ";
         }
        
        if (cmd.isSet("h"))
        {
            DataFile.printHeaderOverview(cmd.getFile("i"), sep);
            return;
        }

        boolean gotheader = !cmd.isSet("noheader");
        dt.readDataFile(cmd.getFile("i"), cmd.isSet("csv"), gotheader, sep, cmd.getOptionalInt("peak", -1));

        if (headers!=null)
        {
            for (String n: dt.header.getHeaderNames())
            {
                if (headers.contains(n))
                    continue;

                DataHeader dh = dt.header.getHeader(n);
                if (dh==null)
                    continue;

                dh.setVisible(false);
            }
        }

        JFrame jf = getDetailViewFrame(dt, null , null);

        jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE) ;

    }
}
