package lmu.utils.swing;

import javax.swing.*;
import javax.swing.table.*;
import java.util.*;

import lmu.utils.*;

public abstract class PagingModel extends AbstractTableModel
{

  protected int pageSize;

  protected int pageOffset;

  protected int data_size;
  

  
  public PagingModel(int numRows, int size) 
  {
      this.pageSize=numRows;
      data_size=size;
      pageOffset=0;
    
  }

  
  // Return values appropriate for the visible table part.
  public int getRowCount() 
  {
    return Math.min(pageSize, data_size-(pageOffset*pageSize));
  }

  
  // Work only on the visible part of the table.
  public abstract Object getRowObject(int row) ;
  
  

  // Use this method to figure out which page you are on.
  public int getPageOffset() 
  {
    return pageOffset;
  }

  public int getPageCount() 
  {
    return (int) Math.ceil((double) data_size / pageSize);
  }

  // Use this method if you want to know how big the real table is . . . we
  // could also write "getRealValueAt()" if needed.
  public int getRealRowCount() 
  {
    return data_size;
  }

  public int getPageSize() 
  {
    return pageSize;
  }

  public void setPageSize(int s) 
  {
    if (s == pageSize) 
    {
      return;
    }
    int oldPageSize = pageSize;
    pageSize = s;
    pageOffset = (oldPageSize * pageOffset) / pageSize;
    fireTableDataChanged();
    
  }


  public void toPage(int p)
  {
      p = Math.max(0,p);
      p = Math.min(p, getPageCount()-1);

      pageOffset=p;
      fireTableDataChanged();

  }
  // Update the page offset and fire a data changed (all rows).
  public void prevPage() 
  {
      toPage(pageOffset-1);
  }

  public void nextPage() 
  {
      toPage(pageOffset+1);
      
  
 }


}

