package ortholog;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map.Entry;
import java.util.PriorityQueue;
import java.util.concurrent.atomic.AtomicInteger;

import algo.alignment.Alignment;
import genome.Gene;
import genome.GenomicUtils;
import genome.Organism;
import genome.Protein;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.ObjectGetter;
import lmu.utils.OptionParser;
import lmu.utils.Pair;
import lmu.utils.SimpleOptionParser;
import lmu.utils.ThreadUtils;
import ortholog.TaxonomyNode.SpeciesDistance;
import protein.algo.ProteinAligner;

public class GeneStructureProteinPath  {
    //private static float MIN_ID = 0.4f;
    private static Logger log = LogConfig.getLogger();

    enum Mode{
        PHYLOGENY,
        SPANNING_TREE
    }
    
    HashMap<Pair<String,String>,Float> pair2Info = new HashMap<>();
    HashMap<String,List<Pair<String,Float>>> ptoPairInfo = new   HashMap<String,List<Pair<String,Float>>> ();

    Dijkstra<String> dj = null;

    //List<Gene> genes = null;
    
    protected HashMap<Integer,List<String>> org2Gene = new HashMap<> ();
    protected HashMap<Integer,Integer> org2Index = new  HashMap<>();
    protected List<Integer> orderedOrgs = new ArrayList<>();
    protected HashMap<String,List<String>> geneId2ProteinId = new HashMap<>();
    protected HashMap<String,String> protein2Seq = new HashMap<String,String>();
    protected Integer startOrg = null;
    
    ProteinAligner aligner = new ProteinAligner();
    
    private int nextOrganismIndex(int start){
        for(int i = start; i< orderedOrgs.size();i++){
            if ( org2Gene.containsKey(orderedOrgs.get(i))) return i;
        }

        return -1;
    }

/*
    public GeneStructureProteinPath(MultiSpeciesGeneMapping gene_mappings,Collection<Organism> orgs,TaxonomyNode tnroot,  Gene gene){
        List<SpeciesDistance> sds = TaxonomyNode.getDistances(gene.organism, tnroot);
        List<Organism> orderedOrgs = new ArrayList<Organism>();
        ObjectGetter.apply(sds, (SpeciesDistance sd) -> {orderedOrgs.addAll(sd.organisms);});

        init(new OrthologGeneGroup(gene_mappings,gene,null).getRelevantGenes(),orderedOrgs,gene);
        generateOptimalPaths();
    }
*/
    public GeneStructureProteinPath(Collection<Gene> genes,List<Integer> orderedOrgs){
        init(genes,orderedOrgs);
        
        generateOptimalPaths();
    }
    protected GeneStructureProteinPath()
    {
        
    }
 /*
    public GeneStructureProteinPath(MultiSpeciesGeneMapping gene_mappings,List<Organism> orderedOrgs,  Gene gene){
        init(new OrthologGeneGroup(gene_mappings,gene,null).getRelevantGenes(),orderedOrgs,gene);
        generateOptimalPaths();
    }
    */
    
    private void init(Collection<Gene> genes,List<Integer> orgs)
    {

        ObjectGetter.apply(genes, (Gene g) -> 
        {
            if ( !org2Gene.containsKey(g.organism.getTaxId()))
                org2Gene.put(g.organism.getTaxId(),new ArrayList<String>()); 
            
            org2Gene.get(g.organism.getTaxId()).add(g.id); 
            if ( g.proteins.size() == 0)
            {
                throw new RuntimeException(g + " has no proteins");
            }
            List<String> proteins = new ArrayList<>();
            for(Protein p : g.proteins.values()){

                String seq = GenomicUtils.getProteinSeq(p,g.organism.getGenome(),false);
                protein2Seq.put(p.id,seq);
                proteins.add(p.id);
            }
            geneId2ProteinId.put(g.id,proteins);
        });
        
        this.orderedOrgs = new ArrayList<>(orgs);
        //ObjectGetter.apply(orgs, (Organism o) -> this.orderedOrgs.add(o.getTaxId()));
        for(int i = 0; i< orderedOrgs.size(); i++){
            if ( !this.org2Gene.containsKey(orderedOrgs.get(i)) )
            {
                continue;
            }
            if ( this.startOrg == null){
                this.startOrg = orderedOrgs.get(i);
            }
            this.org2Index.put(orderedOrgs.get(i), i);
            
        }
        /*
        Collections.sort(this.genes, (String g1, String g2) -> {
            if ( g1 ==startGene ) return -1;
            if ( g2 ==startGene ) return 1;

            return org2Index.get(g1.organism).compareTo(org2Index.get(g2.organism));
        });
        */
        log.info("Number of genes in OGG: %d. Number of proteins %d" ,genes.size(),protein2Seq.size());


    }
/*
    public static List<Protein> spanningTree(OrthologGeneGroup ogg)
    {
        HashMap<Gene,Protein> gene2SelectedProtein = new  HashMap<Gene,Protein>();
        
        HashMap<Pair<Protein,Protein>,OrthologInfo> p2p = new   HashMap<Pair<Protein,Protein>,OrthologInfo>();
        
        PriorityQueue<Pair<Protein,Protein>> seqIds = new PriorityQueue<Pair<Protein,Protein>>( Comparator.comparing((Pair<Protein,Protein> p)->1-p2p.get(p).getSeqId()));
        
        List<Gene> genes = new ArrayList<Gene>(ogg.getRelevantGenes());
        
        log.info("Genes: %d" , ogg.getRelevantGenes().size());
        ProteinAligner aligner = new ProteinAligner();
        
        HashMap<Protein,String> p2Seq = new HashMap<Protein,String>();
        for(int i =  0 ; i< genes.size();i++)
        {
            for(Protein p : genes.get(i).proteins.values())
            {
                String seq = GenomicUtils.getProteinSeq(p,genes.get(i).organism.getGenome(),false);
                p2Seq.put(p,seq);
            }
        }
        
        for(int i =  0 ; i< genes.size();i++)
        {
            for(int j = i+1 ; j<genes.size() ; j++)
            {
                for(Protein p1 : genes.get(i).proteins.values())
                {
                    for(Protein p2 : genes.get(j).proteins.values())
                    {
                        Pair<Protein,Protein> p = new Pair<Protein,Protein>(p1,p2);
                        
                        OrthologInfo oi = new OrthologInfo(p1,p2Seq.get(p1),p2,p2Seq.get(p2), aligner);
                        
                        //OrthologInfo oi = new OrthologInfo(p1,p2,aligner);
                        p2p.put(p, oi);
                        seqIds.add(p); 
                    }
                }
            }
        }
        log.info("Search path");
        List<Pair<Protein,Protein>> edges = new    ArrayList<Pair<Protein,Protein>>();
        List<Protein> proteins = new ArrayList<Protein>();
        
        while(!seqIds.isEmpty())
        {
            if ( proteins.size() == genes.size()) break;
            Pair<Protein,Protein> p = seqIds.poll();
            
            Gene g1 = p.first.parentTranscript.parentGene;
            Gene g2 = p.second.parentTranscript.parentGene;
            
            if ( gene2SelectedProtein.containsKey(g1) && !gene2SelectedProtein.get(g1).equals(p.getFirst()) )
                continue;
            
            if ( gene2SelectedProtein.containsKey(g2) && !gene2SelectedProtein.get(g2).equals(p.getSecond()) )
                continue;
            
           
            if (! gene2SelectedProtein.containsKey(g2) ){
                proteins.add(p.getSecond());
                gene2SelectedProtein.put(g2, p.getSecond());
            }
            
            if (! gene2SelectedProtein.containsKey(g1) ){
                proteins.add(p.getFirst());
                gene2SelectedProtein.put(g1, p.getFirst());
            }
            edges.add(p);
            
        }
        
        

        
        return proteins;
    }
    */
    public static void main(String[] args) throws Exception
    {
        
        SimpleOptionParser cmdLine = new SimpleOptionParser("ogg","orga","taxf","ogg_dir","start_org","threads");
        cmdLine.setInt("start_org");
        cmdLine.setFile("ogg","orga","taxf");
        cmdLine.setDir("ogg_dir");
        cmdLine.setDefault("start_org", "9606");
        cmdLine.setDefault("threads", "8");
        cmdLine.setOptional("ogg","ogg_dir");
        
        LogConfig lc = LogConfig.configureLog(null, true);

        if (!OptionParser.parseParams(args,true, false, true,  cmdLine))
            return;
        
        Integer startOrg = cmdLine.getInt("start_org");

        
        Organism.readOrganisms(cmdLine.getFile("orga"));
        
        Taxonomy tax = new Taxonomy(cmdLine.getFile("taxf"));
        
        TaxonomyNode tnroot = tax.buildSparseTaxonomy(Organism.getOrganisms(), true);

        List<Integer> orderedOrgs = getOrderedOrgs(Organism.getOrganism(startOrg),tnroot);

        
        for(Organism o : Organism.getAllOrganisms())
        {
            o.getAnnot();
            o.getGenome();
        }
        
        List<File> files = new ArrayList<>();
        
        if ( cmdLine.isOptionSet("ogg"))
        {
            files.add(cmdLine.getFile("ogg"));
        }else if ( cmdLine.isOptionSet("ogg_dir"))
        {
            for(File f : cmdLine.getFile("ogg_dir").listFiles())
            {
                if ( f.getName().endsWith(".xml"))
                {
                    files.add(f);
                }
            }
        }
        
        AtomicInteger k = new AtomicInteger();
        ThreadUtils.runLambdas(files, (File f)->
        {
            OrthologGeneGroup ogg = new OrthologGeneGroup(f);
            
            // spanningTree(ogg);
             
             try{
                 GeneStructureProteinPath gss = new GeneStructureProteinPath(ogg.getRelevantGenes(),orderedOrgs);
                 List<Pair<String, String>> best = gss.getBestPath();
                 List<Pair<String,String>> proteins = gss.extend(best);
                 System.err.println(ogg.getRelevantGenes().size() + " " +proteins.size() + " " + k.getAndIncrement() + " " + files.size());
                 
             }catch(Exception e)
             {
                 e.printStackTrace();
                 System.exit(1);
             }
             
        },cmdLine.getInt("threads"));

    }
    private static List<Integer> getOrderedOrgs(Organism fromOrg, TaxonomyNode tnroot) 
    {
        List<SpeciesDistance> sds = TaxonomyNode.getDistances(fromOrg, tnroot);
        List<Integer> orderedOrgs = new ArrayList<Integer>();
        ObjectGetter.apply(sds, (SpeciesDistance sd) -> {
            for(Organism o : sd.organisms)
            {
                orderedOrgs.add(o.getTaxId());
            }
        });
        
        return orderedOrgs;

    }
    protected void generateOptimalPaths(){
     
        for(int i =  0 ; i< orderedOrgs.size()-1 ; i++){
            int j = nextOrganismIndex(i);
            int k = nextOrganismIndex(j+1);
            if ( j == -1 || k == -1) {
                break;
            }

            List<String> from =  org2Gene.get(orderedOrgs.get(j));
            List<String> to = org2Gene.get(orderedOrgs.get(k));
            
            log.debug("From genes: %s; To genes: %s",from,to);
            int speciesConnections=0;
            for(String g1 : from){
                for(String p1 :this.geneId2ProteinId.get(g1) ){
                    
                    if (!ptoPairInfo.containsKey(p1) ){
                        ptoPairInfo.put(p1, new ArrayList<Pair<String,Float>>());
                    }
                    for(String g2 : to){
                        if ( g2 == g1) continue;
                        log.debug("%s  -- %s ",g1, g2);

                        
                        for(String p2 :this.geneId2ProteinId.get(g2) ){
                            
                            String s1 = protein2Seq.get(p1);
                            String s2 = protein2Seq.get(p2);
                          
                           // OrthologInfo oi = new OrthologInfo(p1,s1,p2,s2, aligner);
                            Float seqId = this.getSeqId(s1, s2);
                            
                            //OrthologInfo oi = new OrthologInfo(p1,p2Seq.get(p1),p2,p2Seq.get(p2), aligner);
                            //log.debug("Ortholog: %s %s; seqId:%.2f; gstrId:%.2f; comId:%.2f",oi.src_id , oi.dst_id , oi.getSeqId() ,oi.getGenomicAlignment().getGeneStructureId() , oi.getSeqGeneStructureId());
                            speciesConnections++;
                            ptoPairInfo.get(p1).add(new Pair<String,Float>(p2,seqId));
                            Pair<String,String> p = new Pair<String,String>(p1,p2);
                            pair2Info.put(p, seqId);
                            
                          
                        }
                    }

                }
            }
            
        }

        dj = new Dijkstra<String>("Start");

        dj.computePaths( 
                (String p1,String p2) -> {
                    if ( p1.equals("Start") ) return 0;
                    
                    Float seqId =  pair2Info.get(new Pair<String,String>(p1,p2));
                    return (float)1f- seqId;
                }, 
                (String p) -> {
                    if ( p.equals("Start"))
                    {
                        List<String> ret = new ArrayList<String>();
                        for(String gId : org2Gene.get(startOrg))
                        {
                            ret.addAll(this.geneId2ProteinId.get(gId));
                        }
                        return ret;
                    }

//                  
//                    if (!ptoPairInfo.containsKey(p) || ptoPairInfo.get(p).size()==0) return ret;
                    List<String> ret = new ArrayList<String>();
                    ObjectGetter.apply(ptoPairInfo.get(p),(Pair<String,Float> pa) ->
                    {
                        ret.add(pa.first);
                        //if ( pa.first.getOrganism() != p.getOrganism()) ret.add(pa.first); 
                    });
                    /*
                    if ( ret.size() == 0)
                    {
                        throw new RuntimeException("No valid path for:" + p);
                    }
                    */
                    return ret;
                }
                ); 

    }
    private float getSeqId(String p1seq, String p2seq){
        Alignment ali = aligner.align(p1seq, p2seq);
        float num_id = ali.getNumId(p1seq, p2seq);
        float seqid = num_id/(0.5f*(p1seq.length()+p2seq.length()));
        return seqid;
    }
    public List<Pair<String,String>> extend(List<Pair<String,String>> geneProteins){
        
        List<Pair<String,String>> ret = new ArrayList<>(geneProteins);
        HashSet<String> consideredGenes = new  HashSet<>();
        ObjectGetter.apply(geneProteins, (Pair<String,String> p)-> consideredGenes.add(p.getFirst()));
        List<Pair<String,String>> additional = new ArrayList<>();
        for(String g1 : this.geneId2ProteinId.keySet())
        {
            if ( consideredGenes.contains(g1)) continue;

            Float bestId = null;
            
            String bestPId = null;
            
            for(String p1 : this.geneId2ProteinId.get(g1))
            {
                for(String g2 : consideredGenes)
                {
                    for(String p2 : this.geneId2ProteinId.get(g2))
                    {
                        Pair<String,String> p = new Pair<String,String>(p1,p2);
                        Float seqId=pair2Info.get(p);
                        if ( seqId == null){
                            seqId =this.getSeqId(this.protein2Seq.get(p1),this.protein2Seq.get(p2));// new OrthologInfo(p1,protein2Seq.get(p1),p2,protein2Seq.get(p2),aligner);
                        }
                        
                        if (  (bestId == null || seqId > bestId))
                        {
                            bestPId = p1;
                            bestId = seqId;
                        }
                    }
                } 
            }
           
            additional.add(Pair.create(g1,bestPId));
            
        }
        ret.addAll(additional);

        return ret;
    }
    public List<Pair<String,String>> getBestPath(){
       // Organism o = null;
        for(int i = orderedOrgs.size()-1 ; i>=0;i--){
            if ( org2Gene.containsKey(orderedOrgs.get(i))){
                Integer taxId = orderedOrgs.get(i);
               
                List<Pair<String,String>> proteins = getBestPath(taxId);
                if ( proteins.size() > 0) 
                    return proteins;
            }
        }
        
        throw new RuntimeException("Could not find path");
    }

    public List<Pair<String,String>> getBestPath(Integer targetTaxId){
        List<Pair<String,String>> ret = new ArrayList<>();  
        
        List<Dijkstra<String>.PathElement> best = null;
        
        if ( !org2Gene.containsKey(targetTaxId)) return ret; 
        
        for(String g : org2Gene.get(targetTaxId)){
            for(String p : this.geneId2ProteinId.get(g) ){
                List<Dijkstra<String>.PathElement> path = dj.getPath(p);
                if ( path == null) continue;
                if( best == null || best.get(best.size()-1).distance > path.get(path.size()-1).distance){
                    best =path;
                }
            }
        } 
      
        for(Dijkstra<String>.PathElement g : best){
            for(Entry<String, List<String>> e : geneId2ProteinId.entrySet())
            {
                if ( e.getValue().contains(g.element))
                    ret.add(Pair.create(e.getKey(),g.element));
                        
            }
            
        }
     
            /*
            ObjectGetter.apply(best,(Dijkstra<String>.PathElement  e) -> {
            String geneId = null;
            
            for(Dijkstra<String>.PathElement g : best){
                if ( this.geneId2ProteinId.get(g).contains(e.element))
                    ret.add(Pair.create(geneId,e.element));
            }
            
        }); */
            
        return ret;
    }
    public static  class   Dijkstra<E extends Object>{
        public interface WeightGetter<E>{
            public float getWeight(E source,E target);
        }
        public interface AdjacentGetter<E>{
            public List<E> getAdjacent(E source);
        }
        private HashMap<E,Pair<E,Float>> bestPrevious = new HashMap<E,Pair<E,Float>>();

        private E start;
        public Dijkstra(E start){
            this.start = start;
        }
        public class PathElement{
            public E element;
            public Float distance;
            public PathElement(Float distance, E element){
                this.distance = distance;
                this.element = element;
            }
            public String toString(){
                return String.format("%.2f (%s)",distance,element);
            }
        }

        public List<PathElement> getPath( E ... targets){

            List<PathElement> bestPath = null;
            Float bestDistance = null;

            for(E target : targets){
               
                List<PathElement> path = new ArrayList<PathElement>();

                E current = target;

                if ( !bestPrevious.containsKey(current)) 
                    continue;
                
                float distance = bestPrevious.get(current).second;
                while(current != start){
                    path.add(new PathElement( bestPrevious.get(current).second,current));
                    current = bestPrevious.get(current).first;
                }

                Collections.reverse(path);
                if ( bestPath == null || distance < bestDistance  ){
                    bestPath = path;
                    bestDistance = distance;
                }
            }
            return bestPath;
        }

        public void computePaths(WeightGetter<E> weighter,AdjacentGetter<E> estimator){

            bestPrevious.put(start, new Pair<E,Float>(null,0f));
            PriorityQueue<E> vertexQueue = new PriorityQueue<E>(Comparator.comparing(e -> bestPrevious.get(e)));

            vertexQueue.add(start);

            while (!vertexQueue.isEmpty()) {
                E u = vertexQueue.poll();
                for (E e : estimator.getAdjacent(u))
                {
                    float distanceThroughU = bestPrevious.get(u).second + weighter.getWeight(u, e);
                    if ( distanceThroughU < 0) throw new RuntimeException("Bug");
                    if ( !bestPrevious.containsKey(e) || distanceThroughU < bestPrevious.get(e).second) {
                        vertexQueue.remove(e);
                        bestPrevious.put(e, new Pair<E,Float>(u,distanceThroughU));
                        vertexQueue.add(e);
                    }
                }
            }
            
        
        }

    }
}
