package ortholog;

import genome.Gene;
import genome.GenomicProteinAlignment;
import genome.GenomicUtils;
import genome.Organism;
import genome.Protein;
import isoformtransfer.InferredExon;
import isoformtransfer.algo.Sim4cc;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.StringReader;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Locale;
import java.util.Map;
import java.util.Vector;
import java.util.zip.GZIPInputStream;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import lmu.utils.BufferedRandomAccessFile;
import lmu.utils.FileUtils;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.ObjectGetter;
import lmu.utils.OptionParser;
import lmu.utils.Pair;
import lmu.utils.Region1D;
import lmu.utils.RegionVector;
import lmu.utils.SimpleOptionParser;
import lmu.utils.XMLUtils;

import org.jdom.Element;

import protein.algo.ProteinAligner;

public class OrthologGeneGroup 
{
    private static double MIN_SEQ_ID = 0.4;

    private double min_seq_id = MIN_SEQ_ID;
    
    protected ProteinAligner aligner = new ProteinAligner();

    protected Logger log = LogConfig.getLogger();

//    HashMap<Gene, HashSet<Gene>> ortholog_edges = new HashMap<Gene, HashSet<Gene>>();

    Vector<Gene> paralogs  = new Vector<Gene>();

  //  HashMap<Gene, Vector<OrthologInfo>> orthologes_nopara = new HashMap<>();

    public HashMap<Gene, Vector<Pair<String, RegionVector>>> transfers = null;

    protected Vector<OrthologInfo> transfer_infos = new Vector<OrthologInfo>();

    HashSet<Gene> noparalog_relevant = null;

    MultiSpeciesGeneMapping gene_mappings = null;
    
    public Gene start;

    protected HashSet<Gene> relevant_genes = new HashSet<Gene>();

    public HashMap<Gene, Vector<OrthologInfo>> orthologes = new HashMap<>();

    public OrthologInfo getInfo(Gene from, Gene to)
    {
        Vector<OrthologInfo>  v = getOrthologs(from);
        if (v==null)
            return null;

        for (OrthologInfo oi : v)
        {
            if (oi.getTargetGene()!=to)
                continue;

            return oi;
        }
        return null;
    }
    
    public HashMap<Organism, HashSet<Gene>> getRelevantMap()
    {
        HashMap<Organism, HashSet<Gene>> rv = new HashMap<>();

        for (Gene g : relevant_genes)
        {
            Organism o = g.organism;
            HashSet<Gene> h = rv.get(o);
            if (h==null)
            {
                rv.put(o, h = new HashSet<>());
            }
            h.add(g);
        }
        return rv;
    }
    /** will transfer all annotated isoforms between all relevant genes 
     * if transitively until now new annotation is there
     */
    



    HashSet<Gene> getNoParalogRelevans()
    {
        if (noparalog_relevant!=null)
            return noparalog_relevant;

        noparalog_relevant = new HashSet<>();

        HashMap<Gene, Double> g2score = new HashMap<Gene, Double>();
        g2score.put(start, 1.0);

        boolean gotnew = true;
        while (gotnew)
        {
            gotnew = false;
            HashMap<Gene, Double> update = new HashMap<Gene, Double>();

            for (Map.Entry<Gene,Double> e: g2score.entrySet())
            {
                Gene g = e.getKey();
                double base = e.getValue();

                for(OrthologInfo oi : orthologes.get(g))
                {
                    double sc = Math.min(oi.getDNASeqId(), base);

                    Gene tg = oi.getTargetGene();

                    Double old1 = g2score.get(tg);
                    Double old2 = update.get(tg);

                    if (old1==null || old1<sc || (old2!=null && old2<sc))
                    {
                        update.put(tg,sc);
                        gotnew = true;
                    }
                }
            }

            g2score.putAll(update);
        }
        HashMap<Organism, Gene> o2g = new HashMap<>();

        for (Map.Entry<Gene,Double> e: g2score.entrySet())
        {
            Gene old = o2g.get(e.getKey().organism);

            if (old!=null && g2score.get(old) >= e.getValue())
            {
                continue;
            }

            o2g.put(e.getKey().organism, e.getKey());
        }

        noparalog_relevant.addAll(o2g.values());
        return noparalog_relevant;
    }

    public Vector<OrthologInfo> getSortedInfos(boolean noparalogs)
    {
        HashSet<Gene> npg = getNoParalogRelevans();

        Vector<OrthologInfo> rv = new Vector<OrthologInfo>();
        for (Vector<OrthologInfo> v : orthologes.values())
        {
            if (!noparalogs)
            {
                rv.addAll(v);
                continue;
            }
            for (OrthologInfo oi : v)
            {
                if (!npg.contains(oi.getSrcGene()) || !npg.contains(oi.getTargetGene()))
                    continue;
                rv.add(oi);
            }
        }
        ObjectGetter.sortBy(rv, "getDNASeqId()", false );
        return rv;
    }
    public Collection<Gene> getRelevantGenes()
    {
        return relevant_genes;
    }
    public Vector<OrthologInfo> getOrthologs(Gene g)
    {
        return orthologes.get(g);
    }
    public Element toXML(){
        Element root = new Element("orthgroup");
        root.setAttribute("start", start.id);
        for (Gene g : relevant_genes)
        {
            Element ge = new Element("gene");
            ge.setAttribute("id", g.id);
            ge.setAttribute("tax", ""+g.organism.getTaxId());
            root.addContent(ge);
        }
        for (Vector<OrthologInfo> v : orthologes.values())
        {
            for (OrthologInfo oi : v)
            {
                root.addContent(oi.toXML());
            }

        }

        if(transfers!=null)
        {
            Element t = new Element("transfers");
            root.addContent(t);
            for (Map.Entry<Gene, Vector<Pair<String, RegionVector>>> e : transfers.entrySet())
            {
                Element g = new Element(e.getKey().id);
                t.addContent(g);
                for (Pair<String,RegionVector> ee: e.getValue())
                {
                    Element rvE = ee.getSecond().getXML();
                    rvE.setAttribute("id", ee.getFirst());
                    g.addContent(rvE);
                }
            }                       
            Element ti = new Element("transferinfos");
            root.addContent(ti);


            for(int i=0; i<transfer_infos.size(); i++)
            {
                OrthologInfo oi = transfer_infos.get(i);
                Element e = oi.toXML();
                e.setAttribute("regvec_idx", ""+i);
                ti.addContent(e);
            }


        }
        return root;
    }
    public void writeXML(PrintWriter pw)
    {
        Element root = toXML();
        pw.println(XMLUtils.prettyXML(root));
        pw.flush();
    }

    protected Pair<String,RegionVector> getSimilarIsoform(RegionVector totest, Collection<Pair<String,RegionVector>> c, int tolerance)
    {


        for (Pair<String,RegionVector> p : c)
        {
            if (totest.getNumRegions()!=p.getSecond().getNumRegions())
                continue;

            RegionVector old = p.getSecond();
            boolean same=true;
            for (int i=0; same && i<totest.getNumRegions(); i++)
            {
                Region1D r1 = old.getRegion(i);
                Region1D r2 = totest.getRegion(i);

                if (Math.abs(r1.getX1()-r2.getX1())<=tolerance && Math.abs(r1.getX2()-r2.getX2())<=tolerance)
                    continue;

                same=false;
            }
            if(same)
                return p;
        }
        return null;
    }


    public OrthologInfo getTransferInfo(RegionVector rv)
    {
        if(rv.getIndex()<0 || rv.getIndex()>=transfer_infos.size())
            return null;

        return transfer_infos.get(rv.getIndex());
    }

    public Vector<Pair<String, RegionVector>> getTransfers(Gene g)
    {
        if(transfers==null)
            return null;
        return transfers.get(g);
    }

  
    public Gene getStartGene()
    {
        return start;
    }
    void initFromStreamingXML(Organism org, Reader r, File xml, boolean onlydirectmappings)
    {
        HashMap<String, Gene> genemap = new HashMap<String, Gene>();

        try
        {
            int northread=0;
            String startgn = null;
            XMLStreamReader xsr = XMLUtils.infactory.createXMLStreamReader(r); //new FileReader(xml));
            while (xsr.hasNext())
            {
                int event = xsr.next();

                if (event!=XMLStreamConstants.START_ELEMENT)
                    continue;

                String en = xsr.getLocalName();
                if (en.equals("orthgroup"))
                {
                    startgn = XMLUtils.getAttrib(xsr, "start", null, true);
                    if (org!=null)
                    {
                        Gene g = org.getAnnot().getGene(startgn);
                        genemap.put(g.id, g);
                        relevant_genes.add(g);
                        
                    }
                    continue;
                }
                if (en.equals("gene"))
                {
                    Gene g = Organism.getOrganism(XMLUtils.getIntAttrib(xsr, "tax", null)).getAnnot().getGene(XMLUtils.getAttrib(xsr, "id", null, true));
                    genemap.put(g.id, g);
                    relevant_genes.add(g);
                    continue;
                }
                if (en.equals(OrthologInfo.ORTHO_ELEM))
                {

                    if (onlydirectmappings && !XMLUtils.getAttrib(xsr, "from", null, true).equals(startgn))
                    {
                        continue;
                    }
                    OrthologInfo  oi = new OrthologInfo(xsr, genemap);

                    Vector<OrthologInfo> v = orthologes.get(oi.src_gene);
                    if (v==null)
                    {
                        orthologes.put(oi.src_gene, v = new Vector<>());
                    }
                    v.add(oi);
                    northread++;

                    continue;
                }
                //todo transfers
            }
            log.debug("got %d relevant genes and %d orth mappings", relevant_genes.size(), northread);
            xsr.close();
            start = genemap.get(startgn);
        }
        catch(XMLStreamException xse)
        {
            throw new RuntimeException("xml error: "+xse.getMessage()+" while reading "+xml.getAbsolutePath());
        }

    }

    void initFromDOMXML(File xml)
    {
        HashMap<String, Gene> genemap = new HashMap<String, Gene>();
        Element root = XMLUtils.getRootElement(xml);

        for (Element ge : XMLUtils.getChildren(root, "gene"))
        {
            Gene g = Organism.getOrganism(XMLUtils.getIntAttrib(ge, "tax", null)).getAnnot().getGene(XMLUtils.getAttrib(ge, "id", null, true));
            genemap.put(g.id, g);
            relevant_genes.add(g);
        }

        start = genemap.get(XMLUtils.getAttrib(root, "start", null, true));

        int northread=0;
        for (Element orth : XMLUtils.getChildren(root, OrthologInfo.ORTHO_ELEM))
        {
            OrthologInfo  oi = new OrthologInfo(orth, genemap);

            Vector<OrthologInfo> v = orthologes.get(oi.src_gene);
            if (v==null)
            {
                orthologes.put(oi.src_gene, v = new Vector<>());
            }
            v.add(oi);
            northread++;
        }
        log.info("got %d relevant genes and %d orth mappings", relevant_genes.size(), northread);

        Element transfers = root.getChild("transfers");
        if (transfers==null)
        {
            return;
        }

        this.transfers = new HashMap<>();

        for (Element e: XMLUtils.getChildren(transfers))
        {
            Gene g = genemap.get(e.getName());
            Vector<Pair<String, RegionVector>> gtrans = new Vector<>();
            this.transfers.put(g, gtrans);
            for (Element re : XMLUtils.getChildren(e, RegionVector.ELEM_NAME))
            {
                RegionVector rv = new RegionVector(re);
                String id = XMLUtils.getAttrib(re, "id" , null, true);
                gtrans.add(new Pair<String, RegionVector>(id, rv));
            }
        }
        Element transinfos = root.getChild("transferinfos");

        for (Element e : XMLUtils.getChildren(transinfos))
        {
            transfer_infos.add(new OrthologInfo(e,genemap));
        }


    }
    protected OrthologGeneGroup()
    {
        
    }
    
    public OrthologGeneGroup(File xml, Organism org)
    {
        this(xml, org, false);
    }
    public OrthologGeneGroup(Element element){

    }
    public OrthologGeneGroup(File xml)
    {
        this(xml, null, false);
    }
    public OrthologGeneGroup(File xml, boolean onlydirect)
    {
        this(xml, null, onlydirect);
    }
    public OrthologGeneGroup(File xml, Organism org, boolean onlydirect)
    {
        long t1 = System.currentTimeMillis();
        try
        {
            if ( xml.getName().endsWith(".gz")){
                //GZIPInputStream gzip = new GZIPInputStream(new FileInputStream(xml));
                //BufferedReader in = new BufferedReader(gzip);

                GZIPInputStream gzip = new GZIPInputStream(new FileInputStream(xml));
                BufferedReader br = new BufferedReader(new InputStreamReader(gzip));
                StringBuffer content = new StringBuffer();
                String line = null;
                while ( (line = br.readLine())!=null){
                    content.append(line);
                }

                gzip.close();
                initFromStreamingXML(org, new StringReader(content.toString()), xml, onlydirect);

                return;
            }
            BufferedRandomAccessFile barf = new BufferedRandomAccessFile(xml, "r", (int)xml.length());
            String buffer =  barf.readString((int)xml.length());
            barf.close();
            initFromStreamingXML(org, new StringReader(buffer), xml, onlydirect);
        }
        catch(IOException ie)
        {
            throw new RuntimeException("i/o error: "+ie.getMessage()+" while reading "+xml.getAbsolutePath());
        }


    }



    public OrthologGeneGroup(File orthdir,  Gene gene, Vector<Gene> paralogs)
    {
        this(new MultiSpeciesGeneMapping(orthdir), gene, paralogs);
    }
    public interface OrthologProteinPairSelector{
        OrthologInfo getOrthologInfo(Gene src,Gene target,ProteinAligner aligner);
    }
    public class DefaultOrthologProteinPairSelector implements OrthologProteinPairSelector{

        @Override
        public OrthologInfo getOrthologInfo(Gene src, Gene target, ProteinAligner aligner) {
            return new OrthologInfo(src,target,aligner);
        }

    }

    public OrthologGeneGroup(MultiSpeciesGeneMapping gene_mappings, Gene gene,  Vector<Gene> paralogs)
    {
        this(gene_mappings, gene,(src,target,aligner) -> { return new OrthologInfo(src,target,aligner); }, paralogs);

    }


    private void addParalogs(Gene gene){
        if (paralogs!=null && paralogs.size()>0)
        {
            Organism o = start.organism;

            int added=0;
            for (Gene g : paralogs)
            {
                OrthologInfo oi = new OrthologInfo(start, g, aligner);
                if (oi.seqid < min_seq_id)
                {
                    log.warn("skip opossible paralog %s\n", oi);
                    continue;
                }
                gene_mappings.addParalogMapping(o, gene.id, g.id);
                added++;
            }
            log.debug("added %d/%d pararalogs\n", added, paralogs.size());


        }
    }


    /**
     * @param gene start_gene 
     * @param gene_mappings the gene mapping
     * @return returns all gene transitively reached via the gene mappings (also those <min_seq_id)
     */
    public static Vector<Gene> getGenes(Gene gene, MultiSpeciesGeneMapping gene_mappings){

        
        
        Vector<Organism> orgs = new Vector<Organism>(gene_mappings.getOrganisms());

        Vector<HashSet<String>> relevants = new Vector<HashSet<String>>();
        int src_idx = gene_mappings.getOrgIdx(gene.organism);
        for (int i=0; i<orgs.size(); i++)
        {
            relevants.add(new HashSet<String>());
        }
        relevants.get(src_idx).add(gene.id);
   //     relevants.get(src_idx).addAll(gene_mappings.getMappings(gene.organism, gene.organism, gene.id));

        int old_relevants=0;
        int new_relevants=1;
        int iter=0;
        while (new_relevants!=old_relevants)
        {
            old_relevants=new_relevants;

            for (int i=0; i<orgs.size(); i++)
            {
                HashSet<String> rs = relevants.get(i);
                if (rs.size()==0)
                    continue;

                // Organism o = v.get(i);
                for (int j=0; j<orgs.size(); j++)
                {
                    if (i==j)
                        continue;


                    HashSet<String> target = relevants.get(j);
                    for (String r: relevants.get(i))
                    {
                        target.addAll(gene_mappings.getMappings(i, j, r));
                    }
                }

            }

            new_relevants = 0;
            for (HashSet<String> rs: relevants)
            {
                new_relevants+= rs.size();
            }
            iter++;
        }

        Vector<Gene> genes = new Vector<Gene>();
        for (int i=0; i<orgs.size(); i++)
        {
            Organism o = orgs.get(i);
            HashSet<String> rs =relevants.get(i);
            for (String s: rs)
            {
                if (!o.getAnnot().containsGeneId(s))
                        continue;
                
                Gene g = o.getAnnot().getGene(s);
                
                if (g == null)
                {
                    continue;
                }
                genes.add(g);
            }

        }
        return genes;
    }
    public OrthologGeneGroup(MultiSpeciesGeneMapping gene_mappings, Gene gene, OrthologProteinPairSelector selector, Vector<Gene> paralogs)
    {
        this(gene_mappings,gene,selector,paralogs,MIN_SEQ_ID);
    }
    public OrthologGeneGroup(MultiSpeciesGeneMapping gene_mappings, Gene gene, OrthologProteinPairSelector selector, Vector<Gene> paralogs, double min_seq_id)
    {
        this.min_seq_id = min_seq_id;
        
        start = gene;
        this.gene_mappings = gene_mappings;

        addParalogs(gene);

        //collect the info for the gene
        Vector<Gene> genes = getGenes(gene,gene_mappings);
        genes.sort((Gene g1, Gene g2) -> g1.id.compareTo(g2.id));
        
        HashMap<Gene, Vector<OrthologInfo>> gene2orth = new HashMap<Gene, Vector<OrthologInfo>>();

        for (int i=0; i<genes.size(); i++)
        {
            Gene g1 = genes.get(i);
            for (int j=i+1; j<genes.size(); j++)
            {
                Gene g2 = genes.get(j);

                if ( g1 == gene){

                }
                OrthologInfo oi = selector.getOrthologInfo(g1, g2,aligner);//= new OrthologInfo(g1,g2, aligner);
                if ( oi == null) {
                    continue;
                }
                if (oi.seqid < min_seq_id)
                {
                    if ( log.isTraceEnabled()){
                        log.trace("skipped: %s\n", oi);
                        log.trace("%s:\n%s\n", oi.dst_id, oi.p2seq);
                        log.trace("\n"+oi.ali.getFormattedBy(80, oi.src_id, oi.p1seq, oi.dst_id, oi.p2seq));
                    }
                    continue;
                }
            //    System.out.println(g1.id + " " + g2.id  + " " + oi.seqid);
                Vector<OrthologInfo>  ov = gene2orth.get(oi.src_gene);
                if (ov == null)
                {
                    gene2orth.put(oi.src_gene, ov = new Vector<OrthologInfo>());
                }
                Vector<OrthologInfo>  vr = gene2orth.get(oi.target_gene);
                if (vr == null)
                {
                    gene2orth.put(oi.target_gene, vr = new Vector<OrthologInfo>());
                }

                ov.add(oi);
                vr.add(oi.reverse());


            }
        }
        if (gene2orth.get(start)==null)
        {
            return;
        }

      //collect the relevant mappings
        LinkedList<OrthologInfo> q = new LinkedList<OrthologInfo>();
        q.addAll(gene2orth.get(start));
        HashSet<Gene> sources = new HashSet<Gene>();
        sources.add(start);
        relevant_genes.add(start);
        while (!q.isEmpty())
        {
            OrthologInfo oi = q.poll();
            relevant_genes.add(oi.target_gene);
            Vector<OrthologInfo> ov = orthologes.get(oi.src_gene);
            if(ov==null)
            {
                orthologes.put(oi.src_gene, ov = new Vector<>());
            }
            ov.add(oi);

            log.debug(oi);

            if (sources.contains(oi.target_gene))
                continue;

            sources.add(oi.target_gene);
            Vector<OrthologInfo> v2 = gene2orth.get(oi.target_gene);
            if (v2==null)
                continue;

            q.addAll(v2);

        }

    }
    public static void main(String[] args) throws Exception
    {
       // ortholog.runnable.OrthologGeneGroup.main(args);
    }
   
}
