package ortholog;

import org.jdom.*;
import genome.*;
import lmu.utils.*;

public class PartialCodingExon 
{
    public static final String ELEM_NAME="pcexon";
    public String exon_name;
    public CodingExon exon;
    @Deprecated
    public Pair<Integer,Integer> region;
    public int start_in_exon;
    public int end_in_exon;
    
    
    double coverage=0.0;

    public int length=0;
    public PartialCodingExon(CodingExon exon, int start_in_exon, int end_in_exon)
    {
        this.exon=exon; 
        exon_name=exon.exon.id;
        
        region = new Pair<Integer,Integer>(start_in_exon, end_in_exon);
        length=end_in_exon-start_in_exon;
        this.start_in_exon = start_in_exon;
        this.end_in_exon = end_in_exon;
        coverage=(length/(double)exon.length);
        if ( coverage > 1 || this.start_in_exon > exon.length || this.end_in_exon > exon.length) throw new RuntimeException("Stop/start/length outside of exon boundaries" +start_in_exon + " " +end_in_exon + " " + exon.length );
    }


    public int getLength()
    {
        return length;
    }
    public int getExonLength()
    {
        return exon.length;
    }
    public double getCoverage()
    {
        return coverage;
    }

    public int getStartInGene() {
    	return start_in_exon+this.exon.start_in_gene;
    }
    public int getEndInGene() {
    	return end_in_exon+this.exon.start_in_gene;
    }
    public PartialCodingExon(Element e)
    {
        region = Pair.parseIntPair(XMLUtils.getAttrib(e,"region", null, true));
        exon_name=XMLUtils.getAttrib(e,"exon", null, true);
        length=region.getSecond()-region.getFirst();
        coverage=XMLUtils.getDoubleAttrib(e, "coverage", null);
        this.start_in_exon = region.getFirst();
        this.end_in_exon = region.getSecond();
        
    }
    
    public void setExon(Protein p)
    {
        for (int i=0; i<p.getNumExons(); i++)
        {
            CodingExon ce = p.getExon(i);
            if (!ce.exon.id.equals(exon_name))
                continue;
            
            exon=ce;
        }
    }
    public Element getXML()
    {
        Element el = new Element(ELEM_NAME);
        el.setAttribute("region", ""+region);
        el.setAttribute("exon", exon_name);
        el.setAttribute("coverage", ""+coverage);
        return el;
    }

    public String toString()
    {
        return String.format("%s %d-%d (cov: %2.2f%%)",exon.exon.id, region.getFirst(), region.getSecond(), 100.0*coverage);
    }
}
