package ortholog.intron;

import java.io.*;
import java.util.*;
import lmu.utils.*;
import protein.algo.*;
import algo.alignment.*;
import ortholog.*;

import genome.*;

public class ExonMapping
{
    static Logger log = null;

    synchronized void initLogger()
    {
        if (log!=null)
            return;
        
        log = LogConfig.getLogger();
    }
    RegionVector protein;

    TreeMap<Region1D, HashSet<Region1D> > exon2_aliparts = new TreeMap<>();
    TreeMap<Region1D, Region1D> alipart2_exon = new TreeMap<>();

    TreeMap<Region1D, Vector<RegionAlignment>> exon2_alis = new TreeMap<Region1D, Vector<RegionAlignment>>();

    public ExonMapping(RegionVector rv, GenomicProteinAlignment gpa, boolean src)
    {
        initLogger();
        this.protein = rv;

        for (Region1D r: protein.getRegions())
        {
            log.trace("next exon (src: %s): %s\n", src, r);
            exon2_aliparts.put(r, new HashSet<Region1D>());
            exon2_alis.put(r, new Vector<RegionAlignment>());
        }

        for (RegionAlignment ra : gpa.cdnaAli)
        {
            Region1D alipart = (src) ? ra.src.getRegions().get(0) : ra.dst.getRegions().get(0);

            log.trace("next alipart: (src: %s): %s\n", src, alipart);
            for (Region1D exon : protein.getRegions())
            {
                log.trace("\tcheck in exon: "+exon);
                if (!(exon.getX1()<=alipart.getX1() && exon.getX2()>= alipart.getX2()))
                    continue;
                
                log.trace("\tFOUND: "+exon);
                exon2_aliparts.get(exon).add(alipart);
                alipart2_exon.put(alipart, exon);
                
                exon2_alis.get(exon).add(ra);
                break;
            }

        }
    }

    public Vector<RegionAlignment> getAlis(Region1D exon)
    {
        return exon2_alis.get(exon);
    }
    public HashSet<Region1D> getMultiParts(Region1D exon)
    {
        return exon2_aliparts.get(exon);
    }

    public Region1D getExon(Region1D alipart)
    {
        return alipart2_exon.get(alipart);
    }
    public int getExonAliParts(Region1D alipart)
    {
        Region1D exon = alipart2_exon.get(alipart);

        if (exon==null)
            return 0;
        
        HashSet<Region1D> r = exon2_aliparts.get(exon);

        return (r==null) ? 0 : r.size();

    }
}


