package genome;

import lmu.utils.Pair;
/** represents the (part) of exon in a protein coding transcript
 * which codes for the final protein sequence
 */
public class CodingExon
{
   public Exon exon;
   public Transcript parentTranscript;
   public Protein parentProtein;
   public int start_position_in_transcript;
   public int start_position_in_coding;
   public int start_position_in_protein;

   public int frame;

   /** 1 -based, end is inclusive ! */
   public int abs_start;
   public int abs_end;


   //* 0 based, end is inclusive! */
   public int start_in_gene;
   public int end_in_gene;



   public int start_in_exon;
   public int end_in_exon;
   public int length;
   public int peptide_length;

   public void updateRelativePositions()
   {
       final Gene g = parentTranscript.parentGene;
       if (g.strand)
       {
           start_in_gene = abs_start-g.start;
           end_in_gene = abs_end-g.start;
       }
       else
       {
           start_in_gene = g.end-abs_end;
           end_in_gene = g.end-abs_start;
       }
   }
   
        
   public CodingExon(Pair<Integer,Integer> abs_region, Protein protein, Transcript transcript, Exon exon, int tr_pos_start, int coding_pos_start)
   {
       this.exon=exon; this.parentTranscript=transcript; this.parentProtein=protein;
       abs_start=abs_region.getFirst();
       abs_end = abs_region.getSecond();
       //correct length to be within the gene
       while (abs_end > transcript.parentGene.end)
       {
           abs_end -= 3;
       }

       length = abs_end-abs_start+1;
       start_position_in_transcript=tr_pos_start;
       start_position_in_coding = coding_pos_start;

       frame=coding_pos_start%3;
       frame=(3-frame)%3;
       start_position_in_protein=(int)(start_position_in_coding/3.0);

       if (frame!=0)
       {
           start_position_in_protein++;
       }
       peptide_length=(int)((length-frame)/3.0);

       start_in_exon=abs_start-exon.abs_start;
       end_in_exon=abs_end-exon.abs_end;

       updateRelativePositions();
   }

   
   public String toString()
   {
       return String.format("%s frame: %d %d-%d (l: %d peptide length: %d) in coding: %d-%d in prot %d-%d", 
                                    exon.id, frame,abs_start, abs_end, length,peptide_length,
                                  start_position_in_coding, start_position_in_coding+length,
                                  start_position_in_protein, start_position_in_protein+peptide_length
           );
                            

   }
}
