package genome;

import lmu.utils.*;
 
public class Codon
{
    int idx;
    char aa;
    char codon[] = new char[3];
    public Region1D exon;
    public Region1D next_exon;

    String codon_str = null;
    
    public int getIdx()
    {
        return idx;
    }
    public char getAA()
    {
        return aa;
    }
    
    public boolean isAdjacent(Codon c, int maxdiff)
    {
        if (exon!=c.exon)
            return false;
        
        return (Math.abs(idx-c.idx)<=maxdiff);
    }
    public String getCodonStr()
    {
        if (codon_str!=null)
            return codon_str;
        
        StringBuffer sb = new StringBuffer();
        sb.append(codon);
        return codon_str = sb.toString();
    }
    public char[] getCodon()
    {
        return codon;
    }
    public RegionVector getRegionVector()
    {
        RegionVector rv = new RegionVector();
        rv.addRegion(exon.getX1()+exon_offset, (isSplit())?exon.getX2():exon.getX1()+exon_offset+3);
        if (!isSplit())
        {
            return rv;
        }
        int leftover = exon.getX2()-(exon.getX1()+exon_offset);
        rv.addRegion(next_exon.getX1(), next_exon.getX1()+3-leftover);
        return rv;
    }

    public Pair<Integer,Integer> getFirstRegion()
    {
        return new Pair<Integer, Integer>(exon.getX1()+exon_offset, (isSplit())?exon.getX2():exon.getX1()+exon_offset+3);
    }
    public Pair<Integer,Integer> getSecondRegion()
    {
        if (!isSplit())
            return null;
        
        int leftover = exon.getX2()-(exon.getX1()+exon_offset);
        return new Pair<Integer, Integer>(next_exon.getX1(), next_exon.getX1()+3-leftover);
    }

    public int exon_offset;

    public Codon(int idx, Region1D exon, int pos)
    {
        this.idx = idx; this.exon=exon; this.exon_offset=pos;
    }
    public boolean isSplit()
    {
        return next_exon!=null;
    }
    public String toString()
    {
        if (isSplit())
        {
            int leftover = exon.getX2()-(exon.getX1()+exon_offset);

            return String.format("%d-%d, %d-%d: %c %c%c%c(%c) (e1: %d-%d, e2: %d-%d)",
                                 exon.getX1()+exon_offset, exon.getX2(),
                                 next_exon.getX1(), next_exon.getX1()+leftover,
                                  aa, codon[0],codon[1],codon[2], protein.aminoacid.AminoAcidType.get(codon,0),
                             exon.getX1(), exon.getX2(), next_exon.getX1(), next_exon.getX2());

        }

        return String.format("%d-%d:%c: %c%c%c(%c) (e: %d-%d)", exon.getX1()+exon_offset, exon.getX1()+exon_offset+3,
                              aa, codon[0],codon[1],codon[2], protein.aminoacid.AminoAcidType.get(codon,0),
                             exon.getX1(), exon.getX2());

    }
}


