package genome;

import lmu.utils.*;

import java.util.*;

import algo.alignment.*;

import org.jdom.*;

import ortholog.intron.ExonMapping;

import javax.xml.stream.*;

public class GenomicProteinAlignment
{
	Logger log = LogConfig.getLogger();

    public final String ELEM_NAME = "genomic_protali";
    public double pseqid=0.0;
    public double dnaseqid=0.0;
    private Double geneStructureId = null;
    
    
    int num_aa_id=0;
    int num_dna_id=0;

    
    RegionVector  p1;
    RegionVector  p2;

    String cdna1;
    String cdna2;

    String protseq1;
    String protseq2;

    
    public Alignment protAli;
    //Alignment cdnaAli;

    public RegionVector getSrc()
    {
        return p1;
    }

    public RegionVector getDst()
    {
        return p2;
    }

    public int getNumSrcExons()
    {
        return p1.getNumRegions();
    }
    public int getNumDstExons()
    {
        return p2.getNumRegions();
    }
    public int getNumAlignedRegions()
    {
        return cdnaAli.size();
    }

    public String getSrcSequence(){
    	return protseq1;
    }
    public String getDstSequence(){
    	return protseq2;
    }
    
    
    public Vector<RegionAlignment> cdnaAli = new Vector<RegionAlignment>();

    public Vector<Pair<String,String>> protRegions = new Vector<Pair<String,String>>();
    public Vector<Alignment> regionProtAlis = new Vector<Alignment>();

    public Element getXML()
    {
        return getXML(ELEM_NAME);
    }

    public Element getXML(String elem)
    {
        Element e = new Element(elem);
        e.setAttribute("num_aa_id", ""+num_aa_id);
        e.setAttribute("num_dna_id", ""+num_dna_id);
        e.setAttribute("pseqid", ""+pseqid);
        e.setAttribute("dnaseqid", ""+dnaseqid);

        e.setAttribute("protseq1", protseq1);
        e.setAttribute("protseq2", protseq2);

        e.setAttribute("cdna1", cdna1);
        e.setAttribute("cdna2", cdna2);

        e.addContent(p1.getXML("p1"));
        e.addContent(p2.getXML("p2"));
        e.addContent(protAli.getXML("protali"));

        for(RegionAlignment ra : cdnaAli)
        {
            e.addContent(ra.getXML());
        }


        return e;
    }
    public GenomicProteinAlignment(XMLStreamReader xsr) throws XMLStreamException
    {
        num_aa_id = XMLUtils.getIntAttrib(xsr, "num_aa_id", null);
        num_dna_id = XMLUtils.getIntAttrib(xsr, "num_dna_id", null);
        pseqid = XMLUtils.getDoubleAttrib(xsr, "pseqid", null);
        dnaseqid = XMLUtils.getDoubleAttrib(xsr, "dnaseqid", null);
        
        protseq1 = XMLUtils.getAttrib(xsr, "protseq1", null, true);
        protseq2 = XMLUtils.getAttrib(xsr, "protseq2", null, true);

        cdna1 = XMLUtils.getAttrib(xsr, "cdna1", null, true);
        cdna2 = XMLUtils.getAttrib(xsr, "cdna2", null, true);


        final String en = xsr.getLocalName();
        while (xsr.hasNext())
        {
            int event = xsr.next();
            if (event == XMLStreamConstants.END_ELEMENT && en.equals(xsr.getLocalName()))
            {
                return;
            }
            if (event != XMLStreamConstants.START_ELEMENT)
            {
                continue;
            }
            final String ln = xsr.getLocalName();
            if (ln.equals("p1"))
            {
                p1 = new RegionVector(xsr);
                continue;
            }
            if (ln.equals("p2"))
            {
                p2 = new RegionVector(xsr);
                continue;
            }
            if (ln.equals("protali"))
            {
                protAli = Alignment.parseAliFromXML(xsr);
                continue;
            }
            if (ln.equals(RegionAlignment.RA_ELEM_NAME))
            {
                cdnaAli.add(new RegionAlignment(xsr));
                continue;
            }
        }
        
    }
    public GenomicProteinAlignment(Element e)
    {
        num_aa_id = XMLUtils.getIntAttrib(e, "num_aa_id", null);
        num_dna_id = XMLUtils.getIntAttrib(e, "num_dna_id", null);
        pseqid = XMLUtils.getDoubleAttrib(e, "pseqid", null);
        dnaseqid = XMLUtils.getDoubleAttrib(e, "dnaseqid", null);
        
        protseq1 = XMLUtils.getAttrib(e, "protseq1", null, true);
        protseq2 = XMLUtils.getAttrib(e, "protseq2", null, true);

        cdna1 = XMLUtils.getAttrib(e, "cdna1", null, true);
        cdna2 = XMLUtils.getAttrib(e, "cdna2", null, true);

        p1 = new RegionVector(e.getChild("p1"));
        p2 = new RegionVector(e.getChild("p2"));

        protAli = Alignment.parseAliFromXML(e.getChild("protali"));

        for (Element rae : XMLUtils.getChildren(e, RegionAlignment.RA_ELEM_NAME))
        {
            cdnaAli.add(new RegionAlignment(rae));
        }
        

    }

    private List<Region1D> getDstRegions(Vector<RegionAlignment> ras){
    	List<Region1D> dst = new ArrayList<Region1D>();
    	
    	ObjectGetter.apply(ras, (RegionAlignment ra) -> {
    		for(Region1D r : ra.dst.getRegions()){
    			if ( r.getLength() > MIN_MATCH_LENGTH) dst.add(r);
    		}
    	});
    	
    	return dst;
    }
    public Pair<Integer,Integer> getMatchedStartEndExons(boolean src){
    	ExonMapping ex = this.getExonMapping(src) ;
    	RegionVector exons = src==true?this.getSrc():this.getDst();
    	
    	return getMatchedStartEndExons(ex,exons);
    }
    public Pair<Integer,Integer> getMatchedStartEndExons(ExonMapping ex, RegionVector exons){
    	
    	int start = exons.size();
    	int end = exons.size();

    	for(int i =0;i<end;i++ ){
        	Region1D exon = exons.getRegions().get(i);
        	if ( exon.getLength() < MIN_MATCH_LENGTH) continue;
        	
        	Vector<RegionAlignment> ras = ex.getAlis(exon);
        	List<Region1D> dst = getDstRegions(ras);
        	
        	if ( dst.size() > 0 ) {
        		start = i;
        		break;
        	}
        }
        for(int i =exons.size()-1;i>=start;i-- ){
        	
        	Region1D exon = exons.getRegions().get(i);
        	if ( exon.getLength() < MIN_MATCH_LENGTH) continue;
        	
        	Vector<RegionAlignment> ras = ex.getAlis(exon);
        
        	List<Region1D> dst = getDstRegions(ras);
        	
        	if ( dst.size() >0) {
        		end = i+1;
        		break;
        	}
        }
    	
    	return new Pair<Integer,Integer>(start,end);
    }
    
    public ExonMapping getExonMapping(boolean src){
    	ExonMapping ex = new ExonMapping(src==true?this.getSrc():this.getDst(),this,src);
        return ex;
    }
    
    public HashMap<Region1D,Set<Region1D>> getExonExonMapping(boolean src){
    	HashMap<Region1D,Set<Region1D>> map = new HashMap<Region1D,Set<Region1D>>();
        
    	ExonMapping exSrc = this.getExonMapping(src);
    	ExonMapping exDst = this.getExonMapping(!src);
    	
    	Pair<Integer,Integer> start_end = getMatchedStartEndExons(src);
    	  
        for(int i =start_end.first;i<start_end.second;i++ ){
        	Region1D exon = src?this.getSrc().getRegions().get(i):this.getDst().getRegions().get(i);
        	Vector<RegionAlignment> ras = exSrc.getAlis(exon);
        	
        	Set<Region1D> mapped = new HashSet<Region1D>();
        	for(RegionAlignment ra : ras){
        		
        		if ( ra.src.getLength()<MIN_MATCH_LENGTH) continue;
        		
        		for(Region1D dst : ra.dst.getRegions()){

        			if ( dst.getLength() < MIN_MATCH_LENGTH) continue;
        			Region1D dstExon = exDst.getExon(dst);
        			mapped.add(dstExon);
        		}
        	
        	}
        	map.put(exon,mapped);
        }
    	
    	return map;
    }
    
    private static int MIN_MATCH_LENGTH=5;
    public double getGeneStructureId(){
    	if ( geneStructureId == null){
	        
	        ExonMapping exSrc = getExonMapping(true);
	        ExonMapping exDst = getExonMapping(false);
	        
	        float idents = 0;
	        
	        Pair<Integer,Integer> src_start_end = getMatchedStartEndExons(exSrc,this.getSrc());
	        Pair<Integer,Integer> dst_start_end = getMatchedStartEndExons(exDst,this.getDst());
  
	        for(int i =src_start_end.first;i<src_start_end.second;i++ ){
	        	Region1D exon = this.getSrc().getRegions().get(i);
	        	Vector<RegionAlignment> ras = exSrc.getAlis(exon);
	        	
	        	boolean ident=true;
	        	if ( ras.size() == 0) ident=false;
	        	
	        	boolean foundOne = false;
	        	for(RegionAlignment ra : ras){
	        		
	        		if ( ra.src.getLength()<MIN_MATCH_LENGTH) continue;
	        		
	        		for(Region1D dst : ra.dst.getRegions()){

	        			if ( dst.getLength() < MIN_MATCH_LENGTH) continue;
	        			Region1D dstExon = exDst.getExon(dst);
	        			
	        			for(RegionAlignment srcMapping : exDst.getAlis(dstExon)){
	        				for(Region1D backSrc : srcMapping.src.getRegions()){
	        					if ( backSrc.getLength() < MIN_MATCH_LENGTH) continue;
	        					foundOne = true;
	        					if ( !exSrc.getExon(backSrc).equals(exon)) ident=false;
	        				}
	        			}
	        		}
	        	
	        	}
	        	if ( foundOne == false) ident = false;
	        	if ( ident) idents+=2;
	        }
	        
	        float norm = (src_start_end.second-src_start_end.first)+(dst_start_end.second-dst_start_end.first);
	    	geneStructureId= (double)idents/norm;

    	}
    	if( geneStructureId > 1 || geneStructureId < 0) {
    		
    		throw new RuntimeException(String.format("Bug"));
    	}
    	return geneStructureId;
    }
    
    public GenomicProteinAlignment(Protein p1, Protein p2, Alignment protali,
                                   GenomeSequenceExtractor gex1, GenomeSequenceExtractor gex2, int max_gap)
    {
        this(p1, p2, GenomicUtils.getProteinSeq(p1, gex1, false),GenomicUtils.getProteinSeq(p2, gex2, false), protali, gex1, gex2, max_gap);
    }


    public GenomicProteinAlignment(Protein p1, Protein p2, String pseq1, String pseq2, 
                                   Alignment protali,
                                   GenomeSequenceExtractor gex1, GenomeSequenceExtractor gex2, int max_gap)
    {
        this(p1.getCodingRegionVector(false),
             p2.getCodingRegionVector(false),
             pseq1, pseq2, 
             GenomicUtils.getProteinSeq(p1, gex1, true), GenomicUtils.getProteinSeq(p2, gex2, true), protali, max_gap);
    }


    public GenomicProteinAlignment(RegionVector prot1vec, RegionVector prot2vec, String pseq1, String pseq2, String cdna1, String cdna2, Alignment protali, int max_gap)
                                   
    {
        this.p1 = prot1vec;
        this.p2 = prot2vec;
        this.protseq1 = pseq1;
        this.protseq2 = pseq2;
        this.cdna1 = cdna1;
        this.cdna2 = cdna2;
        
        this.protAli = protali;
        if (protali.alignment == null || protali.alignment.size() ==0 )
        {
            return;
        }
        Vector<Codon> codons1 = GenomicUtils.getCodons(prot1vec.getRegions(), pseq1, cdna1);
        Vector<Codon> codons2 = GenomicUtils.getCodons(prot2vec.getRegions(), pseq2, cdna2);

        Collections.sort(protali.alignment);

        Codon start1 = codons1.get(protali.alignment.get(0).getFirst());
        Codon start2 = codons2.get(protali.alignment.get(0).getSecond());
        
        int alistart=0;
        Codon end1 = start1;
        Codon end2 = start2;

        for (Pair<Integer,Integer> p: protAli.alignment)
        {
            Codon c1 = codons1.get(p.getFirst());
            Codon c2 = codons2.get(p.getSecond());
            if (c1.getAA() == c2.getAA())
            {
                num_aa_id++;
            }
            int base1 = c1.getIdx()*3;
            int base2 = c2.getIdx()*3;

            final char[] cod1 = c1.getCodon();
            final char[] cod2 = c2.getCodon();
            
        
        
            for (int k=0; k<3; k++)
            {
                if(cod1[k]!=cod2[k])
                    continue;
        
                num_dna_id++;
            }
        
        }

        
        

        for (int i=1; i<=protali.alignment.size(); i++)
        {
            Pair<Integer,Integer> p = (protali.alignment.size()<=i) ? null : protali.alignment.get(i);

            Codon c1 = (p==null)?null:codons1.get(p.getFirst());
            Codon c2 = (p==null)?null:codons2.get(p.getSecond());

            if (p!=null && end1.isAdjacent(c1, max_gap) && end2.isAdjacent(c2, max_gap))
            {
                end1 = c1;
                end2 = c2;
                continue;     
            }

            RegionAlignment ra = new RegionAlignment();

            int ra_num_aa_id = 0;
            int ra_num_dna_id = 0;
            ra.ali = new Alignment(pseq1.length(), pseq2.length());
            ra.ali.initAlignment();

            int shift1 = protali.alignment.get(alistart).getFirst();
            int shift2 = protali.alignment.get(alistart).getSecond();

            ra.src_prot_region = new Region1D(shift1, protali.alignment.get(i-1).getFirst()+1, null);
            ra.dst_prot_region = new Region1D(shift2, protali.alignment.get(i-1).getSecond()+1, null);


            for (int j=alistart; j<i; j++)
            {
                Pair<Integer,Integer> ap = protali.alignment.get(j);

                ra.ali.alignment.add(new Pair<Integer,Integer>(ap.getFirst()-shift1, ap.getSecond()-shift2));
                Codon cc1 = codons1.get(ap.getFirst());
                Codon cc2 = codons2.get(ap.getSecond());
                char[] cod1 = cc1.getCodon();
                char[] cod2 = cc2.getCodon();

                if (cc1.getAA()==cc2.getAA())
                {
                    ra_num_aa_id++;
                }

                for (int k=0; k<3; k++)
                {
                    if(cod1[k] != cod2[k])
                        continue;
                    
                    ra_num_dna_id++;
                }
            }
            ra.ali.setLengths(ra.src_prot_region.getLength(), ra.dst_prot_region.getLength());
            int l1 = end1.getIdx()-start1.getIdx()+1;
            int l2 = end2.getIdx()-start2.getIdx()+1;

            double meanl = (l1+l2)*0.5;
            double Dmeanl = 3*meanl;
            
            ra.dna_seqid = ra_num_dna_id/Dmeanl;
            ra.prot_seqid = ra_num_aa_id/meanl;

            ra.src = new RegionVector();
            ra.src.addRegion(new Region1D(start1.getFirstRegion().getFirst(), end1.getFirstRegion().getSecond(),null));
            if (end1.isSplit())
            {
                ra.src.addRegion(new Region1D(end1.getSecondRegion()));
            }
            ra.dst = new RegionVector();
            ra.dst.addRegion(new Region1D(start2.getFirstRegion().getFirst(), end2.getFirstRegion().getSecond(),null));
            if (end2.isSplit())
            {
                ra.dst.addRegion(new Region1D(end2.getSecondRegion()));
            }

            cdnaAli.add(ra);

            /*new ali block found
            System.out.printf("prot: %2.3f%%, dna: %2.3f%%, %d-%d: %d-%d vs %d-%d %d-%d\n", 
                              100.0*ra.prot_seqid, 100.0*ra.dna_seqid, 
                              start1.getIdx(), end1.getIdx(),
                              start1.getFirstRegion().getFirst(),
                              end1.getFirstRegion().getSecond(),

                              start2.getIdx(), end2.getIdx(),
                              start2.getFirstRegion().getFirst(),
                              end2.getFirstRegion().getSecond()
                              );
            */
            start1 = end1 = c1;
            start2 = end2 = c2;
            alistart = i;
        }

        double ml = (0.5*(pseq1.length()+pseq2.length()));

        pseqid = num_aa_id / ml;;
        dnaseqid = num_dna_id / (3.0*ml);


        
        
    }


    public String getCheckAlignmentInfo()
    {
        return getCheckAlignmentInfo(80,false,"","");
    }

 
    public String getCheckAlignmentInfo(int formattedBy,boolean includeExon, String id1, String id2)
    {

        Vector<Pair<Integer,Integer>> cdnali = new Vector<Pair<Integer,Integer>>();

        final String pseq1 = protseq1;
        final String pseq2 = protseq2;
    
        for (Pair<Integer,Integer> p: protAli.alignment)
        {
            int base1 = p.getFirst()*3;
            int base2 = p.getSecond()*3;
            
            for (int i=base1, j=base2, k=0; k<3; i++, j++, k++)
            {
                cdnali.add(new Pair<Integer,Integer>(i, j));
            }
        
        }
        
        
        StringBuffer check1 = new StringBuffer();
        
        int w = 0;
        int currentExon = 0;
        Region1D exon = this.getSrc().get(currentExon);
        StringBuffer checkExon1 = new StringBuffer();
        
        for (int i=0, j=0; i<pseq1.length(); i++, j+=3)
        {
        	if( includeExon)
        	{
	        	for(int k = j ; k < j+3;k++,w++)
	        	{
	        	
	        		if ( w > exon.getLength()-1)
	        		{
	        			w = 0;
	        			currentExon = currentExon+1;
	        			exon = this.getSrc().get(currentExon);
		
	        		}
	        		if ( currentExon%2==0 ) 
	        			checkExon1.append("0");
	        		else
	        			checkExon1.append(1);
	        	}
	            check1.append(pseq1.charAt(i));
	            check1.append(' ');
	            check1.append(' ');
        	}else{
        		check1.append(pseq1.charAt(i));
                check1.append(protein.aminoacid.AminoAcidType.get(cdna1.substring(j, j+3)));
                check1.append('.');
                  	
        	}
            
        }
        
        w = 0;
        currentExon = 0;
        exon = this.getDst().get(currentExon);
        StringBuffer check2 = new StringBuffer();
        StringBuffer checkExon2 = new StringBuffer();
        
        for (int i=0, j=0; i<pseq2.length(); i++, j+=3)
        {
        	if( includeExon)
        	{
        		for(int k = j ; k < j+3;k++,w++)
        		{
        			if ( w > exon.getLength()-1){
        				w = 0;
        				currentExon = currentExon+1;
        				exon = this.getDst().get(currentExon);

        			}
        			if ( currentExon%2==0 ) 
        				checkExon2.append("0");
        			else
        				checkExon2.append(1);
        		}
	            check2.append(pseq2.charAt(i));
	            check2.append(' ');
	            check2.append(' ');
        	}else{
        		check2.append(pseq2.charAt(i));
        		check2.append(protein.aminoacid.AminoAcidType.get(cdna2.substring(j, j+3)));
        		check2.append(':');
        	}

        }
   
        Vector<Pair<String,String>> in1 = new Vector<Pair<String,String>>();
        Vector<Pair<String,String>> in2 = new Vector<Pair<String,String>>();

        this.cdna1 = this.cdna1.substring(0,check1.length());
        this.cdna2 = this.cdna2.substring(0,check2.length());
        
    
        
        if ( includeExon) in1.add(new Pair<String,String>("exons", checkExon1.toString()));
        in1.add(new Pair<String,String>(id1, check1.toString()));
        in1.add(new Pair<String,String>("fromcdna", cdna1));
        
        in2.add(new Pair<String,String>("tocdna", cdna2));
        in2.add(new Pair<String,String>(id2, check2.toString()));
        if ( includeExon)in2.add(new Pair<String,String>("exons", checkExon2.toString()));
        
    
        Alignment ca = new Alignment(cdna1.length(), cdna2.length(), cdnali);

        return ca.getFormattedBy(formattedBy, in1, in2);
    }
}



