package genome;

import lmu.utils.*;
import lmu.utils.tuple.*;
import java.util.*;
import java.util.regex.*;
import java.io.*;

public class  GenomicUtils
{
      
    static Pattern chrP = Pattern.compile(":-\\(\\)\\s+");
    static FileUtils.StringSplitter chrInfoSplitter = FileUtils.getTokenizer(":- ()/,;");

    static public Tuple3<String, Integer,Integer> parseChrRegion(String line)
    {
        String sp[] = chrInfoSplitter.split(line);
        
        if (sp.length != 3)
            return null;
        
        try
        {
            return Tuple3.create(sp[0], (sp[1].equals(".")) ? 0 : Integer.parseInt(sp[1]), (sp[2].equals(".")) ? Integer.MAX_VALUE : Integer.parseInt(sp[2]));
        }
        catch(NumberFormatException nfe)
        {
            
        }
        return null;
    }
    
            
    public static char getStrand(Boolean b)
    {
        return (b == null) ? ' ' : (b) ? '+' : '-';
    }
    

    public static Boolean getStrand(String s)
    {
         return (s == null || s.length() == 0) ? null : s.charAt(0) == '+';
    }

    public static final String CANONICAL_SPLICE_MOTIF = "GTAG";

    static public Vector<Codon> getCodons(Vector<Region1D> genomic_coding_regions, String pseq, String cdna_seq)
    {
        //System.out.printf("in pseq:\n%s\ncdnaseq:\n%s\n",      
        //                  pseq, cdna_seq);
        Vector<Codon> codons= new Vector<Codon>();

        int idx=0;
        int leftover=0;
        Region1D last = null;
        Codon splitC = null;
        int ppos = 0;
        int tr_offset=0;
        int r_offset=0;
        for (Region1D r: genomic_coding_regions)
        {
            final int REND = r.getX2();
            //System.out.println("next exon: "+r);
            r_offset = 0;
            if (splitC!=null)
            {
                splitC.next_exon=r;
                r_offset=3-leftover;
                //System.out.printf("will complete split codong with %d leftover\n", leftover);
                for (int j=leftover; j<3; j++, tr_offset++)
                {
                    //System.out.printf("set %d: %c\n", leftover+j, cdna_seq.charAt(tr_offset));
                    splitC.codon[j] = cdna_seq.charAt(tr_offset);
                }

                //System.out.println("complete split codon: "+splitC);
                splitC=null;
            }
            for (int rpos=r.getX1(); rpos+r_offset<REND; r_offset+=3)
            {
                
                if (REND-(rpos+r_offset) >= 3)
                {

                    if (pseq.length()<=ppos)
                    {
                        break;
                    }
                    //got complete codon
                    Codon c = new Codon(idx++, r, r_offset);
                    c.aa=pseq.charAt(ppos++);
                    for (int k=0; k<3; k++, tr_offset++)
                    {
                        c.codon[k]=cdna_seq.charAt(tr_offset);
                    }
                    //System.out.println("add complete codon: "+c);
                    codons.add(c);
                    continue;
                }
                //got split codon
                
                if (pseq.length()<=ppos)
                {
                    break;
                }
                //split exon
                Codon c = new Codon(idx++, r, r_offset);
                c.aa=pseq.charAt(ppos++);
                for (int j=r.getX1()+r_offset, k=0; j<REND; j++,k++, tr_offset++)
                {
                    c.codon[k]=cdna_seq.charAt(tr_offset);
                }
                codons.add(c);
                leftover = (REND-(r.getX1()+r_offset));
                splitC = c; 
            }
            
        }

        return codons;
    }
    
    
    static public Vector<Region1D> convertTranscriptRegions2GenomicRegions(Vector<Region1D> genomic_coding_regions, Vector<Region1D> toconvert)
    {
        Vector<Region1D> rv = new Vector<Region1D>();
        
        int start_exon=0;
        Region1D ce = genomic_coding_regions.get(0);
        
        int tr_start = 0;
        int tr_end = ce.getLength();

        for(Region1D r : toconvert)
        {
            int r_trstart = r.getX1()*3;
            int r_trend = r.getX2()*3;
            while (tr_end < r_trstart)
            {
                start_exon++;
                ce = genomic_coding_regions.get(start_exon);
                tr_start = tr_end;
                tr_end = tr_start+ ce.getLength();
                
            }
            
            
            while (r_trend > tr_end)
            {
                int offset = Math.max(0, r_trstart-tr_start);
                rv.add(new Region1D(ce.getX1()+offset, ce.getX2(), null));

                start_exon++;
                if (start_exon == genomic_coding_regions.size())
                {
                    return ObjectGetter.filter(rv, (_r) -> _r.getLength() > 0);                   
                }
                ce = genomic_coding_regions.get(start_exon);
                tr_start = tr_end;
                tr_end = tr_start+ce.getLength();

            }
            
            int offset = Math.max(0, r_trstart-tr_start);
            rv.add(new Region1D(ce.getX1()+offset, ce.getX1()+(r_trend-tr_start), null));
        }
        
        return ObjectGetter.filter(rv, (r) -> r.getLength() > 0); 
    }


    static void transcribed2RegionVector(Pair<Integer,Integer> tr_region, RegionVector global_regions, RegionVector addto)
    {
        int prelength=0;
        int trlength=0;
        boolean in_region=false;
        for (Region1D r : global_regions.getRegions())
        {
            prelength = trlength;
            trlength += r.getLength();

            if (tr_region.getSecond()<prelength)
                return;
            
            if (tr_region.getFirst()>=trlength)
                continue;
            

            int in_exstart = (tr_region.getFirst()<=prelength)?r.getX1():r.getX1()+tr_region.getFirst()-prelength;

            int in_exend = (trlength < tr_region.getSecond())?r.getX2(): r.getX1()+tr_region.getSecond()-prelength;
            
            addto.addRegion(new Region1D(in_exstart, in_exend, r.getObject()));
            
        }
    }

    static public RegionVector transcribed2RegionVector(Pair<Integer,Integer> tr_region, RegionVector global_regions)
    {
        RegionVector rv = new RegionVector();
        transcribed2RegionVector(tr_region,global_regions, rv);
        return rv;
    }
    static public RegionVector transcribed2RegionVector(Vector<Pair<Integer,Integer>> tr_regions, RegionVector global_regions)
    {
        RegionVector rv = new RegionVector();
        for (Pair<Integer,Integer> p: tr_regions)
        {
            transcribed2RegionVector(p,global_regions, rv);
        }
        
        return rv;
    }
    
    public static int toGlobal(int local, int g_start, int g_end,  boolean strand)
    {
        return (strand) ? g_start+local : g_end+1-local;
    }

    public static int toGlobal(int local, Gene g)
    {
        return (g.strand) ? g.start + local : g.end + 1 - local;
    }

    public static Pair<Integer,Integer> toGlobal(Pair<Integer,Integer> local, int g_start, int g_end,  boolean strand)
    {
        if (strand)
        {
            return new Pair<Integer,Integer>(g_start+local.getFirst(), g_start+local.getSecond());
        }
        return new Pair<Integer,Integer>(g_end+1-local.getSecond(), g_end+1-local.getFirst());
    }

    public static Region1D toGlobal(Region1D local, Gene g)
    {
        return toGlobal(local, g.start, g.end, g.strand);
    }

    public static Region1D toGlobal(Region1D local, int g_start, int g_end,  boolean strand)
    {
        if (strand)
        {
            return new Region1D(g_start+local.getX1(), g_start+local.getX2(), null);
        }
        return new Region1D(g_end+1-local.getX2(), g_end+1-local.getX1(), null);
    }

    public static RegionVector toGlobal(RegionVector local, Gene g)
    {
        return toGlobal(local, g.start, g.end, g.strand);
    }
    public static RegionVector toGlobal(RegionVector local, int g_start, int g_end,  boolean strand)
    {
        RegionVector global = new RegionVector();
        
    
        if (strand)
        {
            for (Region1D r : local.getRegions())
            {
                if (local.getLength()==0)
                {
                    continue;
                }
                global.addRegion(r.getX1()+g_start, r.getX2()+g_start);
            }
            
            return global;
        }
        //reverse 
        Vector<Region1D> v = local.getRegions();
    
        for (int i=v.size()-1; i>=0; i--)
        {
            Region1D r = v.get(i);
    
            if (r.getLength()==0)
            {
                continue;
            }
    
            global.addRegion(g_end-r.getX2()+1, g_end-r.getX1()+1);
        }
    
        return global;
    
    }

    public static int toLocal(int pos, int g_start, int g_end, boolean strand)
    {
        return (strand) ? pos - g_start : g_end + 1 - pos;
    }

    public static int toLocal(int pos, Gene g)
    {
        if (pos < 0)
            return pos;
        
        return toLocal(pos, g.start, g.end, g.strand);
    }

    public static Pair<Integer,Integer> toLocal(Pair<Integer,Integer> global, int g_start, int g_end, boolean strand)
    {
        if(strand)
        {
            return new Pair<Integer,Integer>(global.getFirst()-g_start, global.getSecond()-g_start);
        }
        return new Pair<Integer,Integer>(g_end+1-global.getSecond(), g_end+1-global.getFirst());
    }
    public static Region1D toLocal(Region1D global, Gene g)
    {
        return toLocal(global, g.start, g.end, g.strand);
    }

    public static Region1D toLocal(Region1D global, int g_start, int g_end, boolean strand)
    {
        if(strand)
        {
            return new Region1D(global.getX1()-g_start, global.getX2()-g_start, global.getObject());
        }
        return new Region1D(g_end+1-global.getX2(), g_end+1-global.getX1(), global.getObject());
    }
    public static RegionVector toLocal(RegionVector global, Gene g)
    {
        return toLocal(global, g.start, g.end, g.strand);
    }
    public static RegionVector toLocal(RegionVector global, int g_start, int g_end, boolean strand)
    {
        RegionVector rv = new RegionVector();
        if (strand)
        {
            for (Region1D r : global.getRegions())
            {
                rv.addRegion(toLocal(r,g_start,g_end,strand));
            }
        }
        else
        {
            Vector<Region1D> v = global.getRegions();
            for (int i=v.size()-1; i>=0; i--)
            {
                Region1D r = v.get(i);
                rv.addRegion(toLocal(r,g_start,g_end,strand));
            }
            

        }
        return rv;
    }

    /*


    public static class ProteinSeqCheck
    {
        Logger log = LogConfig.getLogger();
        Protein p;
        public Vector<Integer> shifts = new Vector<Integer>();
        public Vector<Pair<Integer,Integer>> cuts =new Vector<Pair<Integer,Integer>> ();

        public Vector<String> translated = new Vector<String> ();
        
        public boolean multi_src;
        public boolean single_src;

        public int getNumSolutions()
        {
            return shifts.size();
        }

        public String getTranslated()
        {
            if (!single_src)
            {
                throw new RuntimeException("no single valid solution!");
            }

            return getTranslated(0);
        }
        public String getTranslated(int sol)
        {
            return translated.get(sol);
        }
        public Vector<Pair<Integer,Integer>> getFixedProteinRegionsInGene()
        {
            if (!single_src)
            {
                throw new RuntimeException("no single valid solution!");
            }
            return getFixedProteinRegionsInGene(0);
        }
        /*
        public Vector<Pair<Integer,Integer>> getFixedProteinRegionsInGene(int sol)
        {
            Vector<Pair<Integer,Integer>> rv = new Vector<Pair<Integer,Integer>>();
            int shift=shifts.get(sol);
            Pair<Integer,Integer> cut = cuts.get(sol);
            String ps = translated.get(sol);
            


            
            int n2skip = shift+cut.getFirst()*3;

            int coding_length=0;
            for (CodingExon ce: p.coding_exons)
            {
                coding_length+=ce.length;
            }
            int endcut=coding_length-n2skip-3*ps.length();
            int lastpos=coding_length-endcut;
            
            int celength=0;
            for (CodingExon ce: p.coding_exons)
            {
                int start=ce.start_in_gene;
                int end = ce.end_in_gene+1;

                if (n2skip>0)
                {
                    start=Math.min(start+n2skip, end);
                    n2skip-=(start-n2skip);
                }
                celength+=(end-start);
                boolean stophere=false;
                if (celength>lastpos)
                {
                    end-=(celength-lastpos);
                    stophere=true;
                }


                rv.add(new Pair<Integer,Integer>(start,end));

                if (stophere)
                {
                    break;
                }


            }
            return rv;
        }
        public ProteinSeqCheck(Protein p, int max_no_aa_accepted, GenomeSequenceExtractor extractor, String gene_seq)
        {
            this.p=p;
            String prot_dna_seq = null;
            String coding_prot_dna_seq = null;
            
            if (extractor!=null)
            {
                prot_dna_seq=GenomicUtils.getProteinSeq(p, extractor, true);
            }
            else
            {
                RegionVector v = p.getCodingRegionVector(false);
                prot_dna_seq = GenomicUtils.getSplicedSeq(gene_seq, v, true);
            }

            

            for (int i=0; i<3; i++)
            {
                String ps=GenomicUtils.translate(prot_dna_seq, 0, i);
                    
                int nx=0;
                Pair<Integer,Integer> cutPos = protein.DigeUtils.getXCut(ps);
                
                if (cutPos.getSecond()<0)
                {
                    continue;
                }
            
                for (int j=cutPos.getFirst(); j<cutPos.getSecond()+1; j++)
                {
                    if (ps.charAt(j)!='X')
                        continue;
            
                    nx++;
                }
                if (nx>max_no_aa_accepted)
                {
                    //too much x , skip it
                    continue;
                }
                
                translated.add(ps.substring(cutPos.getFirst(),cutPos.getSecond()));
                shifts.add(i);
                cuts.add(cutPos);
                
            }
            
            multi_src=translated.size()>1;
            single_src=translated.size()==1;

        }
        


    }
    */
    static char REVERSE_DNA[]=new char[Character.MAX_VALUE];
    static char DNA2RNA[]=new char[Character.MAX_VALUE];

    static 
    {
        for (char c=0; c<Character.MAX_VALUE; c++)
        {
            REVERSE_DNA[c]='N';
            DNA2RNA[c]='N';
        }
        REVERSE_DNA['T']='A';
        REVERSE_DNA['A']='T';
        REVERSE_DNA['G']='C';
        REVERSE_DNA['C']='G';

        REVERSE_DNA['t']='A';
        REVERSE_DNA['a']='T';
        REVERSE_DNA['g']='C';
        REVERSE_DNA['c']='G';


        DNA2RNA['T']='A';
        DNA2RNA['A']='U';
        DNA2RNA['G']='C';
        DNA2RNA['C']='G';

        DNA2RNA['t']='A';
        DNA2RNA['a']='U';
        DNA2RNA['g']='C';
        DNA2RNA['c']='G';
    }

    public static String antisense(String s)
    {
        return antisense(s,false);
    }
    public static String antisense(String s, boolean get_rna)
    {
        final char[] conv = (get_rna)? DNA2RNA : REVERSE_DNA;

        StringBuffer sb = new StringBuffer();
        for (int i=0; i<s.length(); i++)
        {
            sb.append(conv[s.charAt(i)]);
            
        }
        return sb.toString();
    }
    public static String reverse_complement(String s)
    {
        StringBuffer sb = new StringBuffer();
        for (int i=s.length()-1; i>=0; i--)
        {
            sb.append(REVERSE_DNA[s.charAt(i)]);
            
        }
        return sb.toString();
    }

    public static String translate(String dna)
    {
        return translate(dna,0,0);
    }
    public static String translate(String dna, int start, int frame)
    {
        char[] ar = dna.toCharArray();

        StringBuffer sb = new StringBuffer();
        for (int i=start+frame; i<=dna.length()-3; i+=3)
        {
            sb.append(protein.aminoacid.AminoAcidType.get(ar, i));
        }
        return sb.toString();
    }

    static public void fixProteinTranslationStart(Protein p, GenomeSequenceExtractor fromdna)
    {
        String trseq = getTranscriptSeq(p.parentTranscript,fromdna);
        final int transend = trseq.length()-p.transl_end;
        if (transend<=p.transl_start)
        {
            throw new RuntimeException(String.format("invalid protein spec for %s, transcript seq is %d long and trans range is: %d-> -%d",
                                                     p.id, trseq.length(), p.transl_start, p.transl_end));
        }
        String coding = trseq.substring(p.transl_start, transend);

        Logger log = LogConfig.getLogger();

        //check which is the best frame
        int orig_codons=0;
        int bestframe=0;
        int best_num_codons=0;

        for (int i=0; i<3; i++)
        {
            String trans = translate(coding, 0, i);
            int num_real_codons=0;
            
            for (int j=0; j<trans.length(); j++)
            {
                if (trans.charAt(j)=='X')
                    continue;
                
                num_real_codons++;
            }

            if (i==0)
            {
                orig_codons=num_real_codons;
            }
            log.debug(String.format("fixing %s length: %d frame: %d codons: %d\n", p.id, trans.length(),  i, num_real_codons));
            if (num_real_codons<=best_num_codons+1)
                continue;
            
            bestframe=i;
            best_num_codons=num_real_codons;
        }

        if (bestframe==0)
            return;

        
        
        log.warn(String.format("correct translation start for %s (gene: %s) original codons: %d corrected: %d", p.id, p.parentTranscript.parentGene.id, 
                 orig_codons, best_num_codons));

        p.transl_start+=bestframe;
        p.init();
    
    }
    
    public static String getTranscriptSeq(Transcript t, GenomeSequenceExtractor fromdna)
    {
        final Gene from = t.parentGene;
        return getSplicedSeq(from.chr, from.strand, t.getRegionVector(true), fromdna, true);
    }
    
    public static String getGeneSeq(Gene g, GenomeSequenceExtractor fromdna)
    {
        String gseq = fromdna.getGenomeSeq(g.chr, g.start, g.end+1);

        if (g.strand)
            return gseq;

        return GenomicUtils.reverse_complement(gseq);
        
    }
    
    static int SHIFT_TABLE [][]={
        
                {0, 1, 2}, 
                {1, 2, 0}, 
                {2, 0, 1}
            };


    public static String getSplicedSeq(final String chr, boolean strand, RegionVector regvec, GenomeSequenceExtractor fromdna, boolean dna)
    {
        return getSplicedSeq(chr,strand,regvec.getRegionPairs(),fromdna,dna);
    }

    public static String getSplicedSeq(final String chr, boolean strand, Vector<Pair<Integer,Integer>> regions, GenomeSequenceExtractor fromdna, boolean dna)
    {
        return getSplicedSeq(strand,chr,regions,fromdna,null,dna);
    }

    public static String getSplicedSeq(String seq, RegionVector regionvec, boolean dna)
    {
        return getSplicedSeq(seq, regionvec.getRegionPairs(), dna);
    }

    public static String getSplicedSeq(String seq, Vector<Pair<Integer,Integer>> regions, boolean dna)
    {
        return getSplicedSeq(true,null,regions, null, seq,dna);
    }
   
    public static String getSplicedSeq(boolean strand, final String chr, Vector<Pair<Integer,Integer>> regions, GenomeSequenceExtractor fromdna, 
                                       String relative, boolean dna)
    {
        
        Collections.sort(regions);
        StringBuffer sb = new StringBuffer();
        int real_length=0;
        for(int i=0; i<regions.size(); i++)
        {
            Pair<Integer,Integer> region = regions.get(i);
            if (region.getSecond()<=region.getFirst())
            {
                continue; //nothing to extract
            }
            String genomeseq = null;
            if (relative!=null)
            {
                //checks
                if(region.getFirst()<0 || region.getFirst()>relative.length())
                {
                    throw new RuntimeException("Invalid region: "+region+" (outof bounds in relative of length: "+relative.length()+") regions: "+regions);
                    
                }
                if (region.getSecond()>relative.length())
                {
                    throw new RuntimeException("Invalid region: "+region+" (outof bounds in relative of length: "+relative.length()+") regions: "+regions);
                    
                }

                if (region.getFirst()<0 || region.getSecond()<0 || region.getSecond()<=region.getFirst() || region.getSecond()>relative.length())
                {
                    throw new RuntimeException("invalid region: "+region+" in regions: "+
                                               regions+" relative length: "+relative.length());
                }
                genomeseq = relative.substring(region.getFirst(), region.getSecond());
            }
            else
            {
                genomeseq = fromdna.getGenomeSeq(chr, region.getFirst(), region.getSecond());
            }
            
            if (genomeseq==null)
            {
                throw new RuntimeException(String.format("error by extracting %s for %s\n", ""+region, chr));
                                           
            }
            //System.out.println("APPEND:\n"+genomeseq.toString());
            sb.append(genomeseq);
        }

        if (!strand)
        {
            String revsb = GenomicUtils.reverse_complement(sb.toString());
            sb.setLength(0);
            sb.append(revsb);
        }

        String protdna = sb.toString();
        if (dna)
        {
            
            return protdna;
        }
        /*if (corrected_regions!=null)
        {
            System.out.println("protdna:\n"+protdna);
        }
        */

        return GenomicUtils.translate(protdna, 0, 0);
        
    }

    

    public static String getProteinSeqUntilStopCodon(Protein fromP, GenomeSequenceExtractor fromdna, boolean dna, Vector<Pair<Integer,Integer> > regions)
    {
        if (regions==null)
        {
            regions = new Vector<Pair<Integer,Integer> >();
        }
        
        
        final boolean strand = fromP.parentTranscript.parentGene.strand;


        final String chr = fromP.parentTranscript.parentGene.chr;


        Vector<String> ce_dnas = new Vector<String>();
        
        final int ne = fromP.getNumExons();
        for (int i=0; i<ne; i++)
        {
            CodingExon ce = fromP.getExon(i);
            ce_dnas.add(fromdna.getGenomeSeq(chr, ce.abs_start, ce.abs_end+1, strand));
        }
        StringBuffer dnabuff =new StringBuffer();
        StringBuffer protBuff = new StringBuffer();
        StringBuffer codon = new StringBuffer();
        boolean gotstop = false;
        for (int i=0; !gotstop && i<ne; i++)
        {
            CodingExon ce = fromP.getExon(i);
            String cseq = ce_dnas.get(i);
            for (int j=0; !gotstop && j<cseq.length(); j++)
            {
                codon.append(cseq.charAt(j));
                if (codon.length()!=3)
                    continue;

                dnabuff.append(codon.toString());
                char AA = protein.aminoacid.AminoAcidType.get(codon.toString());
                if (AA == 'X')
                {
                    if (strand)
                    {
                        regions.add(new Pair<Integer,Integer>(ce.abs_start, ce.abs_start+j));
                    }
                    else
                    {
                        regions.add(new Pair<Integer,Integer>(ce.abs_end+1-j, ce.abs_end+1));
                    }
                    
                    gotstop=true;
                    break;
                }
                codon.setLength(0);
                protBuff.append(AA);

            }
            if (gotstop)
                break;
            
            regions.add(new Pair<Integer,Integer>(ce.abs_start, ce.abs_end+1));
        }
        if (!strand)
        {
            Collections.reverse(regions);
        }
        if (dna)
        {
            return dnabuff.toString();
        }
        
        return protBuff.toString();
    }

    public static  HashMap<Protein, String> getProteinSequences(Gene g, GenomeSequenceExtractor fromdna, boolean dna)
    {
        HashMap<Protein, String> rv = new HashMap<>();
        for (Protein p : g.proteins.values())
        {
            rv.put(p, getProteinSeq(p, fromdna, dna));
        }
        return rv;
    }

    
    
    public static String getProteinSeq(Protein fromP, String gene_seq, boolean dna)
    {
        return getSplicedSeq(gene_seq, fromP.getCodingRegionVector(false), dna);

    }
    public static String getProteinSeq(Protein fromP, GenomeSequenceExtractor fromdna, boolean dna)
    {
        
        final Gene from = fromP.parentTranscript.parentGene;
        //System.out.printf("%s: %c\n", from.id, (from.strand)?'+':'-');
        
        return getSplicedSeq(from.chr, from.strand, fromP.getCodingRegionVector(true), fromdna, dna);
        //return getSplicedSeq(from.strand, from.chr,  fromP.getCodingRegionVector(true), fromdna, dna);
        /*
        StringBuffer sb = new StringBuffer();
        
        int real_length=0;
        for(int i=0; i<fromP.coding_exons.size(); i++)
        {
            CodingExon ce = fromP.coding_exons.get(i);
            String genomeseq = fromdna.getGenomeSeq(from.chr, ce.abs_start, ce.abs_end);
            if (genomeseq==null)
            {
                throw new RuntimeException(String.format("error by extracting %s for %s\n", ""+ce, fromP));
                                           
            }
            if (!from.strand)
            {
                genomeseq = GenomicUtils.reverse_complement(genomeseq);
            }
            int shift=0;
            if (fixes!=null)
            {
                int fix=fixes.get(i);
                int clength=sb.length()%3;
                shift = SHIFT_TABLE[clength][fix];
                
            }
            
            
            if (shift>0 && genomeseq.length()>shift)
            {
                genomeseq=genomeseq.substring(shift);
            }
            sb.append(genomeseq);
        }
        String protdna = sb.toString();
        if (dna)
        {
            
            return protdna;
        }

        return GenomicUtils.translate(protdna, 0, 0);
        */
    }
    

    static final String FAILED_OTHER ="failed: other";
    public static class OrgCheckInfo
    {
        GenomeAnnotationHandler gah;
        String prefix;

        public OrgCheckInfo(String prefix, GenomeAnnotationHandler gah)
        {
            this.gah = gah;
            this.prefix=prefix;
            cu = new CountUpdater(prefix+" proteins");
            cu.update(FAILED_OTHER, 0);
            stop_cu = new CountUpdater(prefix+" db stopcodons");
            gstop_cu = new CountUpdater(prefix+" genome stopcodons");
            splice = new CountUpdater(prefix+" splicesites");
            prots2splice = new CountUpdater(prefix+" prot2splicesites");
            if (gah instanceof GTFAnnotation)
            {                 
                numannprot = ((GTFAnnotation)gah).getNumProteins();
            }
        }

        public void writeInfos(PrintWriter pw)
        {
            pw.printf("%s ok proteins: %d/%d (%2.3f%%) vs annotated: %d/%d: %2.3f%%\n", 
                      prefix, valid_prots, cu.getTotalCount(), (100.0*valid_prots)/cu.getTotalCount(), 
                      valid_prots, numannprot, (numannprot==0)?0.0:(100.0*valid_prots)/numannprot);
    
            cu.printBestN(pw, -1);
            pw.printf("prots with annotated long tail after stop codons: %d\n", num_prots_with_tail_stopcodons);
            stop_cu.printBestN(pw, -1);
            gstop_cu.printBestN(pw, 10);
            splice.printBestN(pw, 6);
            prots2splice.printBestN(pw, 5);
            pw.flush();
        }

        public static void writeHeader(PrintWriter pw)
        {
            pw.printf("prefix\tnumgenes\tnumprots\tannotpros\tnumvalids\tnumvalidpercent\tnumvalidannotpercent\tnumfailed\tnumfailedpercent\n");
            pw.flush();

        }

        public void writeInfo(PrintWriter pw)
        {
            int failed = cu.getCount(FAILED_OTHER);
            pw.printf("%s\t%d\t%d\t%d\t%d\t%2.2f\t%2.2f\t%d\t%2.2f\n",
                      prefix, ng, np, numannprot, valid_prots, 
                      (100.0*valid_prots)/np, 
                      (100.0*valid_prots)/numannprot, 
                      failed,
                      (100.0*failed)/np);
            pw.flush();
        }
        int numannprot=0;
        public CountUpdater cu;
        public CountUpdater stop_cu;
        public CountUpdater gstop_cu;
        public CountUpdater splice;
        public CountUpdater prots2splice;
        public int np=0;
        public int valid_prots=0;
        public int ng=0;
        public int num_prots_with_tail_stopcodons;

    }

    public static OrgCheckInfo testAnnotation(GenomeSequenceExtractor gex, GenomeAnnotationHandler gah)
    {
        return testAnnotation(new PrintWriter(System.out), "", gex,gah, false, 0);
        
    }
    public static void reportProteinCheck(Protein p, GenomeSequenceExtractor gex, String check)
    {
        reportProteinCheck(new PrintWriter(System.out),p ,gex,check);
    }
    public static void reportProteinCheck(PrintWriter pw, Protein p, GenomeSequenceExtractor gex, String check)
    {
        
        Gene g = p.parentTranscript.parentGene;
        String pseq = GenomicUtils.getSplicedSeq(g.chr, g.strand, p.getCodingRegionVector(true).getRegionPairs(), gex, false);

        pw.printf("%s(%s(%c) %d-%d): %s transl start: %d (%s) transl end: %d (%s)\n", g.id,  
                  g.chr, (g.strand)?'+':'-', g.start, g.end,
                  p.id, p.transl_start, p.start_exon, p.transl_end, p.end_exon);
        if (check!=null)
        {
            pw.printf("annot protseq:\n%s\n", check);
        }
            
        String gene_seq = GenomicUtils.getGeneSeq(g, gex);
        pw.printf("genomic protseq:\n%s\n", pseq);
        for (int i=0; i< p.getNumExons(); i++)
        {
            CodingExon ce = p.getExon(i);
            
            String exon_cdna = gene_seq.substring(ce.start_in_gene, ce. end_in_gene+1);
            pw.printf("\texon: %d: %s %d-%d (exon: %d-%d)\n", i, ce.exon.id, ce.abs_start, ce.abs_end, ce.exon.abs_start, ce.exon.abs_end);
            for (int j=ce.frame; j<exon_cdna.length()-3; j+=3)
            {
                String codon = exon_cdna.substring(j, j+3);
                if('X' != protein.aminoacid.AminoAcidType.get(codon))
                    continue;

                pw.printf("\t\tSTOP CODON at %d: %s\n", j, codon);
            }
            String pep = translate(exon_cdna, 0, ce.frame);
            pw.printf("\t\tpeptide: %s\n\t\tcdna: %s\n",  pep, exon_cdna);
        }
        pw.flush();
    }
    public static OrgCheckInfo testAnnotation(PrintWriter pw, String prefix, GenomeSequenceExtractor gex, GenomeAnnotationHandler gah, boolean verbose, int verbose_failed)
    {
        Logger log = LogConfig.getLogger();
        
        final String CANONICAL = CANONICAL_SPLICE_MOTIF;
        Iterator<Gene> it = gah.getGenes(null, null, null);
        
        OrgCheckInfo oci = new OrgCheckInfo(prefix, gah);

        int num_reported_genomic_failed = 0;
        int num_reported_failed = 0;
        int num_reported_longtail = 0;
        int num_reported_subsequence = 0;
        int num_prots_with_tail_stopcodons=0;

        HashSet<String> missing_chr = new HashSet<String>();
        while (it.hasNext())
        {

            if ((++oci.ng) % 1001 == 0 && verbose)
            {
                 pw.printf("after %d genes\n", oci.ng-1);

                 oci.writeInfos(pw);
            }
            Gene g = it.next();
            if (!gex.gotChr(g.chr))
            {
                if (!missing_chr.contains(g.chr))
                {
                    log.warn("missing chr: %s requested by: %s", g.chr, g.id);
                }
                missing_chr.add(g.chr);
                continue;

            }
            String gene_seq = GenomicUtils.getGeneSeq(g, gex);

            for (Protein p : g.proteins.values())
            {
                oci.np++;
                int noncan = 0;
                RegionVector introns = p.getCodingRegionVector(true).getInverseRegionVector();
                for (Region1D r : introns.getRegions())
                {
                     String donor = gex.getGenomeSeq(g.chr, r.getX1(), r.getX1()+2, g.strand);
                     String acceptor = gex.getGenomeSeq(g.chr, r.getX2()-2, r.getX2(), g.strand);
                     String ss = (g.strand)?donor+acceptor:acceptor+donor;
                     oci.splice.update(ss, 1);
                     if (CANONICAL.equals(ss))
                         continue;
                     noncan++;
                }
                oci.prots2splice.update("prots with "+noncan+" non-can splicesites", 1);
                String pseq = GenomicUtils.getSplicedSeq(g.chr, g.strand, p.getCodingRegionVector(true).getRegionPairs(), gex, false);
                //String pseq = GenomicUtils.getProteinSeq(p, db.gex, false);
                String check = gah.getProteinSeq(p.id);

                
                try
                {
                    RegionVector v = p.getCodingRegionVector(false);
                    String prot_dna_seq = GenomicUtils.getSplicedSeq(gene_seq, v, true);
                }
                catch(Exception e)
                {
                    throw new RuntimeException("error at protein;: "+ p.id+" gene: "+g.id, e);
                }
                

                String genomic_to_first_stop = pseq;
                if (pseq.indexOf('X')>0)
                {
                    genomic_to_first_stop = pseq.substring(0, pseq.indexOf('X'));
                }
                
                if (pseq.length()-genomic_to_first_stop.length()>20)
                {
                    /*
                    Vector<Pair<Integer,Integer> > cut_regions = new Vector<Pair<Integer,Integer> >();
                    String gcutted = GenomicUtils.getProteinSeq(p, db.gex, false, true, cut_regions);
                    System.out.println("regions: "+cut_regions);
                    String recheck = GenomicUtils.getSplicedSeq(g.chr, g.strand, cut_regions, db.gex, false);
                    RegionVector local = GenomicUtils.toLocal(new RegionVector(cut_regions), g);

                    String gseq = GenomicUtils.getGeneSeq(g, db.gex);

                    String lseq = GenomicUtils.getSplicedSeq(gseq, local, false);

                    System.out.printf("\n\nlength diff at %s(%c):%s genomic: %d until first stop codon: %d\nfull: %s\ncut : %s\ncheck: %s\nrecheck: %s\nlcheck: %s\nANNOT: %s\n",
                                      g.id, (g.strand)?'+':'-', p.id, pseq.length(), genomic_to_first_stop.length(),
                                      pseq, genomic_to_first_stop, gcutted, recheck, lseq,
                                      check);
                                      */

                }
                if (check==null)
                    continue;
                
                String check_to_first_stop = check;

                int nstops=0;
                for (int i=0; i<check.length(); i++)
                {
                    if (check.charAt(i)!='X')
                        continue;
                    
                    nstops++;
                }
                oci.stop_cu.update("db info with "+nstops+" stop codons", 1);

                int ngstops=0;
                for (int i=0; i<pseq.length(); i++)
                {
                    if (pseq.charAt(i)!='X')
                        continue;
                    
                    ngstops++;
                }


                //System.err.printf("%s:%s : %d stop codons\n", g.id, p.id, ngstops);
                
                if (ngstops>= 10)
                {
                    oci.gstop_cu.update("genomic  info with >=10 stop codons", 1);
                }
                else
                {
                    oci.gstop_cu.update("genomic  info with "+ngstops+" stop codons", 1);
                }

                if (check.indexOf('X')>0)
                {
                    check_to_first_stop = check.substring(0, check.indexOf('X'));
                }

                if (genomic_to_first_stop.equals(check_to_first_stop))
                {
                    if (ngstops>3)
                    {
                        if (num_reported_longtail<verbose_failed)
                        {
                            num_reported_longtail++;
                            pw.printf("genomic stops: %d %s: %s last exon: %s transl stop: %d\n", ngstops, g.id, p.id, p.end_exon, p.transl_end);
                        }
                        log.debug("genomic stops: %d %s: %s last exon: %s transl stop: %d\n", ngstops, g.id, p.id, p.end_exon, p.transl_end);
                        num_prots_with_tail_stopcodons++;
                    }
                    oci.valid_prots++;
                    oci.cu.update("fullok", 1);
                    continue;
                }
                if (pseq.contains(check) || check.contains(pseq))
                {
                    int lengthdiff = Math.max(pseq.length(), check.length())-Math.min(pseq.length(), check.length());
                    
                    if (lengthdiff<2)
                    {
                        oci.cu.update("subsequence, lengthdiff == 1", 1);
                        oci.valid_prots++;
                    } 
                    else if (lengthdiff <= 5)
                    {
                        oci.cu.update("subsequence, lengthdiff <= 5", 1);
                    }
                    else if (lengthdiff <= 10)
                    {
                        oci.cu.update("subsequence, lengthdiff <= 10", 1);
                    }
                    else
                    {
                        oci.cu.update("subsequence, lengthdiff > 10", 1);
                    }
                    


                    if (num_reported_subsequence < verbose_failed)
                    {
                        num_reported_subsequence++;
                        pw.printf("SUBSEQUENCE:\n%s:%s\n", g.id, p.id);
                        reportProteinCheck(pw,p,gex,check);
                        
                    }
                    //System.out.printf("SUBSEQ (genome: %d vs db: %d):\n%10s:%s\n%10s:%s\n", pseq.length(), check.length(),  "genome", pseq, "db", check);
                    continue;
                }

                
                String cdna = GenomicUtils.getProteinSeq(p, gex, true);
                boolean foundok = false;
                for (int i=0; !foundok && i<3; i++)
                {
                    String ti = GenomicUtils.translate(cdna, 0, i);
                    if (ti.equals(check))
                    {
                        oci.cu.update("failed: frameshift", 1);
                        foundok=true;
                        break;
                    }
                    if (check.contains(ti) || ti.contains(check))
                    {
                        oci.cu.update("failed: frameshift subsequence", 1);
                        foundok=true;
                        break;
                    }
                }
                if (foundok)
                {
                    continue;
                }
                //count stop codons
                int nstop_check = 0;
                for (int i = 0; i<check.length(); i++)
                {
                    if (check.charAt(i)!='X')
                        continue;
                    nstop_check++;
                    
                }
                int nstop_read = 0;
                for (int i = 0; i<pseq.length(); i++)
                {
                    if (pseq.charAt(i)!='X')
                        continue;
                    nstop_read++;
                }

                if (nstop_check > 1 && nstop_read > 1)
                {
                    oci.cu.update("failed: both db and genomic multiple stop codons", 1);
                    continue;
                }
                if (nstop_check > 1)
                {
                    oci.cu.update("failed: db multiple stop codons",1);
                    continue;
                }
                if (nstop_read > 1)
                {
                    if(num_reported_genomic_failed < verbose_failed)
                    {
                        num_reported_genomic_failed++;
                        pw.printf("FAILED: (%d stop codons)", nstop_read);
                        reportProteinCheck(pw,p,gex,check);
                        
                    }
                    oci.cu.update("failed: genomic multiple stop codons",1);
                    continue;
                }

                //calc the longest subsequence quick&dirty
                int lcs = 0;
                for(int skip = -check.length(); skip<pseq.length(); skip++)
                {
                    int tmatches = 0;
                    for (int i=0, j=skip; i<check.length() && j<pseq.length(); i++, j++)
                    {
                        if (j<0 || check.charAt(i)!=pseq.charAt(j))
                            continue;
                    
                        tmatches++;
                    }
                    if (tmatches <= lcs)
                        continue;
                    
                    lcs = tmatches;
                }

                double nmatches = lcs/(double)check.length();
                if (lcs >= 0.95)
                {
                    oci.cu.update("no subseq but >=95% seqid", 1);
                    oci.valid_prots++;
                }
                else if (lcs >= 0.8)
                {
                    oci.cu.update("no subseq but >=80%% seqid", 1);
                }
                else
                {
                    oci.cu.update(FAILED_OTHER, 1);

                    if (num_reported_failed < verbose_failed)
                    {
                        num_reported_failed++;
    
                        pw.printf("FAILED: other\n");
                        reportProteinCheck(pw,p,gex,check);
                    }
                }
                

                
                
            }
            
        }
        
        oci.writeInfos(pw);
                
        return oci;
    }

}
