package genome;

import java.io.*;
import java.util.*;

import javax.xml.stream.*;

import lmu.utils.*;

import org.jdom.Element;

public class Organism 
{
    Logger log = LogConfig.getLogger();

    final static String ORG_ELEM = "org";
    final static String GENOMIC = "GENOMIC";
    final static String DERIVED = "DERIVED";
    final static String STAR_INDEX = "STAR_INDEX";
    final static String BOWTIE1_INDEX = "BOWTIE1_INDEX";
    final static String BOWTIE2_INDEX = "BOWTIE2_INDEX";
    final static String TRANSBOWTIE1_INDEX = "TRANSCRIPTOME_BOWTIE1_INDEX";
    final static String TRANSBOWTIE2_INDEX = "TRANSCRIPTOME_BOWTIE2_INDEX";

    final static String GENOME_LOOKUP = ".genome.lookup";
    final static String GTF_GENE_EXTRACT = ".gene_extract";
    final static String GTF_LOOKUP_SUFFIX = ".lookup";
    final static String GTF_GENE2PROTEIN_LOOKUP_SUFFIX = ".prot2gene";
    final static String GTF_GENE2TR_LOOKUP_SUFFIX = ".tr2gene";
    final static String GTF_NAME2GENE_LOOKUP_SUFFIX = ".g2namelookup";

    final static String GENOME_LOOKUP_ELEM = "genome_lookup";
    final static String GTF_GENE_EXTRACT_ELEM = "gtf_gene_extract";
    final static String GTF_GENE_EXTRACT_LOOKUP_ELEM = "gtf_gene_extract_lookup";
    final static String GTF_LOOKUP_ELEM = "gtf_lookup";
    final static String GTF_GENE2PROTEN_ELEM = "gtf_gene2prot";
    final static String GTF_GENE2TR_ELEM = "gtf_gene2tr";
    final static String GTF_NAME2GENE_ELEM = "gtf_name2gene";

    final static String INDEXDIR_ELEM = "indexdir";
    final static String BLASTDB_ELEM = "blastdb";
    final static String DNA_ELEM = "dna";
    final static String GTF_ELEM = "gtf";
    final static String PROTEIN_ELEM = "protein";
    final static String TRANSCRIPTOME_ELEM = "transcriptome";

    final static String STAR_INDEX_ELEM = "star";
    final static String BOWTIE1_INDEX_ELEM = "bowtie1";
    final static String BOWTIE2_INDEX_ELEM = "bowtie2";
    final static String TRANSBOWTIE1_INDEX_ELEM = "transbowtie1";
    final static String TRANSBOWTIE2_INDEX_ELEM = "transbowtie2";

    public String name;
    String path;
    String symbol;
    HashSet<String> aliases = new HashSet<String>();
    int taxid;

    //annot section
    File gtf_file;
    File gtf_file_lookup;
    File gtf_file_prot2gene;
    File gtf_file_tr2gene;
    File gtf_file_name2id;
    File gtf_geneseq_extract;
    File gtf_geneseq_extract_lookup;

    File indexDir = null;

    HashMap<String, Long> gene2fp = null;

    SQLConnectionOptionParser annot_db;
    GTFAnnotation handler; 


    File cdna_fasta;
    File prot_fasta;

    //genome
    File genome_dir;
    File genome_lookup;
    GenomeSequenceExtractor gex;

    boolean proteins_loaded=false;


    String genome_version;
    String transcriptome_version;

    static HashMap<String, Organism> lookup = new HashMap<String, Organism>();
    static HashMap<Integer, Organism> tax2org = new HashMap<Integer, Organism>();

    public Organism(String name, File genome, File gtf, File protfasta)   
    {
        this.name = name;
        this.genome_dir = genome;
        gtf_file = gtf;
        prot_fasta = protfasta;
        this.path = genome.getParent();
    }
    public Organism(Element e)
    {
        addAlias(name = XMLUtils.getAttrib(e, "name", null, true));
        setTaxId(XMLUtils.getIntAttrib(e, "taxid", null));
        setAliases(XMLUtils.getAttrib(e, "syns", null, false));

        path = XMLUtils.getAttrib(e, "path", null, true);

        addAlias(symbol = XMLUtils.getAttrib(e, "symbol", null, false));
        if (symbol == null || symbol.length() == 0)
        {
            symbol = name;
        }

        cdna_fasta = getPath(e, TRANSCRIPTOME_ELEM); 
        gtf_file = getPath(e, GTF_ELEM);
        gtf_file_lookup = getPath(e, GTF_LOOKUP_ELEM);
        gtf_file_prot2gene = getPath(e, GTF_GENE2PROTEN_ELEM);
        gtf_file_tr2gene = getPath(e, GTF_GENE2TR_ELEM);
        gtf_geneseq_extract =  getPath(e, GTF_GENE_EXTRACT_ELEM);
        gtf_geneseq_extract_lookup = getPath(e, GTF_GENE_EXTRACT_LOOKUP_ELEM);
        Element name2gene = e.getChild(GTF_NAME2GENE_ELEM);
        
        gtf_file_name2id = getPath(e, GTF_NAME2GENE_ELEM, null, false);
        
        prot_fasta = getPath(e, PROTEIN_ELEM);
        indexDir = getPath(e, INDEXDIR_ELEM);

        genome_dir = getPath(e, DNA_ELEM);
        genome_lookup = getPath(e, GENOME_LOOKUP_ELEM);

        addAlias(genome_dir.getParentFile().getName());

    }

    static public File getPath(Element e, String elem)
    {
        return getPath(e, elem, null, true);
    }

    static public File getPath(Element e, String elem, String version, boolean throw_exception)
    {
        LinkedList<Element> v = XMLUtils.getChildren(e, elem);
        Element rve = ObjectGetter.filterOne(v, (_e) -> { String _v = XMLUtils.getAttrib(_e, "version", null, false); return ((version == null || version.length() == 0) && _v == null) || 
                                            (_v != null && version != null && version.equals(_v)); });

        if (rve == null && throw_exception)
            throw new FRuntimeException("no path found to %s in version >%s<", elem, version);

        if (rve == null)
            return null;
        
        return XMLUtils.readPath(rve);
    }

    public File getDNAMappingIndexDir()
    {
        return indexDir;
    }

    public synchronized String getGeneSeq(String id)
    {
        /*
        if (gene2fp==null && gtf_geneseq_extract.exists() && gtf_geneseq_extract_lookup.exists())
        {
            gene2fp = new HashMap<String, Long>();

            for (String[] sp : FileUtils.readCols(gtf_geneseq_extract_lookup, "\t", 0, 1))
            {
                gene2fp.put(sp[0], Long.parseLong(sp[1]));
            }
        }

        if (gene2fp!=null)
        {
            Long fp = gene2fp.get(id);
            if (fp == null)
            {
                log.warn("no fp found for gene: "+id);
                return null;
            }
            String gseq = null;
            try
            {

                BufferedRandomAccessFile barf = new BufferedRandomAccessFile(gtf_geneseq_extract, "r", 30000);
                barf.seek(fp);
                String gn = barf.readLine();
                if (!gn.equals(id))
                {
                    throw new FRuntimeException("invalid gene seq entry found for %s (%s) in %s", id, gn, gtf_geneseq_extract.getAbsolutePath());
                }

                gseq = barf.readLine();
                barf.close();
            }
            catch (IOException e)
            {
                throw new FRuntimeException("i/o exception: %s while reading extracted gene seq from: %s for %s",e, e.getMessage(), gtf_geneseq_extract.getAbsolutePath(), id);
            }
            return gseq;

        }
         
        */ 
        GTFAnnotation gtf =  getAnnot();
        GenomeSequenceExtractor gex = getGenome();
        GenomicRegion gr = gtf.getGeneRegion(id);
        if (gr==null)
        {
            return null;
        }
        return gex.getGenomeSeq(gr.getChr(), gr.getX1(), gr.getX2()+1, gr.getStrand());
    }

    public void extractGenes()
    {
        GTFAnnotation gtf =  getAnnot();
        GenomeSequenceExtractor gex = getGenome();
        try
        {

            gene2fp = new HashMap<String, Long>();
            PrintWriter pw = new PrintWriter(gtf_geneseq_extract);
            long fp=0;            
            long tl=0;
            for (String gene: gtf.getGeneNames())
            {
                GenomicRegion gr = gtf.getGeneRegion(gene);
                int length = gr.getLength();
                tl+=length;
                System.out.printf("will take region: %s %d-%d for %s length: %d total: %2.2f Mio bp\n", gr.getChr(), gr.getX1(), gr.getX2(), gene, length, tl/1000000.0);
                String seq = gex.getGenomeSeq(gr.getChr(), gr.getX1(), gr.getX2()+1, gr.getStrand());


                gene2fp.put(gene, fp);
                fp+=(gene.length()+seq.length()+2);

                pw.println(gene);
                pw.println(seq);

            }
            pw.close();

            pw = new PrintWriter(gtf_geneseq_extract_lookup);

            for (Map.Entry<String,Long> e: gene2fp.entrySet())
            {
                pw.printf("%s\t%d\n", e.getKey(), e.getValue());
            }
            pw.close();
        }
        catch (IOException e)
        {
            throw new FRuntimeException("i/o error: %s while writing gene extraction to %s and %s", e, e.getMessage(), 
                                        gtf_geneseq_extract.getAbsolutePath(), gtf_geneseq_extract_lookup.getAbsolutePath());
        }

    }

    public synchronized GenomeSequenceExtractor getGenome()
    {
        if (gex!=null)
            return gex;

        return gex = new GenomeSequenceExtractor(genome_dir, genome_lookup);

    }

    public void close()
    {
        gex = null;
        handler = null;

    }
    public GTFAnnotation getAnnot(boolean loadproteins)
    {
        GTFAnnotation gah = getAnnot();
        if (!loadproteins || proteins_loaded)
            return gah;

        loadProteins(gah);
        return gah;
    }

    synchronized void  loadProteins(GTFAnnotation gah)
    {
        if (proteins_loaded)
            return;

        getAnnot();
        gah.setProteinFasta(prot_fasta);
        proteins_loaded = true;

    }
    public synchronized GTFAnnotation getAnnot()
    {
        if (handler!=null)
            return handler;

        if (gtf_file!=null && gtf_file.exists())
        {
            gtf_file_lookup.getParentFile().mkdirs();
            gtf_file_prot2gene.getParentFile().mkdirs();
            
            handler = new GTFAnnotation(this, gtf_file, gtf_file_lookup, gtf_file_prot2gene, gtf_file_tr2gene);

            if (gtf_file_name2id != null)
            {
                if (!gtf_file_name2id.exists())
                {
                    handler.createGeneNameLookup(gtf_file_name2id);
                }
                handler.setGeneNameLookup(gtf_file_name2id);
            }
            return handler;
        }
        /*
        if (annot_db!=null && annot_db.isInited())
        {
           proteins_loaded  = true;
           return handler = new extsource.EnsemblDbHandler(annot_db.getKeepAliveConnection());
        }
        */

        throw new RuntimeException("no annotation is configured for "+name+" gtf file: "+gtf_file);
    }
    public void setAliases(String syns)
    {
        if (syns==null)
            return;

        for (String s: StringUtils.toVector(syns.split("\\|")))
        {
            addAlias(s);
        }

    }
    public void addAlias(String n)
    {
        if (n==null)
            return;

        aliases.add(n);
        lookup.put(n, this);
        lookup.put(n.toLowerCase(), this);
    }

    public void setTaxId(int taxid)
    {
        this.taxid=taxid;
        tax2org.put(taxid, this);
    }

    static boolean check_dir(Logger log, File d, String cat)
    {
        if (d.exists() && d.isDirectory())
            return true;

        log.warn("missing "+cat+" dir: "+d.getAbsolutePath());
        return false;
    }

    static Element getBasePathDef(String name, String path, String relative)
    {
        Element base = new Element("basepath");
        base.setAttribute("name", name);
        base.setAttribute("path",  path);
        if (relative!=null)
        {
            base.setAttribute("relative", relative);
        }

        return base;
    }
    static Element getVarDef(String name, String value)
    {
        Element e = new Element("vardef");
        e.setAttribute("name", name);
        e.setAttribute("value", value);
        return e;
    }
    static Element getPathDef(String name,String path, String relative)
    {
        Element e = new Element(name);
        e.setAttribute("path", path);
        if (relative!=null)
        {
            e.setAttribute("relative", relative);
        }
        return e;
    }

    static String getRelativeRef(File ref, File path)
    {
     //   System.out.printf("path. %s\nref: %s\n", path.getAbsolutePath(), ref.getAbsolutePath());
        return path.getAbsolutePath().substring(ref.getAbsolutePath().length()+1);
    }

    public static Element createOrganismsXML(File rootdir, File derivedroot, File orgfile)
    {
        DataTable dt = DataTable.readFile(orgfile);

        Logger log = LogConfig.getLogger();
        Element root = new Element("config");





        /*<star relative="bovbase" path="STAR_INDEX"/>
              <bowtie1 relative="bovbase" path="BOWTIE1_INDEX/${bovine_genome_version}.dna"/>
              <bowtie2 relative="bovbase" path="BOWTIE2_INDEX/${bovine_genome_version}.dna"/>
              <transbowtie1 relative="bovbase" path="TRANSCRIPTOME_BOWTIE1_INDEX/${bovine_genome_version}.dna"/>
          */
        root.addContent(getBasePathDef(GENOMIC, rootdir.getAbsolutePath(), null));
        root.addContent(getBasePathDef(DERIVED, derivedroot.getAbsolutePath(), null));



        DataTable.TupleGetter tg = dt.getTupleGetter("path", "name" , "taxid", "symbol", "syns");
        
        HashSet<Integer> taxIds = new HashSet<Integer>();
        
        for (int i=0; i<dt.getSize(); i++)
        {
            Tuple t = tg.get(i);

           // System.out.println("process next .: " +t);
            String path = (String)t.get(0);
            String orgname = (String)t.get(1);
            String symbol = (String) t.get(3);
            int taxid  = (Integer)t.get(2);
            
            if ( taxIds.contains(taxid))
            {
                log.warn(taxid + " already contained. Skip.");
                continue;
            }
            taxIds.add(taxid);
            
            String syns = (String)t.get(4);

            File orgroot = new File(rootdir, path);

            File droot = new File(derivedroot, path);

            droot.mkdirs();

            if (!droot.exists() || !droot.isDirectory())
            {
                log.error("could not create derived directory: %s", droot.getAbsolutePath());
                continue;
            }

            File dna = new File(orgroot, "dna");
           if (!check_dir(log, dna, "genome"))
               continue;

            File pep = new File(orgroot, "pep");
            if (!check_dir(log, pep, "protein"))
                continue;

            File cdna = new File(orgroot, "cdna");
            if (!check_dir(log, cdna, "cdna"))
                continue;


            //check for the genome version
            String genome_prefix = null;
            String genome_version = null;
            for (File f : FileUtils.getFiles(dna, ".*\\.dna.*\\.fa.*"))
            {
                String prefix = f.getName().split(".dna")[0];
                String sp[]  = prefix.split("\\.");
                if (sp.length<2)
                {
                    log.warn("could not figure out the dna version from prefix: "+prefix+" file: "+f.getName());
                    continue;
                }
                genome_prefix = sp[0];
                genome_version = sp[1];
                break;
            }

            log.info("org: %s genome version: %s", orgname, genome_version);
            //look for the gtf file
            String gtf_prefix = genome_prefix+"."+genome_version;
            File gtf_file = null;
            Vector<File> gtfs = new Vector<File>();

            for (File f : FileUtils.getFiles(orgroot, ".*\\.gtf"))
            {
                if (!f.getName().startsWith(gtf_prefix))
                    continue;

                gtfs.add(f);
                gtf_file = f;
            }

            if (gtf_file==null)
            {
                log.error("no gtf found for "+path);
                continue;
            }
            if (gtfs.size()>1)
            {
                log.error("multiple gtf found for: %s\n", path);
            }

            File[] pepfiles = FileUtils.getFiles(pep, ".*\\.all\\.fa(\\.[g|b]z2?)?");
            if (pepfiles.length==0)
            {
                log.error("no protein  file found in %s", pep.getAbsolutePath());
                continue;
            }
            if (pepfiles.length > 1)
            {
                log.warn("multiple protein files (%d) found", pepfiles.length);
            }

            File[] cdnafiles = FileUtils.getFiles(cdna, ".*\\.all\\.fa(\\.[g|b]z2?)?");
            if (cdnafiles.length==0)
            {
                log.warn("no transcriptome file found in %s", cdna.getAbsolutePath());
                continue;
            }
            if (cdnafiles.length > 1)
            {
                log.warn("multiple cnda files (%d) found", cdnafiles.length);
            }

            log.info("%s taxid: %d got gtf: %s", path, taxid, gtf_file.getName());
            Element orgE = new Element(ORG_ELEM);

            String baseref = orgroot.getName()+"_base";
            orgE.addContent(getBasePathDef(baseref, path, GENOMIC));

            String derivedref = orgroot.getName()+"_derived";

            orgE.addContent(getBasePathDef(derivedref, path, DERIVED));

            orgE.setAttribute("path", path);
            orgE.setAttribute("taxid", ""+taxid);
            orgE.setAttribute("name", orgname);
            if (symbol!=null)
            {
                orgE.setAttribute("symbol", symbol);
            }
            if (syns!=null)
            {
                orgE.setAttribute("syns", syns);
            }

            orgE.addContent(getPathDef(GENOME_LOOKUP_ELEM, GENOME_LOOKUP, derivedref));
            orgE.addContent(getPathDef(GTF_GENE_EXTRACT_ELEM, gtf_file.getName()+GTF_GENE_EXTRACT, derivedref));
            orgE.addContent(getPathDef(GTF_GENE_EXTRACT_LOOKUP_ELEM, gtf_file.getName()+GTF_GENE_EXTRACT+GTF_LOOKUP_SUFFIX, derivedref));

            orgE.addContent(getPathDef(GTF_LOOKUP_ELEM, gtf_file.getName()+GTF_LOOKUP_SUFFIX, derivedref));
            orgE.addContent(getPathDef(GTF_GENE2PROTEN_ELEM, gtf_file.getName()+GTF_GENE2PROTEIN_LOOKUP_SUFFIX, derivedref));
            orgE.addContent(getPathDef(GTF_GENE2TR_ELEM, gtf_file.getName() + GTF_GENE2TR_LOOKUP_SUFFIX, derivedref));
            orgE.addContent(getPathDef(GTF_NAME2GENE_ELEM, gtf_file.getName() + GTF_NAME2GENE_LOOKUP_SUFFIX, derivedref));
            

            orgE.addContent(getPathDef(DNA_ELEM, getRelativeRef(orgroot, dna), baseref)); 

            orgE.addContent(getPathDef(INDEXDIR_ELEM, getRelativeRef(orgroot, new File(orgroot, "index")), derivedref)); 
            orgE.addContent(getPathDef(BLASTDB_ELEM, getRelativeRef(orgroot, new File(orgroot, "blastdb")), derivedref)); 

            orgE.addContent(getPathDef(GTF_ELEM, getRelativeRef(orgroot, gtf_file), baseref)); 
            orgE.addContent(getPathDef(PROTEIN_ELEM, getRelativeRef(orgroot, pepfiles[0]), baseref)); 
            orgE.addContent(getPathDef(TRANSCRIPTOME_ELEM, getRelativeRef(orgroot, cdnafiles[0]), baseref)); 

            root.addContent(orgE);

        }

        return root;
    }

    public static HashSet<Integer> getTaxIds(Collection<Organism> orgs)
    {
        HashSet<Integer> rv = new HashSet<>();
        for (Organism org: orgs)
        {
            rv.add(org.getTaxId());
        }
        return rv;
    }
    public static Collection<Organism> getAllOrganisms()
    {
        return tax2org.values();
    }

    public static Organism readOrganism(File xml, String key)
    {
        return new Organism(readOrganismElement(xml,key));
    }

    public static Element readOrganismElement(File xml, String key)
    {
        Element nroot = new Element("config");
        boolean found = false;

        try
        {
            int taxid = -1;
            try
            {
                taxid = Integer.parseInt(key);
            }
            catch (NumberFormatException nfe)
            {
                //silent error
            }
            XMLStreamReader xsr = XMLUtils.infactory.createXMLStreamReader(new FileReader(xml));

            while (xsr.hasNext())
            {
                int event = xsr.next();
                if (event != XMLStreamConstants.START_ELEMENT)
                {
                    continue;
                }
                if (xsr.getLocalName().equals("config"))
                    continue;

                if (!xsr.getLocalName().equals(ORG_ELEM))
                {
                    nroot.addContent(XMLUtils.readElemFromStream(xsr));
                    continue;
                }

                if (found)
                    continue;



                found = (taxid>=0 && taxid == XMLUtils.getIntAttrib(xsr, "taxid", -1));

                if (!found)
                {
                    found = key.equals(XMLUtils.getAttrib(xsr, "symbol", "", false));
                }

                String path = XMLUtils.getAttrib(xsr, "path", null, true);
                String path_sp[] = path.split(File.separator);
                if (path_sp[path_sp.length -1 ].equals(key))
                {
                    found = true;
                }

                if (!found)
                {
                    FileUtils.RecordIterator rit = new FileUtils.RecordIterator('|');
                    rit.setLine(XMLUtils.getAttrib(xsr, "syns", "", false));
                    while (!found && rit.next())
                    {
                        found = key.equals(rit.getRecord());
                    }

                }
                if (!found)
                    continue;

                nroot.addContent(XMLUtils.readElemFromStream(xsr));

            }
            xsr.close();
        }
        catch (IOException ie)
        {
            throw new FRuntimeException("i/o error: %s while reading from %s and looked for %s", ie, ie.getMessage(), xml.getAbsolutePath(), key);
        }
        catch (XMLStreamException xse)
        {
            throw new FRuntimeException("xml error: %s while reading from %s and looked for %s", xse, xse.getMessage(), xml.getAbsolutePath(), key);
        }
        if (!found)
        {
            throw new FRuntimeException("cannot find organism in: %s, looked for %s", xml.getAbsolutePath(), key);
        }
        nroot = XMLUtils.resolveAll(nroot);

        return nroot.getChild(ORG_ELEM);

    }
    public static Collection<Organism> readOrganisms(File xml)
    {
        return readOrganisms(XMLUtils.getResolvedRootElement(xml));
    }
    public static Collection<Organism> readOrganisms(Element root)
    {
        Logger log  = LogConfig.getLogger();
        //log.info("caller stack: %s", ExceptionManager.logCallerStack());
        for (Element e: XMLUtils.getChildren(root, ORG_ELEM))
        {
            Organism o = new Organism(e);
            //System.out.println(XMLUtils.prettyXML(e));
        }

        log.info("read %d organisms with %d aliases", tax2org.size(), lookup.size());
        log.debug("aliases: %s", lookup.keySet());
        return tax2org.values();
    }


    public String getPath()
    {
        return path;
    }
    public int getTaxId()
    {
        return taxid;
    }

    public String getSymbol()
    {
        return symbol;
    }
    public String getName()
    {
        return name;
    }
    public String getShortName()
    {
        return getShortName(false);
    }
    public String getShortName(boolean abbreviation){
        String[] tokens =  name.split("\\s+");
        if ( abbreviation)
        {
            int length = Math.min(3,tokens[1].length());
            return tokens[0].charAt(0) + tokens[1].substring(0, length).toUpperCase();
            
        }
        return tokens[0].charAt(0) + ". " + tokens[1];
    }
    public static Collection<Organism> getOrganisms()
    {
        return tax2org.values();
    }
    public static Organism getOrganism(String name)
    {
        return getOrganism(name, false);

    }

    public static Organism getOrganism(String name, boolean error_if_missing)
    {
        Logger log = LogConfig.getLogger();
        log.debug("check org >%s<\ncaller: %s", name, ExceptionManager.logCallerStack());
        Organism o = lookup.get(name);
        if (o!=null)
            return o;

        if (null != (o = lookup.get(name.toLowerCase())))
            return o;


        try
        {
            System.err.printf("get with taxid: >%s<\n", name);
            o = getOrganism(Integer.parseInt(name.trim()), false);
            if (o != null)
                return o;
        }
        catch (NumberFormatException nfe)
        {
            
            //silent error, was just a try
        }
        if (error_if_missing)
        {
            throw new RuntimeException("unknown organism >"+name+"< requested (got "+lookup.size()+" known lookups to "+tax2org.size()+" organisms");
        }
        return null;

    }

    public static Organism getOrganism(int taxid)
    {
        return getOrganism(taxid, false);
    }
    public static Organism getOrganism(int taxid, boolean error_if_missing)
    {
        Organism o = tax2org.get(taxid);
        if (!error_if_missing || o !=null)
            return o;

        throw new RuntimeException("unknown organism taxid: "+taxid+" requested (got "+lookup.size()+" known lookups to "+tax2org.size()+" organisms");

    }

    public void writeProteinFasta(PrintWriter out)
    {
        for (String gn : getAnnot().getGenesByType("protein_coding"))
        {
            Gene g = getAnnot().getGene(gn);
            for (Protein p : g.proteins.values())
            {
                String seq = GenomicUtils.getProteinSeq(p, getGenome(), false);
                out.printf(">%s:%s\n%s\n", p.id, g.id, seq);

            }
        }
    }


    public static void writeRepresentatives(File outf, File outdir, int fastastep, Organism org, Collection<String> protids)
    {
        Logger log = LogConfig.getLogger();

        if (outf == null && outdir == null)
        {
            throw new RuntimeException("invalid input need outf or outdir");
        }
        log.info("start extracting %d representative protids for %s", protids.size(), org.getName());
        int nentries = 0;
        int fastaidx = 1;
        PrintWriter pw = (outf != null) ? FileUtils.getWriter(outf) : FileUtils.getWriter(outdir, "fasta_%d.fa", fastaidx);

        GTFAnnotation annot = org.getAnnot();

        for (String s: protids)
        {
            String gid = annot.getGeneIdByProteinId(s);
            if (gid == null)
            {
                System.err.println("unknown gene to prot "+s+" in " +org.getName());
                continue;
            }
            Protein p = annot.getGene(gid).proteins.get(s);
            pw.printf(">%s\n", s);
            String seq = GenomicUtils.getProteinSeq(p, org.getGenome(), false);
            fasta.FastaReader.writeFasta(seq, pw);
            nentries++;

            if (nentries % 100 == 0)
            {
                log.info("after extracting %d/%d representative protids for %s", nentries, protids.size(), org.getName());
            }

            if (outdir!=null && fastastep>0 && nentries % fastastep == 0)
            {
                pw.close();
                fastaidx++;
                pw = FileUtils.getWriter(outdir, "fasta_%d.fa", fastaidx);
            }

        }

        pw.close();



    }
    public String toString()
    {
        return name;
    }

    public static void main (String args[]) 
    {
        LogConfig lc = LogConfig.configureLog(null, true);

        SimpleOptionParser cmd = new SimpleOptionParser("root", "derivedroot", "o", "outxml", "orglist", 
                                                        "xml", "orgs", "org", "extract", "writeproteins", "version",
                                                        "overview", "od", "path", "checkpeptides", "protset", "fastastep",
                                                        "ovgene", "getpath");

        cmd.setFile("orglist", "xml", "protset");
        cmd.setDir("root", "derivedroot", "od");
        cmd.setInt("fastastep");
        cmd.setOptional("orglist","orgs", "root", "outxml", "xml", "derivedroot", "org", "writeproteins", "od", "path", "protset", "fastastep", "o", "ovgene", "version");
        cmd.setSwitches("extract", "overview", "checkpeptides", "writeproteins", "getpath");

        if (!OptionParser.parseParams(args, false, true, true, cmd, lc))
            return;

        if (args.length == 0)
        {
         //   System.out.println(cmd.getHelp());
        }


        if (cmd.isSet("getpath"))
        {
            if (!cmd.isOptionSet("xml") || !cmd.isOptionSet("org"))
            {
                System.err.println("please set -xml <file> and -org <name, syononym or taxid>");
                return;
            }
            Organism.readOrganisms(cmd.getFile("xml"));
            Organism o = Organism.getOrganism(cmd.getValue("org"));
            System.out.println(o.getPath());
            return;
        }
        if (cmd.isOptionSet("path"))
        {
            if (!cmd.isOptionSet("xml") || !cmd.isOptionSet("org"))
            {
                System.err.println("please set -xml <file> and -org <name, syononym or taxid>");
                return;
            }
            String version = cmd.getOptionalValue("version", "");
            Element e = Organism.readOrganismElement(cmd.getFile("xml"), cmd.getValue("org"));

            try
            {
                System.out.printf("%s\n", Organism.getPath(e, cmd.getValue("path"), version, true).getAbsolutePath());
                
            }
            catch (Exception ex)
            {
                System.err.println(ex.getMessage());
                return;
            }
            

            return;
        }

        if (cmd.isOptionSet("xml"))
        {
            for (Organism o : Organism.readOrganisms(XMLUtils.getResolvedRootElement(cmd.getFile("xml"))))
            {
                //System.out.println(o.genome_dir.getParent());

                GTFAnnotation annot = o.getAnnot(false);
                o.getGenome();
            }

        }
        if (cmd.isOptionSet("outxml"))
        {
            Element e = createOrganismsXML(cmd.getFile("root"), cmd.getFile("derivedroot"), cmd.getFile("orglist"));
            PrintWriter pw = cmd.getWriter("outxml");
            pw.println(XMLUtils.prettyXML(e));
            pw.close();
          //  System.out.println(XMLUtils.prettyXML(e));
        }
        Vector<Organism> orgs = new Vector<Organism>();

        if (cmd.isOptionSet("org"))
        {
            Organism o = Organism.getOrganism(cmd.getValue("org"));
            
            orgs.add(o);

            if (cmd.isOptionSet("ovgene"))
            {
                String sp[] = cmd.getValue("ovgene").split("[:-]");

                if (sp.length != 3)
                {
                    System.err.printf("invalid ovgene need in form chr:start-end got: >%s<\n", cmd.getValue("ovgene"));
                    return;
                }
                String chr = sp[0];
                int start = Integer.parseInt(sp[1]);
                int end = Integer.parseInt(sp[2]);
                System.out.printf("will get genes overlapping region: %s %d-%d\n", chr, start, end);

                Iterator<Gene> git = o.getAnnot().getGenes(chr, start, end);
                while (git.hasNext())
                {
                    Gene g = git.next();
                    System.out.printf("%s (%c): %s: %d-%d\n", g.id, (g.strand) ? '+' : '-', g.chr, g.start, g.end);
                }
                return;
            }


            if (cmd.isSet("writeproteins"))
            {


                if (cmd.isOptionSet("protset"))
                {
                    HashSet<String> pset = FileUtils.readSet(cmd.getFile("protset"));
                    Organism.writeRepresentatives(cmd.getOptionalFile("o"), cmd.getOptionalFile("od"), cmd.getOptionalInt("fastastep", -1), o, pset);
                    return;
                }

                PrintWriter pw = cmd.getWriter("o");
                o.writeProteinFasta(pw);
                pw.close();
                return;
            }
        }
        if (cmd.isOptionSet("orgs"))
        {
            for (String s : FileUtils.readSet(cmd.getFile("orgs")))
            {
                Organism o = Organism.getOrganism(s);
                if (o==null)
                {
                    System.err.println("cannot find organismus named: "+s);
                    continue;
                }
                orgs.add(o);
            }
        }
        if (orgs.size() == 0)
        {
            orgs.addAll(Organism.getAllOrganisms());
        }
        File od = cmd.getOptionalFile("od");


        for (Organism o : orgs)
        {
            if (cmd.isSet("checkpeptides"))
            {
                HashMap<String, HashSet<String>> pep2gn = new HashMap<String, HashSet<String>>();

                //System.out.println("check: %s %s\n", o.getName(), o.getAnnot());
                Collection<String> genes = o.getAnnot().getGenesByType("protein_coding");
                int np = 0;
                for (String gn : genes)
                {
                    Gene g = o.getAnnot().getGene(gn);
                    for (Protein p : g.proteins.values())
                    {
                        String pseq = GenomicUtils.getProteinSeq(p, o.getGenome(), false);
                        Vector<String> peptides = protein.DigeUtils.digeTryptic(pseq, true, 6);

                        //System.out.printf("%s %s: %d: %s\n", g.id, p.id, peptides.size(), peptides);
                        for (String ps : peptides)
                        {
                            HashSet<String> hs = pep2gn.get(ps);
                            if (hs == null)
                            {
                                pep2gn.put(ps, hs = new HashSet<String>());
                            }
                            hs.add(g.id);
                        }

                    }
                    if (++np % 100 ==0)
                    {
                        System.out.printf("processed %d/%d genes\n", np, genes.size());
                    }
                }
                int maxoverlap = NumUtils.max(ObjectGetter.map(pep2gn.values(), (HashSet<String> h) -> h.size()));
                System.out.printf("max overlap: %d, %d peptides\n", maxoverlap, pep2gn.size());
                final double norm = 100.0/pep2gn.size();
                for (int i=1; i<=maxoverlap; i++)
                {
                    final int totest = i;
                    int eqn = ObjectGetter.filter(pep2gn.values(), (HashSet<String> h) -> h.size()==totest).size();
                    int n = ObjectGetter.filter(pep2gn.values(), (HashSet<String> h) -> h.size()>totest).size();
                    System.out.printf("%d peps with =%d  (%d with >%d) genes (%2.2f%% / %2.2f%%)\n", eqn, i, n, i,
                                      norm*eqn, norm*n);
                }
                continue;
            }
            if (cmd.isSet("overview") && od!=null)
            {
                System.out.printf("o: %s\n", o);
                PrintWriter pw = FileUtils.getWriter(od, "%s.info", o.getTaxId());

                pw.printf("gene\tbiotype\tchr\tisoforms\tprotisoforms\tmaxnumexons\tnovelintronisoforms\tnosubisoforms\n");
                GTFAnnotation gtf = o.getAnnot();

                Iterator<Gene> it = gtf.getGenes();
                while (it.hasNext())
                {
                    Gene g = it.next();
                    int maxexons = 0;
                    Protein lead = null;

                    for (Protein p : g.proteins.values())
                    {
                        maxexons = Math.max(maxexons, p.getNumExons());
                        if (lead!=null && lead.getNumExons() > p.getNumExons())
                            continue;

                        lead = p;
                    }

                    int novel_intron_prot = g.getNovelIntronIsoforms(lead).size();
                    int n_nosub = g.getNonSubProts(lead).size();

                    pw.printf("%s\t%s\t%s\t%d\t%d\t%d\t%d\t%d\n", g.id, g.biotype, g.chr, g.transcripts.size(), g.proteins.size(), maxexons, novel_intron_prot, n_nosub);

                }
                pw.close();

            }




            if (cmd.isSet("extract"))
            {
                o.extractGenes();

            }
            if (o!=null)
            {
                System.out.println(o.getName()+" got "+ o.getAnnot().getGenesByType("protein_coding").size()+" protein coding genes");

            }
            else
            {
                System.out.println("no organism selected help: ");
                System.out.println(cmd.getHelp());
            }

            /*
            for (String g : o.getAnnot().getGeneNames())
            {
                
                String gs=o.getGeneSeq(g);
                GTFAnnotation annot = o.getAnnot(true);
                Gene gg = annot.getGene(g);
                if (!gg.biotype.equals("protein_coding"))
                {
                    continue;
                }
                GenomicRegion gr = annot.getGeneRegion(g);

                System.out.printf("gene: %s %s(%c) %d-%d region: %s %s(%c) %d-%d\n", gg.id, gg.chr,  (gg.strand)?'+':'-', gg.start, gg.end,
                                  g, gr.chr, (gr.strand)?'+':'-', gr.getX1(),gr.getX2());

                //gs = o.getGenome().getGenomeSeq(gr.getChr(), gr.getX1(), gr.getX2()+1, gr.getStrand());
                String gexseq = GenomicUtils.getGeneSeq(gg, o.getGenome());
                /*for (int i=0; i<gexseq.length() && i<gs.length(); i+=80)
                {
                    int e1 = Math.min(i+80, gs.length());
                    int e2 = Math.min(i+80, gexseq.length());

                    System.out.printf(">%d-%d\n%s\n%s\n\n", i, i+80, gs.substring(i, e1), gexseq.substring(i, e2));
                }
                

                for (Protein p : gg.proteins.values())
                {
                    RegionVector rv = p.getCodingRegionVector(false);
                    
                    String seq = GenomicUtils.getSplicedSeq(gs, rv, false);

                    

                    
                    String check = annot.getProteinSeq(p.id);
                    System.out.printf("gexeq length: %d\n", gexseq.length());
                    System.out.printf("%s (%c) length: %d gseq lenght: %d: %s\n%s\ncheck:\n%s\ngs:\n%s\n", gg.id, (gg.strand)?'+':'-', gg.end-gg.start+1, gs.length(), p.id, seq, check,
                                      GenomicUtils.getSplicedSeq(gexseq, rv, false));
                    
                }
                
            }
            */
        }

    }
}
