 package genome;

 import lmu.utils.*;
 import java.util.*;

 public class Protein
 {
     static Logger log = null;

     public String id;
     public Transcript parentTranscript = null;
     //offset of first translated position in the first exon
     public int transl_start;
     //negative offset of last translated position in the last exon
     public int transl_end;
     public String start_exon = null;
     public String end_exon = null;
     Vector<Pair<Integer,Integer> > v;
     

     static synchronized Logger getLogger()
     {
         if (log!=null)
             return log;
         
         return log = LogConfig.getLogger();
     }
     
     public int getCDSLength()
     {
         int l = 0;
         for (CodingExon ce : coding_exons)
         {
             l+= ce.length;
         }
         return l;
     }
     public CodingExon getExon(int i)
     {
         return coding_exons.get(i);
     }

     public int getNumExons()
     {
         return coding_exons.size();
     }
     public  Vector<CodingExon> coding_exons=null;
     
     public Protein(Protein p)
     {
         this.id=p.id;
         this.parentTranscript=p.parentTranscript;
         this.transl_start=p.transl_start;
         this.transl_end=p.transl_end;
         this.start_exon=p.start_exon;
         this.end_exon=p.end_exon;
         this.v=p.v;
         this.sequence=p.sequence;
         if (p.coding_exons!=null)
         {
             this.coding_exons=new Vector<CodingExon>();
             this.coding_exons.addAll(p.coding_exons);
         }
     }

     

     public RegionVector getCodingRegionVector(boolean global)
     {
         
         //Logger log = getLogger();
         //log.error("%s:%s thread: %s stack:\n%s\n", id, parentTranscript.parentGene.id, Thread.currentThread().getId(), ExceptionManager.logCallerStack());
         RegionVector l = new RegionVector();
         for (CodingExon ce : coding_exons)
         {
             l.addRegion(ce.start_in_gene, ce.end_in_gene+1);
         }

         return (global)?GenomicUtils.toGlobal(l,parentTranscript.parentGene):l;
     }

     public Vector<Pair<Integer,Integer> > getGenomeRegion()
     {
         return v;
     }

     public void init()
     {
         v = new Vector<Pair<Integer,Integer> >();

         
         coding_exons = new Vector<CodingExon>();

         boolean rev = !parentTranscript.parentGene.strand;
         boolean gotfirst=false;

         
         
         /*log.debug("%s parentT: %s exons: %d start: %s end: %s transstart: %d transend: %d\n", 
                            id, parentTranscript.id, parentTranscript.getExons().size(), 
                            start_exon, end_exon, transl_start, transl_end);
         */
         int trpos=0;
         int prot_pos=0;
         for (Exon ex : parentTranscript.getExons())
         {
             boolean f =ex.id.equals(start_exon);
             boolean e = ex.id.equals(end_exon);

             
             if (false && f && ex.phase!=null)
             {
                 int abs_phase = (ex.phase+3)%3;

                 //System.out.printf("%s will correct transl start for first exon: phase: %d trans start: %d (current trpos: %d)\n", 
                 //                  id, ex.phase, trpos+transl_start, trpos);
                 while ((trpos+transl_start)%3!=abs_phase)
                 {
                     //System.out.printf("\t++\n");
                     transl_start++;
                 }
             }
             //System.out.printf("%s f: %s e: %s %d-%d (%d-%d)\n", ex.id, ""+f, ""+e, ex.abs_start, ex.abs_end, trpos , trpos+ex.getLength());
             Pair<Integer,Integer> p = new Pair<Integer,Integer>(ex.abs_start,ex.abs_end);
             if (f)
             {
                 if (rev)
                 {
                     p.setSecond(ex.abs_end-transl_start);
                 }
                 else
                 {
                     p.setFirst(ex.abs_start+transl_start);
                 }
                 
             }

             if (e)
             {
                 //substract stop codon
                 /**/
                 if (rev)
                 {
                     p.setFirst(ex.abs_end-transl_end);
                 }
                 else
                 {
                     p.setSecond(ex.abs_start+transl_end);
                 }
                 /**/
                 
             }
             /*
             if (f && e)
             {
                 if(rev)
                 {
                     p.setFirst(ex.abs_end-transl_end);
                     p.setSecond(ex.abs_end-transl_start);
                 }
                 else
                 {
                     p.setFirst(ex.abs_start+transl_start);
                     p.setSecond(ex.abs_start+transl_end);
                 }
             }
             */
             if (f)
             {
                 gotfirst=true;
             }

             trpos+=ex.getLength();
             if (!gotfirst)
             {
                 continue;
             }
                 
             CodingExon ce = new CodingExon(p, this,parentTranscript, ex, trpos-ex.getLength(), prot_pos);


             
             //System.out.printf("REGION: %d-%d\n", p.getFirst(), p.getSecond());
             v.add(p);
             if (ce.length>0)
             {
                 prot_pos+=ce.length;
                 coding_exons.add(ce);
             }
             
             
             if (e)
                 break;
             
             

         }

     }
     public Organism getOrganism()
     {
    	 return this.parentTranscript.parentGene.getOrganism();
     }
     public String sequence = null;

     public Protein(Transcript t, String id)
     {
         this.id=id; parentTranscript=t;
     }
     public Protein(Transcript t, String id, String start_exon, String end_exon, int transl_start, int transl_end)
     {
         this.id=id; parentTranscript=t;
         this.start_exon=start_exon;
         this.end_exon=end_exon;
         this.transl_start=transl_start;
         this.transl_end=transl_end;
         init();
     }
     @Override
     public int hashCode()
     {
    	 return toString().hashCode();
     }
     @Override
     public boolean equals(Object o)
     {
         if ( o instanceof Protein)
         {
             if ( ((Protein) o).id.equals(this.id)) return true;
         }
    	 return false;
     }
     public String toString()
     {
         return String.format("%s (t: %s g: %s) from %s (%d) to %s (%d)", id, parentTranscript.id, parentTranscript.parentGene.id,
                              start_exon, transl_start, end_exon, transl_end);
     }

 }
