package genome.config;

import java.io.File;

import org.jdom.Element;

import config.Module;
import lmu.utils.ConfigFileGetter;
import lmu.utils.LogConfig;
import lmu.utils.Logger;
import lmu.utils.XMLUtils;
import lmu.utils.ConfigFileGetter.ConfigFileNotFound;

public class GenomicConfigManager  {
    
    //duplicated from ConfigManager ...
    static String default_data_config="datasource.config";

    static Boolean inited = false;

    static public File getDataSourceConfig()
    {
        try
        {
            return ConfigFileGetter.getConfigFile(default_data_config);

        }
        catch (ConfigFileNotFound e)
        {
            return null;
        }

    }

    public static void init()
    {
        synchronized(inited)
        {
            if (inited)
                return;
            
            init(getDataSourceConfig());
        }
    }
    
    public static void init(File f)
    {

        synchronized(inited)
        {
            if (inited)
                return;

            Logger log = LogConfig.getLogger();

            Element root = XMLUtils.getResolvedRootElement(f);

            for (Element e: XMLUtils.getChildren(root))
            {
                log.debug("got next elem: "+e.getName());
                Module m = Module.get(e.getName());

                if (m == null)
                    continue;
                
                
                switch (m)
                {
                    case ORGXML:
                        init(e);
                        break;

                }
            }
        }
    }
    
    public static void init(Element e)
    {
        synchronized(inited)
        {
            if (inited)
                return;

            Logger log = LogConfig.getLogger();

            System.out.println(XMLUtils.prettyXML(e));
            
            File orgxml = XMLUtils.readPath(e);
            log.debug("load organisms from "+orgxml.getAbsolutePath());
            
            genome.Organism.readOrganisms(orgxml);
            inited = true;
        }
    }
    
    public static genome.Organism getGenomicOrganism(int id)
    {
        init();
        return getGenomicOrganism(id, false);
    }
    public static genome.Organism getGenomicOrganism(int id, boolean error_if_missing)
    {
        init();
        return getGenomicOrganism(""+id, error_if_missing);
    }
    public static genome.Organism getGenomicOrganism(String id, boolean error_if_missing)
    {
        init();
        return genome.Organism.getOrganism(id, error_if_missing);
    }
    
    public static genome.Organism getGenomicOrganism(String id)
    {
        return getGenomicOrganism(id, false);
    }

   
}
