package protein;

import lmu.utils.*;
import java.util.*;

public class DigeUtils 
{

    public static final char TRYPTIC_SITES[]={'K', 'R'};


    public static Vector<String> digeTryptic(final String protseq, boolean removeX, int minsize)
    {
        return digeTryptic(protseq, 0, removeX,minsize);
    }

    public static Vector<String> digeTryptic(final String protseq, int maxskip, boolean removeX, int minsize)
    {
        Vector<String> vs= dige(protseq, maxskip, TRYPTIC_SITES);
        if (minsize<=1 && ! removeX)
        {
            return vs;
        }
        Vector<String> filtered = new Vector<String>();
        for (String s : vs)
        {
            String qs = (removeX)?cutX(s):s;
            if (removeX && qs.indexOf('X')>=0)
            {
                continue;
            }
            if (qs.length()<minsize)
            {
                continue;
            }
            filtered.add(qs);
        }
        return filtered;
    }

    public static Vector<String> digeTryptic(final String protseq)
    {
        return digeTryptic(protseq,false, 0);
    }

    static boolean is_stop(char ch, char ...sites)
    {
        for (char sch : sites)
        {
            if (sch == ch)
                return true;
        }
        return false;
    }


    public static Vector<String> dige(final String protseq, int num_skipped, char ... sites)
    {
        Vector<String> rv = new Vector<String>();
        Vector<Integer> cut_sites = new Vector<>();
        cut_sites.add(0);
        if (num_skipped > 0)
        {
            cut_sites.add(1);
        }

        final int N = protseq.length() -1;
        for (int i=0; i<= N; i++)
        {

            if(i != N && !is_stop(protseq.charAt(i), sites))
                continue;

            final int end = i;
            for (int j = cut_sites.size()-1, k=0; j >= 0 && k <= num_skipped; j--, k++)
            {
                int start = cut_sites.get(j);
                rv.add(protseq.substring(start, i+1));
            }
            cut_sites.add(i+1);
        }

        return rv;
    }

    public static Vector<String> dige(final String protseq, char ... sites)
    {
        return dige(protseq, 0, sites);
    }

    
    public static  Pair<Integer,Integer> getXCut(String protseq)
    {
         int startx=0;
        while (startx<protseq.length() && protseq.charAt(startx)=='X')
        {
            startx++;
        }
        int endx=protseq.length()-1;  
        while (endx>=0 && protseq.charAt(endx)=='X')
        {
            endx--;
        }
        return new Pair<Integer,Integer>(startx,endx);
    }
    public static String cutX(String protseq)
    {
        Pair<Integer,Integer> p = getXCut(protseq);
        if (p.getFirst()>=p.getSecond())
        {
            return "";
        }
        return protseq.substring(p.getFirst(), p.getSecond()+1);
    }

    
}
