package protein.algo.external;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public 	class MAFFT extends MultipleAlignmentAlgo{
	private File algo;
	public MAFFT(File algo){
		this.algo = algo;
	}
	private static Pattern MAFFT = Pattern.compile("([\\w+\\.\\:-]+)\\s+([\\w-]+)");
	@Override
	public StringBuffer[] align(File sequencesTmpFile, List<String> ids) throws ExternalException {
		List<String> command = new ArrayList<String>();
		command.add(algo.getAbsolutePath());
		command.add("--clustalout");
		command.add("--auto");
		command.add("--quiet");
		command.add(sequencesTmpFile.getAbsolutePath());
		StringBuffer[] output = new StringBuffer[ids.size()];
		for(int i = 0 ; i < ids.size() ; i ++){
			output[i] = new StringBuffer();
		}
		ProcessBuilder builder = new ProcessBuilder(command  );
		builder.redirectErrorStream(true);
		try
		{
			Process process = builder.start();

			BufferedReader reader = new BufferedReader (new InputStreamReader(process.getInputStream()));

			String line = null;

			int id = 0;
			while ((line = reader.readLine ()) != null)
			{   //System.out.println(line);
				Matcher match = MAFFT.matcher(line);
				if ( match.matches()){
				    
					String alignment = match.group(2);
					output[id].append(alignment);
					//System.out.println(">" + line);
					//System.out.println(id + " " + alignment);
                    
					id = (++id)%ids.size();
				}else{
				}
			}

			reader.close();

			process.waitFor();


			int exitValue = process.exitValue();

			if ( exitValue != 0)
			{
				throw new ExternalException("exit with return value != 0");

			}

		}
		catch (IOException e)
		{
			throw new ExternalException("i/o error: "+e.getMessage()+" while running mul. ali", e);
		}
		catch (InterruptedException iex)
		{
			throw new ExternalException("run interrupted: "+iex.getMessage(), iex);
		}
		return output;
	}
    @Override
    public boolean isoformStructureSupport() {
        return false;
    }
	
}