package protein.algo.external;

import java.io.*;
import java.util.*;
import java.util.regex.*;
public class PRRN extends MultipleAlignmentAlgo{
	private File algo;
	private boolean local =true;
	private int structureScore = 10;
	
	public PRRN(File algo){
	    this(algo,true,10);
	}
	
	public PRRN(File algo,boolean local,int structureScore){
		this.algo = algo;
		this.local = local;
		this.structureScore = structureScore;
	}
	public static Pattern prrAliPattern = Pattern.compile("\\s+\\d+\\s+([\\w-]+)\\s*\\|\\s+([\\w+\\.\\:-]*)");
	
	@Override
	public StringBuffer[] align(File sequencesTmpFile, List<String> ids) throws ExternalException {
		List<String> command = new ArrayList<String>();
		command.add(algo.getAbsolutePath());
		if ( local) command.add("-L");
		command.add("-yJ"+structureScore);
		command.add(sequencesTmpFile.getAbsolutePath());
		
		
		StringBuffer[] output = new StringBuffer[ids.size()];
		HashMap<String,Integer> seqToIdx  = new HashMap<String,Integer> ();
        
		for(int i = 0 ; i < ids.size() ; i++){
			output[i] = new StringBuffer();
			seqToIdx.put(ids.get(i),i);
		}
		ProcessBuilder builder = new ProcessBuilder(command  );
		
		  File tabDir = new File(algo.getParentFile()+ "/../table");
          
          builder.environment().put("ALN_TAB",tabDir.getAbsolutePath() );
      
		
		builder.redirectErrorStream(true);
		StringBuffer out = new StringBuffer();
		try
		{
			Process process = builder.start();

			BufferedReader reader = new BufferedReader (new InputStreamReader(process.getInputStream()));

			String line = null;
			
			while ((line = reader.readLine ()) != null)
			{	out.append(line + "\n");
				Matcher match = prrAliPattern.matcher(line);
				if ( match.matches()){
					
				    
					String alignment = match.group(1).trim();
					String seqId = match.group(2).trim();
					//System.out.println(seqId + "\t" + alignment + " " +seqToIdx.get(seqId));
					output[seqToIdx.get(seqId)].append(alignment);
				}else{
				//	System.out.println("Skipped:" + line);
				}
			}

			reader.close();

			process.waitFor();

			int exitValue = process.exitValue();

			if ( exitValue != 0)
			{
				throw new ExternalException(out.toString());
			}

		}
		catch (IOException e)
		{
			throw new ExternalException("i/o error: "+e.getMessage()+" while running mul. ali", e);
		}
		catch (InterruptedException iex)
		{
			throw new ExternalException("run interrupted: "+iex.getMessage(), iex);
		}
		return output;
	}

    @Override
    public boolean isoformStructureSupport() {
        return true;
    }
	
}
