package protein.aminoacid;

import java.io.*;
import java.util.*;
/** represents a log-odd substitutionsmatrix 
 */
public class AminoAcidExchangeMatrix 
{

    static HashMap<String, AminoAcidExchangeMatrix> matrices = new HashMap<String, AminoAcidExchangeMatrix>();


    public static void register(String name, AminoAcidExchangeMatrix m)
    {
        matrices.put(name, m);
    }
    public static AminoAcidExchangeMatrix getMatrix(String name)
    {
        AminoAcidExchangeMatrix rv = matrices.get(name);
        if (rv!=null) return rv;
        
        File f = null;
        try
        {
            f = lmu.utils.ConfigFileGetter.getMatrixFile(AminoAcidExchangeMatrix.class, name);
        }
        catch(IOException e)
        {
            throw new RuntimeException("matrixfile not found for name:"+name+" error:"+e.getMessage(), e);
        }
        
        if (!f.exists())
        {
            throw new RuntimeException("cannot find matrixfile (name: "+name+" path "+f.getAbsolutePath()+")");
        }
        rv= new AminoAcidExchangeMatrix();
        rv.parseMatrixfile(f);
        matrices.put(name, rv);
        return rv;
    }
    long factor=1000;
    final static int MSIZE = 'Z'-'A'+1;

    boolean similarityMatrix;

    String matrixName;
    Double [][] M = new Double[MSIZE][MSIZE];

    long[][] lM = new long[MSIZE][MSIZE];

    public Double[] getRow(char c)
    {
        return M[c-'A'];
    }

    public long[] getLRow(char c)
    {
        if (c-'A'<0 || c-'A'>=lM.length)
        {
            throw new RuntimeException("invlaid character: "+c);
        }
        return lM[c-'A'];
    }
    public long getL(char c1, char c2)
    {
        return lM[c1-'A'][c2-'A'];
        
    }
    public Double get(char c1, char c2)
    {
        return M[c1-'A'][c2-'A'];
        
    }
    public long getFactor()
    {
        return factor;
    }



    void parseMatrixfile(File matrixFile)
    {
        boolean foundName = false;
        boolean foundScore = false;
        boolean foundColIndex = false;
        boolean foundRowIndex = false;
        boolean foundColNum = false;
        boolean foundRowNum = false;
        
        Vector<String> matrixLines = new Vector<String>();
        
        int numRow = -1;
        int numCol = -1;
        
        String line = null;
        
        String rowIndex=null;
        String colIndex=null;


        try
        {
    
            BufferedReader matrixReader = new BufferedReader(new FileReader(matrixFile));
            
            // parse matrix file
            while((line=matrixReader.readLine()) != null)
            {
                    if(line.startsWith("#"))
                        continue;
                     
                    if(line.startsWith("NAME"))
                    {
                        try
                        {
                                matrixName = line.split("\\s+")[1];
                        }
                        catch(ArrayIndexOutOfBoundsException ex)
                        {
                                throw new RuntimeException("Illegal NAME parameter in matrix file " + matrixFile + ", no name provided in " + line + "!");
                        }
                        
                        foundName = true;
                        continue;
                    }
                    
                    
                    if(line.startsWith("SCORE"))
                    {
                        try
                        {
                                String score = line.split("\\s+")[1];
                                
                                if(score.equalsIgnoreCase("similarity"))
                                {
                                        similarityMatrix = true;
                                }
                                else if(score.equalsIgnoreCase("distance"))
                                {
                                        similarityMatrix = false;
                                }
                                else
                                {
                                        throw new RuntimeException("Illegal SCORE parameter in matrix file " + matrixFile + ", no score parameter (similarity | distance) provided in " + line + "!");
                                }
                        }
                        catch(ArrayIndexOutOfBoundsException ex)
                        {
                                throw new RuntimeException("Illegal SCORE parameter in matrix file " + matrixFile + ", no score parameter (similarity | distance) provided in " + line + "!");
                        }
                        
                        foundScore = true;
                        continue;
                    }
                    
                    
                    if(line.startsWith("ROWINDEX"))
                    {
                        try
                        {
                                rowIndex = line.split("\\s+")[1];
                        }
                        catch(ArrayIndexOutOfBoundsException ex)
                        {
                                throw new RuntimeException("Illegal ROWINDEX parameter in matrix file " + matrixFile + ", no row index provided in " + line + "!");
                        }

                        foundRowIndex = true;
                        continue;
                    }
                    
                    
                    if(line.startsWith("COLINDEX"))
                    {
                        try
                        {
                            colIndex = line.split("\\s+")[1];
                        }
                        catch(ArrayIndexOutOfBoundsException ex)
                        {
                                throw new RuntimeException("Illegal COLINDEX parameter in matrix file " + matrixFile + ", no row index provided in " + line + "!");
                        }

                        foundColIndex = true;
                        continue;
                    }
                    
                    if(line.startsWith("NUMROW"))
                    {
                            try
                            {
                                    numRow = Integer.parseInt(line.split("\\s+")[1]);
                            }
                            catch(ArrayIndexOutOfBoundsException ex)
                            {
                                    throw new RuntimeException("Illegal NUMROW parameter in matrix file " + matrixFile + ", no number for rows provided in " + line + "!");
                            }
                            catch(NumberFormatException ex)
                            {
                                    throw new RuntimeException("Illegal NUMROW parameter in matrix file " + matrixFile + ", no numeric value provided for row number in " + line + "!");
                            }

                            foundRowNum= true;
                            continue;
                    }
                    
                    if(line.startsWith("NUMCOL"))
                    {
                        try
                        {
                                numCol = Integer.parseInt(line.split("\\s+")[1]);
                        }
                        catch(ArrayIndexOutOfBoundsException ex)
                        {
                                throw new RuntimeException("Illegal NUMCOL parameter in matrix file " + matrixFile + ", no number for cols provided in " + line + "!");
                        }
                        catch(NumberFormatException ex)
                        {
                                throw new RuntimeException("Illegal NUMCOL parameter in matrix file " + matrixFile + ", no numeric value provided for col number in " + line + "!");
                        }

                        foundColNum= true;
                    }
                    else if(line.startsWith("MATRIX"))
                    {
                        try
                        {
                                matrixLines.add(line);
                        }
                        catch(ArrayIndexOutOfBoundsException ex)
                        {
                               throw new RuntimeException("Illegal MATRIX parameter in matrix file " + matrixFile + ", no matrix entry provided in " + line + "!");
                        }
                    }
            }
            
            // close reader
            matrixReader.close();
        }
        catch(IOException e)
        {
            throw new RuntimeException("error:"+e.getMessage()+" while reading AA-matrix:"+matrixFile, e);
        }
        // check if all parameters were provided correctly
        if(!foundName)
        {
            throw new RuntimeException("NAME parameter missing in matrix file " + matrixFile + "!");
        }
        else if(!foundScore)
        {
            throw new RuntimeException("SCORE parameter missing in matrix file " + matrixFile + "!");
        }
        else if(!foundRowIndex)
        {
            throw new RuntimeException("ROWINDEX parameter missing in matrix file " + matrixFile + "!");
        }
        else if(!foundColIndex)
        {
            throw new RuntimeException("COLINDEX parameter missing in matrix file " + matrixFile + "!");
        }
        else if(!foundRowNum)
        {
            throw new RuntimeException("NUMROW parameter missing in matrix file " + matrixFile + "!");
        }
        else if(!foundColNum)
        {
            throw new RuntimeException("NUMCOL parameter missing in matrix file " + matrixFile + "!");
        }
        // check if number of matrix lines is the same as the specified row number
        else if(numRow != matrixLines.size())
        {
            throw new RuntimeException("Number of MATRIX entries ("+matrixLines.size()+") does not match specified number of rows " + numRow + " in matrix file: " + matrixFile + "!");
        }
        
        
        if (matrixLines.size()!=colIndex.length())
        {
            throw new RuntimeException(matrixFile+":: matrix lines "+matrixLines.size()+" and colIndex size: "+colIndex.length()+" differ!");
        }
        double expValue = 0;
        double v=0;
        int nRead=0;
        // parse matrix entries
        //System.out.printf("found "+matrixLines.size()+" mlines\n");
        for(int i=0; i<matrixLines.size(); i++)
        {
            String[] matrixEntries = ((String)matrixLines.get(i)).split("\\s+");
                
            if (matrixEntries.length>rowIndex.length()+1)
            {
                throw new RuntimeException(matrixFile+":: matrix entries ("+matrixEntries.length+" for row "+i+" and rowIndex size: "+rowIndex.length()+" differ!");
            }
            
            // for all entries: parse and store in the matrix
            for(int j=1; j<matrixEntries.length; j++)
            {
                    try
                    {
                        /*System.out.println("row: "+i+" read col: "+j+" M: "+M.length);
                        System.out.println("read colname: "+colIndex.charAt(j-1));
                        System.out.println("read rowname: "+rowIndex.charAt(i));
                        */

                        v = Double.parseDouble(matrixEntries[j]);
                        M[rowIndex.charAt(i)-'A'][colIndex.charAt(j-1)-'A'] = v;
                        
                        expValue+=v;
                        nRead++;
                    }
                    catch(NumberFormatException ex)
                    {
                          throw new RuntimeException("Matrix:" +matrixFile+" contains non numeric values ("+i+","+j+"): " + matrixEntries[j]);
                    }
            }
        }
        if (nRead==0)
        {
            throw new RuntimeException("no matrix value found for:" +matrixFile);
        }
        expValue/=nRead;
        
        // assign symmetric values to all Float.MINValue values
        for(int i=0; i<M.length-1; i++)
        {
            for(int j=0; j<M[i].length-1; j++)
            {
                if (M[i][j]==null && M[j][i]!=null)
                {
                    M[i][j]=M[j][i];
                    continue;
                }
                if (M[i][j]==null)
                {
                    M[i][j]=expValue;
                }
                
                
            }
        }
    
        for(int i=0; i<M.length-1; i++)
        {
            for(int j=0; j<M[i].length-1; j++)
            {
                lM[i][j]=(long)(factor*M[i][j]);
            }
        }
        /*
        for (int i=0; i<rowIndex.length(); i++)
        {
            for (int j=0; j<colIndex.length(); j++)
            {
                char c1=rowIndex.charAt(i); char c2=colIndex.charAt(j);
                System.out.printf("%c - %c: %2.3f %d\n", c1, c2, get(c1,c2), getL(c1,c2));
            }
        }
        */
    }
}
