package protein.aminoacid;

import java.util.HashMap;
import java.util.Vector;

import lmu.utils.FRuntimeException;
import lmu.utils.Pair;

public class AtomType
{
    public String label;
    public String full_name;
    public Double mono_mass;
    public Double avg_mass;

    public static HashMap<String, AtomType> atoms = new HashMap<String, AtomType>();
    
    public AtomType()
    {

    }
    public AtomType(String l)
    {
        this(l,null);
    }
    public AtomType(String label, String name)
    {
        this.label=label; this.full_name=name;
    }

    static public AtomType get(String label, String name)
    {
         return get(label,name,false);
    }
    static public AtomType get(String label)
    {
        return get(label,null);
    }
    static public AtomType get(String label, String name, boolean create_if_not_there)
    {
        AtomType at = atoms.get(label);
        if (at!=null || !create_if_not_there)
        {
            return at;
        }

        at = new AtomType(label,name);
        atoms.put(label, at);

        return at;
    }
    static 
    {
        String[][]  standards = {{"H", "Hydrogen"},{"C", "Carbon"},{"N", "Nitrogen"},{"O", "Oxygen"},{"Na", "Sodium"},
            {"P", "Phosphorous"}, {"Ca", "Calcium"}, {"S", "Sulfur"}, {"Cl", "Chlorine"},{"K", "Protassium"},
            {"Zn", "Zinc"}, {"Br", "Bromine"},{"Se", "Selenium"},{"Fe", "Iron"},{"Ni", "Nickel"},
            {"Cu", "Copper"},{"Mn", "Manganese"},{"Mg", "Magnesium"},{"Cd", "Cadmium"}, {"Cr", "Chromium"},
            {"Pd", "Palladium"}
                };

       for(String[] i : standards)
       {
           get(i[0], i[1], true);
       }
            
    }

    public static double getMass(String formula)
    {
        StringBuffer label = new StringBuffer();
        StringBuffer count = new StringBuffer();

        boolean last_label = true;

        Vector<Pair<String, Integer>> infos = new Vector<>();

        for (int i=0; i<formula.length(); i++)
        {
            char ch = formula.charAt(i);

            if (Character.isDigit(ch))
            {
                count.append(ch);
                last_label = false;
                continue;
            }
            
            if (last_label && Character.isLowerCase(ch))
            {
                label.append(ch);
                continue;
            }

            int c = (count.length() == 0) ? 1 : Integer.parseInt(count.toString());
        
            if (label.length() > 0)
            {
                infos.add(Pair.create(label.toString(), c));
            }
            

            label.setLength(0);
            count.setLength(0);
            label.append(ch);
            
            last_label = true;

        }

        int c = (count.length() == 0) ? 1 : Integer.parseInt(count.toString());
        infos.add(Pair.create(label.toString(), c));
        
        System.out.println(infos);

        double mass = 0.0;
        for (Pair<String,Integer> p : infos)
        {
            AtomType at = get(p.getFirst());
            if (at == null)
            {
                throw new FRuntimeException("unknow atom type: %s in formula: %s", p.getFirst(), formula);
            }
            if (at.mono_mass == null)
            {
                throw new FRuntimeException("no mass configured yet for atom atom type: %s in formula: %s", p.getFirst(), formula);
            }
            mass += (at.mono_mass * p.getSecond());
        }

        return mass;
    }

    public static void main (String args[]) 
    {
        /*
        try
        {
            extsource.UnimodReader.init(ConfigFileGetter.getConfigFile("MS/unimod.xml"));
        }
        catch(ConfigFileGetter.ConfigFileNotFound cnfe)
        {
            throw new FRuntimeException("error at init: %s", cnfe, cnfe.getMessage());
        }

        System.out.println(getMass(args[0]));
        */
    }
}
